(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Efficient_LPO
imports RPO
begin


fun anyNstriE :: "('b \<Rightarrow> 'a  \<Rightarrow> ((bool \<times> bool) \<times> 'b)) \<Rightarrow> 'b \<Rightarrow> 'a list \<Rightarrow> (bool \<times> 'b)"
where
  "anyNstriE f m [] = (False,m)" |
  "anyNstriE f m (s # ss) = (let ((stri,nstri),m') = f m s in 
     if nstri then (nstri,m') else anyNstriE f m' ss)"
 
lemma anyNstriE [fundef_cong] : 
  assumes "\<And> s m. s \<in> set ss \<Longrightarrow> f m s = g m s"
  shows "anyNstriE f m ss = anyNstriE g m ss"
using assms proof (induct ss arbitrary: m, simp)
  case (Cons s ss)
  show ?case 
  proof (cases "f m s")
    case (Pair res m')
    with Cons have rec: "anyNstriE f m' ss = anyNstriE g m' ss" by auto
    from Pair rec show ?thesis 
      by (simp add: Cons(2)[of s m, simplified], cases res, auto)    
  qed
qed

fun allStriE :: "('b \<Rightarrow> 'a  \<Rightarrow> ((bool \<times> bool) \<times> 'b)) \<Rightarrow> 'b \<Rightarrow> 'a list \<Rightarrow> (bool \<times> 'b)"
where
  "allStriE f m [] = (True,m)" |
  "allStriE f m (t # ts) = (let ((stri,nstri),m') = f m t in
     if stri then allStriE f m' ts else (False,m'))"

lemma allStriE [fundef_cong] : 
  assumes "\<And> t m. t \<in> set ts \<Longrightarrow> f m t = g m t"
  shows "allStriE f m ts = allStriE g m ts"
using assms proof (induct ts arbitrary: m, simp)
  case (Cons t ts)
  show ?case 
  proof (cases "f m t")
    case (Pair res m')
    with Cons have rec: "allStriE f m' ts = allStriE g m' ts" by auto
    from Pair rec show ?thesis 
      by (simp add: Cons(2)[of t m, simplified], cases res, auto)    
  qed
qed

fun lex_extE :: "('b \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> ((bool \<times> bool) \<times> 'b)) \<Rightarrow> 'b \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> ((bool \<times> bool) \<times> 'b)"
where
  "lex_extE f m [] [] = ((False,True),m)" |
  "lex_extE f m (_#_) [] = ((True,True),m)" |
  "lex_extE f m [] (_#_) = ((False,False),m)" |
  "lex_extE f m (a#as) (b#bs) = (let ((stri,nstri),m') = f m a b in
     if stri then ((True,True),m')
     else if nstri then lex_extE f m' as bs 
     else ((False,False),m'))"

lemma lex_extE_cong[fundef_cong] : fixes f g ss ts m
  assumes "\<And> s t m. \<lbrakk>s \<in> set ss; t \<in> set ts\<rbrakk> \<Longrightarrow> f m s t = g m s t" 
  shows "lex_extE f m ss ts = lex_extE g m ss ts" 
using assms 
proof (induct ss arbitrary: ts m)
  case Nil
  thus ?case by (cases ts, auto)
next
  case (Cons s ss) note oCons = this
  show ?case 
  proof (cases ts)
    case Nil
    with Cons show ?thesis by auto
  next
    case (Cons t ts')
    with oCons have id: "g m s t = f m s t" by simp
    show ?thesis
    proof (cases "f m s t")
      case (Pair res m')
      from oCons Cons have id2: "lex_extE f m' ss ts' = lex_extE g m' ss ts'" by auto
      show ?thesis
        by (simp add: Pair Cons id, cases res, auto simp: id2)
    qed
  qed
qed

abbreviation (input) m_lookup where "m_lookup \<equiv> \<lambda> (x,y). x"
abbreviation (input) m_store where "m_store \<equiv> \<lambda> (x,y). y"

function lpoE_m :: "
     ('m  \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> (bool \<times> bool) option)
  \<times> ('m \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> (bool \<times> bool) \<Rightarrow> 'm)
      \<Rightarrow> 'm \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> ((bool \<times> bool) \<times> 'm)"
where
  "lpoE_m m_ops m pr s t = (case (m_lookup m_ops) m s t of
       Some v \<Rightarrow> (v,m)
     | None \<Rightarrow> let (v,new_m) = (case (s,t) of
         (Var x, Var y) \<Rightarrow> ((False, x=y),m)
       | (Var x, Fun g ts) \<Rightarrow> ((False, ts=[] \<and> pr g 0 = 0),m)
       | (Fun f ss, Var y) \<Rightarrow> let (con,m') = anyNstriE (\<lambda> lm s. lpoE_m m_ops lm pr s t) m ss in ((con,con), m')
       | (Fun f ss, Fun g ts) \<Rightarrow> (let (ssNstri,m') = anyNstriE (\<lambda> lm s. lpoE_m m_ops lm pr s t) m ss in
           if ssNstri then ((True,True), m')
           else (let prl = pr f (length ss); prr = pr g (length ts); (sAllStri,m'') = (allStriE (\<lambda>lm. lpoE_m m_ops lm pr s) m' ts) in
             if prl \<ge> prr \<and> sAllStri then
               if prl > prr then ((True,True), m'')
               else lex_extE (\<lambda>lm. lpoE_m m_ops lm pr) m'' ss ts
             else ((False,False), m'')
           )
         )
       ) in (v, (m_store m_ops) new_m s t v)
   )"
by pat_completeness auto

termination lpoE_m
proof (relation "measures [\<lambda> (_,_,_,s,t). size s + size t]", (auto simp: size_simps)[4])
  fix m_ops m "pr" s t x y f list fa lista xa xb ya xc xd xe xf yb sa ta ma
  assume xy: "(x,y) = (s,t)" and x: "x = Fun f list" and y: "y = Fun fa lista" and sa: "sa \<in> set list" and ta: "ta \<in> set lista" 
  from sa have sa1: "size sa \<le> list_size size list" by (induct list, auto)
  from ta have ta1: "size ta \<le> list_size size lista" by (induct lista, auto)
  from x sa1 y ta1 xy[symmetric] show "((m_ops,ma,pr,sa,ta), m_ops,m, pr, s, t) \<in> measures [\<lambda>(_,_,_, s, t). size s + size t]" by auto
qed


locale memory =
  fixes m_lookup :: "'m  \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> (bool \<times> bool) option" 
  fixes m_store  :: "'m \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> (bool \<times> bool) \<Rightarrow> 'm"
  fixes m_valid  :: "(('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> (bool \<times> bool)) \<Rightarrow> 'm \<Rightarrow> bool" 
  fixes m_empty :: "'m \<Rightarrow> bool"
  assumes store_sound:"m_valid f m \<and> f s t = v \<longrightarrow> m_valid f (m_store m s t v)"
  assumes empty_is_valid:"m_empty m \<longrightarrow> m_valid f m"
  assumes lookup_value_sound:"m_valid f m \<and> m_lookup m s t = Some v \<longrightarrow> f s t = v"
begin

abbreviation m_ops where "m_ops \<equiv> (m_lookup,m_store)"
abbreviation lpoE where "lpoE \<equiv> lpoE_m m_ops"

definition m_lpo_valid where "m_lpo_valid pr m = m_valid (lpo_unbounded pr) m"

lemma empty_m_is_lpo_valid : "m_empty m \<longrightarrow> m_lpo_valid pr m" unfolding m_lpo_valid_def using empty_is_valid by blast

lemma anyNstriE_sound:
  assumes "\<And>s m. s \<in> set ss \<and> m_valid f m \<Longrightarrow> fst (fE m s) = f s t \<and> m_valid f (snd (fE m s))"
  and "m_valid f m"
  shows "m_valid f (snd (anyNstriE fE m ss)) \<and> fst (anyNstriE fE m ss) = list_ex (\<lambda>s. snd (f s t)) ss"
using assms proof (induct ss arbitrary: m)
  case Nil
  hence "m_valid f m" by simp
  moreover have "anyNstriE fE m [] = (False,m)" by simp
  moreover have "list_ex (\<lambda>s. snd (f s t)) [] = False" by simp
  ultimately show ?case by simp
next
  case (Cons x xs)
  thus ?case proof (cases "fE m x")
    case (Pair v m') note fE_val=this
    hence m':"snd (fE m x) = m'" by simp
    have "x \<in> set (x # xs)" by simp
    from this Cons have x_vals_sound:"fst (fE m x) = f x t \<and> m_valid f (snd (fE m x))" by blast
    from this m' have m'_valid:"m_valid f m'" by (simp only: assms)
    show ?thesis proof (cases v) 
      case (Pair stri nstri) 
      from x_vals_sound have "fst (fE m x) = f x t" by simp
      from this fE_val Pair have f_val:"f x t = (stri,nstri)" by simp
      show ?thesis proof (cases nstri)
        case True
        from True Pair fE_val have "snd (anyNstriE fE m (x # xs) ) = m'" by simp
        from m'_valid this have valid_memory:"m_valid f (snd (anyNstriE fE m (x # xs)))" by simp
        from True Pair fE_val have anyNstriE_val:"fst (anyNstriE fE m (x # xs)) = True" by simp
        from True f_val have list_ex_val:"list_ex (\<lambda>s. snd (f s t)) (x # xs) = True" by simp
        from list_ex_val anyNstriE_val valid_memory show ?thesis by simp
      next
        case False
        from False Pair fE_val have anyNstriE_val:"anyNstriE fE m (x # xs) = anyNstriE fE m' xs" by simp
        from Cons(2) have "\<And>s m. s \<in> set xs \<and> m_valid f m \<Longrightarrow> fst (fE m s) = f s t \<and> m_valid f (snd (fE m s ))" by simp 
        from this Cons(1)[of m'] m'_valid 
        have rec:"m_valid f (snd (anyNstriE fE m' xs)) \<and> fst (anyNstriE fE m' xs) = list_ex (\<lambda>s. snd (f s t)) xs" by blast
        from f_val False have list_ex_val:"list_ex (\<lambda>s. snd (f s t)) (x # xs) = list_ex (\<lambda>s. snd (f s t)) xs" by simp
        from rec anyNstriE_val list_ex_val show ?thesis by simp 
      qed
    qed
  qed
qed

lemma allStriE_sound:
  assumes "\<And>t m. t \<in> set ts \<and> m_valid f m \<Longrightarrow> fst (fE m t) = f s t \<and> m_valid f (snd (fE m t))"
  and "m_valid f m"
  shows "m_valid f (snd (allStriE fE m ts)) \<and> fst (allStriE fE m ts) = list_all (\<lambda>t. fst (f s t)) ts"
using assms proof (induct ts arbitrary: m)
  case Nil
  hence "m_valid f m" by simp
  moreover have "allStriE fE m [] = (True,m)" by simp
  moreover have "list_all (\<lambda>t. fst (f s t)) [] = True" by simp
  ultimately show ?case by simp
next
  case (Cons x xs)
  thus ?case proof (cases "fE m x")
    case (Pair v m') note fE_val=this
    hence m':"snd (fE m x) = m'" by simp
    have "x \<in> set (x # xs)" by simp
    from this Cons have x_vals_sound:"fst (fE m x) = f s x \<and> m_valid f (snd (fE m x))" by blast
    from this m' have m'_valid:"m_valid f m'" by (simp only: assms)
    show ?thesis proof (cases v) 
      case (Pair stri nstri) 
      from x_vals_sound have "fst (fE m x) = f s x" by simp
      from this fE_val Pair have f_val:"f s x = (stri,nstri)" by simp
      show ?thesis proof (cases stri)
        case False
        from False Pair fE_val have "snd (allStriE fE m (x # xs) ) = m'" by simp
        from m'_valid this have valid_memory:"m_valid f (snd (allStriE fE m (x # xs)))" by simp
        from False Pair fE_val have allStriE_val:"fst (allStriE fE m (x # xs)) = False" by simp
        from this False f_val have list_all_val:"list_all (\<lambda>t. fst (f s t)) (x # xs) = False" by simp
        from list_all_val allStriE_val valid_memory show ?thesis by simp
      next
        case True
        from True Pair fE_val have allStriE_val:"allStriE fE m (x # xs) = allStriE fE m' xs" by simp
        from Cons(2) have "\<And>t m. t \<in> set xs \<and> m_valid f m \<Longrightarrow> fst (fE m t) = f s t \<and> m_valid f (snd (fE m t))" by simp 
        from this Cons(1)[of m'] m'_valid 
        have rec:"m_valid f (snd (allStriE fE m' xs)) \<and> fst (allStriE fE m' xs) = list_all (\<lambda>t. fst (f s t)) xs" by blast
        from f_val True have list_all_val:"list_all (\<lambda>t. fst (f s t)) (x # xs) = list_all (\<lambda>t. fst (f s t)) xs" by simp
        from rec allStriE_val list_all_val show ?thesis by simp 
      qed
    qed
  qed
qed

declare lex_ext_unbounded.simps[simp]

lemma lex_extE_sound:
  assumes "\<And>t s m.  s \<in> set ss \<and> t \<in> set ts \<and> m_valid f m \<Longrightarrow> fst (fE m s t) = f s t \<and> m_valid f (snd (fE m s t))"
  and "m_valid f m"
  shows "m_valid f (snd (lex_extE fE m ss ts)) \<and> fst (lex_extE fE m ss ts) = lex_ext_unbounded f ss ts"
using assms proof (induct ss ts arbitrary: m rule: List.list_induct2')
  case (1 m)
  hence "m_valid f m" by simp
  moreover have "lex_extE fE m [] [] = ((False,True),m)" by simp
  moreover have "lex_ext_unbounded f [] [] = (False, True)" by simp
  ultimately show ?case by simp
next
  case (2 a as m)
  hence "m_valid f m" by simp
  moreover have "lex_extE fE m (a # as) [] = ((True,True),m)" by simp
  moreover have "lex_ext_unbounded f (a # as) [] = (True, True)" by simp
  ultimately show ?case by simp
next
  case (3 b bs m)
  hence "m_valid f m" by simp
  moreover have "lex_extE fE m [] (b # bs)  = ((False,False),m)" by simp
  moreover have "lex_ext_unbounded f [] (b # bs)  = (False, False)" by simp
  ultimately show ?case by simp
next
  case (4 a as b bs m)
  show ?case proof (cases "fE m a b")
    case (Pair v m') note fE_val=this
    from 4(2)[of a b m] 4(3) have fE_sound:"fst (fE m a b) = f a b \<and> m_valid f (snd (fE m a b))" by (simp only: assms, auto)
    from this fE_val have m'_valid: "m_valid f m'" by simp
    show ?thesis proof (cases v)
      case (Pair stri nstri)
      from fE_sound Pair fE_val have f_val:"f a b = (stri,nstri)" by simp
      show ?thesis proof (cases stri)
        case True
        from fE_val Pair True have lex_extE_val:"lex_extE fE m (a # as) (b # bs) = ((True,True),m')" by simp
        from f_val True have lex_ext_unbounded_val:"lex_ext_unbounded f (a # as) (b # bs) = (True,True)" by simp
        from m'_valid lex_extE_val lex_ext_unbounded_val show ?thesis by simp
      next
        case False note striFalse = this
        show ?thesis proof (cases nstri)
          case True
          from 4(2) have "\<And>s t m. s \<in> set as \<and> t \<in> set bs \<and> m_valid f m \<Longrightarrow> fst (fE m s t) = f s t \<and> m_valid f (snd (fE m s t))" by simp 
          from this 4(1)[of m'] m'_valid have rec:
            "m_valid f (snd (lex_extE fE m' as bs)) \<and> fst (lex_extE fE m' as bs) = lex_ext_unbounded f as bs" by (simp only: assms, blast)
          from striFalse True Pair fE_val have lex_extE_val:"lex_extE fE m (a # as) (b # bs) = lex_extE fE m' as bs" by simp
          from striFalse True f_val have lex_ext_unbounded_val:"lex_ext_unbounded f (a # as) (b # bs) = lex_ext_unbounded f as bs" by simp
          from lex_extE_val lex_ext_unbounded_val rec show ?thesis by simp
        next
          case False
          from striFalse False Pair fE_val have lex_extE_val:"lex_extE fE m (a # as) (b # bs) = ((False,False),m')" by simp
          from striFalse False f_val have lex_ext_unbounded_val:"lex_ext_unbounded f (a # as) (b # bs) = (False,False)" by simp
          from lex_extE_val lex_ext_unbounded_val m'_valid show ?thesis by simp
        qed
      qed
    qed
  qed
qed

declare lex_ext_unbounded.simps[simp del]

lemma lpoE_memoized_sound :
  assumes "m_lpo_valid pr m"
  and "m_lookup m s t = Some v"
  shows "lpo_unbounded pr s t = v"
proof (cases v)
  case (Pair stri nstri)
  from this assms lookup_value_sound show ?thesis unfolding m_lpo_valid_def by blast
qed

lemma store_lpo_sound :
  assumes "m_lpo_valid pr m"
  and "lpo_unbounded pr s t = v"
  shows "m_lpo_valid pr (m_store m s t v)"
proof -
  from assms store_sound show ?thesis unfolding m_lpo_valid_def by blast 
qed

lemma lpoE_sound :
assumes "m_lpo_valid pr m"
shows "fst(lpoE m pr s t) = lpo_unbounded pr s t \<and> m_lpo_valid pr (snd (lpoE m pr s t))"
using assms proof (induct s arbitrary: m t rule: term_induct)
  case (Var x) note oVar=this
  show ?case proof (cases "m_lookup m (Var x) t")
    case (Some v)
    from Var this lpoE_memoized_sound[of "pr" m "Var x" t v] show ?thesis by simp
  next
    case None
    show ?thesis proof (cases t)
      case (Var y)
      from Var None have lpo_val:"lpo_unbounded pr (Var x) t = (False,x=y)" by simp
      from Var None have lpoE_val:"lpoE m pr (Var x) t = ((False,x=y), m_store m (Var x) t (False,x=y))" by simp
      from lpo_val lpoE_val store_lpo_sound[of "pr" m "Var x" t "(False,x=y)"] oVar show ?thesis by simp
    next
      case (Fun g ts)
      from Fun None have lpo_val: "lpo_unbounded pr (Var x) t = (False, ts=[] \<and> pr g 0 = 0)" by simp
      from Fun None have lpoE_val: "lpoE m pr (Var x) t = ((False,ts=[] \<and> pr g 0 = 0),m_store m (Var x) t (False,ts=[] \<and> pr g 0 = 0))" by simp
      from oVar lpo_val lpoE_val store_lpo_sound[of "pr" m "Var x" t "(False,ts=[] \<and> pr g 0 = 0)"] show ?thesis by simp
    qed
  qed
next
  case (Fun f ss) note oFun = this
  thus ?case proof (induct t arbitrary: m)
    case (Var y)
    show ?case proof (cases "m_lookup m (Fun f ss) (Var y)")
      case (Some v)
      from this Var(2) lpoE_memoized_sound[of "pr" m "Fun f ss" "Var y" v] show ?thesis by simp
    next
      case None
      from Var(1) have "\<And>s m. s \<in> set ss \<and> m_lpo_valid pr m \<Longrightarrow> fst (lpoE m pr s (Var y)) = lpo_unbounded pr s (Var y) 
        \<and> m_lpo_valid pr (snd (lpoE m pr s (Var y)))" by auto 
      from anyNstriE_sound[of ss "lpo_unbounded pr" "(\<lambda>lm s. lpoE lm pr s (Var y))" "(Var y)" m] this Var(2) 
      have anyNstriE:"m_lpo_valid pr (snd (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss)) \<and>
        fst (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss) = list_ex (\<lambda>s. snd (lpo_unbounded pr s (Var y))) ss" by (simp only: m_lpo_valid_def, blast)
      from this have list_ex:"fst (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss) = list_ex (\<lambda>s. snd (lpo_unbounded pr s (Var y))) ss" by simp
      have lpoE_val:
        "fst (lpoE m pr (Fun f ss) (Var y)) = (fst (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss),fst (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss))" 
        by (simp only: lpoE_m.simps[of m_ops m "pr" "(Fun f ss)" "(Var y)"], simp del: lpoE_m.simps add: None, simp add: split_def)
      have lpo_val: 
        "lpo_unbounded pr (Fun f ss) (Var y) = (fst (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss), fst (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss))" 
        by (simp only: "lpo_unbounded.simps", simp only: Let_def, simp only: list_ex)
      from lpoE_val lpo_val have correct_value: "fst (lpoE m pr (Fun f ss) (Var y)) = lpo_unbounded pr (Fun f ss) (Var y)" by simp
      have lpoE_memory: 
        "snd (lpoE m pr (Fun f ss) (Var y)) = 
        m_store (snd (anyNstriE (\<lambda>lm s. lpoE lm pr s (Var y)) m ss)) (Fun f ss) (Var y) (lpo_unbounded pr (Fun f ss) (Var y))" 
        by (simp only: lpoE_m.simps[of m_ops m "pr" "(Fun f ss)" "(Var y)"], 
          simp del: lpo_unbounded.simps lpoE_m.simps add: None, simp del: lpoE_m.simps lpo_unbounded.simps add:split_def lpo_val)
      from this anyNstriE have "m_lpo_valid pr (snd (lpoE m pr (Fun f ss) (Var y)))" by (simp add: store_lpo_sound) 
      from this correct_value show ?thesis by simp
    qed
  next 
    case (Fun g ts)
    show ?case proof (cases "m_lookup m (Fun f ss) (Fun g ts)")
      case (Some v)
      from this Fun(3) lpoE_memoized_sound[of "pr" m "(Fun f ss)" "(Fun g ts)" v] show ?thesis by simp
    next
      case None
      from Fun(2) have "\<And>s m. s \<in> set ss \<and> m_lpo_valid pr m \<Longrightarrow> fst (lpoE m pr s (Fun g ts)) = lpo_unbounded pr s (Fun g ts) 
        \<and> m_lpo_valid pr (snd (lpoE m pr s (Fun g  ts)))" by auto 
      from anyNstriE_sound[of ss "lpo_unbounded pr" "(\<lambda>lm s. lpoE lm pr s (Fun g ts))" "(Fun g ts)" m] this Fun(3) 
      have anyNstriE:"m_lpo_valid pr (snd (anyNstriE (\<lambda>lm s. lpoE lm pr s (Fun g ts)) m ss)) \<and>
        fst (anyNstriE (\<lambda>lm s. lpoE lm pr s (Fun g ts)) m ss) = list_ex (\<lambda>s. snd (lpo_unbounded pr s (Fun g ts))) ss" by (simp only:m_lpo_valid_def, blast)
      show ?thesis proof (cases "anyNstriE (\<lambda>lm s. lpoE lm pr s (Fun g ts)) m ss")
        case (Pair ssNstri m') note anyNstriE_Pair=this
        from this anyNstriE have m'_valid:"m_lpo_valid pr m'" by simp
        show ?thesis proof (cases ssNstri)
          case True
          from None Pair True have lpoE_val: "fst (lpoE m pr (Fun f ss) (Fun g ts)) = (True,True)" by simp
          from this None Pair True  have lpoE_memory: "snd (lpoE m pr (Fun f ss) (Fun g ts)) = m_store m' (Fun f ss) (Fun g ts) (True,True)" by simp
          from Pair anyNstriE True have lpo_val:"lpo_unbounded pr (Fun f ss) (Fun g ts) = (True,True)" by simp
          from lpoE_val lpo_val lpoE_memory m'_valid show ?thesis by (simp add: store_lpo_sound)
        next
          case False note ssNstri_False = this
          from Fun have "\<And>t m. t \<in> set ts \<and> m_lpo_valid pr m \<Longrightarrow> fst (lpoE m pr (Fun f ss) t) = lpo_unbounded pr (Fun f ss) t 
            \<and> m_lpo_valid pr (snd (lpoE m pr (Fun f ss) t))" by auto 
          from allStriE_sound[of ts "lpo_unbounded pr" "(\<lambda>lm. lpoE lm pr (Fun f ss))" "(Fun f ss)" m'] this m'_valid
          have allStriE:"m_lpo_valid pr (snd (allStriE (\<lambda>lm. lpoE lm pr (Fun f ss)) m' ts)) \<and>
            fst (allStriE (\<lambda>lm. lpoE lm pr (Fun f ss)) m' ts) = list_all (\<lambda>t. fst (lpo_unbounded pr (Fun f ss) t)) ts" 
            by (simp only: m_lpo_valid_def, blast)
          show ?thesis proof (cases "allStriE (\<lambda>lm. lpoE lm pr (Fun f ss)) m' ts")
            case (Pair sAllStri m'')
            from this allStriE have m''_valid: "m_lpo_valid pr m''" by simp
            show ?thesis proof (cases "pr f (length ss) \<ge> pr g (length ts) \<and> sAllStri")
              case True note geqTrue=this
              show ?thesis proof (cases "pr f (length ss) > pr g (length ts)")
                case True note geTrue=this
                from None Pair anyNstriE_Pair geqTrue geTrue ssNstri_False have lpoE_val: "fst (lpoE m pr (Fun f ss) (Fun g ts)) = (True,True)" by simp
                from this None Pair anyNstriE_Pair geqTrue geTrue ssNstri_False 
                have lpoE_memory: "snd (lpoE m pr (Fun f ss) (Fun g ts)) = m_store m'' (Fun f ss) (Fun g ts) (True,True)" by simp 
                from Pair allStriE geTrue geqTrue have lpo_val:"lpo_unbounded pr (Fun f ss) (Fun g ts) = (True,True)" by simp
                from lpoE_val lpo_val lpoE_memory m''_valid show ?thesis by (simp add: store_lpo_sound)
              next
                case False note geFalse=this
                from oFun(1) have "\<And>s m t. s \<in> set ss \<and> t \<in> set ts \<and> m_lpo_valid pr m \<Longrightarrow> fst (lpoE m pr s t) = lpo_unbounded pr s t 
                  \<and> m_lpo_valid pr (snd (lpoE m pr s t))" by auto 
                from lex_extE_sound[of ss ts "(lpo_unbounded pr)" "(\<lambda>lm. lpoE lm pr)" m''] this m''_valid
                have lex_extE:"m_lpo_valid pr (snd (lex_extE (\<lambda>lm. lpoE lm pr) m'' ss ts)) \<and>
                  fst (lex_extE (\<lambda>lm. lpoE lm pr) m'' ss ts) = lex_ext_unbounded (lpo_unbounded pr) ss ts" by (simp only: m_lpo_valid_def, blast)
                from None anyNstriE_Pair ssNstri_False Pair geqTrue geFalse have 
                  lpoE_val:"fst (lpoE m pr (Fun f ss) (Fun g ts)) = fst (lex_extE (\<lambda>lm. lpoE lm pr) m'' ss ts)" 
                  by (simp only: lpoE_m.simps[of m_ops m "pr" "Fun f ss" "Fun g ts"], simp del: lpoE_m.simps, simp add: split_def)
                from anyNstriE ssNstri_False anyNstriE_Pair have anyFalse:"list_ex (\<lambda>s. snd (lpo_unbounded pr s (Fun g ts))) ss = False" by simp
                from allStriE Pair geqTrue have forallTrue:
                  "pr g (length ts) \<le> pr f (length ss) \<and> list_all (\<lambda>t. fst (lpo_unbounded pr (Fun f ss) t)) ts = True" by simp
                have lpo_val: "lpo_unbounded pr (Fun f ss) (Fun g ts) = fst (lex_extE (\<lambda>lm. lpoE lm pr) m'' ss ts)" 
                  by (simp only: "lpo_unbounded.simps", simp only: anyFalse Let_def forallTrue, simp add: geFalse lex_extE)
                from lpoE_val lpo_val have correct_value: "fst (lpoE m pr (Fun f ss) (Fun g ts)) = lpo_unbounded pr (Fun f ss) (Fun g ts)" by simp
                have lpoE_memory:
                  "snd (lpoE m pr (Fun f ss) (Fun g ts)) = 
                  m_store (snd (lex_extE (\<lambda>lm. lpoE lm pr) m'' ss ts)) (Fun f ss) (Fun g ts) (lpo_unbounded pr (Fun f ss) (Fun g ts))" 
                  by (simp only: lpoE_m.simps[of m_ops m "pr" "(Fun f ss)" "(Fun g ts)"], 
                    simp del: lpo_unbounded.simps lpoE_m.simps add: None ssNstri_False Pair anyNstriE_Pair geqTrue geFalse, 
                    simp del: lpoE_m.simps lpo_unbounded.simps add:split_def lpo_val)
                from this lex_extE have "m_lpo_valid pr (snd (lpoE m pr (Fun f ss) (Fun g ts)))" by (simp add: store_lpo_sound) 
                from this correct_value show ?thesis by simp
              qed
            next
              case False note geqFalse = this
              from None Pair anyNstriE_Pair geqFalse ssNstri_False have lpoE_val: "fst (lpoE m pr (Fun f ss) (Fun g ts)) = (False,False)" by auto 
              from this None Pair anyNstriE_Pair geqFalse ssNstri_False 
              have lpoE_memory: "snd (lpoE m pr (Fun f ss) (Fun g ts)) = m_store m'' (Fun f ss) (Fun g ts) (False,False)" by auto
              from Pair anyNstriE_Pair allStriE geqFalse anyNstriE ssNstri_False 
              have lpo_val:"lpo_unbounded pr (Fun f ss) (Fun g ts) = (False,False)" by (simp only: lpo_unbounded.simps, simp only: Let_def, auto)
              from lpoE_val lpo_val lpoE_memory m''_valid show ?thesis by (simp add: store_lpo_sound)
            qed
          qed
        qed
      qed
    qed
  qed
qed

lemma lpoE_sound_empty : assumes empty: "m_empty m"
shows "fst(lpoE m pr s t) = lpo_unbounded pr s t"
  by (rule lpoE_sound[THEN conjunct1], rule mp[OF empty_m_is_lpo_valid], rule empty)
end

type_synonym ('i,'o)memory_1 = "(('i \<times> 'i) \<times> 'o)list"

definition store_1 :: "('i,'o)memory_1 \<Rightarrow> 'i \<Rightarrow> 'i \<Rightarrow> 'o \<Rightarrow> ('i,'o)memory_1"
  where "store_1 m s t v = ((s,t),v) # m"

definition lookup_1 :: "('i,'o)memory_1 \<Rightarrow> 'i \<Rightarrow> 'i \<Rightarrow> 'o option"
  where "lookup_1 m s t = lookup (s,t) m"

definition empty_1 :: "('i,'o)memory_1 \<Rightarrow> bool"
  where "empty_1 m = (m = [])"

definition valid_1 :: "('i \<Rightarrow> 'i \<Rightarrow> 'o) \<Rightarrow> ('i,'o)memory_1 \<Rightarrow> bool"
  where "valid_1 f m \<equiv> \<forall>s t v. ((s,t),v) \<in> set m \<longrightarrow> f s t = v"

interpretation memory_1 : memory lookup_1 store_1 valid_1 empty_1 
proof (unfold_locales)
  fix f m s t v
  have "valid_1 f m \<and> f s t = v \<Longrightarrow> valid_1 f (store_1 m s t v)" unfolding valid_1_def store_1_def by auto
  moreover have "valid_1 f m \<and> lookup_1 m s t = Some v \<Longrightarrow> f s t = v" unfolding lookup_1_def valid_1_def proof -
    from lookup_sound[of "(s,t)" m v] show "(\<forall>s t v. ((s, t), v) \<in> set m \<longrightarrow> f s t = v) \<and> lookup (s, t) m = Some v \<Longrightarrow> f s t = v" by auto
  qed
  moreover have "empty_1 m \<Longrightarrow> valid_1 f m" unfolding empty_1_def valid_1_def by auto
  ultimately show "valid_1 f m \<and> f s t = v \<longrightarrow> valid_1 f (store_1 m s t v)"
    and "valid_1 f m \<and> lookup_1 m s t = Some v \<longrightarrow> f s t = v" 
    and "empty_1 m \<longrightarrow> valid_1 f m" by auto
qed

definition lpo_E_1
  where "lpo_E_1 pr s t = fst(memory_1.lpoE [] pr s t)"

theorem lpo_E_1_to_lpo_unbounded : "lpo_E_1 pr s t = lpo_unbounded pr s t"
  unfolding lpo_E_1_def
  by (rule memory_1.lpoE_sound_empty, unfold empty_1_def, simp)


(*
declare lpo_unbounded.simps[code del]
lemma [code]: "lpo_unbounded = lpo_E_1"
  by (intro ext, simp add: lpo_E_1_to_lpo_unbounded)
*)

type_synonym ('i,'o)memory_2 = "('i \<times> ('i\<times>'o)list) list"

primrec store_2 :: "('i,'o)memory_2 \<Rightarrow> 'i \<Rightarrow> 'i \<Rightarrow> 'o \<Rightarrow> ('i,'o)memory_2"
  where
  "store_2 [] s t v = [(s,[(t,v)])]" |
  "store_2 (head # tail) s t v = (
     if fst head = s then (s, (t,v) # (snd head)) # tail 
     else head # store_2 tail s t v )"

primrec lookup_2 :: "('i,'o)memory_2 \<Rightarrow> 'i \<Rightarrow> 'i \<Rightarrow> 'o option"
where
  "lookup_2 [] s t  = None" | 
  "lookup_2 (head # tail) s t = (
     if fst head = s then lookup t (snd head)
     else lookup_2 tail s t)"
    
definition valid_2
  where "valid_2 f m \<equiv> (\<forall>s m'. (s,m') \<in> set m \<longrightarrow> (\<forall>t v. (t,v) \<in> set m' \<longrightarrow> f s t = v))"

definition empty_2
  where "empty_2 m \<equiv> m = []"

lemma store_2_sound:
  assumes "valid_2 f m"
  and "f s t = v"
  shows "valid_2 f (store_2 m s t v)"
using assms unfolding valid_2_def proof (induct m)
  case Nil
  thus ?case by auto
next
  case (Cons head tail)
  from Cons(2) have tailvalid: "\<forall>s m'. (s,m') \<in> set tail \<longrightarrow> (\<forall>t v. (t, v) \<in> set m' \<longrightarrow> f s t = v)" by auto
  from spec[OF spec[OF Cons(2), of "fst head"], of "snd head"] have headvalid: "\<forall>t v. (t,v) \<in> set (snd head) \<longrightarrow> f (fst head) t = v" by auto
  show ?case proof (cases "fst head = s")
    case True
    from True have storedef:"store_2 (head # tail) s t v = (s, (t,v) # (snd head)) # tail" by simp
    from this True Cons(3) tailvalid headvalid show ?thesis by auto
  next
    case False
    from False have rec: "store_2 (head # tail) s t v = head # (store_2 tail  s t v)" by simp
    from Cons(1) tailvalid this Cons(3) have "\<forall>sa m'. (sa, m') \<in> set (store_2 tail s t v) \<longrightarrow> (\<forall>t v. (t, v) \<in> set m' \<longrightarrow> f sa t = v)" by blast
    from this rec headvalid show ?thesis by auto
  qed
qed

lemma lookup_2_sound:
  assumes "lookup_2 m s t = Some v"
  shows "\<exists>m'. (s,m') \<in> set m \<and> (t,v) \<in> set m'"
using assms proof (induct m arbitrary: m')
  case Nil
  thus ?case by simp
next
  case (Cons head tail)
  have head_in:"(fst head, snd head) \<in> set (head # tail)" by simp
  show ?case proof (cases "fst head = s")
    case True
    from True have "lookup_2 (head # tail) s t = lookup t (snd head)" by simp
    from Cons this have "lookup t (snd head)  = Some v" by simp
    from lookup_sound[OF this] have "(t,v) \<in> set (snd head)"by simp
    from head_in True this show ?thesis by blast
  next
    case False
    hence "lookup_2 (head # tail) s t = lookup_2 tail s t" by simp
    from this Cons show ?thesis by auto
  qed
qed

lemma lookup_value_sound_2:
  assumes "valid_2 f m"
  and "lookup_2 m s t = Some v"
  shows "f s t = v"
using assms proof - 
  from assms lookup_2_sound[of m s t v] have exists: "\<exists>m'. (s, m') \<in> set m \<and> (t, v) \<in> set m'" by simp
  from assms(1) have "\<forall>s t v. (\<exists>m'. (s, m') \<in> set m \<and> (t, v) \<in> set m') \<longrightarrow> f s t = v" 
    unfolding valid_2_def by blast
  from spec[OF spec[OF spec[OF this, of s], of t], of v] have "(\<exists>m'. (s, m') \<in> set m \<and> (t, v) \<in> set m') \<longrightarrow> f s t = v" .
  from this exists show ?thesis by auto
qed

lemma empty_valid_2 : "empty_2 m \<Longrightarrow> valid_2 f m"
  unfolding valid_2_def empty_2_def by simp

interpretation memory_2 : memory lookup_2 store_2 valid_2 empty_2 
proof (unfold_locales)
  fix f m s t v
  have "valid_2 f m \<and> f s t = v \<Longrightarrow> valid_2 f (store_2 m s t v)" using store_2_sound[of f m s t v] by blast
  moreover have "valid_2 f m \<and> lookup_2 m s t = Some v \<Longrightarrow> f s t = v" using lookup_value_sound_2[of f m s t v] by blast
  moreover have "empty_2 m \<Longrightarrow> valid_2 f m" using empty_valid_2 by blast
  ultimately show "valid_2 f m \<and> f s t = v \<longrightarrow> valid_2 f (store_2 m s t v)"
    and "valid_2 f m \<and> lookup_2 m s t = Some v \<longrightarrow> f s t = v" 
    and "empty_2 m \<longrightarrow> valid_2 f m" by auto
qed

definition lpo_E_2
  where "lpo_E_2 pr s t = fst(memory_2.lpoE [] pr s t)"

theorem lpo_E_2_to_lpo_unbounded : "lpo_E_2 pr s t = lpo_unbounded pr s t"
  unfolding lpo_E_2_def
  by (rule memory_2.lpoE_sound_empty, unfold empty_2_def, simp)

declare lpo_unbounded.simps[code del]
lemma [code]: "lpo_unbounded = lpo_E_2"
  by (intro ext, simp add: lpo_E_2_to_lpo_unbounded)

hide_const(open) empty_1 empty_2 lookup_1 lookup_2 store_1 store_2
hide_const valid_1 valid_2 

end
