(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Semantic_Labeling_Framework
imports DP_Framework Relative_Rewriting
begin

datatype ('f,'l)lab = Lab "('f,'l)lab" 'l | FunLab "('f,'l)lab" "('f,'l)lab list" | UnLab 'f | Sharp "('f,'l)lab"

text {* An alternative induction scheme for labs. *}
lemma
  fixes P :: "('f,'l)lab \<Rightarrow> bool"
  assumes "\<And>f. P(UnLab f)"
    and "\<And>f l. P f \<Longrightarrow> P(Lab f l)"
    and "\<And>f. P f \<Longrightarrow> P(Sharp f)"
    and "\<And>f ls. P f \<Longrightarrow> (\<And>l. l \<in> set ls \<Longrightarrow> P l) \<Longrightarrow> P(FunLab f ls)"
  shows lab_induct[case_names UnLab Lab Sharp FunLab,induct type: lab]: "P f"
    and "\<And>f. f \<in> set fs \<Longrightarrow> P f"
using assms by (induct f and fs) auto

instantiation lab :: ("show", "show") "show" begin
  fun shows_lab :: "('f::show,'l::show)lab \<Rightarrow> shows"
  where "shows_lab(UnLab f) = shows f"
      | "shows_lab(Lab f l) = (shows_lab f +@+ ''['' +#+ shows l +@+ shows_string '']'')"
      | "shows_lab(Sharp f) = (shows_lab f +@+ shows_string ''#'')"
      | "shows_lab(FunLab f l) = (shows_lab f +@+ ''['' +#+ shows_map shows_lab l +@+ shows_string '']'')"
  definition "shows_prec (d::nat) l = shows_lab l"
  definition "shows_list_lab (fs::('f::show,'l::show)lab list) \<equiv> shows_list_aux shows fs"

  lemma assoc_lab: "shows_prec d (f::('f::show,'l::show)lab) r @ s = shows_prec d f (r @ s)" by (induct f arbitrary: r s) (auto simp: shows_prec_lab_def)
  lemma assoc_lab_elt: "\<forall>(f::('f::show,'l::show)lab)\<in>set fs. shows_prec d f r @ s = shows_prec d f (r @ s)"
    using assoc_lab by auto
  instance by
    (intro_classes,rule assoc_lab,unfold shows_list_lab_def)
    (rule shows_list_aux_assoc[OF assoc_lab_elt])
end

text {* Resolve problems with recurisve dependencies in generated code. *}
lemma [code]:
  "HOL.eq (FunLab f ls) (FunLab g ks) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ls ks"
  by (simp_all add: list_all2_eq [symmetric]) 

(* unlab removes one layer of labelings *)
fun unlab :: "('f,'l)lab \<Rightarrow> ('f,'l)lab"
where "unlab (Lab f l) = f"
 | "unlab (FunLab f l) = f"
 | "unlab x = x"

abbreviation unlab_term :: "(('f,'l)lab,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
where "unlab_term \<equiv> map_term unlab"

abbreviation unlab_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
where "unlab_trs \<equiv> map_trs unlab"

abbreviation unlab_rules :: "(('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules"
where "unlab_rules \<equiv> map (map_rule unlab)"

abbreviation unlab_subst :: "(('f,'l)lab,'v)subst \<Rightarrow> (('f,'l)lab,'v)subst"
where "unlab_subst \<equiv> map_subst unlab"

(* init_lab adds empty labelings to all function symbols *)
fun init_lab :: "'f \<Rightarrow> ('f,'l)lab"
where "init_lab f = UnLab f"

abbreviation init_lab_term :: "('f,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
where "init_lab_term \<equiv> map_term init_lab"

abbreviation init_lab_subst :: "('f,'v)subst \<Rightarrow> (('f,'l)lab,'v)subst"
where "init_lab_subst \<equiv> map_subst init_lab"

abbreviation init_lab_rule :: "('f,'v)rule \<Rightarrow> (('f,'l)lab,'v)rule"
where "init_lab_rule \<equiv> map_rule init_lab"

abbreviation init_lab_trs :: "('f,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
where "init_lab_trs \<equiv> map_trs init_lab"

abbreviation init_lab_rules :: "('f,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules"
where "init_lab_rules \<equiv> map init_lab_rule"


(* remove_lab switches back to an unlabeled signature and thereby
   removes all labelings *)
fun remove_lab :: "('f,'l)lab \<Rightarrow> 'f" 
where "remove_lab (Lab f _) = remove_lab f"
 | "remove_lab (UnLab f) = f"
 | "remove_lab (Sharp f) = remove_lab f"
 | "remove_lab (FunLab f _) = remove_lab f"

lemma remove_lab_sound[simp]: "remove_lab (init_lab f) = f" by auto

abbreviation remove_lab_term :: "(('f,'l)lab,'v)term \<Rightarrow> ('f,'v)term"
where "remove_lab_term \<equiv> map_term remove_lab"

abbreviation remove_lab_subst :: "(('f,'l)lab,'v)subst \<Rightarrow> ('f,'v)subst"
where "remove_lab_subst \<equiv> map_subst remove_lab"


lemma remove_lab_term_sound: "remove_lab_term (init_lab_term t) = t"
proof (induct t, simp)
  case (Fun f ss)
  thus ?case by (induct ss, auto)
qed


lemma unlab_init_lab_term_sound: "unlab_term (init_lab_term t) = init_lab_term t"
proof (induct t, simp)
  case (Fun f ss)
  thus ?case by (induct ss, auto)
qed

lemma unlab_init_lab_trs_sound: "unlab_trs (init_lab_trs R) = (init_lab_trs R)"
  by (force simp: unlab_init_lab_term_sound map_trs.simps)


abbreviation remove_lab_rule :: "(('f,'l)lab,'v)rule \<Rightarrow> ('f,'v)rule"
where "remove_lab_rule \<equiv> map_rule remove_lab"

lemma remove_lab_rule_sound: "remove_lab_rule (init_lab_rule lr) = lr"
  by (auto simp: remove_lab_term_sound)

abbreviation remove_lab_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> ('f,'v)trs"
where "remove_lab_trs \<equiv> map_trs remove_lab"

lemma remove_lab_trs_sound: fixes R :: "('f,'v)trs"
  shows "remove_lab_trs (init_lab_trs R) = R"
proof 
  show "R \<subseteq> remove_lab_trs (init_lab_trs R :: (('f,'l)lab,'v)trs)" (is "_ \<subseteq> remove_lab_trs ?IR")
  proof 
    fix lr 
    assume "lr \<in> R"
    hence "remove_lab_rule ((init_lab_rule lr) :: (('f,'l)lab,'v)rule) \<in> (remove_lab_trs ?IR)" by (force simp: map_trs.simps)
    thus "lr \<in> remove_lab_trs ?IR" by (simp only: remove_lab_rule_sound)
  qed
next
  show "remove_lab_trs (init_lab_trs R) \<subseteq> R" by (auto simp: remove_lab_term_sound map_trs.simps)
qed


definition decr_of_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
  where "decr_of_trs R \<equiv> { (l,r) | l r . (l,r) \<in> R \<and> l \<noteq> r \<and> unlab_term l = unlab_term r}"

definition unlab_of_trs :: "(('f,'l)lab,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs" 
  where "unlab_of_trs R \<equiv> map_trs unlab (R - decr_of_trs R)"

definition decrSN :: "(('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> bool"
  where "decrSN R m \<equiv> (\<forall> n < m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))) \<longrightarrow> SN (rstep R)"

lemma unlab_of_trs_mono: "r \<subseteq> s \<Longrightarrow> unlab_of_trs r \<subseteq> unlab_of_trs s"
  unfolding unlab_of_trs_def decr_of_trs_def by (auto simp: map_trs.simps)

lemma unlab_of_trs_union[simp]: "unlab_of_trs (r \<union> s) = unlab_of_trs r \<union> unlab_of_trs s"
  unfolding unlab_of_trs_def decr_of_trs_def by (auto simp: map_trs.simps)

lemma unlab_of_trs_unlab_trs: "(unlab_of_trs ^^ n) R \<subseteq> (unlab_trs ^^ n) R"
  by (induct n, auto simp: unlab_of_trs_def map_trs.simps)

lemma decr_of_trs_mono: "r \<subseteq> s \<Longrightarrow> decr_of_trs r \<subseteq> decr_of_trs s"
  unfolding decr_of_trs_def by auto

lemma decr_of_trs_union[simp]: "decr_of_trs (r \<union> s) = decr_of_trs r \<union> decr_of_trs s"
  unfolding decr_of_trs_def by auto

lemma decr_of_trs_subset_trs: "decr_of_trs r \<subseteq> r"
  unfolding decr_of_trs_def by auto

lemma unlab_trs_union: "unlab_trs (R \<union> S) = unlab_trs R \<union> unlab_trs S"
  by (auto simp: map_trs.simps)

definition sound_lab_tt :: "((('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> (('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> bool) \<Rightarrow> bool"
  where "sound_lab_tt tt \<equiv> \<forall> R m rR rm. tt R m rR rm \<longrightarrow> (decrSN rR rm \<longrightarrow> decrSN R m)"


definition sound_lab_tt_impl :: "((('f,'l)lab,'v)rules \<Rightarrow> nat \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> nat \<Rightarrow> bool) \<Rightarrow> bool"
  where "sound_lab_tt_impl tt \<equiv> \<forall> R m rR rm. tt R m rR rm \<longrightarrow> (decrSN (set rR) rm \<longrightarrow> decrSN (set R) m)"


lemma SN_unlab:
  fixes R :: "(('f,'l)lab,'v)trs" and t :: "(('f,'l)lab,'v)term"
  assumes SN1: "SN (rstep (decr_of_trs R))"
  and SN2: "SN_elt (rstep (unlab_of_trs R)) (unlab_term t)"
  shows "SN_elt (rstep R) t"
proof (rule ccontr)
  let ?D = "decr_of_trs R"
  let ?ND = "R - ?D"
  have R: "R = ?D \<union> ?ND" using decr_of_trs_subset_trs by auto
  assume "\<not> SN_elt (rstep R) t"
  from this obtain tt where "\<forall> i. (tt i, tt (Suc i)) \<in> rstep R" and t: "tt 0 = t" unfolding SN_elt_def by auto
  hence steps: "\<forall> i. (tt i, tt (Suc i)) \<in> (rstep ?D \<union> rstep ?ND)" by (simp only: rstep_union[symmetric], simp only: R[symmetric])
  show False
  proof (cases "\<exists> i. (\<forall> j. j \<ge> i \<longrightarrow> (tt j, tt (Suc j)) \<in> rstep ?D)")
    case True
    from this obtain i where steps2: "\<forall> j.  j \<ge> i \<longrightarrow> (tt j, tt (Suc j)) \<in> rstep ?D" ..
    have "\<forall> j. ((shift tt i) j, (shift tt i) (Suc j)) \<in> rstep ?D" using steps2 by auto
    with SN1 show False unfolding SN_defs by blast
  next
    case False
    have id: "unlab_trs ?ND = unlab_of_trs R" unfolding unlab_of_trs_def by simp
    have "unlab_trs ?D \<subseteq> Id" unfolding decr_of_trs_def by (auto simp: map_trs.simps)
    with rstep_id rstep_mono have subset: "rstep (unlab_trs ?D) \<subseteq> Id" by blast
    have compat: "Id O rstep (unlab_of_trs R) \<subseteq> rstep (unlab_of_trs R)" by auto
    from steps False have infND: "\<forall> i. (\<exists> j. j \<ge> i \<and> (tt j, tt (Suc j)) \<in> rstep ?ND)" by force
    from steps have "\<forall> i. (unlab_term (tt i), unlab_term (tt (Suc i))) \<in> Id \<union> rstep (unlab_of_trs R)" 
      using rstep_imp_map_rstep subset by (simp only: id[symmetric], blast)
    from non_strict_ending[OF this compat] t SN2 have 
      "\<exists> j. \<forall> i \<ge> j. (unlab_term (tt i), unlab_term (tt (Suc i))) \<in> Id - rstep (unlab_of_trs R)" by force
    from this obtain j where nInf: "\<forall> i \<ge> j. (unlab_term (tt i), unlab_term (tt (Suc i))) \<notin> rstep (unlab_of_trs R)" by auto
    from infND have "\<forall> i. (\<exists> j. j \<ge> i \<and> (unlab_term (tt j), unlab_term (tt (Suc j))) \<in> rstep (unlab_of_trs R))"
      using rstep_imp_map_rstep[of _ _ ?ND] by (simp only: id[symmetric]) blast+
    with nInf show False by blast
  qed
qed
   

lemma unlab_tt_sound: "sound_lab_tt (\<lambda> R m rR rm. rR = unlab_of_trs R \<and> Suc rm = m)"
unfolding sound_lab_tt_def decrSN_def
proof (intro allI, intro impI, erule conjE, simp)
  fix R m rR rm
  assume imp: "(\<forall> n < rm. SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) (unlab_of_trs R))))) \<longrightarrow> SN(rstep (unlab_of_trs R))"
    and  SN: "\<forall> n < m. SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))"
    and m: "Suc rm = m"
    and unlab: "rR = unlab_of_trs R"
  have "\<forall> n < rm. SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) (unlab_of_trs R))))"
  proof (rule allI, rule impI)
    fix n
    assume "n < rm"
    with m have le: "Suc n < m" by auto
    have id: "((unlab_of_trs ^^ n) (unlab_of_trs R)) = (unlab_of_trs ^^ Suc n) R" by (simp add: funpow_swap1)
    from SN le show "SN(rstep (decr_of_trs ((unlab_of_trs ^^ n) (unlab_of_trs R))))" by (simp only: id)
  qed
  with imp have SN2: "SN(rstep (unlab_of_trs R))" by auto
  from m have "0 < m" by auto
  with SN have "SN(rstep (decr_of_trs ((unlab_of_trs ^^ 0) R)))" by blast
  hence SN3: "SN(rstep (decr_of_trs R))" by auto
  from SN2 SN_unlab[OF SN3]
  show "SN(rstep R)" unfolding SN_defs by auto
qed

definition lab_minimal_term :: "(('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> (('f,'l)lab,'v)term \<Rightarrow> bool"
where "lab_minimal_term R m t \<equiv> \<forall> n \<le> m. SN_elt(rstep ((unlab_of_trs ^^ n) R)) ((unlab_term ^^ n) t)"


definition lab_minimal_cond :: "(('f,'l)lab,'v)trs \<Rightarrow> nat \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)subst) \<Rightarrow> bool"
where "lab_minimal_cond R m s t \<sigma> \<equiv> \<forall> i. lab_minimal_term R m (t i \<cdot> \<sigma> i)"

lemmas lab_minimal_defs = lab_minimal_term_def lab_minimal_cond_def 

lemma lab_minimal_cond_imp_minimal_cond[simp]: "lab_minimal_cond R m s t \<sigma> \<Longrightarrow> minimal_cond R s t \<sigma>"
proof -
  assume "lab_minimal_cond R m s t \<sigma>"
  from this[unfolded lab_minimal_defs, THEN spec, THEN spec[of _ 0]]
  show ?thesis unfolding minimal_cond_def
    by auto
qed

lemma funpow_mono:
  fixes f :: "'a set \<Rightarrow> 'a set"
  assumes mono: "\<And> x y. x \<subseteq> y \<Longrightarrow> f x \<subseteq> f y"
  and subset: "x \<subseteq> y"
  shows "(f ^^ n) x \<subseteq> (f ^^ n) y"
  by (induct n, simp, rule subset, auto simp: mono)

lemma funpow_ci_mono:
  fixes f :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
  assumes mono: "\<And> x y. x \<subseteq>ci y \<Longrightarrow> f x \<subseteq>ci f y"
  and subset: "x \<subseteq>ci y"
  shows "(f ^^ n) x \<subseteq>ci (f ^^ n) y"
  by (induct n, simp, rule subset, auto simp: mono)


lemma rstep_decr_of_trs_mono: assumes "rstep R \<subseteq> rstep S"
  shows "rstep (decr_of_trs R) \<subseteq> rstep (decr_of_trs S)"
(* 
   this lemma does not hold, consider R = f_1(a) \<rightarrow> f_0(a) and S = f_1(x) \<rightarrow> f_0(a) 
   then decr_of_trs R = R, whereas decr_of_trs S = {} 

   this is the reason why ci_subset is not defined as \<rightarrow>R \<subseteq> \<rightarrow>S
*)
oops

lemma decr_of_trs_mono_ci: assumes "R \<subseteq>ci S"
  shows "decr_of_trs R \<subseteq>ci decr_of_trs S"
proof (unfold ci_subset_def, intro allI impI)
  fix l r
  assume "(l,r) \<in> decr_of_trs R"
  hence inR: "(l,r) \<in> R" and noteq: "l \<noteq> r" and unlabeq: "unlab_term l = unlab_term r"
    unfolding decr_of_trs_def by auto
  from assms inR ci_subset_def
  obtain l' r' C where inS: "(l',r') \<in> S" and l: "l = C\<langle>l'\<rangle>" and r: "r = C\<langle>r'\<rangle>" by blast
  with noteq have noteq2: "l' \<noteq> r'" by auto
  from l r unlabeq have unlabeq2: "unlab_term l' = unlab_term r'" by auto
  have "(l',r') \<in> decr_of_trs S" unfolding decr_of_trs_def using inS noteq2 unlabeq2 by auto
  from this l r show "\<exists> l' r' C. (l',r') \<in> decr_of_trs S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>" by blast
qed
 
lemma unlab_of_trs_mono_ci: assumes "R \<subseteq>ci S"
  shows "unlab_of_trs R \<subseteq>ci unlab_of_trs S"
proof (unfold ci_subset_def, intro allI impI)
  fix l r
  assume "(l,r) \<in> unlab_of_trs R"  
  from this obtain ll lr where inR: "(ll,lr) \<in> R" and l: "l = unlab_term ll" and r: "r = unlab_term lr" and cond: "(ll = lr \<or> unlab_term ll \<noteq> unlab_term lr)"
    unfolding unlab_of_trs_def decr_of_trs_def by (force simp: map_trs.simps)+
  from assms inR ci_subset_def 
  obtain l' r' C where inS: "(l',r') \<in> S" and ll: "ll = C\<langle>l'\<rangle>" and lr: "lr = C\<langle>r'\<rangle>" by blast
  from cond have "l' = r' \<or> unlab_term l' \<noteq> unlab_term r'" (is "?l \<or> ?r")
  proof
    assume "ll = lr" with l r ll lr show ?thesis by auto
  next
    assume "unlab_term ll \<noteq> unlab_term lr" with l r ll lr show ?thesis by auto
  qed
  with inS have mem: "(unlab_term l', unlab_term r') \<in> unlab_of_trs S" unfolding unlab_of_trs_def decr_of_trs_def by (cases "l' = r'", (force simp: map_trs.simps)+)
  show "\<exists> l' r' C. (l',r') \<in> unlab_of_trs S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    by (rule exI[of _ "unlab_term l'"], 
         rule exI[of _ "unlab_term r'"],
         rule exI[of _ "map_ctxt unlab C"],
      auto simp: l r ll lr mem)
qed

lemma lab_minimal_term_mono: 
  assumes subset: "R \<subseteq>ci R'" and cond: "lab_minimal_term R' m t" 
  shows "lab_minimal_term R m t"
unfolding lab_minimal_term_def
proof (intro allI, intro impI)
  fix n 
  assume "n \<le> m"
  with cond have SN: "SN_elt (rstep ((unlab_of_trs ^^ n) R')) ((unlab_term ^^ n) t)"
    unfolding lab_minimal_term_def by auto
  show "SN_elt (rstep ((unlab_of_trs ^^ n) R)) ((unlab_term ^^ n) t)"
    by (rule SN_elt_subset[OF SN rstep_ci_mono[OF funpow_ci_mono[OF unlab_of_trs_mono_ci subset]]], simp)
qed


lemma lab_minimal_cond_mono: 
  assumes subset: "R \<subseteq>ci R'" and cond: "lab_minimal_cond R' m s t \<sigma>" 
  shows "lab_minimal_cond R m s t \<sigma>"
  using lab_minimal_term_mono[OF subset] cond
unfolding lab_minimal_cond_def
by auto

fun min_lab_ichain ::
 "(('f,'l)lab,'v)dpp \<Rightarrow> nat \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)subst) \<Rightarrow> bool"
where "min_lab_ichain (P,R) m s t \<sigma> = (ichain (P,R) s t \<sigma> \<and> lab_minimal_cond R m s t \<sigma>)"

lemma min_lab_ichainI[intro]: 
  assumes "R \<subseteq>ci R'" and "lab_minimal_cond R' m s t \<sigma>" and "ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)" 
  shows "min_lab_ichain (P,R) m (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  from assms have "lab_minimal_cond R m s t \<sigma>" using lab_minimal_cond_mono by auto
  hence "lab_minimal_cond R m (shift s i) (shift t i) (shift \<sigma> i)" 
    unfolding lab_minimal_cond_def by auto
  with assms show ?thesis by auto
qed


definition finite_lab_dpp :: "(('f,'l)lab,'v)dpp \<Rightarrow> nat \<Rightarrow> bool" where "finite_lab_dpp DPP m = (\<not>(\<exists>s t \<sigma>. min_lab_ichain DPP m s t \<sigma>))"



lemma weak_match_unlab: assumes wm: "weak_match t s" shows "weak_match ((unlab_term ^^ n) t) ((unlab_term ^^ n) s)"
proof (induct n)
  case 0
  thus ?case using wm by simp
next
  case (Suc n)
  show ?case
    by (simp, rule weak_match_map_term, rule Suc)
qed


lemma left_linear_unlab: assumes ll: "left_linear_trs R" shows "left_linear_trs ((unlab_of_trs ^^ n) R)"
proof (induct n, simp add: ll)
  case (Suc n)
  show ?case 
  proof (simp, rule left_linear_mono)
    show "left_linear_trs (map_trs unlab ((unlab_of_trs ^^ n) R))" by (simp only: left_linear_map_trs, rule Suc)
  next
    show "unlab_of_trs ((unlab_of_trs ^^ n) R) \<subseteq> unlab_trs ((unlab_of_trs ^^ n) R)" unfolding unlab_of_trs_def by (auto simp: map_trs.simps)
  qed
qed
  

lemma left_linear_min_lab_ichain_imp_clean_min_lab_ichain_fcc: 
  fixes P R::"(('f,'l)lab,'v)trs" and F::"('f,'l)lab sig"
  defines "c \<equiv> \<lambda>t::(('f,'l)lab, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::(('f,'l)lab, 'v) subst. clean_subst F \<sigma>"
  assumes left_linear: "left_linear_trs R"
  and subset: "funas_below_trs P \<union> funas_trs R \<subseteq> F"
  and non_var: "\<And> s t. (s,t) \<in> P \<Longrightarrow> \<not> is_var s \<and> \<not> is_var t \<and> \<not> defined R (the (root t), num_args t)"
  and chain: "min_lab_ichain (P,R) m s t \<sigma>"
  shows "min_lab_ichain (P,R) m s t (\<lambda> i. cs (\<sigma> i))"
proof -
  from chain have ichain: "ichain (P,R) s t \<sigma>"
    and min: "\<And> i n. n \<le> m \<Longrightarrow> SN_elt (rstep ((unlab_of_trs ^^ n) R)) ((unlab_term^^n) (t i \<cdot> \<sigma> i))" by (auto simp: lab_minimal_defs)
  from left_var_imp_not_SN[where R = R and t = "t 0 \<cdot> \<sigma> 0"] min[of 0 0]
  have nvar: "\<forall> lr \<in> R. \<forall> x. fst lr \<noteq> Var x" by force 
  from ichain_imp_clean_ichain_fcc[OF subset non_var nvar ichain]
  have ichain: "ichain (P,R) s t (\<lambda> i. cs (\<sigma> i))" unfolding cs_def .
  { 
    fix i t
    have wm: "weak_match (t \<cdot> \<sigma> i) (t \<cdot> cs (\<sigma> i))"
    proof (induct t, simp add: cs_def, simp add: clean_weak_match)
      case (Fun f ss)
      from this[unfolded set_conv_nth] show ?case by force 
    qed
  }  note wm = this
  have "lab_minimal_cond R m s t (\<lambda> i. cs (\<sigma> i))" unfolding lab_minimal_defs
  proof (intro allI impI)
    fix n i
    assume n: "n \<le> m"
    from ichain have "(s i, t i) \<in> P" by auto
    show "SN_elt (rstep ((unlab_of_trs ^^ n) R)) ((unlab_term^^n) (t i \<cdot> cs (\<sigma> i)))"
    proof (rule weak_match_SN[OF weak_match_unlab[OF wm] _ min[OF n]])
      show "left_linear_trs ((unlab_of_trs ^^ n) R)"
        by (rule left_linear_unlab[OF left_linear])
    qed
  qed
  with ichain show ?thesis by simp
qed

lemma decr_of_init_is_empty: 
  "decr_of_trs (init_lab_trs R :: (('f,'l)lab,'v)trs) = {}" (is "decr_of_trs ?IR = {}")
proof -
  {
    fix il ir :: "(('f,'l)lab,'v)term"
    assume nid: "il \<noteq> ir"
    assume id: "unlab_term il = unlab_term ir"
    assume mem: "(il,ir) \<in> ?IR"
    let ?i = "init_lab_term :: ('f,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
    from mem obtain l r where "(l,r) \<in> R" and il: "il = ?i l" and ir: "ir = ?i r" by (force simp: map_trs.simps)
    from id nid have "?i l \<noteq> ?i r \<and> unlab_term (?i l) = unlab_term (?i r)" by (auto simp: il ir) 
    hence False by (simp only: unlab_init_lab_term_sound, auto)
  }
  thus ?thesis unfolding decr_of_trs_def  by auto
qed

lemma unlab_of_init_is_init: 
  "(unlab_of_trs ^^ n) (init_lab_trs R :: (('f,'l)lab,'v)trs) = (init_lab_trs R :: (('f,'l)lab,'v)trs)" (is "_ = ?IR")
proof -
  {
    fix n
    have "(unlab_of_trs ^^ n) ?IR = ?IR"
    proof (induct n, simp)
      case (Suc n)
      have "(unlab_of_trs ^^ Suc n) ?IR = unlab_of_trs ((unlab_of_trs ^^ n) ?IR)" by simp
      also have "\<dots> = unlab_of_trs ?IR" by (simp only: Suc)
      also have "\<dots> = unlab_trs ?IR" unfolding unlab_of_trs_def using decr_of_init_is_empty by (auto simp: map_trs.simps)
      also have "\<dots> = ?IR" by (rule unlab_init_lab_trs_sound)
      finally show ?case .
    qed
  } 
  thus ?thesis by auto
qed

lemma unlab_term_init_is_init: 
  "(unlab_term ^^ n) (init_lab_term t :: (('f,'l)lab,'v)term) = (init_lab_term t :: (('f,'l)lab,'v)term)" (is "_ = ?it")
proof (induct n, simp)
  case (Suc n)
  with unlab_init_lab_term_sound[of t]
  show ?case by (simp del: init_lab.simps add: funpow_swap1)
qed

lemma init_lab_sound: assumes finite: "finite_lab_dpp (init_lab_trs P :: (('f,'l)lab,'v)trs, init_lab_trs R) m" (is "finite_lab_dpp (?lP, ?lR) _") 
  shows "finite_dpp (P,R)"
proof (rule ccontr)
  assume "\<not> finite_dpp (P,R)"
  from this obtain s t \<sigma> where chain: "min_ichain (P,R) s t \<sigma>" unfolding finite_dpp_def by auto
  hence ichain: "ichain (P,R) s t \<sigma>" by auto
  let ?it = "init_lab_term :: ('f,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
  let ?ir = "init_lab_trs :: ('f,'v)trs \<Rightarrow> (('f,'l)lab,'v)trs"
  let ?ls = "\<lambda> i. ?it (s i)"
  let ?lt = "\<lambda> i. ?it (t i)"
  let ?lsig = "\<lambda> i. init_lab_subst (\<sigma> i)" 
  let ?lP = "?ir P"
  let ?lR = "?ir R"
  have ichain2: "ichain (?lP,?lR) ?ls ?lt ?lsig" 
    by (rule ichain_imp_map_ichain, rule ichain)  
  have "\<forall> n \<le> m. \<forall> i. SN_elt (rstep ((unlab_of_trs ^^ n) ?lR)) ((unlab_term ^^ n) (?lt i \<cdot> ?lsig i))" 
  proof (intro allI, intro impI, intro allI)
    fix n i
    from chain have SN: "SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
    have id_term: "(unlab_term ^^ n) (?lt i \<cdot> ?lsig i) = ?lt i \<cdot> ?lsig i" by  (simp only: map_subst_distrib[symmetric], rule unlab_term_init_is_init)
    from SN have foo: "SN_elt (rstep (remove_lab_trs ?lR)) (remove_lab_term (?lt i \<cdot> ?lsig i))" 
      by (simp only: remove_lab_trs_sound remove_lab_term_sound map_subst_distrib[symmetric])
    have "SN_elt (rstep ?lR) (?lt i \<cdot> ?lsig i)" by (rule SN_map_imp_SN[OF foo])
    thus "SN_elt (rstep ((unlab_of_trs ^^ n) ?lR)) ((unlab_term ^^ n) (?lt i \<cdot> ?lsig i))" by (simp only: id_term unlab_of_init_is_init) 
  qed
  with ichain2 have "min_lab_ichain (?lP,?lR) m ?ls ?lt ?lsig" by (auto simp: lab_minimal_defs)
  with finite show False unfolding finite_lab_dpp_def by blast
qed

lemma init_lab_sound_trs:
  assumes SN: "SN (rstep (init_lab_trs R))" (is "SN (rstep ?IR)")
  shows "SN (rstep R)"
  using assms SN_map_imp_SN[of init_lab R] unfolding SN_def by auto


lemma rtrancl_rstep_unlab_of_is_unlab: "(rstep (unlab_of_trs R))^* = (rstep (unlab_trs R))^*" (is "?l = ?r")
proof -
  have "?l = (rstep (unlab_of_trs R) \<union> Id)^*" using rtrancl_reflcl by auto
  also have "\<dots> = (rstep (unlab_of_trs R \<union> Id))^*" by (simp only: rstep_id rstep_union)
  also have "\<dots> = (rstep (unlab_trs R \<union> Id))^*" 
  proof -
    have "unlab_of_trs R \<union> Id = unlab_trs R \<union> Id"
      by (rule set_eqI, auto simp: split_paired_all unlab_of_trs_def decr_of_trs_def map_trs.simps)
    thus ?thesis by auto
  qed
  also have "\<dots> = (rstep (unlab_trs R) \<union> Id)^*" by (simp only: rstep_id rstep_union)
  also have "\<dots> = (rstep (unlab_trs R))^*" by (rule rtrancl_reflcl)
  finally show ?thesis .
qed

lemma unlab_proc_sound: fixes R :: "(('f,'l)lab,'v)trs"
  assumes finite: "finite_lab_dpp (unlab_trs P, unlab_of_trs R) m"
  shows "finite_lab_dpp (P,R) (Suc m)"
proof (rule ccontr)
  assume "\<not> finite_lab_dpp (P,R) (Suc m)"
  from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) (Suc m) s t \<sigma>" unfolding finite_lab_dpp_def by auto
  hence ichain: "ichain (P,R) s t \<sigma>" by auto
  let ?ls = "\<lambda> i. unlab_term (s i)"
  let ?lt = "\<lambda> i. unlab_term (t i)"
  let ?lsig = "\<lambda> i. unlab_subst (\<sigma> i)" 
  let ?lP = "unlab_trs P"
  let ?lR = "unlab_of_trs R"
  from ichain have "ichain (?lP, unlab_trs R) ?ls ?lt ?lsig" using ichain_imp_map_ichain by blast
  hence ichain2: "ichain (?lP, ?lR) ?ls ?lt ?lsig" using rtrancl_rstep_unlab_of_is_unlab by auto
  have cond: "lab_minimal_cond ?lR m ?ls ?lt ?lsig" 
    unfolding lab_minimal_defs
  proof (intro allI impI)
    fix n i
    assume le: "n \<le> m"
    with chain have "SN_elt (rstep ((unlab_of_trs ^^ (Suc n)) R)) ((unlab_term ^^ (Suc n)) (t i \<cdot> \<sigma> i))"    
      by (simp only: min_lab_ichain.simps lab_minimal_defs)
    thus "SN_elt (rstep ((unlab_of_trs ^^ n) ?lR)) ((unlab_term ^^ n) (?lt i \<cdot> ?lsig i))"
      by (simp add: funpow_swap1)
  qed
  from ichain2 cond have "min_lab_ichain (?lP, ?lR) m ?ls ?lt ?lsig" by (auto simp: lab_minimal_cond_def)
  with finite show False unfolding finite_lab_dpp_def by blast
qed


definition sound_lab_proc 
  where "sound_lab_proc proc \<equiv> \<forall> P R m rP rR rm. (proc P R m rP rR rm \<longrightarrow> (finite_lab_dpp (rP, rR) rm \<longrightarrow> finite_lab_dpp (P, R) m))"

definition sound_lab_proc_impl
  where "sound_lab_proc_impl proc \<equiv> \<forall> P R m rP rR rm. (proc P R m (computeRuleMap R) rP rR rm \<longrightarrow> (finite_lab_dpp (set rP, set rR) rm \<longrightarrow> finite_lab_dpp (set P, set R) m))"

definition sound_lab_only_pairs_proc_impl
  where "sound_lab_only_pairs_proc_impl proc \<equiv> \<forall> P R m rP rm. (proc P R m (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP rm \<longrightarrow> (finite_lab_dpp (set rP, set R) rm \<longrightarrow> finite_lab_dpp (set P, set R) m))"


lemma decr_SN_to_SN_init: 
  assumes "decrSN (init_lab_trs R :: (('f,'l)lab,'v)trs) m" (is "decrSN ?IR _")
  shows "SN (rstep (init_lab_trs R :: (('f,'l)lab,'v)trs))"
proof -
  have SN2: "\<forall>n. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) ?IR)))" 
  proof
    fix n
    have "decr_of_trs ((unlab_of_trs ^^ n) ?IR) = decr_of_trs ?IR" by (simp only: unlab_of_init_is_init)
    also have "\<dots> = {}" by (rule decr_of_init_is_empty)
    finally show "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) ?IR)))" by auto
  qed
  with assms show ?thesis unfolding decrSN_def by auto
qed

lemma decr_SN_to_SN:
  assumes decr: "decrSN (init_lab_trs R) m"
  shows "SN (rstep R)"
  by (rule init_lab_sound_trs, rule decr_SN_to_SN_init, rule decr)

lemma decr_SN_to_SN_II:
  assumes decr: "decrSN R 0"
  shows "SN (rstep R)"
using assms unfolding decrSN_def by auto

lemma SN_to_decr_SN:
  assumes "SN (rstep R)"
  shows "decrSN R m"
using assms unfolding decrSN_def by auto


lemma weak_ci_tt_to_sound_lab_tt: assumes ci_proc: "weak_ci_tt tt"
  shows "sound_lab_tt (\<lambda> R m rR rm. tt R rR \<and> m = rm)"
unfolding sound_lab_tt_def decrSN_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume m: "m = rm" and cond: "tt R rR" and step: "(\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))) \<longrightarrow> SN (rstep rR)" and SN1: "\<forall> n < m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))"
  from cond ci_proc have subset: "rR \<subseteq>ci R" and SN: "SN (rstep rR) \<longrightarrow> SN (rstep R)" unfolding weak_ci_tt_def by auto
  have "\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))"
  proof (intro allI impI)
    fix n
    assume "n < rm"
    with m have SN2: "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) R)))" using SN1 by auto
    show "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) rR)))"
      by (rule SN_subset, rule SN2, rule rstep_ci_mono, rule decr_of_trs_mono_ci, rule funpow_ci_mono, rule unlab_of_trs_mono_ci, auto simp: subset)
  qed
  with step have "SN (rstep rR)" by auto
  with SN show "SN (rstep R)" ..
qed




lemma weak_ci_tt_impl_to_sound_lab_tt: assumes ci_proc: "weak_ci_tt_impl tt"
  shows "sound_lab_tt_impl (\<lambda> R m rR rm. tt R rR \<and> m = rm)"
unfolding sound_lab_tt_impl_def decrSN_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume m: "m = rm" and cond: "tt R rR" and step: "(\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set rR))))) \<longrightarrow> SN (rstep (set rR))" and SN1: "\<forall> n < m. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set R))))"
  from cond ci_proc have subset: "set rR \<subseteq>ci set R" and SN: "SN (rstep (set rR)) \<longrightarrow> SN (rstep (set R))" unfolding weak_ci_tt_impl_def by auto
  have "\<forall> n < rm. SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set rR))))"
  proof (intro allI impI)
    fix n
    assume "n < rm"
    with m have SN2: "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set R))))" using SN1 by auto
    show "SN (rstep (decr_of_trs ((unlab_of_trs ^^ n) (set rR))))"
      by (rule SN_subset, rule SN2, rule rstep_ci_mono, rule decr_of_trs_mono_ci, rule funpow_ci_mono, rule unlab_of_trs_mono_ci, auto simp: subset)
  qed
  with step have "SN (rstep (set rR))" by auto
  with SN show "SN (rstep (set R))" ..
qed

lemma ci_proc_to_sound_lab_proc: assumes ci_proc: "ci_proc proc"
  shows "sound_lab_proc (\<lambda> P R m rP rR rm. m = rm \<and> proc P R rP rR)"
unfolding sound_lab_proc_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume m: "m = rm" and cond: "proc P R rP rR" and finite: "finite_lab_dpp (rP, rR) rm"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (P,R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "rR \<subseteq>ci R \<and> ichain (rP,rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_proc_def by force
    have "min_lab_ichain (rP,rR) rm (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = R and R = rR] chain ml m by auto
    hence "\<not> finite_lab_dpp (rP,rR) rm" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed

(* the following two techniques are only possible due to the explicit index m,
   they allow to use non-chain-identifying processors even in the labeled semantics,
   although afterwards one cannot perform unlabeling any more
*)
lemma sound_proc_to_sound_lab_proc: assumes sound_proc: "sound_proc proc"
  shows "sound_lab_proc (\<lambda> P R m rP rR rm. rm = 0 \<and> proc P R rP rR)"
unfolding sound_lab_proc_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume rm: "rm = 0" and cond: "proc P R rP rR" and finite: "finite_lab_dpp (rP, rR) rm"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (P,R) s t \<sigma>" by auto
    hence "\<not> finite_dpp (P,R)" unfolding finite_dpp_def by auto
    with sound_proc cond have "\<not> finite_dpp (rP, rR)" unfolding sound_proc_def by auto
    from this obtain s t \<sigma> where "min_ichain (rP, rR) s t \<sigma>" unfolding finite_dpp_def by auto
    hence "min_lab_ichain (rP, rR) 0 s t \<sigma>" by (auto simp: minimal_cond_def lab_minimal_defs)
    hence "\<not> finite_lab_dpp (rP,rR) 0" unfolding finite_lab_dpp_def by blast
    with finite rm show False by simp
  qed
qed

lemma sound_tt_impl_to_sound_lab_tt_impl: assumes sound: "sound_tt_impl tt"
  shows "sound_lab_tt_impl (\<lambda> R m rR rm. rm = 0 \<and> tt R rR)"
unfolding sound_lab_tt_impl_def
proof (intro allI impI, simp)
  fix R m rR rm
  assume tt: "rm = 0 \<and> tt R rR" and decr: "decrSN (set rR) 0"
  from decr have "SN (rstep (set rR))" unfolding decrSN_def by force
  with tt sound have "SN (rstep (set R))" unfolding sound_tt_impl_def by auto
  thus "decrSN (set R) m" by (rule SN_to_decr_SN)
qed
 

lemma sound_proc_to_sound_lab_proc_impl: assumes sound_proc: "sound_proc_impl proc"
  shows "sound_lab_proc_impl (\<lambda> P R m rmap rP rR rm. rm = 0 \<and> proc P R rmap rP rR)"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume rm: "rm = 0" and cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_lab_dpp (set rP, set rR) rm"
  show "finite_lab_dpp (set P, set R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    hence "\<not> finite_dpp (set P,set R)" unfolding finite_dpp_def by auto
    with sound_proc cond have "\<not> finite_dpp (set rP, set rR)" unfolding sound_proc_impl_def by auto
    from this obtain s t \<sigma> where "min_ichain (set rP, set rR) s t \<sigma>" unfolding finite_dpp_def by auto
    hence "min_lab_ichain (set rP, set rR) 0 s t \<sigma>" by (auto simp: minimal_cond_def lab_minimal_defs)
    hence "\<not> finite_lab_dpp (set rP,set rR) 0" unfolding finite_lab_dpp_def by blast
    with finite rm show False by simp
  qed
qed 
 
lemma ci_proc_to_sound_lab_proc_impl[simp]: assumes ci_proc: "ci_proc_impl proc"
  shows "sound_lab_proc_impl (\<lambda> P R m rmap rP rR rm. m = rm \<and> proc P R rmap rP rR)"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m rP rR rm
  assume m: "m = rm" and cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_lab_dpp (set rP, set rR) rm"
  show "finite_lab_dpp (set P, set R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "set rR \<subseteq>ci set R \<and> ichain (set rP,set rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_proc_impl_def by force
    have "min_lab_ichain (set rP,set rR) rm (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = "set R" and R = "set rR"] chain ml m by auto
    hence "\<not> finite_lab_dpp (set rP,set rR) rm" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed 

lemma ci_proc_to_sound_lab_only_pairs_proc_impl[simp]: assumes ci_proc: "ci_only_pairs_proc_impl proc"
  shows "sound_lab_only_pairs_proc_impl (\<lambda> P R m rmap1 rmap2 rP rm. m = rm \<and> proc P R rmap1 rmap2 rP)"
unfolding sound_lab_only_pairs_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m rP rm
  assume m: "m = rm" and cond: "proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP" and finite: "finite_lab_dpp (set rP, set R) rm"
  show "finite_lab_dpp (set P, set R) m"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R) m"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding ci_only_pairs_proc_impl_def by blast
    have "min_lab_ichain (set rP,set R) rm (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = "set R" and R = "set R"] chain ml m ci_subset_refl[of "set R"] by auto
    hence "\<not> finite_lab_dpp (set rP,set R) rm" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed 


lemma apply_ci_only_pairs_lab_proc: assumes ci: "ci_only_pairs_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set R) m"
  and proc: "proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from ci have "sound_lab_only_pairs_proc_impl (\<lambda> P R m rmap1 rmap2 rP rm. m = rm \<and> proc P R rmap1 rmap2 rP)" by simp
  thus ?thesis unfolding sound_lab_only_pairs_proc_impl_def using fin proc by blast
qed

lemma apply_ci_lab_proc: assumes ci: "ci_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR) m"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from ci have "sound_lab_proc_impl (\<lambda> P R m rmap rP rR rm. m = rm \<and> proc P R rmap rP rR)" by simp
  thus ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed

lemma apply_sound_lab_proc: assumes sound: "sound_lab_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR) rm"
  and proc: "proc P R m (computeRuleMap R) rP rR rm"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from sound show ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed

lemma apply_sound_proc: assumes sound: "sound_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR) 0"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_lab_dpp (set P, set R) m"
proof -
  from sound_proc_to_sound_lab_proc_impl[OF sound]
  show ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed

lemma apply_sound_lab_tt: assumes sound: "sound_lab_tt_impl tt"
  and fin: "decrSN (set rR) rm"
  and proc: "tt R m rR rm"
  shows "decrSN (set R) m"
proof -
  from sound
  show ?thesis unfolding sound_lab_tt_impl_def using fin proc sound by blast
qed 

lemma apply_weak_ci_lab_tt: assumes ci: "weak_ci_tt_impl tt"
  and fin: "decrSN (set rR) m"
  and proc: "tt R rR"
  shows "decrSN (set R) m"
  using apply_sound_lab_tt[OF weak_ci_tt_impl_to_sound_lab_tt[OF ci] fin] proc by simp

lemma apply_ci_lab_tt: assumes ci: "ci_tt_impl tt"
  and fin: "decrSN (set rR) m"
  and proc: "tt R rR"
  shows "decrSN (set R) m"
  using apply_weak_ci_lab_tt[OF ci_tt_to_weak_ci_tt_impl[OF ci] fin] proc  by simp


definition unlab_of_trs_list :: "(('f,'l)lab,'v)rules \<Rightarrow> (('f,'l)lab,'v)rules"
 where "unlab_of_trs_list R \<equiv> remdups (map_trs_list unlab (filter (\<lambda> (l,r). l = r \<or> unlab_term l \<noteq> unlab_term r) R))"

lemma unlab_of_trs_list_sound: "set (unlab_of_trs_list R) = unlab_of_trs (set R)"
unfolding unlab_of_trs_list_def unlab_of_trs_def decr_of_trs_def 
by (auto simp: map_trs.simps)

definition
  unlab_trs_tt :: "(('f::show,'l::show)lab,'v::show)rules \<Rightarrow> (('f,'l)lab,'v)rules \<Rightarrow> shows check"
where
  "unlab_trs_tt R rR \<equiv> let u =
  unlab_of_trs_list R in (check_subseteq rR u \<guillemotright> check_subseteq u rR)
    <+? (\<lambda>s. shows_rules rR +@+ shows_nl +@+ '' does not correspond to unlabeled TRS ''
      +#+ shows_nl +@+ shows_rules u +@+ shows_nl)"

lemma unlab_trs_tt_sound: "sound_lab_tt_impl (\<lambda> R m rR rm. m = Suc rm \<and> isOK(unlab_trs_tt R rR))"
unfolding sound_lab_tt_impl_def
proof (intro allI impI, erule conjE)
  fix R m rR rm
  assume "isOK(unlab_trs_tt R rR)" and SN: "decrSN (set rR) rm" and m: "m = Suc rm"
  hence "set rR = set (unlab_of_trs_list R)" unfolding unlab_trs_tt_def by (auto simp: Let_def)
  hence "set rR = unlab_of_trs (set R)" using unlab_of_trs_list_sound by auto
  with SN have "decrSN(unlab_of_trs (set R)) rm" by auto
  with unlab_tt_sound m show "decrSN (set R) m" unfolding sound_lab_tt_def by auto
qed


definition
  check_unlab_proc ::
    "(('f::show,'l::show)lab,'v::show)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     (('f,'l)lab,'v)rules \<Rightarrow> 
     shows check"
where
  "check_unlab_proc P R ulP ulR \<equiv> do {
     let ULR = unlab_of_trs_list R;
     check_subseteq (map_trs_list unlab P) ulP
         <+? (\<lambda>rl. ''unlabeled pair '' +#+ shows_rule rl +@+ shows '' is missing'' +@+ shows_nl);
     check_subseteq ULR ulR
         <+? (\<lambda>rl. ''unlabeled rule '' +#+ shows_rule rl +@+ shows '' is missing'' +@+ shows_nl);
     check_subseteq ulR ULR
        <+? (\<lambda>rl. ''rule '' +#+ shows_rule rl +@+ shows '' is not an unlabeled rule'' +@+ shows_nl)
   }"

lemma check_unlab_proc_sound: 
  "sound_lab_proc_impl (\<lambda> P R m rmap ulP ulR rm. m = Suc rm \<and> isOK(check_unlab_proc P R ulP ulR))"
unfolding sound_lab_proc_impl_def
proof (intro allI impI, erule conjE)
  fix P R m ulP ulR rm
  assume m: "m = Suc rm" 
    and  check: "isOK(check_unlab_proc P R ulP ulR)"
    and  finite: "finite_lab_dpp (set ulP, set ulR) rm"
  hence id: "unlab_of_trs (set R) = set ulR" and subs: "unlab_trs (set P) \<subseteq> set ulP" 
    unfolding check_unlab_proc_def by (auto simp: Let_def unlab_of_trs_list_sound)
  {
    fix s t \<sigma>
    assume "min_lab_ichain (unlab_trs (set P), set ulR) rm s t \<sigma>"
    with subs have "min_lab_ichain (set ulP, set ulR) rm s t \<sigma>" by auto
  }
  hence fin: "finite_lab_dpp (unlab_trs (set P), set ulR) rm" using finite m unfolding finite_lab_dpp_def by blast
  show "finite_lab_dpp (set P, set R) m"
    by (simp only: m, rule unlab_proc_sound, auto simp only: id, rule fin)
qed



lemmas rel_SN_unlab = rel_SN_map_imp_SN[where fg = unlab]
lemmas rel_SN_init_lab = rel_SN_map_imp_SN[where fg = init_lab]

definition
  unlab_reltrs_tt :: "(('f::show,'l::show)lab,'v::show)reltrsL \<Rightarrow> (('f,'l)lab,'v)reltrsL \<Rightarrow> shows check"
where
  "unlab_reltrs_tt RS rRS \<equiv>  (check_subseteq (map_trs_list unlab (fst RS)) (fst rRS) \<guillemotright> check_all (\<lambda> r. (fst r = snd r) \<or> r \<in> set (snd rRS)) (map_trs_list unlab (snd RS)))
    <+? (\<lambda>s. shows_string ''cannot find unlabeled rule '' +@+ shows_nl +@+ shows_rule s)"

lemma unlab_reltrs_tt: "sound_rel_tt_impl (\<lambda> RS rRS. isOK(unlab_reltrs_tt RS rRS))"
unfolding sound_rel_tt_impl_def
proof (intro allI impI)
  fix R S rR rS
  assume check: "isOK (unlab_reltrs_tt (R,S) (rR, rS))" and SN: "rel_SN_alt (rstep (set rR), rstep (set rS))"
  from check have R: "unlab_trs (set R) \<subseteq> set rR" and S: "\<And> r. r \<in> unlab_trs (set S) \<Longrightarrow> fst r = snd r \<or> r \<in> set rS" unfolding unlab_reltrs_tt_def by auto
  hence S: "unlab_trs (set S) \<subseteq> set rS \<union> Id" by force
  from SN have "rel_SN_alt (rstep (set rR), rstep (set rS \<union> Id))" 
    unfolding rel_SN_conv[symmetric] rel_SN_def rel_rstep.simps[symmetric] rel_rstep_Id .
  from rel_SN_alt_mono[OF rstep_mono[OF R] rstep_mono[OF S] this]
  show "rel_SN_alt (rstep (set R), rstep (set S))" by (rule rel_SN_unlab)
qed

(* towards equivalence of finiteness and relative termination *)


  
theorem finite_dpp_iff_rel_SN: "finite_lab_dpp (P,R) m = rel_SN ((CS P \<inter> {(s,t) | s t. lab_minimal_term R m t}), rstep R)" (is "_ = rel_SN (?R,?S)")
proof
  assume SN: "rel_SN (?R,?S)"
  show "finite_lab_dpp (P,R) m"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain s t \<sigma> where chain: "min_lab_ichain (P,R) m s t \<sigma>" unfolding finite_lab_dpp_def by auto
    obtain f where f: "\<And> i. f i = s i \<cdot> \<sigma> i" by auto
    from chain have "\<And> i. (s i \<cdot> \<sigma> i, t i \<cdot> \<sigma> i) \<in> ?R" and "\<And> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?S^*"  
      by (auto simp: lab_minimal_term_def lab_minimal_cond_def)
    hence "\<And> i. (f i,f (Suc i)) \<in> ?S^* O ?R O ?S^*" unfolding f by auto
    with SN[unfolded rel_SN_def rel_step.simps] 
    show False unfolding SN_defs by blast
  qed
next
  assume finite: "finite_lab_dpp (P,R) m"
  show "rel_SN (?R,?S)" 
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain t where steps: "\<And> i. (t i, t (Suc i)) \<in> ?S^* O ?R O ?S^*"
      unfolding rel_SN_def rel_step.simps SN_defs by auto
    let ?prop = "\<lambda> s u i.  (t i, s) \<in> ?S^* \<and> (s,u) \<in> ?R \<and> (u, t (Suc i)) \<in> ?S^*"
    have "\<forall> i. \<exists> s u. ?prop s u i"
    proof
      fix i
      from steps[of i] show "\<exists> s u. ?prop s u i" by auto
    qed
    from choice[OF this] obtain s where "\<forall> i. \<exists> u. ?prop (s i) u i" ..
    from choice[OF this] obtain u where p: "\<And> i. ?prop (s i) (u i) i" by auto
    let ?prop2 = "\<lambda> l r \<sigma> i. (l, r) \<in> P \<and> lab_minimal_term R m (r \<cdot> \<sigma>) \<and> r \<cdot> \<sigma> = u i \<and> l \<cdot> \<sigma> = s i"
    have "\<forall> i. \<exists> l r \<sigma>. ?prop2 l r \<sigma> i"
    proof
      fix i
      from p[of i] have "(s i, u i) \<in> CS P" by auto
      with subst_closed_decompose[OF this] obtain l r \<sigma> where  "s i = l \<cdot> \<sigma> \<and> u i = r \<cdot> \<sigma> \<and> (l,r) \<in> P" by auto
      with p[of i] have "?prop2 l r \<sigma> i" by auto
      thus "\<exists> l r \<sigma>. ?prop2 l r \<sigma> i" by blast
    qed
    from choice[OF this] obtain l where "\<forall> i. \<exists> r \<sigma> . ?prop2 (l i) r \<sigma> i" ..
    from choice[OF this] obtain r where "\<forall> i. \<exists> \<sigma> . ?prop2 (l i) (r i) \<sigma> i" ..
    from choice[OF this] obtain \<sigma> where p2: "\<And> i. ?prop2 (l i) (r i) (\<sigma> i) i" by auto
    {
      fix i
      from p2[of i] p[of i] p[of "Suc i"] p2[of "Suc i"]
      have "(r i \<cdot> \<sigma> i, l (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?S^* " by auto
    }
    with p2 have "min_lab_ichain (P,R) m l r \<sigma>" unfolding lab_minimal_term_def
      by (auto simp: lab_minimal_defs p2)
    with finite show False unfolding finite_lab_dpp_def by blast
  qed
qed


lemma rstep_imp_unlab_of_steps: assumes "(s,t) \<in> rstep R"
  shows "((unlab_term ^^ n) s, (unlab_term ^^ n) t) \<in> (rstep ((unlab_of_trs ^^ n) R))^*"
  using assms
proof (induct n arbitrary: s t)
  case 0 
  thus ?case by auto
next
  case (Suc n)
  let ?un = "\<lambda> n t. (unlab_term ^^ n) t"
  let ?u = "?un n"
  let ?us = "?un (Suc n)"
  obtain ss tt where ss: "ss = ?u s" and tt: "tt = ?u t" by auto
  have s: "?us s = unlab_term ss" and t: "?us t = unlab_term tt" unfolding ss tt by auto
  from Suc have "(ss,tt) \<in> (rstep ((unlab_of_trs ^^ n) R))^*" unfolding ss tt by auto
  from rsteps_imp_map_rsteps[OF this]
  have "(unlab_term ss, unlab_term tt) \<in> (rstep (unlab_trs ((unlab_of_trs ^^ n) R)))^*" .
  hence "(unlab_term ss, unlab_term tt) \<in> (rstep (unlab_of_trs ((unlab_of_trs ^^ n) R)))^*" unfolding  rtrancl_rstep_unlab_of_is_unlab . 
  thus ?case unfolding s t by simp
qed

lemma rsteps_imp_unlab_of_steps: assumes "(s,t) \<in> (rstep R)^*"
  shows "((unlab_term ^^ n) s, (unlab_term ^^ n) t) \<in> (rstep ((unlab_of_trs ^^ n) R))^*"
using assms
proof (induct)
  case (step y z)
  from rstep_imp_unlab_of_steps[OF step(2), of n] step(3) show ?case by auto
qed simp
  

lemma lab_minimal_term_rsteps: assumes min: "lab_minimal_term R m t" and 
  steps: "(t,s) \<in> (rstep R)^*" 
  shows "lab_minimal_term R m s"
proof(unfold lab_minimal_term_def, intro allI impI)
  fix n
  assume "n \<le> m"
  with min have "SN_elt (rstep ((unlab_of_trs ^^ n) R)) ((unlab_term ^^ n) t)"
    unfolding lab_minimal_term_def by auto
  from steps_preserve_SN_elt[OF rsteps_imp_unlab_of_steps[OF steps, of n] this]
  show "SN_elt (rstep ((unlab_of_trs ^^ n) R)) ((unlab_term ^^ n) s)" .
qed



end
