object Natural {

  def apply(numeral: BigInt): Natural = new Natural(numeral max 0)
  def apply(numeral: Int): Natural = Natural(BigInt(numeral))
  def apply(numeral: String): Natural = Natural(BigInt(numeral))

}

class Natural private(private val value: BigInt) {

  override def hashCode(): Int = this.value.hashCode()

  override def equals(that: Any): Boolean = that match {
    case that: Natural => this equals that
    case _ => false
  }

  override def toString(): String = this.value.toString

  def equals(that: Natural): Boolean = this.value == that.value

  def as_BigInt: BigInt = this.value
  def as_Int: Int = if (this.value >= scala.Int.MinValue && this.value <= scala.Int.MaxValue)
      this.value.intValue
    else error("Int value out of range: " + this.value.toString)

  def +(that: Natural): Natural = new Natural(this.value + that.value)
  def -(that: Natural): Natural = Natural(this.value - that.value)
  def *(that: Natural): Natural = new Natural(this.value * that.value)

  def /%(that: Natural): (Natural, Natural) = if (that.value == 0) (new Natural(0), this)
    else {
      val (k, l) = this.value /% that.value
      (new Natural(k), new Natural(l))
    }

  def <=(that: Natural): Boolean = this.value <= that.value

  def <(that: Natural): Boolean = this.value < that.value

}


object Ceta {

def id[A]: A => A = (x: A) => x

trait equal[A] {
  val `Ceta.equal`: (A, A) => Boolean
}
def equal[A](a: A, b: A)(implicit A: equal[A]): Boolean = A.`Ceta.equal`(a, b)

def eq[A: equal](a: A, b: A): Boolean = equal[A](a, b)

abstract sealed class nat
final case object Zero_nat extends nat
final case class Suc(a: nat) extends nat

trait ord[A] {
  val `Ceta.less_eq`: (A, A) => Boolean
  val `Ceta.less`: (A, A) => Boolean
}
def less_eq[A](a: A, b: A)(implicit A: ord[A]): Boolean = A.`Ceta.less_eq`(a, b)
def less[A](a: A, b: A)(implicit A: ord[A]): Boolean = A.`Ceta.less`(a, b)

trait preorder[A] extends ord[A] {
}

trait order[A] extends preorder[A] {
}

trait linorder[A] extends order[A] {
}

abstract sealed class color
final case object R extends color
final case object B extends color

abstract sealed class rbta[+A, +B]
final case object Emptya extends rbta[Nothing, Nothing]
final case class
  Branch[A, B](a: color, b: rbta[A, B], c: A, d: B, e: rbta[A, B])
  extends rbta[A, B]

abstract sealed class rbt[+A, +B]
final case class Rbt[A, B](a: rbta[A, B]) extends rbt[A, B]

abstract sealed class rat
final case class Frct(a: (BigInt, BigInt)) extends rat

def nat_aux(i: BigInt, n: nat): nat =
  (if (i <= BigInt(0)) n else nat_aux(i - BigInt(1), Suc(n)))

def nat(i: BigInt): nat = nat_aux(i, Zero_nat)

def hd[A](x0: List[A]): A = x0 match {
  case x :: xs => x
}

def tl[A](x0: List[A]): List[A] = x0 match {
  case Nil => Nil
  case x :: xs => xs
}

def collect[A](p: A => Boolean): A => Boolean = p

def is_none[A](x0: Option[A]): Boolean = x0 match {
  case Some(x) => false
  case None => true
}

def dom[A, B](m: A => Option[B]): A => Boolean =
  collect[A]((a: A) => ! (is_none[B](m(a))))

abstract sealed class order_tag
final case object Lex extends order_tag
final case object Mul extends order_tag

abstract sealed class term[+A, +B]
final case class Var[B, A](a: B) extends term[A, B]
final case class Fun[A, B](a: A, b: List[term[A, B]]) extends term[A, B]

def nulla[A](x0: List[A]): Boolean = x0 match {
  case Nil => true
  case x :: xs => false
}

def snd[A, B](x0: (A, B)): B = x0 match {
  case (a, b) => b
}

def list_ex[A](p: A => Boolean, x1: List[A]): Boolean = (p, x1) match {
  case (p, Nil) => false
  case (p, x :: xs) => p(x) || list_ex[A](p, xs)
}

def fst[A, B](x0: (A, B)): A = x0 match {
  case (a, b) => a
}

def plus_nat(x0: nat, n: nat): nat = (x0, n) match {
  case (Suc(m), n) => plus_nat(m, Suc(n))
  case (Zero_nat, n) => n
}

def size_list[A](x0: List[A]): nat = x0 match {
  case Nil => Zero_nat
  case a :: list => plus_nat(size_list[A](list), Suc(Zero_nat))
}

def list_all[A](p: A => Boolean, x1: List[A]): Boolean = (p, x1) match {
  case (p, Nil) => true
  case (p, x :: xs) => p(x) && list_all[A](p, xs)
}

def or2(x0: (Boolean, Boolean), x1: (Boolean, Boolean)): (Boolean, Boolean) =
  (x0, x1) match {
  case ((b1, b2), (b3, b4)) => (b1 || b3, b2 || b4)
}

def lifta(x0: (Boolean, Boolean)): (Boolean, Boolean) = x0 match {
  case (uu, b) => (b, b)
}

def remove1[A: equal](x: A, xa1: List[A]): List[A] = (x, xa1) match {
  case (x, Nil) => Nil
  case (x, y :: xs) => (if (eq[A](x, y)) xs else y :: remove1[A](x, xs))
}

def filter[A](p: A => Boolean, x1: List[A]): List[A] = (p, x1) match {
  case (p, Nil) => Nil
  case (p, x :: xs) => (if (p(x)) x :: filter[A](p, xs) else filter[A](p, xs))
}

def ex_dominating[A: equal](f: A => A => (Boolean, Boolean), y: A, x2: List[A],
                             uw: List[A], ux: List[A]):
      (Boolean, Boolean) =
  (f, y, x2, uw, ux) match {
  case (f, y, Nil, uw, ux) => (false, false)
  case (f, y, b :: bs, xs, ys) =>
    (f(b))(y) match {
      case (true, ns) =>
        or2(lifta(mul_ext_impl[A](f, remove1[A](b, xs),
                                   filter[A]((ya: A) =>
       ! (fst[Boolean, Boolean]((f(b))(ya))),
      ys))),
             ex_dominating[A](f, y, bs, xs, ys))
      case (false, true) =>
        or2(mul_ext_impl[A](f, remove1[A](b, xs), ys),
             ex_dominating[A](f, y, bs, xs, ys))
      case (false, false) => ex_dominating[A](f, y, bs, xs, ys)
    }
}

def mul_ext_impl[A: equal](f: A => A => (Boolean, Boolean), x1: List[A],
                            x2: List[A]):
      (Boolean, Boolean) =
  (f, x1, x2) match {
  case (f, Nil, Nil) => (false, true)
  case (f, Nil, v :: va) => (false, false)
  case (f, v :: va, Nil) => (true, true)
  case (f, v :: va, y :: ys) => ex_dominating[A](f, y, v :: va, v :: va, ys)
}

def equal_lista[A: equal](x0: List[A], x1: List[A]): Boolean = (x0, x1) match {
  case (a :: list, Nil) => false
  case (Nil, a :: list) => false
  case (aa :: lista, a :: list) => eq[A](aa, a) && equal_lista[A](lista, list)
  case (Nil, Nil) => true
}

def equal_terma[A: equal, B: equal](x0: term[A, B], x1: term[A, B]): Boolean =
  (x0, x1) match {
  case (Fun(f, list), Var(v)) => false
  case (Var(v), Fun(f, list)) => false
  case (Fun(fa, lista), Fun(f, list)) =>
    eq[A](fa, f) && equal_lista[term[A, B]](lista, list)
  case (Var(va), Var(v)) => eq[B](va, v)
}

implicit def equal_term[A: equal, B: equal]: equal[term[A, B]] = new
  equal[term[A, B]] {
  val `Ceta.equal` = (a: term[A, B], b: term[A, B]) => equal_terma[A, B](a, b)
}

def mpo_impl[A: equal,
              B: equal](uu: (A => nat => A => nat => (Boolean, Boolean),
                              A => Boolean),
                         x1: term[A, B], x2: term[A, B]):
      (Boolean, Boolean) =
  (uu, x1, x2) match {
  case (uu, Var(x), Var(y)) => (false, eq[B](x, y))
  case (pr, Var(x), Fun(g, ts)) =>
    (false,
      nulla[term[A, B]](ts) &&
        (snd[A => nat => A => nat => (Boolean, Boolean),
              A => Boolean](pr)).apply(g))
  case (pr, Fun(f, ss), Var(y)) =>
    {
      val con: Boolean =
        list_ex[term[A, B]]((s: term[A, B]) =>
                              snd[Boolean,
                                   Boolean](mpo_impl[A,
              B](pr, s, Var[B, A](y))),
                             ss);
      (con, con)
    }
  case (pr, Fun(f, ss), Fun(g, ts)) =>
    (if (list_ex[term[A, B]]((s: term[A, B]) =>
                               snd[Boolean,
                                    Boolean](mpo_impl[A,
               B](pr, s, Fun[A, B](g, ts))),
                              ss))
      (true, true)
      else {
             val (prs, prns): (Boolean, Boolean) =
               (fst[A => nat => A => nat => (Boolean, Boolean),
                     A => Boolean](pr)).apply(f).apply(size_list[term[A,
                               B]](ss)).apply(g).apply(size_list[term[A,
                               B]](ts));
             (if (prns &&
                    list_all[term[A, B]]((t: term[A, B]) =>
   fst[Boolean, Boolean](mpo_impl[A, B](pr, Fun[A, B](f, ss), t)),
  ts))
               (if (prs) (true, true)
                 else mul_ext_impl[term[A,
 B]]((a: term[A, B]) => (b: term[A, B]) => mpo_impl[A, B](pr, a, b), ss, ts))
               else (false, false))
           })
}

def mpo[A: equal, B: equal]:
      ((A => nat => A => nat => (Boolean, Boolean), A => Boolean)) =>
        nat => (term[A, B]) => (term[A, B]) => (Boolean, Boolean) =
  (p: (A => nat => A => nat => (Boolean, Boolean), A => Boolean)) =>
    (_: nat) => (a: term[A, B]) => (b: term[A, B]) => mpo_impl[A, B](p, a, b)

abstract sealed class xml
final case class
  Xml(a: List[Char], b: List[(List[Char], List[Char])], c: List[xml],
       d: Option[List[Char]])
  extends xml

abstract sealed class sum[+A, +B]
final case class Inl[A, B](a: A) extends sum[A, B]
final case class Inr[B, A](a: B) extends sum[A, B]

def binda[A, B, C](m: sum[A, B], f: B => sum[A, C]): sum[A, C] =
  m match {
    case Inl(a) => Inl[A, C](a)
    case Inr(a) => f(a)
  }

implicit def equal_char: equal[Char] = new equal[Char] {
  val `Ceta.equal` = (a: Char, b: Char) => a == b
}

def the[A](x0: Option[A]): A = x0 match {
  case Some(x) => x
}

def foldr[A, B](f: A => B => B, x1: List[A], a: B): B = (f, x1, a) match {
  case (f, Nil, a) => a
  case (f, x :: xs, a) => (f(x))(foldr[A, B](f, xs, a))
}

def concat[A](xss: List[List[A]]): List[A] =
  foldr[List[A], List[A]]((a: List[A]) => (b: List[A]) => a ++ b, xss, Nil)

abstract sealed class nibble
final case object Nibble0 extends nibble
final case object Nibble1 extends nibble
final case object Nibble2 extends nibble
final case object Nibble3 extends nibble
final case object Nibble4 extends nibble
final case object Nibble5 extends nibble
final case object Nibble6 extends nibble
final case object Nibble7 extends nibble
final case object Nibble8 extends nibble
final case object Nibble9 extends nibble
final case object NibbleA extends nibble
final case object NibbleB extends nibble
final case object NibbleC extends nibble
final case object NibbleD extends nibble
final case object NibbleE extends nibble
final case object NibbleF extends nibble

def newline: Char = '\12'

def comp[A, B, C](f: A => B, g: C => A): C => B = (x: C) => f(g(x))

def shows_string: (List[Char]) => (List[Char]) => List[Char] =
  (a: List[Char]) => (b: List[Char]) => a ++ b

trait show[A] {
  val `Ceta.shows_prec`: (nat, A, List[Char]) => List[Char]
  val `Ceta.shows_list`: (List[A], List[Char]) => List[Char]
}
def shows_prec[A](a: nat, b: A, c: List[Char])(implicit A: show[A]): List[Char]
  = A.`Ceta.shows_prec`(a, b, c)
def shows_list[A](a: List[A], b: List[Char])(implicit A: show[A]): List[Char] =
  A.`Ceta.shows_list`(a, b)

def shows_prec_list[A: show](d: nat, l: List[A]): (List[Char]) => List[Char] =
  (a: List[Char]) => shows_list[A](l, a)

def shows_prec_char(d: nat, c: Char): (List[Char]) => List[Char] =
  (a: List[Char]) => c :: a

def shows_list_char(cs: List[Char]): (List[Char]) => List[Char] =
  shows_string.apply(cs)

implicit def show_char: show[Char] = new show[Char] {
  val `Ceta.shows_prec` = (a: nat, b: Char, c: List[Char]) =>
    (shows_prec_char(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[Char], b: List[Char]) =>
    (shows_list_char(a)).apply(b)
}

def shows_map[A](s: A => (List[Char]) => List[Char], x1: List[A]):
      (List[Char]) => List[Char] =
  (s, x1) match {
  case (s, Nil) => id[List[Char]]
  case (s, x :: xs) =>
    comp[List[Char], List[Char], List[Char]](s(x), shows_map[A](s, xs))
}

def dquote: List[Char] = List('\"')

def shows_attr(av: (List[Char], List[Char])): (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_prec_list[Char](Zero_nat,
   fst[List[Char], List[Char]](av)),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('=')),
shows_string.apply(dquote ++ (snd[List[Char], List[Char]](av) ++ dquote))))

def shows_attrs(as: List[(List[Char], List[Char])]):
      (List[Char]) => List[Char] =
  shows_map[(List[Char],
              List[Char])]((a: (List[Char], List[Char])) =>
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List(' ')),
        shows_attr(a)),
                            as)

def replicate[A](xa0: nat, x: A): List[A] = (xa0, x) match {
  case (Zero_nat, x) => Nil
  case (Suc(n), x) => x :: replicate[A](n, x)
}

def shows_XML_indent(ind: List[Char], i: nat, x2: xml):
      (List[Char]) => List[Char] =
  (ind, i, x2) match {
  case (ind, i, Xml(n, a, c, t)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(ind),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('<')),
  comp[List[Char], List[Char],
        List[Char]](shows_prec_list[Char](Zero_nat, n),
                     comp[List[Char], List[Char],
                           List[Char]](shows_attrs(a),
(if (nulla[xml](c) && is_none[List[Char]](t)) shows_string.apply(List('/', '>'))
  else comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('>')),
                          comp[List[Char], List[Char],
                                List[Char]]((if (! (nulla[xml](c)))
      comp[List[Char], List[Char],
            List[Char]](shows_map[xml]((aa: xml) =>
 shows_XML_indent(ind ++ replicate[Char](i, ' '), i, aa),
c),
                         shows_string.apply(ind))
      else id[List[Char]]),
     comp[List[Char], List[Char],
           List[Char]]((if (is_none[List[Char]](t)) id[List[Char]]
                         else shows_string.apply(the[List[Char]](t))),
                        comp[List[Char], List[Char],
                              List[Char]](shows_string.apply(List('<', '/')),
   comp[List[Char], List[Char],
         List[Char]](shows_prec_list[Char](Zero_nat, n),
                      shows_string.apply(List('>'))))))))))))
}

def shows_prec_xml(d: nat, xml: xml): (List[Char]) => List[Char] =
  shows_XML_indent(List(newline), nat(BigInt(2)), xml)

def text(tag: List[Char], x1: xml): sum[List[Char], List[Char]] = (tag, x1)
  match {
  case (tag, Xml(n, atts, cs, t)) =>
    (if (equal_lista[Char](n, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (nulla[xml](cs) &&
               (! (is_none[List[Char]](t)) &&
                 ! (nulla[Char](the[List[Char]](t)))))))
      Inr[List[Char], List[Char]](the[List[Char]](t))
      else Inl[List[Char],
                List[Char]](concat[Char](List(List('c', 'o', 'u', 'l', 'd', ' ',
            'n', 'o', 't', ' ', 'e', 'x', 't', 'r', 'a', 'c', 't', ' ', 't',
            'e', 'x', 't', ' ', 'f', 'o', 'r', ' '),
       tag, List(' ', 'f', 'r', 'o', 'm', ' '), List(newline),
       (shows_prec_xml(Zero_nat, Xml(n, atts, cs, t))).apply(Nil)))))
}

def take[A](n: nat, x1: List[A]): List[A] = (n, x1) match {
  case (n, Nil) => Nil
  case (n, x :: xs) =>
    n match {
      case Zero_nat => Nil
      case Suc(m) => x :: take[A](m, xs)
    }
}

def one_nat: nat = Suc(Zero_nat)

def digit2int(x: Char): sum[List[Char], BigInt] =
  (if (x == '0') Inr[BigInt, List[Char]](BigInt(0))
    else (if (x == '1') Inr[BigInt, List[Char]](BigInt(1))
           else (if (x == '2') Inr[BigInt, List[Char]](BigInt(2))
                  else (if (x == '3') Inr[BigInt, List[Char]](BigInt(3))
                         else (if (x == '4') Inr[BigInt, List[Char]](BigInt(4))
                                else (if (x == '5')
                                       Inr[BigInt, List[Char]](BigInt(5))
                                       else (if (x == '6')
      Inr[BigInt, List[Char]](BigInt(6))
      else (if (x == '7') Inr[BigInt, List[Char]](BigInt(7))
             else (if (x == '8') Inr[BigInt, List[Char]](BigInt(8))
                    else (if (x == '9') Inr[BigInt, List[Char]](BigInt(9))
                           else Inl[List[Char],
                                     BigInt](List(x) ++
       List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'd', 'i', 'g',
             'i', 't'))))))))))))

def string2int_aux(n: BigInt, x1: List[Char]): sum[List[Char], BigInt] = (n, x1)
  match {
  case (n, Nil) => Inr[BigInt, List[Char]](n)
  case (n, d :: s) =>
    binda[List[Char], BigInt,
           BigInt](digit2int(d),
                    (m: BigInt) => string2int_aux(BigInt(10) * n + m, s))
}

def string2int(s: List[Char]): sum[List[Char], BigInt] =
  (if (nulla[Char](s))
    Inl[List[Char],
         BigInt](List('c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o', 'n', 'v',
                       'e', 'r', 't', ' ', 'e', 'm', 'p', 't', 'y', ' ', 's',
                       't', 'r', 'i', 'n', 'g', ' ', 'i', 'n', 't', 'o', ' ',
                       'n', 'u', 'm', 'b', 'e', 'r'))
    else (if (equal_lista[Char](take[Char](one_nat, s), List('-')))
           binda[List[Char], BigInt,
                  BigInt](string2int_aux(BigInt(0), tl[Char](s)),
                           (i: BigInt) =>
                             Inr[BigInt, List[Char]](BigInt(0) - i))
           else string2int_aux(BigInt(0), s)))

def int(tag: List[Char], x: xml): sum[List[Char], BigInt] =
  binda[List[Char], List[Char],
         BigInt](text(tag, x), (a: List[Char]) => string2int(a))

def sequence[A, B](x0: List[sum[A, B]]): sum[A, List[B]] = x0 match {
  case Nil => Inr[List[B], A](Nil)
  case m :: ms =>
    binda[A, B,
           List[B]](m, (x: B) =>
                         binda[A, List[B],
                                List[B]](sequence[A, B](ms),
  (xs: List[B]) => Inr[List[B], A](x :: xs)))
}

def mapa[A, B](f: A => B, x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x :: xs) => f(x) :: mapa[A, B](f, xs)
}

def mapm[A, B, C](f: A => sum[B, C], ms: List[A]): sum[B, List[C]] =
  sequence[B, C](mapa[A, sum[B, C]](f, ms))

def map[A]:
      (xml => sum[List[Char], A]) => (List[xml]) => sum[List[Char], List[A]] =
  (a: xml => sum[List[Char], A]) =>
    (b: List[xml]) => mapm[xml, List[Char], A](a, b)

def nata(tag: List[Char], x: xml): sum[List[Char], nat] =
  binda[List[Char], List[Char],
         nat](text(tag, x),
               (txt: List[Char]) =>
                 binda[List[Char], BigInt,
                        nat](string2int(txt),
                              (i: BigInt) => Inr[nat, List[Char]](nat(i))))

def maps[A, B](f: A => List[B], x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x :: xs) => f(x) ++ maps[A, B](f, xs)
}

def lookupa[A: equal, B](uu: A, x1: List[(A, B)]): Option[B] = (uu, x1) match {
  case (uu, Nil) => None
  case (aa, (a, b) :: xs) =>
    (if (eq[A](aa, a)) Some[B](b) else lookupa[A, B](aa, xs))
}

implicit def equal_list[A: equal]: equal[List[A]] = new equal[List[A]] {
  val `Ceta.equal` = (a: List[A], b: List[A]) => equal_lista[A](a, b)
}

def get_name(x0: xml): List[Char] = x0 match {
  case Xml(name, uu, uv, uw) => name
}

def xml_options[A](ps: List[(List[Char], xml => sum[List[Char], A])], x: xml):
      sum[List[Char], A] =
  lookupa[List[Char], xml => sum[List[Char], A]](get_name(x), ps) match {
    case None =>
      Inl[List[Char],
           A](concat[Char](List(List('e', 'x', 'p', 'e', 'c', 't', 'e', 'd',
                                      ' ', 'o', 'n', 'e', ' ', 'o', 'f', ':',
                                      ' '),
                                 maps[(List[Char], xml => sum[List[Char], A]),
                                       Char]((p:
        (List[Char], xml => sum[List[Char], A]))
       => fst[List[Char], xml => sum[List[Char], A]](p) ++ List(' '),
      ps),
                                 List(newline),
                                 List('b', 'u', 't', ' ', 'f', 'o', 'u', 'n',
                                       'd'),
                                 List(newline),
                                 (shows_prec_xml(Zero_nat, x)).apply(Nil))))
    case Some(p) => p(x)
  }

def xml_change[A, B](p: xml => sum[List[Char], A], f: A => B, x: xml):
      sum[List[Char], B] =
  binda[List[Char], A, B](p(x), (a: A) => Inr[B, List[Char]](f(a)))

def abs_int(i: BigInt): BigInt = (if (i < BigInt(0)) (- i) else i)

def apsnd[A, B, C](f: A => B, x1: (C, A)): (C, B) = (f, x1) match {
  case (f, (x, y)) => (x, f(y))
}

def sgn_int(i: BigInt): BigInt =
  (if (i == BigInt(0)) BigInt(0)
    else (if (BigInt(0) < i) BigInt(1) else (- BigInt(1))))

def divmod_int(k: BigInt, l: BigInt): (BigInt, BigInt) =
  (if (k == BigInt(0)) (BigInt(0), BigInt(0))
    else (if (l == BigInt(0)) (BigInt(0), k)
           else apsnd[BigInt, BigInt,
                       BigInt]((a: BigInt) => sgn_int(l) * a,
                                (if (sgn_int(k) == sgn_int(l))
                                  ((k: BigInt) => (l: BigInt) => if (l == 0)
                                    (BigInt(0), k) else
                                    (k.abs /% l.abs)).apply(k).apply(l)
                                  else {
 val (r, s): (BigInt, BigInt) =
   ((k: BigInt) => (l: BigInt) => if (l == 0) (BigInt(0), k) else
     (k.abs /% l.abs)).apply(k).apply(l);
 (if (s == BigInt(0)) ((- r), BigInt(0))
   else ((- r) - BigInt(1), abs_int(l) - s))
                                       }))))

def mod_int(a: BigInt, b: BigInt): BigInt =
  snd[BigInt, BigInt](divmod_int(a, b))

def gcd_int(k: BigInt, l: BigInt): BigInt =
  abs_int((if (l == BigInt(0)) k
            else gcd_int(l, mod_int(abs_int(k), abs_int(l)))))

def div_int(a: BigInt, b: BigInt): BigInt =
  fst[BigInt, BigInt](divmod_int(a, b))

def normalize(p: (BigInt, BigInt)): (BigInt, BigInt) =
  (if (BigInt(0) < snd[BigInt, BigInt](p))
    {
      val a: BigInt = gcd_int(fst[BigInt, BigInt](p), snd[BigInt, BigInt](p));
      (div_int(fst[BigInt, BigInt](p), a), div_int(snd[BigInt, BigInt](p), a))
    }
    else (if (snd[BigInt, BigInt](p) == BigInt(0)) (BigInt(0), BigInt(1))
           else {
                  val a: BigInt =
                    (- (gcd_int(fst[BigInt, BigInt](p),
                                 snd[BigInt, BigInt](p))));
                  (div_int(fst[BigInt, BigInt](p), a),
                    div_int(snd[BigInt, BigInt](p), a))
                }))

def fract(a: BigInt, b: BigInt): rat = Frct(normalize((a, b)))

def equal_nata(x0: nat, x1: nat): Boolean = (x0, x1) match {
  case (Suc(nat), Zero_nat) => false
  case (Zero_nat, Suc(nat)) => false
  case (Suc(nata), Suc(nat)) => equal_nata(nata, nat)
  case (Zero_nat, Zero_nat) => true
}

def nth[A](x0: List[A], x1: nat): A = (x0, x1) match {
  case (x :: xs, Suc(n)) => nth[A](xs, n)
  case (x :: xs, Zero_nat) => x
}

def no_transform[A](tag: List[Char], xml: xml): sum[List[Char], A] =
  Inl[List[Char],
       A](concat[Char](List(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                                  ' ', 't', 'r', 'a', 'n', 's', 'f', 'o', 'r',
                                  'm', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l',
                                  'l', 'o', 'w', 'i', 'n', 'g', '\12', ' ', ' ',
                                  ' ', ' ', ' ', 'x', 'm', 'l', ' ', 'e', 'l',
                                  'e', 'm', 'e', 'n', 't', ' ', '(', 'e', 'x',
                                  'p', 'e', 'c', 't', 'e', 'd', ' '),
                             tag, List(')'), List(newline),
                             (shows_prec_xml(Zero_nat, xml)).apply(Nil))))

def xml2elements[A, B,
                  C](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B], f: A => B => C, x4: xml):
      sum[List[Char], C] =
  (tag, p1, p2, f, x4) match {
  case (tag, p1, p2, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(2))) &&
               is_none[List[Char]](text))))
      binda[List[Char], A,
             C](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   binda[List[Char], B,
                          C](p2(nth[xml](cs, one_nat)),
                              (x2: B) => Inr[C, List[Char]]((f(x1))(x2))))
      else no_transform[C](tag, Xml(name, atts, cs, text)))
}

def rat: xml => sum[List[Char], rat] =
  (a: xml) =>
    xml_options[rat](List((List('i', 'n', 't', 'e', 'g', 'e', 'r'),
                            (aa: xml) =>
                              xml_change[BigInt,
  rat]((ab: xml) => int(List('i', 'n', 't', 'e', 'g', 'e', 'r'), ab),
        (x: BigInt) => fract(x, BigInt(1)), aa)),
                           (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l'),
                             (aa: xml) =>
                               xml2elements[BigInt, BigInt,
     rat](List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l'),
           (ab: xml) =>
             int(List('n', 'u', 'm', 'e', 'r', 'a', 't', 'o', 'r'), ab),
           (ab: xml) =>
             int(List('d', 'e', 'n', 'o', 'm', 'i', 'n', 'a', 't', 'o', 'r'),
                  ab),
           (ab: BigInt) => (b: BigInt) => fract(ab, b), aa))),
                      a)

abstract sealed class hterm[+A, +B]
final case object HHole extends hterm[Nothing, Nothing]
final case class HFun[A, B](a: A, b: List[hterm[A, B]]) extends hterm[A, B]

def mergeHTerms[A: equal, B](x0: List[hterm[A, B]], x1: List[hterm[A, B]]):
      Option[List[hterm[A, B]]] =
  (x0, x1) match {
  case (Nil, Nil) => Some[List[hterm[A, B]]](Nil)
  case (Nil, uu :: uv) => None
  case (uw :: ux, Nil) => None
  case (HHole :: hsl, h :: hsr) =>
    mergeHTerms[A, B](hsl, hsr) match {
      case None => None
      case Some(res) => Some[List[hterm[A, B]]](h :: res)
    }
  case (HFun(f, hs) :: hsl, HHole :: hsr) =>
    mergeHTerms[A, B](hsl, hsr) match {
      case None => None
      case Some(res) => Some[List[hterm[A, B]]](HFun[A, B](f, hs) :: res)
    }
  case (HFun(f, fs) :: hsl, HFun(g, gs) :: hsr) =>
    (if (eq[A](f, g))
      mergeHTerms[A, B](fs, gs) match {
        case None => None
        case Some(mhs) =>
          mergeHTerms[A, B](hsl, hsr) match {
            case None => None
            case Some(res) => Some[List[hterm[A, B]]](HFun[A, B](f, mhs) :: res)
          }
      }
      else None)
}

def merge[A: equal, B](hl: hterm[A, B], hr: hterm[A, B]): Option[hterm[A, B]] =
  mergeHTerms[A, B](List(hl), List(hr)) match {
    case None => None
    case Some(res) => Some[hterm[A, B]](nth[hterm[A, B]](res, Zero_nat))
  }

def mergeX[A: equal,
            B: equal](x: A, h: hterm[B, A], xa2: List[(hterm[B, A], A)]):
      Option[List[(hterm[B, A], A)]] =
  (x, h, xa2) match {
  case (x, h, Nil) => Some[List[(hterm[B, A], A)]](Nil)
  case (x, h, (ht, y) :: ps) =>
    (if (eq[A](x, y))
      merge[B, A](h, ht) match {
        case None => None
        case Some(m) => mergeX[A, B](x, m, ps)
      }
      else mergeX[A, B](x, h, ps) match {
             case None => None
             case Some(pairs) => Some[List[(hterm[B, A], A)]]((ht, y) :: pairs)
           })
}

def mergeAll[A: equal, B: equal](x0: List[(hterm[A, B], B)]): Boolean = x0 match
  {
  case Nil => true
  case (h, x) :: ps =>
    mergeX[B, A](x, h, ps) match {
      case None => false
      case Some(a) => mergeAll[A, B](a)
    }
}

def decomp[A, B](x0: List[A], x1: List[B]): Option[List[(A, B)]] = (x0, x1)
  match {
  case (Nil, Nil) => Some[List[(A, B)]](Nil)
  case (uu :: uv, Nil) => None
  case (Nil, uw :: ux) => None
  case (s :: ss, t :: ts) =>
    decomp[A, B](ss, ts) match {
      case None => None
      case Some(pairs) => Some[List[(A, B)]]((s, t) :: pairs)
    }
}

def gcMatchList[A: equal, B](x0: List[(hterm[A, B], term[A, B])]):
      Option[List[(hterm[A, B], B)]] =
  x0 match {
  case Nil => Some[List[(hterm[A, B], B)]](Nil)
  case (HHole, uu) :: ps => gcMatchList[A, B](ps)
  case (HFun(f, hs), Fun(g, ss)) :: ps =>
    (if (eq[A](f, g))
      decomp[hterm[A, B], term[A, B]](hs, ss) match {
        case None => None
        case Some(pairs) => gcMatchList[A, B](pairs ++ ps)
      }
      else None)
  case (HFun(f, hs), Var(x)) :: ps =>
    gcMatchList[A, B](ps) match {
      case None => None
      case Some(pairs) =>
        Some[List[(hterm[A, B], B)]]((HFun[A, B](f, hs), x) :: pairs)
    }
}

def gcMatchI[A: equal, B: equal](h: hterm[A, B], t: term[A, B]): Boolean =
  gcMatchList[A, B](List((h, t))) match {
    case None => false
    case Some(a) => mergeAll[A, B](a)
  }

def capI[A: equal,
          B: equal](uu: List[(term[A, B], term[A, B])], x1: term[A, B]):
      hterm[A, B] =
  (uu, x1) match {
  case (uu, Var(uv)) => HHole
  case (r, Fun(f, ts)) =>
    {
      val h: hterm[A, B] =
        HFun[A, B](f, mapa[term[A, B],
                            hterm[A, B]]((a: term[A, B]) => capI[A, B](r, a),
  ts));
      (if (list_ex[(term[A, B],
                     term[A, B])]((ra: (term[A, B], term[A, B])) =>
                                    gcMatchI[A,
      B](h, fst[term[A, B], term[A, B]](ra)),
                                   r))
        HHole else h)
    }
}

def fold[A, B](f: A => B => B, x1: List[A]): B => B = (f, x1) match {
  case (f, Nil) => id[B]
  case (f, x :: xs) => comp[B, B, B](fold[A, B](f, xs), f(x))
}

def rev[A](xs: List[A]): List[A] =
  (fold[A, List[A]]((a: A) => (b: List[A]) => a :: b, xs)).apply(Nil)

trait bot[A] extends preorder[A] {
  val `Ceta.bot`: A
}
def bot[A](implicit A: bot[A]): A = A.`Ceta.bot`

def bot_fun[A, B: bot]: A => B = (_: A) => bot[B]

def less_eq_bool(x0: Boolean, b: Boolean): Boolean = (x0, b) match {
  case (true, b) => b
  case (false, b) => true
}

def less_bool(x0: Boolean, b: Boolean): Boolean = (x0, b) match {
  case (true, b) => false
  case (false, b) => b
}

implicit def ord_bool: ord[Boolean] = new ord[Boolean] {
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

implicit def preorder_bool: preorder[Boolean] = new preorder[Boolean] {
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

def bot_boola: Boolean = false

implicit def bot_bool: bot[Boolean] = new bot[Boolean] {
  val `Ceta.bot` = bot_boola
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

def inserta[A: equal](y: A, a: A => Boolean, x: A): Boolean =
  eq[A](y, x) || a(x)

def set[A: equal](x0: List[A]): A => Boolean = x0 match {
  case Nil => bot_fun[A, Boolean]
  case x :: xs => (a: A) => inserta[A](x, set[A](xs), a)
}

def less_eq_nat(x0: nat, n: nat): Boolean = (x0, n) match {
  case (Suc(m), n) => less_nat(m, n)
  case (Zero_nat, n) => true
}

def less_nat(m: nat, x1: nat): Boolean = (m, x1) match {
  case (m, Suc(n)) => less_eq_nat(m, n)
  case (n, Zero_nat) => false
}

def upt(i: nat, j: nat): List[nat] =
  (if (less_nat(i, j)) i :: upt(Suc(i), j) else Nil)

def zip[A, B](xs: List[A], ys: List[B]): List[(A, B)] = (xs, ys) match {
  case (x :: xs, y :: ys) => (x, y) :: zip[A, B](xs, ys)
  case (xs, Nil) => Nil
  case (Nil, ys) => Nil
}

abstract sealed class ctxt[+A, +B]
final case object Hole extends ctxt[Nothing, Nothing]
final case class
  More[A, B](a: A, b: List[term[A, B]], c: ctxt[A, B], d: List[term[A, B]])
  extends ctxt[A, B]

def string2bool(s: List[Char]): sum[List[Char], Boolean] =
  (if (equal_lista[Char](s, List('t', 'r', 'u', 'e')))
    Inr[Boolean, List[Char]](true)
    else (if (equal_lista[Char](s, List('f', 'a', 'l', 's', 'e')))
           Inr[Boolean, List[Char]](false)
           else Inl[List[Char],
                     Boolean](List('c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o',
                                    'n', 'v', 'e', 'r', 't', ' ') ++
                                (s ++ List(' ', 'i', 'n', 't', 'o', ' ', 'B',
    'o', 'o', 'l', 'e', 'a', 'n')))))

def bool(tag: List[Char], node: xml): sum[List[Char], Boolean] =
  binda[List[Char], List[Char],
         Boolean](text(tag, node), (a: List[Char]) => string2bool(a))

def capRM2[A: equal,
            B: equal](uu: ((A, nat)) => List[(term[A, B], term[A, B])],
                       x1: term[A, B]):
      hterm[A, B] =
  (uu, x1) match {
  case (uu, Var(uv)) => HHole
  case (rm, Fun(f, ts)) =>
    {
      val h: hterm[A, B] =
        HFun[A, B](f, mapa[term[A, B],
                            hterm[A, B]]((a: term[A, B]) => capRM2[A, B](rm, a),
  ts))
      val n: nat = size_list[term[A, B]](ts);
      (if (list_ex[(term[A, B],
                     term[A, B])]((r: (term[A, B], term[A, B])) =>
                                    gcMatchI[A,
      B](h, fst[term[A, B], term[A, B]](r)),
                                   rm((f, n))))
        HHole else h)
    }
}

def capRM[A: equal,
           B: equal](r: Option[((A, nat)) => List[(term[A, B], term[A, B])]]):
      (term[A, B]) => hterm[A, B] =
  r match {
    case None => (_: term[A, B]) => HHole
    case Some(a) => (b: term[A, B]) => capRM2[A, B](a, b)
  }

def drop[A](n: nat, x1: List[A]): List[A] = (n, x1) match {
  case (n, Nil) => Nil
  case (n, x :: xs) =>
    n match {
      case Zero_nat => x :: xs
      case Suc(m) => drop[A](m, xs)
    }
}

def last[A](x0: List[A]): A = x0 match {
  case x :: xs => (if (nulla[A](xs)) x else last[A](xs))
}

def empty[A: linorder, B]: rbt[A, B] = Rbt[A, B](Emptya)

trait times[A] {
  val `Ceta.times`: (A, A) => A
}
def times[A](a: A, b: A)(implicit A: times[A]): A = A.`Ceta.times`(a, b)

trait dvd[A] extends times[A] {
}

def args[A, B](x0: term[A, B]): List[term[A, B]] = x0 match {
  case Var(x) => Nil
  case Fun(f, ts) => ts
}

def root[A, B](x0: term[A, B]): Option[A] = x0 match {
  case Var(uu) => None
  case Fun(f, uv) => Some[A](f)
}

def membera[A: equal](x0: List[A], y: A): Boolean = (x0, y) match {
  case (Nil, y) => false
  case (x :: xs, y) => eq[A](x, y) || membera[A](xs, y)
}

def supt_impl[A: equal, B: equal](x0: term[A, B], t: term[A, B]): Boolean =
  (x0, t) match {
  case (Var(x), t) => false
  case (Fun(f, ss), t) =>
    membera[term[A, B]](ss, t) ||
      list_ex[term[A, B]]((s: term[A, B]) => supt_impl[A, B](s, t), ss)
}

def supt[A: equal, B: equal]: ((term[A, B], term[A, B])) => Boolean =
  (a: (term[A, B], term[A, B])) =>
    {
      val (aa, b): (term[A, B], term[A, B]) = a;
      supt_impl[A, B](aa, b)
    }

def find[A](uu: A => Boolean, x1: List[A]): Option[A] = (uu, x1) match {
  case (uu, Nil) => None
  case (p, x :: xs) => (if (p(x)) Some[A](x) else find[A](p, xs))
}

trait one[A] {
  val `Ceta.one`: A
}
def one[A](implicit A: one[A]): A = A.`Ceta.one`

implicit def times_int: times[BigInt] = new times[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def dvd_int: dvd[BigInt] = new dvd[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

def one_inta: BigInt = BigInt(1)

implicit def one_int: one[BigInt] = new one[BigInt] {
  val `Ceta.one` = one_inta
}

implicit def ord_int: ord[BigInt] = new ord[BigInt] {
  val `Ceta.less_eq` = (a: BigInt, b: BigInt) => a <= b
  val `Ceta.less` = (a: BigInt, b: BigInt) => a < b
}

def foldl[A, B](f: A => B => A, a: A, x2: List[B]): A = (f, a, x2) match {
  case (f, a, Nil) => a
  case (f, a, x :: xs) => foldl[A, B](f, (f(a))(x), xs)
}

def vec[A](n: nat, x: List[A]): Boolean = equal_nata(size_list[A](x), n)

def mat[A](nr: nat, nc: nat, m: List[List[A]]): Boolean =
  equal_nata(size_list[List[A]](m), nc) &&
    list_all[List[A]]((a: List[A]) => vec[A](nr, a), m)

def funpow[A](x0: nat, f: A => A): A => A = (x0, f) match {
  case (Zero_nat, f) => id[A]
  case (Suc(n), f) => comp[A, A, A](f, funpow[A](n, f))
}

implicit def ord_nat: ord[nat] = new ord[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

def paint[A, B](c: color, x1: rbta[A, B]): rbta[A, B] = (c, x1) match {
  case (c, Emptya) => Emptya
  case (c, Branch(uu, l, k, v, r)) => Branch[A, B](c, l, k, v, r)
}

def balance[A, B](x0: rbta[A, B], s: A, t: B, x3: rbta[A, B]): rbta[A, B] =
  (x0, s, t, x3) match {
  case (Branch(R, a, w, x, b), s, t, Branch(R, c, y, z, d)) =>
    Branch[A, B](R, Branch[A, B](B, a, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, d))
  case (Branch(R, Branch(R, a, w, x, b), s, t, c), y, z, Emptya) =>
    Branch[A, B](R, Branch[A, B](B, a, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, Emptya))
  case (Branch(R, Branch(R, a, w, x, b), s, t, c), y, z,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](R, Branch[A, B](B, a, w, x, b), s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Branch(R, Emptya, w, x, Branch(R, b, s, t, c)), y, z, Emptya) =>
    Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, Emptya))
  case (Branch(R, Branch(B, va, vb, vc, vd), w, x, Branch(R, b, s, t, c)), y, z,
         Emptya)
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t, Branch[A, B](B, c, y, z, Emptya))
  case (Branch(R, Emptya, w, x, Branch(R, b, s, t, c)), y, z,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Branch(R, Branch(B, ve, vf, vg, vh), w, x, Branch(R, b, s, t, c)), y, z,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, ve, vf, vg, vh), w, x,
                                     b),
                     s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Emptya, w, x, Branch(R, b, s, t, Branch(R, c, y, z, d))) =>
    Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, d))
  case (Branch(B, va, vb, vc, vd), w, x,
         Branch(R, b, s, t, Branch(R, c, y, z, d)))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t, Branch[A, B](B, c, y, z, d))
  case (Emptya, w, x, Branch(R, Branch(R, b, s, t, c), y, z, Emptya)) =>
    Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, Emptya))
  case (Emptya, w, x,
         Branch(R, Branch(R, b, s, t, c), y, z, Branch(B, va, vb, vc, vd)))
    => Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Branch(B, va, vb, vc, vd), w, x,
         Branch(R, Branch(R, b, s, t, c), y, z, Emptya))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t, Branch[A, B](B, c, y, z, Emptya))
  case (Branch(B, va, vb, vc, vd), w, x,
         Branch(R, Branch(R, b, s, t, c), y, z, Branch(B, ve, vf, vg, vh)))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, ve, vf, vg, vh)))
  case (Emptya, s, t, Emptya) => Branch[A, B](B, Emptya, s, t, Emptya)
  case (Emptya, s, t, Branch(B, va, vb, vc, vd)) =>
    Branch[A, B](B, Emptya, s, t, Branch[A, B](B, va, vb, vc, vd))
  case (Emptya, s, t, Branch(v, Emptya, vb, vc, Emptya)) =>
    Branch[A, B](B, Emptya, s, t, Branch[A, B](v, Emptya, vb, vc, Emptya))
  case (Emptya, s, t, Branch(v, Branch(B, ve, vf, vg, vh), vb, vc, Emptya)) =>
    Branch[A, B](B, Emptya, s, t,
                  Branch[A, B](v, Branch[A, B](B, ve, vf, vg, vh), vb, vc,
                                Emptya))
  case (Emptya, s, t, Branch(v, Emptya, vb, vc, Branch(B, vf, vg, vh, vi))) =>
    Branch[A, B](B, Emptya, s, t,
                  Branch[A, B](v, Emptya, vb, vc,
                                Branch[A, B](B, vf, vg, vh, vi)))
  case (Emptya, s, t,
         Branch(v, Branch(B, ve, vj, vk, vl), vb, vc,
                 Branch(B, vf, vg, vh, vi)))
    => Branch[A, B](B, Emptya, s, t,
                     Branch[A, B](v, Branch[A, B](B, ve, vj, vk, vl), vb, vc,
                                   Branch[A, B](B, vf, vg, vh, vi)))
  case (Branch(B, va, vb, vc, vd), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t, Emptya)
  case (Branch(B, va, vb, vc, vd), s, t, Branch(B, ve, vf, vg, vh)) =>
    Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                  Branch[A, B](B, ve, vf, vg, vh))
  case (Branch(B, va, vb, vc, vd), s, t, Branch(v, Emptya, vf, vg, Emptya)) =>
    Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                  Branch[A, B](v, Emptya, vf, vg, Emptya))
  case (Branch(B, va, vb, vc, vd), s, t,
         Branch(v, Branch(B, vi, vj, vk, vl), vf, vg, Emptya))
    => Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                     Branch[A, B](v, Branch[A, B](B, vi, vj, vk, vl), vf, vg,
                                   Emptya))
  case (Branch(B, va, vb, vc, vd), s, t,
         Branch(v, Emptya, vf, vg, Branch(B, vj, vk, vl, vm)))
    => Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                     Branch[A, B](v, Emptya, vf, vg,
                                   Branch[A, B](B, vj, vk, vl, vm)))
  case (Branch(B, va, vb, vc, vd), s, t,
         Branch(v, Branch(B, vi, vn, vo, vp), vf, vg,
                 Branch(B, vj, vk, vl, vm)))
    => Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                     Branch[A, B](v, Branch[A, B](B, vi, vn, vo, vp), vf, vg,
                                   Branch[A, B](B, vj, vk, vl, vm)))
  case (Branch(v, Emptya, vb, vc, Emptya), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](v, Emptya, vb, vc, Emptya), s, t, Emptya)
  case (Branch(v, Emptya, vb, vc, Branch(B, ve, vf, vg, vh)), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](v, Emptya, vb, vc,
                                  Branch[A, B](B, ve, vf, vg, vh)),
                  s, t, Emptya)
  case (Branch(v, Branch(B, vf, vg, vh, vi), vb, vc, Emptya), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vf, vg, vh, vi), vb, vc,
                                  Emptya),
                  s, t, Emptya)
  case (Branch(v, Branch(B, vf, vg, vh, vi), vb, vc, Branch(B, ve, vj, vk, vl)),
         s, t, Emptya)
    => Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vf, vg, vh, vi), vb, vc,
                                     Branch[A, B](B, ve, vj, vk, vl)),
                     s, t, Emptya)
  case (Branch(v, Emptya, vf, vg, Emptya), s, t, Branch(B, va, vb, vc, vd)) =>
    Branch[A, B](B, Branch[A, B](v, Emptya, vf, vg, Emptya), s, t,
                  Branch[A, B](B, va, vb, vc, vd))
  case (Branch(v, Emptya, vf, vg, Branch(B, vi, vj, vk, vl)), s, t,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](B, Branch[A, B](v, Emptya, vf, vg,
                                     Branch[A, B](B, vi, vj, vk, vl)),
                     s, t, Branch[A, B](B, va, vb, vc, vd))
  case (Branch(v, Branch(B, vj, vk, vl, vm), vf, vg, Emptya), s, t,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vj, vk, vl, vm), vf, vg,
                                     Emptya),
                     s, t, Branch[A, B](B, va, vb, vc, vd))
  case (Branch(v, Branch(B, vj, vk, vl, vm), vf, vg, Branch(B, vi, vn, vo, vp)),
         s, t, Branch(B, va, vb, vc, vd))
    => Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vj, vk, vl, vm), vf, vg,
                                     Branch[A, B](B, vi, vn, vo, vp)),
                     s, t, Branch[A, B](B, va, vb, vc, vd))
}

def ins[A: linorder, B](f: A => B => B => B, k: A, v: B, x3: rbta[A, B]):
      rbta[A, B] =
  (f, k, v, x3) match {
  case (f, k, v, Emptya) => Branch[A, B](R, Emptya, k, v, Emptya)
  case (f, k, v, Branch(B, l, x, y, r)) =>
    (if (less[A](k, x)) balance[A, B](ins[A, B](f, k, v, l), x, y, r)
      else (if (less[A](x, k)) balance[A, B](l, x, y, ins[A, B](f, k, v, r))
             else Branch[A, B](B, l, x, ((f(k))(y))(v), r)))
  case (f, k, v, Branch(R, l, x, y, r)) =>
    (if (less[A](k, x)) Branch[A, B](R, ins[A, B](f, k, v, l), x, y, r)
      else (if (less[A](x, k)) Branch[A, B](R, l, x, y, ins[A, B](f, k, v, r))
             else Branch[A, B](R, l, x, ((f(k))(y))(v), r)))
}

def insert_with_key[A: linorder,
                     B](f: A => B => B => B, k: A, v: B, t: rbta[A, B]):
      rbta[A, B] =
  paint[A, B](B, ins[A, B](f, k, v, t))

def insertb[A: linorder, B]: A => B => (rbta[A, B]) => rbta[A, B] =
  (a: A) =>
    (b: B) =>
      (c: rbta[A, B]) =>
        insert_with_key[A, B]((_: A) => (_: B) => (nv: B) => nv, a, b, c)

def impl_of[A: linorder, B](x0: rbt[A, B]): rbta[A, B] = x0 match {
  case Rbt(x) => x
}

def insert[A: linorder, B](k: A, v: B, t: rbt[A, B]): rbt[A, B] =
  Rbt[A, B](insertb[A, B].apply(k).apply(v).apply(impl_of[A, B](t)))

def lookupb[A: linorder, B](x0: rbta[A, B], k: A): Option[B] = (x0, k) match {
  case (Emptya, k) => None
  case (Branch(uu, l, x, y, r), k) =>
    (if (less[A](k, x)) lookupb[A, B](l, k)
      else (if (less[A](x, k)) lookupb[A, B](r, k) else Some[B](y)))
}

def lookup[A: linorder, B](t: rbt[A, B]): A => Option[B] =
  (a: A) => lookupb[A, B](impl_of[A, B](t), a)

def quotient_of(x0: rat): (BigInt, BigInt) = x0 match {
  case Frct(x) => x
}

def times_rata(p: rat, q: rat): rat =
  Frct({
         val (a, c): (BigInt, BigInt) = quotient_of(p)
         val (b, d): (BigInt, BigInt) = quotient_of(q);
         normalize((a * b, c * d))
       })

implicit def times_rat: times[rat] = new times[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def dvd_rat: dvd[rat] = new dvd[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

def one_rata: rat = Frct((one_inta, one_inta))

implicit def one_rat: one[rat] = new one[rat] {
  val `Ceta.one` = one_rata
}

def less_eq_rat(p: rat, q: rat): Boolean =
  {
    val (a, c): (BigInt, BigInt) = quotient_of(p)
    val (b, d): (BigInt, BigInt) = quotient_of(q);
    a * d <= c * b
  }

def less_rat(p: rat, q: rat): Boolean =
  {
    val (a, c): (BigInt, BigInt) = quotient_of(p)
    val (b, d): (BigInt, BigInt) = quotient_of(q);
    a * d < c * b
  }

implicit def ord_rat: ord[rat] = new ord[rat] {
  val `Ceta.less_eq` = (a: rat, b: rat) => less_eq_rat(a, b)
  val `Ceta.less` = (a: rat, b: rat) => less_rat(a, b)
}

def member[A](x: A, a: A => Boolean): Boolean = a(x)

def debug[A](i: List[Char], t: List[Char], x: A): A = x

def fun_upd[A: equal, B](f: A => B, a: A, b: B): A => B =
  (x: A) => (if (eq[A](x, a)) b else f(x))

trait plus[A] {
  val `Ceta.plus`: (A, A) => A
}
def plus[A](a: A, b: A)(implicit A: plus[A]): A = A.`Ceta.plus`(a, b)

trait zero[A] {
  val `Ceta.zero`: A
}
def zero[A](implicit A: zero[A]): A = A.`Ceta.zero`

def implies(p: Boolean, pa: Boolean): Boolean = (p, pa) match {
  case (p, true) => true
  case (p, false) => ! p
  case (true, p) => p
  case (false, p) => true
}

implicit def plus_int: plus[BigInt] = new plus[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

def zero_inta: BigInt = BigInt(0)

implicit def zero_int: zero[BigInt] = new zero[BigInt] {
  val `Ceta.zero` = zero_inta
}

abstract sealed class pos
final case object Empty extends pos
final case class PCons(a: nat, b: pos) extends pos

trait power[A] extends one[A] with times[A] {
}

def powera[A](one: A, times: A => A => A, a: A, x3: nat): A =
  (one, times, a, x3) match {
  case (one, times, a, Suc(n)) => (times(a))(powera[A](one, times, a, n))
  case (one, times, a, Zero_nat) => one
}

def power[A: power]: A => nat => A =
  (a: A) =>
    (b: nat) => powera[A](one[A], (aa: A) => (ba: A) => times[A](aa, ba), a, b)

def prc_nat[A](pr: A => nat => nat):
      A => nat => A => nat => (Boolean, Boolean) =
  (f: A) =>
    (n: nat) =>
      (g: A) =>
        (m: nat) =>
          {
            val pf: nat = (pr(f))(n)
            val pg: nat = (pr(g))(m);
            (less_nat(pg, pf), less_eq_nat(pg, pf))
          }

def prl_nat[A](pr: A => nat => nat): A => Boolean =
  (f: A) => equal_nata((pr(f))(Zero_nat), Zero_nat)

def plus_rata(p: rat, q: rat): rat =
  Frct({
         val (a, c): (BigInt, BigInt) = quotient_of(p)
         val (b, d): (BigInt, BigInt) = quotient_of(q);
         normalize((a * d + b * c, c * d))
       })

implicit def plus_rat: plus[rat] = new plus[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

def zero_rata: rat = Frct((zero_inta, one_inta))

implicit def zero_rat: zero[rat] = new zero[rat] {
  val `Ceta.zero` = zero_rata
}

def is_var[A, B](x0: term[A, B]): Boolean = x0 match {
  case Var(uu) => true
  case Fun(uv, uw) => false
}

def supteq[A: equal, B: equal]: ((term[A, B], term[A, B])) => Boolean =
  (a: (term[A, B], term[A, B])) =>
    {
      val (x, y): (term[A, B], term[A, B]) = a;
      equal_terma[A, B](x, y) || supt_impl[A, B](x, y)
    }

def shows_nl: (List[Char]) => List[Char] = shows_prec_char(Zero_nat, newline)

def shows_between(l: (List[Char]) => List[Char], p: (List[Char]) => List[Char],
                   r: (List[Char]) => List[Char]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](l, comp[List[Char], List[Char], List[Char]](p, r))

def shows_sep[A](s: A => (List[Char]) => List[Char],
                  sep: (List[Char]) => List[Char], x2: List[A]):
      (List[Char]) => List[Char] =
  (s, sep, x2) match {
  case (s, sep, Nil) => shows_string.apply(Nil)
  case (s, sep, List(x)) => s(x)
  case (s, sep, x :: v :: va) =>
    comp[List[Char], List[Char],
          List[Char]](s(x),
                       comp[List[Char], List[Char],
                             List[Char]](sep, shows_sep[A](s, sep, v :: va)))
}

def shows_list_gen[A](elt: A => (List[Char]) => List[Char], e: List[Char],
                       l: List[Char], s: List[Char], r: List[Char],
                       xs: List[A]):
      (List[Char]) => List[Char] =
  (if (nulla[A](xs)) shows_string.apply(e)
    else shows_between(shows_string.apply(l),
                        shows_sep[A](elt, shows_string.apply(s), xs),
                        shows_string.apply(r)))

def shows_term[A: show, B: show](x0: term[A, B]): (List[Char]) => List[Char] =
  x0 match {
  case Var(x) => (a: List[Char]) => shows_prec[B](Zero_nat, x, a)
  case Fun(f, ts) =>
    comp[List[Char], List[Char],
          List[Char]]((a: List[Char]) => shows_prec[A](Zero_nat, f, a),
                       shows_list_gen[(List[Char]) =>
List[Char]](id[(List[Char]) => List[Char]], Nil, List('('), List(',', ' '),
             List(')'),
             mapa[term[A, B],
                   (List[Char]) =>
                     List[Char]]((a: term[A, B]) => shows_term[A, B](a), ts)))
}

def shows_prec_term[A: show, B: show](d: nat, t: term[A, B]):
      (List[Char]) => List[Char] =
  shows_term[A, B](t)

def shows_rule[A: show, B: show](x0: (term[A, B], term[A, B])):
      (List[Char]) => List[Char] =
  x0 match {
  case (l, r) =>
    comp[List[Char], List[Char],
          List[Char]](shows_prec_term[A, B](Zero_nat, l),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List(' ', '-', '>',
                          ' ')),
  shows_prec_term[A, B](Zero_nat, r)))
}

def shows_rules[A: show, B: show](trs: List[(term[A, B], term[A, B])]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_list_gen[(term[A, B],
                                     term[A,
   B])]((a: (term[A, B], term[A, B])) => shows_rule[A, B](a), Nil, Nil,
         List(newline), Nil, trs),
                     shows_nl)

def shows_p[A: show, B: show](p: List[(term[A, B], term[A, B])]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('d', 'e', 'p', 'e', 'n', 'd', 'e',
     'n', 'c', 'y', ' ', 'p', 'a', 'i', 'r', 's', ':')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, shows_rules[A, B](p)))

implicit def equal_int: equal[BigInt] = new equal[BigInt] {
  val `Ceta.equal` = (a: BigInt, b: BigInt) => a == b
}

implicit def power_int: power[BigInt] = new power[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.one` = one_inta
}

def remdups[A: equal](x0: List[A]): List[A] = x0 match {
  case Nil => Nil
  case x :: xs =>
    (if (membera[A](xs, x)) remdups[A](xs) else x :: remdups[A](xs))
}

def it_add[A, B,
            C](update: A => B => C => C,
                iterate: (A => B => C => C) => C => C => C, m1: C, m2: C):
      C =
  ((iterate(update))(m2))(m1)

def vec0I[A](ze: A, n: nat): List[A] = replicate[A](n, ze)

def mat0I[A](ze: A, nr: nat, nc: nat): List[List[A]] =
  replicate[List[A]](nc, vec0I[A](ze, nr))

def minus_nat(m: nat, n: nat): nat = (m, n) match {
  case (Suc(m), Suc(n)) => minus_nat(m, n)
  case (Zero_nat, n) => Zero_nat
  case (m, Zero_nat) => m
}

def vec1I[A](ze: A, on: A, n: nat, i: nat): List[A] =
  replicate[A](i, ze) ++
    (on :: replicate[A](minus_nat(minus_nat(n, one_nat), i), ze))

def mat1I[A](ze: A, on: A, n: nat): List[List[A]] =
  mapa[nat, List[A]]((a: nat) => vec1I[A](ze, on, n, a), upt(Zero_nat, n))

def less_eq_prod_aux[A: equal: ord, B: ord](x0: (A, B), x1: (A, B)): Boolean =
  (x0, x1) match {
  case ((a1, a2), (b1, b2)) =>
    less[A](a1, b1) || eq[A](a1, b1) && less_eq[B](a2, b2)
}

def less_eq_prod[A: equal: ord, B: ord](a: (A, B), b: (A, B)): Boolean =
  less_eq_prod_aux[A, B](a, b)

def equal_proda[A: equal, B: equal](x0: (A, B), x1: (A, B)): Boolean = (x0, x1)
  match {
  case ((aa, ba), (a, b)) => eq[A](aa, a) && eq[B](ba, b)
}

def less_prod[A: equal: ord, B: equal: ord](a: (A, B), b: (A, B)): Boolean =
  ! (equal_proda[A, B](a, b)) && less_eq_prod_aux[A, B](a, b)

implicit def ord_prod[A: equal: ord, B: equal: ord]: ord[(A, B)] = new
  ord[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

implicit def equal_nat: equal[nat] = new equal[nat] {
  val `Ceta.equal` = (a: nat, b: nat) => equal_nata(a, b)
}

implicit def preorder_nat: preorder[nat] = new preorder[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

implicit def order_nat: order[nat] = new order[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

def equal_rata(a: rat, b: rat): Boolean =
  equal_proda[BigInt, BigInt](quotient_of(a), quotient_of(b))

implicit def equal_rat: equal[rat] = new equal[rat] {
  val `Ceta.equal` = (a: rat, b: rat) => equal_rata(a, b)
}

implicit def power_rat: power[rat] = new power[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.one` = one_rata
}

trait non_strict_order[A] {
  val `Ceta.ge`: (A, A) => Boolean
}
def ge[A](a: A, b: A)(implicit A: non_strict_order[A]): Boolean =
  A.`Ceta.ge`(a, b)

def digit2string(n: nat): List[Char] =
  (if (equal_nata(n, Zero_nat)) List('0')
    else (if (equal_nata(n, one_nat)) List('1')
           else (if (equal_nata(n, nat(BigInt(2)))) List('2')
                  else (if (equal_nata(n, nat(BigInt(3)))) List('3')
                         else (if (equal_nata(n, nat(BigInt(4)))) List('4')
                                else (if (equal_nata(n, nat(BigInt(5))))
                                       List('5')
                                       else (if (equal_nata(n, nat(BigInt(6))))
      List('6')
      else (if (equal_nata(n, nat(BigInt(7)))) List('7')
             else (if (equal_nata(n, nat(BigInt(8)))) List('8')
                    else (if (equal_nata(n, nat(BigInt(9)))) List('9')
                           else List('n', 'o', 't', ' ', 'a', ' ', 'd', 'i',
                                      'g', 'i', 't')))))))))))

def divmod_nat(m: nat, n: nat): (nat, nat) =
  (if (equal_nata(n, Zero_nat) || less_nat(m, n)) (Zero_nat, m)
    else {
           val (q, a): (nat, nat) = divmod_nat(minus_nat(m, n), n);
           (Suc(q), a)
         })

def div_nat(m: nat, n: nat): nat = fst[nat, nat](divmod_nat(m, n))

def mod_nat(m: nat, n: nat): nat = snd[nat, nat](divmod_nat(m, n))

def shows_nat(n: nat): (List[Char]) => List[Char] =
  (if (less_nat(n, nat(BigInt(10)))) shows_string.apply(digit2string(n))
    else comp[List[Char], List[Char],
               List[Char]](shows_nat(div_nat(n, nat(BigInt(10)))),
                            shows_string.apply(digit2string(mod_nat(n,
                             nat(BigInt(10)))))))

def shows_prec_nat(d: nat, n: nat): (List[Char]) => List[Char] = shows_nat(n)

def shows_prec_int(d: nat, i: BigInt): (List[Char]) => List[Char] =
  (if (i < BigInt(0))
    comp[List[Char], List[Char],
          List[Char]](shows_prec_char(Zero_nat, '-'),
                       shows_prec_nat(Zero_nat, nat(BigInt(0) - i)))
    else shows_prec_nat(Zero_nat, nat(i)))

def shows_list_aux[A](s: A => (List[Char]) => List[Char], xs: List[A]):
      (List[Char]) => List[Char] =
  shows_list_gen[A](s, List('[', ']'), List('['), List(','), List(']'), xs)

def shows_list_int(is: List[BigInt]): (List[Char]) => List[Char] =
  shows_list_aux[BigInt]((a: BigInt) => shows_prec_int(Zero_nat, a), is)

implicit def show_int: show[BigInt] = new show[BigInt] {
  val `Ceta.shows_prec` = (a: nat, b: BigInt, c: List[Char]) =>
    (shows_prec_int(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[BigInt], b: List[Char]) =>
    (shows_list_int(a)).apply(b)
}

def shows_list_nat(ns: List[nat]): (List[Char]) => List[Char] =
  shows_list_aux[nat]((a: nat) => shows_prec_nat(Zero_nat, a), ns)

implicit def show_nat: show[nat] = new show[nat] {
  val `Ceta.shows_prec` = (a: nat, b: nat, c: List[Char]) =>
    (shows_prec_nat(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[nat], b: List[Char]) =>
    (shows_list_nat(a)).apply(b)
}

def show_rata(x: rat): List[Char] =
  {
    val (den, num): (BigInt, BigInt) = quotient_of(x);
    (if (num == BigInt(1)) (shows_prec_int(Zero_nat, den)).apply(Nil)
      else (comp[List[Char], List[Char],
                  List[Char]](comp[List[Char], List[Char],
                                    List[Char]](shows_prec_int(Zero_nat, den),
         shows_prec_char(Zero_nat, '/')),
                               shows_prec_int(Zero_nat, num))).apply(Nil))
  }

def shows_prec_rat(d: nat, x: rat): (List[Char]) => List[Char] =
  shows_string.apply(show_rata(x))

def shows_list_rat(rs: List[rat]): (List[Char]) => List[Char] =
  shows_list_aux[rat]((a: rat) => shows_prec_rat(Zero_nat, a), rs)

implicit def show_rat: show[rat] = new show[rat] {
  val `Ceta.shows_prec` = (a: nat, b: rat, c: List[Char]) =>
    (shows_prec_rat(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[rat], b: List[Char]) =>
    (shows_list_rat(a)).apply(b)
}

def subt_at[A, B](s: term[A, B], x1: pos): term[A, B] = (s, x1) match {
  case (s, Empty) => s
  case (Fun(f, ss), PCons(i, p)) => subt_at[A, B](nth[term[A, B]](ss, i), p)
}

def the_var[A, B](x0: term[A, B]): B = x0 match {
  case Var(x) => x
}

def map_term[A, B, C](fun: A => B, x1: term[A, C]): term[B, C] = (fun, x1) match
  {
  case (fun, Var(x)) => Var[C, B](x)
  case (fun, Fun(f, ts)) =>
    Fun[B, C](fun(f),
               mapa[term[A, C],
                     term[B, C]]((a: term[A, C]) => map_term[A, B, C](fun, a),
                                  ts))
}

def map_rule[A, B, C](fg: A => B, lr: (term[A, C], term[A, C])):
      (term[B, C], term[B, C]) =
  (map_term[A, B, C](fg, fst[term[A, C], term[A, C]](lr)),
    map_term[A, B, C](fg, snd[term[A, C], term[A, C]](lr)))

abstract sealed class lab[+A, +B]
final case class Lab[A, B](a: lab[A, B], b: B) extends lab[A, B]
final case class FunLab[A, B](a: lab[A, B], b: List[lab[A, B]]) extends
  lab[A, B]
final case class UnLab[A, B](a: A) extends lab[A, B]
final case class Sharp[A, B](a: lab[A, B]) extends lab[A, B]

def maxa[A](less_eq: A => A => Boolean, a: A, b: A): A =
  (if ((less_eq(a))(b)) b else a)

def max[A: ord]: A => A => A =
  (a: A) => (b: A) => maxa[A]((aa: A) => (ba: A) => less_eq[A](aa, ba), a, b)

def max_list(x0: List[nat]): nat = x0 match {
  case Nil => Zero_nat
  case x :: xs => max[nat].apply(x).apply(max_list(xs))
}

def label_depth[A, B](x0: lab[A, B]): nat = x0 match {
  case UnLab(uu) => Zero_nat
  case Lab(f, uv) => Suc(label_depth[A, B](f))
  case FunLab(f, uw) => Suc(label_depth[A, B](f))
  case Sharp(f) => Suc(label_depth[A, B](f))
}

def equal_laba[A: equal, B: equal](x0: lab[A, B], x1: lab[A, B]): Boolean =
  (x0, x1) match {
  case (Sharp(lab), UnLab(f)) => false
  case (UnLab(f), Sharp(lab)) => false
  case (Sharp(laba), FunLab(lab, list)) => false
  case (FunLab(laba, list), Sharp(lab)) => false
  case (UnLab(f), FunLab(lab, list)) => false
  case (FunLab(lab, list), UnLab(f)) => false
  case (Sharp(laba), Lab(lab, l)) => false
  case (Lab(laba, l), Sharp(lab)) => false
  case (UnLab(f), Lab(lab, l)) => false
  case (Lab(lab, l), UnLab(f)) => false
  case (FunLab(laba, list), Lab(lab, l)) => false
  case (Lab(laba, l), FunLab(lab, list)) => false
  case (Sharp(laba), Sharp(lab)) => equal_laba[A, B](laba, lab)
  case (UnLab(fa), UnLab(f)) => eq[A](fa, f)
  case (FunLab(laba, lista), FunLab(lab, list)) =>
    equal_laba[A, B](laba, lab) && equal_lista[lab[A, B]](lista, list)
  case (Lab(laba, la), Lab(lab, l)) =>
    equal_laba[A, B](laba, lab) && eq[B](la, l)
}

implicit def equal_lab[A: equal, B: equal]: equal[lab[A, B]] = new
  equal[lab[A, B]] {
  val `Ceta.equal` = (a: lab[A, B], b: lab[A, B]) => equal_laba[A, B](a, b)
}

def gen_label[A, B](f: lab[A, B], x1: nat): lab[A, B] = (f, x1) match {
  case (f, Zero_nat) => f
  case (f, Suc(n)) => FunLab[A, B](gen_label[A, B](f, n), Nil)
}

def fmap[A: equal,
          B: equal](a: lab[A, B], nn: nat,
                     sml: List[((lab[A, B], nat), List[lab[A, B]])]):
      (lab[A, B]) => nat => lab[A, B] =
  {
    val m: nat =
      Suc(max_list(mapa[lab[A, B],
                         nat]((aa: lab[A, B]) => label_depth[A, B](aa),
                               a :: maps[((lab[A, B], nat), List[lab[A, B]]),
  lab[A, B]]((aa: ((lab[A, B], nat), List[lab[A, B]])) =>
               snd[(lab[A, B], nat), List[lab[A, B]]](aa),
              sml))));
    (f: lab[A, B]) =>
      (n: nat) =>
        (if (equal_proda[lab[A, B], nat]((f, n), (a, nn))) a
          else gen_label[A, B](f, m))
  }

def extract[A](p: A => Boolean, x1: List[A]): Option[(List[A], (A, List[A]))] =
  (p, x1) match {
  case (p, Nil) => None
  case (p, x :: xs) =>
    (if (p(x)) Some[(List[A], (A, List[A]))]((Nil, (x, xs)))
      else extract[A](p, xs) match {
             case None => None
             case Some((ys, (y, zs))) =>
               Some[(List[A], (A, List[A]))]((x :: ys, (y, zs)))
           })
}

def distinct[A: equal](x0: List[A]): Boolean = x0 match {
  case Nil => true
  case x :: xs => ! (membera[A](xs, x)) && distinct[A](xs)
}

def vec_comp_all[A](rel: A => A => Boolean, v: List[A], w: List[A]): Boolean =
  list_all[(A, A)]((a: (A, A)) => {
                                    val (aa, b): (A, A) = a;
                                    (rel(aa))(b)
                                  },
                    zip[A, A](v, w))

def mat_comp_all[A](rel: A => A => Boolean, m1: List[List[A]],
                     m2: List[List[A]]):
      Boolean =
  list_all[(List[A],
             List[A])]((a: (List[A], List[A])) =>
                         {
                           val (aa, b): (List[A], List[A]) = a;
                           vec_comp_all[A](rel, aa, b)
                         },
                        zip[List[A], List[A]](m1, m2))

def mat_ge[A: non_strict_order]: (List[List[A]]) => (List[List[A]]) => Boolean =
  (a: List[List[A]]) =>
    (b: List[List[A]]) =>
      mat_comp_all[A]((aa: A) => (ba: A) => ge[A](aa, ba), a, b)

def times_nat(x0: nat, n: nat): nat = (x0, n) match {
  case (Zero_nat, n) => Zero_nat
  case (Suc(m), n) => plus_nat(n, times_nat(m, n))
}

def mina[A](less_eq: A => A => Boolean, a: A, b: A): A =
  (if ((less_eq(a))(b)) a else b)

def min[A: ord]: A => A => A =
  (a: A) => (b: A) => mina[A]((aa: A) => (ba: A) => less_eq[A](aa, ba), a, b)

trait top[A] extends preorder[A] {
  val `Ceta.top`: A
}
def top[A](implicit A: top[A]): A = A.`Ceta.top`

def minus_rat(p: rat, q: rat): rat =
  Frct({
         val (a, c): (BigInt, BigInt) = quotient_of(p)
         val (b, d): (BigInt, BigInt) = quotient_of(q);
         normalize((a * d - b * c, c * d))
       })

def shows_list_list[A: show](ls: List[List[A]]): (List[Char]) => List[Char] =
  shows_list_aux[List[A]]((a: List[A]) => shows_prec_list[A](Zero_nat, a), ls)

implicit def show_list[A: show]: show[List[A]] = new show[List[A]] {
  val `Ceta.shows_prec` = (a: nat, b: List[A], c: List[Char]) =>
    (shows_prec_list[A](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[List[A]], b: List[Char]) =>
    (shows_list_list[A](a)).apply(b)
}

def shows_paren(p: (List[Char]) => List[Char]): (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_prec_char(Zero_nat, '('),
                     comp[List[Char], List[Char],
                           List[Char]](p, shows_prec_char(Zero_nat, ')')))

def shows_prec_prod[A: show, B: show](d: nat, p: (A, B)):
      (List[Char]) => List[Char] =
  shows_paren(comp[List[Char], List[Char],
                    List[Char]]((a: List[Char]) =>
                                  shows_prec[A](Zero_nat, fst[A, B](p), a),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_prec_char(Zero_nat,
                            ','),
            (a: List[Char]) => shows_prec[B](Zero_nat, snd[A, B](p), a))))

def shows_list_prod[A: show, B: show](ps: List[(A, B)]):
      (List[Char]) => List[Char] =
  shows_list_aux[(A, B)]((a: (A, B)) => shows_prec_prod[A, B](Zero_nat, a), ps)

implicit def show_prod[A: show, B: show]: show[(A, B)] = new show[(A, B)] {
  val `Ceta.shows_prec` = (a: nat, b: (A, B), c: List[Char]) =>
    (shows_prec_prod[A, B](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[(A, B)], b: List[Char]) =>
    (shows_list_prod[A, B](a)).apply(b)
}

def map_vars[A, B, C](m: A => B, x1: term[C, A]): term[C, B] = (m, x1) match {
  case (m, Var(x)) => Var[B, C](m(x))
  case (m, Fun(f, ts)) =>
    Fun[C, B](f, mapa[term[C, A],
                       term[C, B]]((a: term[C, A]) => map_vars[A, B, C](m, a),
                                    ts))
}

def num_args[A, B](x0: term[A, B]): nat = x0 match {
  case Var(x) => Zero_nat
  case Fun(f, ts) => size_list[term[A, B]](ts)
}

def shows_list_term[A: show, B: show](is: List[term[A, B]]):
      (List[Char]) => List[Char] =
  shows_list_aux[term[A, B]]((a: term[A, B]) =>
                               shows_prec_term[A, B](Zero_nat, a),
                              is)

implicit def show_term[A: show, B: show]: show[term[A, B]] = new
  show[term[A, B]] {
  val `Ceta.shows_prec` = (a: nat, b: term[A, B], c: List[Char]) =>
    (shows_prec_term[A, B](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[term[A, B]], b: List[Char]) =>
    (shows_list_term[A, B](a)).apply(b)
}

def shows_trs[A: show, B: show](r: List[(term[A, B], term[A, B])]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('r', 'e', 'w', 'r', 'i', 't', 'e',
     ' ', 's', 'y', 's', 't', 'e', 'm', ':')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, shows_rules[A, B](r)))

def shows_dpp[A: show,
               B: show](p: List[(term[A, B], term[A, B])],
                         r: List[(term[A, B], term[A, B])]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char], List[Char]](shows_p[A, B](p), shows_trs[A, B](r))

def unapp[A: equal, B](a: A, x1: term[A, B]): (term[A, B], List[term[A, B]]) =
  (a, x1) match {
  case (a, Var(x)) => (Var[B, A](x), Nil)
  case (a, Fun(f, ss)) =>
    (if (eq[A](f, a) && equal_nata(size_list[term[A, B]](ss), nat(BigInt(2))))
      {
        val (r, ts): (term[A, B], List[term[A, B]]) =
          unapp[A, B](a, nth[term[A, B]](ss, Zero_nat));
        (r, ts ++ List(nth[term[A, B]](ss, one_nat)))
      }
      else (Fun[A, B](f, ss), Nil))
}

abstract sealed class af_entry
final case class Collapse(a: nat) extends af_entry
final case class AFList(a: List[nat]) extends af_entry

def xml_many_elements[A, B](tag: List[Char], p: xml => sum[List[Char], A],
                             f: (List[A]) => B, x3: xml):
      sum[List[Char], B] =
  (tag, p, f, x3) match {
  case (tag, p, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) && is_none[List[Char]](text)))
      binda[List[Char], List[A],
             B](map[A].apply(p).apply(cs),
                 comp[B, sum[List[Char], B],
                       List[A]]((a: B) => Inr[B, List[Char]](a), f))
      else no_transform[B](tag, Xml(name, atts, cs, text)))
}

def xml3elements[A, B, C,
                  D](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B],
                      p3: xml => sum[List[Char], C], f: A => B => C => D,
                      x5: xml):
      sum[List[Char], D] =
  (tag, p1, p2, p3, f, x5) match {
  case (tag, p1, p2, p3, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(3))) &&
               is_none[List[Char]](text))))
      binda[List[Char], A,
             D](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   binda[List[Char], B,
                          D](p2(nth[xml](cs, one_nat)),
                              (x2: B) =>
                                binda[List[Char], C,
                                       D](p3(nth[xml](cs, nat(BigInt(2)))),
   (x3: C) => Inr[D, List[Char]](((f(x1))(x2))(x3)))))
      else no_transform[D](tag, Xml(name, atts, cs, text)))
}

def position: xml => sum[List[Char], nat] =
  (a: xml) =>
    xml_change[nat, nat]((aa: xml) =>
                           nata(List('p', 'o', 's', 'i', 't', 'i', 'o', 'n'),
                                 aa),
                          (n: nat) => minus_nat(n, one_nat), a)

def afs[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[((A, nat), af_entry)]] =
  (a: xml) =>
    xml_many_elements[((A, nat), af_entry),
                       List[((A, nat),
                              af_entry)]](List('a', 'r', 'g', 'u', 'm', 'e',
        'n', 't', 'F', 'i', 'l', 't', 'e', 'r'),
   (aa: xml) =>
     xml3elements[A, nat, af_entry,
                   ((A, nat),
                     af_entry)](List('a', 'r', 'g', 'u', 'm', 'e', 'n', 't',
                                      'F', 'i', 'l', 't', 'e', 'r', 'E', 'n',
                                      't', 'r', 'y'),
                                 xml2name,
                                 (ab: xml) =>
                                   nata(List('a', 'r', 'i', 't', 'y'), ab),
                                 (ab: xml) =>
                                   xml_options[af_entry](List((List('c', 'o',
                             'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'),
                        (ac: xml) =>
                          xml_change[nat, af_entry]((ad: xml) =>
              nata(List('c', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'), ad),
             (i: nat) => Collapse(minus_nat(i, one_nat)), ac)),
                       (List('n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's',
                              'i', 'n', 'g'),
                         (ac: xml) =>
                           xml_many_elements[nat,
      af_entry](List('n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n',
                      'g'),
                 position, (ad: List[nat]) => AFList(ad), ac))),
                  ab),
                                 (f: A) =>
                                   (ab: nat) => (b: af_entry) => ((f, ab), b),
                                 aa),
   id[List[((A, nat), af_entry)]], a)

def list2position(x0: List[nat]): pos = x0 match {
  case Nil => Empty
  case n :: ns => PCons(n, list2position(ns))
}

def pos: xml => sum[List[Char], pos] =
  (a: xml) =>
    xml_many_elements[nat, pos](List('p', 'o', 's', 'i', 't', 'i', 'o', 'n',
                                      'I', 'n', 'T', 'e', 'r', 'm'),
                                 position, (aa: List[nat]) => list2position(aa),
                                 a)

def xml1many_elements_gen[A, B,
                           C](tag: List[Char], p1: xml => sum[List[Char], A],
                               p2: A => xml => sum[List[Char], B],
                               f: A => (List[B]) => C, x4: xml):
      sum[List[Char], C] =
  (tag, p1, p2, f, x4) match {
  case (tag, p1, p2, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (is_none[List[Char]](text) && ! (nulla[xml](cs)))))
      binda[List[Char], A,
             C](p1(hd[xml](cs)),
                 (x: A) =>
                   binda[List[Char], List[B],
                          C](map[B].apply(p2(x)).apply(tl[xml](cs)),
                              (xs: List[B]) => Inr[C, List[Char]]((f(x))(xs))))
      else no_transform[C](tag, Xml(name, atts, cs, text)))
}

def xml1many_elements[A, B,
                       C](tag: List[Char], p1: xml => sum[List[Char], A],
                           p2: xml => sum[List[Char], B]):
      (A => (List[B]) => C) => xml => sum[List[Char], C] =
  (a: A => (List[B]) => C) =>
    (b: xml) => xml1many_elements_gen[A, B, C](tag, p1, (_: A) => p2, a, b)

def xml1element[A, B](tag: List[Char], p1: xml => sum[List[Char], A], f: A => B,
                       x3: xml):
      sum[List[Char], B] =
  (tag, p1, f, x3) match {
  case (tag, p1, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), one_nat) &&
               is_none[List[Char]](text))))
      binda[List[Char], A,
             B](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) => Inr[B, List[Char]](f(x1)))
      else no_transform[B](tag, Xml(name, atts, cs, text)))
}

def vara[A](x: xml): sum[List[Char], term[A, List[Char]]] =
  binda[List[Char], List[Char],
         term[A, List[Char]]](text(List('v', 'a', 'r'), x),
                               comp[term[A, List[Char]],
                                     sum[List[Char], term[A, List[Char]]],
                                     List[Char]]((a: term[A, List[Char]]) =>
           Inr[term[A, List[Char]], List[Char]](a),
          (a: List[Char]) => Var[List[Char], A](a)))

def term[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], term[A, List[Char]]] =
  xml_options[term[A, List[Char]]](List((List('v', 'a', 'r'),
  (a: xml) => vara[A](a)),
 (List('f', 'u', 'n', 'a', 'p', 'p'),
   (xml1many_elements[A, term[A, List[Char]],
                       term[A, List[Char]]](List('f', 'u', 'n', 'a', 'p', 'p'),
     xml2name,
     (a: xml) =>
       xml1element[term[A, List[Char]],
                    term[A, List[Char]]](List('a', 'r', 'g'),
  (aa: xml) => term[A](xml2name, aa), id[term[A, List[Char]]],
  a))).apply((a: A) =>
               (b: List[term[A, List[Char]]]) => Fun[A, List[Char]](a, b)))),
                                    x)

def rule[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], (term[A, List[Char]], term[A, List[Char]])] =
  (a: xml) =>
    xml2elements[term[A, List[Char]], term[A, List[Char]],
                  (term[A, List[Char]],
                    term[A, List[Char]])](List('r', 'u', 'l', 'e'),
   (aa: xml) =>
     xml1element[term[A, List[Char]],
                  term[A, List[Char]]](List('l', 'h', 's'),
(ab: xml) => term[A](xml2name, ab), id[term[A, List[Char]]], aa),
   (aa: xml) =>
     xml1element[term[A, List[Char]],
                  term[A, List[Char]]](List('r', 'h', 's'),
(ab: xml) => term[A](xml2name, ab), id[term[A, List[Char]]], aa),
   (aa: term[A, List[Char]]) => (b: term[A, List[Char]]) => (aa, b), a)

def scg_position: xml => sum[List[Char], nat] =
  (a: xml) => nata(List('p', 'o', 's', 'i', 't', 'i', 'o', 'n'), a)

def map_filter[A, B](f: A => Option[B], x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x :: xs) =>
    f(x) match {
      case None => map_filter[A, B](f, xs)
      case Some(y) => y :: map_filter[A, B](f, xs)
    }
}

def scg[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             ((term[A, List[Char]], term[A, List[Char]]),
               (List[(nat, nat)], List[(nat, nat)]))] =
  (xml1many_elements[(term[A, List[Char]], term[A, List[Char]]),
                      (Boolean, (nat, nat)),
                      ((term[A, List[Char]], term[A, List[Char]]),
                        (List[(nat, nat)],
                          List[(nat, nat)]))](List('s', 'i', 'z', 'e', 'C', 'h',
            'a', 'n', 'g', 'e', 'G', 'r', 'a', 'p', 'h'),
       rule[A](xml2name),
       (a: xml) =>
         xml3elements[nat, Boolean, nat,
                       (Boolean,
                         (nat, nat))](List('e', 'd', 'g', 'e'), scg_position,
                                       (aa: xml) =>
 bool(List('s', 't', 'r', 'i', 'c', 't'), aa),
                                       scg_position,
                                       (p: nat) =>
 (s: Boolean) => (q: nat) => (s, (p, q)),
                                       a))).apply((lr:
             (term[A, List[Char]], term[A, List[Char]]))
            => (edges: List[(Boolean, (nat, nat))]) =>
                 (lr, (map_filter[(Boolean, (nat, nat)),
                                   (nat, nat)]((x: (Boolean, (nat, nat))) =>
         (if (fst[Boolean, (nat, nat)](x))
           Some[(nat, nat)](snd[Boolean, (nat, nat)](x)) else None),
        edges),
                        map_filter[(Boolean, (nat, nat)),
                                    (nat, nat)]((x: (Boolean, (nat, nat))) =>
          (if (! (fst[Boolean, (nat, nat)](x)))
            Some[(nat, nat)](snd[Boolean, (nat, nat)](x)) else None),
         edges))))

def equal_hterma[A: equal, B](x0: hterm[A, B], x1: hterm[A, B]): Boolean =
  (x0, x1) match {
  case (HFun(f, list), HHole) => false
  case (HHole, HFun(f, list)) => false
  case (HFun(fa, lista), HFun(f, list)) =>
    eq[A](fa, f) && equal_lista[hterm[A, B]](lista, list)
  case (HHole, HHole) => true
}

implicit def equal_hterm[A: equal, B]: equal[hterm[A, B]] = new
  equal[hterm[A, B]] {
  val `Ceta.equal` = (a: hterm[A, B], b: hterm[A, B]) =>
    equal_hterma[A, B](a, b)
}

def partition[A](p: A => Boolean, x1: List[A]): (List[A], List[A]) = (p, x1)
  match {
  case (p, Nil) => (Nil, Nil)
  case (p, x :: xs) =>
    {
      val (yes, no): (List[A], List[A]) = partition[A](p, xs);
      (if (p(x)) (x :: yes, no) else (yes, x :: no))
    }
}

def vec_pre_gtI[A](gt: A => A => Boolean, v: List[A], w: List[A]): Boolean =
  list_ex[(A, A)]((a: (A, A)) => {
                                   val (aa, b): (A, A) = a;
                                   (gt(aa))(b)
                                 },
                   zip[A, A](v, w))

def mat_pre_gtI[A](gt: A => A => Boolean, m1: List[List[A]], m2: List[List[A]]):
      Boolean =
  list_ex[(List[A],
            List[A])]((a: (List[A], List[A])) =>
                        {
                          val (aa, b): (List[A], List[A]) = a;
                          vec_pre_gtI[A](gt, aa, b)
                        },
                       zip[List[A], List[A]](m1, m2))

def sub_vec[A]: nat => (List[A]) => List[A] =
  (a: nat) => (b: List[A]) => take[A](a, b)

def sub_mat[A](nr: nat, nc: nat, m: List[List[A]]): List[List[A]] =
  mapa[List[A], List[A]](sub_vec[A].apply(nr), take[List[A]](nc, m))

def mat_gtI[A: non_strict_order](gt: A => A => Boolean, sd: nat,
                                  m1: List[List[A]], m2: List[List[A]]):
      Boolean =
  mat_ge[A].apply(m1).apply(m2) &&
    mat_pre_gtI[A](gt, sub_mat[A](sd, sd, m1), sub_mat[A](sd, sd, m2))

def vec_map[A]: (A => A) => (List[A]) => List[A] =
  (a: A => A) => (b: List[A]) => mapa[A, A](a, b)

def mat_map[A](f: A => A): (List[List[A]]) => List[List[A]] =
  (a: List[List[A]]) => mapa[List[A], List[A]](vec_map[A].apply(f), a)

implicit def preorder_prod[A: equal: order, B: equal: order]: preorder[(A, B)] =
  new preorder[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

implicit def order_prod[A: equal: order, B: equal: order]: order[(A, B)] = new
  order[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

def check[A](b: Boolean, e: A): sum[A, Unit] =
  (if (b) Inr[Unit, A](()) else Inl[A, Unit](e))

def mpo_strict[A: equal: show, B: equal: show](pr: A => nat => nat, n: nat):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](fst[Boolean,
                               Boolean](mpo[A,
     B].apply((prc_nat[A](pr), prl_nat[A](pr))).apply(n).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_term[A, B](Zero_nat, s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            'M', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_prec_term[A, B](Zero_nat, t), shows_nl)))))
    }

trait semigroup_add[A] extends plus[A] {
}

trait ab_semigroup_add[A] extends semigroup_add[A] {
}

trait semigroup_mult[A] extends times[A] {
}

trait semiring[A] extends ab_semigroup_add[A] with semigroup_mult[A] {
}

trait monoid_add[A] extends semigroup_add[A] with zero[A] {
}

trait ordered_ab_semigroup[A]
  extends ab_semigroup_add[A] with non_strict_order[A] {
}

trait max_ordered_monoid_add[A]
  extends monoid_add[A] with ordered_ab_semigroup[A] {
  val `Ceta.max0`: A => A
}
def max0[A](a: A)(implicit A: max_ordered_monoid_add[A]): A = A.`Ceta.max0`(a)

def s_ins[A, B, C](update: A => Unit => B => C, x: A, s: B): C =
  ((update(x))(()))(s)

def poss_list[A, B](x0: term[A, B]): List[pos] = x0 match {
  case Var(x) => List(Empty)
  case Fun(f, ss) =>
    Empty ::
      maps[(nat, List[pos]),
            pos]((a: (nat, List[pos])) =>
                   {
                     val (i, aa): (nat, List[pos]) = a;
                     mapa[pos, pos]((ab: pos) => PCons(i, ab), aa)
                   },
                  zip[nat, List[pos]](upt(Zero_nat, size_list[term[A, B]](ss)),
                                       mapa[term[A, B],
     List[pos]]((a: term[A, B]) => poss_list[A, B](a), ss)))
}

def sharp_term[A, B](shp: A => A, x1: term[A, B]): term[A, B] = (shp, x1) match
  {
  case (shp, Var(x)) => Var[B, A](x)
  case (shp, Fun(f, ss)) => Fun[A, B](shp(f), ss)
}

def aarity[A](sm: A => nat => List[A], f: A, n: nat): nat =
  minus_nat(size_list[A]((sm(f))(n)), one_nat)

def list_diff[A: equal](x0: List[A], ys: List[A]): List[A] = (x0, ys) match {
  case (Nil, ys) => Nil
  case (x :: xs, ys) =>
    {
      val zs: List[A] = list_diff[A](xs, ys);
      (if (membera[A](ys, x)) zs else x :: zs)
    }
}

def min_list[A: linorder](x0: List[A]): A = x0 match {
  case List(x) => x
  case x :: v :: va => min[A].apply(x).apply(min_list[A](v :: va))
}

def quote(s: (List[Char]) => List[Char]): (List[Char]) => List[Char] =
  shows_between(shows_prec_char(Zero_nat, '\''), s,
                 shows_prec_char(Zero_nat, '\''))

def expecting[A: show, B](msg: List[Char], ts: List[A]):
      sum[List[Char], (B, List[A])] =
  Inl[List[Char],
       (B, List[A])](List('e', 'x', 'p', 'e', 'c', 't', 'i', 'n', 'g', ':',
                           ' ') ++
                       (msg ++
                         (List(',', ' ', 'b', 'u', 't', ' ', 'f', 'o', 'u', 'n',
                                'd', ':', ' ') ++
                           (quote(shows_prec_list[A](Zero_nat,
              take[A](nat(BigInt(30)), ts)))).apply(Nil))))

def eoi[A: show](x0: List[A]): sum[List[Char], (Unit, List[A])] = x0 match {
  case Nil => Inr[(Unit, List[A]), List[Char]](((), Nil))
  case v :: va =>
    expecting[A, Unit](List('e', 'n', 'd', ' ', 'o', 'f', ' ', 'i', 'n', 'p',
                             'u', 't'),
                        v :: va)
}

def xml_choice[A](e: List[Char], xa1: List[xml => sum[List[Char], A]], x: xml):
      sum[List[Char], A] =
  (e, xa1, x) match {
  case (e, Nil, x) =>
    Inl[List[Char],
         A](concat[Char](List(List('e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                                    'p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 'c',
                                    'h', 'o', 'i', 'c', 'e', ' ', 'f', 'o', 'r',
                                    ' '),
                               e, List(newline),
                               (shows_prec_xml(Zero_nat, x)).apply(Nil))))
  case (e, p :: ps, x) =>
    p(x) match {
      case Inl(_) => xml_choice[A](e, ps, x)
      case Inr(a) => Inr[A, List[Char]](a)
    }
}

abstract sealed class alter[+A, +B, +C, +D, +E, +F, +G, +H, +I, +J]
final case class Alter1[A, B, C, D, E, F, G, H, I, J](a: A) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter2[B, A, C, D, E, F, G, H, I, J](a: B) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter3[C, A, B, D, E, F, G, H, I, J](a: C) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter4[D, A, B, C, E, F, G, H, I, J](a: D) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter5[E, A, B, C, D, F, G, H, I, J](a: E) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter6[F, A, B, C, D, E, G, H, I, J](a: F) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter7[G, A, B, C, D, E, F, H, I, J](a: G) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter8[H, A, B, C, D, E, F, G, I, J](a: H) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter9[I, A, B, C, D, E, F, G, H, J](a: I) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter10[J, A, B, C, D, E, F, G, H, I](a: J) extends
  alter[A, B, C, D, E, F, G, H, I, J]

def xml0elements[A](tag: List[Char], x: A, xa2: xml): sum[List[Char], A] =
  (tag, x, xa2) match {
  case (tag, x, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (nulla[xml](cs) && is_none[List[Char]](text))))
      Inr[A, List[Char]](x)
      else no_transform[A](tag, Xml(name, atts, cs, text)))
}

def xml4elements[A, B, C, D,
                  E](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B],
                      p3: xml => sum[List[Char], C],
                      p4: xml => sum[List[Char], D], f: A => B => C => D => E,
                      x6: xml):
      sum[List[Char], E] =
  (tag, p1, p2, p3, p4, f, x6) match {
  case (tag, p1, p2, p3, p4, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(4))) &&
               is_none[List[Char]](text))))
      binda[List[Char], A,
             E](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   binda[List[Char], B,
                          E](p2(nth[xml](cs, one_nat)),
                              (x2: B) =>
                                binda[List[Char], C,
                                       E](p3(nth[xml](cs, nat(BigInt(2)))),
   (x3: C) =>
     binda[List[Char], D,
            E](p4(nth[xml](cs, nat(BigInt(3)))),
                (x4: D) => Inr[E, List[Char]]((((f(x1))(x2))(x3))(x4))))))
      else no_transform[E](tag, Xml(name, atts, cs, text)))
}

def ctxt[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], ctxt[A, List[Char]]] =
  xml_options[ctxt[A, List[Char]]](List((List('b', 'o', 'x'),
  (a: xml) => xml0elements[ctxt[A, List[Char]]](List('b', 'o', 'x'), Hole, a)),
 (List('f', 'u', 'n', 'C', 'o', 'n', 't', 'e', 'x', 't'),
   (a: xml) =>
     xml4elements[A, List[term[A, List[Char]]], ctxt[A, List[Char]],
                   List[term[A, List[Char]]],
                   ctxt[A, List[Char]]](List('f', 'u', 'n', 'C', 'o', 'n', 't',
      'e', 'x', 't'),
 xml2name,
 (aa: xml) =>
   xml_many_elements[term[A, List[Char]],
                      List[term[A, List[Char]]]](List('b', 'e', 'f', 'o', 'r',
               'e'),
          (ab: xml) => term[A](xml2name, ab), id[List[term[A, List[Char]]]],
          aa),
 (aa: xml) => ctxt[A](xml2name, aa),
 (aa: xml) =>
   xml_many_elements[term[A, List[Char]],
                      List[term[A, List[Char]]]](List('a', 'f', 't', 'e', 'r'),
          (ab: xml) => term[A](xml2name, ab), id[List[term[A, List[Char]]]],
          aa),
 (aa: A) =>
   (b: List[term[A, List[Char]]]) =>
     (c: ctxt[A, List[Char]]) =>
       (d: List[term[A, List[Char]]]) => More[A, List[Char]](aa, b, c, d),
 a))),
                                    x)

def xml3to4elements[A, B, C, D,
                     E](tag: List[Char], p1: xml => sum[List[Char], A],
                         p2: xml => sum[List[Char], B],
                         p3: xml => sum[List[Char], C],
                         p4: xml => sum[List[Char], D],
                         f: A => B => Option[C] => D => E, x6: xml):
      sum[List[Char], E] =
  (tag, p1, p2, p3, p4, f, x6) match {
  case (tag, p1, p2, p3, p4, f, Xml(name, atts, cs, text)) =>
    {
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (less_eq_nat(nat(BigInt(3)), l) &&
                 (less_eq_nat(l, nat(BigInt(4))) &&
                   is_none[List[Char]](text)))))
        binda[List[Char], A,
               E](p1(nth[xml](cs, Zero_nat)),
                   (x1: A) =>
                     binda[List[Char], B,
                            E](p2(nth[xml](cs, one_nat)),
                                (x2: B) =>
                                  (if (equal_nata(l, nat(BigInt(4))))
                                    binda[List[Char], C,
   E](p3(nth[xml](cs, nat(BigInt(2)))),
       (x3: C) =>
         binda[List[Char], D,
                E](p4(nth[xml](cs, nat(BigInt(3)))),
                    (x4: D) =>
                      Inr[E, List[Char]]((((f(x1))(x2))(Some[C](x3)))(x4))))
                                    else binda[List[Char], D,
        E](p4(nth[xml](cs, nat(BigInt(2)))),
            (x4: D) => Inr[E, List[Char]]((((f(x1))(x2))(None))(x4))))))
        else no_transform[E](tag, Xml(name, atts, cs, text)))
    }
}

def rstep[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (pos, ((term[A, List[Char]], term[A, List[Char]]),
                     (Boolean, term[A, List[Char]])))] =
  (a: xml) =>
    xml3to4elements[pos, (term[A, List[Char]], term[A, List[Char]]), Unit,
                     term[A, List[Char]],
                     (pos, ((term[A, List[Char]], term[A, List[Char]]),
                             (Boolean,
                               term[A, List[Char]])))](List('r', 'e', 'w', 'r',
                     'i', 't', 'e', 'S', 't', 'e', 'p'),
                pos, rule[A](xml2name),
                (aa: xml) =>
                  xml0elements[Unit](List('r', 'e', 'l', 'a', 't', 'i', 'v',
   'e'),
                                      (), aa),
                (aa: xml) => term[A](xml2name, aa),
                (p: pos) =>
                  (r: (term[A, List[Char]], term[A, List[Char]])) =>
                    (rel: Option[Unit]) =>
                      (t: term[A, List[Char]]) =>
                        (p, (r, (is_none[Unit](rel), t))),
                a)

def rsteps[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (term[A, List[Char]],
               List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                            (Boolean, term[A, List[Char]])))])] =
  (xml1many_elements[term[A, List[Char]],
                      (pos, ((term[A, List[Char]], term[A, List[Char]]),
                              (Boolean, term[A, List[Char]]))),
                      (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     (Boolean,
                                       term[A,
     List[Char]])))])](List('r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q',
                             'u', 'e', 'n', 'c', 'e'),
                        (a: xml) =>
                          xml1element[term[A, List[Char]],
                                       term[A,
     List[Char]]](List('s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'),
                   (aa: xml) => term[A](xml2name, aa), id[term[A, List[Char]]],
                   a),
                        rstep[A](xml2name))).apply((a: term[A, List[Char]]) =>
             (b: List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                              (Boolean, term[A, List[Char]])))])
               => (a, b))

def subst[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[(List[Char], term[A, List[Char]])]] =
  (a: xml) =>
    xml_many_elements[(List[Char], term[A, List[Char]]),
                       List[(List[Char],
                              term[A, List[Char]])]](List('s', 'u', 'b', 's',
                   't', 'i', 't', 'u', 't', 'i', 'o', 'n'),
              (aa: xml) =>
                xml2elements[List[Char], term[A, List[Char]],
                              (List[Char],
                                term[A, List[Char]])](List('s', 'u', 'b', 's',
                    't', 'E', 'n', 't', 'r', 'y'),
               (ab: xml) => text(List('v', 'a', 'r'), ab),
               (ab: xml) => term[A](xml2name, ab),
               (ab: List[Char]) => (b: term[A, List[Char]]) => (ab, b), aa),
              id[List[(List[Char], term[A, List[Char]])]], a)

def loop[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (term[A, List[Char]],
               (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                             (Boolean, term[A, List[Char]])))],
                 (List[(List[Char], term[A, List[Char]])],
                   ctxt[A, List[Char]])))] =
  (a: xml) =>
    xml3elements[(term[A, List[Char]],
                   List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                (Boolean, term[A, List[Char]])))]),
                  List[(List[Char], term[A, List[Char]])], ctxt[A, List[Char]],
                  (term[A, List[Char]],
                    (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                  (Boolean, term[A, List[Char]])))],
                      (List[(List[Char], term[A, List[Char]])],
                        ctxt[A, List[Char]])))](List('l', 'o', 'o', 'p'),
         rsteps[A](xml2name), subst[A](xml2name),
         (aa: xml) => ctxt[A](xml2name, aa),
         (aa: (term[A, List[Char]],
                List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                             (Boolean, term[A, List[Char]])))]))
           => {
                val (s, rseq):
                      (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     (Boolean, term[A, List[Char]])))])
                  = aa;
                (sigma: List[(List[Char], term[A, List[Char]])]) =>
                  (c: ctxt[A, List[Char]]) => (s, (rseq, (sigma, c)))
              },
         a)

abstract sealed class projL[+A]
final case class Projection[A](a: List[(A, nat)]) extends projL[A]

def proj[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], projL[A]] =
  (a: xml) =>
    xml_change[List[((A, nat), af_entry)],
                projL[A]](afs[A](xml2name),
                           (afl: List[((A, nat), af_entry)]) =>
                             Projection[A](mapa[((A, nat), af_entry),
         (A, nat)]((aa: ((A, nat), af_entry)) =>
                     {
                       val (ab, b): ((A, nat), af_entry) = aa;
                       ({
                          val (f, ac): (A, nat) = ab;
                          (ba: af_entry) =>
                            ba match {
                              case Collapse(ad) => (f, ad)
                              case AFList(_) => (f, ac)
                            }
                        })(b)
                     },
                    afl)),
                           a)

def start_term[A: equal](xml2name: xml => sum[List[Char], A],
                          t: term[A, List[Char]]):
      xml => sum[List[Char], term[A, List[Char]]] =
  (a: xml) =>
    xml1element[term[A, List[Char]],
                 term[A, List[Char]]](List('s', 't', 'a', 'r', 't', 'T', 'e',
    'r', 'm'),
                                       (xml: xml) =>
 binda[List[Char], term[A, List[Char]],
        term[A, List[Char]]](term[A](xml2name, xml),
                              (s: term[A, List[Char]]) =>
                                (if (equal_terma[A, List[Char]](s, t))
                                  Inr[term[A, List[Char]], List[Char]](t)
                                  else Inl[List[Char],
    term[A, List[Char]]](List('<', 's', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm',
                               '>', ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't',
                               ' ', 'm', 'a', 't', 'c', 'h', ' ', 'l', 'h',
                               's')))),
                                       id[term[A, List[Char]]], a)

def proj_term[A, B](p: A => nat, x1: term[A, B]): term[A, B] = (p, x1) match {
  case (p, Var(x)) => Var[B, A](x)
  case (p, Fun(f, ts)) =>
    (if (less_nat(p(f), size_list[term[A, B]](ts))) nth[term[A, B]](ts, p(f))
      else Fun[A, B](f, ts))
}

def rseq[A: equal](xml2name: xml => sum[List[Char], A], pi: A => nat,
                    r: (term[A, List[Char]], term[A, List[Char]])):
      xml =>
        sum[List[Char],
             ((term[A, List[Char]], term[A, List[Char]]),
               List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                            term[A, List[Char]]))])] =
  (xml1many_elements[term[A, List[Char]],
                      (pos, ((term[A, List[Char]], term[A, List[Char]]),
                              (Boolean, term[A, List[Char]]))),
                      ((term[A, List[Char]], term[A, List[Char]]),
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     term[A,
   List[Char]]))])](List('r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u',
                          'e', 'n', 'c', 'e'),
                     start_term[A](xml2name,
                                    proj_term[A,
       List[Char]](pi, fst[term[A, List[Char]], term[A, List[Char]]](r))),
                     rstep[A](xml2name))).apply((_: term[A, List[Char]]) =>
          (rseq: List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                              (Boolean, term[A, List[Char]])))])
            => (r, mapa[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                (Boolean, term[A, List[Char]]))),
                         (pos, ((term[A, List[Char]], term[A, List[Char]]),
                                 term[A, List[Char]]))]((a:
                   (pos, ((term[A, List[Char]], term[A, List[Char]]),
                           (Boolean, term[A, List[Char]]))))
                  => {
                       val (x, (y, (_, z))):
                             (pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     (Boolean, term[A, List[Char]])))
                         = a;
                       (x, (y, z))
                     },
                 rseq)))

implicit def semigroup_add_int: semigroup_add[BigInt] = new
  semigroup_add[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def ab_semigroup_add_int: ab_semigroup_add[BigInt] = new
  ab_semigroup_add[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def semigroup_mult_int: semigroup_mult[BigInt] = new
  semigroup_mult[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def semiring_int: semiring[BigInt] = new semiring[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

def mat_max0[A: max_ordered_monoid_add]: (List[List[A]]) => List[List[A]] =
  mat_map[A]((a: A) => max0[A](a))

implicit def linorder_nat: linorder[nat] = new linorder[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

abstract sealed class tpoly[+A, +B]
final case class PVar[A, B](a: A) extends tpoly[A, B]
final case class PNum[B, A](a: B) extends tpoly[A, B]
final case class PSum[A, B](a: List[tpoly[A, B]]) extends tpoly[A, B]
final case class PMult[A, B](a: List[tpoly[A, B]]) extends tpoly[A, B]

def mpo_nstrict[A: equal: show, B: equal: show](pr: A => nat => nat, n: nat):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](snd[Boolean,
                               Boolean](mpo[A,
     B].apply((prc_nat[A](pr), prl_nat[A](pr))).apply(n).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_term[A, B](Zero_nat, s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            '=', 'M', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_prec_term[A, B](Zero_nat, t), shows_nl)))))
    }

implicit def semigroup_add_rat: semigroup_add[rat] = new semigroup_add[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def ab_semigroup_add_rat: ab_semigroup_add[rat] = new
  ab_semigroup_add[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def semigroup_mult_rat: semigroup_mult[rat] = new semigroup_mult[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def semiring_rat: semiring[rat] = new semiring[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

trait mult_zero[A] extends times[A] with zero[A] {
}

def s_memb[A, B, C](lookup: A => B => Option[C], x: A, s: B): Boolean =
  ! (is_none[C]((lookup(x))(s)))

def ctxt_apply[A, B](x0: ctxt[A, B], s: term[A, B]): term[A, B] = (x0, s) match
  {
  case (Hole, s) => s
  case (More(f, ss1, c, ss2), s) =>
    Fun[A, B](f, ss1 ++ (ctxt_apply[A, B](c, s) :: ss2))
}

def equal_ctxt[A: equal, B: equal](x0: ctxt[A, B], x1: ctxt[A, B]): Boolean =
  (x0, x1) match {
  case (More(f, list1, ctxt, list2), Hole) => false
  case (Hole, More(f, list1, ctxt, list2)) => false
  case (More(fa, list1a, ctxta, list2a), More(f, list1, ctxt, list2)) =>
    eq[A](fa, f) &&
      (equal_lista[term[A, B]](list1a, list1) &&
        (equal_ctxt[A, B](ctxta, ctxt) &&
          equal_lista[term[A, B]](list2a, list2)))
  case (Hole, Hole) => true
}

def shows_ctxt[A: show, B: show](x0: ctxt[A, B]): (List[Char]) => List[Char] =
  x0 match {
  case Hole => shows_string.apply(List('[', ']'))
  case More(f, ss1, d, ss2) =>
    comp[List[Char], List[Char],
          List[Char]]((a: List[Char]) => shows_prec[A](Zero_nat, f, a),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('(')),
  comp[List[Char], List[Char],
        List[Char]](shows_list_gen[term[A,
 B]]((a: term[A, B]) => shows_prec_term[A, B](Zero_nat, a), Nil, Nil,
      List(',', ' '), List(',', ' '), ss1),
                     comp[List[Char], List[Char],
                           List[Char]](shows_ctxt[A, B](d),
shows_list_gen[term[A, B]]((a: term[A, B]) =>
                             shows_prec_term[A, B](Zero_nat, a),
                            List(')'), List(',', ' '), List(',', ' '),
                            List(')'), ss2)))))
}

def hvf_top[A: equal, B](a: A, n: nat, x2: term[A, B]): Boolean = (a, n, x2)
  match {
  case (a, n, Fun(f, ts)) =>
    implies(eq[A](f, a) && equal_nata(size_list[term[A, B]](ts), n),
             ! (is_var[A, B](hd[term[A, B]](ts))))
  case (a, n, Var(uu)) => false
}

def list_inter[A: equal](x0: List[A], bs: List[A]): List[A] = (x0, bs) match {
  case (Nil, bs) => Nil
  case (a :: as, bs) =>
    (if (membera[A](bs, a)) a :: list_inter[A](as, bs)
      else list_inter[A](as, bs))
}

def list_union[A: equal](x0: List[A], ys: List[A]): List[A] = (x0, ys) match {
  case (Nil, ys) => ys
  case (x :: xs, ys) =>
    {
      val zs: List[A] = list_union[A](xs, ys);
      (if (membera[A](zs, x)) zs else x :: zs)
    }
}

def bind[A, B,
          C](m: (List[A]) => sum[List[Char], (B, List[A])],
              f: B => (List[A]) => sum[List[Char], (C, List[A])], ts: List[A]):
      sum[List[Char], (C, List[A])] =
  binda[List[Char], (B, List[A]),
         (C, List[A])](m(ts),
                        (a: (B, List[A])) =>
                          {
                            val (aa, b): (B, List[A]) = a;
                            (f(aa))(b)
                          })

def many(p: Char => Boolean, x1: List[Char]):
      sum[List[Char], (List[Char], List[Char])] =
  (p, x1) match {
  case (p, Nil) => Inr[(List[Char], List[Char]), List[Char]]((Nil, Nil))
  case (p, t :: ts) =>
    (if (p(t))
      binda[List[Char], (List[Char], List[Char]),
             (List[Char],
               List[Char])](many(p, ts),
                             (a: (List[Char], List[Char])) =>
                               {
                                 val (rs, tsa): (List[Char], List[Char]) = a;
                                 Inr[(List[Char], List[Char]),
                                      List[Char]]((t :: rs, tsa))
                               })
      else Inr[(List[Char], List[Char]), List[Char]]((Nil, t :: ts)))
}

def wspace: List[Char] = List(newline, '\11', '\15', ' ')

def trim(x0: List[Char]): List[Char] = x0 match {
  case Nil => Nil
  case c :: s => (if (membera[Char](wspace, c)) trim(s) else c :: s)
}

abstract sealed class domain
final case class Naturala(a: nat) extends domain
final case object Arctic extends domain
final case class Arctic_rat(a: rat) extends domain
final case class Int_mat(a: nat, b: nat) extends domain
final case class Arc_mat(a: nat) extends domain
final case class Arc_rat_mat(a: rat, b: nat) extends domain
final case class Rational(a: rat, b: nat) extends domain
final case class Rat_mat(a: rat, b: nat, c: nat) extends domain

def label[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], sum[List[nat], List[A]]] =
  (a: xml) =>
    xml_options[sum[List[nat],
                     List[A]]](List((List('n', 'u', 'm', 'b', 'e', 'r', 'L',
   'a', 'b', 'e', 'l'),
                                      (aa: xml) =>
xml_many_elements[nat, sum[List[nat],
                            List[A]]](List('n', 'u', 'm', 'b', 'e', 'r', 'L',
    'a', 'b', 'e', 'l'),
                                       (ab: xml) =>
 nata(List('n', 'u', 'm', 'b', 'e', 'r'), ab),
                                       (ab: List[nat]) =>
 Inl[List[nat], List[A]](ab),
                                       aa)),
                                     (List('s', 'y', 'm', 'b', 'o', 'l', 'L',
    'a', 'b', 'e', 'l'),
                                       (aa: xml) =>
 xml_many_elements[A, sum[List[nat],
                           List[A]]](List('s', 'y', 'm', 'b', 'o', 'l', 'L',
   'a', 'b', 'e', 'l'),
                                      xml2name,
                                      (ab: List[A]) =>
Inr[List[A], List[nat]](ab),
                                      aa))),
                                a)

def rules[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[(term[A, List[Char]], term[A, List[Char]])]] =
  (a: xml) =>
    xml_many_elements[(term[A, List[Char]], term[A, List[Char]]),
                       List[(term[A, List[Char]],
                              term[A, List[Char]])]](List('r', 'u', 'l', 'e',
                   's'),
              rule[A](xml2name),
              id[List[(term[A, List[Char]], term[A, List[Char]])]], a)

def state: xml => sum[List[Char], List[Char]] =
  (a: xml) => text(List('s', 't', 'a', 't', 'e'), a)

def isOK[A, B](m: sum[A, B]): Boolean =
  m match {
    case Inl(_) => false
    case Inr(_) => true
  }

implicit def mult_zero_int: mult_zero[BigInt] = new mult_zero[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

def list_update[A](x0: List[A], i: nat, y: A): List[A] = (x0, i, y) match {
  case (x :: xs, Suc(i), y) => x :: list_update[A](xs, i, y)
  case (x :: xs, Zero_nat, y) => y :: xs
  case (Nil, i, y) => Nil
}

def base[A](x0: (A, nat)): A = x0 match {
  case (f, h) => f
}

def lift[A](h: nat, f: A): (A, nat) = (f, h)

def vars_term_lista[A, B](x0: term[A, B]): List[B] = x0 match {
  case Var(x) => List(x)
  case Fun(uu, ts) =>
    maps[term[A, B], B]((a: term[A, B]) => vars_term_lista[A, B](a), ts)
}

def vars_term_list[A, B: equal]: (term[A, B]) => List[B] =
  comp[List[B], List[B],
        term[A, B]]((a: List[B]) => remdups[B](a),
                     (a: term[A, B]) => vars_term_lista[A, B](a))

def roof[A, B: equal](x0: (term[A, B], term[A, B])): (term[A, B]) => Boolean =
  x0 match {
  case (l, r) =>
    {
      val xs: List[B] = vars_term_list[A, B].apply(r);
      (t: term[A, B]) =>
        {
          val xt: List[B] = vars_term_list[A, B].apply(t);
          list_all[B]((a: B) => membera[B](xt, a), xs)
        }
    }
}

def mat_monoI[A](geq1: A => Boolean, sd: nat, m: List[List[A]]): Boolean =
  list_all[List[A]]((a: List[A]) => list_ex[A](geq1, a), sub_mat[A](sd, sd, m))

def scalar_prodI[A](ze: A, pl: A => A => A, ti: A => A => A, v: List[A],
                     w: List[A]):
      A =
  foldr[(A, A),
         A]((a: (A, A)) => {
                             val (x, y): (A, A) = a;
                             pl((ti(x))(y))
                           },
             zip[A, A](v, w), ze)

def matT_vec_multI[A](ze: A, pl: A => A => A, ti: A => A => A, m: List[List[A]],
                       v: List[A]):
      List[A] =
  mapa[List[A], A]((w: List[A]) => scalar_prodI[A](ze, pl, ti, w, v), m)

def transpose[A](nr: nat, x1: List[List[A]]): List[List[A]] = (nr, x1) match {
  case (nr, Nil) => replicate[List[A]](nr, Nil)
  case (nr, v :: m) =>
    mapa[(A, List[A]),
          List[A]]((a: (A, List[A])) =>
                     {
                       val (aa, b): (A, List[A]) = a;
                       aa :: b
                     },
                    zip[A, List[A]](v, transpose[A](nr, m)))
}

def mat_multI[A](ze: A, pl: A => A => A, ti: A => A => A, nr: nat,
                  m1: List[List[A]], m2: List[List[A]]):
      List[List[A]] =
  mapa[List[A],
        List[A]]((a: List[A]) =>
                   matT_vec_multI[A](ze, pl, ti, transpose[A](nr, m1), a),
                  m2)

def vec_plusI[A](pl: A => A => A, v: List[A], w: List[A]): List[A] =
  mapa[(A, A),
        A]((xy: (A, A)) => (pl(fst[A, A](xy)))(snd[A, A](xy)), zip[A, A](v, w))

def mat_plusI[A](pl: A => A => A, m1: List[List[A]], m2: List[List[A]]):
      List[List[A]] =
  mapa[(List[A], List[A]),
        List[A]]((uv: (List[A], List[A])) =>
                   vec_plusI[A](pl, fst[List[A], List[A]](uv),
                                 snd[List[A], List[A]](uv)),
                  zip[List[A], List[A]](m1, m2))

def equal_color(x0: color, x1: color): Boolean = (x0, x1) match {
  case (B, R) => false
  case (R, B) => false
  case (B, B) => true
  case (R, R) => true
}

def equal_rbtb[A: equal, B: equal](x0: rbta[A, B], x1: rbta[A, B]): Boolean =
  (x0, x1) match {
  case (Branch(color, rbt1, a, b, rbt2), Emptya) => false
  case (Emptya, Branch(color, rbt1, a, b, rbt2)) => false
  case (Branch(colora, rbt1a, aa, ba, rbt2a), Branch(color, rbt1, a, b, rbt2))
    => equal_color(colora, color) &&
         (equal_rbtb[A, B](rbt1a, rbt1) &&
           (eq[A](aa, a) && (eq[B](ba, b) && equal_rbtb[A, B](rbt2a, rbt2))))
  case (Emptya, Emptya) => true
}

def equal_rbta[A: equal: linorder, B: equal](ra: rbt[A, B], r: rbt[A, B]):
      Boolean =
  equal_rbtb[A, B](impl_of[A, B](ra), impl_of[A, B](r))

implicit def equal_rbt[A: equal: linorder, B: equal]: equal[rbt[A, B]] = new
  equal[rbt[A, B]] {
  val `Ceta.equal` = (a: rbt[A, B], b: rbt[A, B]) => equal_rbta[A, B](a, b)
}

implicit def mult_zero_rat: mult_zero[rat] = new mult_zero[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

trait comm_monoid_add[A] extends ab_semigroup_add[A] with monoid_add[A] {
}

trait semiring_0[A]
  extends comm_monoid_add[A] with mult_zero[A] with semiring[A] {
}

trait monoid_mult[A] extends semigroup_mult[A] with power[A] {
}

trait zero_neq_one[A] extends one[A] with zero[A] {
}

trait semiring_1[A]
  extends monoid_mult[A] with semiring_0[A] with zero_neq_one[A] {
}

def s_alpha[A, B](alpha: A => B => Option[Unit], m: A): B => Boolean =
  dom[B, Unit](alpha(m))

def s_empty[A](empt: A): A = empt

def s_union[A](add: A): A = add

def linear_term_impl[A: equal, B](xs: List[A], x1: term[B, A]):
      Option[List[A]] =
  (xs, x1) match {
  case (xs, Var(x)) => (if (membera[A](xs, x)) None else Some[List[A]](x :: xs))
  case (xs, Fun(uu, Nil)) => Some[List[A]](xs)
  case (xs, Fun(f, t :: ts)) =>
    linear_term_impl[A, B](xs, t) match {
      case None => None
      case Some(ys) => linear_term_impl[A, B](ys, Fun[B, A](f, ts))
    }
}

def linear_term[A, B: equal](t: term[A, B]): Boolean =
  linear_term_impl[B, A](Nil, t) match {
    case None => false
    case Some(_) => true
  }

def map_term_wa[A, B, C](fg: A => nat => B, x1: term[A, C]): term[B, C] =
  (fg, x1) match {
  case (fg, Var(x)) => Var[C, B](x)
  case (fg, Fun(f, ts)) =>
    Fun[B, C]((fg(f))(size_list[term[A, C]](ts)),
               mapa[term[A, C],
                     term[B, C]]((a: term[A, C]) => map_term_wa[A, B, C](fg, a),
                                  ts))
}

abstract sealed class ta_rule[+A, +B]
final case class TA_rule[B, A](a: B, b: List[A], c: A) extends ta_rule[A, B]

def rq[A, B](x0: ta_rule[A, B]): A = x0 match {
  case TA_rule(f, qs, q) => q
}

def map_rules_wa[A, B, C](fg: A => nat => B, r: List[(term[A, C], term[A, C])]):
      List[(term[B, C], term[B, C])] =
  mapa[(term[A, C], term[A, C]),
        (term[B, C],
          term[B, C])]((a: (term[A, C], term[A, C])) =>
                         {
                           val (l, ra): (term[A, C], term[A, C]) = a;
                           (map_term_wa[A, B, C](fg, l),
                             map_term_wa[A, B, C](fg, ra))
                         },
                        r)

def map_trs_list[A, B, C](fg: A => B, r: List[(term[A, C], term[A, C])]):
      List[(term[B, C], term[B, C])] =
  mapa[(term[A, C], term[A, C]),
        (term[B, C],
          term[B, C])]((a: (term[A, C], term[A, C])) =>
                         map_rule[A, B, C](fg, a),
                        r)

def supteq_list[A, B](x0: term[A, B]): List[term[A, B]] = x0 match {
  case Var(x) => List(Var[B, A](x))
  case Fun(f, ts) =>
    Fun[A, B](f, ts) ::
      maps[term[A, B], term[A, B]]((a: term[A, B]) => supteq_list[A, B](a), ts)
}

def dP_list[A: equal,
             B: equal](shp: A => A, r: List[(term[A, B], term[A, B])],
                        rm: Option[((A, nat)) =>
                                     List[(term[A, B], term[A, B])]]):
      List[(term[A, B], term[A, B])] =
  {
    val rma: ((A, nat)) => List[(term[A, B], term[A, B])] =
      the[((A, nat)) => List[(term[A, B], term[A, B])]](rm);
    maps[(term[A, B], term[A, B]),
          (term[A, B],
            term[A, B])]((lr: (term[A, B], term[A, B])) =>
                           {
                             val l: term[A, B] = fst[term[A, B], term[A, B]](lr)
                             val s: term[A, B] = sharp_term[A, B](shp, l);
                             map_filter[term[A, B],
 (term[A, B],
   term[A, B])]((x: term[A, B]) =>
                  (if (! (member[(term[A, B],
                                   term[A, B])]((l, x), supt[A, B])) &&
                         (! (is_var[A, B](x)) &&
                           ! (nulla[(term[A, B],
                                      term[A,
    B])](rma((the[A](root[A, B](x)), num_args[A, B](x)))))))
                    Some[(term[A, B],
                           term[A, B])]((s, sharp_term[A, B](shp, x)))
                    else None),
                 supteq_list[A, B](snd[term[A, B], term[A, B]](lr)))
                           },
                          r)
  }

def hvf_term[A: equal, B](a: A, t: term[A, B]): Boolean =
  unapp[A, B](a, t) match {
    case (Var(_), ts) => nulla[term[A, B]](ts)
    case (Fun(_, us), ts) =>
      list_all[term[A, B]]((b: term[A, B]) => hvf_term[A, B](a, b), us ++ ts)
  }

def oneof_aux(allowed: List[List[Char]], x1: List[List[Char]], ts: List[Char]):
      sum[List[Char], (List[Char], List[Char])] =
  (allowed, x1, ts) match {
  case (allowed, Nil, ts) =>
    expecting[Char,
               List[Char]](List('o', 'n', 'e', ' ', 'o', 'f', ' ') ++
                             (shows_list_list[Char](allowed)).apply(Nil),
                            ts)
  case (allowed, x :: xs, ts) =>
    (if (equal_lista[Char](mapa[(Char, Char),
                                 Char]((a: (Char, Char)) => snd[Char, Char](a),
zip[Char, Char](x, ts)),
                            x))
      Inr[(List[Char], List[Char]),
           List[Char]]((x, trim(drop[Char](size_list[Char](x), ts))))
      else oneof_aux(allowed, xs, ts))
}

def oneof(xs: List[List[Char]]):
      (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) => oneof_aux(xs, xs, a)

abstract sealed class xmldoc
final case class Xmldoc(a: List[Char], b: xml) extends xmldoc

def xml5elements[A, B, C, D, E,
                  F](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B],
                      p3: xml => sum[List[Char], C],
                      p4: xml => sum[List[Char], D],
                      p5: xml => sum[List[Char], E],
                      f: A => B => C => D => E => F, x7: xml):
      sum[List[Char], F] =
  (tag, p1, p2, p3, p4, p5, f, x7) match {
  case (tag, p1, p2, p3, p4, p5, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(5))) &&
               is_none[List[Char]](text))))
      binda[List[Char], A,
             F](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   binda[List[Char], B,
                          F](p2(nth[xml](cs, one_nat)),
                              (x2: B) =>
                                binda[List[Char], C,
                                       F](p3(nth[xml](cs, nat(BigInt(2)))),
   (x3: C) =>
     binda[List[Char], D,
            F](p4(nth[xml](cs, nat(BigInt(3)))),
                (x4: D) =>
                  binda[List[Char], E,
                         F](p5(nth[xml](cs, nat(BigInt(4)))),
                             (x5: E) =>
                               Inr[F, List[Char]](((((f(x1))(x2))(x3))(x4))(x5)))))))
      else no_transform[F](tag, Xml(name, atts, cs, text)))
}

def catcha[A, B, C](m: sum[A, B], f: A => sum[C, B]): sum[C, B] =
  m match {
    case Inl(a) => f(a)
    case Inr(a) => Inr[B, C](a)
  }

def forallm[A, B](f: A => sum[B, Unit], x1: List[A]): sum[(A, B), Unit] =
  (f, x1) match {
  case (f, Nil) => Inr[Unit, (A, B)](())
  case (f, x :: xs) =>
    binda[(A, B), Unit,
           Unit](catcha[B, Unit,
                         (A, B)](f(x), (xa: B) => Inl[(A, B), Unit]((x, xa))),
                  (_: Unit) => forallm[A, B](f, xs))
}

def check_edges[A](ss: A => (List[Char]) => List[Char], g: ((A, A)) => Boolean,
                    c: List[A], d: List[A]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(A, (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[A, (List[Char]) =>
                                     List[Char]]((n: A) =>
           catcha[(A, (List[Char]) => List[Char]), Unit,
                   (List[Char]) =>
                     List[Char]](forallm[A,
  (List[Char]) =>
    List[Char]]((m: A) =>
                  check[(List[Char]) =>
                          List[Char]](! (g((m, n))),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('e', 'd', 'g', 'e', ' ', 'f', 'r', 'o',
  'm', ' ')),
                  comp[List[Char], List[Char],
                        List[Char]](ss(m),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 't', 'o', ' ')), ss(n))))),
                 d),
                                  (x: (A, (List[Char]) => List[Char])) =>
                                    Inl[(List[Char]) => List[Char],
 Unit](snd[A, (List[Char]) => List[Char]](x))),
          c),
                         (x: (A, (List[Char]) => List[Char])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](snd[A, (List[Char]) => List[Char]](x)))

implicit def monoid_add_int: monoid_add[BigInt] = new monoid_add[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def comm_monoid_add_int: comm_monoid_add[BigInt] = new
  comm_monoid_add[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def semiring_0_int: semiring_0[BigInt] = new semiring_0[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
}

implicit def monoid_mult_int: monoid_mult[BigInt] = new monoid_mult[BigInt] {
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def zero_neq_one_int: zero_neq_one[BigInt] = new zero_neq_one[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.one` = one_inta
}

implicit def semiring_1_int: semiring_1[BigInt] = new semiring_1[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

def matcha[A, B]: ((term[A, B], term[A, B])) => (term[A, B]) => Boolean =
  (_: (term[A, B], term[A, B])) => (_: term[A, B]) => true

implicit def
  linorder_prod[A: equal: linorder, B: equal: linorder]: linorder[(A, B)] = new
  linorder[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

def top_boola: Boolean = true

implicit def top_bool: top[Boolean] = new top[Boolean] {
  val `Ceta.top` = top_boola
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

def top_fun[A, B: top]: A => B = (_: A) => top[B]

def equal_posa(x0: pos, x1: pos): Boolean = (x0, x1) match {
  case (PCons(nat, pos), Empty) => false
  case (Empty, PCons(nat, pos)) => false
  case (PCons(nata, posa), PCons(nat, pos)) =>
    equal_nata(nata, nat) && equal_posa(posa, pos)
  case (Empty, Empty) => true
}

implicit def equal_pos: equal[pos] = new equal[pos] {
  val `Ceta.equal` = (a: pos, b: pos) => equal_posa(a, b)
}

abstract sealed class arctic
final case object MinInfty extends arctic
final case class Num_arc(a: BigInt) extends arctic

abstract sealed class arctic_rat
final case object MinInfty_rat extends arctic_rat
final case class Num_arc_rat(a: rat) extends arctic_rat

abstract sealed class redpair[+A]
final case class Int_carrier[A](a: List[((A, nat), (BigInt, List[BigInt]))])
  extends redpair[A]
final case class
  Int_nl_carrier[A](a: List[((A, nat), List[(List[(nat, nat)], BigInt)])])
  extends redpair[A]
final case class Rat_carrier[A](a: rat, b: List[((A, nat), (rat, List[rat]))])
  extends redpair[A]
final case class
  Rat_nl_carrier[A](a: rat, b: List[((A, nat), List[(List[(nat, nat)], rat)])])
  extends redpair[A]
final case class Arctic_carrier[A](a: List[((A, nat), (arctic, List[arctic]))])
  extends redpair[A]
final case class
  Arctic_rat_carrier[A](a: rat,
                         b: List[((A, nat), (arctic_rat, List[arctic_rat]))])
  extends redpair[A]
final case class
  Int_mat_carrier[A](a: nat, b: nat,
                      c: List[((A, nat),
                                (List[List[BigInt]],
                                  List[List[List[BigInt]]]))])
  extends redpair[A]
final case class
  Rat_mat_carrier[A](a: rat, b: nat, c: nat,
                      d: List[((A, nat),
                                (List[List[rat]], List[List[List[rat]]]))])
  extends redpair[A]
final case class
  Arc_mat_carrier[A](a: nat,
                      b: List[((A, nat),
                                (List[List[arctic]],
                                  List[List[List[arctic]]]))])
  extends redpair[A]
final case class
  Arc_rat_mat_carrier[A](a: rat, b: nat,
                          c: List[((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]]))])
  extends redpair[A]
final case class
  Lpo[A](a: List[((A, nat), (nat, order_tag))], b: List[((A, nat), af_entry)])
  extends redpair[A]
final case class
  Mpo[A](a: List[((A, nat), (nat, order_tag))], b: List[((A, nat), af_entry)])
  extends redpair[A]
final case class
  Rpo[A](a: List[((A, nat), (nat, order_tag))], b: List[((A, nat), af_entry)])
  extends redpair[A]

abstract sealed class arithFun
final case class Arg(a: nat) extends arithFun
final case class Const(a: nat) extends arithFun
final case class Sum(a: List[arithFun]) extends arithFun
final case class Max(a: List[arithFun]) extends arithFun
final case class Min(a: List[arithFun]) extends arithFun
final case class Prod(a: List[arithFun]) extends arithFun

abstract sealed class sl_inter[+A, +B]
final case class SL_Inter[A, B](a: nat, b: List[((lab[A, B], nat), arithFun)])
  extends sl_inter[A, B]

abstract sealed class dPProof[+A, +B, +C]
final case object PisEmpty extends dPProof[Nothing, Nothing, Nothing]
final case class
  RedPairProc[A, B,
               C](a: redpair[lab[A, B]],
                   b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   c: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  RedPairUrProc[A, B,
                 C](a: redpair[lab[A, B]],
                     b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                     c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                     d: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  DepGraphProc[A, B,
                C](a: List[(Option[dPProof[A, B, C]],
                             List[(term[lab[A, B], C], term[lab[A, B], C])])])
  extends dPProof[A, B, C]
final case class
  MonoRedPairProc[A, B,
                   C](a: redpair[lab[A, B]],
                       b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                       c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                       d: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  MonoRedPairUrProc[A, B,
                     C](a: redpair[lab[A, B]],
                         b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         e: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  UnlabProc[A, B,
             C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                 b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                 c: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  SemlabProc[A, B,
              C](a: sl_inter[A, B],
                  b: sum[(lab[A, B]) => (List[nat]) => sum[B, List[lab[A, B]]],
                          (C, Boolean)],
                  c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  e: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  RootlabProc[A, B,
               C](a: Option[lab[A, B]],
                   b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   d: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  SubtermCriterionProc[A, B,
                        C](a: projL[lab[A, B]],
                            b: List[((term[lab[A, B], C], term[lab[A, B], C]),
                                      List[(pos,
     ((term[lab[A, B], C], term[lab[A, B], C]), term[lab[A, B], C]))])],
                            c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                            d: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  SizeChangeSubTerm[A, B,
                     C](a: List[((term[lab[A, B], C], term[lab[A, B], C]),
                                  (List[(nat, nat)], List[(nat, nat)]))])
  extends dPProof[A, B, C]
final case class
  SizeChangeRedPair[A, B,
                     C](a: redpair[lab[A, B]],
                         b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         c: List[((term[lab[A, B], C], term[lab[A, B], C]),
                                   (List[(nat, nat)], List[(nat, nat)]))])
  extends dPProof[A, B, C]
final case class
  UncurryProc[A, B,
               C](a: Option[nat],
                   b: (lab[A, B],
                        (List[((lab[A, B], nat), List[lab[A, B]])],
                          (List[(term[lab[A, B], C], term[lab[A, B], C])],
                            List[(term[lab[A, B], C], term[lab[A, B], C])]))),
                   c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   e: dPProof[A, B, C])
  extends dPProof[A, B, C]
final case class
  FccProc[A, B,
           C](a: lab[A, B], b: List[ctxt[lab[A, B], C]],
               c: List[(term[lab[A, B], C], term[lab[A, B], C])],
               d: List[(term[lab[A, B], C], term[lab[A, B], C])],
               e: dPProof[A, B, C])
  extends dPProof[A, B, C]

abstract sealed class boundstype
final case object Roof extends boundstype
final case object Match extends boundstype

abstract sealed class tree_automaton[+A, +B]
final case class
  Tree_Automaton[A, B](a: List[A], b: List[ta_rule[A, B]], c: List[(A, A)])
  extends tree_automaton[A, B]

abstract sealed class bounds_info[+A, +B]
final case class
  Bounds_Info[B, A](a: boundstype, b: nat, c: List[B],
                     d: tree_automaton[B, (A, nat)])
  extends bounds_info[A, B]

abstract sealed class tRSProof[+A, +B, +C]
final case class
  DPTrans[A, B,
           C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
               b: dPProof[A, B, C])
  extends tRSProof[A, B, C]
final case class
  RuleRemoval[A, B,
               C](a: redpair[lab[A, B]],
                   b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   c: tRSProof[A, B, C])
  extends tRSProof[A, B, C]
final case class
  Unlab[A, B,
         C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
             b: tRSProof[A, B, C])
  extends tRSProof[A, B, C]
final case class
  Semlab[A, B,
          C](a: sl_inter[A, B],
              b: sum[(lab[A, B]) => (List[nat]) => sum[B, List[lab[A, B]]], C],
              c: List[(term[lab[A, B], C], term[lab[A, B], C])],
              d: tRSProof[A, B, C])
  extends tRSProof[A, B, C]
final case class
  Rootlab[A, B,
           C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
               b: tRSProof[A, B, C])
  extends tRSProof[A, B, C]
final case class
  Fcc[A, B,
       C](a: List[ctxt[lab[A, B], C]],
           b: List[(term[lab[A, B], C], term[lab[A, B], C])],
           c: tRSProof[A, B, C])
  extends tRSProof[A, B, C]
final case class
  StringReversal[A, B,
                  C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                      b: tRSProof[A, B, C])
  extends tRSProof[A, B, C]
final case class Bounds[A, B, C](a: bounds_info[lab[A, B], C]) extends
  tRSProof[A, B, C]
final case class
  Uncurry[A, B,
           C](a: (lab[A, B],
                   (List[((lab[A, B], nat), List[lab[A, B]])],
                     (List[(term[lab[A, B], C], term[lab[A, B], C])],
                       List[(term[lab[A, B], C], term[lab[A, B], C])]))),
               b: List[(term[lab[A, B], C], term[lab[A, B], C])],
               c: tRSProof[A, B, C])
  extends tRSProof[A, B, C]
final case object RisEmpty extends tRSProof[Nothing, Nothing, Nothing]

def rm_update[A: linorder, B]: A => B => (rbt[A, B]) => rbt[A, B] =
  (a: A) => (b: B) => (c: rbt[A, B]) => insert[A, B](a, b, c)

def itoi_iterateo[A, B, C,
                   D](it: (A => Boolean) => (B => C => A => A) => D => A => A):
      (B => C => A => A) => D => A => A =
  it((_: A) => true)

def rm_iterateoi[A, B,
                  C](c: A => Boolean, f: B => C => A => A, x2: rbta[B, C],
                      sigma: A):
      A =
  (c, f, x2, sigma) match {
  case (c, f, Emptya, sigma) => sigma
  case (c, f, Branch(col, l, k, v, r), sigma) =>
    (if (c(sigma))
      {
        val sigmaa: A = rm_iterateoi[A, B, C](c, f, l, sigma);
        (if (c(sigmaa)) rm_iterateoi[A, B, C](c, f, r, ((f(k))(v))(sigmaa))
          else sigmaa)
      }
      else sigma)
}

def rm_iterateoia[A, B: linorder,
                   C](c: A => Boolean, f: B => C => A => A, r: rbt[B, C]):
      A => A =
  (a: A) => rm_iterateoi[A, B, C](c, f, impl_of[B, C](r), a)

def rm_iterateo[A: linorder, B, C]:
      (A => B => C => C) => (rbt[A, B]) => C => C =
  itoi_iterateo[C, A, B,
                 rbt[A, B]]((a: C => Boolean) =>
                              (b: A => B => C => C) =>
                                (c: rbt[A, B]) =>
                                  rm_iterateoia[C, A, B](a, b, c))

def rm_iterate[A: linorder, B, C]: (A => B => C => C) => (rbt[A, B]) => C => C =
  rm_iterateo[A, B, C]

def rm_add[A: linorder, B]: (rbt[A, B]) => (rbt[A, B]) => rbt[A, B] =
  (a: rbt[A, B]) =>
    (b: rbt[A, B]) =>
      it_add[A, B,
              rbt[A, B]](rm_update[A, B], rm_iterate[A, B, rbt[A, B]], a, b)

def rs_ins[A: linorder]: A => (rbt[A, Unit]) => rbt[A, Unit] =
  (a: A) =>
    (b: rbt[A, Unit]) =>
      s_ins[A, rbt[A, Unit], rbt[A, Unit]](rm_update[A, Unit], a, b)

def anyNstriE[A, B](f: A => B => ((Boolean, Boolean), A), m: A, x2: List[B]):
      (Boolean, A) =
  (f, m, x2) match {
  case (f, m, Nil) => (false, m)
  case (f, m, s :: ss) =>
    {
      val (a, b): ((Boolean, Boolean), A) = (f(m))(s);
      ({
         val (_, nstri): (Boolean, Boolean) = a;
         (ma: A) => (if (nstri) (nstri, ma) else anyNstriE[A, B](f, ma, ss))
       })(b)
    }
}

def allStriE[A, B](f: A => B => ((Boolean, Boolean), A), m: A, x2: List[B]):
      (Boolean, A) =
  (f, m, x2) match {
  case (f, m, Nil) => (true, m)
  case (f, m, t :: ts) =>
    {
      val (a, b): ((Boolean, Boolean), A) = (f(m))(t);
      ({
         val (stri, _): (Boolean, Boolean) = a;
         (ma: A) => (if (stri) allStriE[A, B](f, ma, ts) else (false, ma))
       })(b)
    }
}

def lex_extE[A, B](f: A => B => B => ((Boolean, Boolean), A), m: A, x2: List[B],
                    x3: List[B]):
      ((Boolean, Boolean), A) =
  (f, m, x2, x3) match {
  case (f, m, Nil, Nil) => ((false, true), m)
  case (f, m, uu :: uv, Nil) => ((true, true), m)
  case (f, m, Nil, uw :: ux) => ((false, false), m)
  case (f, m, a :: as, b :: bs) =>
    {
      val (aa, ba): ((Boolean, Boolean), A) = ((f(m))(a))(b);
      ({
         val (stri, nstri): (Boolean, Boolean) = aa;
         (ma: A) =>
           (if (stri) ((true, true), ma)
             else (if (nstri) lex_extE[A, B](f, ma, as, bs)
                    else ((false, false), ma)))
       })(ba)
    }
}

def lpoE_m[A, B,
            C: equal](m_ops:
                        (A => (term[B, C]) =>
                                (term[B, C]) => Option[(Boolean, Boolean)],
                          A => (term[B, C]) =>
                                 (term[B, C]) => ((Boolean, Boolean)) => A),
                       m: A, pr: B => nat => nat, s: term[B, C], t: term[B, C]):
      ((Boolean, Boolean), A) =
  ((({
       val (x, _):
             (A => (term[B, C]) => (term[B, C]) => Option[(Boolean, Boolean)],
               A => (term[B, C]) => (term[B, C]) => ((Boolean, Boolean)) => A)
         = m_ops;
       x
     })(m))(s))(t)
    match {
    case None =>
      {
        val (v, new_m): ((Boolean, Boolean), A) =
          (s, t) match {
            case (Var(x), Var(y)) => ((false, eq[C](x, y)), m)
            case (Var(x), Fun(g, ts)) =>
              ((false,
                 nulla[term[B, C]](ts) &&
                   equal_nata((pr(g))(Zero_nat), Zero_nat)),
                m)
            case (Fun(f, ss), Var(_)) =>
              {
                val (con, a): (Boolean, A) =
                  anyNstriE[A, term[B, C]]((lm: A) =>
     (sa: term[B, C]) => lpoE_m[A, B, C](m_ops, lm, pr, sa, t),
    m, ss);
                ((con, con), a)
              }
            case (Fun(f, ss), Fun(g, ts)) =>
              anyNstriE[A, term[B, C]]((lm: A) =>
 (sa: term[B, C]) => lpoE_m[A, B, C](m_ops, lm, pr, sa, t),
m, ss)
                match {
                case (true, ma) => ((true, true), ma)
                case (false, ma) =>
                  {
                    val prl: nat = (pr(f))(size_list[term[B, C]](ss))
                    val prr: nat = (pr(g))(size_list[term[B, C]](ts))
                    val (sAllStri, mb): (Boolean, A) =
                      allStriE[A, term[B,
C]]((lm: A) => (a: term[B, C]) => lpoE_m[A, B, C](m_ops, lm, pr, s, a), ma, ts);
                    (if (less_eq_nat(prr, prl) && sAllStri)
                      (if (less_nat(prr, prl)) ((true, true), mb)
                        else lex_extE[A, term[B,
       C]]((lm: A) =>
             (a: term[B, C]) =>
               (b: term[B, C]) => lpoE_m[A, B, C](m_ops, lm, pr, a, b),
            mb, ss, ts))
                      else ((false, false), mb))
                  }
              }
          };
        (v, (((({
                  val (_, y):
                        (A => (term[B, C]) =>
                                (term[B, C]) => Option[(Boolean, Boolean)],
                          A => (term[B, C]) =>
                                 (term[B, C]) => ((Boolean, Boolean)) => A)
                    = m_ops;
                  y
                })(new_m))(s))(t))(v))
      }
    case Some(v) => (v, m)
  }

def lookup_2[A: equal, B](x0: List[(A, List[(A, B)])], s: A, t: A): Option[B] =
  (x0, s, t) match {
  case (Nil, s, t) => None
  case (head :: tail, s, t) =>
    (if (eq[A](fst[A, List[(A, B)]](head), s))
      lookupa[A, B](t, snd[A, List[(A, B)]](head))
      else lookup_2[A, B](tail, s, t))
}

def store_2[A: equal, B](x0: List[(A, List[(A, B)])], s: A, t: A, v: B):
      List[(A, List[(A, B)])] =
  (x0, s, t, v) match {
  case (Nil, s, t, v) => List((s, List((t, v))))
  case (head :: tail, s, t, v) =>
    (if (eq[A](fst[A, List[(A, B)]](head), s))
      (s, (t, v) :: snd[A, List[(A, B)]](head)) :: tail
      else head :: store_2[A, B](tail, s, t, v))
}

def lpo_E_2[A: equal,
             B: equal](pr: A => nat => nat, s: term[A, B], t: term[A, B]):
      (Boolean, Boolean) =
  fst[(Boolean, Boolean),
       List[(term[A, B],
              List[(term[A, B],
                     (Boolean,
                       Boolean))])]](lpoE_m[List[(term[A, B],
           List[(term[A, B], (Boolean, Boolean))])],
     A, B](((a: List[(term[A, B], List[(term[A, B], (Boolean, Boolean))])]) =>
              (b: term[A, B]) =>
                (c: term[A, B]) =>
                  lookup_2[term[A, B], (Boolean, Boolean)](a, b, c),
             (a: List[(term[A, B], List[(term[A, B], (Boolean, Boolean))])]) =>
               (b: term[A, B]) =>
                 (c: term[A, B]) =>
                   (d: (Boolean, Boolean)) =>
                     store_2[term[A, B], (Boolean, Boolean)](a, b, c, d)),
            Nil, pr, s, t))

def lpo_unbounded[A: equal, B: equal]:
      (A => nat => nat) => (term[A, B]) => (term[A, B]) => (Boolean, Boolean) =
  (a: A => nat => nat) =>
    (b: term[A, B]) => (c: term[A, B]) => lpo_E_2[A, B](a, b, c)

def equal_order_tag(x0: order_tag, x1: order_tag): Boolean = (x0, x1) match {
  case (Mul, Lex) => false
  case (Lex, Mul) => false
  case (Mul, Mul) => true
  case (Lex, Lex) => true
}

def lex_ext_unbounded[A](f: A => A => (Boolean, Boolean), x1: List[A],
                          x2: List[A]):
      (Boolean, Boolean) =
  (f, x1, x2) match {
  case (f, Nil, Nil) => (false, true)
  case (f, uu :: uv, Nil) => (true, true)
  case (f, Nil, uw :: ux) => (false, false)
  case (f, a :: as, b :: bs) =>
    (f(a))(b) match {
      case (true, nstri) => (true, true)
      case (false, true) => lex_ext_unbounded[A](f, as, bs)
      case (false, false) => (false, false)
    }
}

def rpo_unbounded_impl[A: equal,
                        B: equal](uu: (A =>
 nat => A => nat => (Boolean, Boolean),
A => Boolean),
                                   uv: A => nat => order_tag, x2: term[A, B],
                                   x3: term[A, B]):
      (Boolean, Boolean) =
  (uu, uv, x2, x3) match {
  case (uu, uv, Var(x), Var(y)) => (false, eq[B](x, y))
  case (pr, uw, Var(x), Fun(g, ts)) =>
    (false,
      nulla[term[A, B]](ts) &&
        (snd[A => nat => A => nat => (Boolean, Boolean),
              A => Boolean](pr)).apply(g))
  case (pr, c, Fun(f, ss), Var(y)) =>
    {
      val con: Boolean =
        list_ex[term[A, B]]((s: term[A, B]) =>
                              snd[Boolean,
                                   Boolean](rpo_unbounded_impl[A,
                        B](pr, c, s, Var[B, A](y))),
                             ss);
      (con, con)
    }
  case (pr, c, Fun(f, ss), Fun(g, ts)) =>
    (if (list_ex[term[A, B]]((s: term[A, B]) =>
                               snd[Boolean,
                                    Boolean](rpo_unbounded_impl[A,
                         B](pr, c, s, Fun[A, B](g, ts))),
                              ss))
      (true, true)
      else {
             val (prs, prns): (Boolean, Boolean) =
               (fst[A => nat => A => nat => (Boolean, Boolean),
                     A => Boolean](pr)).apply(f).apply(size_list[term[A,
                               B]](ss)).apply(g).apply(size_list[term[A,
                               B]](ts));
             (if (prns &&
                    list_all[term[A, B]]((t: term[A, B]) =>
   fst[Boolean, Boolean](rpo_unbounded_impl[A, B](pr, c, Fun[A, B](f, ss), t)),
  ts))
               (if (prs) (true, true)
                 else (if (equal_order_tag((c(f))(size_list[term[A, B]](ss)),
    (c(g))(size_list[term[A, B]](ts))))
                        (if (equal_order_tag((c(f))(size_list[term[A, B]](ss)),
      Mul))
                          mul_ext_impl[term[A,
     B]]((a: term[A, B]) =>
           (b: term[A, B]) => rpo_unbounded_impl[A, B](pr, c, a, b),
          ss, ts)
                          else lex_ext_unbounded[term[A,
               B]]((a: term[A, B]) =>
                     (b: term[A, B]) => rpo_unbounded_impl[A, B](pr, c, a, b),
                    ss, ts))
                        else (! (equal_nata(size_list[term[A, B]](ss),
     Zero_nat)) &&
                                equal_nata(size_list[term[A, B]](ts), Zero_nat),
                               equal_nata(size_list[term[A, B]](ts),
   Zero_nat))))
               else (false, false))
           })
}

def rpo_unbounded[A: equal, B: equal]:
      ((A => nat => A => nat => (Boolean, Boolean), A => Boolean)) =>
        (A => nat => order_tag) =>
          (term[A, B]) => (term[A, B]) => (Boolean, Boolean) =
  (a: (A => nat => A => nat => (Boolean, Boolean), A => Boolean)) =>
    (b: A => nat => order_tag) =>
      (c: term[A, B]) => (d: term[A, B]) => rpo_unbounded_impl[A, B](a, b, c, d)

implicit def monoid_add_rat: monoid_add[rat] = new monoid_add[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def comm_monoid_add_rat: comm_monoid_add[rat] = new
  comm_monoid_add[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def semiring_0_rat: semiring_0[rat] = new semiring_0[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
}

implicit def monoid_mult_rat: monoid_mult[rat] = new monoid_mult[rat] {
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def zero_neq_one_rat: zero_neq_one[rat] = new zero_neq_one[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.one` = one_rata
}

implicit def semiring_1_rat: semiring_1[rat] = new semiring_1[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

def equal_suma[A: equal, B: equal](x0: sum[A, B], x1: sum[A, B]): Boolean =
  (x0, x1) match {
  case (Inr(b), Inl(a)) => false
  case (Inl(a), Inr(b)) => false
  case (Inr(ba), Inr(b)) => eq[B](ba, b)
  case (Inl(aa), Inl(a)) => eq[A](aa, a)
}

implicit def equal_sum[A: equal, B: equal]: equal[sum[A, B]] = new
  equal[sum[A, B]] {
  val `Ceta.equal` = (a: sum[A, B], b: sum[A, B]) => equal_suma[A, B](a, b)
}

def match_vars[A: equal, B: equal](x0: List[(A, term[B, A])]):
      Option[A => term[B, A]] =
  x0 match {
  case Nil => Some[A => term[B, A]]((a: A) => Var[A, B](a))
  case (x, t) :: ps =>
    lookupa[A, term[B, A]](x, ps) match {
      case None =>
        match_vars[A, B](ps) match {
          case None => None
          case Some(sigma) =>
            Some[A => term[B, A]](fun_upd[A, term[B, A]](sigma, x, t))
        }
      case Some(s) =>
        (if (equal_terma[B, A](t, s)) match_vars[A, B](ps) else None)
    }
}

def match_term_list[A: equal, B](x0: List[(term[A, B], term[A, B])]):
      Option[List[(B, term[A, B])]] =
  x0 match {
  case Nil => Some[List[(B, term[A, B])]](Nil)
  case (Var(uu), Fun(uv, uw)) :: ps => None
  case (Fun(f, ts), Fun(g, ls)) :: ps =>
    (if (eq[A](f, g))
      decomp[term[A, B], term[A, B]](ts, ls) match {
        case None => None
        case Some(eqs) => match_term_list[A, B](eqs ++ ps)
      }
      else None)
  case (t, Var(x)) :: ps =>
    match_term_list[A, B](ps) match {
      case None => None
      case Some(eqs) => Some[List[(B, term[A, B])]]((x, t) :: eqs)
    }
}

def match_list[A: equal, B: equal](tls: List[(term[A, B], term[A, B])]):
      Option[B => term[A, B]] =
  match_term_list[A, B](tls) match {
    case None => None
    case Some(a) => match_vars[B, A](a)
  }

def instance_rule[A: equal,
                   B: equal](lr: (term[A, B], term[A, B]),
                              st: (term[A, B], term[A, B])):
      Boolean =
  ! (is_none[B => term[A, B]](match_list[A,
  B](List((fst[term[A, B], term[A, B]](lr), fst[term[A, B], term[A, B]](st)),
           (snd[term[A, B], term[A, B]](lr),
             snd[term[A, B], term[A, B]](st))))))

def check_pairwise[A, B](c: A => A => sum[B, Unit], x1: List[A]): sum[B, Unit] =
  (c, x1) match {
  case (c, Nil) => Inr[Unit, B](())
  case (c, x :: xs) =>
    binda[B, Unit,
           Unit](catcha[(A, B), Unit,
                         B](forallm[A, B](c(x), xs),
                             (xa: (A, B)) => Inl[B, Unit](snd[A, B](xa))),
                  (_: Unit) => check_pairwise[A, B](c, xs))
}

def check_disjoint[A: equal](xs: List[A], ys: List[A]): sum[A, Unit] =
  catcha[(A, A), Unit,
          A](forallm[A, A]((x: A) =>
                             (if (! (membera[A](ys, x))) Inr[Unit, A](())
                               else Inl[A, Unit](x)),
                            xs),
              (x: (A, A)) => Inl[A, Unit](snd[A, A](x)))

def check_partition[A: equal](xss: List[List[A]]): sum[A, Unit] =
  check_pairwise[List[A],
                  A]((a: List[A]) => (b: List[A]) => check_disjoint[A](a, b),
                      xss)

implicit def equal_prod[A: equal, B: equal]: equal[(A, B)] = new equal[(A, B)] {
  val `Ceta.equal` = (a: (A, B), b: (A, B)) => equal_proda[A, B](a, b)
}

def shows_lab[A: show, B: show](x0: lab[A, B]): (List[Char]) => List[Char] = x0
  match {
  case UnLab(f) => (a: List[Char]) => shows_prec[A](Zero_nat, f, a)
  case Lab(f, l) =>
    comp[List[Char], List[Char],
          List[Char]](shows_lab[A, B](f),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('[')),
  comp[List[Char], List[Char],
        List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, l, a),
                     shows_string.apply(List(']')))))
  case Sharp(f) =>
    comp[List[Char], List[Char],
          List[Char]](shows_lab[A, B](f), shows_string.apply(List('#')))
  case FunLab(f, l) =>
    comp[List[Char], List[Char],
          List[Char]](shows_lab[A, B](f),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('[')),
  comp[List[Char], List[Char],
        List[Char]](shows_map[lab[A, B]]((a: lab[A, B]) => shows_lab[A, B](a),
  l),
                     shows_string.apply(List(']')))))
}

def shows_prec_lab[A: show, B: show](d: nat, l: lab[A, B]):
      (List[Char]) => List[Char] =
  shows_lab[A, B](l)

def shows_list_lab[A: show, B: show](fs: List[lab[A, B]]):
      (List[Char]) => List[Char] =
  shows_list_aux[lab[A, B]]((a: lab[A, B]) => shows_prec_lab[A, B](Zero_nat, a),
                             fs)

implicit def show_lab[A: show, B: show]: show[lab[A, B]] = new show[lab[A, B]] {
  val `Ceta.shows_prec` = (a: nat, b: lab[A, B], c: List[Char]) =>
    (shows_prec_lab[A, B](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[lab[A, B]], b: List[Char]) =>
    (shows_list_lab[A, B](a)).apply(b)
}

def check_inj[A: equal: show,
               B: equal: show](a: lab[A, B], nn: nat,
                                sml: List[((lab[A, B], nat), List[lab[A, B]])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val symbols: List[List[(lab[A, B], nat)]] =
      mapa[((lab[A, B], nat), List[lab[A, B]]),
            List[(lab[A, B],
                   nat)]]((aa: ((lab[A, B], nat), List[lab[A, B]])) =>
                            {
                              val (ab, b): ((lab[A, B], nat), List[lab[A, B]]) =
                                aa;
                              ({
                                 val (_, n): (lab[A, B], nat) = ab;
                                 (fs: List[lab[A, B]]) =>
                                   mapa[(lab[A, B], nat),
 (lab[A, B],
   nat)]((ac: (lab[A, B], nat)) =>
           {
             val (g, i): (lab[A, B], nat) = ac;
             (g, plus_nat(n, times_nat(i, minus_nat(nn, one_nat))))
           },
          zip[lab[A, B], nat](fs, upt(Zero_nat, size_list[lab[A, B]](fs))))
                               })(b)
                            },
                           sml)
    val fsymbols: List[(lab[A, B], nat)] = concat[(lab[A, B], nat)](symbols);
    binda[(List[Char]) => List[Char], Unit,
           Unit](catcha[(lab[A, B], nat), Unit,
                         (List[Char]) =>
                           List[Char]](check_partition[(lab[A, B],
                 nat)](symbols),
(x: (lab[A, B], nat)) =>
  Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_prec_list[Char](Zero_nat,
              List('s', 'y', 'm', 'b', 'o', 'l', ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_prod[lab[A, B],
                           nat](Zero_nat, x),
           shows_prec_list[Char](Zero_nat,
                                  List(' ', 'o', 'c', 'c', 'u', 'r', 's', ' ',
't', 'w', 'i', 'c', 'e', ')')))))),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](! (membera[(lab[A, B], nat)](fsymbols, (a, nn))),
              comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ',
                     's', 'y', 'm', 'b', 'o', 'l')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_prec_lab[A,
                           B](Zero_nat, a),
            shows_prec_list[Char](Zero_nat,
                                   List(' ', 'm', 'u', 's', 't', ' ', 'n', 'o',
 't', ' ', 'o', 'c', 'c', 'u', 'r', ' ', 'a', 's', ' ', 'n', 'e', 'w', ' ', 's',
 'y', 'm', 'b', 'o', 'l'))))),
                                  (_: Unit) =>
                                    binda[(List[Char]) => List[Char], Unit,
   Unit](check[(List[Char]) =>
                 List[Char]](! (membera[(lab[A, B],
  nat)](mapa[((lab[A, B], nat), List[lab[A, B]]),
              (lab[A, B],
                nat)]((aa: ((lab[A, B], nat), List[lab[A, B]])) =>
                        fst[(lab[A, B], nat), List[lab[A, B]]](aa),
                       sml),
         (a, nn))),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_prec_list[Char](Zero_nat,
                               List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                                     'o', 'n', ' ', 's', 'y', 'm', 'b', 'o',
                                     'l')),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_lab[A, B](Zero_nat, a),
                            shows_prec_list[Char](Zero_nat,
           List(' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ',
                 'u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd'))))),
          (_: Unit) =>
            (if (less_eq_nat(nn, one_nat))
              catcha[(List[(lab[A, B], nat)], (List[Char]) => List[Char]), Unit,
                      (List[Char]) =>
                        List[Char]](forallm[List[(lab[A, B], nat)],
     (List[Char]) =>
       List[Char]]((aa: List[(lab[A, B], nat)]) =>
                     check_pairwise[(lab[A, B], nat),
                                     (List[Char]) =>
                                       List[Char]]((gn1: (lab[A, B], nat)) =>
             (gn2: (lab[A, B], nat)) =>
               check[(List[Char]) =>
                       List[Char]](! (equal_proda[lab[A, B], nat](gn1, gn2)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('s', 'y', 'm', 'b', 'o', 'l', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_prec_prod[lab[A, B], nat](Zero_nat, gn1),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'o', 'c', 'c', 'u', 'r', 's', ' ', 't', 'w', 'i',
                       'c', 'e'))))),
            aa),
                    symbols),
                                     (x:
(List[(lab[A, B], nat)], (List[Char]) => List[Char]))
                                       => Inl[(List[Char]) => List[Char],
       Unit](snd[List[(lab[A, B], nat)], (List[Char]) => List[Char]](x)))
              else Inr[Unit, (List[Char]) => List[Char]](())))))
  }

def manyof(cs: List[Char]):
      (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) => many((aa: Char) => membera[Char](cs, aa), a)

def spaces: (List[Char]) => sum[List[Char], (Unit, List[Char])] =
  (a: List[Char]) =>
    bind[Char, List[Char],
          Unit](manyof(wspace),
                 (_: List[Char]) =>
                   (ts: List[Char]) =>
                     Inr[(Unit, List[Char]), List[Char]](((), ts)),
                 a)

def xml2nd_choice[A, B,
                   C](tag: List[Char], p1: xml => sum[List[Char], A],
                       cn: List[Char], p2: xml => sum[List[Char], B],
                       f: A => Option[B] => C, x5: xml):
      sum[List[Char], C] =
  (tag, p1, cn, p2, f, x5) match {
  case (tag, p1, cn, p2, f, Xml(name, atts, cs, text)) =>
    {
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (less_eq_nat(nat(BigInt(2)), l) && is_none[List[Char]](text))))
        binda[List[Char], A,
               C](p1(nth[xml](cs, Zero_nat)),
                   (x1: A) =>
                     binda[List[Char], Boolean,
                            C](bool(cn, nth[xml](cs, one_nat)),
                                (b: Boolean) =>
                                  (if (b)
                                    binda[List[Char], B,
   C](p2(nth[xml](cs, minus_nat(l, one_nat))),
       (x2: B) => Inr[C, List[Char]]((f(x1))(Some[B](x2))))
                                    else Inr[C, List[Char]]((f(x1))(None)))))
        else no_transform[C](tag, Xml(name, atts, cs, text)))
    }
}

def alter1a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      A =
  x0 match {
  case Alter1(x) => x
}

def alter2a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      B =
  x0 match {
  case Alter2(x) => x
}

def alter3a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      C =
  x0 match {
  case Alter3(x) => x
}

def alter4a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      D =
  x0 match {
  case Alter4(x) => x
}

def alter5a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      E =
  x0 match {
  case Alter5(x) => x
}

def alter6a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      F =
  x0 match {
  case Alter6(x) => x
}

def alter7a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      G =
  x0 match {
  case Alter7(x) => x
}

def alter8a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      H =
  x0 match {
  case Alter8(x) => x
}

def alter9a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      I =
  x0 match {
  case Alter9(x) => x
}

def rat_domain: xml => sum[List[Char], rat] =
  (a: xml) =>
    xml1element[rat, rat](List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'),
                           (aa: xml) =>
                             xml1element[rat,
  rat](List('d', 'e', 'l', 't', 'a'), rat, id[rat], aa),
                           id[rat], a)

def basic_domain: xml => sum[List[Char], nat => domain] =
  (a: xml) =>
    xml_options[nat =>
                  domain](List((List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'),
                                 (aa: xml) =>
                                   xml0elements[nat =>
          domain](List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'),
                   (ab: nat) => Naturala(ab), aa)),
                                (List('a', 'r', 'c', 't', 'i', 'c'),
                                  (aa: xml) =>
                                    xml1element[nat => domain,
         nat =>
           domain](List('a', 'r', 'c', 't', 'i', 'c'),
                    (ab: xml) =>
                      xml1element[nat => domain,
                                   nat =>
                                     domain](List('d', 'o', 'm', 'a', 'i', 'n'),
      (ac: xml) =>
        xml_options[nat =>
                      domain](List((List('n', 'a', 't', 'u', 'r', 'a', 'l',
  's'),
                                     (ad: xml) =>
                                       xml0elements[nat =>
              domain](List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'),
                       (_: nat) => Arctic, ad)),
                                    (List('i', 'n', 't', 'e', 'g', 'e', 'r',
   's'),
                                      (ad: xml) =>
xml0elements[nat =>
               domain](List('i', 'n', 't', 'e', 'g', 'e', 'r', 's'),
                        (_: nat) => Arctic, ad)),
                                    (List('r', 'a', 't', 'i', 'o', 'n', 'a',
   'l', 's'),
                                      (ad: xml) =>
xml_change[rat, nat =>
                  domain](rat_domain, (r: rat) => (_: nat) => Arctic_rat(r),
                           ad))),
                               ac),
      id[nat => domain], ab),
                    id[nat => domain], aa)),
                                (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l',
                                       's'),
                                  (aa: xml) =>
                                    xml_change[rat,
        nat =>
          domain](rat_domain, (ab: rat) => (b: nat) => Rational(ab, b), aa))),
                           a)

def interpretation_type: xml => sum[List[Char], domain] =
  (a: xml) =>
    xml1element[domain,
                 domain](List('t', 'y', 'p', 'e'),
                          (aa: xml) =>
                            xml_options[domain](List((List('p', 'o', 'l', 'y',
                    'n', 'o', 'm', 'i', 'a', 'l'),
               (ab: xml) =>
                 xml2elements[nat => domain, nat,
                               domain](List('p', 'o', 'l', 'y', 'n', 'o', 'm',
     'i', 'a', 'l'),
(ac: xml) =>
  xml1element[nat => domain,
               nat =>
                 domain](List('d', 'o', 'm', 'a', 'i', 'n'),
                          (ad: xml) =>
                            xml_options[nat =>
  domain](List((List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'), basic_domain),
                (List('a', 'r', 'c', 't', 'i', 'c'), basic_domain),
                (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'),
                  basic_domain),
                (List('m', 'a', 't', 'r', 'i', 'c', 'e', 's'),
                  (ae: xml) =>
                    xml3elements[nat, nat, nat => domain,
                                  nat =>
                                    domain](List('m', 'a', 't', 'r', 'i', 'c',
          'e', 's'),
     (af: xml) => nata(List('d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'), af),
     (af: xml) =>
       nata(List('s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e', 'n', 's',
                  'i', 'o', 'n'),
             af),
     (af: xml) =>
       xml1element[nat => domain,
                    nat =>
                      domain](List('d', 'o', 'm', 'a', 'i', 'n'), basic_domain,
                               id[nat => domain], af),
     (di: nat) =>
       (sd: nat) =>
         (domain: nat => domain) =>
           (d: nat) =>
             domain(d) match {
               case Naturala(_) => Int_mat(di, sd)
               case Arctic => Arc_mat(di)
               case Arctic_rat(delta) => Arc_rat_mat(delta, di)
               case Rational(delta, _) => Rat_mat(delta, di, sd)
             },
     ae))),
           ad),
                          id[nat => domain], ac),
(ac: xml) => nata(List('d', 'e', 'g', 'r', 'e', 'e'), ac),
(typea: nat => domain) => typea, ab)),
              (List('m', 'a', 't', 'r', 'i', 'x', 'I', 'n', 't', 'e', 'r', 'p',
                     'r', 'e', 't', 'a', 't', 'i', 'o', 'n'),
                (ab: xml) =>
                  xml3elements[nat => domain, nat, nat,
                                domain](List('m', 'a', 't', 'r', 'i', 'x', 'I',
      'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n'),
 (ac: xml) =>
   xml1element[nat => domain,
                nat =>
                  domain](List('d', 'o', 'm', 'a', 'i', 'n'), basic_domain,
                           id[nat => domain], ac),
 (ac: xml) => nata(List('d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'), ac),
 (ac: xml) =>
   nata(List('s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e', 'n', 's', 'i',
              'o', 'n'),
         ac),
 (domain: nat => domain) =>
   (di: nat) =>
     (sd: nat) =>
       domain(Zero_nat) match {
         case Naturala(_) => Int_mat(di, sd)
         case Arctic => Arc_mat(di)
         case Arctic_rat(delta) => Arc_rat_mat(delta, di)
         case Rational(delta, _) => Rat_mat(delta, di, sd)
       },
 ab))),
         aa),
                          id[domain], a)

def polynomial[A](xml2coeff: xml => sum[List[Char], A], x: xml):
      sum[List[Char], tpoly[nat, A]] =
  xml1element[tpoly[nat, A],
               tpoly[nat, A]](List('p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a',
                                    'l'),
                               (a: xml) =>
                                 xml_options[tpoly[nat,
            A]](List((List('c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n',
                            't'),
                       (aa: xml) =>
                         xml1element[A, tpoly[nat,
       A]](List('c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't'),
            xml2coeff, (ab: A) => PNum[A, nat](ab), aa)),
                      (List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'),
                        (aa: xml) =>
                          xml_change[nat, tpoly[nat,
         A]]((ab: xml) =>
               nata(List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'), ab),
              (n: nat) => PVar[nat, A](minus_nat(n, one_nat)), aa)),
                      (List('s', 'u', 'm'),
                        (aa: xml) =>
                          xml_many_elements[tpoly[nat, A],
     tpoly[nat, A]](List('s', 'u', 'm'),
                     (ab: xml) => polynomial[A](xml2coeff, ab),
                     (ab: List[tpoly[nat, A]]) => PSum[nat, A](ab), aa)),
                      (List('p', 'r', 'o', 'd', 'u', 'c', 't'),
                        (aa: xml) =>
                          xml_many_elements[tpoly[nat, A],
     tpoly[nat, A]](List('p', 'r', 'o', 'd', 'u', 'c', 't'),
                     (ab: xml) => polynomial[A](xml2coeff, ab),
                     (ab: List[tpoly[nat, A]]) => PMult[nat, A](ab), aa))),
                 a),
                               id[tpoly[nat, A]], x)

def int_coeff: xml => sum[List[Char], BigInt] =
  (a: xml) => int(List('i', 'n', 't', 'e', 'g', 'e', 'r'), a)

def sum_lpoly[A](uu: A => A => A, x1: List[A], ys: List[A]): List[A] =
  (uu, x1, ys) match {
  case (uu, Nil, ys) => ys
  case (uv, v :: va, Nil) => v :: va
  case (pl, x :: xs, y :: ys) => (pl(x))(y) :: sum_lpoly[A](pl, xs, ys)
}

def lpoly_of[A: equal](uu: A, uv: A, uw: A => A => A, ux: A => A => A,
                        x4: tpoly[nat, A]):
      sum[List[Char], (A, List[A])] =
  (uu, uv, uw, ux, x4) match {
  case (uu, uv, uw, ux, PNum(i)) => Inr[(A, List[A]), List[Char]]((i, Nil))
  case (ze, on, pl, ti, PVar(x)) =>
    Inr[(A, List[A]), List[Char]]((ze, replicate[A](x, ze) ++ List(on)))
  case (ze, on, pl, ti, PSum(Nil)) => Inr[(A, List[A]), List[Char]]((ze, Nil))
  case (ze, on, pl, ti, PSum(p :: ps)) =>
    binda[List[Char], (A, List[A]),
           (A, List[A])](lpoly_of[A](ze, on, pl, ti, p),
                          (a: (A, List[A])) =>
                            {
                              val (cp, ncp): (A, List[A]) = a;
                              binda[List[Char], (A, List[A]),
                                     (A, List[A])](lpoly_of[A](ze, on, pl, ti,
                        PSum[nat, A](ps)),
            (aa: (A, List[A])) =>
              {
                val (cq, ncq): (A, List[A]) = aa;
                Inr[(A, List[A]),
                     List[Char]](((pl(cp))(cq), sum_lpoly[A](pl, ncp, ncq)))
              })
                            })
  case (ze, on, pl, ti, PMult(Nil)) => Inr[(A, List[A]), List[Char]]((on, Nil))
  case (ze, on, pl, ti, PMult(p :: ps)) =>
    binda[List[Char], (A, List[A]),
           (A, List[A])](lpoly_of[A](ze, on, pl, ti, p),
                          (a: (A, List[A])) =>
                            {
                              val (cp, ncp): (A, List[A]) = a;
                              binda[List[Char], (A, List[A]),
                                     (A, List[A])](lpoly_of[A](ze, on, pl, ti,
                        PMult[nat, A](ps)),
            (aa: (A, List[A])) =>
              {
                val (cq, ncq): (A, List[A]) = aa;
                (if (list_all[A]((ab: A) => eq[A](ze, ab), ncp))
                  Inr[(A, List[A]),
                       List[Char]](((ti(cp))(cq), mapa[A, A](ti(cp), ncq)))
                  else (if (list_all[A]((ab: A) => eq[A](ze, ab), ncq))
                         Inr[(A, List[A]),
                              List[Char]](((ti(cp))(cq),
    mapa[A, A]((x: A) => (ti(x))(cq), ncp)))
                         else Inl[List[Char],
                                   (A, List[A])](List('c', 'a', 'n', 'n', 'o',
               't', ' ', 't', 'r', 'a', 'n', 's', 'f', 'o', 'r', 'm', ' ', 'n',
               'o', 'n', '-', 'l', 'i', 'n', 'e', 'a', 'r', ' ', 'p', 'o', 'l',
               'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 't', 'o', ' ', 'l', 'i',
               'n', 'e', 'a', 'r', ' ', 'p', 'o', 'l', 'y', 'n', 'o', 'm', 'i',
               'a', 'l'))))
              })
                            })
}

def get_comp_0[A](x0: (A => Boolean,
                        (A, (A, (A, (A => A => A,
                                      (A => A => A,
(A => A,
  (A => A => Boolean,
    (A => A => Boolean,
      (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => ze
}

def max0_int(x: BigInt): BigInt = max[BigInt].apply(BigInt(0)).apply(x)

def ge_int(x: BigInt, y: BigInt): Boolean = y <= x

def int_name: List[Char] =
  List('n', 'a', 't', 'u', 'r', 'a', 'l', 's', ' ', 'w', 'i', 't', 'h', ' ',
        'n', 'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't',
        'a', 'n', 't', 's')

def int_mono(x: BigInt): Boolean = BigInt(1) <= x

def get_comp_1[A](x0: (A => Boolean,
                        (A, (A, (A, (A => A => A,
                                      (A => A => A,
(A => A,
  (A => A => Boolean,
    (A => A => Boolean,
      (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => on
}

def get_comp_plus[A](x0: (A => Boolean,
                           (A, (A, (A, (A => A => A,
 (A => A => A,
   (A => A,
     (A => A => Boolean,
       (A => A => Boolean,
         (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => A => A =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => pl
}

def get_comp_times[A](x0: (A => Boolean,
                            (A, (A, (A, (A => A => A,
  (A => A => A,
    (A => A,
      (A => A => Boolean,
        (A => A => Boolean,
          (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => A => A =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => ti
}

def fit_length[A](uu: A, x1: nat, uv: List[A]): List[A] = (uu, x1, uv) match {
  case (uu, Zero_nat, uv) => Nil
  case (a, Suc(v), Nil) => replicate[A](Suc(v), a)
  case (a, Suc(n), b :: as) => b :: fit_length[A](a, n, as)
}

trait ab_semigroup_mult[A] extends semigroup_mult[A] {
}

trait comm_monoid_mult[A] extends ab_semigroup_mult[A] with monoid_mult[A] {
}

trait comm_semiring[A] extends ab_semigroup_mult[A] with semiring[A] {
}

trait comm_semiring_0[A] extends comm_semiring[A] with semiring_0[A] {
}

trait comm_semiring_1[A]
  extends comm_monoid_mult[A] with comm_semiring_0[A] with dvd[A] with
    semiring_1[A]
  {
}

def zero_poly[A, B]: List[(List[(A, nat)], B)] = Nil

def eq_monom[A: equal](x0: List[(A, nat)], n: List[(A, nat)]): Boolean = (x0, n)
  match {
  case (Nil, n) => nulla[(A, nat)](n)
  case ((x, p) :: m, n) =>
    extract[(A, nat)]((yq: (A, nat)) => eq[A](fst[A, nat](yq), x), n) match {
      case None => false
      case Some((n1, ((_, q), n2))) =>
        equal_nata(p, q) && eq_monom[A](m, n1 ++ n2)
    }
}

def poly_add[A: equal,
              B: equal: semiring_0](x0: List[(List[(A, nat)], B)],
                                     q: List[(List[(A, nat)], B)]):
      List[(List[(A, nat)], B)] =
  (x0, q) match {
  case (Nil, q) => q
  case ((m, c) :: p, q) =>
    extract[(List[(A, nat)],
              B)]((mc: (List[(A, nat)], B)) =>
                    eq_monom[A](fst[List[(A, nat)], B](mc), m),
                   q)
      match {
      case None => (m, c) :: poly_add[A, B](p, q)
      case Some((q1, ((_, d), q2))) =>
        (if (eq[B](plus[B](c, d), zero[B])) poly_add[A, B](p, q1 ++ q2)
          else (m, plus[B](c, d)) :: poly_add[A, B](p, q1 ++ q2))
    }
}

def one_poly[A, B: semiring_1]: List[(List[(A, nat)], B)] = List((Nil, one[B]))

def monom_mult[A: equal](x0: List[(A, nat)], n: List[(A, nat)]):
      List[(A, nat)] =
  (x0, n) match {
  case (Nil, n) => n
  case ((x, p) :: m, n) =>
    extract[(A, nat)]((yq: (A, nat)) => eq[A](fst[A, nat](yq), x), n) match {
      case None => (x, p) :: monom_mult[A](m, n)
      case Some((n1, ((_, q), n2))) =>
        (x, plus_nat(p, q)) :: monom_mult[A](m, n1 ++ n2)
    }
}

def monom_mult_poly[A: equal,
                     B: equal: semiring_0](uu: (List[(A, nat)], B),
    x1: List[(List[(A, nat)], B)]):
      List[(List[(A, nat)], B)] =
  (uu, x1) match {
  case (uu, Nil) => Nil
  case ((ma, c), (m, d) :: p) =>
    (if (eq[B](times[B](c, d), zero[B])) monom_mult_poly[A, B]((ma, c), p)
      else (monom_mult[A](ma, m), times[B](c, d)) ::
             monom_mult_poly[A, B]((ma, c), p))
}

def poly_mult[A: equal,
               B: equal: semiring_0](x0: List[(List[(A, nat)], B)],
                                      q: List[(List[(A, nat)], B)]):
      List[(List[(A, nat)], B)] =
  (x0, q) match {
  case (Nil, q) => Nil
  case (mc :: p, q) =>
    poly_add[A, B](monom_mult_poly[A, B](mc, q), poly_mult[A, B](p, q))
}

def poly_of[A: equal, B: equal: comm_semiring_1](x0: tpoly[A, B]):
      List[(List[(A, nat)], B)] =
  x0 match {
  case PNum(i) => (if (eq[B](i, zero[B])) Nil else List((Nil, i)))
  case PVar(x) => List((List((x, one_nat)), one[B]))
  case PSum(Nil) => zero_poly[A, B]
  case PSum(p :: ps) =>
    poly_add[A, B](poly_of[A, B](p), poly_of[A, B](PSum[A, B](ps)))
  case PMult(Nil) => one_poly[A, B]
  case PMult(p :: ps) =>
    poly_mult[A, B](poly_of[A, B](p), poly_of[A, B](PMult[A, B](ps)))
}

implicit def ab_semigroup_mult_int: ab_semigroup_mult[BigInt] = new
  ab_semigroup_mult[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def comm_monoid_mult_int: comm_monoid_mult[BigInt] = new
  comm_monoid_mult[BigInt] {
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def comm_semiring_int: comm_semiring[BigInt] = new
  comm_semiring[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def comm_semiring_0_int: comm_semiring_0[BigInt] = new
  comm_semiring_0[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def comm_semiring_1_int: comm_semiring_1[BigInt] = new
  comm_semiring_1[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

def arctic_coeff: xml => sum[List[Char], arctic] =
  (a: xml) =>
    xml_options[arctic](List((List('i', 'n', 't', 'e', 'g', 'e', 'r'),
                               (aa: xml) =>
                                 xml_change[BigInt,
     arctic]((ab: xml) => int(List('i', 'n', 't', 'e', 'g', 'e', 'r'), ab),
              (ab: BigInt) => Num_arc(ab), aa)),
                              (List('m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i',
                                     'n', 'i', 't', 'y'),
                                (aa: xml) =>
                                  xml0elements[arctic](List('m', 'i', 'n', 'u',
                     's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y'),
                MinInfty, aa))),
                         a)

def equal_arctica(x0: arctic, x1: arctic): Boolean = (x0, x1) match {
  case (Num_arc(int), MinInfty) => false
  case (MinInfty, Num_arc(int)) => false
  case (Num_arc(inta), Num_arc(int)) => inta == int
  case (MinInfty, MinInfty) => true
}

implicit def equal_arctic: equal[arctic] = new equal[arctic] {
  val `Ceta.equal` = (a: arctic, b: arctic) => equal_arctica(a, b)
}

def zero_arctic: arctic = MinInfty

def one_arctic: arctic = Num_arc(BigInt(0))

def plus_arctic(x0: arctic, y: arctic): arctic = (x0, y) match {
  case (MinInfty, y) => y
  case (Num_arc(v), MinInfty) => Num_arc(v)
  case (Num_arc(x), Num_arc(y)) => Num_arc(max[BigInt].apply(x).apply(y))
}

def times_arctic(x0: arctic, y: arctic): arctic = (x0, y) match {
  case (MinInfty, y) => MinInfty
  case (Num_arc(v), MinInfty) => MinInfty
  case (Num_arc(x), Num_arc(y)) => Num_arc(x + y)
}

def ge_arctic(x: arctic, xa1: arctic): Boolean = (x, xa1) match {
  case (x, MinInfty) => true
  case (MinInfty, Num_arc(uu)) => false
  case (Num_arc(x), Num_arc(y)) => y <= x
}

def gt_arctic(uu: arctic, x1: arctic): Boolean = (uu, x1) match {
  case (uu, MinInfty) => true
  case (MinInfty, Num_arc(uv)) => false
  case (Num_arc(x), Num_arc(y)) => y < x
}

def arctic_name: List[Char] =
  List('a', 'r', 'c', 't', 'i', 'c', ' ', 's', 'e', 'm', 'i', 'r', 'i', 'n',
        'g')

def pos_arctic(x0: arctic): Boolean = x0 match {
  case MinInfty => false
  case Num_arc(n) => BigInt(0) <= n
}

def arctic_rat_coeff: xml => sum[List[Char], arctic_rat] =
  (a: xml) =>
    xml_options[arctic_rat](List((List('i', 'n', 't', 'e', 'g', 'e', 'r'),
                                   (aa: xml) =>
                                     xml_change[rat,
         arctic_rat](rat, (ab: rat) => Num_arc_rat(ab), aa)),
                                  (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l'),
                                    (aa: xml) =>
                                      xml_change[rat,
          arctic_rat](rat, (ab: rat) => Num_arc_rat(ab), aa)),
                                  (List('m', 'i', 'n', 'u', 's', 'I', 'n', 'f',
 'i', 'n', 'i', 't', 'y'),
                                    (aa: xml) =>
                                      xml0elements[arctic_rat](List('m', 'i',
                             'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't',
                             'y'),
                        MinInfty_rat, aa))),
                             a)

def equal_arctic_rata(x0: arctic_rat, x1: arctic_rat): Boolean = (x0, x1) match
  {
  case (Num_arc_rat(rat), MinInfty_rat) => false
  case (MinInfty_rat, Num_arc_rat(rat)) => false
  case (Num_arc_rat(rata), Num_arc_rat(rat)) => equal_rata(rata, rat)
  case (MinInfty_rat, MinInfty_rat) => true
}

implicit def equal_arctic_rat: equal[arctic_rat] = new equal[arctic_rat] {
  val `Ceta.equal` = (a: arctic_rat, b: arctic_rat) => equal_arctic_rata(a, b)
}

def zero_arctic_rat: arctic_rat = MinInfty_rat

def one_arctic_rat: arctic_rat = Num_arc_rat(fract(BigInt(0), BigInt(1)))

def plus_arctic_rat(x0: arctic_rat, y: arctic_rat): arctic_rat = (x0, y) match {
  case (MinInfty_rat, y) => y
  case (Num_arc_rat(v), MinInfty_rat) => Num_arc_rat(v)
  case (Num_arc_rat(x), Num_arc_rat(y)) =>
    Num_arc_rat(max[rat].apply(x).apply(y))
}

def times_arctic_rat(x0: arctic_rat, y: arctic_rat): arctic_rat = (x0, y) match
  {
  case (MinInfty_rat, y) => MinInfty_rat
  case (Num_arc_rat(v), MinInfty_rat) => MinInfty_rat
  case (Num_arc_rat(x), Num_arc_rat(y)) => Num_arc_rat(plus_rata(x, y))
}

def ge_arctic_rat(x: arctic_rat, xa1: arctic_rat): Boolean = (x, xa1) match {
  case (x, MinInfty_rat) => true
  case (MinInfty_rat, Num_arc_rat(uu)) => false
  case (Num_arc_rat(x), Num_arc_rat(y)) => less_eq_rat(y, x)
}

def rat_gt(delta: rat): rat => rat => Boolean =
  (x: rat) => (y: rat) => less_eq_rat(delta, minus_rat(x, y))

def gt_arctic_rat(delta: rat, uu: arctic_rat, x2: arctic_rat): Boolean =
  (delta, uu, x2) match {
  case (delta, uu, MinInfty_rat) => true
  case (delta, MinInfty_rat, Num_arc_rat(uv)) => false
  case (delta, Num_arc_rat(x), Num_arc_rat(y)) =>
    (rat_gt(delta)).apply(x).apply(y)
}

def arctic_rat_name: List[Char] =
  List('a', 'r', 'c', 't', 'i', 'c', ' ', 'r', 'a', 't', 'i', 'o', 'n', 'a',
        'l', 's')

def pos_arctic_rat(x0: arctic_rat): Boolean = x0 match {
  case MinInfty_rat => false
  case Num_arc_rat(n) => less_eq_rat(fract(BigInt(0), BigInt(1)), n)
}

def vec_coeff[A](xml2coeff: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[A]] =
  (a: xml) =>
    xml_many_elements[A, List[A]](List('v', 'e', 'c', 't', 'o', 'r'),
                                   (aa: xml) =>
                                     xml1element[A,
          A](List('c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't'),
              xml2coeff, id[A], aa),
                                   id[List[A]], a)

def mat_coeff[A](ze: A, xml2coeff: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[List[A]]] =
  (a: xml) =>
    xml_options[List[List[A]]](List((List('m', 'a', 't', 'r', 'i', 'x'),
                                      (aa: xml) =>
xml_many_elements[List[A],
                   List[List[A]]](List('m', 'a', 't', 'r', 'i', 'x'),
                                   vec_coeff[A](xml2coeff), id[List[List[A]]],
                                   aa)),
                                     (List('v', 'e', 'c', 't', 'o', 'r'),
                                       (aa: xml) =>
 xml_change[List[A],
             List[List[A]]](vec_coeff[A](xml2coeff),
                             (v: List[A]) =>
                               v :: replicate[List[A]](minus_nat(size_list[A](v),
                          one_nat),
                replicate[A](size_list[A](v), ze)),
                             aa))),
                                a)

def get_comp_def[A](x0: (A => Boolean,
                          (A, (A, (A, (A => A => A,
(A => A => A,
  (A => A,
    (A => A => Boolean,
      (A => A => Boolean,
        (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => defa
}

implicit def non_strict_order_int: non_strict_order[BigInt] = new
  non_strict_order[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
}

implicit def ordered_ab_semigroup_int: ordered_ab_semigroup[BigInt] = new
  ordered_ab_semigroup[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def max_ordered_monoid_add_int: max_ordered_monoid_add[BigInt] = new
  max_ordered_monoid_add[BigInt] {
  val `Ceta.max0` = (a: BigInt) => max0_int(a)
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
}

def get_comp_gt[A](x0: (A => Boolean,
                         (A, (A, (A, (A => A => A,
                                       (A => A => A,
 (A => A,
   (A => A => Boolean,
     (A => A => Boolean,
       (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => A => Boolean =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => gt
}

def get_comp_mono[A](x0: (A => Boolean,
                           (A, (A, (A, (A => A => A,
 (A => A => A,
   (A => A,
     (A => A => Boolean,
       (A => A => Boolean,
         (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      Boolean =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => mo
}

def get_comp_description[A](x0: (A => Boolean,
                                  (A, (A,
(A, (A => A => A,
      (A => A => A,
        (A => A,
          (A => A => Boolean,
            (A => A => Boolean,
              (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      List[Char] =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => de
}

def mat_mono[A](sd: nat, m: A => Boolean): (List[List[A]]) => Boolean =
  (a: List[List[A]]) => mat_monoI[A](m, sd, a)

def get_comp_check_mono[A](x0: (A => Boolean,
                                 (A, (A, (A,
   (A => A => A,
     (A => A => A,
       (A => A,
         (A => A => Boolean,
           (A => A => Boolean,
             (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => Boolean =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => cm
}

implicit def non_strict_order_arctic: non_strict_order[arctic] = new
  non_strict_order[arctic] {
  val `Ceta.ge` = (a: arctic, b: arctic) => ge_arctic(a, b)
}

def mat_arc_posI[A](ap: A => Boolean, m: List[List[A]]): Boolean =
  ap(nth[A](nth[List[A]](m, Zero_nat), Zero_nat))

def get_comp_apos[A](x0: (A => Boolean,
                           (A, (A, (A, (A => A => A,
 (A => A => A,
   (A => A,
     (A => A => Boolean,
       (A => A => Boolean,
         (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => Boolean =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => ap
}

implicit def non_strict_order_arctic_rat: non_strict_order[arctic_rat] = new
  non_strict_order[arctic_rat] {
  val `Ceta.ge` = (a: arctic_rat, b: arctic_rat) => ge_arctic_rat(a, b)
}

def max0_rat(x: rat): rat = max[rat].apply(fract(BigInt(0), BigInt(1))).apply(x)

def ge_rat(x: rat, y: rat): Boolean = less_eq_rat(y, x)

def rat_name: List[Char] =
  List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's', ' ', '(', 'w', 'i', 't',
        'h', ' ', 'n', 'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n',
        's', 't', 'a', 'n', 't', 's', ')')

def rat_mono(x: rat): Boolean = less_eq_rat(fract(BigInt(1), BigInt(1)), x)

implicit def ab_semigroup_mult_rat: ab_semigroup_mult[rat] = new
  ab_semigroup_mult[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def comm_monoid_mult_rat: comm_monoid_mult[rat] = new
  comm_monoid_mult[rat] {
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def comm_semiring_rat: comm_semiring[rat] = new comm_semiring[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def comm_semiring_0_rat: comm_semiring_0[rat] = new
  comm_semiring_0[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def comm_semiring_1_rat: comm_semiring_1[rat] = new
  comm_semiring_1[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def non_strict_order_rat: non_strict_order[rat] = new
  non_strict_order[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
}

implicit def ordered_ab_semigroup_rat: ordered_ab_semigroup[rat] = new
  ordered_ab_semigroup[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def max_ordered_monoid_add_rat: max_ordered_monoid_add[rat] = new
  max_ordered_monoid_add[rat] {
  val `Ceta.max0` = (a: rat) => max0_rat(a)
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
}

def alter10a[A, B, C, D, E, F, G, H, I,
              J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      J =
  x0 match {
  case Alter10(x) => x
}

def interpretation[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redpair[A]] =
  (a: xml) =>
    xml1many_elements_gen[domain,
                           alter[((A, nat), (BigInt, List[BigInt])),
                                  ((A, nat), (rat, List[rat])),
                                  ((A, nat), (arctic, List[arctic])),
                                  ((A, nat),
                                    (List[List[BigInt]],
                                      List[List[List[BigInt]]])),
                                  ((A, nat),
                                    (List[List[rat]], List[List[List[rat]]])),
                                  ((A, nat),
                                    (List[List[arctic]],
                                      List[List[List[arctic]]])),
                                  ((A, nat), (arctic_rat, List[arctic_rat])),
                                  ((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]])),
                                  ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                  ((A, nat), List[(List[(nat, nat)], rat)])],
                           redpair[A]](List('i', 'n', 't', 'e', 'r', 'p', 'r',
     'e', 't', 'a', 't', 'i', 'o', 'n'),
interpretation_type,
(aa: domain) =>
  aa match {
    case Naturala(deg) =>
      (if (less_eq_nat(deg, one_nat))
        (ab: xml) =>
          xml3elements[A, nat, (BigInt, List[BigInt]),
                        alter[((A, nat), (BigInt, List[BigInt])),
                               ((A, nat), (rat, List[rat])),
                               ((A, nat), (arctic, List[arctic])),
                               ((A, nat),
                                 (List[List[BigInt]],
                                   List[List[List[BigInt]]])),
                               ((A, nat),
                                 (List[List[rat]], List[List[List[rat]]])),
                               ((A, nat),
                                 (List[List[arctic]],
                                   List[List[List[arctic]]])),
                               ((A, nat), (arctic_rat, List[arctic_rat])),
                               ((A, nat),
                                 (List[List[arctic_rat]],
                                   List[List[List[arctic_rat]]])),
                               ((A, nat), List[(List[(nat, nat)], BigInt)]),
                               ((A, nat),
                                 List[(List[(nat, nat)],
rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
          (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
          (x: xml) =>
            binda[List[Char], tpoly[nat, BigInt],
                   (BigInt,
                     List[BigInt])](polynomial[BigInt](int_coeff, x),
                                     (ac: tpoly[nat, BigInt]) =>
                                       lpoly_of[BigInt](get_comp_0[BigInt](((_:
                                       BigInt)
                                      => true,
                                     (BigInt(0),
                                       (BigInt(1),
 (BigInt(1),
   ((ad: BigInt) => (b: BigInt) => ad + b,
     ((ad: BigInt) => (b: BigInt) => ad * b,
       ((ad: BigInt) => max0_int(ad),
         ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
           ((xa: BigInt) => (y: BigInt) => y < xa,
             (true,
               (int_name,
                 ((_: BigInt) => true,
                   (ad: BigInt) => int_mono(ad)))))))))))))),
                 get_comp_1[BigInt](((_: BigInt) => true,
                                      (BigInt(0),
(BigInt(1),
  (BigInt(1),
    ((ad: BigInt) => (b: BigInt) => ad + b,
      ((ad: BigInt) => (b: BigInt) => ad * b,
        ((ad: BigInt) => max0_int(ad),
          ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
            ((xa: BigInt) => (y: BigInt) => y < xa,
              (true,
                (int_name,
                  ((_: BigInt) => true,
                    (ad: BigInt) => int_mono(ad)))))))))))))),
                 get_comp_plus[BigInt](((_: BigInt) => true,
 (BigInt(0),
   (BigInt(1),
     (BigInt(1),
       ((ad: BigInt) => (b: BigInt) => ad + b,
         ((ad: BigInt) => (b: BigInt) => ad * b,
           ((ad: BigInt) => max0_int(ad),
             ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
               ((xa: BigInt) => (y: BigInt) => y < xa,
                 (true,
                   (int_name,
                     ((_: BigInt) => true,
                       (ad: BigInt) => int_mono(ad)))))))))))))),
                 get_comp_times[BigInt](((_: BigInt) => true,
  (BigInt(0),
    (BigInt(1),
      (BigInt(1),
        ((ad: BigInt) => (b: BigInt) => ad + b,
          ((ad: BigInt) => (b: BigInt) => ad * b,
            ((ad: BigInt) => max0_int(ad),
              ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
                ((xa: BigInt) => (y: BigInt) => y < xa,
                  (true,
                    (int_name,
                      ((_: BigInt) => true,
                        (ad: BigInt) => int_mono(ad)))))))))))))),
                 ac)),
          (f: A) =>
            (ac: nat) =>
              (poly: (BigInt, List[BigInt])) =>
                Alter1[((A, nat), (BigInt, List[BigInt])),
                        ((A, nat), (rat, List[rat])),
                        ((A, nat), (arctic, List[arctic])),
                        ((A, nat),
                          (List[List[BigInt]], List[List[List[BigInt]]])),
                        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                        ((A, nat),
                          (List[List[arctic]], List[List[List[arctic]]])),
                        ((A, nat), (arctic_rat, List[arctic_rat])),
                        ((A, nat),
                          (List[List[arctic_rat]],
                            List[List[List[arctic_rat]]])),
                        ((A, nat), List[(List[(nat, nat)], BigInt)]),
                        ((A, nat),
                          List[(List[(nat, nat)],
                                 rat)])](((f, ac),
   (fst[BigInt, List[BigInt]](poly),
     fit_length[BigInt](get_comp_0[BigInt](((_: BigInt) => true,
     (BigInt(0),
       (BigInt(1),
         (BigInt(1),
           ((ad: BigInt) => (b: BigInt) => ad + b,
             ((ad: BigInt) => (b: BigInt) => ad * b,
               ((ad: BigInt) => max0_int(ad),
                 ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
                   ((x: BigInt) => (y: BigInt) => y < x,
                     (true,
                       (int_name,
                         ((_: BigInt) => true,
                           (ad: BigInt) => int_mono(ad)))))))))))))),
                         ac, snd[BigInt, List[BigInt]](poly))))),
          ab)
        else (ab: xml) =>
               xml3elements[A, nat, List[(List[(nat, nat)], BigInt)],
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 binda[List[Char], tpoly[nat, BigInt],
                        List[(List[(nat, nat)],
                               BigInt)]](polynomial[BigInt](int_coeff, x),
  (p: tpoly[nat, BigInt]) =>
    Inr[List[(List[(nat, nat)], BigInt)], List[Char]](poly_of[nat, BigInt](p))),
               (f: A) =>
                 (ac: nat) =>
                   (poly: List[(List[(nat, nat)], BigInt)]) =>
                     Alter9[((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat),
                               List[(List[(nat, nat)], rat)])](((f, ac), poly)),
               ab))
    case Arctic =>
      (ab: xml) =>
        xml3elements[A, nat, (arctic, List[arctic]),
                      alter[((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])]](List('i', 'n', 't', 'e', 'r',
             'p', 'r', 'e', 't'),
        xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
        (x: xml) =>
          binda[List[Char], tpoly[nat, arctic],
                 (arctic,
                   List[arctic])](polynomial[arctic](arctic_coeff, x),
                                   (ac: tpoly[nat, arctic]) =>
                                     lpoly_of[arctic](get_comp_0[arctic](((_:
                                     arctic)
                                    => true,
                                   (zero_arctic,
                                     (one_arctic,
                                       (one_arctic,
 ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
   ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
     (id[arctic],
       ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
         ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
           (false,
             (arctic_name,
               ((ad: arctic) => pos_arctic(ad),
                 (_: arctic) => false))))))))))))),
               get_comp_1[arctic](((_: arctic) => true,
                                    (zero_arctic,
                                      (one_arctic,
(one_arctic,
  ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
    ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
      (id[arctic],
        ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
          ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
            (false,
              (arctic_name,
                ((ad: arctic) => pos_arctic(ad),
                  (_: arctic) => false))))))))))))),
               get_comp_plus[arctic](((_: arctic) => true,
                                       (zero_arctic,
 (one_arctic,
   (one_arctic,
     ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
       ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
         (id[arctic],
           ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
             ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
               (false,
                 (arctic_name,
                   ((ad: arctic) => pos_arctic(ad),
                     (_: arctic) => false))))))))))))),
               get_comp_times[arctic](((_: arctic) => true,
(zero_arctic,
  (one_arctic,
    (one_arctic,
      ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
        ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
          (id[arctic],
            ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
              ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
                (false,
                  (arctic_name,
                    ((ad: arctic) => pos_arctic(ad),
                      (_: arctic) => false))))))))))))),
               ac)),
        (f: A) =>
          (ac: nat) =>
            (poly: (arctic, List[arctic])) =>
              Alter3[((A, nat), (arctic, List[arctic])),
                      ((A, nat), (BigInt, List[BigInt])),
                      ((A, nat), (rat, List[rat])),
                      ((A, nat),
                        (List[List[BigInt]], List[List[List[BigInt]]])),
                      ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                      ((A, nat),
                        (List[List[arctic]], List[List[List[arctic]]])),
                      ((A, nat), (arctic_rat, List[arctic_rat])),
                      ((A, nat),
                        (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                      ((A, nat), List[(List[(nat, nat)], BigInt)]),
                      ((A, nat),
                        List[(List[(nat, nat)],
                               rat)])](((f, ac),
 (fst[arctic, List[arctic]](poly),
   fit_length[arctic](get_comp_0[arctic](((_: arctic) => true,
   (zero_arctic,
     (one_arctic,
       (one_arctic,
         ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
           ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
             (id[arctic],
               ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
                 ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
                   (false,
                     (arctic_name,
                       ((ad: arctic) => pos_arctic(ad),
                         (_: arctic) => false))))))))))))),
                       ac, snd[arctic, List[arctic]](poly))))),
        ab)
    case Arctic_rat(d) =>
      (ab: xml) =>
        xml3elements[A, nat, (arctic_rat, List[arctic_rat]),
                      alter[((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])]](List('i', 'n', 't', 'e', 'r',
             'p', 'r', 'e', 't'),
        xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
        (x: xml) =>
          binda[List[Char], tpoly[nat, arctic_rat],
                 (arctic_rat,
                   List[arctic_rat])](polynomial[arctic_rat](arctic_rat_coeff,
                      x),
                                       (ac: tpoly[nat, arctic_rat]) =>
 lpoly_of[arctic_rat](get_comp_0[arctic_rat](((_: arctic_rat) => true,
       (zero_arctic_rat,
         (one_arctic_rat,
           (one_arctic_rat,
             ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
               ((ad: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ad, b),
                 (id[arctic_rat],
                   ((ad: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ad, b),
                     ((ad: arctic_rat) =>
                        (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                       (false,
                         (arctic_rat_name,
                           ((ad: arctic_rat) => pos_arctic_rat(ad),
                             (_: arctic_rat) => false))))))))))))),
                       get_comp_1[arctic_rat](((_: arctic_rat) => true,
        (zero_arctic_rat,
          (one_arctic_rat,
            (one_arctic_rat,
              ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
                ((ad: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ad, b),
                  (id[arctic_rat],
                    ((ad: arctic_rat) =>
                       (b: arctic_rat) => ge_arctic_rat(ad, b),
                      ((ad: arctic_rat) =>
                         (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                        (false,
                          (arctic_rat_name,
                            ((ad: arctic_rat) => pos_arctic_rat(ad),
                              (_: arctic_rat) => false))))))))))))),
                       get_comp_plus[arctic_rat](((_: arctic_rat) => true,
           (zero_arctic_rat,
             (one_arctic_rat,
               (one_arctic_rat,
                 ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
                   ((ad: arctic_rat) =>
                      (b: arctic_rat) => times_arctic_rat(ad, b),
                     (id[arctic_rat],
                       ((ad: arctic_rat) =>
                          (b: arctic_rat) => ge_arctic_rat(ad, b),
                         ((ad: arctic_rat) =>
                            (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                           (false,
                             (arctic_rat_name,
                               ((ad: arctic_rat) => pos_arctic_rat(ad),
                                 (_: arctic_rat) => false))))))))))))),
                       get_comp_times[arctic_rat](((_: arctic_rat) => true,
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  ((ad: arctic_rat) =>
                     (b: arctic_rat) => plus_arctic_rat(ad, b),
                    ((ad: arctic_rat) =>
                       (b: arctic_rat) => times_arctic_rat(ad, b),
                      (id[arctic_rat],
                        ((ad: arctic_rat) =>
                           (b: arctic_rat) => ge_arctic_rat(ad, b),
                          ((ad: arctic_rat) =>
                             (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                            (false,
                              (arctic_rat_name,
                                ((ad: arctic_rat) => pos_arctic_rat(ad),
                                  (_: arctic_rat) => false))))))))))))),
                       ac)),
        (f: A) =>
          (ac: nat) =>
            (poly: (arctic_rat, List[arctic_rat])) =>
              Alter7[((A, nat), (arctic_rat, List[arctic_rat])),
                      ((A, nat), (BigInt, List[BigInt])),
                      ((A, nat), (rat, List[rat])),
                      ((A, nat), (arctic, List[arctic])),
                      ((A, nat),
                        (List[List[BigInt]], List[List[List[BigInt]]])),
                      ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                      ((A, nat),
                        (List[List[arctic]], List[List[List[arctic]]])),
                      ((A, nat),
                        (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                      ((A, nat), List[(List[(nat, nat)], BigInt)]),
                      ((A, nat),
                        List[(List[(nat, nat)],
                               rat)])](((f, ac),
 (fst[arctic_rat, List[arctic_rat]](poly),
   fit_length[arctic_rat](get_comp_0[arctic_rat](((_: arctic_rat) => true,
           (zero_arctic_rat,
             (one_arctic_rat,
               (one_arctic_rat,
                 ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
                   ((ad: arctic_rat) =>
                      (b: arctic_rat) => times_arctic_rat(ad, b),
                     (id[arctic_rat],
                       ((ad: arctic_rat) =>
                          (b: arctic_rat) => ge_arctic_rat(ad, b),
                         ((ad: arctic_rat) =>
                            (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                           (false,
                             (arctic_rat_name,
                               ((ad: arctic_rat) => pos_arctic_rat(ad),
                                 (_: arctic_rat) => false))))))))))))),
                           ac, snd[arctic_rat, List[arctic_rat]](poly))))),
        ab)
    case Int_mat(n, sd) =>
      (ab: xml) =>
        xml3elements[A, nat, (List[List[BigInt]], List[List[List[BigInt]]]),
                      alter[((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])]](List('i', 'n', 't', 'e', 'r',
             'p', 'r', 'e', 't'),
        xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
        (x: xml) =>
          binda[List[Char], tpoly[nat, List[List[BigInt]]],
                 (List[List[BigInt]],
                   List[List[List[BigInt]]])](polynomial[List[List[BigInt]]](mat_coeff[BigInt](BigInt(0),
                int_coeff),
                                      x),
       (ac: tpoly[nat, List[List[BigInt]]]) =>
         lpoly_of[List[List[BigInt]]](get_comp_0[List[List[BigInt]]](((ad:
                                 List[List[BigInt]])
                                => mat[BigInt](n, n, ad),
                               (mat0I[BigInt](BigInt(0), n, n),
                                 (mat1I[BigInt](BigInt(0), BigInt(1), n),
                                   (mat1I[BigInt](BigInt(0),
           get_comp_def[BigInt](((_: BigInt) => true,
                                  (BigInt(0),
                                    (BigInt(1),
                                      (BigInt(1),
((ad: BigInt) => (b: BigInt) => ad + b,
  ((ad: BigInt) => (b: BigInt) => ad * b,
    ((ad: BigInt) => max0_int(ad),
      ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
        ((xa: BigInt) => (y: BigInt) => y < xa,
          (true,
            (int_name,
              ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))),
           n),
                                     ((ad: List[List[BigInt]]) =>
(b: List[List[BigInt]]) =>
  mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
                                       ((ad: List[List[BigInt]]) =>
  (b: List[List[BigInt]]) =>
    mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                       (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
 (mat_max0[BigInt],
   (mat_ge[BigInt],
     ((ad: List[List[BigInt]]) =>
        (b: List[List[BigInt]]) =>
          mat_gtI[BigInt](get_comp_gt[BigInt](((_: BigInt) => true,
        (BigInt(0),
          (BigInt(1),
            (BigInt(1),
              ((ae: BigInt) => (ba: BigInt) => ae + ba,
                ((ae: BigInt) => (ba: BigInt) => ae * ba,
                  ((ae: BigInt) => max0_int(ae),
                    ((ae: BigInt) => (ba: BigInt) => ge_int(ae, ba),
                      ((xa: BigInt) => (y: BigInt) => y < xa,
                        (true,
                          (int_name,
                            ((_: BigInt) => true,
                              (ae: BigInt) => int_mono(ae)))))))))))))),
                           sd, ad, b),
       (get_comp_mono[BigInt](((_: BigInt) => true,
                                (BigInt(0),
                                  (BigInt(1),
                                    (BigInt(1),
                                      ((ad: BigInt) => (b: BigInt) => ad + b,
((ad: BigInt) => (b: BigInt) => ad * b,
  ((ad: BigInt) => max0_int(ad),
    ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
      ((xa: BigInt) => (y: BigInt) => y < xa,
        (true,
          (int_name,
            ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))),
         (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'w', 'i', 't', 'h', ' ', 's', 't', 'r',
                                   'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n',
                                   's', 'i', 'o', 'n', ' '),
                             (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                             List(' ', 'o', 'v', 'e', 'r', ' '),
                             get_comp_description[BigInt](((_: BigInt) => true,
                    (BigInt(0),
                      (BigInt(1),
                        (BigInt(1),
                          ((ad: BigInt) => (b: BigInt) => ad + b,
                            ((ad: BigInt) => (b: BigInt) => ad * b,
                              ((ad: BigInt) => max0_int(ad),
                                ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
                                  ((xa: BigInt) => (y: BigInt) => y < xa,
                                    (true,
                                      (int_name,
((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))))),
           ((_: List[List[BigInt]]) => true,
             mat_mono[BigInt](sd, get_comp_check_mono[BigInt](((_: BigInt) =>
                         true,
                        (BigInt(0),
                          (BigInt(1),
                            (BigInt(1),
                              ((ad: BigInt) => (b: BigInt) => ad + b,
                                ((ad: BigInt) => (b: BigInt) => ad * b,
                                  ((ad: BigInt) => max0_int(ad),
                                    ((ad: BigInt) =>
                                       (b: BigInt) => ge_int(ad, b),
                                      ((xa: BigInt) => (y: BigInt) => y < xa,
(true,
  (int_name,
    ((_: BigInt) => true,
      (ad: BigInt) => int_mono(ad)))))))))))))))))))))))))))),
                                       get_comp_1[List[List[BigInt]]](((ad:
                                  List[List[BigInt]])
                                 => mat[BigInt](n, n, ad),
                                (mat0I[BigInt](BigInt(0), n, n),
                                  (mat1I[BigInt](BigInt(0), BigInt(1), n),
                                    (mat1I[BigInt](BigInt(0),
            get_comp_def[BigInt](((_: BigInt) => true,
                                   (BigInt(0),
                                     (BigInt(1),
                                       (BigInt(1),
 ((ad: BigInt) => (b: BigInt) => ad + b,
   ((ad: BigInt) => (b: BigInt) => ad * b,
     ((ad: BigInt) => max0_int(ad),
       ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
         ((xa: BigInt) => (y: BigInt) => y < xa,
           (true,
             (int_name,
               ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))),
            n),
                                      ((ad: List[List[BigInt]]) =>
 (b: List[List[BigInt]]) =>
   mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
((ad: List[List[BigInt]]) =>
   (b: List[List[BigInt]]) =>
     mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                        (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
  (mat_max0[BigInt],
    (mat_ge[BigInt],
      ((ad: List[List[BigInt]]) =>
         (b: List[List[BigInt]]) =>
           mat_gtI[BigInt](get_comp_gt[BigInt](((_: BigInt) => true,
         (BigInt(0),
           (BigInt(1),
             (BigInt(1),
               ((ae: BigInt) => (ba: BigInt) => ae + ba,
                 ((ae: BigInt) => (ba: BigInt) => ae * ba,
                   ((ae: BigInt) => max0_int(ae),
                     ((ae: BigInt) => (ba: BigInt) => ge_int(ae, ba),
                       ((xa: BigInt) => (y: BigInt) => y < xa,
                         (true,
                           (int_name,
                             ((_: BigInt) => true,
                               (ae: BigInt) => int_mono(ae)))))))))))))),
                            sd, ad, b),
        (get_comp_mono[BigInt](((_: BigInt) => true,
                                 (BigInt(0),
                                   (BigInt(1),
                                     (BigInt(1),
                                       ((ad: BigInt) => (b: BigInt) => ad + b,
 ((ad: BigInt) => (b: BigInt) => ad * b,
   ((ad: BigInt) => max0_int(ad),
     ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
       ((xa: BigInt) => (y: BigInt) => y < xa,
         (true,
           (int_name,
             ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))),
          (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil),
                              List('x'),
                              (shows_prec_nat(Zero_nat, n)).apply(Nil),
                              List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                    ' ', 'w', 'i', 't', 'h', ' ', 's', 't', 'r',
                                    'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n',
                                    's', 'i', 'o', 'n', ' '),
                              (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                              List(' ', 'o', 'v', 'e', 'r', ' '),
                              get_comp_description[BigInt](((_: BigInt) => true,
                     (BigInt(0),
                       (BigInt(1),
                         (BigInt(1),
                           ((ad: BigInt) => (b: BigInt) => ad + b,
                             ((ad: BigInt) => (b: BigInt) => ad * b,
                               ((ad: BigInt) => max0_int(ad),
                                 ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
                                   ((xa: BigInt) => (y: BigInt) => y < xa,
                                     (true,
                                       (int_name,
 ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))))),
            ((_: List[List[BigInt]]) => true,
              mat_mono[BigInt](sd, get_comp_check_mono[BigInt](((_: BigInt) =>
                          true,
                         (BigInt(0),
                           (BigInt(1),
                             (BigInt(1),
                               ((ad: BigInt) => (b: BigInt) => ad + b,
                                 ((ad: BigInt) => (b: BigInt) => ad * b,
                                   ((ad: BigInt) => max0_int(ad),
                                     ((ad: BigInt) =>
(b: BigInt) => ge_int(ad, b),
                                       ((xa: BigInt) => (y: BigInt) => y < xa,
 (true,
   (int_name,
     ((_: BigInt) => true,
       (ad: BigInt) => int_mono(ad)))))))))))))))))))))))))))),
                                       get_comp_plus[List[List[BigInt]]](((ad:
                                     List[List[BigInt]])
                                    => mat[BigInt](n, n, ad),
                                   (mat0I[BigInt](BigInt(0), n, n),
                                     (mat1I[BigInt](BigInt(0), BigInt(1), n),
                                       (mat1I[BigInt](BigInt(0),
               get_comp_def[BigInt](((_: BigInt) => true,
                                      (BigInt(0),
(BigInt(1),
  (BigInt(1),
    ((ad: BigInt) => (b: BigInt) => ad + b,
      ((ad: BigInt) => (b: BigInt) => ad * b,
        ((ad: BigInt) => max0_int(ad),
          ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
            ((xa: BigInt) => (y: BigInt) => y < xa,
              (true,
                (int_name,
                  ((_: BigInt) => true,
                    (ad: BigInt) => int_mono(ad)))))))))))))),
               n),
 ((ad: List[List[BigInt]]) =>
    (b: List[List[BigInt]]) =>
      mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
   ((ad: List[List[BigInt]]) =>
      (b: List[List[BigInt]]) =>
        mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                           (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
     (mat_max0[BigInt],
       (mat_ge[BigInt],
         ((ad: List[List[BigInt]]) =>
            (b: List[List[BigInt]]) =>
              mat_gtI[BigInt](get_comp_gt[BigInt](((_: BigInt) => true,
            (BigInt(0),
              (BigInt(1),
                (BigInt(1),
                  ((ae: BigInt) => (ba: BigInt) => ae + ba,
                    ((ae: BigInt) => (ba: BigInt) => ae * ba,
                      ((ae: BigInt) => max0_int(ae),
                        ((ae: BigInt) => (ba: BigInt) => ge_int(ae, ba),
                          ((xa: BigInt) => (y: BigInt) => y < xa,
                            (true,
                              (int_name,
                                ((_: BigInt) => true,
                                  (ae: BigInt) => int_mono(ae)))))))))))))),
                               sd, ad, b),
           (get_comp_mono[BigInt](((_: BigInt) => true,
                                    (BigInt(0),
                                      (BigInt(1),
(BigInt(1),
  ((ad: BigInt) => (b: BigInt) => ad + b,
    ((ad: BigInt) => (b: BigInt) => ad * b,
      ((ad: BigInt) => max0_int(ad),
        ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
          ((xa: BigInt) => (y: BigInt) => y < xa,
            (true,
              (int_name,
                ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))),
             (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil),
                                 List('x'),
                                 (shows_prec_nat(Zero_nat, n)).apply(Nil),
                                 List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e',
                                       's', ' ', 'w', 'i', 't', 'h', ' ', 's',
                                       't', 'r', 'i', 'c', 't', ' ', 'd', 'i',
                                       'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                                 (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                                 List(' ', 'o', 'v', 'e', 'r', ' '),
                                 get_comp_description[BigInt](((_: BigInt) =>
                         true,
                        (BigInt(0),
                          (BigInt(1),
                            (BigInt(1),
                              ((ad: BigInt) => (b: BigInt) => ad + b,
                                ((ad: BigInt) => (b: BigInt) => ad * b,
                                  ((ad: BigInt) => max0_int(ad),
                                    ((ad: BigInt) =>
                                       (b: BigInt) => ge_int(ad, b),
                                      ((xa: BigInt) => (y: BigInt) => y < xa,
(true,
  (int_name, ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))))),
               ((_: List[List[BigInt]]) => true,
                 mat_mono[BigInt](sd, get_comp_check_mono[BigInt](((_: BigInt)
                             => true,
                            (BigInt(0),
                              (BigInt(1),
                                (BigInt(1),
                                  ((ad: BigInt) => (b: BigInt) => ad + b,
                                    ((ad: BigInt) => (b: BigInt) => ad * b,
                                      ((ad: BigInt) => max0_int(ad),
((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
  ((xa: BigInt) => (y: BigInt) => y < xa,
    (true,
      (int_name,
        ((_: BigInt) => true,
          (ad: BigInt) => int_mono(ad)))))))))))))))))))))))))))),
                                       get_comp_times[List[List[BigInt]]](((ad:
                                      List[List[BigInt]])
                                     => mat[BigInt](n, n, ad),
                                    (mat0I[BigInt](BigInt(0), n, n),
                                      (mat1I[BigInt](BigInt(0), BigInt(1), n),
(mat1I[BigInt](BigInt(0),
                get_comp_def[BigInt](((_: BigInt) => true,
                                       (BigInt(0),
 (BigInt(1),
   (BigInt(1),
     ((ad: BigInt) => (b: BigInt) => ad + b,
       ((ad: BigInt) => (b: BigInt) => ad * b,
         ((ad: BigInt) => max0_int(ad),
           ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
             ((xa: BigInt) => (y: BigInt) => y < xa,
               (true,
                 (int_name,
                   ((_: BigInt) => true,
                     (ad: BigInt) => int_mono(ad)))))))))))))),
                n),
  ((ad: List[List[BigInt]]) =>
     (b: List[List[BigInt]]) =>
       mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
    ((ad: List[List[BigInt]]) =>
       (b: List[List[BigInt]]) =>
         mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                            (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
      (mat_max0[BigInt],
        (mat_ge[BigInt],
          ((ad: List[List[BigInt]]) =>
             (b: List[List[BigInt]]) =>
               mat_gtI[BigInt](get_comp_gt[BigInt](((_: BigInt) => true,
             (BigInt(0),
               (BigInt(1),
                 (BigInt(1),
                   ((ae: BigInt) => (ba: BigInt) => ae + ba,
                     ((ae: BigInt) => (ba: BigInt) => ae * ba,
                       ((ae: BigInt) => max0_int(ae),
                         ((ae: BigInt) => (ba: BigInt) => ge_int(ae, ba),
                           ((xa: BigInt) => (y: BigInt) => y < xa,
                             (true,
                               (int_name,
                                 ((_: BigInt) => true,
                                   (ae: BigInt) => int_mono(ae)))))))))))))),
                                sd, ad, b),
            (get_comp_mono[BigInt](((_: BigInt) => true,
                                     (BigInt(0),
                                       (BigInt(1),
 (BigInt(1),
   ((ad: BigInt) => (b: BigInt) => ad + b,
     ((ad: BigInt) => (b: BigInt) => ad * b,
       ((ad: BigInt) => max0_int(ad),
         ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
           ((xa: BigInt) => (y: BigInt) => y < xa,
             (true,
               (int_name,
                 ((_: BigInt) => true,
                   (ad: BigInt) => int_mono(ad)))))))))))))),
              (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil),
                                  List('x'),
                                  (shows_prec_nat(Zero_nat, n)).apply(Nil),
                                  List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e',
's', ' ', 'w', 'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i',
'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                                  (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                                  List(' ', 'o', 'v', 'e', 'r', ' '),
                                  get_comp_description[BigInt](((_: BigInt) =>
                          true,
                         (BigInt(0),
                           (BigInt(1),
                             (BigInt(1),
                               ((ad: BigInt) => (b: BigInt) => ad + b,
                                 ((ad: BigInt) => (b: BigInt) => ad * b,
                                   ((ad: BigInt) => max0_int(ad),
                                     ((ad: BigInt) =>
(b: BigInt) => ge_int(ad, b),
                                       ((xa: BigInt) => (y: BigInt) => y < xa,
 (true,
   (int_name, ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))))),
                ((_: List[List[BigInt]]) => true,
                  mat_mono[BigInt](sd, get_comp_check_mono[BigInt](((_: BigInt)
                              => true,
                             (BigInt(0),
                               (BigInt(1),
                                 (BigInt(1),
                                   ((ad: BigInt) => (b: BigInt) => ad + b,
                                     ((ad: BigInt) => (b: BigInt) => ad * b,
                                       ((ad: BigInt) => max0_int(ad),
 ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
   ((xa: BigInt) => (y: BigInt) => y < xa,
     (true,
       (int_name,
         ((_: BigInt) => true,
           (ad: BigInt) => int_mono(ad)))))))))))))))))))))))))))),
                                       ac)),
        (f: A) =>
          (ac: nat) =>
            (poly: (List[List[BigInt]], List[List[List[BigInt]]])) =>
              Alter4[((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
                      ((A, nat), (BigInt, List[BigInt])),
                      ((A, nat), (rat, List[rat])),
                      ((A, nat), (arctic, List[arctic])),
                      ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                      ((A, nat),
                        (List[List[arctic]], List[List[List[arctic]]])),
                      ((A, nat), (arctic_rat, List[arctic_rat])),
                      ((A, nat),
                        (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                      ((A, nat), List[(List[(nat, nat)], BigInt)]),
                      ((A, nat),
                        List[(List[(nat, nat)],
                               rat)])](((f, ac),
 (fst[List[List[BigInt]], List[List[List[BigInt]]]](poly),
   fit_length[List[List[BigInt]]](get_comp_0[List[List[BigInt]]](((ad:
                             List[List[BigInt]])
                            => mat[BigInt](n, n, ad),
                           (mat0I[BigInt](BigInt(0), n, n),
                             (mat1I[BigInt](BigInt(0), BigInt(1), n),
                               (mat1I[BigInt](BigInt(0),
       get_comp_def[BigInt](((_: BigInt) => true,
                              (BigInt(0),
                                (BigInt(1),
                                  (BigInt(1),
                                    ((ad: BigInt) => (b: BigInt) => ad + b,
                                      ((ad: BigInt) => (b: BigInt) => ad * b,
((ad: BigInt) => max0_int(ad),
  ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
    ((x: BigInt) => (y: BigInt) => y < x,
      (true,
        (int_name,
          ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))),
       n),
                                 ((ad: List[List[BigInt]]) =>
                                    (b: List[List[BigInt]]) =>
                                      mat_plusI[BigInt]((ae: BigInt) =>
                  (ba: BigInt) => ae + ba,
                 ad, b),
                                   ((ad: List[List[BigInt]]) =>
                                      (b: List[List[BigInt]]) =>
mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                   (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
                                     (mat_max0[BigInt],
                                       (mat_ge[BigInt],
 ((ad: List[List[BigInt]]) =>
    (b: List[List[BigInt]]) =>
      mat_gtI[BigInt](get_comp_gt[BigInt](((_: BigInt) => true,
    (BigInt(0),
      (BigInt(1),
        (BigInt(1),
          ((ae: BigInt) => (ba: BigInt) => ae + ba,
            ((ae: BigInt) => (ba: BigInt) => ae * ba,
              ((ae: BigInt) => max0_int(ae),
                ((ae: BigInt) => (ba: BigInt) => ge_int(ae, ba),
                  ((x: BigInt) => (y: BigInt) => y < x,
                    (true,
                      (int_name,
                        ((_: BigInt) => true,
                          (ae: BigInt) => int_mono(ae)))))))))))))),
                       sd, ad, b),
   (get_comp_mono[BigInt](((_: BigInt) => true,
                            (BigInt(0),
                              (BigInt(1),
                                (BigInt(1),
                                  ((ad: BigInt) => (b: BigInt) => ad + b,
                                    ((ad: BigInt) => (b: BigInt) => ad * b,
                                      ((ad: BigInt) => max0_int(ad),
((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
  ((x: BigInt) => (y: BigInt) => y < x,
    (true,
      (int_name,
        ((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))),
     (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                         (shows_prec_nat(Zero_nat, n)).apply(Nil),
                         List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                               'w', 'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c',
                               't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o',
                               'n', ' '),
                         (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                         List(' ', 'o', 'v', 'e', 'r', ' '),
                         get_comp_description[BigInt](((_: BigInt) => true,
                (BigInt(0),
                  (BigInt(1),
                    (BigInt(1),
                      ((ad: BigInt) => (b: BigInt) => ad + b,
                        ((ad: BigInt) => (b: BigInt) => ad * b,
                          ((ad: BigInt) => max0_int(ad),
                            ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
                              ((x: BigInt) => (y: BigInt) => y < x,
                                (true,
                                  (int_name,
                                    ((_: BigInt) => true,
                                      (ad: BigInt) =>
int_mono(ad)))))))))))))))),
       ((_: List[List[BigInt]]) => true,
         mat_mono[BigInt](sd, get_comp_check_mono[BigInt](((_: BigInt) => true,
                    (BigInt(0),
                      (BigInt(1),
                        (BigInt(1),
                          ((ad: BigInt) => (b: BigInt) => ad + b,
                            ((ad: BigInt) => (b: BigInt) => ad * b,
                              ((ad: BigInt) => max0_int(ad),
                                ((ad: BigInt) => (b: BigInt) => ge_int(ad, b),
                                  ((x: BigInt) => (y: BigInt) => y < x,
                                    (true,
                                      (int_name,
((_: BigInt) => true, (ad: BigInt) => int_mono(ad)))))))))))))))))))))))))))),
                                   ac, snd[List[List[BigInt]],
    List[List[List[BigInt]]]](poly))))),
        ab)
    case Arc_mat(n) =>
      (ab: xml) =>
        xml3elements[A, nat, (List[List[arctic]], List[List[List[arctic]]]),
                      alter[((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])]](List('i', 'n', 't', 'e', 'r',
             'p', 'r', 'e', 't'),
        xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
        (x: xml) =>
          binda[List[Char], tpoly[nat, List[List[arctic]]],
                 (List[List[arctic]],
                   List[List[List[arctic]]])](polynomial[List[List[arctic]]](mat_coeff[arctic](MinInfty,
                arctic_coeff),
                                      x),
       (ac: tpoly[nat, List[List[arctic]]]) =>
         lpoly_of[List[List[arctic]]](get_comp_0[List[List[arctic]]](((ad:
                                 List[List[arctic]])
                                => mat[arctic](n, n, ad),
                               (mat0I[arctic](zero_arctic, n, n),
                                 (mat1I[arctic](zero_arctic, one_arctic, n),
                                   (mat1I[arctic](zero_arctic,
           get_comp_def[arctic](((_: arctic) => true,
                                  (zero_arctic,
                                    (one_arctic,
                                      (one_arctic,
((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
  ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
    (id[arctic],
      ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
        ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
          (false,
            (arctic_name,
              ((ad: arctic) => pos_arctic(ad),
                (_: arctic) => false))))))))))))),
           n),
                                     ((ad: List[List[arctic]]) =>
(b: List[List[arctic]]) =>
  mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba), ad, b),
                                       ((ad: List[List[arctic]]) =>
  (b: List[List[arctic]]) =>
    mat_multI[arctic](zero_arctic,
                       (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                       (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n,
                       ad, b),
 (id[List[List[arctic]]],
   (mat_ge[arctic],
     ((ad: List[List[arctic]]) =>
        (b: List[List[arctic]]) =>
          mat_comp_all[arctic](get_comp_gt[arctic](((_: arctic) => true,
             (zero_arctic,
               (one_arctic,
                 (one_arctic,
                   ((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                     ((ae: arctic) => (ba: arctic) => times_arctic(ae, ba),
                       (id[arctic],
                         ((ae: arctic) => (ba: arctic) => ge_arctic(ae, ba),
                           ((ae: arctic) => (ba: arctic) => gt_arctic(ae, ba),
                             (false,
                               (arctic_name,
                                 ((ae: arctic) => pos_arctic(ae),
                                   (_: arctic) => false))))))))))))),
                                ad, b),
       (false,
         (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'o', 'v', 'e', 'r', ' '),
                             get_comp_description[arctic](((_: arctic) => true,
                    (zero_arctic,
                      (one_arctic,
                        (one_arctic,
                          ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
                            ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
                              (id[arctic],
                                ((ad: arctic) =>
                                   (b: arctic) => ge_arctic(ad, b),
                                  ((ad: arctic) =>
                                     (b: arctic) => gt_arctic(ad, b),
                                    (false,
                                      (arctic_name,
((ad: arctic) => pos_arctic(ad), (_: arctic) => false))))))))))))))),
           ((ad: List[List[arctic]]) =>
              mat_arc_posI[arctic](get_comp_apos[arctic](((_: arctic) => true,
                   (zero_arctic,
                     (one_arctic,
                       (one_arctic,
                         ((ae: arctic) => (b: arctic) => plus_arctic(ae, b),
                           ((ae: arctic) => (b: arctic) => times_arctic(ae, b),
                             (id[arctic],
                               ((ae: arctic) => (b: arctic) => ge_arctic(ae, b),
                                 ((ae: arctic) =>
                                    (b: arctic) => gt_arctic(ae, b),
                                   (false,
                                     (arctic_name,
                                       ((ae: arctic) => pos_arctic(ae),
 (_: arctic) => false))))))))))))),
                                    ad),
             (_: List[List[arctic]]) => false))))))))))))),
                                       get_comp_1[List[List[arctic]]](((ad:
                                  List[List[arctic]])
                                 => mat[arctic](n, n, ad),
                                (mat0I[arctic](zero_arctic, n, n),
                                  (mat1I[arctic](zero_arctic, one_arctic, n),
                                    (mat1I[arctic](zero_arctic,
            get_comp_def[arctic](((_: arctic) => true,
                                   (zero_arctic,
                                     (one_arctic,
                                       (one_arctic,
 ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
   ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
     (id[arctic],
       ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
         ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
           (false,
             (arctic_name,
               ((ad: arctic) => pos_arctic(ad),
                 (_: arctic) => false))))))))))))),
            n),
                                      ((ad: List[List[arctic]]) =>
 (b: List[List[arctic]]) =>
   mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba), ad,
                      b),
((ad: List[List[arctic]]) =>
   (b: List[List[arctic]]) =>
     mat_multI[arctic](zero_arctic,
                        (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                        (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n,
                        ad, b),
  (id[List[List[arctic]]],
    (mat_ge[arctic],
      ((ad: List[List[arctic]]) =>
         (b: List[List[arctic]]) =>
           mat_comp_all[arctic](get_comp_gt[arctic](((_: arctic) => true,
              (zero_arctic,
                (one_arctic,
                  (one_arctic,
                    ((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                      ((ae: arctic) => (ba: arctic) => times_arctic(ae, ba),
                        (id[arctic],
                          ((ae: arctic) => (ba: arctic) => ge_arctic(ae, ba),
                            ((ae: arctic) => (ba: arctic) => gt_arctic(ae, ba),
                              (false,
                                (arctic_name,
                                  ((ae: arctic) => pos_arctic(ae),
                                    (_: arctic) => false))))))))))))),
                                 ad, b),
        (false,
          (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil),
                              List('x'),
                              (shows_prec_nat(Zero_nat, n)).apply(Nil),
                              List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                    ' ', 'o', 'v', 'e', 'r', ' '),
                              get_comp_description[arctic](((_: arctic) => true,
                     (zero_arctic,
                       (one_arctic,
                         (one_arctic,
                           ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
                             ((ad: arctic) =>
                                (b: arctic) => times_arctic(ad, b),
                               (id[arctic],
                                 ((ad: arctic) =>
                                    (b: arctic) => ge_arctic(ad, b),
                                   ((ad: arctic) =>
                                      (b: arctic) => gt_arctic(ad, b),
                                     (false,
                                       (arctic_name,
 ((ad: arctic) => pos_arctic(ad), (_: arctic) => false))))))))))))))),
            ((ad: List[List[arctic]]) =>
               mat_arc_posI[arctic](get_comp_apos[arctic](((_: arctic) => true,
                    (zero_arctic,
                      (one_arctic,
                        (one_arctic,
                          ((ae: arctic) => (b: arctic) => plus_arctic(ae, b),
                            ((ae: arctic) => (b: arctic) => times_arctic(ae, b),
                              (id[arctic],
                                ((ae: arctic) =>
                                   (b: arctic) => ge_arctic(ae, b),
                                  ((ae: arctic) =>
                                     (b: arctic) => gt_arctic(ae, b),
                                    (false,
                                      (arctic_name,
((ae: arctic) => pos_arctic(ae), (_: arctic) => false))))))))))))),
                                     ad),
              (_: List[List[arctic]]) => false))))))))))))),
                                       get_comp_plus[List[List[arctic]]](((ad:
                                     List[List[arctic]])
                                    => mat[arctic](n, n, ad),
                                   (mat0I[arctic](zero_arctic, n, n),
                                     (mat1I[arctic](zero_arctic, one_arctic, n),
                                       (mat1I[arctic](zero_arctic,
               get_comp_def[arctic](((_: arctic) => true,
                                      (zero_arctic,
(one_arctic,
  (one_arctic,
    ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
      ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
        (id[arctic],
          ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
            ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
              (false,
                (arctic_name,
                  ((ad: arctic) => pos_arctic(ad),
                    (_: arctic) => false))))))))))))),
               n),
 ((ad: List[List[arctic]]) =>
    (b: List[List[arctic]]) =>
      mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba), ad,
                         b),
   ((ad: List[List[arctic]]) =>
      (b: List[List[arctic]]) =>
        mat_multI[arctic](zero_arctic,
                           (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                           (ae: arctic) => (ba: arctic) => times_arctic(ae, ba),
                           n, ad, b),
     (id[List[List[arctic]]],
       (mat_ge[arctic],
         ((ad: List[List[arctic]]) =>
            (b: List[List[arctic]]) =>
              mat_comp_all[arctic](get_comp_gt[arctic](((_: arctic) => true,
                 (zero_arctic,
                   (one_arctic,
                     (one_arctic,
                       ((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                         ((ae: arctic) => (ba: arctic) => times_arctic(ae, ba),
                           (id[arctic],
                             ((ae: arctic) => (ba: arctic) => ge_arctic(ae, ba),
                               ((ae: arctic) =>
                                  (ba: arctic) => gt_arctic(ae, ba),
                                 (false,
                                   (arctic_name,
                                     ((ae: arctic) => pos_arctic(ae),
                                       (_: arctic) => false))))))))))))),
                                    ad, b),
           (false,
             (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil),
                                 List('x'),
                                 (shows_prec_nat(Zero_nat, n)).apply(Nil),
                                 List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e',
                                       's', ' ', 'o', 'v', 'e', 'r', ' '),
                                 get_comp_description[arctic](((_: arctic) =>
                         true,
                        (zero_arctic,
                          (one_arctic,
                            (one_arctic,
                              ((ad: arctic) =>
                                 (b: arctic) => plus_arctic(ad, b),
                                ((ad: arctic) =>
                                   (b: arctic) => times_arctic(ad, b),
                                  (id[arctic],
                                    ((ad: arctic) =>
                                       (b: arctic) => ge_arctic(ad, b),
                                      ((ad: arctic) =>
 (b: arctic) => gt_arctic(ad, b),
(false,
  (arctic_name,
    ((ad: arctic) => pos_arctic(ad), (_: arctic) => false))))))))))))))),
               ((ad: List[List[arctic]]) =>
                  mat_arc_posI[arctic](get_comp_apos[arctic](((_: arctic) =>
                        true,
                       (zero_arctic,
                         (one_arctic,
                           (one_arctic,
                             ((ae: arctic) => (b: arctic) => plus_arctic(ae, b),
                               ((ae: arctic) =>
                                  (b: arctic) => times_arctic(ae, b),
                                 (id[arctic],
                                   ((ae: arctic) =>
                                      (b: arctic) => ge_arctic(ae, b),
                                     ((ae: arctic) =>
(b: arctic) => gt_arctic(ae, b),
                                       (false,
 (arctic_name,
   ((ae: arctic) => pos_arctic(ae), (_: arctic) => false))))))))))))),
ad),
                 (_: List[List[arctic]]) => false))))))))))))),
                                       get_comp_times[List[List[arctic]]](((ad:
                                      List[List[arctic]])
                                     => mat[arctic](n, n, ad),
                                    (mat0I[arctic](zero_arctic, n, n),
                                      (mat1I[arctic](zero_arctic, one_arctic,
              n),
(mat1I[arctic](zero_arctic,
                get_comp_def[arctic](((_: arctic) => true,
                                       (zero_arctic,
 (one_arctic,
   (one_arctic,
     ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
       ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
         (id[arctic],
           ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
             ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
               (false,
                 (arctic_name,
                   ((ad: arctic) => pos_arctic(ad),
                     (_: arctic) => false))))))))))))),
                n),
  ((ad: List[List[arctic]]) =>
     (b: List[List[arctic]]) =>
       mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                          ad, b),
    ((ad: List[List[arctic]]) =>
       (b: List[List[arctic]]) =>
         mat_multI[arctic](zero_arctic,
                            (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                            (ae: arctic) =>
                              (ba: arctic) => times_arctic(ae, ba),
                            n, ad, b),
      (id[List[List[arctic]]],
        (mat_ge[arctic],
          ((ad: List[List[arctic]]) =>
             (b: List[List[arctic]]) =>
               mat_comp_all[arctic](get_comp_gt[arctic](((_: arctic) => true,
                  (zero_arctic,
                    (one_arctic,
                      (one_arctic,
                        ((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                          ((ae: arctic) => (ba: arctic) => times_arctic(ae, ba),
                            (id[arctic],
                              ((ae: arctic) =>
                                 (ba: arctic) => ge_arctic(ae, ba),
                                ((ae: arctic) =>
                                   (ba: arctic) => gt_arctic(ae, ba),
                                  (false,
                                    (arctic_name,
                                      ((ae: arctic) => pos_arctic(ae),
(_: arctic) => false))))))))))))),
                                     ad, b),
            (false,
              (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil),
                                  List('x'),
                                  (shows_prec_nat(Zero_nat, n)).apply(Nil),
                                  List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e',
's', ' ', 'o', 'v', 'e', 'r', ' '),
                                  get_comp_description[arctic](((_: arctic) =>
                          true,
                         (zero_arctic,
                           (one_arctic,
                             (one_arctic,
                               ((ad: arctic) =>
                                  (b: arctic) => plus_arctic(ad, b),
                                 ((ad: arctic) =>
                                    (b: arctic) => times_arctic(ad, b),
                                   (id[arctic],
                                     ((ad: arctic) =>
(b: arctic) => ge_arctic(ad, b),
                                       ((ad: arctic) =>
  (b: arctic) => gt_arctic(ad, b),
 (false,
   (arctic_name,
     ((ad: arctic) => pos_arctic(ad), (_: arctic) => false))))))))))))))),
                ((ad: List[List[arctic]]) =>
                   mat_arc_posI[arctic](get_comp_apos[arctic](((_: arctic) =>
                         true,
                        (zero_arctic,
                          (one_arctic,
                            (one_arctic,
                              ((ae: arctic) =>
                                 (b: arctic) => plus_arctic(ae, b),
                                ((ae: arctic) =>
                                   (b: arctic) => times_arctic(ae, b),
                                  (id[arctic],
                                    ((ae: arctic) =>
                                       (b: arctic) => ge_arctic(ae, b),
                                      ((ae: arctic) =>
 (b: arctic) => gt_arctic(ae, b),
(false,
  (arctic_name,
    ((ae: arctic) => pos_arctic(ae), (_: arctic) => false))))))))))))),
 ad),
                  (_: List[List[arctic]]) => false))))))))))))),
                                       ac)),
        (f: A) =>
          (ac: nat) =>
            (poly: (List[List[arctic]], List[List[List[arctic]]])) =>
              Alter6[((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
                      ((A, nat), (BigInt, List[BigInt])),
                      ((A, nat), (rat, List[rat])),
                      ((A, nat), (arctic, List[arctic])),
                      ((A, nat),
                        (List[List[BigInt]], List[List[List[BigInt]]])),
                      ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                      ((A, nat), (arctic_rat, List[arctic_rat])),
                      ((A, nat),
                        (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                      ((A, nat), List[(List[(nat, nat)], BigInt)]),
                      ((A, nat),
                        List[(List[(nat, nat)],
                               rat)])](((f, ac),
 (fst[List[List[arctic]], List[List[List[arctic]]]](poly),
   fit_length[List[List[arctic]]](get_comp_0[List[List[arctic]]](((ad:
                             List[List[arctic]])
                            => mat[arctic](n, n, ad),
                           (mat0I[arctic](zero_arctic, n, n),
                             (mat1I[arctic](zero_arctic, one_arctic, n),
                               (mat1I[arctic](zero_arctic,
       get_comp_def[arctic](((_: arctic) => true,
                              (zero_arctic,
                                (one_arctic,
                                  (one_arctic,
                                    ((ad: arctic) =>
                                       (b: arctic) => plus_arctic(ad, b),
                                      ((ad: arctic) =>
 (b: arctic) => times_arctic(ad, b),
(id[arctic],
  ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
    ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
      (false,
        (arctic_name,
          ((ad: arctic) => pos_arctic(ad), (_: arctic) => false))))))))))))),
       n),
                                 ((ad: List[List[arctic]]) =>
                                    (b: List[List[arctic]]) =>
                                      mat_plusI[arctic]((ae: arctic) =>
                  (ba: arctic) => plus_arctic(ae, ba),
                 ad, b),
                                   ((ad: List[List[arctic]]) =>
                                      (b: List[List[arctic]]) =>
mat_multI[arctic](zero_arctic,
                   (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                   (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n, ad,
                   b),
                                     (id[List[List[arctic]]],
                                       (mat_ge[arctic],
 ((ad: List[List[arctic]]) =>
    (b: List[List[arctic]]) =>
      mat_comp_all[arctic](get_comp_gt[arctic](((_: arctic) => true,
         (zero_arctic,
           (one_arctic,
             (one_arctic,
               ((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                 ((ae: arctic) => (ba: arctic) => times_arctic(ae, ba),
                   (id[arctic],
                     ((ae: arctic) => (ba: arctic) => ge_arctic(ae, ba),
                       ((ae: arctic) => (ba: arctic) => gt_arctic(ae, ba),
                         (false,
                           (arctic_name,
                             ((ae: arctic) => pos_arctic(ae),
                               (_: arctic) => false))))))))))))),
                            ad, b),
   (false,
     (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                         (shows_prec_nat(Zero_nat, n)).apply(Nil),
                         List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                               'o', 'v', 'e', 'r', ' '),
                         get_comp_description[arctic](((_: arctic) => true,
                (zero_arctic,
                  (one_arctic,
                    (one_arctic,
                      ((ad: arctic) => (b: arctic) => plus_arctic(ad, b),
                        ((ad: arctic) => (b: arctic) => times_arctic(ad, b),
                          (id[arctic],
                            ((ad: arctic) => (b: arctic) => ge_arctic(ad, b),
                              ((ad: arctic) => (b: arctic) => gt_arctic(ad, b),
                                (false,
                                  (arctic_name,
                                    ((ad: arctic) => pos_arctic(ad),
                                      (_: arctic) => false))))))))))))))),
       ((ad: List[List[arctic]]) =>
          mat_arc_posI[arctic](get_comp_apos[arctic](((_: arctic) => true,
               (zero_arctic,
                 (one_arctic,
                   (one_arctic,
                     ((ae: arctic) => (b: arctic) => plus_arctic(ae, b),
                       ((ae: arctic) => (b: arctic) => times_arctic(ae, b),
                         (id[arctic],
                           ((ae: arctic) => (b: arctic) => ge_arctic(ae, b),
                             ((ae: arctic) => (b: arctic) => gt_arctic(ae, b),
                               (false,
                                 (arctic_name,
                                   ((ae: arctic) => pos_arctic(ae),
                                     (_: arctic) => false))))))))))))),
                                ad),
         (_: List[List[arctic]]) => false))))))))))))),
                                   ac, snd[List[List[arctic]],
    List[List[List[arctic]]]](poly))))),
        ab)
    case Arc_rat_mat(d, n) =>
      (ab: xml) =>
        xml3elements[A, nat,
                      (List[List[arctic_rat]], List[List[List[arctic_rat]]]),
                      alter[((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])]](List('i', 'n', 't', 'e', 'r',
             'p', 'r', 'e', 't'),
        xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
        (x: xml) =>
          binda[List[Char], tpoly[nat, List[List[arctic_rat]]],
                 (List[List[arctic_rat]],
                   List[List[List[arctic_rat]]])](polynomial[List[List[arctic_rat]]](mat_coeff[arctic_rat](MinInfty_rat,
                            arctic_rat_coeff),
      x),
           (ac: tpoly[nat, List[List[arctic_rat]]]) =>
             lpoly_of[List[List[arctic_rat]]](get_comp_0[List[List[arctic_rat]]](((ad:
     List[List[arctic_rat]])
    => mat[arctic_rat](n, n, ad),
   (mat0I[arctic_rat](zero_arctic_rat, n, n),
     (mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
       (mat1I[arctic_rat](zero_arctic_rat,
                           get_comp_def[arctic_rat](((_: arctic_rat) => true,
              (zero_arctic_rat,
                (one_arctic_rat,
                  (one_arctic_rat,
                    ((ad: arctic_rat) =>
                       (b: arctic_rat) => plus_arctic_rat(ad, b),
                      ((ad: arctic_rat) =>
                         (b: arctic_rat) => times_arctic_rat(ad, b),
                        (id[arctic_rat],
                          ((ad: arctic_rat) =>
                             (b: arctic_rat) => ge_arctic_rat(ad, b),
                            ((ad: arctic_rat) =>
                               (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                              (false,
                                (arctic_rat_name,
                                  ((ad: arctic_rat) => pos_arctic_rat(ad),
                                    (_: arctic_rat) => false))))))))))))),
                           n),
         ((ad: List[List[arctic_rat]]) =>
            (b: List[List[arctic_rat]]) =>
              mat_plusI[arctic_rat]((ae: arctic_rat) =>
                                      (ba: arctic_rat) =>
plus_arctic_rat(ae, ba),
                                     ad, b),
           ((ad: List[List[arctic_rat]]) =>
              (b: List[List[arctic_rat]]) =>
                mat_multI[arctic_rat](zero_arctic_rat,
                                       (ae: arctic_rat) =>
 (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                       (ae: arctic_rat) =>
 (ba: arctic_rat) => times_arctic_rat(ae, ba),
                                       n, ad, b),
             (id[List[List[arctic_rat]]],
               (mat_ge[arctic_rat],
                 ((ad: List[List[arctic_rat]]) =>
                    (b: List[List[arctic_rat]]) =>
                      mat_comp_all[arctic_rat](get_comp_gt[arctic_rat](((_:
                                   arctic_rat)
                                  => true,
                                 (zero_arctic_rat,
                                   (one_arctic_rat,
                                     (one_arctic_rat,
                                       ((ae: arctic_rat) =>
  (ba: arctic_rat) => plus_arctic_rat(ae, ba),
 ((ae: arctic_rat) => (ba: arctic_rat) => times_arctic_rat(ae, ba),
   (id[arctic_rat],
     ((ae: arctic_rat) => (ba: arctic_rat) => ge_arctic_rat(ae, ba),
       ((ae: arctic_rat) => (ba: arctic_rat) => gt_arctic_rat(d, ae, ba),
         (false,
           (arctic_rat_name,
             ((ae: arctic_rat) => pos_arctic_rat(ae),
               (_: arctic_rat) => false))))))))))))),
        ad, b),
                   (false,
                     (concat[Char](List((shows_prec_nat(Zero_nat,
                 n)).apply(Nil),
 List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
 List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
       ' '),
 get_comp_description[arctic_rat](((_: arctic_rat) => true,
                                    (zero_arctic_rat,
                                      (one_arctic_rat,
(one_arctic_rat,
  ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
    ((ad: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ad, b),
      (id[arctic_rat],
        ((ad: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ad, b),
          ((ad: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ad, b),
            (false,
              (arctic_rat_name,
                ((ad: arctic_rat) => pos_arctic_rat(ad),
                  (_: arctic_rat) => false))))))))))))))),
                       ((ad: List[List[arctic_rat]]) =>
                          mat_arc_posI[arctic_rat](get_comp_apos[arctic_rat](((_:
 arctic_rat)
=> true,
                                       (zero_arctic_rat,
 (one_arctic_rat,
   (one_arctic_rat,
     ((ae: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ae, b),
       ((ae: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ae, b),
         (id[arctic_rat],
           ((ae: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ae, b),
             ((ae: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ae, b),
               (false,
                 (arctic_rat_name,
                   ((ae: arctic_rat) => pos_arctic_rat(ae),
                     (_: arctic_rat) => false))))))))))))),
            ad),
                         (_: List[List[arctic_rat]]) => false))))))))))))),
       get_comp_1[List[List[arctic_rat]]](((ad: List[List[arctic_rat]]) =>
     mat[arctic_rat](n, n, ad),
    (mat0I[arctic_rat](zero_arctic_rat, n, n),
      (mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
        (mat1I[arctic_rat](zero_arctic_rat,
                            get_comp_def[arctic_rat](((_: arctic_rat) => true,
               (zero_arctic_rat,
                 (one_arctic_rat,
                   (one_arctic_rat,
                     ((ad: arctic_rat) =>
                        (b: arctic_rat) => plus_arctic_rat(ad, b),
                       ((ad: arctic_rat) =>
                          (b: arctic_rat) => times_arctic_rat(ad, b),
                         (id[arctic_rat],
                           ((ad: arctic_rat) =>
                              (b: arctic_rat) => ge_arctic_rat(ad, b),
                             ((ad: arctic_rat) =>
                                (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                               (false,
                                 (arctic_rat_name,
                                   ((ad: arctic_rat) => pos_arctic_rat(ad),
                                     (_: arctic_rat) => false))))))))))))),
                            n),
          ((ad: List[List[arctic_rat]]) =>
             (b: List[List[arctic_rat]]) =>
               mat_plusI[arctic_rat]((ae: arctic_rat) =>
                                       (ba: arctic_rat) =>
 plus_arctic_rat(ae, ba),
                                      ad, b),
            ((ad: List[List[arctic_rat]]) =>
               (b: List[List[arctic_rat]]) =>
                 mat_multI[arctic_rat](zero_arctic_rat,
(ae: arctic_rat) => (ba: arctic_rat) => plus_arctic_rat(ae, ba),
(ae: arctic_rat) => (ba: arctic_rat) => times_arctic_rat(ae, ba), n, ad, b),
              (id[List[List[arctic_rat]]],
                (mat_ge[arctic_rat],
                  ((ad: List[List[arctic_rat]]) =>
                     (b: List[List[arctic_rat]]) =>
                       mat_comp_all[arctic_rat](get_comp_gt[arctic_rat](((_:
                                    arctic_rat)
                                   => true,
                                  (zero_arctic_rat,
                                    (one_arctic_rat,
                                      (one_arctic_rat,
((ae: arctic_rat) => (ba: arctic_rat) => plus_arctic_rat(ae, ba),
  ((ae: arctic_rat) => (ba: arctic_rat) => times_arctic_rat(ae, ba),
    (id[arctic_rat],
      ((ae: arctic_rat) => (ba: arctic_rat) => ge_arctic_rat(ae, ba),
        ((ae: arctic_rat) => (ba: arctic_rat) => gt_arctic_rat(d, ae, ba),
          (false,
            (arctic_rat_name,
              ((ae: arctic_rat) => pos_arctic_rat(ae),
                (_: arctic_rat) => false))))))))))))),
         ad, b),
                    (false,
                      (concat[Char](List((shows_prec_nat(Zero_nat,
                  n)).apply(Nil),
  List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
  List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
        ' '),
  get_comp_description[arctic_rat](((_: arctic_rat) => true,
                                     (zero_arctic_rat,
                                       (one_arctic_rat,
 (one_arctic_rat,
   ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
     ((ad: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ad, b),
       (id[arctic_rat],
         ((ad: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ad, b),
           ((ad: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ad, b),
             (false,
               (arctic_rat_name,
                 ((ad: arctic_rat) => pos_arctic_rat(ad),
                   (_: arctic_rat) => false))))))))))))))),
                        ((ad: List[List[arctic_rat]]) =>
                           mat_arc_posI[arctic_rat](get_comp_apos[arctic_rat](((_:
  arctic_rat)
 => true,
(zero_arctic_rat,
  (one_arctic_rat,
    (one_arctic_rat,
      ((ae: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ae, b),
        ((ae: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ae, b),
          (id[arctic_rat],
            ((ae: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ae, b),
              ((ae: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ae, b),
                (false,
                  (arctic_rat_name,
                    ((ae: arctic_rat) => pos_arctic_rat(ae),
                      (_: arctic_rat) => false))))))))))))),
             ad),
                          (_: List[List[arctic_rat]]) => false))))))))))))),
       get_comp_plus[List[List[arctic_rat]]](((ad: List[List[arctic_rat]]) =>
        mat[arctic_rat](n, n, ad),
       (mat0I[arctic_rat](zero_arctic_rat, n, n),
         (mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
           (mat1I[arctic_rat](zero_arctic_rat,
                               get_comp_def[arctic_rat](((_: arctic_rat) =>
                   true,
                  (zero_arctic_rat,
                    (one_arctic_rat,
                      (one_arctic_rat,
                        ((ad: arctic_rat) =>
                           (b: arctic_rat) => plus_arctic_rat(ad, b),
                          ((ad: arctic_rat) =>
                             (b: arctic_rat) => times_arctic_rat(ad, b),
                            (id[arctic_rat],
                              ((ad: arctic_rat) =>
                                 (b: arctic_rat) => ge_arctic_rat(ad, b),
                                ((ad: arctic_rat) =>
                                   (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                                  (false,
                                    (arctic_rat_name,
                                      ((ad: arctic_rat) => pos_arctic_rat(ad),
(_: arctic_rat) => false))))))))))))),
                               n),
             ((ad: List[List[arctic_rat]]) =>
                (b: List[List[arctic_rat]]) =>
                  mat_plusI[arctic_rat]((ae: arctic_rat) =>
  (ba: arctic_rat) => plus_arctic_rat(ae, ba),
 ad, b),
               ((ad: List[List[arctic_rat]]) =>
                  (b: List[List[arctic_rat]]) =>
                    mat_multI[arctic_rat](zero_arctic_rat,
   (ae: arctic_rat) => (ba: arctic_rat) => plus_arctic_rat(ae, ba),
   (ae: arctic_rat) => (ba: arctic_rat) => times_arctic_rat(ae, ba), n, ad, b),
                 (id[List[List[arctic_rat]]],
                   (mat_ge[arctic_rat],
                     ((ad: List[List[arctic_rat]]) =>
                        (b: List[List[arctic_rat]]) =>
                          mat_comp_all[arctic_rat](get_comp_gt[arctic_rat](((_:
                                       arctic_rat)
                                      => true,
                                     (zero_arctic_rat,
                                       (one_arctic_rat,
 (one_arctic_rat,
   ((ae: arctic_rat) => (ba: arctic_rat) => plus_arctic_rat(ae, ba),
     ((ae: arctic_rat) => (ba: arctic_rat) => times_arctic_rat(ae, ba),
       (id[arctic_rat],
         ((ae: arctic_rat) => (ba: arctic_rat) => ge_arctic_rat(ae, ba),
           ((ae: arctic_rat) => (ba: arctic_rat) => gt_arctic_rat(d, ae, ba),
             (false,
               (arctic_rat_name,
                 ((ae: arctic_rat) => pos_arctic_rat(ae),
                   (_: arctic_rat) => false))))))))))))),
            ad, b),
                       (false,
                         (concat[Char](List((shows_prec_nat(Zero_nat,
                     n)).apply(Nil),
     List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
     List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
           ' '),
     get_comp_description[arctic_rat](((_: arctic_rat) => true,
(zero_arctic_rat,
  (one_arctic_rat,
    (one_arctic_rat,
      ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
        ((ad: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ad, b),
          (id[arctic_rat],
            ((ad: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ad, b),
              ((ad: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                (false,
                  (arctic_rat_name,
                    ((ad: arctic_rat) => pos_arctic_rat(ad),
                      (_: arctic_rat) => false))))))))))))))),
                           ((ad: List[List[arctic_rat]]) =>
                              mat_arc_posI[arctic_rat](get_comp_apos[arctic_rat](((_:
     arctic_rat)
    => true,
   (zero_arctic_rat,
     (one_arctic_rat,
       (one_arctic_rat,
         ((ae: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ae, b),
           ((ae: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ae, b),
             (id[arctic_rat],
               ((ae: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ae, b),
                 ((ae: arctic_rat) =>
                    (b: arctic_rat) => gt_arctic_rat(d, ae, b),
                   (false,
                     (arctic_rat_name,
                       ((ae: arctic_rat) => pos_arctic_rat(ae),
                         (_: arctic_rat) => false))))))))))))),
                ad),
                             (_: List[List[arctic_rat]]) => false))))))))))))),
       get_comp_times[List[List[arctic_rat]]](((ad: List[List[arctic_rat]]) =>
         mat[arctic_rat](n, n, ad),
        (mat0I[arctic_rat](zero_arctic_rat, n, n),
          (mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
            (mat1I[arctic_rat](zero_arctic_rat,
                                get_comp_def[arctic_rat](((_: arctic_rat) =>
                    true,
                   (zero_arctic_rat,
                     (one_arctic_rat,
                       (one_arctic_rat,
                         ((ad: arctic_rat) =>
                            (b: arctic_rat) => plus_arctic_rat(ad, b),
                           ((ad: arctic_rat) =>
                              (b: arctic_rat) => times_arctic_rat(ad, b),
                             (id[arctic_rat],
                               ((ad: arctic_rat) =>
                                  (b: arctic_rat) => ge_arctic_rat(ad, b),
                                 ((ad: arctic_rat) =>
                                    (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                                   (false,
                                     (arctic_rat_name,
                                       ((ad: arctic_rat) => pos_arctic_rat(ad),
 (_: arctic_rat) => false))))))))))))),
                                n),
              ((ad: List[List[arctic_rat]]) =>
                 (b: List[List[arctic_rat]]) =>
                   mat_plusI[arctic_rat]((ae: arctic_rat) =>
   (ba: arctic_rat) => plus_arctic_rat(ae, ba),
  ad, b),
                ((ad: List[List[arctic_rat]]) =>
                   (b: List[List[arctic_rat]]) =>
                     mat_multI[arctic_rat](zero_arctic_rat,
    (ae: arctic_rat) => (ba: arctic_rat) => plus_arctic_rat(ae, ba),
    (ae: arctic_rat) => (ba: arctic_rat) => times_arctic_rat(ae, ba), n, ad, b),
                  (id[List[List[arctic_rat]]],
                    (mat_ge[arctic_rat],
                      ((ad: List[List[arctic_rat]]) =>
                         (b: List[List[arctic_rat]]) =>
                           mat_comp_all[arctic_rat](get_comp_gt[arctic_rat](((_:
arctic_rat)
                                       => true,
                                      (zero_arctic_rat,
(one_arctic_rat,
  (one_arctic_rat,
    ((ae: arctic_rat) => (ba: arctic_rat) => plus_arctic_rat(ae, ba),
      ((ae: arctic_rat) => (ba: arctic_rat) => times_arctic_rat(ae, ba),
        (id[arctic_rat],
          ((ae: arctic_rat) => (ba: arctic_rat) => ge_arctic_rat(ae, ba),
            ((ae: arctic_rat) => (ba: arctic_rat) => gt_arctic_rat(d, ae, ba),
              (false,
                (arctic_rat_name,
                  ((ae: arctic_rat) => pos_arctic_rat(ae),
                    (_: arctic_rat) => false))))))))))))),
             ad, b),
                        (false,
                          (concat[Char](List((shows_prec_nat(Zero_nat,
                      n)).apply(Nil),
      List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
      List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
            ' '),
      get_comp_description[arctic_rat](((_: arctic_rat) => true,
 (zero_arctic_rat,
   (one_arctic_rat,
     (one_arctic_rat,
       ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
         ((ad: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ad, b),
           (id[arctic_rat],
             ((ad: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ad, b),
               ((ad: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                 (false,
                   (arctic_rat_name,
                     ((ad: arctic_rat) => pos_arctic_rat(ad),
                       (_: arctic_rat) => false))))))))))))))),
                            ((ad: List[List[arctic_rat]]) =>
                               mat_arc_posI[arctic_rat](get_comp_apos[arctic_rat](((_:
      arctic_rat)
     => true,
    (zero_arctic_rat,
      (one_arctic_rat,
        (one_arctic_rat,
          ((ae: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ae, b),
            ((ae: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ae, b),
              (id[arctic_rat],
                ((ae: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ae, b),
                  ((ae: arctic_rat) =>
                     (b: arctic_rat) => gt_arctic_rat(d, ae, b),
                    (false,
                      (arctic_rat_name,
                        ((ae: arctic_rat) => pos_arctic_rat(ae),
                          (_: arctic_rat) => false))))))))))))),
                 ad),
                              (_: List[List[arctic_rat]]) => false))))))))))))),
       ac)),
        (f: A) =>
          (ac: nat) =>
            (poly: (List[List[arctic_rat]], List[List[List[arctic_rat]]])) =>
              Alter8[((A, nat),
                       (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                      ((A, nat), (BigInt, List[BigInt])),
                      ((A, nat), (rat, List[rat])),
                      ((A, nat), (arctic, List[arctic])),
                      ((A, nat),
                        (List[List[BigInt]], List[List[List[BigInt]]])),
                      ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                      ((A, nat),
                        (List[List[arctic]], List[List[List[arctic]]])),
                      ((A, nat), (arctic_rat, List[arctic_rat])),
                      ((A, nat), List[(List[(nat, nat)], BigInt)]),
                      ((A, nat),
                        List[(List[(nat, nat)],
                               rat)])](((f, ac),
 (fst[List[List[arctic_rat]], List[List[List[arctic_rat]]]](poly),
   fit_length[List[List[arctic_rat]]](get_comp_0[List[List[arctic_rat]]](((ad:
                                     List[List[arctic_rat]])
                                    => mat[arctic_rat](n, n, ad),
                                   (mat0I[arctic_rat](zero_arctic_rat, n, n),
                                     (mat1I[arctic_rat](zero_arctic_rat,
                 one_arctic_rat, n),
                                       (mat1I[arctic_rat](zero_arctic_rat,
                   get_comp_def[arctic_rat](((_: arctic_rat) => true,
      (zero_arctic_rat,
        (one_arctic_rat,
          (one_arctic_rat,
            ((ad: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(ad, b),
              ((ad: arctic_rat) => (b: arctic_rat) => times_arctic_rat(ad, b),
                (id[arctic_rat],
                  ((ad: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ad, b),
                    ((ad: arctic_rat) =>
                       (b: arctic_rat) => gt_arctic_rat(d, ad, b),
                      (false,
                        (arctic_rat_name,
                          ((ad: arctic_rat) => pos_arctic_rat(ad),
                            (_: arctic_rat) => false))))))))))))),
                   n),
 ((ad: List[List[arctic_rat]]) =>
    (b: List[List[arctic_rat]]) =>
      mat_plusI[arctic_rat]((ae: arctic_rat) =>
                              (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                             ad, b),
   ((ad: List[List[arctic_rat]]) =>
      (b: List[List[arctic_rat]]) =>
        mat_multI[arctic_rat](zero_arctic_rat,
                               (ae: arctic_rat) =>
                                 (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                               (ae: arctic_rat) =>
                                 (ba: arctic_rat) => times_arctic_rat(ae, ba),
                               n, ad, b),
     (id[List[List[arctic_rat]]],
       (mat_ge[arctic_rat],
         ((ad: List[List[arctic_rat]]) =>
            (b: List[List[arctic_rat]]) =>
              mat_comp_all[arctic_rat](get_comp_gt[arctic_rat](((_: arctic_rat)
                          => true,
                         (zero_arctic_rat,
                           (one_arctic_rat,
                             (one_arctic_rat,
                               ((ae: arctic_rat) =>
                                  (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                 ((ae: arctic_rat) =>
                                    (ba: arctic_rat) =>
                                      times_arctic_rat(ae, ba),
                                   (id[arctic_rat],
                                     ((ae: arctic_rat) =>
(ba: arctic_rat) => ge_arctic_rat(ae, ba),
                                       ((ae: arctic_rat) =>
  (ba: arctic_rat) => gt_arctic_rat(d, ae, ba),
 (false,
   (arctic_rat_name,
     ((ae: arctic_rat) => pos_arctic_rat(ae),
       (_: arctic_rat) => false))))))))))))),
ad, b),
           (false,
             (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil),
                                 List('x'),
                                 (shows_prec_nat(Zero_nat, n)).apply(Nil),
                                 List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e',
                                       's', ' ', 'o', 'v', 'e', 'r', ' '),
                                 get_comp_description[arctic_rat](((_:
                              arctic_rat)
                             => true,
                            (zero_arctic_rat,
                              (one_arctic_rat,
                                (one_arctic_rat,
                                  ((ad: arctic_rat) =>
                                     (b: arctic_rat) => plus_arctic_rat(ad, b),
                                    ((ad: arctic_rat) =>
                                       (b: arctic_rat) =>
 times_arctic_rat(ad, b),
                                      (id[arctic_rat],
((ad: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ad, b),
  ((ad: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ad, b),
    (false,
      (arctic_rat_name,
        ((ad: arctic_rat) => pos_arctic_rat(ad),
          (_: arctic_rat) => false))))))))))))))),
               ((ad: List[List[arctic_rat]]) =>
                  mat_arc_posI[arctic_rat](get_comp_apos[arctic_rat](((_:
                                 arctic_rat)
                                => true,
                               (zero_arctic_rat,
                                 (one_arctic_rat,
                                   (one_arctic_rat,
                                     ((ae: arctic_rat) =>
(b: arctic_rat) => plus_arctic_rat(ae, b),
                                       ((ae: arctic_rat) =>
  (b: arctic_rat) => times_arctic_rat(ae, b),
 (id[arctic_rat],
   ((ae: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(ae, b),
     ((ae: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, ae, b),
       (false,
         (arctic_rat_name,
           ((ae: arctic_rat) => pos_arctic_rat(ae),
             (_: arctic_rat) => false))))))))))))),
    ad),
                 (_: List[List[arctic_rat]]) => false))))))))))))),
                                       ac, snd[List[List[arctic_rat]],
        List[List[List[arctic_rat]]]](poly))))),
        ab)
    case Rational(d, deg) =>
      (if (less_eq_nat(deg, one_nat))
        (ab: xml) =>
          xml3elements[A, nat, (rat, List[rat]),
                        alter[((A, nat), (BigInt, List[BigInt])),
                               ((A, nat), (rat, List[rat])),
                               ((A, nat), (arctic, List[arctic])),
                               ((A, nat),
                                 (List[List[BigInt]],
                                   List[List[List[BigInt]]])),
                               ((A, nat),
                                 (List[List[rat]], List[List[List[rat]]])),
                               ((A, nat),
                                 (List[List[arctic]],
                                   List[List[List[arctic]]])),
                               ((A, nat), (arctic_rat, List[arctic_rat])),
                               ((A, nat),
                                 (List[List[arctic_rat]],
                                   List[List[List[arctic_rat]]])),
                               ((A, nat), List[(List[(nat, nat)], BigInt)]),
                               ((A, nat),
                                 List[(List[(nat, nat)],
rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
          (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
          (x: xml) =>
            binda[List[Char], tpoly[nat, rat],
                   (rat, List[rat])](polynomial[rat](rat, x),
                                      (ac: tpoly[nat, rat]) =>
lpoly_of[rat](get_comp_0[rat](((_: rat) => true,
                                (fract(BigInt(0), BigInt(1)),
                                  (fract(BigInt(1), BigInt(1)),
                                    (d, ((ad: rat) =>
   (b: rat) => plus_rata(ad, b),
  ((ad: rat) => (b: rat) => times_rata(ad, b),
    ((ad: rat) => max0_rat(ad),
      ((ad: rat) => (b: rat) => ge_rat(ad, b),
        (rat_gt(d),
          (true,
            (rat_name,
              ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
               get_comp_1[rat](((_: rat) => true,
                                 (fract(BigInt(0), BigInt(1)),
                                   (fract(BigInt(1), BigInt(1)),
                                     (d, ((ad: rat) =>
    (b: rat) => plus_rata(ad, b),
   ((ad: rat) => (b: rat) => times_rata(ad, b),
     ((ad: rat) => max0_rat(ad),
       ((ad: rat) => (b: rat) => ge_rat(ad, b),
         (rat_gt(d),
           (true,
             (rat_name,
               ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
               get_comp_plus[rat](((_: rat) => true,
                                    (fract(BigInt(0), BigInt(1)),
                                      (fract(BigInt(1), BigInt(1)),
(d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
      ((ad: rat) => (b: rat) => times_rata(ad, b),
        ((ad: rat) => max0_rat(ad),
          ((ad: rat) => (b: rat) => ge_rat(ad, b),
            (rat_gt(d),
              (true,
                (rat_name,
                  ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
               get_comp_times[rat](((_: rat) => true,
                                     (fract(BigInt(0), BigInt(1)),
                                       (fract(BigInt(1), BigInt(1)),
 (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
       ((ad: rat) => (b: rat) => times_rata(ad, b),
         ((ad: rat) => max0_rat(ad),
           ((ad: rat) => (b: rat) => ge_rat(ad, b),
             (rat_gt(d),
               (true,
                 (rat_name,
                   ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
               ac)),
          (f: A) =>
            (ac: nat) =>
              (poly: (rat, List[rat])) =>
                Alter2[((A, nat), (rat, List[rat])),
                        ((A, nat), (BigInt, List[BigInt])),
                        ((A, nat), (arctic, List[arctic])),
                        ((A, nat),
                          (List[List[BigInt]], List[List[List[BigInt]]])),
                        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                        ((A, nat),
                          (List[List[arctic]], List[List[List[arctic]]])),
                        ((A, nat), (arctic_rat, List[arctic_rat])),
                        ((A, nat),
                          (List[List[arctic_rat]],
                            List[List[List[arctic_rat]]])),
                        ((A, nat), List[(List[(nat, nat)], BigInt)]),
                        ((A, nat),
                          List[(List[(nat, nat)],
                                 rat)])](((f, ac),
   (fst[rat, List[rat]](poly),
     fit_length[rat](get_comp_0[rat](((_: rat) => true,
                                       (fract(BigInt(0), BigInt(1)),
 (fract(BigInt(1), BigInt(1)),
   (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
         ((ad: rat) => (b: rat) => times_rata(ad, b),
           ((ad: rat) => max0_rat(ad),
             ((ad: rat) => (b: rat) => ge_rat(ad, b),
               (rat_gt(d),
                 (true,
                   (rat_name,
                     ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
                      ac, snd[rat, List[rat]](poly))))),
          ab)
        else (ab: xml) =>
               xml3elements[A, nat, List[(List[(nat, nat)], rat)],
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 binda[List[Char], tpoly[nat, rat],
                        List[(List[(nat, nat)],
                               rat)]](polynomial[rat](rat, x),
                                       (p: tpoly[nat, rat]) =>
 Inr[List[(List[(nat, nat)], rat)], List[Char]](poly_of[nat, rat](p))),
               (f: A) =>
                 (ac: nat) =>
                   (poly: List[(List[(nat, nat)], rat)]) =>
                     Alter10[((A, nat), List[(List[(nat, nat)], rat)]),
                              ((A, nat), (BigInt, List[BigInt])),
                              ((A, nat), (rat, List[rat])),
                              ((A, nat), (arctic, List[arctic])),
                              ((A, nat),
                                (List[List[BigInt]], List[List[List[BigInt]]])),
                              ((A, nat),
                                (List[List[rat]], List[List[List[rat]]])),
                              ((A, nat),
                                (List[List[arctic]], List[List[List[arctic]]])),
                              ((A, nat), (arctic_rat, List[arctic_rat])),
                              ((A, nat),
                                (List[List[arctic_rat]],
                                  List[List[List[arctic_rat]]])),
                              ((A, nat),
                                List[(List[(nat, nat)],
                                       BigInt)])](((f, ac), poly)),
               ab))
    case Rat_mat(d, n, sd) =>
      (ab: xml) =>
        xml3elements[A, nat, (List[List[rat]], List[List[List[rat]]]),
                      alter[((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])]](List('i', 'n', 't', 'e', 'r',
             'p', 'r', 'e', 't'),
        xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
        (x: xml) =>
          binda[List[Char], tpoly[nat, List[List[rat]]],
                 (List[List[rat]],
                   List[List[List[rat]]])](polynomial[List[List[rat]]](mat_coeff[rat](fract(BigInt(0),
             BigInt(1)),
       rat),
                                x),
    (ac: tpoly[nat, List[List[rat]]]) =>
      lpoly_of[List[List[rat]]](get_comp_0[List[List[rat]]](((ad:
                        List[List[rat]])
                       => mat[rat](n, n, ad),
                      (mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                        (mat1I[rat](fract(BigInt(0), BigInt(1)),
                                     fract(BigInt(1), BigInt(1)), n),
                          (mat1I[rat](fract(BigInt(0), BigInt(1)),
                                       get_comp_def[rat](((_: rat) => true,
                   (fract(BigInt(0), BigInt(1)),
                     (fract(BigInt(1), BigInt(1)),
                       (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                             ((ad: rat) => (b: rat) => times_rata(ad, b),
                               ((ad: rat) => max0_rat(ad),
                                 ((ad: rat) => (b: rat) => ge_rat(ad, b),
                                   (rat_gt(d),
                                     (true,
                                       (rat_name,
 ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
                                       n),
                            ((ad: List[List[rat]]) =>
                               (b: List[List[rat]]) =>
                                 mat_plusI[rat]((ae: rat) =>
          (ba: rat) => plus_rata(ae, ba),
         ad, b),
                              ((ad: List[List[rat]]) =>
                                 (b: List[List[rat]]) =>
                                   mat_multI[rat](fract(BigInt(0), BigInt(1)),
           (ae: rat) => (ba: rat) => plus_rata(ae, ba),
           (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                                (mat_max0[rat],
                                  (mat_ge[rat],
                                    ((ad: List[List[rat]]) =>
                                       (b: List[List[rat]]) =>
 mat_gtI[rat](get_comp_gt[rat](((_: rat) => true,
                                 (fract(BigInt(0), BigInt(1)),
                                   (fract(BigInt(1), BigInt(1)),
                                     (d, ((ae: rat) =>
    (ba: rat) => plus_rata(ae, ba),
   ((ae: rat) => (ba: rat) => times_rata(ae, ba),
     ((ae: rat) => max0_rat(ae),
       ((ae: rat) => (ba: rat) => ge_rat(ae, ba),
         (rat_gt(d),
           (true,
             (rat_name,
               ((_: rat) => true, (ae: rat) => rat_mono(ae)))))))))))))),
               sd, ad, b),
                                      (get_comp_mono[rat](((_: rat) => true,
                    (fract(BigInt(0), BigInt(1)),
                      (fract(BigInt(1), BigInt(1)),
                        (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                              ((ad: rat) => (b: rat) => times_rata(ad, b),
                                ((ad: rat) => max0_rat(ad),
                                  ((ad: rat) => (b: rat) => ge_rat(ad, b),
                                    (rat_gt(d),
                                      (true,
(rat_name, ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
(concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                    (shows_prec_nat(Zero_nat, n)).apply(Nil),
                    List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w',
                          'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                          'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                    (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                    List(' ', 'o', 'v', 'e', 'r', ' '),
                    get_comp_description[rat](((_: rat) => true,
        (fract(BigInt(0), BigInt(1)),
          (fract(BigInt(1), BigInt(1)),
            (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                  ((ad: rat) => (b: rat) => times_rata(ad, b),
                    ((ad: rat) => max0_rat(ad),
                      ((ad: rat) => (b: rat) => ge_rat(ad, b),
                        (rat_gt(d),
                          (true,
                            (rat_name,
                              ((_: rat) => true,
                                (ad: rat) => rat_mono(ad)))))))))))))))),
  ((_: List[List[rat]]) => true,
    mat_mono[rat](sd, get_comp_check_mono[rat](((_: rat) => true,
         (fract(BigInt(0), BigInt(1)),
           (fract(BigInt(1), BigInt(1)),
             (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                   ((ad: rat) => (b: rat) => times_rata(ad, b),
                     ((ad: rat) => max0_rat(ad),
                       ((ad: rat) => (b: rat) => ge_rat(ad, b),
                         (rat_gt(d),
                           (true,
                             (rat_name,
                               ((_: rat) => true,
                                 (ad: rat) =>
                                   rat_mono(ad)))))))))))))))))))))))))))),
                                 get_comp_1[List[List[rat]]](((ad:
                         List[List[rat]])
                        => mat[rat](n, n, ad),
                       (mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                         (mat1I[rat](fract(BigInt(0), BigInt(1)),
                                      fract(BigInt(1), BigInt(1)), n),
                           (mat1I[rat](fract(BigInt(0), BigInt(1)),
get_comp_def[rat](((_: rat) => true,
                    (fract(BigInt(0), BigInt(1)),
                      (fract(BigInt(1), BigInt(1)),
                        (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                              ((ad: rat) => (b: rat) => times_rata(ad, b),
                                ((ad: rat) => max0_rat(ad),
                                  ((ad: rat) => (b: rat) => ge_rat(ad, b),
                                    (rat_gt(d),
                                      (true,
(rat_name, ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
n),
                             ((ad: List[List[rat]]) =>
                                (b: List[List[rat]]) =>
                                  mat_plusI[rat]((ae: rat) =>
           (ba: rat) => plus_rata(ae, ba),
          ad, b),
                               ((ad: List[List[rat]]) =>
                                  (b: List[List[rat]]) =>
                                    mat_multI[rat](fract(BigInt(0), BigInt(1)),
            (ae: rat) => (ba: rat) => plus_rata(ae, ba),
            (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                                 (mat_max0[rat],
                                   (mat_ge[rat],
                                     ((ad: List[List[rat]]) =>
(b: List[List[rat]]) =>
  mat_gtI[rat](get_comp_gt[rat](((_: rat) => true,
                                  (fract(BigInt(0), BigInt(1)),
                                    (fract(BigInt(1), BigInt(1)),
                                      (d,
((ae: rat) => (ba: rat) => plus_rata(ae, ba),
  ((ae: rat) => (ba: rat) => times_rata(ae, ba),
    ((ae: rat) => max0_rat(ae),
      ((ae: rat) => (ba: rat) => ge_rat(ae, ba),
        (rat_gt(d),
          (true,
            (rat_name,
              ((_: rat) => true, (ae: rat) => rat_mono(ae)))))))))))))),
                sd, ad, b),
                                       (get_comp_mono[rat](((_: rat) => true,
                     (fract(BigInt(0), BigInt(1)),
                       (fract(BigInt(1), BigInt(1)),
                         (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                               ((ad: rat) => (b: rat) => times_rata(ad, b),
                                 ((ad: rat) => max0_rat(ad),
                                   ((ad: rat) => (b: rat) => ge_rat(ad, b),
                                     (rat_gt(d),
                                       (true,
 (rat_name, ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
 (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                     (shows_prec_nat(Zero_nat, n)).apply(Nil),
                     List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w',
                           'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't',
                           ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n',
                           ' '),
                     (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                     List(' ', 'o', 'v', 'e', 'r', ' '),
                     get_comp_description[rat](((_: rat) => true,
         (fract(BigInt(0), BigInt(1)),
           (fract(BigInt(1), BigInt(1)),
             (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                   ((ad: rat) => (b: rat) => times_rata(ad, b),
                     ((ad: rat) => max0_rat(ad),
                       ((ad: rat) => (b: rat) => ge_rat(ad, b),
                         (rat_gt(d),
                           (true,
                             (rat_name,
                               ((_: rat) => true,
                                 (ad: rat) => rat_mono(ad)))))))))))))))),
   ((_: List[List[rat]]) => true,
     mat_mono[rat](sd, get_comp_check_mono[rat](((_: rat) => true,
          (fract(BigInt(0), BigInt(1)),
            (fract(BigInt(1), BigInt(1)),
              (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                    ((ad: rat) => (b: rat) => times_rata(ad, b),
                      ((ad: rat) => max0_rat(ad),
                        ((ad: rat) => (b: rat) => ge_rat(ad, b),
                          (rat_gt(d),
                            (true,
                              (rat_name,
                                ((_: rat) => true,
                                  (ad: rat) =>
                                    rat_mono(ad)))))))))))))))))))))))))))),
                                 get_comp_plus[List[List[rat]]](((ad:
                            List[List[rat]])
                           => mat[rat](n, n, ad),
                          (mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                            (mat1I[rat](fract(BigInt(0), BigInt(1)),
 fract(BigInt(1), BigInt(1)), n),
                              (mat1I[rat](fract(BigInt(0), BigInt(1)),
   get_comp_def[rat](((_: rat) => true,
                       (fract(BigInt(0), BigInt(1)),
                         (fract(BigInt(1), BigInt(1)),
                           (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                                 ((ad: rat) => (b: rat) => times_rata(ad, b),
                                   ((ad: rat) => max0_rat(ad),
                                     ((ad: rat) => (b: rat) => ge_rat(ad, b),
                                       (rat_gt(d),
 (true, (rat_name, ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
   n),
                                ((ad: List[List[rat]]) =>
                                   (b: List[List[rat]]) =>
                                     mat_plusI[rat]((ae: rat) =>
              (ba: rat) => plus_rata(ae, ba),
             ad, b),
                                  ((ad: List[List[rat]]) =>
                                     (b: List[List[rat]]) =>
                                       mat_multI[rat](fract(BigInt(0),
                     BigInt(1)),
               (ae: rat) => (ba: rat) => plus_rata(ae, ba),
               (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                                    (mat_max0[rat],
                                      (mat_ge[rat],
((ad: List[List[rat]]) =>
   (b: List[List[rat]]) =>
     mat_gtI[rat](get_comp_gt[rat](((_: rat) => true,
                                     (fract(BigInt(0), BigInt(1)),
                                       (fract(BigInt(1), BigInt(1)),
 (d, ((ae: rat) => (ba: rat) => plus_rata(ae, ba),
       ((ae: rat) => (ba: rat) => times_rata(ae, ba),
         ((ae: rat) => max0_rat(ae),
           ((ae: rat) => (ba: rat) => ge_rat(ae, ba),
             (rat_gt(d),
               (true,
                 (rat_name,
                   ((_: rat) => true, (ae: rat) => rat_mono(ae)))))))))))))),
                   sd, ad, b),
  (get_comp_mono[rat](((_: rat) => true,
                        (fract(BigInt(0), BigInt(1)),
                          (fract(BigInt(1), BigInt(1)),
                            (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                                  ((ad: rat) => (b: rat) => times_rata(ad, b),
                                    ((ad: rat) => max0_rat(ad),
                                      ((ad: rat) => (b: rat) => ge_rat(ad, b),
(rat_gt(d),
  (true, (rat_name, ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
    (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                        (shows_prec_nat(Zero_nat, n)).apply(Nil),
                        List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                              'w', 'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c',
                              't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o',
                              'n', ' '),
                        (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                        List(' ', 'o', 'v', 'e', 'r', ' '),
                        get_comp_description[rat](((_: rat) => true,
            (fract(BigInt(0), BigInt(1)),
              (fract(BigInt(1), BigInt(1)),
                (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                      ((ad: rat) => (b: rat) => times_rata(ad, b),
                        ((ad: rat) => max0_rat(ad),
                          ((ad: rat) => (b: rat) => ge_rat(ad, b),
                            (rat_gt(d),
                              (true,
                                (rat_name,
                                  ((_: rat) => true,
                                    (ad: rat) => rat_mono(ad)))))))))))))))),
      ((_: List[List[rat]]) => true,
        mat_mono[rat](sd, get_comp_check_mono[rat](((_: rat) => true,
             (fract(BigInt(0), BigInt(1)),
               (fract(BigInt(1), BigInt(1)),
                 (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                       ((ad: rat) => (b: rat) => times_rata(ad, b),
                         ((ad: rat) => max0_rat(ad),
                           ((ad: rat) => (b: rat) => ge_rat(ad, b),
                             (rat_gt(d),
                               (true,
                                 (rat_name,
                                   ((_: rat) => true,
                                     (ad: rat) =>
                                       rat_mono(ad)))))))))))))))))))))))))))),
                                 get_comp_times[List[List[rat]]](((ad:
                             List[List[rat]])
                            => mat[rat](n, n, ad),
                           (mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                             (mat1I[rat](fract(BigInt(0), BigInt(1)),
  fract(BigInt(1), BigInt(1)), n),
                               (mat1I[rat](fract(BigInt(0), BigInt(1)),
    get_comp_def[rat](((_: rat) => true,
                        (fract(BigInt(0), BigInt(1)),
                          (fract(BigInt(1), BigInt(1)),
                            (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                                  ((ad: rat) => (b: rat) => times_rata(ad, b),
                                    ((ad: rat) => max0_rat(ad),
                                      ((ad: rat) => (b: rat) => ge_rat(ad, b),
(rat_gt(d),
  (true, (rat_name, ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
    n),
                                 ((ad: List[List[rat]]) =>
                                    (b: List[List[rat]]) =>
                                      mat_plusI[rat]((ae: rat) =>
               (ba: rat) => plus_rata(ae, ba),
              ad, b),
                                   ((ad: List[List[rat]]) =>
                                      (b: List[List[rat]]) =>
mat_multI[rat](fract(BigInt(0), BigInt(1)),
                (ae: rat) => (ba: rat) => plus_rata(ae, ba),
                (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                                     (mat_max0[rat],
                                       (mat_ge[rat],
 ((ad: List[List[rat]]) =>
    (b: List[List[rat]]) =>
      mat_gtI[rat](get_comp_gt[rat](((_: rat) => true,
                                      (fract(BigInt(0), BigInt(1)),
(fract(BigInt(1), BigInt(1)),
  (d, ((ae: rat) => (ba: rat) => plus_rata(ae, ba),
        ((ae: rat) => (ba: rat) => times_rata(ae, ba),
          ((ae: rat) => max0_rat(ae),
            ((ae: rat) => (ba: rat) => ge_rat(ae, ba),
              (rat_gt(d),
                (true,
                  (rat_name,
                    ((_: rat) => true, (ae: rat) => rat_mono(ae)))))))))))))),
                    sd, ad, b),
   (get_comp_mono[rat](((_: rat) => true,
                         (fract(BigInt(0), BigInt(1)),
                           (fract(BigInt(1), BigInt(1)),
                             (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                                   ((ad: rat) => (b: rat) => times_rata(ad, b),
                                     ((ad: rat) => max0_rat(ad),
                                       ((ad: rat) => (b: rat) => ge_rat(ad, b),
 (rat_gt(d),
   (true, (rat_name, ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
     (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                         (shows_prec_nat(Zero_nat, n)).apply(Nil),
                         List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                               'w', 'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c',
                               't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o',
                               'n', ' '),
                         (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                         List(' ', 'o', 'v', 'e', 'r', ' '),
                         get_comp_description[rat](((_: rat) => true,
             (fract(BigInt(0), BigInt(1)),
               (fract(BigInt(1), BigInt(1)),
                 (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                       ((ad: rat) => (b: rat) => times_rata(ad, b),
                         ((ad: rat) => max0_rat(ad),
                           ((ad: rat) => (b: rat) => ge_rat(ad, b),
                             (rat_gt(d),
                               (true,
                                 (rat_name,
                                   ((_: rat) => true,
                                     (ad: rat) => rat_mono(ad)))))))))))))))),
       ((_: List[List[rat]]) => true,
         mat_mono[rat](sd, get_comp_check_mono[rat](((_: rat) => true,
              (fract(BigInt(0), BigInt(1)),
                (fract(BigInt(1), BigInt(1)),
                  (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                        ((ad: rat) => (b: rat) => times_rata(ad, b),
                          ((ad: rat) => max0_rat(ad),
                            ((ad: rat) => (b: rat) => ge_rat(ad, b),
                              (rat_gt(d),
                                (true,
                                  (rat_name,
                                    ((_: rat) => true,
                                      (ad: rat) =>
rat_mono(ad)))))))))))))))))))))))))))),
                                 ac)),
        (f: A) =>
          (ac: nat) =>
            (poly: (List[List[rat]], List[List[List[rat]]])) =>
              Alter5[((A, nat), (List[List[rat]], List[List[List[rat]]])),
                      ((A, nat), (BigInt, List[BigInt])),
                      ((A, nat), (rat, List[rat])),
                      ((A, nat), (arctic, List[arctic])),
                      ((A, nat),
                        (List[List[BigInt]], List[List[List[BigInt]]])),
                      ((A, nat),
                        (List[List[arctic]], List[List[List[arctic]]])),
                      ((A, nat), (arctic_rat, List[arctic_rat])),
                      ((A, nat),
                        (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                      ((A, nat), List[(List[(nat, nat)], BigInt)]),
                      ((A, nat),
                        List[(List[(nat, nat)],
                               rat)])](((f, ac),
 (fst[List[List[rat]], List[List[List[rat]]]](poly),
   fit_length[List[List[rat]]](get_comp_0[List[List[rat]]](((ad:
                       List[List[rat]])
                      => mat[rat](n, n, ad),
                     (mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                       (mat1I[rat](fract(BigInt(0), BigInt(1)),
                                    fract(BigInt(1), BigInt(1)), n),
                         (mat1I[rat](fract(BigInt(0), BigInt(1)),
                                      get_comp_def[rat](((_: rat) => true,
                  (fract(BigInt(0), BigInt(1)),
                    (fract(BigInt(1), BigInt(1)),
                      (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                            ((ad: rat) => (b: rat) => times_rata(ad, b),
                              ((ad: rat) => max0_rat(ad),
                                ((ad: rat) => (b: rat) => ge_rat(ad, b),
                                  (rat_gt(d),
                                    (true,
                                      (rat_name,
((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
                                      n),
                           ((ad: List[List[rat]]) =>
                              (b: List[List[rat]]) =>
                                mat_plusI[rat]((ae: rat) =>
         (ba: rat) => plus_rata(ae, ba),
        ad, b),
                             ((ad: List[List[rat]]) =>
                                (b: List[List[rat]]) =>
                                  mat_multI[rat](fract(BigInt(0), BigInt(1)),
          (ae: rat) => (ba: rat) => plus_rata(ae, ba),
          (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                               (mat_max0[rat],
                                 (mat_ge[rat],
                                   ((ad: List[List[rat]]) =>
                                      (b: List[List[rat]]) =>
mat_gtI[rat](get_comp_gt[rat](((_: rat) => true,
                                (fract(BigInt(0), BigInt(1)),
                                  (fract(BigInt(1), BigInt(1)),
                                    (d, ((ae: rat) =>
   (ba: rat) => plus_rata(ae, ba),
  ((ae: rat) => (ba: rat) => times_rata(ae, ba),
    ((ae: rat) => max0_rat(ae),
      ((ae: rat) => (ba: rat) => ge_rat(ae, ba),
        (rat_gt(d),
          (true,
            (rat_name,
              ((_: rat) => true, (ae: rat) => rat_mono(ae)))))))))))))),
              sd, ad, b),
                                     (get_comp_mono[rat](((_: rat) => true,
                   (fract(BigInt(0), BigInt(1)),
                     (fract(BigInt(1), BigInt(1)),
                       (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                             ((ad: rat) => (b: rat) => times_rata(ad, b),
                               ((ad: rat) => max0_rat(ad),
                                 ((ad: rat) => (b: rat) => ge_rat(ad, b),
                                   (rat_gt(d),
                                     (true,
                                       (rat_name,
 ((_: rat) => true, (ad: rat) => rat_mono(ad)))))))))))))),
                                       (concat[Char](List((shows_prec_nat(Zero_nat,
                                   n)).apply(Nil),
                   List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                   List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w',
                         'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                         'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                   (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                   List(' ', 'o', 'v', 'e', 'r', ' '),
                   get_comp_description[rat](((_: rat) => true,
       (fract(BigInt(0), BigInt(1)),
         (fract(BigInt(1), BigInt(1)),
           (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                 ((ad: rat) => (b: rat) => times_rata(ad, b),
                   ((ad: rat) => max0_rat(ad),
                     ((ad: rat) => (b: rat) => ge_rat(ad, b),
                       (rat_gt(d),
                         (true,
                           (rat_name,
                             ((_: rat) => true,
                               (ad: rat) => rat_mono(ad)))))))))))))))),
 ((_: List[List[rat]]) => true,
   mat_mono[rat](sd, get_comp_check_mono[rat](((_: rat) => true,
        (fract(BigInt(0), BigInt(1)),
          (fract(BigInt(1), BigInt(1)),
            (d, ((ad: rat) => (b: rat) => plus_rata(ad, b),
                  ((ad: rat) => (b: rat) => times_rata(ad, b),
                    ((ad: rat) => max0_rat(ad),
                      ((ad: rat) => (b: rat) => ge_rat(ad, b),
                        (rat_gt(d),
                          (true,
                            (rat_name,
                              ((_: rat) => true,
                                (ad: rat) =>
                                  rat_mono(ad)))))))))))))))))))))))))))),
                                ac, snd[List[List[rat]],
 List[List[List[rat]]]](poly))))),
        ab)
  },
(typea: domain) =>
  (pi: List[alter[((A, nat), (BigInt, List[BigInt])),
                   ((A, nat), (rat, List[rat])),
                   ((A, nat), (arctic, List[arctic])),
                   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
                   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
                   ((A, nat), (arctic_rat, List[arctic_rat])),
                   ((A, nat),
                     (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                   ((A, nat), List[(List[(nat, nat)], BigInt)]),
                   ((A, nat), List[(List[(nat, nat)], rat)])]])
    => typea match {
         case Naturala(deg) =>
           (if (less_eq_nat(deg, one_nat))
             Int_carrier[A](mapa[alter[((A, nat), (BigInt, List[BigInt])),
((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
((A, nat), (List[List[rat]], List[List[List[rat]]])),
((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
((A, nat), (arctic_rat, List[arctic_rat])),
((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
((A, nat), List[(List[(nat, nat)], BigInt)]),
((A, nat), List[(List[(nat, nat)], rat)])],
                                  ((A, nat),
                                    (BigInt,
                                      List[BigInt]))]((aa:
                 alter[((A, nat), (BigInt, List[BigInt])),
                        ((A, nat), (rat, List[rat])),
                        ((A, nat), (arctic, List[arctic])),
                        ((A, nat),
                          (List[List[BigInt]], List[List[List[BigInt]]])),
                        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                        ((A, nat),
                          (List[List[arctic]], List[List[List[arctic]]])),
                        ((A, nat), (arctic_rat, List[arctic_rat])),
                        ((A, nat),
                          (List[List[arctic_rat]],
                            List[List[List[arctic_rat]]])),
                        ((A, nat), List[(List[(nat, nat)], BigInt)]),
                        ((A, nat), List[(List[(nat, nat)], rat)])])
                => alter1a[((A, nat), (BigInt, List[BigInt])),
                            ((A, nat), (rat, List[rat])),
                            ((A, nat), (arctic, List[arctic])),
                            ((A, nat),
                              (List[List[BigInt]], List[List[List[BigInt]]])),
                            ((A, nat),
                              (List[List[rat]], List[List[List[rat]]])),
                            ((A, nat),
                              (List[List[arctic]], List[List[List[arctic]]])),
                            ((A, nat), (arctic_rat, List[arctic_rat])),
                            ((A, nat),
                              (List[List[arctic_rat]],
                                List[List[List[arctic_rat]]])),
                            ((A, nat), List[(List[(nat, nat)], BigInt)]),
                            ((A, nat), List[(List[(nat, nat)], rat)])](aa),
               pi))
             else Int_nl_carrier[A](mapa[alter[((A, nat),
         (BigInt, List[BigInt])),
        ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
        ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
        ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
        ((A, nat), (arctic_rat, List[arctic_rat])),
        ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
        ((A, nat), List[(List[(nat, nat)], BigInt)]),
        ((A, nat), List[(List[(nat, nat)], rat)])],
  ((A, nat),
    List[(List[(nat, nat)],
           BigInt)])]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                  ((A, nat), (rat, List[rat])),
                                  ((A, nat), (arctic, List[arctic])),
                                  ((A, nat),
                                    (List[List[BigInt]],
                                      List[List[List[BigInt]]])),
                                  ((A, nat),
                                    (List[List[rat]], List[List[List[rat]]])),
                                  ((A, nat),
                                    (List[List[arctic]],
                                      List[List[List[arctic]]])),
                                  ((A, nat), (arctic_rat, List[arctic_rat])),
                                  ((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]])),
                                  ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                  ((A, nat), List[(List[(nat, nat)], rat)])])
                        => alter9a[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)], rat)])](aa),
                       pi)))
         case Arctic =>
           Arctic_carrier[A](mapa[alter[((A, nat), (BigInt, List[BigInt])),
 ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
 ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
 ((A, nat), (List[List[rat]], List[List[List[rat]]])),
 ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
 ((A, nat), (arctic_rat, List[arctic_rat])),
 ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
 ((A, nat), List[(List[(nat, nat)], BigInt)]),
 ((A, nat), List[(List[(nat, nat)], rat)])],
                                   ((A, nat),
                                     (arctic,
                                       List[arctic]))]((aa:
                  alter[((A, nat), (BigInt, List[BigInt])),
                         ((A, nat), (rat, List[rat])),
                         ((A, nat), (arctic, List[arctic])),
                         ((A, nat),
                           (List[List[BigInt]], List[List[List[BigInt]]])),
                         ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                         ((A, nat),
                           (List[List[arctic]], List[List[List[arctic]]])),
                         ((A, nat), (arctic_rat, List[arctic_rat])),
                         ((A, nat),
                           (List[List[arctic_rat]],
                             List[List[List[arctic_rat]]])),
                         ((A, nat), List[(List[(nat, nat)], BigInt)]),
                         ((A, nat), List[(List[(nat, nat)], rat)])])
                 => alter3a[((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                pi))
         case Arctic_rat(d) =>
           Arctic_rat_carrier[A](d, mapa[alter[((A, nat),
         (BigInt, List[BigInt])),
        ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
        ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
        ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
        ((A, nat), (arctic_rat, List[arctic_rat])),
        ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
        ((A, nat), List[(List[(nat, nat)], BigInt)]),
        ((A, nat), List[(List[(nat, nat)], rat)])],
  ((A, nat),
    (arctic_rat,
      List[arctic_rat]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                      ((A, nat), (rat, List[rat])),
                                      ((A, nat), (arctic, List[arctic])),
                                      ((A, nat),
(List[List[BigInt]], List[List[List[BigInt]]])),
                                      ((A, nat),
(List[List[rat]], List[List[List[rat]]])),
                                      ((A, nat),
(List[List[arctic]], List[List[List[arctic]]])),
                                      ((A, nat),
(arctic_rat, List[arctic_rat])),
                                      ((A, nat),
(List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                                      ((A, nat),
List[(List[(nat, nat)], BigInt)]),
                                      ((A, nat),
List[(List[(nat, nat)], rat)])])
                            => alter7a[((A, nat), (BigInt, List[BigInt])),
((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
((A, nat), (List[List[rat]], List[List[List[rat]]])),
((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
((A, nat), (arctic_rat, List[arctic_rat])),
((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
((A, nat), List[(List[(nat, nat)], BigInt)]),
((A, nat), List[(List[(nat, nat)], rat)])](aa),
                           pi))
         case Int_mat(n, sd) =>
           Int_mat_carrier[A](n, sd,
                               mapa[alter[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat), List[(List[(nat, nat)], rat)])],
                                     ((A, nat),
                                       (List[List[BigInt]],
 List[List[List[BigInt]]]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
 ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
 ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
 ((A, nat), (List[List[rat]], List[List[List[rat]]])),
 ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
 ((A, nat), (arctic_rat, List[arctic_rat])),
 ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
 ((A, nat), List[(List[(nat, nat)], BigInt)]),
 ((A, nat), List[(List[(nat, nat)], rat)])])
                               => alter4a[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                              pi))
         case Arc_mat(n) =>
           Arc_mat_carrier[A](n, mapa[alter[((A, nat), (BigInt, List[BigInt])),
     ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
     ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
     ((A, nat), (List[List[rat]], List[List[List[rat]]])),
     ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
     ((A, nat), (arctic_rat, List[arctic_rat])),
     ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
     ((A, nat), List[(List[(nat, nat)], BigInt)]),
     ((A, nat), List[(List[(nat, nat)], rat)])],
                                       ((A, nat),
 (List[List[arctic]],
   List[List[List[arctic]]]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat), List[(List[(nat, nat)], rat)])])
                                 => alter6a[((A, nat), (BigInt, List[BigInt])),
     ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
     ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
     ((A, nat), (List[List[rat]], List[List[List[rat]]])),
     ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
     ((A, nat), (arctic_rat, List[arctic_rat])),
     ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
     ((A, nat), List[(List[(nat, nat)], BigInt)]),
     ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                                pi))
         case Arc_rat_mat(d, n) =>
           Arc_rat_mat_carrier[A](d, n,
                                   mapa[alter[((A, nat),
        (BigInt, List[BigInt])),
       ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
       ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
       ((A, nat), (List[List[rat]], List[List[List[rat]]])),
       ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
       ((A, nat), (arctic_rat, List[arctic_rat])),
       ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
       ((A, nat), List[(List[(nat, nat)], BigInt)]),
       ((A, nat), List[(List[(nat, nat)], rat)])],
 ((A, nat),
   (List[List[arctic_rat]],
     List[List[List[arctic_rat]]]))]((aa:
alter[((A, nat), (BigInt, List[BigInt])), ((A, nat), (rat, List[rat])),
       ((A, nat), (arctic, List[arctic])),
       ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
       ((A, nat), (List[List[rat]], List[List[List[rat]]])),
       ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
       ((A, nat), (arctic_rat, List[arctic_rat])),
       ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
       ((A, nat), List[(List[(nat, nat)], BigInt)]),
       ((A, nat), List[(List[(nat, nat)], rat)])])
                                       => alter8a[((A, nat),
            (BigInt, List[BigInt])),
           ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
           ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
           ((A, nat), (List[List[rat]], List[List[List[rat]]])),
           ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
           ((A, nat), (arctic_rat, List[arctic_rat])),
           ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
           ((A, nat), List[(List[(nat, nat)], BigInt)]),
           ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                                      pi))
         case Rational(d, deg) =>
           (if (less_eq_nat(deg, one_nat))
             Rat_carrier[A](d, mapa[alter[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat), List[(List[(nat, nat)], rat)])],
                                     ((A, nat),
                                       (rat,
 List[rat]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                          ((A, nat), (rat, List[rat])),
                          ((A, nat), (arctic, List[arctic])),
                          ((A, nat),
                            (List[List[BigInt]], List[List[List[BigInt]]])),
                          ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                          ((A, nat),
                            (List[List[arctic]], List[List[List[arctic]]])),
                          ((A, nat), (arctic_rat, List[arctic_rat])),
                          ((A, nat),
                            (List[List[arctic_rat]],
                              List[List[List[arctic_rat]]])),
                          ((A, nat), List[(List[(nat, nat)], BigInt)]),
                          ((A, nat), List[(List[(nat, nat)], rat)])])
                => alter2a[((A, nat), (BigInt, List[BigInt])),
                            ((A, nat), (rat, List[rat])),
                            ((A, nat), (arctic, List[arctic])),
                            ((A, nat),
                              (List[List[BigInt]], List[List[List[BigInt]]])),
                            ((A, nat),
                              (List[List[rat]], List[List[List[rat]]])),
                            ((A, nat),
                              (List[List[arctic]], List[List[List[arctic]]])),
                            ((A, nat), (arctic_rat, List[arctic_rat])),
                            ((A, nat),
                              (List[List[arctic_rat]],
                                List[List[List[arctic_rat]]])),
                            ((A, nat), List[(List[(nat, nat)], BigInt)]),
                            ((A, nat), List[(List[(nat, nat)], rat)])](aa),
               pi))
             else Rat_nl_carrier[A](d, mapa[alter[((A, nat),
            (BigInt, List[BigInt])),
           ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
           ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
           ((A, nat), (List[List[rat]], List[List[List[rat]]])),
           ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
           ((A, nat), (arctic_rat, List[arctic_rat])),
           ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
           ((A, nat), List[(List[(nat, nat)], BigInt)]),
           ((A, nat), List[(List[(nat, nat)], rat)])],
     ((A, nat),
       List[(List[(nat, nat)],
              rat)])]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                  ((A, nat), (rat, List[rat])),
                                  ((A, nat), (arctic, List[arctic])),
                                  ((A, nat),
                                    (List[List[BigInt]],
                                      List[List[List[BigInt]]])),
                                  ((A, nat),
                                    (List[List[rat]], List[List[List[rat]]])),
                                  ((A, nat),
                                    (List[List[arctic]],
                                      List[List[List[arctic]]])),
                                  ((A, nat), (arctic_rat, List[arctic_rat])),
                                  ((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]])),
                                  ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                  ((A, nat), List[(List[(nat, nat)], rat)])])
                        => alter10a[((A, nat), (BigInt, List[BigInt])),
                                     ((A, nat), (rat, List[rat])),
                                     ((A, nat), (arctic, List[arctic])),
                                     ((A, nat),
                                       (List[List[BigInt]],
 List[List[List[BigInt]]])),
                                     ((A, nat),
                                       (List[List[rat]],
 List[List[List[rat]]])),
                                     ((A, nat),
                                       (List[List[arctic]],
 List[List[List[arctic]]])),
                                     ((A, nat), (arctic_rat, List[arctic_rat])),
                                     ((A, nat),
                                       (List[List[arctic_rat]],
 List[List[List[arctic_rat]]])),
                                     ((A, nat),
                                       List[(List[(nat, nat)], BigInt)]),
                                     ((A, nat),
                                       List[(List[(nat, nat)], rat)])](aa),
                       pi)))
         case Rat_mat(d, n, sd) =>
           Rat_mat_carrier[A](d, n, sd,
                               mapa[alter[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat), List[(List[(nat, nat)], rat)])],
                                     ((A, nat),
                                       (List[List[rat]],
 List[List[List[rat]]]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                      ((A, nat), (rat, List[rat])),
                                      ((A, nat), (arctic, List[arctic])),
                                      ((A, nat),
(List[List[BigInt]], List[List[List[BigInt]]])),
                                      ((A, nat),
(List[List[rat]], List[List[List[rat]]])),
                                      ((A, nat),
(List[List[arctic]], List[List[List[arctic]]])),
                                      ((A, nat),
(arctic_rat, List[arctic_rat])),
                                      ((A, nat),
(List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                                      ((A, nat),
List[(List[(nat, nat)], BigInt)]),
                                      ((A, nat),
List[(List[(nat, nat)], rat)])])
                            => alter5a[((A, nat), (BigInt, List[BigInt])),
((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
((A, nat), (List[List[rat]], List[List[List[rat]]])),
((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
((A, nat), (arctic_rat, List[arctic_rat])),
((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
((A, nat), List[(List[(nat, nat)], BigInt)]),
((A, nat), List[(List[(nat, nat)], rat)])](aa),
                           pi))
       },
a)

def xml1to2elements[A, B,
                     C](tag: List[Char], p1: xml => sum[List[Char], A],
                         p2: xml => sum[List[Char], B], f: A => Option[B] => C,
                         x4: xml):
      sum[List[Char], C] =
  (tag, p1, p2, f, x4) match {
  case (tag, p1, p2, f, Xml(name, atts, cs, text)) =>
    {
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (less_eq_nat(one_nat, l) &&
                 (less_eq_nat(l, nat(BigInt(2))) &&
                   is_none[List[Char]](text)))))
        binda[List[Char], A,
               C](p1(nth[xml](cs, Zero_nat)),
                   (x1: A) =>
                     (if (equal_nata(l, nat(BigInt(2))))
                       binda[List[Char], B,
                              C](p2(nth[xml](cs, one_nat)),
                                  (x2: B) =>
                                    Inr[C, List[Char]]((f(x1))(Some[B](x2))))
                       else Inr[C, List[Char]]((f(x1))(None))))
        else no_transform[C](tag, Xml(name, atts, cs, text)))
    }
}

def status_precedence[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (List[((A, nat), (nat, order_tag))], Option[order_tag])] =
  (a: xml) =>
    xml_many_elements[((A, nat), (nat, order_tag)),
                       (List[((A, nat), (nat, order_tag))],
                         Option[order_tag])](List('s', 't', 'a', 't', 'u', 's',
           'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e'),
      (aa: xml) =>
        xml4elements[A, nat, nat, order_tag,
                      ((A, nat),
                        (nat, order_tag))](List('s', 't', 'a', 't', 'u', 's',
         'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', 'E', 'n', 't', 'r',
         'y'),
    xml2name, (ab: xml) => nata(List('a', 'r', 'i', 't', 'y'), ab),
    (ab: xml) =>
      nata(List('p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e'), ab),
    (ab: xml) =>
      xml_options[order_tag](List((List('l', 'e', 'x'),
                                    (ac: xml) =>
                                      xml0elements[order_tag](List('l', 'e',
                            'x'),
                       Lex, ac)),
                                   (List('m', 'u', 'l'),
                                     (ac: xml) =>
                                       xml0elements[order_tag](List('m', 'u',
                             'l'),
                        Mul, ac))),
                              ab),
    (f: A) => (ab: nat) => (p: nat) => (s: order_tag) => ((f, ab), (p, s)), aa),
      (ps: List[((A, nat), (nat, order_tag))]) =>
        (ps, {
               val s: List[order_tag] =
                 mapa[((A, nat), (nat, order_tag)),
                       order_tag](comp[(nat, order_tag), order_tag,
((A, nat),
  (nat, order_tag))]((aa: (nat, order_tag)) => snd[nat, order_tag](aa),
                      (aa: ((A, nat), (nat, order_tag))) =>
                        snd[(A, nat), (nat, order_tag)](aa)),
                                   ps);
               (if (list_all[order_tag]((aa: order_tag) =>
  equal_order_tag(Lex, aa),
 s))
                 Some[order_tag](Lex)
                 else (if (list_all[order_tag]((aa: order_tag) =>
         equal_order_tag(Mul, aa),
        s))
                        Some[order_tag](Mul) else None))
             }),
      a)

def path_order[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redpair[A]] =
  (a: xml) =>
    xml1to2elements[(List[((A, nat), (nat, order_tag))], Option[order_tag]),
                     List[((A, nat), af_entry)],
                     redpair[A]](List('p', 'a', 't', 'h', 'O', 'r', 'd', 'e',
                                       'r'),
                                  status_precedence[A](xml2name),
                                  afs[A](xml2name),
                                  (prec_tau:
                                     (List[((A, nat), (nat, order_tag))],
                                       Option[order_tag]))
                                    => (afo: Option[List[((A, nat), af_entry)]])
 => {
      val afsa: List[((A, nat), af_entry)] =
        afo match {
          case None => Nil
          case Some(af) => af
        }
      val ord: (List[((A, nat), (nat, order_tag))]) =>
                 (List[((A, nat), af_entry)]) => redpair[A]
        = snd[List[((A, nat), (nat, order_tag))], Option[order_tag]](prec_tau)
            match {
            case None =>
              (aa: List[((A, nat), (nat, order_tag))]) =>
                (b: List[((A, nat), af_entry)]) => Rpo[A](aa, b)
            case Some(Lex) =>
              (aa: List[((A, nat), (nat, order_tag))]) =>
                (b: List[((A, nat), af_entry)]) => Lpo[A](aa, b)
            case Some(Mul) =>
              (aa: List[((A, nat), (nat, order_tag))]) =>
                (b: List[((A, nat), af_entry)]) => Mpo[A](aa, b)
          };
      (ord(fst[List[((A, nat), (nat, order_tag))],
                Option[order_tag]](prec_tau)))(afsa)
    },
                                  a)

def redpair[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redpair[A]] =
  (a: xml) =>
    xml1element[redpair[A],
                 redpair[A]](List('r', 'e', 'd', 'P', 'a', 'i', 'r'),
                              (aa: xml) =>
                                xml_options[redpair[A]](List((List('i', 'n',
                            't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i',
                            'o', 'n'),
                       interpretation[A](xml2name)),
                      (List('p', 'a', 't', 'h', 'O', 'r', 'd', 'e', 'r'),
                        path_order[A](xml2name))),
                 aa),
                              id[redpair[A]], a)

def height[A](x0: (A, nat)): nat = x0 match {
  case (f, h) => h
}

def rm_lookup[A: linorder, B](k: A, m: rbt[A, B]): Option[B] =
  (lookup[A, B](m)).apply(k)

def rs_memb[A: linorder]: A => (rbt[A, Unit]) => Boolean =
  (a: A) =>
    (b: rbt[A, Unit]) =>
      s_memb[A, rbt[A, Unit],
              Unit]((aa: A) => (ba: rbt[A, Unit]) => rm_lookup[A, Unit](aa, ba),
                     a, b)

def shows_lpo_repr[A: show](prs: List[((A, nat), (nat, order_tag))]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('L', 'P', 'O', ' ', 'w', 'i', 't',
     'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
     ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_map[((A, nat),
                              (nat, order_tag))]((a:
            ((A, nat), (nat, order_tag)))
           => {
                val (aa, b): ((A, nat), (nat, order_tag)) = a;
                ({
                   val (f, n): (A, nat) = aa;
                   (ab: (nat, order_tag)) =>
                     {
                       val (pr, _): (nat, order_tag) = ab;
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('p', 'r', 'e',
                          'c', 'e', 'd', 'e', 'n', 'c', 'e', '(')),
  comp[List[Char], List[Char],
        List[Char]]((ac: List[Char]) => shows_prec[A](Zero_nat, f, ac),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('[')),
comp[List[Char], List[Char],
      List[Char]](shows_prec_nat(Zero_nat, n),
                   comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(']', ')', ' ', '=',
                      ' ')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_nat(Zero_nat, pr), shows_nl))))))
                     }
                 })(b)
              },
          prs),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'e', 'c', 'e', 'd', 'e', 'n',
 'c', 'e', '(', '_', ')', ' ', '=', ' ', '0')),
                 shows_nl)))))

def shows_mpo_repr[A: show](prs: List[((A, nat), (nat, order_tag))]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('M', 'P', 'O', ' ', 'w', 'i', 't',
     'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
     ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_map[((A, nat),
                              (nat, order_tag))]((a:
            ((A, nat), (nat, order_tag)))
           => {
                val (aa, b): ((A, nat), (nat, order_tag)) = a;
                ({
                   val (f, n): (A, nat) = aa;
                   (ab: (nat, order_tag)) =>
                     {
                       val (pr, _): (nat, order_tag) = ab;
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('p', 'r', 'e',
                          'c', 'e', 'd', 'e', 'n', 'c', 'e', '(')),
  comp[List[Char], List[Char],
        List[Char]]((ac: List[Char]) => shows_prec[A](Zero_nat, f, ac),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('[')),
comp[List[Char], List[Char],
      List[Char]](shows_prec_nat(Zero_nat, n),
                   comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(']', ')', ' ', '=',
                      ' ')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_nat(Zero_nat, pr), shows_nl))))))
                     }
                 })(b)
              },
          prs),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'e', 'c', 'e', 'd', 'e', 'n',
 'c', 'e', '(', '_', ')', ' ', '=', ' ', '0')),
                 shows_nl)))))

def shows_rpo_repr[A: show](prs: List[((A, nat), (nat, order_tag))]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('R', 'P', 'O', ' ', 'w', 'i', 't',
     'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
     ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_map[((A, nat),
                              (nat, order_tag))]((a:
            ((A, nat), (nat, order_tag)))
           => {
                val (aa, b): ((A, nat), (nat, order_tag)) = a;
                ({
                   val (f, n): (A, nat) = aa;
                   (ab: (nat, order_tag)) =>
                     {
                       val (pr, _): (nat, order_tag) = ab;
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('p', 'r', 'e',
                          'c', 'e', 'd', 'e', 'n', 'c', 'e', '(')),
  comp[List[Char], List[Char],
        List[Char]]((ac: List[Char]) => shows_prec[A](Zero_nat, f, ac),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('[')),
comp[List[Char], List[Char],
      List[Char]](shows_prec_nat(Zero_nat, n),
                   comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(']', ')', ' ', '=',
                      ' ')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_nat(Zero_nat, pr), shows_nl))))))
                     }
                 })(b)
              },
          prs),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'e', 'c', 'e', 'd', 'e', 'n',
 'c', 'e', '(', '_', ')', ' ', '=', ' ', '0')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('a', 'n', 'd', ' ', 't', 'h', 'e', ' ',
                                       'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n',
                                       'g', ' ', 's', 't', 'a', 't', 'u', 's')),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_map[((A, nat),
                        (nat, order_tag))]((a: ((A, nat), (nat, order_tag))) =>
     {
       val (aa, b): ((A, nat), (nat, order_tag)) = a;
       ({
          val (f, n): (A, nat) = aa;
          (ab: (nat, order_tag)) =>
            {
              val (_, s): (nat, order_tag) = ab;
              comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('s', 't', 'a', 't', 'u',
                 's', '(')),
                                 comp[List[Char], List[Char],
                                       List[Char]]((ac: List[Char]) =>
             shows_prec[A](Zero_nat, f, ac),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('[')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_nat(Zero_nat, n),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(']', ')', ' ', '=', ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(s match {
                            case Lex => List('l', 'e', 'x')
                            case Mul => List('m', 'u', 'l')
                          }),
        shows_nl))))))
            }
        })(b)
     },
    prs),
             comp[List[Char], List[Char],
                   List[Char]](shows_nl,
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('s',
                                   't', 'a', 't', 'u', 's', '(', '_', ')', ' ',
                                   '=', ' ', 'l', 'e', 'x')),
           shows_nl)))))))))))

def s_isEmpty[A](isEmpty: A): A = isEmpty

def wf_terms_list[A: equal, B](uu: List[(A, nat)], x1: term[A, B]): Boolean =
  (uu, x1) match {
  case (uu, Var(uv)) => true
  case (fa, Fun(f, ts)) =>
    membera[(A, nat)](fa, (f, size_list[term[A, B]](ts))) &&
      list_all[term[A, B]]((a: term[A, B]) => wf_terms_list[A, B](fa, a), ts)
}

def full_af[A](f: A, n: nat, i: nat): Boolean = true

def defs_list[A: equal, B](x0: List[(term[A, B], term[A, B])]): List[(A, nat)] =
  x0 match {
  case Nil => Nil
  case (Fun(f, ss), r) :: ts =>
    {
      val fs: List[(A, nat)] = defs_list[A, B](ts)
      val n: nat = size_list[term[A, B]](ss);
      (if (membera[(A, nat)](fs, (f, n))) fs else (f, n) :: fs)
    }
  case (Var(x), r) :: ts => defs_list[A, B](ts)
}

def shows_pos(x0: pos): (List[Char]) => List[Char] = x0 match {
  case PCons(i, Empty) => shows_prec_nat(Zero_nat, i)
  case PCons(i, PCons(v, va)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_prec_nat(Zero_nat, i), shows_pos(PCons(v, va)))
  case Empty => shows_string.apply(List('e'))
}

def apply_args[A, B](a: A, t: term[A, B], x2: List[term[A, B]]): term[A, B] =
  (a, t, x2) match {
  case (a, t, Nil) => t
  case (a, t, s :: ss) => apply_args[A, B](a, Fun[A, B](a, List(t, s)), ss)
}

def get_symbol[A](sm: A => nat => List[A], f: A, n: nat, i: nat): A =
  nth[A]((sm(f))(n), i)

def sig_list_to_sig_map[A: equal](a: A, sml: List[((A, nat), List[A])],
                                   fmap: (List[((A, nat), List[A])]) =>
   A => nat => A):
      A => nat => List[A] =
  {
    val fm: A => nat => A = fmap(sml);
    (f: A) =>
      (n: nat) =>
        lookupa[(A, nat), List[A]]((f, n), sml) match {
          case None => List((fm(f))(n))
          case Some(xs) => (if (nulla[A](xs)) List((fm(f))(n)) else xs)
        }
  }

def vars_rule_list[A, B](r: (term[A, B], term[A, B])): List[B] =
  vars_term_lista[A, B](fst[term[A, B], term[A, B]](r)) ++
    vars_term_lista[A, B](snd[term[A, B], term[A, B]](r))

def vars_rule_lista[A, B: equal](r: (term[A, B], term[A, B])): List[B] =
  remdups[B](vars_rule_list[A, B](r))

def aarity_term[A: equal, B](a: A, sm: A => nat => List[A], t: term[A, B]):
      Option[nat] =
  unapp[A, B](a, t) match {
    case (Var(_), ts) => None
    case (Fun(f, ss), ts) =>
      Some[nat](minus_nat(aarity[A](sm, f, size_list[term[A, B]](ss)),
                           size_list[term[A, B]](ts)))
  }

def eta_closed_rules[A: equal: show,
                      B: equal: show](a: A, sm: A => nat => List[A],
                                       ra: List[(term[A, B], term[A, B])],
                                       r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, B], term[A, B]),
                                 (List[Char]) =>
                                   List[Char]]((b: (term[A, B], term[A, B])) =>
         {
           val (l, raa): (term[A, B], term[A, B]) = b;
           aarity_term[A, B](a, sm, l) match {
             case None => Inr[Unit, (List[Char]) => List[Char]](())
             case Some(Zero_nat) => Inr[Unit, (List[Char]) => List[Char]](())
             case Some(Suc(_)) =>
               check[(List[Char]) =>
                       List[Char]](list_ex[(term[A, B],
     term[A, B])]((ba: (term[A, B], term[A, B])) =>
                    {
                      val (lll, rrr): (term[A, B], term[A, B]) = ba;
                      (lll, rrr) match {
                        case (Var(_), bb) => false
                        case (Fun(f, Nil), bb) => false
                        case (Fun(f, List(ll)), bb) => false
                        case (Fun(f, List(ll, Var(x))), Var(_)) => false
                        case (Fun(f, List(ll, Var(x))), Fun(g, Nil)) => false
                        case (Fun(f, List(ll, Var(x))), Fun(g, List(rr))) =>
                          false
                        case (Fun(f, List(ll, Var(x))),
                               Fun(g, List(rr, Var(y))))
                          => eq[A](f, a) &&
                               (eq[A](g, a) &&
                                 (eq[B](x, y) &&
                                   (! (membera[B](vars_rule_lista[A,
                           B]((ll, rr)),
           x)) &&
                                     instance_rule[A, B]((l, raa), (ll, rr)))))
                        case (Fun(f, List(ll, Var(x))),
                               Fun(g, rr :: Var(y) :: _ :: _))
                          => false
                        case (Fun(f, List(ll, Var(x))),
                               Fun(g, rr :: Fun(_, _) :: list))
                          => false
                        case (Fun(f, ll :: Var(x) :: _ :: _), bb) => false
                        case (Fun(f, ll :: Fun(_, _) :: list), bb) => false
                      }
                    },
                   r),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('e', 't', 'a', ' ', 'e', 'x', 'p',
   'a', 'n', 's', 'i', 'o', 'n', ' ', 'o', 'f', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_rule[A, B]((l, raa)),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))
           }
         },
        ra),
                         (x: ((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](x)))

def check_subseteq[A: equal](xs: List[A], ys: List[A]): sum[A, Unit] =
  catcha[(A, A), Unit,
          A](forallm[A, A]((x: A) =>
                             (if (membera[A](ys, x)) Inr[Unit, A](())
                               else Inl[A, Unit](x)),
                            xs),
              (x: (A, A)) => Inl[A, Unit](snd[A, A](x)))

def uncurry_term[A: equal, B](a: A, sm: A => nat => List[A], t: term[A, B]):
      term[A, B] =
  unapp[A, B](a, t) match {
    case (Var(x), ts) =>
      apply_args[A, B](a, Var[B, A](x),
                        mapa[term[A, B],
                              term[A, B]]((b: term[A, B]) =>
    uncurry_term[A, B](a, sm, b),
   ts))
    case (Fun(f, ss), ts) =>
      {
        val n: nat = size_list[term[A, B]](ss)
        val uss: List[term[A, B]] =
          mapa[term[A, B],
                term[A, B]]((b: term[A, B]) => uncurry_term[A, B](a, sm, b), ss)
        val uts: List[term[A, B]] =
          mapa[term[A, B],
                term[A, B]]((b: term[A, B]) => uncurry_term[A, B](a, sm, b), ts)
        val aa: nat = aarity[A](sm, f, n)
        val m: nat = min[nat].apply(size_list[term[A, B]](ts)).apply(aa)
        val fm: A = get_symbol[A](sm, f, n, m);
        apply_args[A, B](a, Fun[A, B](fm, uss ++ take[term[A, B]](m, uts)),
                          drop[term[A, B]](m, uts))
      }
  }

def uncurry_rules[A: equal, B](a: A, sm: A => nat => List[A]):
      (List[(term[A, B], term[A, B])]) => List[(term[A, B], term[A, B])] =
  (b: List[(term[A, B], term[A, B])]) =>
    mapa[(term[A, B], term[A, B]),
          (term[A, B],
            term[A, B])]((ba: (term[A, B], term[A, B])) =>
                           {
                             val (l, r): (term[A, B], term[A, B]) = ba;
                             (uncurry_term[A, B](a, sm, l),
                               uncurry_term[A, B](a, sm, r))
                           },
                          b)

def check_CS_subseteq[A: equal,
                       B: equal](r: List[(term[A, B], term[A, B])],
                                  s: List[(term[A, B], term[A, B])]):
      sum[(term[A, B], term[A, B]), Unit] =
  catcha[((term[A, B], term[A, B]), (term[A, B], term[A, B])), Unit,
          (term[A, B],
            term[A, B])](forallm[(term[A, B], term[A, B]),
                                  (term[A, B],
                                    term[A,
  B])]((a: (term[A, B], term[A, B])) =>
         {
           val (l, ra): (term[A, B], term[A, B]) = a;
           check[(term[A, B],
                   term[A, B])](list_ex[(term[A, B],
  term[A, B])]((aa: (term[A, B], term[A, B])) =>
                 instance_rule[A, B]((l, ra), aa),
                s),
                                 (l, ra))
         },
        r),
                          (x: ((term[A, B], term[A, B]),
                                (term[A, B], term[A, B])))
                            => Inl[(term[A, B], term[A, B]),
                                    Unit](snd[(term[A, B], term[A, B]),
       (term[A, B], term[A, B])](x)))

def generate_var(i: nat): List[Char] =
  'x' :: (shows_prec_nat(Zero_nat, i)).apply(Nil)

def generate_f_xs[A](f: A, n: nat): term[A, List[Char]] =
  Fun[A, List[Char]](f, mapa[nat, term[A,
List[Char]]]((i: nat) => Var[List[Char], A](generate_var(i)), upt(Zero_nat, n)))

def uncurry_of_sig_list[A](a: A, sml: List[((A, nat), List[A])],
                            sm: A => nat => List[A]):
      List[(term[A, List[Char]], term[A, List[Char]])] =
  maps[((A, nat), List[A]),
        (term[A, List[Char]],
          term[A, List[Char]])]((b: ((A, nat), List[A])) =>
                                  {
                                    val (ba, c): ((A, nat), List[A]) = b;
                                    ({
                                       val (f, n): (A, nat) = ba;
                                       (_: List[A]) =>
 {
   val g: nat => A = (aa: nat) => get_symbol[A](sm, f, n, aa);
   mapa[nat, (term[A, List[Char]],
               term[A, List[Char]])]((i: nat) =>
                                       (Fun[A,
     List[Char]](a, List(generate_f_xs[A](g(i), plus_nat(n, i)),
                          Var[List[Char], A](generate_var(plus_nat(n, i))))),
 generate_f_xs[A](g(Suc(i)), plus_nat(n, Suc(i)))),
                                      upt(Zero_nat, aarity[A](sm, f, n)))
 }
                                     })(c)
                                  },
                                 sml)

def uncurry_tt[A: equal: show](x0: (A, (List[((A, nat), List[A])],
 (List[(term[A, List[Char]], term[A, List[Char]])],
   List[(term[A, List[Char]], term[A, List[Char]])]))),
                                ra: List[(term[A, List[Char]],
   term[A, List[Char]])],
                                r: List[(term[A, List[Char]],
  term[A, List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  (x0, ra, r) match {
  case ((a, (sml, (u, e))), ra, r) =>
    {
      val sm: A => nat => List[A] =
        sig_list_to_sig_map[A](a, sml,
                                (_: List[((A, nat), List[A])]) =>
                                  (f: A) => (_: nat) => f)
      val r_eta: List[(term[A, List[Char]], term[A, List[Char]])] = e ++ ra;
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[((term[A, List[Char]], term[A, List[Char]]),
                            (List[Char]) => List[Char]),
                           Unit,
                           (List[Char]) =>
                             List[Char]](forallm[(term[A, List[Char]],
           term[A, List[Char]]),
          (List[Char]) =>
            List[Char]]((b: (term[A, List[Char]], term[A, List[Char]])) =>
                          {
                            val (l, _):
                                  (term[A, List[Char]], term[A, List[Char]])
                              = b;
                            check[(List[Char]) =>
                                    List[Char]](hvf_term[A, List[Char]](a, l),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_list[Char](Zero_nat,
          List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_term[A, List[Char]](l),
       shows_prec_list[Char](Zero_nat,
                              List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                                    'w', 'e', 'd')))))
                          },
                         r_eta),
  (x: ((term[A, List[Char]], term[A, List[Char]]), (List[Char]) => List[Char]))
    => Inl[(List[Char]) => List[Char],
            Unit](snd[(term[A, List[Char]], term[A, List[Char]]),
                       (List[Char]) => List[Char]](x))),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](eta_closed_rules[A,
             List[Char]](a, sm, r_eta, r_eta),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], Unit,
     Unit](catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                   (List[Char]) =>
                     List[Char]](check_subseteq[(term[A, List[Char]],
          term[A, List[Char]])]((uncurry_rules[A,
        List[Char]](a, sm)).apply(r_eta),
                                 r),
                                  (x: (term[A, List[Char]],
term[A, List[Char]]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](comp[List[Char], List[Char],
                List[Char]](shows_prec_list[Char](Zero_nat,
           List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l',
                 'e', ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_rule[A, List[Char]](x),
        shows_prec_list[Char](Zero_nat,
                               List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                     'n', 'g')))))),
            (_: Unit) =>
              catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                      (List[Char]) =>
                        List[Char]](check_CS_subseteq[A,
               List[Char]](uncurry_of_sig_list[A](a, sml, sm), r),
                                     (x:
(term[A, List[Char]], term[A, List[Char]]))
                                       => Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_prec_list[Char](Zero_nat,
              List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e',
                    ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_rule[A, List[Char]](x),
           shows_prec_list[Char](Zero_nat,
                                  List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
'i', 'n', 'g')))))))))
    }
}

def concat_lists[A](x0: List[List[A]]): List[List[A]] = x0 match {
  case Nil => List(Nil)
  case as :: xs =>
    {
      val a: List[List[A]] = concat_lists[A](xs);
      maps[List[A],
            List[A]]((vec: List[A]) =>
                       mapa[A, List[A]]((aa: A) => aa :: vec, as),
                      a)
    }
}

def enum_vectors[A, B](c: List[A], x1: List[B]): List[List[(B, A)]] = (c, x1)
  match {
  case (c, Nil) => List(Nil)
  case (c, x :: xs) =>
    {
      val a: List[List[(B, A)]] = enum_vectors[A, B](c, xs);
      maps[List[(B, A)],
            List[(B, A)]]((vec: List[(B, A)]) =>
                            mapa[A, List[(B, A)]]((ca: A) => (x, ca) :: vec, c),
                           a)
    }
}

def letters: List[Char] =
  List('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
        'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B',
        'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
        'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '_', '0', '1', '2',
        '3', '4', '5', '6', '7', '8', '9', '&', ';', ':')

def xml1or2many_elements[A, B, C,
                          D](tag: List[Char], p1: xml => sum[List[Char], A],
                              p2: xml => sum[List[Char], B],
                              p3: xml => sum[List[Char], C],
                              f: A => Option[B] => (List[C]) => D, x5: xml):
      sum[List[Char], D] =
  (tag, p1, p2, p3, f, x5) match {
  case (tag, p1, p2, p3, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (is_none[List[Char]](text) &&
               less_eq_nat(nat(BigInt(2)), size_list[xml](cs)))))
      binda[List[Char], A,
             D](p1(nth[xml](cs, Zero_nat)),
                 (x: A) =>
                   catcha[List[Char], D,
                           List[Char]](binda[List[Char], B,
      D](p2(nth[xml](cs, one_nat)),
          (y: B) =>
            binda[List[Char], List[C],
                   D](map[C].apply(p3).apply(tl[xml](tl[xml](cs))),
                       (xs: List[C]) =>
                         Inr[D, List[Char]](((f(x))(Some[B](y)))(xs)))),
(_: List[Char]) =>
  binda[List[Char], List[C],
         D](map[C].apply(p3).apply(tl[xml](cs)),
             (xs: List[C]) => Inr[D, List[Char]](((f(x))(None))(xs)))))
      else no_transform[D](tag, Xml(name, atts, cs, text)))
}

def arith_fun(xml: xml): sum[List[Char], arithFun] =
  xml1element[arithFun,
               arithFun](List('a', 'r', 'i', 't', 'h', 'F', 'u', 'n', 'c', 't',
                               'i', 'o', 'n'),
                          (a: xml) =>
                            xml_options[arithFun](List((List('n', 'a', 't', 'u',
                      'r', 'a', 'l'),
                 (aa: xml) =>
                   xml_change[nat, arithFun]((ab: xml) =>
       nata(List('n', 'a', 't', 'u', 'r', 'a', 'l'), ab),
      (ab: nat) => Const(ab), aa)),
                (List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'),
                  (aa: xml) =>
                    xml_change[nat, arithFun]((ab: xml) =>
        nata(List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'), ab),
       (n: nat) => Arg(minus_nat(n, one_nat)), aa)),
                (List('s', 'u', 'm'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('s', 'u', 'm'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Sum(ab), aa)),
                (List('p', 'r', 'o', 'd', 'u', 'c', 't'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('p', 'r', 'o', 'd', 'u',
               'c', 't'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Prod(ab), aa)),
                (List('m', 'i', 'n'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('m', 'i', 'n'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Min(ab), aa)),
                (List('m', 'a', 'x'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('m', 'a', 'x'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Max(ab), aa))),
           a),
                          id[arithFun], xml)

def xml0to1elements[A, B](tag: List[Char], p1: xml => sum[List[Char], A],
                           f: Option[A] => B, x3: xml):
      sum[List[Char], B] =
  (tag, p1, f, x3) match {
  case (tag, p1, f, Xml(name, atts, cs, text)) =>
    {
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (less_eq_nat(Zero_nat, l) &&
                 (less_eq_nat(l, one_nat) && is_none[List[Char]](text)))))
        (if (equal_nata(l, one_nat))
          binda[List[Char], A,
                 B](p1(nth[xml](cs, Zero_nat)),
                     (x1: A) => Inr[B, List[Char]](f(Some[A](x1))))
          else Inr[B, List[Char]](f(None)))
        else no_transform[B](tag, Xml(name, atts, cs, text)))
    }
}

def sl_inter[A, B](xml2name: xml => sum[List[Char], lab[A, B]]):
      xml =>
        sum[List[Char], sum[(sl_inter[A, B], Boolean), Option[lab[A, B]]]] =
  (a: xml) =>
    xml1element[sum[(sl_inter[A, B], Boolean), Option[lab[A, B]]],
                 sum[(sl_inter[A, B], Boolean),
                      Option[lab[A, B]]]](List('m', 'o', 'd', 'e', 'l'),
   (aa: xml) =>
     xml_options[sum[(sl_inter[A, B], Boolean),
                      Option[lab[A, B]]]](List((List('f', 'i', 'n', 'i', 't',
              'e', 'M', 'o', 'd', 'e', 'l'),
         (ab: xml) =>
           xml1or2many_elements[nat, Unit, ((lab[A, B], nat), arithFun),
                                 sum[(sl_inter[A, B], Boolean),
                                      Option[lab[A,
          B]]]](List('f', 'i', 'n', 'i', 't', 'e', 'M', 'o', 'd', 'e', 'l'),
                 (ac: xml) =>
                   xml_change[nat, nat]((ad: xml) =>
  nata(List('c', 'a', 'r', 'r', 'i', 'e', 'r', 'S', 'i', 'z', 'e'), ad),
 (n: nat) => minus_nat(n, one_nat), ac),
                 (ac: xml) =>
                   xml1element[Unit,
                                Unit](List('t', 'u', 'p', 'l', 'e', 'O', 'r',
    'd', 'e', 'r'),
                                       (ad: xml) =>
 xml0elements[Unit](List('p', 'o', 'i', 'n', 't', 'W', 'i', 's', 'e'), (), ad),
                                       id[Unit], ac),
                 (ac: xml) =>
                   xml3elements[lab[A, B], nat, arithFun,
                                 ((lab[A, B], nat),
                                   arithFun)](List('i', 'n', 't', 'e', 'r', 'p',
            'r', 'e', 't'),
       xml2name, (ad: xml) => nata(List('a', 'r', 'i', 't', 'y'), ad),
       (ad: xml) => arith_fun(ad),
       (f: lab[A, B]) => (ad: nat) => (b: arithFun) => ((f, ad), b), ac),
                 (n: nat) =>
                   (model: Option[Unit]) =>
                     (xs: List[((lab[A, B], nat), arithFun)]) =>
                       Inl[(sl_inter[A, B], Boolean),
                            Option[lab[A,
B]]]((SL_Inter[A, B](n, xs), is_none[Unit](model))),
                 ab)),
        (List('r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g'),
          (ab: xml) =>
            xml0to1elements[lab[A, B],
                             sum[(sl_inter[A, B], Boolean),
                                  Option[lab[A,
      B]]]](List('r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g'),
             xml2name,
             (ac: Option[lab[A, B]]) =>
               Inr[Option[lab[A, B]], (sl_inter[A, B], Boolean)](ac),
             ab))),
   aa),
   id[sum[(sl_inter[A, B], Boolean), Option[lab[A, B]]]], a)

def xml_term[A: show, B: show](x0: term[A, B]): xml = x0 match {
  case Var(x) =>
    Xml(List('v', 'a', 'r'), Nil, Nil,
         Some[List[Char]](shows_prec[B](Zero_nat, x, Nil)))
  case Fun(f, ts) =>
    Xml(List('f', 'u', 'n', 'a', 'p', 'p'), Nil,
         Xml(List('n', 'a', 'm', 'e'), Nil, Nil,
              Some[List[Char]](shows_prec[A](Zero_nat, f, Nil))) ::
           mapa[term[A, B],
                 xml]((t: term[A, B]) =>
                        Xml(List('a', 'r', 'g'), Nil, List(xml_term[A, B](t)),
                             None),
                       ts),
         None)
}

def less_eq_list[A: equal: order](x0: List[A], xs: List[A]): Boolean = (x0, xs)
  match {
  case (x :: xs, y :: ys) =>
    less[A](x, y) || eq[A](x, y) && less_eq_list[A](xs, ys)
  case (Nil, xs) => true
  case (x :: xs, Nil) => false
}

def less_list[A: equal: order](xs: List[A], x1: List[A]): Boolean = (xs, x1)
  match {
  case (x :: xs, y :: ys) =>
    less[A](x, y) || eq[A](x, y) && less_list[A](xs, ys)
  case (Nil, x :: xs) => true
  case (xs, Nil) => false
}

implicit def ord_list[A: equal: order]: ord[List[A]] = new ord[List[A]] {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

abstract sealed class dPDisproof[+A, +B]
final case class
  DPLoop[A, B](a: term[A, B],
                b: List[(pos, ((term[A, B], term[A, B]),
                                (Boolean, term[A, B])))],
                c: List[(B, term[A, B])], d: ctxt[A, B])
  extends dPDisproof[A, B]
final case class
  DPRuleRemoval[A, B](a: List[(term[A, B], term[A, B])],
                       b: List[(term[A, B], term[A, B])], c: dPDisproof[A, B])
  extends dPDisproof[A, B]

abstract sealed class tRSDisproof[+A, +B]
final case class
  Loop[A, B](a: term[A, B],
              b: List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
              c: List[(B, term[A, B])], d: ctxt[A, B])
  extends tRSDisproof[A, B]
final case object NotWellFormed extends tRSDisproof[Nothing, Nothing]
final case class
  NTRuleRemoval[A, B](a: List[(term[A, B], term[A, B])], b: tRSDisproof[A, B])
  extends tRSDisproof[A, B]
final case class
  NTStringReversal[A, B](a: List[(term[A, B], term[A, B])],
                          b: tRSDisproof[A, B])
  extends tRSDisproof[A, B]
final case class
  NTDPTrans[A, B](a: List[(term[A, B], term[A, B])], b: dPDisproof[A, B])
  extends tRSDisproof[A, B]

def equal_option[A: equal](x0: Option[A], x1: Option[A]): Boolean = (x0, x1)
  match {
  case (Some(a), None) => false
  case (None, Some(a)) => false
  case (Some(aa), Some(a)) => eq[A](aa, a)
  case (None, None) => true
}

def poly_var[A: equal, B](p: List[(List[(A, nat)], B)], v: A): Boolean =
  membera[A](maps[(List[(A, nat)], B),
                   A](comp[List[(A, nat)], List[A],
                            (List[(A, nat)],
                              B)]((a: List[(A, nat)]) =>
                                    mapa[(A, nat),
  A]((aa: (A, nat)) => fst[A, nat](aa), a),
                                   (a: (List[(A, nat)], B)) =>
                                     fst[List[(A, nat)], B](a)),
                       p),
              v)

abstract sealed class certificationResult
final case class Certified(a: List[Char]) extends certificationResult
final case class Unsupported(a: List[Char]) extends certificationResult
final case class Error(a: List[Char]) extends certificationResult

def rm_alpha[A: linorder, B]: (rbt[A, B]) => A => Option[B] =
  (a: rbt[A, B]) => lookup[A, B](a)

def rm_empty[A: linorder, B]: rbt[A, B] = empty[A, B]

def rs_alpha[A: linorder]: (rbt[A, Unit]) => A => Boolean =
  (a: rbt[A, Unit]) => s_alpha[rbt[A, Unit], A](rm_alpha[A, Unit], a)

def rs_empty[A: linorder]: rbt[A, Unit] =
  s_empty[rbt[A, Unit]](rm_empty[A, Unit])

def rs_union[A: linorder]: (rbt[A, Unit]) => (rbt[A, Unit]) => rbt[A, Unit] =
  s_union[(rbt[A, Unit]) => (rbt[A, Unit]) => rbt[A, Unit]](rm_add[A, Unit])

def funs_term_list[A, B](x0: term[A, B]): List[A] = x0 match {
  case Var(uu) => Nil
  case Fun(f, ts) =>
    f :: maps[term[A, B], A]((a: term[A, B]) => funs_term_list[A, B](a), ts)
}

def funs_rule_list[A, B](r: (term[A, B], term[A, B])): List[A] =
  funs_term_list[A, B](fst[term[A, B], term[A, B]](r)) ++
    funs_term_list[A, B](snd[term[A, B], term[A, B]](r))

def funs_trs_list[A: equal, B](r: List[(term[A, B], term[A, B])]): List[A] =
  remdups[A](maps[(term[A, B], term[A, B]),
                   A]((a: (term[A, B], term[A, B])) => funs_rule_list[A, B](a),
                       r))

def unlab[A, B](x0: lab[A, B]): lab[A, B] = x0 match {
  case Lab(f, l) => f
  case FunLab(f, l) => f
  case UnLab(v) => UnLab[A, B](v)
  case Sharp(v) => Sharp[A, B](v)
}

def is_labeled[A, B](x0: lab[A, B]): Boolean = x0 match {
  case UnLab(uu) => false
  case Sharp(uv) => false
  case Lab(v, va) => true
  case FunLab(v, va) => true
}

def label_decomp[A, B](x0: lab[A, B]): (lab[A, B], sum[B, List[lab[A, B]]]) = x0
  match {
  case Lab(f, l) => (f, Inl[B, List[lab[A, B]]](l))
  case FunLab(f, l) => (f, Inr[List[lab[A, B]], B](l))
}

def lge_to_lgr[A, B: equal](lge: A => nat => B => B => Boolean,
                             ls: A => nat => B => Boolean, f: A, n: nat):
      B => B => Boolean =
  {
    val lSfn: B => Boolean = (ls(f))(n)
    val lgefn: B => B => Boolean = (lge(f))(n);
    (l: B) =>
      (la: B) =>
        ! (eq[B](l, la)) &&
          (member[B](l, lSfn) && (member[B](la, lSfn) && (lgefn(l))(la)))
  }

def check_sl_only_lab_and_decr[A: equal: show, B: equal: show,
                                C: equal: show](lge:
          (lab[A, B]) =>
            nat =>
              (sum[B, List[lab[A, B]]]) => (sum[B, List[lab[A, B]]]) => Boolean,
         ls: (lab[A, B]) => nat => (sum[B, List[lab[A, B]]]) => Boolean,
         r: List[(term[lab[A, B], C], term[lab[A, B], C])]):
      (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
        sum[(List[Char]) => List[Char], Unit] =
  (xs: List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
    catcha[((term[lab[A, B], C], term[lab[A, B], C]),
             (List[Char]) => List[Char]),
            Unit,
            (List[Char]) =>
              List[Char]](forallm[(term[lab[A, B], C], term[lab[A, B], C]),
                                   (List[Char]) =>
                                     List[Char]]((a:
            (term[lab[A, B], C], term[lab[A, B], C]))
           => {
                val (ll, lr): (term[lab[A, B], C], term[lab[A, B], C]) = a
                val l: term[lab[A, B], C] =
                  map_term[lab[A, B], lab[A, B],
                            C]((aa: lab[A, B]) => unlab[A, B](aa), ll)
                val ra: term[lab[A, B], C] =
                  map_term[lab[A, B], lab[A, B],
                            C]((aa: lab[A, B]) => unlab[A, B](aa), lr)
                val n: nat = num_args[lab[A, B], C](l);
                check[(List[Char]) =>
                        List[Char]]((if (equal_terma[lab[A, B], C](l, ra))
                                      ! (is_var[lab[A, B], C](ll)) &&
(! (is_var[lab[A, B], C](lr)) &&
  (equal_lista[term[lab[A, B],
                     C]](args[lab[A, B], C](ll), args[lab[A, B], C](lr)) &&
    (is_labeled[A, B](the[lab[A, B]](root[lab[A, B], C](ll))) &&
      (is_labeled[A, B](the[lab[A, B]](root[lab[A, B], C](lr))) &&
        (((ls(the[lab[A, B]](root[lab[A, B],
                                   C](l))))(n))(snd[lab[A, B],
             sum[B, List[lab[A, B]]]](label_decomp[A,
            B](the[lab[A, B]](root[lab[A, B], C](ll))))) &&
          (((ls(the[lab[A, B]](root[lab[A, B],
                                     C](l))))(n))(snd[lab[A, B],
               sum[B, List[lab[A, B]]]](label_decomp[A,
              B](the[lab[A, B]](root[lab[A, B], C](lr))))) &&
            (lge_to_lgr[lab[A, B],
                         sum[B, List[lab[A,
  B]]]](lge, ls, the[lab[A, B]](root[lab[A, B], C](l)),
         n)).apply(snd[lab[A, B],
                        sum[B, List[lab[A,
 B]]]](label_decomp[A, B](the[lab[A, B]](root[lab[A, B],
       C](ll))))).apply(snd[lab[A, B],
                             sum[B, List[lab[A,
      B]]]](label_decomp[A, B](the[lab[A,
B]](root[lab[A, B], C](lr))))))))))) ||
equal_terma[lab[A, B], C](ll, lr) &&
  membera[(term[lab[A, B], C], term[lab[A, B], C])](r, (l, ra))
                                      else membera[(term[lab[A, B], C],
             term[lab[A, B], C])](r, (l, ra))),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List('l', 'a', 'b', 'e', 'l', 'e', 'd', ' ',
'r', 'u', 'l', 'e', ' ')),
                comp[List[Char], List[Char],
                      List[Char]](shows_rule[lab[A, B], C]((l, ra)),
                                   shows_string.apply(List(' ', 'i', 's', ' ',
                    'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'd', 'e', 'c', 'r',
                    'e', 'a', 's', 'i', 'n', 'g', ' ', 'n', 'o', 'r', ' ', 'a',
                    ' ', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'v', 'e', 'r',
                    's', 'i', 'o', 'n', ' ', 'o', 'f', ' ', 'a', 'n', ' ', 'o',
                    'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'r', 'u', 'l',
                    'e')))))
              },
          xs),
                           (x: ((term[lab[A, B], C], term[lab[A, B], C]),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[(term[lab[A, B], C],
         term[lab[A, B], C]),
        (List[Char]) => List[Char]](x)))

def check_sem_lab_tt[A: equal: show, B: equal: show,
                      C: equal: show](valid:
sum[(List[Char]) => List[Char], Unit],
                                       check_decr:
 (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
   sum[(List[Char]) => List[Char], Unit],
                                       check_lab:
 (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
   (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
     sum[(List[Char]) => List[Char], Unit],
                                       check_model_lab:
 (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
   (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
     sum[(List[Char]) => List[Char], Unit],
                                       lge:
 (lab[A, B]) =>
   nat => (sum[B, List[lab[A, B]]]) => (sum[B, List[lab[A, B]]]) => Boolean,
                                       ls:
 (lab[A, B]) => nat => (sum[B, List[lab[A, B]]]) => Boolean,
                                       r:
 List[(term[lab[A, B], C], term[lab[A, B], C])],
                                       lR:
 List[(term[lab[A, B], C], term[lab[A, B], C])]):
      sum[(List[Char]) => List[Char], Unit] =
  binda[(List[Char]) => List[Char], Unit,
         Unit](valid,
                (_: Unit) =>
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](binda[(List[Char]) => List[Char], Unit,
       Unit](check_decr(lR),
              (_: Unit) =>
                binda[(List[Char]) => List[Char], Unit,
                       Unit]((check_model_lab(lR))(r),
                              (_: Unit) =>
                                (check_sl_only_lab_and_decr[A, B,
                     C](lge, ls, r)).apply(lR))),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
                 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'l', 'a', 'b', 'e',
                 'l', 'e', 'd', ' ', 'T', 'R', 'S', ':')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_nl, x)))))

def model_check_valid[A]: sum[A, Unit] = Inr[Unit, A](())

def model_check_decr[A, B]: A => sum[B, Unit] = (_: A) => Inr[Unit, B](())

def eval_lab[A, B, C, D,
              E](i: A => (List[B]) => B, l: A => (List[B]) => C,
                  lc: A => nat => C => D, alpha: E => B, x4: term[A, E]):
      (B, term[D, E]) =
  (i, l, lc, alpha, x4) match {
  case (i, l, lc, alpha, Var(x)) => (alpha(x), Var[E, D](x))
  case (i, l, lc, alpha, Fun(f, ts)) =>
    {
      val clts: List[(B, term[D, E])] =
        mapa[term[A, E],
              (B, term[D, E])]((a: term[A, E]) =>
                                 eval_lab[A, B, C, D, E](i, l, lc, alpha, a),
                                ts)
      val cs: List[B] =
        mapa[(B, term[D, E]),
              B]((a: (B, term[D, E])) => fst[B, term[D, E]](a), clts)
      val c: B = (i(f))(cs)
      val lts: List[term[D, E]] =
        mapa[(B, term[D, E]),
              term[D, E]]((a: (B, term[D, E])) => snd[B, term[D, E]](a), clts);
      (c, Fun[D, E](((lc(f))(size_list[term[A, E]](ts)))((l(f))(cs)), lts))
    }
}

def check_sl_rule_ass[A: show, B: show, C, D: equal: show,
                       E: equal: show](mc: Boolean, i: A => (List[B]) => B,
la: A => (List[B]) => C, lc: A => nat => C => D, cge: B => B => Boolean,
lR: List[(term[D, E], term[D, E])], alpha: E => B,
x7: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  (mc, i, la, lc, cge, lR, alpha, x7) match {
  case (mc, i, la, lc, cge, lR, alpha, (l, r)) =>
    {
      val cl_ll: (B, term[D, E]) = eval_lab[A, B, C, D, E](i, la, lc, alpha, l)
      val cr_lr: (B, term[D, E]) = eval_lab[A, B, C, D, E](i, la, lc, alpha, r);
      binda[(List[Char]) => List[Char], Unit,
             Unit](check[(List[Char]) =>
                           List[Char]](implies(mc,
        (cge(fst[B, term[D, E]](cl_ll)))(fst[B, term[D, E]](cr_lr))),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('r', 'u', 'l', 'e', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_rule[A, E]((l, r)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', 'v', 'i', 'o', 'l', 'a', 't', 'e',
 's', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'd', 'e', 'l', ' ', 'c', 'o', 'n', 'd',
 'i', 't', 'i', 'o', 'n', ',', ' ', '[', 'l', 'h', 's', ']', ' ', '=', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]]((a: List[Char]) =>
                                     shows_prec[B](Zero_nat,
            fst[B, term[D, E]](cl_ll), a),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(',', ' ', '[', 'r', 'h', 's', ']', ' ',
                                       '=', ' ')),
               (a: List[Char]) =>
                 shows_prec[B](Zero_nat, fst[B, term[D, E]](cr_lr), a))))))),
                    (_: Unit) =>
                      check[(List[Char]) =>
                              List[Char]](membera[(term[D, E],
            term[D, E])](lR, (snd[B, term[D, E]](cl_ll),
                               snd[B, term[D, E]](cr_lr))),
   comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('l', 'a', 'b', 'e', 'l', 'e', 'd',
      ' ', 'r', 'u', 'l', 'e', ' ')),
                      comp[List[Char], List[Char],
                            List[Char]](shows_rule[D,
            E]((snd[B, term[D, E]](cl_ll), snd[B, term[D, E]](cr_lr))),
 shows_string.apply(List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g'))))))
    }
}

def assign_from_vector[A: equal, B](vec: List[(A, B)], x: A): B =
  the[B](lookupa[A, B](x, vec))

def check_sl_rule[A: show, B: show, C, D: equal: show,
                   E: equal: show](i: A => (List[B]) => B,
                                    la: A => (List[B]) => C,
                                    lc: A => nat => C => D, c: List[B],
                                    cge: B => B => Boolean, mc: Boolean,
                                    lR: List[(term[D, E], term[D, E])],
                                    x7: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  (i, la, lc, c, cge, mc, lR, x7) match {
  case (i, la, lc, c, cge, mc, lR, (l, r)) =>
    catcha[(E => B, (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[E => B,
                                   (List[Char]) =>
                                     List[Char]]((alpha: E => B) =>
           check_sl_rule_ass[A, B, C, D,
                              E](mc, i, la, lc, cge, lR, alpha, (l, r)),
          mapa[List[(E, B)],
                E => B]((a: List[(E, B)]) =>
                          (b: E) => assign_from_vector[E, B](a, b),
                         enum_vectors[B, E](c,
     list_union[E](vars_term_list[A, E].apply(l),
                    vars_term_list[A, E].apply(r))))),
                           (x: (E => B, (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[E => B,
     (List[Char]) => List[Char]](x)))
}

def labela[A, B](f: lab[A, B], n: nat, x2: sum[B, List[lab[A, B]]]): lab[A, B] =
  (f, n, x2) match {
  case (f, n, Inl(l)) => Lab[A, B](f, l)
  case (f, n, Inr(l)) => FunLab[A, B](f, l)
}

def check_sl_lab_trs[A: equal: show, B: equal: show, C: show,
                      D: equal: show](i: (lab[A, B]) => (List[C]) => C,
                                       l:
 (lab[A, B]) => (List[C]) => sum[B, List[lab[A, B]]],
                                       c: List[C], cge: C => C => Boolean,
                                       lP:
 List[(term[lab[A, B], D], term[lab[A, B], D])],
                                       p:
 List[(term[lab[A, B], D], term[lab[A, B], D])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[lab[A, B], D], term[lab[A, B], D]), (List[Char]) => List[Char]),
          Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[lab[A, B], D], term[lab[A, B], D]),
                                 (List[Char]) =>
                                   List[Char]]((a:
          (term[lab[A, B], D], term[lab[A, B], D]))
         => check_sl_rule[lab[A, B], C, sum[B, List[lab[A, B]]], lab[A, B],
                           D](i, l,
                               (aa: lab[A, B]) =>
                                 (b: nat) =>
                                   (ca: sum[B, List[lab[A, B]]]) =>
                                     labela[A, B](aa, b, ca),
                               c, cge, false, lP, a),
        p),
                         (x: ((term[lab[A, B], D], term[lab[A, B], D]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[lab[A, B], D],
       term[lab[A, B], D]),
      (List[Char]) => List[Char]](x)))

def rl_inter_list[A: equal, B](fs: List[A], d: A, g: A, uu: B): A =
  (if (membera[A](fs, g)) g else d)

def rl_fun[A, B, C](uu: A, cs: B): sum[C, B] = Inr[B, C](cs)

def model_cge[A: equal]: A => A => Boolean = (a: A) => (b: A) => eq[A](a, b)

def check_sl_model_lab_trs[A: equal: show, B: equal: show, C: show,
                            D: equal: show](i: (lab[A, B]) => (List[C]) => C,
     l: (lab[A, B]) => (List[C]) => sum[B, List[lab[A, B]]], c: List[C],
     cge: C => C => Boolean, lR: List[(term[lab[A, B], D], term[lab[A, B], D])],
     r: List[(term[lab[A, B], D], term[lab[A, B], D])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[lab[A, B], D], term[lab[A, B], D]), (List[Char]) => List[Char]),
          Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[lab[A, B], D], term[lab[A, B], D]),
                                 (List[Char]) =>
                                   List[Char]]((a:
          (term[lab[A, B], D], term[lab[A, B], D]))
         => check_sl_rule[lab[A, B], C, sum[B, List[lab[A, B]]], lab[A, B],
                           D](i, l,
                               (aa: lab[A, B]) =>
                                 (b: nat) =>
                                   (ca: sum[B, List[lab[A, B]]]) =>
                                     labela[A, B](aa, b, ca),
                               c, cge, true, lR, a),
        r),
                         (x: ((term[lab[A, B], D], term[lab[A, B], D]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[lab[A, B], D],
       term[lab[A, B], D]),
      (List[Char]) => List[Char]](x)))

def model_lge[A, B, C: equal]: A => B => C => C => Boolean =
  (_: A) => (_: B) => (a: C) => (b: C) => eq[C](a, b)

def model_LS[A, B, C]: A => B => C => Boolean =
  (_: A) => (_: B) => top_fun[C, Boolean]

def rl_tt[A: equal: show,
           B: equal: show](r: List[(term[lab[A, B], List[Char]],
                                     term[lab[A, B], List[Char]])],
                            lR: List[(term[lab[A, B], List[Char]],
                                       term[lab[A, B], List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val fs: List[lab[A, B]] = funs_trs_list[lab[A, B], List[Char]](r);
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (nulla[lab[A, B]](fs)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('r', 'o', 'o', 't', '-', 'l', 'a', 'b',
 'e', 'l', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'a',
 't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ', 'f', 'u', 'n', 'c',
 't', 'i', 'o', 'n', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'i', 'n', ' ', 't',
 'h', 'e', ' ', 's', 'i', 'g', 'n', 'a', 't', 'u', 'r', 'e')),
                 shows_nl)),
                  (_: Unit) =>
                    {
                      val d: lab[A, B] = hd[lab[A, B]](fs);
                      check_sem_lab_tt[A, B,
List[Char]](model_check_valid[(List[Char]) => List[Char]],
             model_check_decr[List[(term[lab[A, B], List[Char]],
                                     term[lab[A, B], List[Char]])],
                               (List[Char]) => List[Char]],
             (a: List[(term[lab[A, B], List[Char]],
                        term[lab[A, B], List[Char]])])
               => (b: List[(term[lab[A, B], List[Char]],
                             term[lab[A, B], List[Char]])])
                    => check_sl_lab_trs[A, B, lab[A, B],
 List[Char]]((aa: lab[A, B]) =>
               (ba: List[lab[A, B]]) =>
                 rl_inter_list[lab[A, B], List[lab[A, B]]](fs, d, aa, ba),
              (aa: lab[A, B]) =>
                (ba: List[lab[A, B]]) =>
                  rl_fun[lab[A, B], List[lab[A, B]], B](aa, ba),
              fs, model_cge[lab[A, B]], a, b),
             (a: List[(term[lab[A, B], List[Char]],
                        term[lab[A, B], List[Char]])])
               => (b: List[(term[lab[A, B], List[Char]],
                             term[lab[A, B], List[Char]])])
                    => check_sl_model_lab_trs[A, B, lab[A, B],
       List[Char]]((aa: lab[A, B]) =>
                     (ba: List[lab[A, B]]) =>
                       rl_inter_list[lab[A, B], List[lab[A, B]]](fs, d, aa, ba),
                    (aa: lab[A, B]) =>
                      (ba: List[lab[A, B]]) =>
                        rl_fun[lab[A, B], List[lab[A, B]], B](aa, ba),
                    fs, model_cge[lab[A, B]], a, b),
             model_lge[lab[A, B], nat, sum[B, List[lab[A, B]]]],
             model_LS[lab[A, B], nat, sum[B, List[lab[A, B]]]], r, lR)
                    })
  }

def uncurry_top[A: equal,
                 B](a: A, n: nat, sm: A => nat => List[A], x3: term[A, B]):
      term[A, B] =
  (a, n, sm, x3) match {
  case (a, n, sm, Fun(f, ts)) =>
    {
      val mt: (List[term[A, B]]) => List[term[A, B]] =
        (aa: List[term[A, B]]) =>
          mapa[term[A, B],
                term[A, B]]((ab: term[A, B]) =>
                              map_term_wa[A, A,
   B]((fa: A) => (na: nat) => get_symbol[A](sm, fa, na, Zero_nat), ab),
                             aa);
      (if (eq[A](f, a) &&
             (equal_nata(size_list[term[A, B]](ts), n) &&
               (! (is_var[A, B](hd[term[A, B]](ts))) &&
                 ! (equal_nata(aarity[A](sm,
  the[A](root[A, B](hd[term[A, B]](ts))), num_args[A, B](hd[term[A, B]](ts))),
                                Zero_nat)))))
        {
          val (Fun(g, ss)): term[A, B] = hd[term[A, B]](ts);
          Fun[A, B](get_symbol[A](sm, g, size_list[term[A, B]](ss), one_nat),
                     mt(ss ++ tl[term[A, B]](ts)))
        }
        else Fun[A, B](get_symbol[A](sm, f, size_list[term[A, B]](ts),
                                      Zero_nat),
                        mt(ts)))
    }
  case (a, n, sm, Var(x)) => Var[B, A](x)
}

def xml_rules[A: show,
               B: show](tag: List[Char], rls: List[(term[A, B], term[A, B])]):
      xml =
  Xml(tag, Nil,
       List(Xml(List('r', 'u', 'l', 'e', 's'), Nil,
                 mapa[(term[A, B], term[A, B]),
                       xml]((a: (term[A, B], term[A, B])) =>
                              {
                                val (l, r): (term[A, B], term[A, B]) = a;
                                Xml(List('r', 'u', 'l', 'e'), Nil,
                                     List(Xml(List('l', 'h', 's'), Nil,
       List(xml_term[A, B](l)), None),
   Xml(List('r', 'h', 's'), Nil, List(xml_term[A, B](r)), None)),
                                     None)
                              },
                             rls),
                 None)),
       None)

def lexord_code[A: equal](lt: A => A => Boolean, uu: List[A], x2: List[A]):
      Boolean =
  (lt, uu, x2) match {
  case (lt, uu, Nil) => false
  case (lt, Nil, v :: va) => true
  case (lt, x :: xs, y :: ys) =>
    (lt(x))(y) || eq[A](x, y) && lexord_code[A](lt, xs, ys)
}

def less_lab[A: equal: ord, B: equal: ord](x0: lab[A, B], x1: lab[A, B]):
      Boolean =
  (x0, x1) match {
  case (UnLab(f), UnLab(g)) => less[A](f, g)
  case (UnLab(f), Lab(v, va)) => true
  case (UnLab(f), FunLab(v, va)) => true
  case (UnLab(f), Sharp(v)) => true
  case (Lab(f, la), Lab(g, l)) =>
    less[B](la, l) || eq[B](la, l) && less_lab[A, B](f, g)
  case (Lab(f, l), UnLab(uv)) => false
  case (Lab(f, l), FunLab(v, va)) => true
  case (Lab(f, l), Sharp(v)) => true
  case (Sharp(f), Sharp(g)) => less_lab[A, B](f, g)
  case (Sharp(f), UnLab(ux)) => false
  case (Sharp(f), Lab(uy, uz)) => false
  case (Sharp(f), FunLab(v, vb)) => true
  case (FunLab(f, fs), FunLab(g, gs)) =>
    less_lab[A, B](f, g) ||
      equal_laba[A, B](f, g) &&
        lexord_code[lab[A, B]]((a: lab[A, B]) =>
                                 (b: lab[A, B]) => less_lab[A, B](a, b),
                                fs, gs)
  case (FunLab(f, fs), Lab(v, va)) => false
  case (FunLab(f, fs), UnLab(v)) => false
  case (FunLab(f, fs), Sharp(v)) => false
}

def less_eq_lab[A: equal: ord, B: equal: ord](xs: lab[A, B], ys: lab[A, B]):
      Boolean =
  less_lab[A, B](xs, ys) || equal_laba[A, B](xs, ys)

implicit def ord_lab[A: equal: ord, B: equal: ord]: ord[lab[A, B]] = new
  ord[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

trait ordered_semiring_0[A] extends semiring_0[A] with ordered_ab_semigroup[A] {
}

trait ordered_semiring_1[A] extends semiring_1[A] with ordered_semiring_0[A] {
}

trait poly_carrier[A] extends comm_semiring_1[A] with ordered_semiring_1[A] {
}

trait max_ordered_semiring_1[A]
  extends max_ordered_monoid_add[A] with ordered_semiring_1[A] {
}

trait poly_carriera[A] extends poly_carrier[A] with max_ordered_semiring_1[A] {
}

def default_I[A: poly_carriera](defa: A, n: nat): List[(List[(nat, nat)], A)] =
  (Nil, defa) ::
    mapa[nat, (List[(nat, nat)],
                A)]((i: nat) => (List((i, one_nat)), one[A]), upt(Zero_nat, n))

def poly_power[A: equal,
                B: equal: comm_semiring_1](uu: List[(List[(A, nat)], B)],
    x1: nat):
      List[(List[(A, nat)], B)] =
  (uu, x1) match {
  case (uu, Zero_nat) => one_poly[A, B]
  case (p, Suc(n)) => poly_mult[A, B](p, poly_power[A, B](p, n))
}

def monom_subst[A, B: equal,
                 C: equal: comm_semiring_1](sigma:
      A => List[(List[(B, nat)], C)],
     x1: List[(A, nat)]):
      List[(List[(B, nat)], C)] =
  (sigma, x1) match {
  case (sigma, Nil) => one_poly[B, C]
  case (sigma, (x, p) :: m) =>
    poly_mult[B, C](poly_power[B, C](sigma(x), p),
                     monom_subst[A, B, C](sigma, m))
}

def poly_subst[A, B: equal,
                C: equal: comm_semiring_1](sigma:
     A => List[(List[(B, nat)], C)],
    x1: List[(List[(A, nat)], C)]):
      List[(List[(B, nat)], C)] =
  (sigma, x1) match {
  case (sigma, Nil) => zero_poly[B, C]
  case (sigma, (m, c) :: p) =>
    poly_add[B, C](poly_mult[B, C](List((Nil, c)),
                                    monom_subst[A, B, C](sigma, m)),
                    poly_subst[A, B, C](sigma, p))
}

def eval_term[A, B: equal: poly_carriera,
               C: equal](uu: A => nat => List[(List[(nat, nat)], B)],
                          x1: term[A, C]):
      List[(List[(C, nat)], B)] =
  (uu, x1) match {
  case (uu, Var(x)) => List((List((x, one_nat)), one[B]))
  case (i, Fun(f, ts)) =>
    {
      val ps: List[List[(List[(C, nat)], B)]] =
        mapa[term[A, C],
              List[(List[(C, nat)],
                     B)]]((a: term[A, C]) => eval_term[A, B, C](i, a), ts)
      val n: nat = size_list[term[A, C]](ts);
      poly_subst[nat, C,
                  B]((ia: nat) =>
                       (if (less_nat(ia, n))
                         nth[List[(List[(C, nat)], B)]](ps, ia)
                         else zero_poly[C, B]),
                      (i(f))(n))
    }
}

def eval_monom[A, B: comm_semiring_1](alpha: A => B, x1: List[(A, nat)]): B =
  (alpha, x1) match {
  case (alpha, Nil) => one[B]
  case (alpha, (x, p) :: m) =>
    times[B](eval_monom[A, B](alpha, m), power[B].apply(alpha(x)).apply(p))
}

def eval_poly[A, B: comm_semiring_1](alpha: A => B,
                                      x1: List[(List[(A, nat)], B)]):
      B =
  (alpha, x1) match {
  case (alpha, Nil) => zero[B]
  case (alpha, mc :: p) =>
    plus[B](times[B](eval_monom[A, B](alpha, fst[List[(A, nat)], B](mc)),
                      snd[List[(A, nat)], B](mc)),
             eval_poly[A, B](alpha, p))
}

def indent(p: (List[Char]) => List[Char], s: List[Char]): List[Char] =
  maps[Char,
        Char]((c: Char) => (if (c == newline) List(newline, ' ') else List(c)),
               p(s))

def vars_trs_list[A, B: equal](trs: List[(term[A, B], term[A, B])]): List[B] =
  remdups[B](maps[(term[A, B], term[A, B]),
                   B]((a: (term[A, B], term[A, B])) => vars_rule_list[A, B](a),
                       trs))

def funas_term_lista[A, B](x0: term[A, B]): List[(A, nat)] = x0 match {
  case Var(uu) => Nil
  case Fun(f, ts) =>
    (f, size_list[term[A, B]](ts)) ::
      maps[term[A, B],
            (A, nat)]((a: term[A, B]) => funas_term_lista[A, B](a), ts)
}

def funas_rule_list[A, B](r: (term[A, B], term[A, B])): List[(A, nat)] =
  funas_term_lista[A, B](fst[term[A, B], term[A, B]](r)) ++
    funas_term_lista[A, B](snd[term[A, B], term[A, B]](r))

def funas_trs_list[A: equal, B](r: List[(term[A, B], term[A, B])]):
      List[(A, nat)] =
  remdups[(A, nat)](maps[(term[A, B], term[A, B]),
                          (A, nat)]((a: (term[A, B], term[A, B])) =>
                                      funas_rule_list[A, B](a),
                                     r))

def shows_prec_ctxt[A: show, B: show](d: nat, c: ctxt[A, B]):
      (List[Char]) => List[Char] =
  shows_ctxt[A, B](c)

def check_flat_ctxt[A: equal: show](vs: List[List[Char]],
                                     c: ctxt[A, List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  c match {
    case Hole =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_ctxt[A, List[Char]](Zero_nat, c),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                                       'a', ' ', 'f', 'l', 'a', 't', ' ', 'c',
                                       'o', 'n', 't', 'e', 'x', 't')),
               shows_nl)))
    case More(_, ss1, Hole, ss2) =>
      {
        val ss: List[term[A, List[Char]]] = ss1 ++ ss2;
        binda[(List[Char]) => List[Char], Unit,
               Unit](check[(List[Char]) =>
                             List[Char]](distinct[term[A, List[Char]]](ss),
  comp[List[Char], List[Char],
        List[Char]](shows_prec_ctxt[A, List[Char]](Zero_nat, c),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'c', 'o',
                        'n', 't', 'a', 'i', 'n', 's', ' ', 'd', 'u', 'p', 'l',
                        'i', 'c', 'a', 't', 'e', ' ', 'v', 'a', 'r', 'i', 'a',
                        'b', 'l', 'e', 's')),
shows_nl))),
                      (_: Unit) =>
                        binda[(List[Char]) => List[Char], Unit,
                               Unit](check[(List[Char]) =>
     List[Char]](list_all[term[A, List[Char]]]((a: term[A, List[Char]]) =>
         is_var[A, List[Char]](a),
        ss),
                  comp[List[Char], List[Char],
                        List[Char]](shows_prec_ctxt[A, List[Char]](Zero_nat, c),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
'f', 'l', 'a', 't', ',', ' ', 'i', '.', 'e', '.', ',', ' ', 'h', 'a', 's', ' ',
'd', 'e', 'p', 't', 'h', ' ', 'g', 'r', 'e', 'a', 't', 'e', 'r', ' ', 't', 'h',
'a', 'n', ' ', 'o', 'n', 'e')),
                shows_nl))),
                                      (_: Unit) =>
check[(List[Char]) =>
        List[Char]](list_all[term[A, List[Char]]]((t: term[A, List[Char]]) =>
            ! (membera[List[Char]](vs, the_var[A, List[Char]](t))),
           ss1 ++ ss2),
                     comp[List[Char], List[Char],
                           List[Char]](shows_prec_ctxt[A,
                List[Char]](Zero_nat, c),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ', 'h', 'a', 's', ' ', 't', 'o',
   ' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', ' ', 'o', 'n', 'l', 'y', ' ', 'f',
   'r', 'e', 's', 'h', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 's')),
                   shows_nl)))))
      }
    case More(_, ss1, More(_, _, _, _), ss2) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_ctxt[A, List[Char]](Zero_nat, c),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                                       'a', ' ', 'f', 'l', 'a', 't', ' ', 'c',
                                       'o', 'n', 't', 'e', 'x', 't')),
               shows_nl)))
  }

def is_flat_ctxt_list[A: equal,
                       B: equal](vs: List[A], fas: List[(B, nat)],
                                  x2: ctxt[B, A]):
      Boolean =
  (vs, fas, x2) match {
  case (vs, fas, More(f, ss1, Hole, ss2)) =>
    {
      val ss: List[term[B, A]] = ss1 ++ ss2;
      membera[(B, nat)](fas, (f, Suc(size_list[term[B, A]](ss)))) &&
        (list_all[term[B, A]]((a: term[B, A]) => is_var[B, A](a), ss) &&
          (distinct[term[B, A]](ss) &&
            nulla[A](list_inter[A](mapa[term[B, A],
 A]((a: term[B, A]) => the_var[B, A](a), ss),
                                    vs))))
    }
  case (vs, fas, Hole) => false
  case (vs, fas, More(v, va, More(vd, ve, vf, vg), vc)) => false
}

def check_is_flat_ctxt[A: equal: show](vs: List[List[Char]],
fas: List[(A, nat)], c: ctxt[A, List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](is_flat_ctxt_list[List[Char], A](vs, fas, c),
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_ctxt[A,
                  List[Char]](Zero_nat, c),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't',
     ' ', 'a', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x',
     't')),
                     shows_nl)))

def all_interval_nat(p: nat => Boolean, i: nat, j: nat): Boolean =
  less_eq_nat(j, i) || p(i) && all_interval_nat(p, Suc(i), j)

def hole_at[A: equal, B](n: nat, i: nat, f: A, x3: ctxt[A, B]): Boolean =
  (n, i, f, x3) match {
  case (n, i, f, More(g, ss1, Hole, ss2)) =>
    eq[A](g, f) &&
      (equal_nata(size_list[term[A, B]](ss1), i) &&
        equal_nata(size_list[term[A, B]](ss2),
                    minus_nat(minus_nat(n, i), one_nat)))
  case (n, i, f, Hole) => false
  case (n, i, f, More(v, va, More(vd, ve, vf, vg), vc)) => false
}

def check_flat_ctxt_complete[A: equal: show](fcs: List[ctxt[A, List[Char]]],
      fa: (A, nat)):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](all_interval_nat((i: nat) =>
 list_ex[ctxt[A, List[Char]]]((a: ctxt[A, List[Char]]) =>
                                hole_at[A,
 List[Char]](snd[A, nat](fa), i, fst[A, nat](fa), a),
                               fcs),
Zero_nat, snd[A, nat](fa)),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'l', 'i', 's', 't', ' ', 'o', 'f', ' ', 'f', 'l',
                          'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', 's',
                          ' ', 'i', 's', ' ', 'i', 'n', 'c', 'o', 'm', 'p', 'l',
                          'e', 't', 'e')),
  shows_nl))

def check_rule_preserving[A: equal: show](fcs: List[ctxt[A, List[Char]]],
   rs: List[(term[A, List[Char]], term[A, List[Char]])],
   rule: (term[A, List[Char]], term[A, List[Char]])):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](membera[(term[A, List[Char]],
                                term[A, List[Char]])](rs, rule) ||
                        list_all[ctxt[A, List[Char]]]((c: ctxt[A, List[Char]])
                => membera[(term[A, List[Char]],
                             term[A, List[Char]])](rs,
            (ctxt_apply[A, List[Char]](c,
fst[term[A, List[Char]], term[A, List[Char]]](rule)),
              ctxt_apply[A, List[Char]](c,
 snd[term[A, List[Char]], term[A, List[Char]]](rule)))),
               fcs),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'u', 'l', 'e', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_rule[A, List[Char]](rule),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'i', 's',
                        ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
                        'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ',
                        't', 'h', 'e', ' ', 'r', 'e', 's', 'u', 'l', 't', 'i',
                        'n', 'g', ' ', 's', 'e', 't', ' ', 'o', 'f', ' ', 'r',
                        'u', 'l', 'e', 's', ' ', 'n', 'o', 'r', ' ', 'c', 'l',
                        'o', 's', 'e', 'd', ' ', 'u', 'n', 'd', 'e', 'r', ' ',
                        'a', 'l', 'l', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o',
                        'n', 't', 'e', 'x', 't', 's')),
shows_nl))))

def check_rule_reflecting[A: equal: show](fcs: List[ctxt[A, List[Char]]],
   rs: List[(term[A, List[Char]], term[A, List[Char]])],
   rule: (term[A, List[Char]], term[A, List[Char]])):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](list_ex[(term[A, List[Char]],
                                term[A, List[Char]])]((a:
                 (term[A, List[Char]], term[A, List[Char]]))
                => {
                     val (l, r): (term[A, List[Char]], term[A, List[Char]]) = a;
                     list_ex[ctxt[A, List[Char]]]((c: ctxt[A, List[Char]]) =>
            equal_terma[A, List[Char]](fst[term[A, List[Char]],
    term[A, List[Char]]](rule),
ctxt_apply[A, List[Char]](c, l)) &&
              equal_terma[A, List[Char]](snd[term[A, List[Char]],
      term[A, List[Char]]](rule),
  ctxt_apply[A, List[Char]](c, r)),
           Hole :: fcs)
                   },
               rs),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'u', 'l', 'e', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_rule[A, List[Char]](rule),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'i', 's',
                        ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
                        'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ',
                        't', 'h', 'e', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a',
                        'l', ' ', 's', 'e', 't', ' ', 'o', 'f', ' ', 'r', 'u',
                        'l', 'e', 's', ' ', 'n', 'o', 'r', ' ', 'o', 'b', 't',
                        'a', 'i', 'n', 'e', 'd', ' ', 'b', 'y', ' ', 'a', 'p',
                        'p', 'l', 'y', 'i', 'n', 'g', ' ', 'a', ' ', 'f', 'l',
                        'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't')),
shows_nl))))

def fcc_tt[A: equal: show](r: List[(term[A, List[Char]], term[A, List[Char]])],
                            fcs: List[ctxt[A, List[Char]]],
                            lR: List[(term[A, List[Char]],
                                       term[A, List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val vs: List[List[Char]] = vars_trs_list[A, List[Char]](r)
    val fas: List[(A, nat)] = funas_trs_list[A, List[Char]](r);
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (nulla[ctxt[A, List[Char]]](fcs)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', 'a', 't', ' ', 'l', 'e', 'a', 's',
 't', ' ', 'o', 'n', 'e', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e',
 'x', 't', ' ', 'i', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd', ' ', 'f',
 'o', 'r', ' ', 't', 'h', 'e', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'u',
 'n', 'd', 'e', 'r', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x',
 't', 's')),
                 shows_nl)),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](catcha[(ctxt[A, List[Char]],
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[ctxt[A, List[Char]],
                        (List[Char]) =>
                          List[Char]]((a: ctxt[A, List[Char]]) =>
check_flat_ctxt[A](vs, a),
                                       fcs),
                (x: (ctxt[A, List[Char]], (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[ctxt[A, List[Char]],
                                  (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    binda[(List[Char]) => List[Char], Unit,
   Unit](catcha[(ctxt[A, List[Char]], (List[Char]) => List[Char]), Unit,
                 (List[Char]) =>
                   List[Char]](forallm[ctxt[A, List[Char]],
(List[Char]) =>
  List[Char]]((a: ctxt[A, List[Char]]) => check_is_flat_ctxt[A](vs, fas, a),
               fcs),
                                (x: (ctxt[A, List[Char]],
                                      (List[Char]) => List[Char]))
                                  => Inl[(List[Char]) => List[Char],
  Unit](snd[ctxt[A, List[Char]], (List[Char]) => List[Char]](x))),
          (_: Unit) =>
            binda[(List[Char]) => List[Char], Unit,
                   Unit](catcha[((A, nat), (List[Char]) => List[Char]), Unit,
                                 (List[Char]) =>
                                   List[Char]](forallm[(A, nat),
                (List[Char]) =>
                  List[Char]]((a: (A, nat)) =>
                                check_flat_ctxt_complete[A](fcs, a),
                               fas),
        (x: ((A, nat), (List[Char]) => List[Char])) =>
          Inl[(List[Char]) => List[Char],
               Unit](snd[(A, nat), (List[Char]) => List[Char]](x))),
                          (_: Unit) =>
                            binda[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[((term[A, List[Char]],
           term[A, List[Char]]),
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[(term[A, List[Char]], term[A, List[Char]]),
                                (List[Char]) =>
                                  List[Char]]((a:
         (term[A, List[Char]], term[A, List[Char]]))
        => check_rule_preserving[A](fcs, lR, a),
       r),
                        (x: ((term[A, List[Char]], term[A, List[Char]]),
                              (List[Char]) => List[Char]))
                          => Inl[(List[Char]) => List[Char],
                                  Unit](snd[(term[A, List[Char]],
      term[A, List[Char]]),
     (List[Char]) => List[Char]](x))),
  (_: Unit) =>
    catcha[((term[A, List[Char]], term[A, List[Char]]),
             (List[Char]) => List[Char]),
            Unit,
            (List[Char]) =>
              List[Char]](forallm[(term[A, List[Char]], term[A, List[Char]]),
                                   (List[Char]) =>
                                     List[Char]]((a:
            (term[A, List[Char]], term[A, List[Char]]))
           => check_rule_reflecting[A](fcs, r, a),
          lR),
                           (x: ((term[A, List[Char]], term[A, List[Char]]),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[(term[A, List[Char]],
         term[A, List[Char]]),
        (List[Char]) => List[Char]](x))))))))
  }

def check_supt[A: equal: show, B: equal: show](s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](member[(term[A, B], term[A, B])]((s, t), supt[A, B]),
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_term[A, B](Zero_nat, t),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't',
     ' ', 'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r', ' ', 's', 'u', 'b', 't', 'e',
     'r', 'm', ' ', 'o', 'f', ' ')),
                     shows_prec_term[A, B](Zero_nat, s))))

def eq_rule_mod_vars[A: equal,
                      B: equal](lr: (term[A, B], term[A, B]),
                                 st: (term[A, B], term[A, B])):
      Boolean =
  instance_rule[A, B](lr, st) && instance_rule[A, B](st, lr)

def shows_prec_pos(d: nat, p: pos): (List[Char]) => List[Char] = shows_pos(p)

def ctxt_of_pos_term[A, B](x0: pos, t: term[A, B]): ctxt[A, B] = (x0, t) match {
  case (Empty, t) => Hole
  case (PCons(i, ps), Fun(f, ts)) =>
    More[A, B](f, take[term[A, B]](i, ts),
                ctxt_of_pos_term[A, B](ps, nth[term[A, B]](ts, i)),
                drop[term[A, B]](Suc(i), ts))
}

def check_rstep[A: equal: show,
                 B: equal: show](ra: List[(term[A, B], term[A, B])], p: pos,
                                  r: (term[A, B], term[A, B]), s: term[A, B],
                                  t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val sps: List[pos] = poss_list[A, B](s)
    val tps: List[pos] = poss_list[A, B](t);
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](list_ex[(term[A, B],
       term[A, B])]((a: (term[A, B], term[A, B])) =>
                      eq_rule_mod_vars[A, B](r, a),
                     ra),
                                      comp[List[Char], List[Char],
    List[Char]](shows_rule[A, B](r),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                    'n', 'o', 't', ' ', 'a', ' ', 'r', 'u', 'l', 'e', ' ', 'o',
                    'f')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl,
               comp[List[Char], List[Char],
                     List[Char]](shows_trs[A, B](ra), shows_nl))))),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](membera[pos](sps, p),
              comp[List[Char], List[Char],
                    List[Char]](shows_prec_pos(Zero_nat, p),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_string.apply(List(' ',
                                    'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                                    'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ',
                                    'o', 'f', ' ')),
            comp[List[Char], List[Char],
                  List[Char]](shows_prec_term[A, B](Zero_nat, s), shows_nl)))),
                                  (_: Unit) =>
                                    binda[(List[Char]) => List[Char], Unit,
   Unit](check[(List[Char]) =>
                 List[Char]](membera[pos](tps, p),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_prec_pos(Zero_nat, p),
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o',
            't', ' ', 'a', ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ',
            'o', 'f', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_prec_term[A,
                       B](Zero_nat, t),
       shows_nl)))),
          (_: Unit) =>
            {
              val c: ctxt[A, B] = ctxt_of_pos_term[A, B](p, s)
              val d: ctxt[A, B] = ctxt_of_pos_term[A, B](p, t)
              val u: term[A, B] = subt_at[A, B](s, p)
              val v: term[A, B] = subt_at[A, B](t, p);
              match_list[A, B](List((u, fst[term[A, B], term[A, B]](r)),
                                     (v, snd[term[A, B], term[A, B]](r))))
                match {
                case None =>
                  Inl[(List[Char]) => List[Char],
                       Unit](comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List('t', 'h',
                                'e', ' ', 't', 'e', 'r', 'm', ' ')),
        comp[List[Char], List[Char],
              List[Char]](shows_prec_term[A, B](Zero_nat, t),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List(' ', 'd',
                              'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'r', 'e',
                              's', 'u', 'l', 't', ' ', 'f', 'r', 'o', 'm', ' ',
                              'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a',
                              'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n',
                              ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_nl,
                         comp[List[Char], List[Char],
                               List[Char]](shows_rule[A, B](r),
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List(' ', 'a', 't', ' ', 'p', 'o', 's',
       'i', 't', 'i', 'o', 'n', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_pos(Zero_nat, p),
  shows_nl))))))))
                case Some(_) =>
                  check[(List[Char]) =>
                          List[Char]](equal_ctxt[A, B](c, d),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 't', 'e', 'r', 'm',
  ' ')),
                  comp[List[Char], List[Char],
                        List[Char]](shows_prec_term[A, B](Zero_nat, t),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
't', ' ', 'r', 'e', 's', 'u', 'l', 't', ' ', 'f', 'r', 'o', 'm', ' ', 'a', ' ',
'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
'o', 'n', ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e', ' ')),
                comp[List[Char], List[Char],
                      List[Char]](shows_nl,
                                   comp[List[Char], List[Char],
 List[Char]](shows_rule[A, B](r),
              comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List(' ', 'a', 't', ' ', 'p',
                 'o', 's', 'i', 't', 'i', 'o', 'n', ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_prec_pos(Zero_nat, p),
            shows_nl))))))))
              }
            })))
  }

def check_rrstep[A: equal: show,
                  B: equal: show](ra: List[(term[A, B], term[A, B])],
                                   r: (term[A, B], term[A, B]), s: term[A, B],
                                   t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check_rstep[A, B](ra, Empty, r, s, t)

def check_steps[A: equal: show,
                 B: equal: show](p: List[(term[A, B], term[A, B])],
                                  r: List[(term[A, B], term[A, B])],
                                  x2: List[(pos,
     ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                  s: term[A, B], u: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (p, r, x2, s, u) match {
  case (p, r, Nil, s, u) =>
    check[(List[Char]) =>
            List[Char]](equal_terma[A, B](s, u),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('t', 'h',
                            'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r',
                            'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'r',
                            'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                            'u', 'e', 'n', 'c', 'e')),
    comp[List[Char], List[Char],
          List[Char]](shows_nl,
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_term[A, B](Zero_nat, s),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(shows_nl.apply(List('d', 'o', 'e', 's',
                    ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p',
                    'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'g',
                    'o', 'a', 'l', ' ', 't', 'e', 'r', 'm'))),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_prec_term[A, B](Zero_nat, u), shows_nl)))))))
  case (p, ra, (uu, (r, (true, t))) :: prts, s, u) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](check_rrstep[A, B](p, r, s, t),
                  (_: Unit) => check_steps[A, B](p, ra, prts, t, u))
  case (pa, ra, (p, (r, (false, t))) :: prts, s, u) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](check_rstep[A, B](ra, p, r, s, t),
                  (_: Unit) => check_steps[A, B](pa, ra, prts, t, u))
}

def only_eta_rules[A: equal: show,
                    B: equal: show](e: List[(term[A, B], term[A, B])],
                                     r_eta: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, B], term[A, B]),
                                 (List[Char]) =>
                                   List[Char]]((a: (term[A, B], term[A, B])) =>
         {
           val (l, r): (term[A, B], term[A, B]) = a;
           check[(List[Char]) =>
                   List[Char]]((l, r) match {
                                 case (Var(_), b) => false
                                 case (Fun(f, ls), Var(_)) => false
                                 case (Fun(f, ls), Fun(g, rs)) =>
                                   eq[A](f, g) &&
                                     (equal_nata(size_list[term[A, B]](ls),
          nat(BigInt(2))) &&
                                       (equal_nata(size_list[term[A, B]](rs),
            nat(BigInt(2))) &&
 (equal_terma[A, B](nth[term[A, B]](ls, one_nat),
                     nth[term[A, B]](rs, one_nat)) &&
   list_ex[(term[A, B],
             term[A, B])]((aa: (term[A, B], term[A, B])) =>
                            {
                              val (la, ra): (term[A, B], term[A, B]) = aa;
                              instance_rule[A,
     B]((hd[term[A, B]](ls), hd[term[A, B]](rs)), (la, ra))
                            },
                           r_eta))))
                               },
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_list[Char](Zero_nat,
                                 List('r', 'u', 'l', 'e', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_rule[A, B]((l, r)),
                              shows_prec_list[Char](Zero_nat,
             List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ', '(',
                   'e', 't', 'a', '-', 'e', 'x', 'p', 'a', 'n', 'd', 'e', 'd',
                   ')', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'r',
                   'u', 'l', 'e')))))
         },
        e),
                         (x: ((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](x)))

def uncurry_proc[A: equal: show](x0: (A, (List[((A, nat), List[A])],
   (List[(term[A, List[Char]], term[A, List[Char]])],
     List[(term[A, List[Char]], term[A, List[Char]])]))),
                                  fmap: (List[((A, nat), List[A])]) =>
  A => nat => A,
                                  check_inj:
                                    A => nat =>
   (List[((A, nat), List[A])]) => sum[(List[Char]) => List[Char], Unit],
                                  pa: List[(term[A, List[Char]],
     term[A, List[Char]])],
                                  ra: List[(term[A, List[Char]],
     term[A, List[Char]])],
                                  p: List[(term[A, List[Char]],
    term[A, List[Char]])],
                                  r: List[(term[A, List[Char]],
    term[A, List[Char]])],
                                  rm: Option[((A, nat)) =>
       List[(term[A, List[Char]], term[A, List[Char]])]]):
      sum[(List[Char]) => List[Char], Unit] =
  (x0, fmap, check_inj, pa, ra, p, r, rm) match {
  case ((a, (sml, (uu, e))), fmap, check_inj, pa, ra, p, r, rm) =>
    {
      val sm: A => nat => List[A] = sig_list_to_sig_map[A](a, sml, fmap)
      val r_eta: List[(term[A, List[Char]], term[A, List[Char]])] = e ++ ra
      val u: List[(term[A, List[Char]], term[A, List[Char]])] =
        (uncurry_rules[A, List[Char]](a, sm)).apply(r_eta)
      val s: List[(term[A, List[Char]], term[A, List[Char]])] =
        uncurry_of_sig_list[A](a, sml, sm);
      binda[(List[Char]) => List[Char], Unit,
             Unit](only_eta_rules[A, List[Char]](e, r_eta),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](((check_inj(a))(nat(BigInt(2))))(sml),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], Unit,
     Unit](catcha[((term[A, List[Char]], term[A, List[Char]]),
                    (List[Char]) => List[Char]),
                   Unit,
                   (List[Char]) =>
                     List[Char]](forallm[(term[A, List[Char]],
   term[A, List[Char]]),
  (List[Char]) =>
    List[Char]]((b: (term[A, List[Char]], term[A, List[Char]])) =>
                  {
                    val (l, _): (term[A, List[Char]], term[A, List[Char]]) = b;
                    check[(List[Char]) =>
                            List[Char]](hvf_term[A, List[Char]](a, l),
 comp[List[Char], List[Char],
       List[Char]](shows_prec_list[Char](Zero_nat,
  List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ',
        'i', 'n', ' ', 'l', 'h', 's', ' ')),
                    comp[List[Char], List[Char],
                          List[Char]](shows_term[A, List[Char]](l),
                                       shows_prec_list[Char](Zero_nat,
                      List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                            'e', 'd')))))
                  },
                 pa),
                                  (x: ((term[A, List[Char]],
 term[A, List[Char]]),
(List[Char]) => List[Char]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](snd[(term[A, List[Char]], term[A, List[Char]]),
               (List[Char]) => List[Char]](x))),
            (_: Unit) =>
              binda[(List[Char]) => List[Char], Unit,
                     Unit](catcha[((term[A, List[Char]], term[A, List[Char]]),
                                    (List[Char]) => List[Char]),
                                   Unit,
                                   (List[Char]) =>
                                     List[Char]](forallm[(term[A, List[Char]],
                   term[A, List[Char]]),
                  (List[Char]) =>
                    List[Char]]((b: (term[A, List[Char]], term[A, List[Char]]))
                                  => {
                                       val
 (l, _): (term[A, List[Char]], term[A, List[Char]]) = b;
                                       check[(List[Char]) =>
       List[Char]](hvf_term[A, List[Char]](a, l),
                    comp[List[Char], List[Char],
                          List[Char]](shows_prec_list[Char](Zero_nat,
                     List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b',
                           'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
                                       comp[List[Char], List[Char],
     List[Char]](shows_term[A, List[Char]](l),
                  shows_prec_list[Char](Zero_nat,
 List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd')))))
                                     },
                                 r_eta),
          (x: ((term[A, List[Char]], term[A, List[Char]]),
                (List[Char]) => List[Char]))
            => Inl[(List[Char]) => List[Char],
                    Unit](snd[(term[A, List[Char]], term[A, List[Char]]),
                               (List[Char]) => List[Char]](x))),
                            (_: Unit) =>
                              binda[(List[Char]) => List[Char], Unit,
                                     Unit](eta_closed_rules[A,
                     List[Char]](a, sm, r_eta, r_eta),
    (_: Unit) =>
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                           (List[Char]) =>
                             List[Char]](check_subseteq[(term[A, List[Char]],
                  term[A, List[Char]])]((uncurry_rules[A,
                List[Char]](a, sm)).apply(pa),
 p),
  (x: (term[A, List[Char]], term[A, List[Char]])) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'p', 'a',
                      'i', 'r', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_rule[A, List[Char]](x),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
  'i', 'n', 'g')))))),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](catcha[(term[A, List[Char]],
    term[A, List[Char]]),
   Unit,
   (List[Char]) =>
     List[Char]](check_subseteq[(term[A, List[Char]],
                                  term[A, List[Char]])](u, r),
                  (x: (term[A, List[Char]], term[A, List[Char]])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](comp[List[Char], List[Char],
                                     List[Char]](shows_prec_list[Char](Zero_nat,
                                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e',
                                      'd', ' ', 'r', 'u', 'l', 'e', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_rule[A, List[Char]](x),
                             shows_prec_list[Char](Zero_nat,
            List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], Unit,
     Unit](catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                   (List[Char]) =>
                     List[Char]](check_CS_subseteq[A, List[Char]](s, r),
                                  (x: (term[A, List[Char]],
term[A, List[Char]]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](comp[List[Char], List[Char],
                List[Char]](shows_prec_list[Char](Zero_nat,
           List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e',
                 ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_rule[A, List[Char]](x),
        shows_prec_list[Char](Zero_nat,
                               List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                     'n', 'g')))))),
            (_: Unit) =>
              catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                      (List[Char]) =>
                        List[Char]](check_CS_subseteq[A, List[Char]](r, u ++ s),
                                     (x:
(term[A, List[Char]], term[A, List[Char]]))
                                       => Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_prec_list[Char](Zero_nat,
              List('a', 'd', 'd', 'i', 't', 'i', 'o', 'n', 'a', 'l', ' ', 'r',
                    'u', 'l', 'e', ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_rule[A, List[Char]](x),
           shows_prec_list[Char](Zero_nat,
                                  List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
'a', 'l', 'l', 'o', 'w', 'e', 'd'))))))))))))))
    }
}

def pi_full[A](uu: A, uv: nat, uw: nat): Boolean = true

def update_tokens[A](f: (List[A]) => List[A], ts: List[A]):
      sum[List[Char], (List[A], List[A])] =
  Inr[(List[A], List[A]), List[Char]]((ts, f(ts)))

def remove_comments_aux_acc(a: List[Char], uu: Boolean, x2: List[Char]):
      List[Char] =
  (a, uu, x2) match {
  case (a, false, c :: cs) =>
    (if (c == '<' &&
           equal_lista[Char](take[Char](nat(BigInt(3)), cs),
                              List('!', '-', '-')))
      remove_comments_aux_acc(a, true, tl[Char](cs))
      else remove_comments_aux_acc(c :: a, false, cs))
  case (a, true, c :: cs) =>
    (if (c == '-' &&
           equal_lista[Char](take[Char](nat(BigInt(2)), cs), List('-', '>')))
      remove_comments_aux_acc(a, false, drop[Char](nat(BigInt(2)), cs))
      else remove_comments_aux_acc(a, true, cs))
  case (a, uu, Nil) => a
}

def remove_comments_aux(uu: Boolean, x1: List[Char]): List[Char] = (uu, x1)
  match {
  case (false, c :: cs) =>
    (if (c == '<' &&
           equal_lista[Char](take[Char](nat(BigInt(3)), cs),
                              List('!', '-', '-')))
      remove_comments_aux(true, tl[Char](cs))
      else c :: remove_comments_aux(false, cs))
  case (true, c :: cs) =>
    (if (c == '-' &&
           equal_lista[Char](take[Char](nat(BigInt(2)), cs), List('-', '>')))
      remove_comments_aux(false, drop[Char](nat(BigInt(2)), cs))
      else remove_comments_aux(true, cs))
  case (uu, Nil) => Nil
}

def remove_comments: (List[Char]) => List[Char] =
  (if (true)
    comp[List[Char], List[Char],
          List[Char]]((a: List[Char]) => rev[Char](a),
                       (a: List[Char]) =>
                         remove_comments_aux_acc(Nil, false, a))
    else (a: List[Char]) => remove_comments_aux(false, a))

def scan_upto(end: List[Char], x1: List[Char]):
      sum[List[Char], (List[Char], List[Char])] =
  (end, x1) match {
  case (end, Nil) =>
    Inl[List[Char],
         (List[Char],
           List[Char])](List('d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                              'n', 'd', ' ', 'e', 'n', 'd', '-', 'm', 'a', 'r',
                              'k', 'e', 'r', ' ') ++
                          (quote(shows_prec_list[Char](Zero_nat,
                end))).apply(Nil))
  case (end, t :: ts) =>
    (if (equal_lista[Char](mapa[(Char, Char),
                                 Char]((a: (Char, Char)) => snd[Char, Char](a),
zip[Char, Char](end, t :: ts)),
                            end))
      Inr[(List[Char], List[Char]),
           List[Char]]((end, drop[Char](size_list[Char](end), t :: ts)))
      else binda[List[Char], (List[Char], List[Char]),
                  (List[Char],
                    List[Char])](scan_upto(end, ts),
                                  (a: (List[Char], List[Char])) =>
                                    {
                                      val (res, tsa): (List[Char], List[Char]) =
a;
                                      Inr[(List[Char], List[Char]),
   List[Char]]((t :: res, tsa))
                                    }))
}

def parse_header(ts: List[Char]): sum[List[Char], (List[Char], List[Char])] =
  (if (equal_lista[Char](take[Char](nat(BigInt(2)), trim(ts)), List('<', '?')))
    bind[Char, List[Char],
          List[Char]]((a: List[Char]) => scan_upto(List('?', '>'), a),
                       (h: List[Char]) =>
                         (a: List[Char]) =>
                           bind[Char, List[Char],
                                 List[Char]]((aa: List[Char]) =>
       parse_header(aa),
      (hs: List[Char]) =>
        (tsa: List[Char]) =>
          Inr[(List[Char], List[Char]), List[Char]]((h ++ hs, tsa)),
      a),
                       ts)
    else bind[Char, Unit,
               List[Char]](spaces,
                            (_: Unit) =>
                              (tsa: List[Char]) =>
                                Inr[(List[Char], List[Char]),
                                     List[Char]]((Nil, tsa)),
                            ts))

def parse_name: (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) =>
    bind[Char, List[Char],
          List[Char]](manyof(letters),
                       (n: List[Char]) =>
                         (aa: List[Char]) =>
                           bind[Char, Unit,
                                 List[Char]](spaces,
      (_: Unit) =>
        (if (nulla[Char](n))
          (_: List[Char]) => Inl[List[Char], (List[Char], List[Char])](letters)
          else (ts: List[Char]) =>
                 Inr[(List[Char], List[Char]), List[Char]]((n, ts))),
      aa),
                       a)

def parse_attribute_value:
      (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) =>
    bind[Char, List[Char],
          List[Char]](oneof(List(dquote)),
                       (_: List[Char]) =>
                         (aa: List[Char]) =>
                           bind[Char, List[Char],
                                 List[Char]]((ab: List[Char]) =>
       many((c: Char) => ! (equal_lista[Char](List(c), dquote)), ab),
      (v: List[Char]) =>
        (ab: List[Char]) =>
          bind[Char, List[Char],
                List[Char]](oneof(List(dquote)),
                             (_: List[Char]) =>
                               (ts: List[Char]) =>
                                 Inr[(List[Char], List[Char]),
                                      List[Char]]((v, ts)),
                             ab),
      aa),
                       a)

def parse_attributes(x0: List[Char]):
      sum[List[Char], (List[(List[Char], List[Char])], List[Char])] =
  x0 match {
  case Nil =>
    Inr[(List[(List[Char], List[Char])], List[Char]), List[Char]]((Nil, Nil))
  case c :: s =>
    (if (membera[Char](List('/', '>'), c))
      Inr[(List[(List[Char], List[Char])], List[Char]),
           List[Char]]((Nil, c :: s))
      else bind[Char, List[Char],
                 List[(List[Char],
                        List[Char])]](parse_name,
                                       (k: List[Char]) =>
 (a: List[Char]) =>
   bind[Char, List[Char],
         List[(List[Char],
                List[Char])]](oneof(List(List('='))),
                               (_: List[Char]) =>
                                 (aa: List[Char]) =>
                                   bind[Char, List[Char],
 List[(List[Char],
        List[Char])]](parse_attribute_value,
                       (v: List[Char]) =>
                         (ab: List[Char]) =>
                           bind[Char, List[(List[Char], List[Char])],
                                 List[(List[Char],
List[Char])]]((ac: List[Char]) => parse_attributes(ac),
               (atts: List[(List[Char], List[Char])]) =>
                 (ts: List[Char]) =>
                   Inr[(List[(List[Char], List[Char])], List[Char]),
                        List[Char]](((k, v) :: atts, ts)),
               ab),
                       aa),
                               a),
                                       c :: s))
}

def parse_text:
      (List[Char]) => sum[List[Char], (Option[List[Char]], List[Char])] =
  (a: List[Char]) =>
    bind[Char, List[Char],
          Option[List[Char]]]((aa: List[Char]) =>
                                many((c: Char) =>
                                       ! (equal_lista[Char](List(c),
                     List('<'))),
                                      aa),
                               (ts: List[Char]) =>
                                 {
                                   val text: List[Char] = trim(ts);
                                   (if (nulla[Char](text))
                                     (tsa: List[Char]) =>
                                       Inr[(Option[List[Char]], List[Char]),
    List[Char]]((None, tsa))
                                     else (tsa: List[Char]) =>
    Inr[(Option[List[Char]], List[Char]),
         List[Char]]((Some[List[Char]](rev[Char](trim(rev[Char](text)))), tsa)))
                                 },
                               a)

def parse_nodes(ts: List[Char]): sum[List[Char], (List[xml], List[Char])] =
  (if (nulla[Char](ts) ||
         equal_lista[Char](take[Char](nat(BigInt(2)), ts), List('<', '/')))
    Inr[(List[xml], List[Char]), List[Char]]((Nil, ts))
    else bind[Char, List[Char],
               List[xml]](oneof(List(List('<'))),
                           (_: List[Char]) =>
                             (a: List[Char]) =>
                               bind[Char, List[Char],
                                     List[xml]](parse_name,
         (n: List[Char]) =>
           (aa: List[Char]) =>
             bind[Char, List[(List[Char], List[Char])],
                   List[xml]]((ab: List[Char]) => parse_attributes(ab),
                               (atts: List[(List[Char], List[Char])]) =>
                                 (ab: List[Char]) =>
                                   bind[Char, List[Char],
 List[xml]](oneof(List(List('/', '>'), List('>'))),
             (e: List[Char]) =>
               (tsa: List[Char]) =>
                 (if (equal_lista[Char](e, List('/', '>')))
                   bind[Char, List[xml],
                         List[xml]]((ac: List[Char]) => parse_nodes(ac),
                                     (cs: List[xml]) =>
                                       (tsb: List[Char]) =>
 Inr[(List[xml], List[Char]), List[Char]]((Xml(n, atts, Nil, None) :: cs, tsb)),
                                     tsa)
                   else bind[Char, Option[List[Char]],
                              List[xml]](parse_text,
  (t: Option[List[Char]]) =>
    (ac: List[Char]) =>
      bind[Char, List[xml],
            List[xml]]((ad: List[Char]) => parse_nodes(ad),
                        (cs: List[xml]) =>
                          (ad: List[Char]) =>
                            bind[Char, List[Char],
                                  List[xml]](oneof(List(List('<', '/') ++
                  (n ++ List('>')))),
      (_: List[Char]) =>
        (ae: List[Char]) =>
          bind[Char, List[xml],
                List[xml]]((af: List[Char]) => parse_nodes(af),
                            (ns: List[xml]) =>
                              (tsb: List[Char]) =>
                                Inr[(List[xml], List[Char]),
                                     List[Char]]((Xml(n, atts, cs, t) :: ns,
           tsb)),
                            ae),
      ad),
                        ac),
  tsa)),
             ab),
                               aa),
         a),
                           ts))

def parse_node: (List[Char]) => sum[List[Char], (xml, List[Char])] =
  (a: List[Char]) =>
    bind[Char, List[Char],
          xml](oneof(List(List('<'))),
                (_: List[Char]) =>
                  (aa: List[Char]) =>
                    bind[Char, List[Char],
                          xml](parse_name,
                                (n: List[Char]) =>
                                  (ab: List[Char]) =>
                                    bind[Char, List[(List[Char], List[Char])],
  xml]((ac: List[Char]) => parse_attributes(ac),
        (atts: List[(List[Char], List[Char])]) =>
          (ac: List[Char]) =>
            bind[Char, List[Char],
                  xml](oneof(List(List('/', '>'), List('>'))),
                        (e: List[Char]) =>
                          (if (equal_lista[Char](e, List('/', '>')))
                            (ts: List[Char]) =>
                              Inr[(xml, List[Char]),
                                   List[Char]]((Xml(n, atts, Nil, None), ts))
                            else (ad: List[Char]) =>
                                   bind[Char, Option[List[Char]],
 xml](parse_text,
       (t: Option[List[Char]]) =>
         (ae: List[Char]) =>
           bind[Char, List[xml],
                 xml]((af: List[Char]) => parse_nodes(af),
                       (cs: List[xml]) =>
                         (af: List[Char]) =>
                           bind[Char, List[Char],
                                 xml](oneof(List(List('<', '/') ++
           (n ++ List('>')))),
                                       (_: List[Char]) =>
 (ts: List[Char]) =>
   Inr[(xml, List[Char]), List[Char]]((Xml(n, atts, cs, t), ts)),
                                       af),
                       ae),
       ad)),
                        ac),
        ab),
                                aa),
                a)

def parse_doc: (List[Char]) => sum[List[Char], (xmldoc, List[Char])] =
  debug[(List[Char]) =>
          sum[List[Char],
               (xmldoc,
                 List[Char])]](List('0'),
                                List('p', 'a', 'r', 's', 'i', 'n', 'g', ' ',
                                      't', 'x', 't', ' ', 't', 'o', ' ', 'x',
                                      'm', 'l'),
                                (a: List[Char]) =>
                                  bind[Char, List[Char],
xmldoc]((aa: List[Char]) => update_tokens[Char](remove_comments, aa),
         (_: List[Char]) =>
           (aa: List[Char]) =>
             bind[Char, List[Char],
                   xmldoc]((ab: List[Char]) => parse_header(ab),
                            (h: List[Char]) =>
                              (ab: List[Char]) =>
                                bind[Char, xml,
                                      xmldoc](parse_node,
       (xml: xml) =>
         (ac: List[Char]) =>
           bind[Char, Unit,
                 xmldoc]((ad: List[Char]) => eoi[Char](ad),
                          (_: Unit) =>
                            (ts: List[Char]) =>
                              Inr[(xmldoc, List[Char]),
                                   List[Char]]((Xmldoc(h, xml), ts)),
                          ac),
       ab),
                            aa),
         a))

abstract sealed class filtered[+A]
final case class FPair[A](a: A, b: nat) extends filtered[A]

def plain_name(x: xml): sum[List[Char], lab[List[Char], List[nat]]] =
  xml_options[lab[List[Char],
                   List[nat]]](List((List('n', 'a', 'm', 'e'),
                                      (a: xml) =>
xml_change[List[Char],
            lab[List[Char],
                 List[nat]]]((aa: xml) => text(List('n', 'a', 'm', 'e'), aa),
                              (aa: List[Char]) =>
                                UnLab[List[Char], List[nat]](aa),
                              a)),
                                     (List('s', 'h', 'a', 'r', 'p'),
                                       (a: xml) =>
 xml1element[lab[List[Char], List[nat]],
              lab[List[Char],
                   List[nat]]](List('s', 'h', 'a', 'r', 'p'),
                                (aa: xml) => plain_name(aa),
                                (aa: lab[List[Char], List[nat]]) =>
                                  Sharp[List[Char], List[nat]](aa),
                                a)),
                                     (List('l', 'a', 'b', 'e', 'l', 'e', 'd',
    'S', 'y', 'm', 'b', 'o', 'l'),
                                       (a: xml) =>
 xml2elements[lab[List[Char], List[nat]],
               sum[List[nat], List[lab[List[Char], List[nat]]]],
               lab[List[Char],
                    List[nat]]](List('l', 'a', 'b', 'e', 'l', 'e', 'd', 'S',
                                      'y', 'm', 'b', 'o', 'l'),
                                 (aa: xml) => plain_name(aa),
                                 label[lab[List[Char],
    List[nat]]]((aa: xml) => plain_name(aa)),
                                 (f: lab[List[Char], List[nat]]) =>
                                   (aa: sum[List[nat],
     List[lab[List[Char], List[nat]]]])
                                     => aa match {
  case Inl(ab) => Lab[List[Char], List[nat]](f, ab)
  case Inr(ab) => FunLab[List[Char], List[nat]](f, ab)
},
                                 a))),
                                x)

def xml2many_elements[A, B, C,
                       D](tag: List[Char], p1: xml => sum[List[Char], A],
                           p2: xml => sum[List[Char], B],
                           p3: xml => sum[List[Char], C],
                           f: A => B => (List[C]) => D, x5: xml):
      sum[List[Char], D] =
  (tag, p1, p2, p3, f, x5) match {
  case (tag, p1, p2, p3, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (is_none[List[Char]](text) &&
               less_eq_nat(nat(BigInt(2)), size_list[xml](cs)))))
      binda[List[Char], A,
             D](p1(nth[xml](cs, Zero_nat)),
                 (x: A) =>
                   binda[List[Char], B,
                          D](p2(nth[xml](cs, one_nat)),
                              (y: B) =>
                                binda[List[Char], List[C],
                                       D](map[C].apply(p3).apply(tl[xml](tl[xml](cs))),
   (xs: List[C]) => Inr[D, List[Char]](((f(x))(y))(xs)))))
      else no_transform[D](tag, Xml(name, atts, cs, text)))
}

def transition_lhs[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], sum[((A, nat), List[List[Char]]), List[Char]]] =
  (if ({
         val (Xml(_, _, cs, _)): xml = x;
         equal_nata(size_list[xml](cs), one_nat)
       })
    xml1element[List[Char],
                 sum[((A, nat), List[List[Char]]),
                      List[Char]]](List('l', 'h', 's'), state,
                                    (a: List[Char]) =>
                                      Inr[List[Char],
   ((A, nat), List[List[Char]])](a),
                                    x)
    else xml2many_elements[A, nat, List[Char],
                            sum[((A, nat), List[List[Char]]),
                                 List[Char]]](List('l', 'h', 's'), xml2name,
       (a: xml) => nata(List('h', 'e', 'i', 'g', 'h', 't'), a), state,
       (f: A) =>
         (h: nat) =>
           (qs: List[List[Char]]) =>
             Inl[((A, nat), List[List[Char]]), List[Char]](((f, h), qs)),
       x))

def transition[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])]] =
  (a: xml) =>
    xml2elements[sum[((A, nat), List[List[Char]]), List[Char]], List[Char],
                  sum[ta_rule[List[Char], (A, nat)],
                       (List[Char],
                         List[Char])]](List('t', 'r', 'a', 'n', 's', 'i', 't',
     'i', 'o', 'n'),
(aa: xml) => transition_lhs[A](xml2name, aa),
(aa: xml) =>
  xml1element[List[Char],
               List[Char]](List('r', 'h', 's'), state, id[List[Char]], aa),
(lhs: sum[((A, nat), List[List[Char]]), List[Char]]) =>
  (q: List[Char]) =>
    lhs match {
      case Inl((f, qs)) =>
        Inl[ta_rule[List[Char], (A, nat)],
             (List[Char], List[Char])](TA_rule[(A, nat), List[Char]](f, qs, q))
      case Inr(qa) =>
        Inr[(List[Char], List[Char]), ta_rule[List[Char], (A, nat)]]((qa, q))
    },
a)

implicit def ord_char: ord[Char] = new ord[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

implicit def preorder_list[A: equal: order]: preorder[List[A]] = new
  preorder[List[A]] {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

implicit def order_list[A: equal: order]: order[List[A]] = new order[List[A]] {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

def gen_list_to_map_aux[A, B,
                         C](upd: A => B => C => C, accs: C, x2: List[(A, B)]):
      C =
  (upd, accs, x2) match {
  case (upd, accs, Nil) => accs
  case (upd, accs, (k, v) :: l) =>
    gen_list_to_map_aux[A, B, C](upd, ((upd(k))(v))(accs), l)
}

def gen_list_to_map[A, B, C](empt: A, upd: B => C => A => A, l: List[(B, C)]):
      A =
  gen_list_to_map_aux[B, C, A](upd, empt, rev[(B, C)](l))

abstract sealed class relTRSDisproof[+A, +B]
final case class
  RelLoop[A, B](a: term[A, B],
                 b: List[(pos, ((term[A, B], term[A, B]),
                                 (Boolean, term[A, B])))],
                 c: List[(B, term[A, B])], d: ctxt[A, B])
  extends relTRSDisproof[A, B]
final case object RelNotWellFormed extends relTRSDisproof[Nothing, Nothing]
final case class
  NTRelRuleRemoval[A, B](a: List[(term[A, B], term[A, B])],
                          b: List[(term[A, B], term[A, B])],
                          c: relTRSDisproof[A, B])
  extends relTRSDisproof[A, B]
final case class RNotSN[A, B](a: tRSDisproof[A, B]) extends relTRSDisproof[A, B]

def shows_monom[A: show](x0: List[(A, nat)]): (List[Char]) => List[Char] = x0
  match {
  case List((x, p)) =>
    (if (equal_nata(p, one_nat))
      (a: List[Char]) => shows_prec[A](Zero_nat, x, a)
      else comp[List[Char], List[Char],
                 List[Char]]((a: List[Char]) => shows_prec[A](Zero_nat, x, a),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('^')),
         shows_prec_nat(Zero_nat, p))))
  case (x, p) :: v :: va =>
    comp[List[Char], List[Char],
          List[Char]]((if (equal_nata(p, one_nat))
                        (a: List[Char]) => shows_prec[A](Zero_nat, x, a)
                        else comp[List[Char], List[Char],
                                   List[Char]]((a: List[Char]) =>
         shows_prec[A](Zero_nat, x, a),
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List('^')),
                           shows_prec_nat(Zero_nat, p)))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('*')),
  shows_monom[A](v :: va)))
  case Nil => shows_string.apply(List('1'))
}

def shows_poly[A: show,
                B: equal: poly_carriera: show](x0: List[(List[(A, nat)], B)]):
      (List[Char]) => List[Char] =
  x0 match {
  case Nil => shows_string.apply(List('0'))
  case (m, c) :: p =>
    comp[List[Char], List[Char],
          List[Char]]((if (eq[B](c, one[B])) shows_monom[A](m)
                        else (if (nulla[(A, nat)](m))
                               (a: List[Char]) => shows_prec[B](Zero_nat, c, a)
                               else comp[List[Char], List[Char],
  List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, c, a),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('*')),
                                  shows_monom[A](m))))),
                       (if (nulla[(List[(A, nat)], B)](p))
                         shows_string.apply(Nil)
                         else comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List(' ',
                                 '+', ' ')),
         shows_poly[A, B](p))))
}

def poly_split[A: equal,
                B: zero](m: List[(A, nat)], p: List[(List[(A, nat)], B)]):
      (B, List[(List[(A, nat)], B)]) =
  extract[(List[(A, nat)],
            B)]((a: (List[(A, nat)], B)) =>
                  {
                    val (n, _): (List[(A, nat)], B) = a;
                    eq_monom[A](m, n)
                  },
                 p)
    match {
    case None => (zero[B], p)
    case Some((p1, ((_, c), p2))) => (c, p1 ++ p2)
  }

def list_to_rm[A: linorder, B]: (List[(A, B)]) => rbt[A, B] =
  (a: List[(A, B)]) =>
    gen_list_to_map[rbt[A, B], A, B](rm_empty[A, B], rm_update[A, B], a)

def rm_isEmpty[A: equal: linorder, B: equal](m: rbt[A, B]): Boolean =
  eq[rbt[A, B]](m, empty[A, B])

def rito_map_to_sorted_list[A, B,
                             C](iterate:
                                  (A => B => (List[(A, B)]) => List[(A, B)]) =>
                                    C => (List[(A, B)]) => List[(A, B)],
                                 m: C):
      List[(A, B)] =
  ((iterate((u: A) => (v: B) => (a: List[(A, B)]) => (u, v) :: a))(m))(Nil)

def itoi_reverse_iterateo[A, B, C,
                           D](it: (A => Boolean) =>
                                    (B => C => A => A) => D => A => A):
      (B => C => A => A) => D => A => A =
  it((_: A) => true)

def rm_reverse_iterateoi[A, B,
                          C](c: A => Boolean, f: B => C => A => A,
                              x2: rbta[B, C], sigma: A):
      A =
  (c, f, x2, sigma) match {
  case (c, f, Emptya, sigma) => sigma
  case (c, f, Branch(col, l, k, v, r), sigma) =>
    (if (c(sigma))
      {
        val sigmaa: A = rm_reverse_iterateoi[A, B, C](c, f, r, sigma);
        (if (c(sigmaa))
          rm_reverse_iterateoi[A, B, C](c, f, l, ((f(k))(v))(sigmaa))
          else sigmaa)
      }
      else sigma)
}

def rm_reverse_iterateoia[A, B: linorder,
                           C](c: A => Boolean, f: B => C => A => A,
                               r: rbt[B, C]):
      A => A =
  (a: A) => rm_reverse_iterateoi[A, B, C](c, f, impl_of[B, C](r), a)

def rm_reverse_iterateo[A: linorder, B, C]:
      (A => B => C => C) => (rbt[A, B]) => C => C =
  itoi_reverse_iterateo[C, A, B,
                         rbt[A, B]]((a: C => Boolean) =>
                                      (b: A => B => C => C) =>
(c: rbt[A, B]) => rm_reverse_iterateoia[C, A, B](a, b, c))

def rm_to_list[A: linorder, B]: (rbt[A, B]) => List[(A, B)] =
  (a: rbt[A, B]) =>
    rito_map_to_sorted_list[A, B,
                             rbt[A, B]](rm_reverse_iterateo[A, B, List[(A, B)]],
 a)

def gen_list_to_set_aux[A, B](ins: A => B => B, accs: B, x2: List[A]): B =
  (ins, accs, x2) match {
  case (ins, accs, Nil) => accs
  case (ins, accs, x :: l) => gen_list_to_set_aux[A, B](ins, (ins(x))(accs), l)
}

def gen_list_to_set[A, B](empt: A, ins: B => A => A): (List[B]) => A =
  (a: List[B]) => gen_list_to_set_aux[B, A](ins, empt, a)

def list_to_rs[A: linorder]: (List[A]) => rbt[A, Unit] =
  gen_list_to_set[rbt[A, Unit], A](rs_empty[A], rs_ins[A])

def equal_unita(u: Unit, v: Unit): Boolean = true

implicit def equal_unit: equal[Unit] = new equal[Unit] {
  val `Ceta.equal` = (a: Unit, b: Unit) => equal_unita(a, b)
}

def rs_isEmpty[A: equal: linorder]: (rbt[A, Unit]) => Boolean =
  s_isEmpty[(rbt[A, Unit]) =>
              Boolean]((a: rbt[A, Unit]) => rm_isEmpty[A, Unit](a))

def rito_set_to_sorted_list[A, B](iterate:
                                    (A => (List[A]) => List[A]) =>
                                      B => (List[A]) => List[A],
                                   m: B):
      List[A] =
  ((iterate((a: A) => (b: List[A]) => a :: b))(m))(Nil)

def s_reverse_iterateo[A: linorder, B,
                        C](iterate: (A => Unit => B => B) => C => B => B,
                            f: A => B => B, s: C, sigma_0: B):
      B =
  ((iterate((k: A) => (_: Unit) => f(k)))(s))(sigma_0)

def rs_reverse_iterateo[A: linorder, B]:
      (A => B => B) => (rbt[A, Unit]) => B => B =
  (a: A => B => B) =>
    (b: rbt[A, Unit]) =>
      (c: B) =>
        s_reverse_iterateo[A, B,
                            rbt[A, Unit]](rm_reverse_iterateo[A, Unit, B], a, b,
   c)

def rs_to_list[A: linorder]: (rbt[A, Unit]) => List[A] =
  (a: rbt[A, Unit]) =>
    rito_set_to_sorted_list[A, rbt[A, Unit]](rs_reverse_iterateo[A, List[A]], a)

def funas_below_term_list[A, B](t: term[A, B]): List[(A, nat)] =
  maps[term[A, B],
        (A, nat)]((a: term[A, B]) => funas_term_lista[A, B](a), args[A, B](t))

def funas_below_rule_list[A, B](r: (term[A, B], term[A, B])): List[(A, nat)] =
  funas_below_term_list[A, B](fst[term[A, B], term[A, B]](r)) ++
    funas_below_term_list[A, B](snd[term[A, B], term[A, B]](r))

def funas_below_trs_list[A: equal, B](trs: List[(term[A, B], term[A, B])]):
      List[(A, nat)] =
  remdups[(A, nat)](maps[(term[A, B], term[A, B]),
                          (A, nat)]((a: (term[A, B], term[A, B])) =>
                                      funas_below_rule_list[A, B](a),
                                     trs))

def check_sem_lab_proc[A: equal: show, B: equal: show,
                        C: equal: show](valid:
  sum[(List[Char]) => List[Char], Unit],
 check_decr:
   (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
     sum[(List[Char]) => List[Char], Unit],
 check_lab:
   (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
     (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
       sum[(List[Char]) => List[Char], Unit],
 check_model_lab:
   (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
     (List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
       sum[(List[Char]) => List[Char], Unit],
 lge: (lab[A, B]) =>
        nat =>
          (sum[B, List[lab[A, B]]]) => (sum[B, List[lab[A, B]]]) => Boolean,
 ls: (lab[A, B]) => nat => (sum[B, List[lab[A, B]]]) => Boolean,
 p: List[(term[lab[A, B], C], term[lab[A, B], C])],
 r: List[(term[lab[A, B], C], term[lab[A, B], C])],
 lP: List[(term[lab[A, B], C], term[lab[A, B], C])],
 lR: List[(term[lab[A, B], C], term[lab[A, B], C])]):
      sum[(List[Char]) => List[Char], Unit] =
  binda[(List[Char]) => List[Char], Unit,
         Unit](valid,
                (_: Unit) =>
                  binda[(List[Char]) => List[Char], Unit,
                         Unit](check_decr(lR),
                                (_: Unit) =>
                                  binda[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]]((check_lab(lP))(p),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('n', 'o', 't', ' ', 'a', 'l', 'l',
      ' ', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'D', 'P', 's', ' ', 'p', 'r',
      'e', 's', 'e', 'n', 't', ':')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          binda[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]]((check_model_lab(lR))(r),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'l', 'a',
                      'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S', ':')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))),
                        (_: Unit) =>
                          catcha[(List[Char]) => List[Char], Unit,
                                  (List[Char]) =>
                                    List[Char]]((check_sl_only_lab_and_decr[A,
                                     B, C](lge, ls, r)).apply(lR),
         (x: (List[Char]) => List[Char]) =>
           Inl[(List[Char]) => List[Char],
                Unit](comp[List[Char], List[Char],
                            List[Char]](x, shows_nl)))))))

def rl_fun_delt[A: equal, B](x0: Option[A], uu: List[A], uv: A, cs: List[A]):
      sum[B, List[A]] =
  (x0, uu, uv, cs) match {
  case (None, uu, uv, cs) => Inr[List[A], B](cs)
  case (Some(delt), fs, f, cs) =>
    Inr[List[A],
         B]((if (membera[A](delt :: fs, f)) cs
              else mapa[A, A]((_: A) => delt, cs)))
}

def rl_proc[A: equal: show,
             B: equal: show](delt_opt: Option[lab[A, B]],
                              p: List[(term[lab[A, B], List[Char]],
term[lab[A, B], List[Char]])],
                              r: List[(term[lab[A, B], List[Char]],
term[lab[A, B], List[Char]])],
                              lP: List[(term[lab[A, B], List[Char]],
 term[lab[A, B], List[Char]])],
                              lR: List[(term[lab[A, B], List[Char]],
 term[lab[A, B], List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val fs: List[lab[A, B]] =
      filter[lab[A, B]]((f: lab[A, B]) =>
                          ! (equal_option[lab[A,
       B]](Some[lab[A, B]](f), delt_opt)),
                         list_union[lab[A,
 B]](funs_trs_list[lab[A, B], List[Char]](r),
      mapa[(lab[A, B], nat),
            lab[A, B]]((a: (lab[A, B], nat)) => fst[lab[A, B], nat](a),
                        funas_below_trs_list[lab[A, B], List[Char]](p))));
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (nulla[lab[A, B]](fs)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('r', 'o', 'o', 't', '-', 'l', 'a', 'b',
 'e', 'l', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'a',
 't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ', 'f', 'u', 'n', 'c',
 't', 'i', 'o', 'n', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'i', 'n', ' ', 't',
 'h', 'e', ' ', 's', 'i', 'g', 'n', 'a', 't', 'u', 'r', 'e')),
                 shows_nl)),
                  (_: Unit) =>
                    {
                      val d: lab[A, B] = hd[lab[A, B]](fs);
                      check_sem_lab_proc[A, B,
  List[Char]](model_check_valid[(List[Char]) => List[Char]],
               model_check_decr[List[(term[lab[A, B], List[Char]],
                                       term[lab[A, B], List[Char]])],
                                 (List[Char]) => List[Char]],
               (a: List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])])
                 => (b: List[(term[lab[A, B], List[Char]],
                               term[lab[A, B], List[Char]])])
                      => check_sl_lab_trs[A, B, lab[A, B],
   List[Char]]((aa: lab[A, B]) =>
                 (ba: List[lab[A, B]]) =>
                   rl_inter_list[lab[A, B], List[lab[A, B]]](fs, d, aa, ba),
                (aa: lab[A, B]) =>
                  (ba: List[lab[A, B]]) =>
                    rl_fun_delt[lab[A, B], B](delt_opt, fs, aa, ba),
                fs, model_cge[lab[A, B]], a, b),
               (a: List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])])
                 => (b: List[(term[lab[A, B], List[Char]],
                               term[lab[A, B], List[Char]])])
                      => check_sl_model_lab_trs[A, B, lab[A, B],
         List[Char]]((aa: lab[A, B]) =>
                       (ba: List[lab[A, B]]) =>
                         rl_inter_list[lab[A, B],
List[lab[A, B]]](fs, d, aa, ba),
                      (aa: lab[A, B]) =>
                        (ba: List[lab[A, B]]) =>
                          rl_fun_delt[lab[A, B], B](delt_opt, fs, aa, ba),
                      fs, model_cge[lab[A, B]], a, b),
               model_lge[lab[A, B], nat, sum[B, List[lab[A, B]]]],
               model_LS[lab[A, B], nat, sum[B, List[lab[A, B]]]], p, r, lP, lR)
                    })
  }

def subst_apply_term[A, B, C](x0: term[A, B], sigma: B => term[A, C]):
      term[A, C] =
  (x0, sigma) match {
  case (Var(x), sigma) => sigma(x)
  case (Fun(f, ss), sigma) =>
    Fun[A, C](f, mapa[term[A, B],
                       term[A, C]]((t: term[A, B]) =>
                                     subst_apply_term[A, B, C](t, sigma),
                                    ss))
}

def rp_af[A, B](x0: (sum[(List[Char]) => List[Char], Unit],
                      (((term[A, B], term[A, B])) =>
                         sum[(List[Char]) => List[Char], Unit],
                        (((term[A, B], term[A, B])) =>
                           sum[(List[Char]) => List[Char], Unit],
                          (A => nat => nat => Boolean,
                            (sum[(List[Char]) => List[Char], Unit],
                              (List[Char]) => List[Char])))))):
      A => nat => nat => Boolean =
  x0 match {
  case (v, (s, (n, (a, (m, d))))) => a
}

abstract sealed class ta_rule_impl[+A, +B]
final case class TA_rule_impl[B, A](a: B, b: List[A], c: A, d: rbt[A, Unit])
  extends ta_rule_impl[A, B]

def rq_impl[A, B](x0: ta_rule_impl[A, B]): rbt[A, Unit] = x0 match {
  case TA_rule_impl(f, qsa, q, qs) => qs
}

def check_rsteps[A: equal: show,
                  B: equal: show](r: List[(term[A, B], term[A, B])],
                                   prts: List[(pos,
        ((term[A, B], term[A, B]), term[A, B]))],
                                   s: term[A, B], u: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check_steps[A, B](Nil, r,
                     mapa[(pos, ((term[A, B], term[A, B]), term[A, B])),
                           (pos, ((term[A, B], term[A, B]),
                                   (Boolean,
                                     term[A,
   B])))]((a: (pos, ((term[A, B], term[A, B]), term[A, B]))) =>
            {
              val (p, (ra, t)): (pos, ((term[A, B], term[A, B]), term[A, B])) =
                a;
              (p, (ra, (false, t)))
            },
           prts),
                     s, u)

def check_wf_trs[A: show, B: equal: show](r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[A, B], term[A, B]),
                                 (List[Char]) => List[Char]),
                                Unit,
                                (List[Char]) =>
                                  List[Char]](forallm[(term[A, B], term[A, B]),
               (List[Char]) =>
                 List[Char]]((rule: (term[A, B], term[A, B])) =>
                               catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](binda[(List[Char]) => List[Char], Unit,
                    Unit](check[(List[Char]) =>
                                  List[Char]](! (is_var[A,
                 B](fst[term[A, B], term[A, B]](rule))),
       shows_string.apply(List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'l',
                                'e', 'f', 't', '-', 'h', 'a', 'n', 'd', ' ',
                                's', 'i', 'd', 'e', ' ', 'i', 'n'))),
                           (_: Unit) =>
                             catcha[B, Unit,
                                     (List[Char]) =>
                                       List[Char]](check_subseteq[B](vars_term_list[A,
     B].apply(snd[term[A, B], term[A, B]](rule)),
                              vars_term_list[A,
      B].apply(fst[term[A, B], term[A, B]](rule))),
            (x: B) =>
              Inl[(List[Char]) => List[Char],
                   Unit](comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('f', 'r',
                            'e', 'e', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
                            'e', ' ')),
    comp[List[Char], List[Char],
          List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, x, a),
                       shows_string.apply(List(' ', 'i', 'n', ' ', 'r', 'i',
        'g', 'h', 't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', ' ',
        'o', 'f'))))))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](x,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 'r', 'u', 'l', 'e', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_rule[A, B](rule),
    shows_nl))))),
                              r),
       (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
         Inl[(List[Char]) => List[Char],
              Unit](snd[(term[A, B], term[A, B]),
                         (List[Char]) => List[Char]](x))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'T', 'R', 'S', ' ',
 'i', 's', ' ', 'n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm',
 'e', 'd')),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))

def bounds_type: xml => sum[List[Char], boundstype] =
  (a: xml) =>
    xml1element[boundstype,
                 boundstype](List('t', 'y', 'p', 'e'),
                              (aa: xml) =>
                                xml_options[boundstype](List((List('r', 'o',
                            'o', 'f'),
                       (ab: xml) =>
                         xml0elements[boundstype](List('r', 'o', 'o', 'f'),
           Roof, ab)),
                      (List('m', 'a', 't', 'c', 'h'),
                        (ab: xml) =>
                          xml0elements[boundstype](List('m', 'a', 't', 'c',
                 'h'),
            Match, ab))),
                 aa),
                              id[boundstype], a)

def bounds_bound: xml => sum[List[Char], nat] =
  (a: xml) => nata(List('b', 'o', 'u', 'n', 'd'), a)

def final_states: xml => sum[List[Char], List[List[Char]]] =
  (a: xml) =>
    xml_many_elements[List[Char],
                       List[List[Char]]](List('f', 'i', 'n', 'a', 'l', 'S', 't',
       'a', 't', 'e', 's'),
  state, id[List[List[Char]]], a)

def transitions[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (List[ta_rule[List[Char], (A, nat)]],
               List[(List[Char], List[Char])])] =
  (a: xml) =>
    xml_many_elements[sum[ta_rule[List[Char], (A, nat)],
                           (List[Char], List[Char])],
                       (List[ta_rule[List[Char], (A, nat)]],
                         List[(List[Char],
                                List[Char])])](List('t', 'r', 'a', 'n', 's',
             'i', 't', 'i', 'o', 'n', 's'),
        transition[A](xml2name),
        (rls: List[sum[ta_rule[List[Char], (A, nat)],
                        (List[Char], List[Char])]])
          => {
               val (rules, eps):
                     (List[sum[ta_rule[List[Char], (A, nat)],
                                (List[Char], List[Char])]],
                       List[sum[ta_rule[List[Char], (A, nat)],
                                 (List[Char], List[Char])]])
                 = partition[sum[ta_rule[List[Char], (A, nat)],
                                  (List[Char],
                                    List[Char])]]((aa:
             sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])])
            => aa match {
                 case Inl(_) => true
                 case Inr(_) => false
               },
           rls)
               val ruls: List[ta_rule[List[Char], (A, nat)]] =
                 mapa[sum[ta_rule[List[Char], (A, nat)],
                           (List[Char], List[Char])],
                       ta_rule[List[Char],
                                (A, nat)]]((aa:
      sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])])
     => {
          val (Inl(r)):
                sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])]
            = aa;
          r
        },
    rules)
               val aa: List[(List[Char], List[Char])] =
                 mapa[sum[ta_rule[List[Char], (A, nat)],
                           (List[Char], List[Char])],
                       (List[Char],
                         List[Char])]((aa:
 sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])])
=> {
     val (Inr(e)): sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])]
       = aa;
     e
   },
                                       eps);
               (ruls, aa)
             },
        a)

def bounds_tree_automaton[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], tree_automaton[List[Char], (A, nat)]] =
  (a: xml) =>
    xml2elements[List[List[Char]],
                  (List[ta_rule[List[Char], (A, nat)]],
                    List[(List[Char], List[Char])]),
                  tree_automaton[List[Char],
                                  (A, nat)]](List('t', 'r', 'e', 'e', 'A', 'u',
           't', 'o', 'm', 'a', 't', 'o', 'n'),
      final_states, transitions[A](xml2name),
      (f: List[List[Char]]) =>
        (aa: (List[ta_rule[List[Char], (A, nat)]],
               List[(List[Char], List[Char])]))
          => {
               val (ab, b):
                     (List[ta_rule[List[Char], (A, nat)]],
                       List[(List[Char], List[Char])])
                 = aa;
               Tree_Automaton[List[Char], (A, nat)](f, ab, b)
             },
      a)

def bounds_info[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], bounds_info[A, List[Char]]] =
  (a: xml) =>
    xml4elements[boundstype, nat, List[List[Char]],
                  tree_automaton[List[Char], (A, nat)],
                  bounds_info[A, List[Char]]](List('b', 'o', 'u', 'n', 'd',
            's'),
       bounds_type, bounds_bound, final_states,
       bounds_tree_automaton[A](xml2name),
       (aa: boundstype) =>
         (b: nat) =>
           (c: List[List[Char]]) =>
             (d: tree_automaton[List[Char], (A, nat)]) =>
               Bounds_Info[List[Char], A](aa, b, c, d),
       a)

def id_labeling[A, B](uu: lab[A, List[B]], cs: List[B]):
      sum[List[B], List[lab[A, List[B]]]] =
  Inl[List[B], List[lab[A, List[B]]]](cs)

def ordering_constraint_proof[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redpair[A]] =
  (a: xml) =>
    xml1element[redpair[A],
                 redpair[A]](List('o', 'r', 'd', 'e', 'r', 'i', 'n', 'g', 'C',
                                   'o', 'n', 's', 't', 'r', 'a', 'i', 'n', 't',
                                   'P', 'r', 'o', 'o', 'f'),
                              redpair[A](xml2name), id[redpair[A]], a)

def xml1many2elements_gen[A, B, C, D,
                           E](tag: List[Char], p1: xml => sum[List[Char], A],
                               p2: A => xml => sum[List[Char], B],
                               p3: xml => sum[List[Char], C],
                               p4: xml => sum[List[Char], D],
                               f: A => (List[B]) => C => D => E, x6: xml):
      sum[List[Char], E] =
  (tag, p1, p2, p3, p4, f, x6) match {
  case (tag, p1, p2, p3, p4, f, Xml(name, atts, cs, text)) =>
    {
      val ds: List[xml] = rev[xml](cs)
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (is_none[List[Char]](text) && less_eq_nat(nat(BigInt(3)), l))))
        binda[List[Char], A,
               E](p1(nth[xml](cs, Zero_nat)),
                   (x: A) =>
                     binda[List[Char], List[B],
                            E](map[B].apply(p2(x)).apply(tl[xml](take[xml](minus_nat(l,
      nat(BigInt(2))),
                                    cs))),
                                (xs: List[B]) =>
                                  binda[List[Char], C,
 E](p3(nth[xml](ds, one_nat)),
     (y: C) =>
       binda[List[Char], D,
              E](p4(nth[xml](ds, Zero_nat)),
                  (z: D) => Inr[E, List[Char]]((((f(x))(xs))(y))(z))))))
        else no_transform[E](tag, Xml(name, atts, cs, text)))
    }
}

def projected_rseq[A: equal](xml2name: xml => sum[List[Char], A], pi: A => nat,
                              x2: xml):
      sum[List[Char],
           ((term[A, List[Char]], term[A, List[Char]]),
             List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                          term[A, List[Char]]))])] =
  (xml2name, pi, x2) match {
  case (xml2name, pi, Xml(name, atts, cs, text)) =>
    {
      val tag: List[Char] =
        List('p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', 'R', 'e', 'w', 'r',
              'i', 't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e');
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (is_none[List[Char]](text) &&
                 equal_nata(size_list[xml](cs), nat(BigInt(2))))))
        binda[List[Char], (term[A, List[Char]], term[A, List[Char]]),
               ((term[A, List[Char]], term[A, List[Char]]),
                 List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                              term[A, List[Char]]))])]((rule[A](xml2name)).apply(nth[xml](cs,
           Zero_nat)),
                (r: (term[A, List[Char]], term[A, List[Char]])) =>
                  (rseq[A](xml2name, pi, r)).apply(nth[xml](cs, one_nat)))
        else no_transform[((term[A, List[Char]], term[A, List[Char]]),
                            List[(pos, ((term[A, List[Char]],
  term[A, List[Char]]),
 term[A, List[Char]]))])](tag, Xml(name, atts, cs, text)))
    }
}

def create_proj[A: equal](x0: projL[A], f: A): nat = (x0, f) match {
  case (Projection(p), f) =>
    lookupa[A, nat](f, p) match {
      case None => Zero_nat
      case Some(n) => n
    }
}

def flat_contexts[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[ctxt[A, List[Char]]]] =
  (a: xml) =>
    xml_many_elements[ctxt[A, List[Char]],
                       List[ctxt[A, List[Char]]]](List('f', 'l', 'a', 't', 'C',
                'o', 'n', 't', 'e', 'x', 't', 's'),
           (aa: xml) => ctxt[A](xml2name, aa), id[List[ctxt[A, List[Char]]]], a)

def uncurry_info[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (A, (List[((A, nat), List[A])],
                   (List[(term[A, List[Char]], term[A, List[Char]])],
                     List[(term[A, List[Char]], term[A, List[Char]])])))] =
  (a: xml) =>
    xml4elements[A, List[((A, nat), List[A])],
                  List[(term[A, List[Char]], term[A, List[Char]])],
                  List[(term[A, List[Char]], term[A, List[Char]])],
                  (A, (List[((A, nat), List[A])],
                        (List[(term[A, List[Char]], term[A, List[Char]])],
                          List[(term[A, List[Char]],
                                 term[A, List[Char]])])))](List('u', 'n', 'c',
                         'u', 'r', 'r', 'y', 'I', 'n', 'f', 'o', 'r', 'm', 'a',
                         't', 'i', 'o', 'n'),
                    xml2name,
                    (aa: xml) =>
                      xml_many_elements[((A, nat), List[A]),
 List[((A, nat),
        List[A])]](List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y',
                         'm', 'b', 'o', 'l', 's'),
                    (ab: xml) =>
                      xml2many_elements[A, nat, A,
 ((A, nat),
   List[A])](List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y', 'm',
                   'b', 'o', 'l', 'E', 'n', 't', 'r', 'y'),
              xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
              xml2name, (f: A) => (n: nat) => (ac: List[A]) => ((f, n), ac),
              ab),
                    id[List[((A, nat), List[A])]], aa),
                    (aa: xml) =>
                      xml1element[List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]](List('u', 'n', 'c', 'u', 'r', 'r', 'y', 'R', 'u', 'l',
                               'e', 's'),
                          rules[A](xml2name),
                          id[List[(term[A, List[Char]], term[A, List[Char]])]],
                          aa),
                    (aa: xml) =>
                      xml1element[List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]](List('e', 't', 'a', 'R', 'u', 'l', 'e', 's'),
                          rules[A](xml2name),
                          id[List[(term[A, List[Char]], term[A, List[Char]])]],
                          aa),
                    (aa: A) =>
                      (sml: List[((A, nat), List[A])]) =>
                        (u: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                          (e: List[(term[A, List[Char]], term[A, List[Char]])])
                            => (aa, (sml, (u, e))),
                    a)

def xml2dpProof[A: equal](xml2name: xml => sum[List[Char], lab[A, List[nat]]],
                           x: xml):
      sum[List[Char], dPProof[A, List[nat], List[Char]]] =
  xml1element[dPProof[A, List[nat], List[Char]],
               dPProof[A, List[nat],
                        List[Char]]](List('d', 'p', 'P', 'r', 'o', 'o', 'f'),
                                      (a: xml) =>
xml_options[dPProof[A, List[nat],
                     List[Char]]](List((List('p', 'I', 's', 'E', 'm', 'p', 't',
      'y'),
 (aa: xml) =>
   xml0elements[dPProof[A, List[nat],
                         List[Char]]](List('p', 'I', 's', 'E', 'm', 'p', 't',
    'y'),
                                       PisEmpty, aa)),
(List('d', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml_many_elements[(Option[dPProof[A, List[nat], List[Char]]],
                        List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])]),
                       dPProof[A, List[nat],
                                List[Char]]](List('d', 'e', 'p', 'G', 'r', 'a',
           'p', 'h', 'P', 'r', 'o', 'c'),
      (ab: xml) =>
        xml2nd_choice[List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                       dPProof[A, List[nat], List[Char]],
                       (Option[dPProof[A, List[nat], List[Char]]],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]],
                                      List[Char]])])](List('c', 'o', 'm', 'p',
                    'o', 'n', 'e', 'n', 't'),
               (ac: xml) =>
                 xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                    term[lab[A, List[nat]], List[Char]])],
                              List[(term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]],
   List[Char]])]](List('d', 'p', 's'), rules[lab[A, List[nat]]](xml2name),
                   id[List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])]],
                   ac),
               List('r', 'e', 'a', 'l', 'S', 'c', 'c'),
               (ac: xml) => xml2dpProof[A](xml2name, ac),
               (dps: List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])])
                 => (prfOpt: Option[dPProof[A, List[nat], List[Char]]]) =>
                      (prfOpt, dps),
               ab),
      (ab: List[(Option[dPProof[A, List[nat], List[Char]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])])])
        => DepGraphProc[A, List[nat], List[Char]](ab),
      aa)),
(List('r', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml3elements[redpair[lab[A, List[nat]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  dPProof[A, List[nat], List[Char]],
                  dPProof[A, List[nat],
                           List[Char]]](List('r', 'e', 'd', 'P', 'a', 'i', 'r',
      'P', 'r', 'o', 'c'),
 ordering_constraint_proof[lab[A, List[nat]]](xml2name),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('d', 'p', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) => xml2dpProof[A](xml2name, ab),
 (ab: redpair[lab[A, List[nat]]]) =>
   (b: List[(term[lab[A, List[nat]], List[Char]],
              term[lab[A, List[nat]], List[Char]])])
     => (c: dPProof[A, List[nat], List[Char]]) =>
          RedPairProc[A, List[nat], List[Char]](ab, b, c),
 aa)),
(List('r', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml4elements[redpair[lab[A, List[nat]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  dPProof[A, List[nat], List[Char]],
                  dPProof[A, List[nat],
                           List[Char]]](List('r', 'e', 'd', 'P', 'a', 'i', 'r',
      'U', 'r', 'P', 'r', 'o', 'c'),
 ordering_constraint_proof[lab[A, List[nat]]](xml2name),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('d', 'p', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('u', 's', 'a', 'b', 'l', 'e',
          'R', 'u', 'l', 'e', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) => xml2dpProof[A](xml2name, ab),
 (ab: redpair[lab[A, List[nat]]]) =>
   (b: List[(term[lab[A, List[nat]], List[Char]],
              term[lab[A, List[nat]], List[Char]])])
     => (c: List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])])
          => (d: dPProof[A, List[nat], List[Char]]) =>
               RedPairUrProc[A, List[nat], List[Char]](ab, b, c, d),
 aa)),
(List('m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o',
       'c'),
  (aa: xml) =>
    xml4elements[redpair[lab[A, List[nat]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  dPProof[A, List[nat], List[Char]],
                  dPProof[A, List[nat],
                           List[Char]]](List('m', 'o', 'n', 'o', 'R', 'e', 'd',
      'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c'),
 ordering_constraint_proof[lab[A, List[nat]]](xml2name),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('d', 'p', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('t', 'r', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) => xml2dpProof[A](xml2name, ab),
 (ab: redpair[lab[A, List[nat]]]) =>
   (b: List[(term[lab[A, List[nat]], List[Char]],
              term[lab[A, List[nat]], List[Char]])])
     => (c: List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])])
          => (d: dPProof[A, List[nat], List[Char]]) =>
               MonoRedPairProc[A, List[nat], List[Char]](ab, b, c, d),
 aa)),
(List('m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r',
       'o', 'c'),
  (aa: xml) =>
    xml5elements[redpair[lab[A, List[nat]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  dPProof[A, List[nat], List[Char]],
                  dPProof[A, List[nat],
                           List[Char]]](List('m', 'o', 'n', 'o', 'R', 'e', 'd',
      'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c'),
 ordering_constraint_proof[lab[A, List[nat]]](xml2name),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('d', 'p', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('t', 'r', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('u', 's', 'a', 'b', 'l', 'e',
          'R', 'u', 'l', 'e', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) => xml2dpProof[A](xml2name, ab),
 (ab: redpair[lab[A, List[nat]]]) =>
   (b: List[(term[lab[A, List[nat]], List[Char]],
              term[lab[A, List[nat]], List[Char]])])
     => (c: List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])])
          => (d: List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])])
               => (e: dPProof[A, List[nat], List[Char]]) =>
                    MonoRedPairUrProc[A, List[nat], List[Char]](ab, b, c, d, e),
 aa)),
(List('s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml4elements[sum[(sl_inter[A, List[nat]], Boolean),
                      Option[lab[A, List[nat]]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  dPProof[A, List[nat], List[Char]],
                  dPProof[A, List[nat],
                           List[Char]]](List('s', 'e', 'm', 'l', 'a', 'b', 'P',
      'r', 'o', 'c'),
 sl_inter[A, List[nat]](xml2name),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('d', 'p', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('t', 'r', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) => xml2dpProof[A](xml2name, ab),
 (ab: sum[(sl_inter[A, List[nat]], Boolean), Option[lab[A, List[nat]]]]) =>
   ab match {
     case Inl((i, true)) =>
       (ac: List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])])
         => (b: List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])])
              => (c: dPProof[A, List[nat], List[Char]]) =>
                   SemlabProc[A, List[nat],
                               List[Char]](i,
    Inl[(lab[A, List[nat]]) =>
          (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]],
         (List[Char],
           Boolean)]((ad: lab[A, List[nat]]) =>
                       (ba: List[nat]) => id_labeling[A, nat](ad, ba)),
    ac, b, c)
     case Inl((i, false)) =>
       (ac: List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])])
         => (b: List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])])
              => (c: dPProof[A, List[nat], List[Char]]) =>
                   SemlabProc[A, List[nat],
                               List[Char]](i,
    Inr[(List[Char], Boolean),
         (lab[A, List[nat]]) =>
           (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]]]((Nil, true)),
    ac, b, c)
     case Inr(ac) =>
       (b: List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])])
         => (c: List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])])
              => (d: dPProof[A, List[nat], List[Char]]) =>
                   RootlabProc[A, List[nat], List[Char]](ac, b, c, d)
   },
 aa)),
(List('s', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml1many2elements_gen[projL[lab[A, List[nat]]],
                           ((term[lab[A, List[nat]], List[Char]],
                              term[lab[A, List[nat]], List[Char]]),
                             List[(pos, ((term[lab[A, List[nat]], List[Char]],
   term[lab[A, List[nat]], List[Char]]),
  term[lab[A, List[nat]], List[Char]]))]),
                           List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                           dPProof[A, List[nat], List[Char]],
                           dPProof[A, List[nat],
                                    List[Char]]](List('s', 'u', 'b', 't', 'e',
               'r', 'm', 'P', 'r', 'o', 'c'),
          proj[lab[A, List[nat]]](xml2name),
          (pi: projL[lab[A, List[nat]]]) =>
            (ab: xml) =>
              projected_rseq[lab[A, List[nat]]](xml2name,
         (ac: lab[A, List[nat]]) => create_proj[lab[A, List[nat]]](pi, ac), ab),
          (ab: xml) =>
            xml1element[List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]],
                                      List[Char]])]](List('d', 'p', 's'),
              rules[lab[A, List[nat]]](xml2name),
              id[List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])]],
              ab),
          (ab: xml) => xml2dpProof[A](xml2name, ab),
          (ab: projL[lab[A, List[nat]]]) =>
            (b: List[((term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]]),
                       List[(pos, ((term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]]),
                                    term[lab[A, List[nat]], List[Char]]))])])
              => (c: List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])])
                   => (d: dPProof[A, List[nat], List[Char]]) =>
                        SubtermCriterionProc[A, List[nat],
      List[Char]](ab, b, c, d),
          aa)),
(List('f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o', 's',
       'u', 'r', 'e', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml5elements[lab[A, List[nat]], List[ctxt[lab[A, List[nat]], List[Char]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  dPProof[A, List[nat], List[Char]],
                  dPProof[A, List[nat],
                           List[Char]]](List('f', 'l', 'a', 't', 'C', 'o', 'n',
      't', 'e', 'x', 't', 'C', 'l', 'o', 's', 'u', 'r', 'e', 'P', 'r', 'o',
      'c'),
 (ab: xml) =>
   xml1element[lab[A, List[nat]],
                lab[A, List[nat]]](List('f', 'r', 'e', 's', 'h', 'S', 'y', 'm',
 'b', 'o', 'l'),
                                    xml2name, id[lab[A, List[nat]]], ab),
 flat_contexts[lab[A, List[nat]]](xml2name),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('d', 'p', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('t', 'r', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) => xml2dpProof[A](xml2name, ab),
 (ab: lab[A, List[nat]]) =>
   (b: List[ctxt[lab[A, List[nat]], List[Char]]]) =>
     (c: List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])])
       => (d: List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])])
            => (e: dPProof[A, List[nat], List[Char]]) =>
                 FccProc[A, List[nat], List[Char]](ab, b, c, d, e),
 aa)),
(List('u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml_choice[dPProof[A, List[nat],
                        List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r', 'y',
   'P', 'r', 'o', 'c'),
                                      List((ab: xml) =>
     xml4elements[(lab[A, List[nat]],
                    (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
                      (List[(term[lab[A, List[nat]], List[Char]],
                              term[lab[A, List[nat]], List[Char]])],
                        List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])]))),
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])],
                   dPProof[A, List[nat], List[Char]],
                   dPProof[A, List[nat],
                            List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r', 'y',
       'P', 'r', 'o', 'c'),
  uncurry_info[lab[A, List[nat]]](xml2name),
  (ac: xml) =>
    xml1element[List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])],
                 List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]],
                              List[Char]])]](List('d', 'p', 's'),
      rules[lab[A, List[nat]]](xml2name),
      id[List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])]],
      ac),
  (ac: xml) =>
    xml1element[List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])],
                 List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]],
                              List[Char]])]](List('t', 'r', 's'),
      rules[lab[A, List[nat]]](xml2name),
      id[List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])]],
      ac),
  (ac: xml) => xml2dpProof[A](xml2name, ac),
  (ac: (lab[A, List[nat]],
         (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
           (List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])],
             List[(term[lab[A, List[nat]], List[Char]],
                    term[lab[A, List[nat]], List[Char]])]))))
    => (b: List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])])
         => (c: List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])])
              => (d: dPProof[A, List[nat], List[Char]]) =>
                   UncurryProc[A, List[nat], List[Char]](None, ac, b, c, d),
  ab),
    (ab: xml) =>
      xml5elements[nat, (lab[A, List[nat]],
                          (List[((lab[A, List[nat]], nat),
                                  List[lab[A, List[nat]]])],
                            (List[(term[lab[A, List[nat]], List[Char]],
                                    term[lab[A, List[nat]], List[Char]])],
                              List[(term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]])]))),
                    List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])],
                    List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])],
                    dPProof[A, List[nat], List[Char]],
                    dPProof[A, List[nat],
                             List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r',
        'y', 'P', 'r', 'o', 'c'),
   (ac: xml) =>
     nata(List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'v', 'e', 'T', 'o',
                'p'),
           ac),
   uncurry_info[lab[A, List[nat]]](xml2name),
   (ac: xml) =>
     xml1element[List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]],
                               List[Char]])]](List('d', 'p', 's'),
       rules[lab[A, List[nat]]](xml2name),
       id[List[(term[lab[A, List[nat]], List[Char]],
                 term[lab[A, List[nat]], List[Char]])]],
       ac),
   (ac: xml) =>
     xml1element[List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]],
                               List[Char]])]](List('t', 'r', 's'),
       rules[lab[A, List[nat]]](xml2name),
       id[List[(term[lab[A, List[nat]], List[Char]],
                 term[lab[A, List[nat]], List[Char]])]],
       ac),
   (ac: xml) => xml2dpProof[A](xml2name, ac),
   (n: nat) =>
     (ac: (lab[A, List[nat]],
            (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
              (List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])]))))
       => (b: List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])])
            => (c: List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])])
                 => (d: dPProof[A, List[nat], List[Char]]) =>
                      UncurryProc[A, List[nat],
                                   List[Char]](Some[nat](n), ac, b, c, d),
   ab)),
                                      aa)),
(List('u', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c'),
  (aa: xml) =>
    xml3elements[List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                  dPProof[A, List[nat], List[Char]],
                  dPProof[A, List[nat],
                           List[Char]]](List('u', 'n', 'l', 'a', 'b', 'P', 'r',
      'o', 'c'),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('d', 'p', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) =>
   xml1element[List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]],
                             List[Char]])]](List('t', 'r', 's'),
     rules[lab[A, List[nat]]](xml2name),
     id[List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
     ab),
 (ab: xml) => xml2dpProof[A](xml2name, ab),
 (ab: List[(term[lab[A, List[nat]], List[Char]],
             term[lab[A, List[nat]], List[Char]])])
   => (b: List[(term[lab[A, List[nat]], List[Char]],
                 term[lab[A, List[nat]], List[Char]])])
        => (c: dPProof[A, List[nat], List[Char]]) =>
             UnlabProc[A, List[nat], List[Char]](ab, b, c),
 aa)),
(List('s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'P', 'r', 'o', 'c'),
  (xml1many_elements[Option[(redpair[lab[A, List[nat]]],
                              List[(term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]])])],
                      ((term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]]),
                        (List[(nat, nat)], List[(nat, nat)])),
                      dPProof[A, List[nat],
                               List[Char]]](List('s', 'i', 'z', 'e', 'C', 'h',
          'a', 'n', 'g', 'e', 'P', 'r', 'o', 'c'),
     (aa: xml) =>
       xml_options[Option[(redpair[lab[A, List[nat]]],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])])]](List((List('s', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r', 'i',
                              't', 'e', 'r', 'i', 'o', 'n'),
                         (ab: xml) =>
                           xml0elements[Option[(redpair[lab[A, List[nat]]],
         List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]],
                      List[Char]])])]](List('s', 'u', 'b', 't', 'e', 'r', 'm',
     'C', 'r', 'i', 't', 'e', 'r', 'i', 'o', 'n'),
None, ab)),
                        (List('r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', 'P',
                               'a', 'i', 'r'),
                          (ab: xml) =>
                            xml2elements[redpair[lab[A, List[nat]]],
  List[(term[lab[A, List[nat]], List[Char]],
         term[lab[A, List[nat]], List[Char]])],
  Option[(redpair[lab[A, List[nat]]],
           List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]],
                        List[Char]])])]](List('r', 'e', 'd', 'u', 'c', 't', 'i',
       'o', 'n', 'P', 'a', 'i', 'r'),
  ordering_constraint_proof[lab[A, List[nat]]](xml2name),
  (ac: xml) =>
    xml1element[List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])],
                 List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]],
                              List[Char]])]](List('u', 's', 'a', 'b', 'l', 'e',
           'R', 'u', 'l', 'e', 's'),
      rules[lab[A, List[nat]]](xml2name),
      id[List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])]],
      ac),
  (redp: redpair[lab[A, List[nat]]]) =>
    (ur: List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])])
      => Some[(redpair[lab[A, List[nat]]],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])])]((redp, ur)),
  ab))),
                   aa),
     scg[lab[A, List[nat]]](xml2name))).apply((aa:
         Option[(redpair[lab[A, List[nat]]],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])])])
        => aa match {
             case None =>
               (ab: List[((term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]]),
                           (List[(nat, nat)], List[(nat, nat)]))])
                 => SizeChangeSubTerm[A, List[nat], List[Char]](ab)
             case Some(redp_ur) =>
               (ab: List[((term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]]),
                           (List[(nat, nat)], List[(nat, nat)]))])
                 => SizeChangeRedPair[A, List[nat],
                                       List[Char]](fst[redpair[lab[A,
                            List[nat]]],
                List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])]](redp_ur),
            snd[redpair[lab[A, List[nat]]],
                 List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])]](redp_ur),
            ab)
           }))),
                                   a),
                                      id[dPProof[A, List[nat], List[Char]]], x)

implicit def preorder_lab[A: equal: order, B: equal: order]: preorder[lab[A, B]]
  = new preorder[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

implicit def order_lab[A: equal: order, B: equal: order]: order[lab[A, B]] = new
  order[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

def unsharp[A, B](x0: lab[A, B]): lab[A, B] = x0 match {
  case Sharp(f) => f
  case Lab(v, va) => Lab[A, B](v, va)
  case FunLab(v, va) => FunLab[A, B](v, va)
  case UnLab(v) => UnLab[A, B](v)
}

abstract sealed class relTRSProof[+A, +B, +C]
final case class
  RelRuleRemoval[A, B,
                  C](a: redpair[lab[A, B]],
                      b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                      c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                      d: relTRSProof[A, B, C])
  extends relTRSProof[A, B, C]
final case class RelSisEmpty[A, B, C](a: tRSProof[A, B, C]) extends
  relTRSProof[A, B, C]
final case class
  RelUnlab[A, B,
            C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                c: relTRSProof[A, B, C])
  extends relTRSProof[A, B, C]
final case class
  RelSemlab[A, B,
             C](a: sl_inter[A, B],
                 b: sum[(lab[A, B]) => (List[nat]) => sum[B, List[lab[A, B]]],
                         C],
                 c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                 d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                 e: relTRSProof[A, B, C])
  extends relTRSProof[A, B, C]
final case class
  RelRootlab[A, B,
              C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  c: relTRSProof[A, B, C])
  extends relTRSProof[A, B, C]
final case class
  RelUncurry[A, B,
              C](a: (lab[A, B],
                      (List[((lab[A, B], nat), List[lab[A, B]])],
                        (List[(term[lab[A, B], C], term[lab[A, B], C])],
                          List[(term[lab[A, B], C], term[lab[A, B], C])]))),
                  b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  d: relTRSProof[A, B, C])
  extends relTRSProof[A, B, C]
final case object RelRisEmpty extends relTRSProof[Nothing, Nothing, Nothing]

abstract sealed class certificationProblem[+A, +B, +C]
final case class
  TRSProof[A, C, B](a: List[(term[A, C], term[A, C])], b: tRSProof[A, B, C])
  extends certificationProblem[A, B, C]
final case class
  TRSDisproof[A, C,
               B](a: List[(term[A, C], term[A, C])],
                   b: tRSDisproof[lab[A, B], C])
  extends certificationProblem[A, B, C]
final case class
  RelTRSProof[A, C,
               B](a: (List[(term[A, C], term[A, C])],
                       List[(term[A, C], term[A, C])]),
                   b: relTRSProof[A, B, C])
  extends certificationProblem[A, B, C]
final case class
  RelTRSDisproof[A, C,
                  B](a: (List[(term[A, C], term[A, C])],
                          List[(term[A, C], term[A, C])]),
                      b: relTRSDisproof[lab[A, B], C])
  extends certificationProblem[A, B, C]

def lpo_strict_unbounded[A: equal: show, B: equal: show](pr: A => nat => nat):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](fst[Boolean,
                               Boolean](lpo_unbounded[A,
               B].apply(pr).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_term[A, B](Zero_nat, s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            'L', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_prec_term[A, B](Zero_nat, t), shows_nl)))))
    }

def lpo_nstrict_unbounded[A: equal: show, B: equal: show](pr: A => nat => nat):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](snd[Boolean,
                               Boolean](lpo_unbounded[A,
               B].apply(pr).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_term[A, B](Zero_nat, s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            '=', 'L', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_prec_term[A, B](Zero_nat, t), shows_nl)))))
    }

def create_LPO_redpair[A: show, B: equal: show,
                        C: equal: show](prec_repr_to_pr:
  (List[((A, nat), (nat, order_tag))]) => B => nat => nat,
 pr: List[((A, nat), (nat, order_tag))]):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
          (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
            (B => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  {
    val p: B => nat => nat = prec_repr_to_pr(pr);
    (Inr[Unit, (List[Char]) => List[Char]](()),
      (lpo_strict_unbounded[B, C](p),
        (lpo_nstrict_unbounded[B, C](p),
          ((a: B) => (b: nat) => (c: nat) => full_af[B](a, b, c),
            (Inr[Unit, (List[Char]) => List[Char]](()),
              shows_lpo_repr[A](pr))))))
  }

def create_MPO_redpair[A: show, B: equal: show,
                        C: equal: show](prec_repr_to_pr:
  (List[((A, nat), (nat, order_tag))]) => B => nat => nat,
 pr: List[((A, nat), (nat, order_tag))]):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
          (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
            (B => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  {
    val p: B => nat => nat = prec_repr_to_pr(pr);
    (Inr[Unit, (List[Char]) => List[Char]](()),
      (mpo_strict[B, C](p, Zero_nat),
        (mpo_nstrict[B, C](p, Zero_nat),
          ((a: B) => (b: nat) => (c: nat) => full_af[B](a, b, c),
            (Inr[Unit, (List[Char]) => List[Char]](()),
              shows_mpo_repr[A](pr))))))
  }

def rpo_strict_unbounded[A: equal: show,
                          B: equal: show](pr:
    (A => nat => A => nat => (Boolean, Boolean), A => Boolean),
   c: A => nat => order_tag):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](fst[Boolean,
                               Boolean](rpo_unbounded[A,
               B].apply(pr).apply(c).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_term[A, B](Zero_nat, s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            'R', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_prec_term[A, B](Zero_nat, t), shows_nl)))))
    }

def rpo_nstrict_unbounded[A: equal: show,
                           B: equal: show](pr:
     (A => nat => A => nat => (Boolean, Boolean), A => Boolean),
    c: A => nat => order_tag):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](snd[Boolean,
                               Boolean](rpo_unbounded[A,
               B].apply(pr).apply(c).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_term[A, B](Zero_nat, s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            '=', 'R', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_prec_term[A, B](Zero_nat, t), shows_nl)))))
    }

def create_RPO_redpair[A: show, B: equal: show,
                        C: equal: show](prec_repr_to_pr:
  (List[((A, nat), (nat, order_tag))]) =>
    (B => nat => nat, B => nat => order_tag),
 pr: List[((A, nat), (nat, order_tag))]):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
          (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
            (B => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  {
    val (p, tau): (B => nat => nat, B => nat => order_tag) =
      prec_repr_to_pr(pr);
    (Inr[Unit, (List[Char]) => List[Char]](()),
      (rpo_strict_unbounded[B, C]((prc_nat[B](p), prl_nat[B](p)), tau),
        (rpo_nstrict_unbounded[B, C]((prc_nat[B](p), prl_nat[B](p)), tau),
          ((a: B) => (b: nat) => (c: nat) => full_af[B](a, b, c),
            (Inr[Unit, (List[Char]) => List[Char]](()),
              shows_rpo_repr[A](pr))))))
  }

def check_left_linear_trs[A: show,
                           B: equal: show](trs: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(term[A, B], term[A, B]), Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[A, B], term[A, B]),
                                 (term[A, B], term[A, B])),
                                Unit,
                                (term[A, B],
                                  term[A,
B])](forallm[(term[A, B], term[A, B]),
              (term[A, B],
                term[A, B])]((x: (term[A, B], term[A, B])) =>
                               (if (linear_term[A,
         B](fst[term[A, B], term[A, B]](x)))
                                 Inr[Unit, (term[A, B], term[A, B])](())
                                 else Inl[(term[A, B], term[A, B]), Unit](x)),
                              trs),
      (x: ((term[A, B], term[A, B]), (term[A, B], term[A, B]))) =>
        Inl[(term[A, B], term[A, B]),
             Unit](snd[(term[A, B], term[A, B]), (term[A, B], term[A, B])](x))),
                         (_: (term[A, B], term[A, B])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_prec_list[(term[A, B], term[A, B])](Zero_nat, trs),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('i', 's', ' ', 'n', 'o', 't', ' ', 'l',
                                       'e', 'f', 't', '-', 'l', 'i', 'n', 'e',
                                       'a', 'r')),
               shows_nl)))))

def check_no_var[A: show, B: show](t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (is_var[A, B](t)),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('v', 'a', 'r',
                          'i', 'a', 'b', 'l', 'e', ' ', 'f', 'o', 'u', 'n',
                          'd')),
  shows_nl))

def check_no_defined_root_defs[A: equal: show,
                                B: show](f: List[(A, nat)], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (membera[(A, nat)](f,
    (the[A](root[A, B](t)), num_args[A, B](t)))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_prec_term[A, B](Zero_nat, t),
                     shows_string.apply(List(' ', 'i', 's', ' ', 'd', 'e', 'f',
      'i', 'n', 'e', 'd')))))

def block_term[A, B](f: A, x1: term[A, B]): term[A, B] = (f, x1) match {
  case (f, Var(x)) => Var[B, A](x)
  case (f, Fun(g, ts)) =>
    Fun[A, B](g, mapa[term[A, B],
                       term[A, B]]((t: term[A, B]) => Fun[A, B](f, List(t)),
                                    ts))
}

def block_rule[A, B](f: A, r: (term[A, B], term[A, B])):
      (term[A, B], term[A, B]) =
  (block_term[A, B](f, fst[term[A, B], term[A, B]](r)),
    block_term[A, B](f, snd[term[A, B], term[A, B]](r)))

def check_superset_of_blocked[A: equal: show](f: A,
       ra: List[(term[A, List[Char]], term[A, List[Char]])],
       r: List[(term[A, List[Char]], term[A, List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                                (List[Char]) =>
                                  List[Char]](catcha[((term[A, List[Char]],
                term[A, List[Char]]),
               (term[A, List[Char]], term[A, List[Char]])),
              Unit,
              (term[A, List[Char]],
                term[A, List[Char]])](forallm[(term[A, List[Char]],
        term[A, List[Char]]),
       (term[A, List[Char]],
         term[A, List[Char]])]((x: (term[A, List[Char]], term[A, List[Char]]))
                                 => (if (membera[(term[A, List[Char]],
           term[A, List[Char]])](ra, block_rule[A, List[Char]](f, x)))
                                      Inr[Unit,
   (term[A, List[Char]], term[A, List[Char]])](())
                                      else Inl[(term[A, List[Char]],
         term[A, List[Char]]),
        Unit](x)),
                                r),
                                       (x:
  ((term[A, List[Char]], term[A, List[Char]]),
    (term[A, List[Char]], term[A, List[Char]])))
 => Inl[(term[A, List[Char]], term[A, List[Char]]),
         Unit](snd[(term[A, List[Char]], term[A, List[Char]]),
                    (term[A, List[Char]], term[A, List[Char]])](x))),
       (x: (term[A, List[Char]], term[A, List[Char]])) =>
         Inl[(List[Char]) => List[Char],
              Unit](comp[List[Char], List[Char],
                          List[Char]](shows_string.apply(List('t', 'h', 'e',
                       ' ', 'r', 'u', 'l', 'e', ' ')),
                                       comp[List[Char], List[Char],
     List[Char]](shows_rule[A, List[Char]](block_rule[A, List[Char]](f, x)),
                  comp[List[Char], List[Char],
                        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                     'm', 'i', 's', 's', 'i', 'n', 'g')),
                                     shows_nl))))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_trs[A, List[Char]](r),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List('i', 's', ' ', 'n',
                    'o', 't', ' ', 'a', ' ', 's', 'u', 'b', 's', 'e', 't', ' ',
                    'o', 'f')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_trs[A, List[Char]](ra),
               comp[List[Char], List[Char], List[Char]](x, shows_nl))))))

def fcc_proc[A: equal: show,
              B: equal: show](n: nat,
                               pa: List[(term[lab[A, B], List[Char]],
  term[lab[A, B], List[Char]])],
                               ra: List[(term[lab[A, B], List[Char]],
  term[lab[A, B], List[Char]])],
                               f: lab[A, B],
                               fcs: List[ctxt[lab[A, B], List[Char]]],
                               p: List[(term[lab[A, B], List[Char]],
 term[lab[A, B], List[Char]])],
                               r: List[(term[lab[A, B], List[Char]],
 term[lab[A, B], List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val fa: (lab[A, B], nat) = (f, Suc(Zero_nat))
    val cf: ctxt[lab[A, B], List[Char]] =
      More[lab[A, B], List[Char]](f, Nil, Hole, Nil)
    val fcsa: List[ctxt[lab[A, B], List[Char]]] = cf :: fcs
    val vs: List[List[Char]] = vars_trs_list[lab[A, B], List[Char]](ra)
    val fs: List[(lab[A, B], nat)] =
      list_union[(lab[A, B],
                   nat)](funas_below_trs_list[lab[A, B], List[Char]](pa),
                          funas_trs_list[lab[A, B], List[Char]](ra))
    val fas: List[(lab[A, B], nat)] = fa :: fs
    val unlabn: (lab[A, B]) => lab[A, B] =
      funpow[lab[A, B]](n, (a: lab[A, B]) => unlab[A, B](a))
    val ds: List[(lab[A, B], nat)] =
      defs_list[lab[A, B],
                 List[Char]](map_trs_list[lab[A, B], lab[A, B],
   List[Char]](unlabn, ra));
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (membera[(lab[A, B],
          nat)](ds, (unlabn(f), one_nat))),
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_lab[A, B](Zero_nat, unlabn(f)),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                    'n', 'o', 't', ' ', 'f', 'r', 'e', 's', 'h')),
                                    shows_nl))),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](check_left_linear_trs[lab[A, B],
                List[Char]](ra),
                                  (_: Unit) =>
                                    binda[(List[Char]) => List[Char], Unit,
   Unit](check_wf_trs[lab[A, B], List[Char]](ra),
          (_: Unit) =>
            binda[(List[Char]) => List[Char], Unit,
                   Unit](catcha[((term[lab[A, B], List[Char]],
                                   term[lab[A, B], List[Char]]),
                                  (List[Char]) => List[Char]),
                                 Unit,
                                 (List[Char]) =>
                                   List[Char]](forallm[(term[lab[A, B],
                      List[Char]],
                 term[lab[A, B], List[Char]]),
                (List[Char]) =>
                  List[Char]]((rb: (term[lab[A, B], List[Char]],
                                     term[lab[A, B], List[Char]]))
                                => binda[(List[Char]) => List[Char], Unit,
  Unit](check_no_var[lab[A, B],
                      List[Char]](fst[term[lab[A, B], List[Char]],
                                       term[lab[A, B], List[Char]]](rb)),
         (_: Unit) =>
           binda[(List[Char]) => List[Char], Unit,
                  Unit](check_no_var[lab[A, B],
                                      List[Char]](snd[term[lab[A, B],
                    List[Char]],
               term[lab[A, B], List[Char]]](rb)),
                         (_: Unit) =>
                           check_no_defined_root_defs[lab[A, B],
               List[Char]](ds, map_term[lab[A, B], lab[A, B],
 List[Char]](unlabn,
              snd[term[lab[A, B], List[Char]],
                   term[lab[A, B], List[Char]]](rb))))),
                               pa),
        (x: ((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
              (List[Char]) => List[Char]))
          => Inl[(List[Char]) => List[Char],
                  Unit](snd[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]]),
                             (List[Char]) => List[Char]](x))),
                          (_: Unit) =>
                            binda[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[(ctxt[lab[A, B], List[Char]],
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[ctxt[lab[A, B], List[Char]],
                                (List[Char]) =>
                                  List[Char]]((a: ctxt[lab[A, B], List[Char]])
        => check_flat_ctxt[lab[A, B]](vs, a),
       fcsa),
                        (x: (ctxt[lab[A, B], List[Char]],
                              (List[Char]) => List[Char]))
                          => Inl[(List[Char]) => List[Char],
                                  Unit](snd[ctxt[lab[A, B], List[Char]],
     (List[Char]) => List[Char]](x))),
  (_: Unit) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](catcha[(ctxt[lab[A, B], List[Char]],
                          (List[Char]) => List[Char]),
                         Unit,
                         (List[Char]) =>
                           List[Char]](forallm[ctxt[lab[A, B], List[Char]],
        (List[Char]) =>
          List[Char]]((a: ctxt[lab[A, B], List[Char]]) =>
                        check_is_flat_ctxt[lab[A, B]](vs, fas, a),
                       fcsa),
(x: (ctxt[lab[A, B], List[Char]], (List[Char]) => List[Char])) =>
  Inl[(List[Char]) => List[Char],
       Unit](snd[ctxt[lab[A, B], List[Char]], (List[Char]) => List[Char]](x))),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](catcha[((lab[A, B], nat),
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[(lab[A, B], nat),
                        (List[Char]) =>
                          List[Char]]((a: (lab[A, B], nat)) =>
check_flat_ctxt_complete[lab[A, B]](fcsa, a),
                                       fas),
                (x: ((lab[A, B], nat), (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[(lab[A, B], nat),
                                  (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    binda[(List[Char]) => List[Char], Unit,
   Unit](catcha[((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
                  (List[Char]) => List[Char]),
                 Unit,
                 (List[Char]) =>
                   List[Char]](forallm[(term[lab[A, B], List[Char]],
 term[lab[A, B], List[Char]]),
(List[Char]) =>
  List[Char]]((a: (term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])) =>
                check_rule_preserving[lab[A, B]](fcsa, r, a),
               ra),
                                (x: ((term[lab[A, B], List[Char]],
                                       term[lab[A, B], List[Char]]),
                                      (List[Char]) => List[Char]))
                                  => Inl[(List[Char]) => List[Char],
  Unit](snd[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
             (List[Char]) => List[Char]](x))),
          (_: Unit) =>
            binda[(List[Char]) => List[Char], Unit,
                   Unit](catcha[((term[lab[A, B], List[Char]],
                                   term[lab[A, B], List[Char]]),
                                  (List[Char]) => List[Char]),
                                 Unit,
                                 (List[Char]) =>
                                   List[Char]](forallm[(term[lab[A, B],
                      List[Char]],
                 term[lab[A, B], List[Char]]),
                (List[Char]) =>
                  List[Char]]((a: (term[lab[A, B], List[Char]],
                                    term[lab[A, B], List[Char]]))
                                => check_rule_reflecting[lab[A,
                      B]](fcsa, ra, a),
                               r),
        (x: ((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
              (List[Char]) => List[Char]))
          => Inl[(List[Char]) => List[Char],
                  Unit](snd[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]]),
                             (List[Char]) => List[Char]](x))),
                          (_: Unit) =>
                            check_superset_of_blocked[lab[A,
                   B]](f, p, pa))))))))))
  }

def check_supteq[A: equal: show, B: equal: show](s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](member[(term[A, B], term[A, B])]((s, t), supteq[A, B]),
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_term[A, B](Zero_nat, t),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't',
     ' ', 'a', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ', 'o', 'f', ' ')),
                     shows_prec_term[A, B](Zero_nat, s))))

def rqs_impl[A, B](x0: ta_rule_impl[A, B]): List[A] = x0 match {
  case TA_rule_impl(f, qsa, q, qs) => qsa
}

def uncurry_eta_split[A: equal,
                       B: equal](eboth: List[(term[A, B], term[A, B])],
                                  rtest: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  {
    val test: ((term[A, B], term[A, B])) => Boolean =
      (a: (term[A, B], term[A, B])) =>
        {
          val (l, r): (term[A, B], term[A, B]) = a;
          less_nat(Zero_nat, num_args[A, B](l)) &&
            (less_nat(Zero_nat, num_args[A, B](r)) &&
              list_ex[(term[A, B],
                        term[A, B])]((aa: (term[A, B], term[A, B])) =>
                                       eq_rule_mod_vars[A,
                 B]((hd[term[A, B]](args[A, B](l)),
                      hd[term[A, B]](args[A, B](r))),
                     aa),
                                      rtest))
        };
    partition[(term[A, B], term[A, B])](test, eboth)
  }

def uncurry_rel_tt[A: equal: show](x0: (A,
 (List[((A, nat), List[A])],
   (List[(term[A, List[Char]], term[A, List[Char]])],
     List[(term[A, List[Char]], term[A, List[Char]])]))),
                                    ra: List[(term[A, List[Char]],
       term[A, List[Char]])],
                                    rwa: List[(term[A, List[Char]],
        term[A, List[Char]])],
                                    r: List[(term[A, List[Char]],
      term[A, List[Char]])],
                                    rw: List[(term[A, List[Char]],
       term[A, List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  (x0, ra, rwa, r, rw) match {
  case ((a, (sml, (u, eb))), ra, rwa, r, rw) =>
    {
      val (e, ew):
            (List[(term[A, List[Char]], term[A, List[Char]])],
              List[(term[A, List[Char]], term[A, List[Char]])])
        = uncurry_eta_split[A, List[Char]](eb, ra)
      val sm: A => nat => List[A] =
        sig_list_to_sig_map[A](a, sml,
                                (_: List[((A, nat), List[A])]) =>
                                  (f: A) => (_: nat) => f)
      val r_eta: List[(term[A, List[Char]], term[A, List[Char]])] = e ++ ra
      val rw_eta: List[(term[A, List[Char]], term[A, List[Char]])] = ew ++ rwa
      val rb_eta: List[(term[A, List[Char]], term[A, List[Char]])] =
        r_eta ++ rw_eta;
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[((term[A, List[Char]], term[A, List[Char]]),
                            (List[Char]) => List[Char]),
                           Unit,
                           (List[Char]) =>
                             List[Char]](forallm[(term[A, List[Char]],
           term[A, List[Char]]),
          (List[Char]) =>
            List[Char]]((b: (term[A, List[Char]], term[A, List[Char]])) =>
                          {
                            val (l, _):
                                  (term[A, List[Char]], term[A, List[Char]])
                              = b;
                            check[(List[Char]) =>
                                    List[Char]](hvf_term[A, List[Char]](a, l),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_list[Char](Zero_nat,
          List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_term[A, List[Char]](l),
       shows_prec_list[Char](Zero_nat,
                              List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                                    'w', 'e', 'd')))))
                          },
                         r_eta),
  (x: ((term[A, List[Char]], term[A, List[Char]]), (List[Char]) => List[Char]))
    => Inl[(List[Char]) => List[Char],
            Unit](snd[(term[A, List[Char]], term[A, List[Char]]),
                       (List[Char]) => List[Char]](x))),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](catcha[((term[A, List[Char]],
     term[A, List[Char]]),
    (List[Char]) => List[Char]),
   Unit,
   (List[Char]) =>
     List[Char]](forallm[(term[A, List[Char]], term[A, List[Char]]),
                          (List[Char]) =>
                            List[Char]]((b:
   (term[A, List[Char]], term[A, List[Char]]))
  => {
       val (l, _): (term[A, List[Char]], term[A, List[Char]]) = b;
       check[(List[Char]) =>
               List[Char]](hvf_term[A, List[Char]](a, l),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_prec_list[Char](Zero_nat,
                             List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i',
                                   'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l',
                                   'h', 's', ' ')),
       comp[List[Char], List[Char],
             List[Char]](shows_term[A, List[Char]](l),
                          shows_prec_list[Char](Zero_nat,
         List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd')))))
     },
 rw_eta),
                  (x: ((term[A, List[Char]], term[A, List[Char]]),
                        (List[Char]) => List[Char]))
                    => Inl[(List[Char]) => List[Char],
                            Unit](snd[(term[A, List[Char]],
term[A, List[Char]]),
                                       (List[Char]) => List[Char]](x))),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], Unit,
     Unit](eta_closed_rules[A, List[Char]](a, sm, r_eta, r_eta),
            (_: Unit) =>
              binda[(List[Char]) => List[Char], Unit,
                     Unit](eta_closed_rules[A,
     List[Char]](a, sm, rb_eta, rb_eta),
                            (_: Unit) =>
                              binda[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[((term[A, List[Char]],
             term[A, List[Char]]),
            (List[Char]) => List[Char]),
           Unit,
           (List[Char]) =>
             List[Char]](forallm[(term[A, List[Char]], term[A, List[Char]]),
                                  (List[Char]) =>
                                    List[Char]]((aa:
           (term[A, List[Char]], term[A, List[Char]]))
          => {
               val (l, rb): (term[A, List[Char]], term[A, List[Char]]) = aa;
               check[(List[Char]) =>
                       List[Char]](! (is_var[A, List[Char]](l)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('l', 'h', 's', ' ', 'm', 'u', 's',
   't', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ', 'a', ' ', 'v', 'a', 'r', 'i',
   'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', ' ')),
               shows_rule[A, List[Char]]((l, rb))))
             },
         rw_eta),
                          (x: ((term[A, List[Char]], term[A, List[Char]]),
                                (List[Char]) => List[Char]))
                            => Inl[(List[Char]) => List[Char],
                                    Unit](snd[(term[A, List[Char]],
        term[A, List[Char]]),
       (List[Char]) => List[Char]](x))),
    (_: Unit) =>
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                           (List[Char]) =>
                             List[Char]](check_subseteq[(term[A, List[Char]],
                  term[A, List[Char]])]((uncurry_rules[A,
                List[Char]](a, sm)).apply(r_eta),
 r),
  (x: (term[A, List[Char]], term[A, List[Char]])) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u',
                      'l', 'e', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_rule[A, List[Char]](x),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
  'i', 'n', 'g')))))),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](catcha[(term[A, List[Char]],
    term[A, List[Char]]),
   Unit,
   (List[Char]) =>
     List[Char]](check_subseteq[(term[A, List[Char]],
                                  term[A,
List[Char]])]((uncurry_rules[A, List[Char]](a, sm)).apply(rw_eta), rw),
                  (x: (term[A, List[Char]], term[A, List[Char]])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](comp[List[Char], List[Char],
                                     List[Char]](shows_prec_list[Char](Zero_nat,
                                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e',
                                      'd', ' ', 'r', 'u', 'l', 'e', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_rule[A, List[Char]](x),
                             shows_prec_list[Char](Zero_nat,
            List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))),
                                    (_: Unit) =>
                                      catcha[(term[A, List[Char]],
       term[A, List[Char]]),
      Unit,
      (List[Char]) =>
        List[Char]](check_CS_subseteq[A, List[Char]](uncurry_of_sig_list[A](a,
                                     sml, sm),
              rw),
                     (x: (term[A, List[Char]], term[A, List[Char]])) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](shows_prec_list[Char](Zero_nat,
                                   List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ',
 'r', 'u', 'l', 'e', ' ')),
             comp[List[Char], List[Char],
                   List[Char]](shows_rule[A, List[Char]](x),
                                shows_prec_list[Char](Zero_nat,
               List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n',
                     'g')))))))))))))
    }
}

def xml2trsInput[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (List[(term[A, List[Char]], term[A, List[Char]])],
               Option[List[(term[A, List[Char]], term[A, List[Char]])]])] =
  (a: xml) =>
    xml1element[(List[(term[A, List[Char]], term[A, List[Char]])],
                  Option[List[(term[A, List[Char]], term[A, List[Char]])]]),
                 (List[(term[A, List[Char]], term[A, List[Char]])],
                   Option[List[(term[A, List[Char]],
                                 term[A, List[Char]])]])](List('i', 'n', 'p',
                        'u', 't'),
                   (aa: xml) =>
                     xml1to2elements[List[(term[A, List[Char]],
    term[A, List[Char]])],
                                      List[(term[A, List[Char]],
     term[A, List[Char]])],
                                      (List[(term[A, List[Char]],
      term[A, List[Char]])],
Option[List[(term[A, List[Char]],
              term[A, List[Char]])]])](List('t', 'r', 's', 'I', 'n', 'p', 'u',
     't'),
(ab: xml) =>
  xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
               List[(term[A, List[Char]],
                      term[A, List[Char]])]](List('t', 'r', 's'),
      rules[A](xml2name), id[List[(term[A, List[Char]], term[A, List[Char]])]],
      ab),
(ab: xml) =>
  xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
               List[(term[A, List[Char]],
                      term[A, List[Char]])]](List('r', 'e', 'l', 'a', 't', 'i',
           'v', 'e', 'R', 'u', 'l', 'e', 's'),
      rules[A](xml2name), id[List[(term[A, List[Char]], term[A, List[Char]])]],
      ab),
(ab: List[(term[A, List[Char]], term[A, List[Char]])]) =>
  (b: Option[List[(term[A, List[Char]], term[A, List[Char]])]]) => (ab, b),
aa),
                   id[(List[(term[A, List[Char]], term[A, List[Char]])],
                        Option[List[(term[A, List[Char]],
                                      term[A, List[Char]])]])],
                   a)

def xml2trsProof[A: equal](xml2name: xml => sum[List[Char], lab[A, List[nat]]],
                            x: xml):
      sum[List[Char], tRSProof[A, List[nat], List[Char]]] =
  xml1element[tRSProof[A, List[nat], List[Char]],
               tRSProof[A, List[nat],
                         List[Char]]](List('t', 'r', 's', 'T', 'e', 'r', 'm',
    'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
                                       (a: xml) =>
 xml_options[tRSProof[A, List[nat],
                       List[Char]]](List((List('r', 'I', 's', 'E', 'm', 'p',
        't', 'y'),
   (aa: xml) =>
     xml0elements[tRSProof[A, List[nat],
                            List[Char]]](List('r', 'I', 's', 'E', 'm', 'p', 't',
       'y'),
  RisEmpty, aa)),
  (List('d', 'p', 'T', 'r', 'a', 'n', 's'),
    (aa: xml) =>
      xml3elements[List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])],
                    Boolean, dPProof[A, List[nat], List[Char]],
                    tRSProof[A, List[nat],
                              List[Char]]](List('d', 'p', 'T', 'r', 'a', 'n',
         's'),
    (ab: xml) =>
      xml1element[List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, List[nat]]](xml2name),
        id[List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]],
        ab),
    (ab: xml) =>
      bool(List('m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l',
                 's'),
            ab),
    (ab: xml) => xml2dpProof[A](xml2name, ab),
    (dps: List[(term[lab[A, List[nat]], List[Char]],
                 term[lab[A, List[nat]], List[Char]])])
      => (_: Boolean) =>
           (ab: dPProof[A, List[nat], List[Char]]) =>
             DPTrans[A, List[nat], List[Char]](dps, ab),
    aa)),
  (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
    (aa: xml) =>
      xml3elements[redpair[lab[A, List[nat]]],
                    List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])],
                    tRSProof[A, List[nat], List[Char]],
                    tRSProof[A, List[nat],
                              List[Char]]](List('r', 'u', 'l', 'e', 'R', 'e',
         'm', 'o', 'v', 'a', 'l'),
    ordering_constraint_proof[lab[A, List[nat]]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, List[nat]]](xml2name),
        id[List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]],
        ab),
    (ab: xml) => xml2trsProof[A](xml2name, ab),
    (ab: redpair[lab[A, List[nat]]]) =>
      (b: List[(term[lab[A, List[nat]], List[Char]],
                 term[lab[A, List[nat]], List[Char]])])
        => (c: tRSProof[A, List[nat], List[Char]]) =>
             RuleRemoval[A, List[nat], List[Char]](ab, b, c),
    aa)),
  (List('u', 'n', 'l', 'a', 'b'),
    (aa: xml) =>
      xml2elements[List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])],
                    tRSProof[A, List[nat], List[Char]],
                    tRSProof[A, List[nat],
                              List[Char]]](List('u', 'n', 'l', 'a', 'b'),
    (ab: xml) =>
      xml1element[List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, List[nat]]](xml2name),
        id[List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]],
        ab),
    (ab: xml) => xml2trsProof[A](xml2name, ab),
    (ab: List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])])
      => (b: tRSProof[A, List[nat], List[Char]]) =>
           Unlab[A, List[nat], List[Char]](ab, b),
    aa)),
  (List('s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'),
    (aa: xml) =>
      xml2elements[List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])],
                    tRSProof[A, List[nat], List[Char]],
                    tRSProof[A, List[nat],
                              List[Char]]](List('s', 't', 'r', 'i', 'n', 'g',
         'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'),
    (ab: xml) =>
      xml1element[List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, List[nat]]](xml2name),
        id[List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]],
        ab),
    (ab: xml) => xml2trsProof[A](xml2name, ab),
    (ab: List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])])
      => (b: tRSProof[A, List[nat], List[Char]]) =>
           StringReversal[A, List[nat], List[Char]](ab, b),
    aa)),
  (List('s', 'e', 'm', 'l', 'a', 'b'),
    (aa: xml) =>
      xml3elements[sum[(sl_inter[A, List[nat]], Boolean),
                        Option[lab[A, List[nat]]]],
                    List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])],
                    tRSProof[A, List[nat], List[Char]],
                    tRSProof[A, List[nat],
                              List[Char]]](List('s', 'e', 'm', 'l', 'a', 'b'),
    sl_inter[A, List[nat]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, List[nat]]](xml2name),
        id[List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]],
        ab),
    (ab: xml) => xml2trsProof[A](xml2name, ab),
    (ab: sum[(sl_inter[A, List[nat]], Boolean), Option[lab[A, List[nat]]]]) =>
      ab match {
        case Inl((i, true)) =>
          (ac: List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])])
            => (b: tRSProof[A, List[nat], List[Char]]) =>
                 Semlab[A, List[nat],
                         List[Char]](i, Inl[(lab[A, List[nat]]) =>
      (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]],
     List[Char]]((ad: lab[A, List[nat]]) =>
                   (ba: List[nat]) => id_labeling[A, nat](ad, ba)),
                                      ac, b)
        case Inl((i, false)) =>
          (ac: List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])])
            => (b: tRSProof[A, List[nat], List[Char]]) =>
                 Semlab[A, List[nat],
                         List[Char]](i, Inr[List[Char],
     (lab[A, List[nat]]) =>
       (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]]](Nil),
                                      ac, b)
        case Inr(_) =>
          (ac: List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])])
            => (b: tRSProof[A, List[nat], List[Char]]) =>
                 Rootlab[A, List[nat], List[Char]](ac, b)
      },
    aa)),
  (List('b', 'o', 'u', 'n', 'd', 's'),
    (xml: xml) =>
      binda[List[Char], bounds_info[lab[A, List[nat]], List[Char]],
             tRSProof[A, List[nat],
                       List[Char]]]((bounds_info[lab[A,
              List[nat]]](xml2name)).apply(xml),
                                     comp[tRSProof[A, List[nat], List[Char]],
   sum[List[Char], tRSProof[A, List[nat], List[Char]]],
   bounds_info[lab[A, List[nat]],
                List[Char]]]((aa: tRSProof[A, List[nat], List[Char]]) =>
                               Inr[tRSProof[A, List[nat], List[Char]],
                                    List[Char]](aa),
                              (aa: bounds_info[lab[A, List[nat]], List[Char]])
                                => Bounds[A, List[nat], List[Char]](aa)))),
  (List('u', 'n', 'c', 'u', 'r', 'r', 'y'),
    (aa: xml) =>
      xml3elements[(lab[A, List[nat]],
                     (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
                       (List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]], List[Char]])]))),
                    List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])],
                    tRSProof[A, List[nat], List[Char]],
                    tRSProof[A, List[nat],
                              List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r',
         'y'),
    uncurry_info[lab[A, List[nat]]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, List[nat]]](xml2name),
        id[List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]],
        ab),
    (ab: xml) => xml2trsProof[A](xml2name, ab),
    (ab: (lab[A, List[nat]],
           (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
             (List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])],
               List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])]))))
      => (b: List[(term[lab[A, List[nat]], List[Char]],
                    term[lab[A, List[nat]], List[Char]])])
           => (c: tRSProof[A, List[nat], List[Char]]) =>
                Uncurry[A, List[nat], List[Char]](ab, b, c),
    aa)),
  (List('f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o',
         's', 'u', 'r', 'e'),
    (aa: xml) =>
      xml3elements[List[ctxt[lab[A, List[nat]], List[Char]]],
                    List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])],
                    tRSProof[A, List[nat], List[Char]],
                    tRSProof[A, List[nat],
                              List[Char]]](List('f', 'l', 'a', 't', 'C', 'o',
         'n', 't', 'e', 'x', 't', 'C', 'l', 'o', 's', 'u', 'r', 'e'),
    flat_contexts[lab[A, List[nat]]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, List[nat]]](xml2name),
        id[List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]],
        ab),
    (ab: xml) => xml2trsProof[A](xml2name, ab),
    (ab: List[ctxt[lab[A, List[nat]], List[Char]]]) =>
      (b: List[(term[lab[A, List[nat]], List[Char]],
                 term[lab[A, List[nat]], List[Char]])])
        => (c: tRSProof[A, List[nat], List[Char]]) =>
             Fcc[A, List[nat], List[Char]](ab, b, c),
    aa))),
                                     a),
                                       id[tRSProof[A, List[nat], List[Char]]],
                                       x)

implicit def preorder_char: preorder[Char] = new preorder[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

implicit def order_char: order[Char] = new order[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

abstract sealed class l_poly[+A, +B]
final case class LPoly[B, A](a: B, b: List[(A, B)]) extends l_poly[A, B]

def sym_collect[A, B](p: (term[A, B]) => Boolean, x1: term[A, B]): List[A] =
  (p, x1) match {
  case (p, Var(x)) => Nil
  case (p, Fun(f, ts)) =>
    (if (p(Fun[A, B](f, ts))) List(f) else Nil) ++
      maps[term[A, B], A]((a: term[A, B]) => sym_collect[A, B](p, a), ts)
}

def fresh_strings_list(name: List[Char], offset: nat, used: List[List[Char]],
                        n: nat):
      List[List[Char]] =
  take[List[Char]](n, filter[List[Char]]((s: List[Char]) =>
   ! (membera[List[Char]](remdups[List[Char]](used), s)),
  mapa[nat, List[Char]]((i: nat) =>
                          name ++
                            (shows_prec_nat(Zero_nat,
     plus_nat(i, offset))).apply(Nil),
                         upt(Zero_nat,
                              plus_nat(n,
size_list[List[Char]](remdups[List[Char]](used)))))))

def subst_of[A: equal, B](xts: List[(A, term[B, A])]): A => term[B, A] =
  (x: A) =>
    lookupa[A, term[B, A]](x, xts) match {
      case None => Var[A, B](x)
      case Some(t) => t
    }

def rm_update_dj[A: linorder, B]: A => B => (rbt[A, B]) => rbt[A, B] =
  rm_update[A, B]

def check_sem_lab_rel_tt[A: show, B: show,
                          C: show](valid: sum[(List[Char]) => List[Char], Unit],
                                    check_decr:
                                      (List[(term[lab[A, B], C],
      term[lab[A, B], C])]) =>
sum[(List[Char]) => List[Char], Unit],
                                    check_lab:
                                      (List[(term[lab[A, B], C],
      term[lab[A, B], C])]) =>
(List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
  sum[(List[Char]) => List[Char], Unit],
                                    check_model_lab:
                                      (List[(term[lab[A, B], C],
      term[lab[A, B], C])]) =>
(List[(term[lab[A, B], C], term[lab[A, B], C])]) =>
  sum[(List[Char]) => List[Char], Unit],
                                    ls: (lab[A, B]) =>
  nat => (sum[B, List[lab[A, B]]]) => Boolean,
                                    rs: (List[(term[lab[A, B], C],
        term[lab[A, B], C])],
  List[(term[lab[A, B], C], term[lab[A, B], C])]),
                                    lRS: (List[(term[lab[A, B], C],
         term[lab[A, B], C])],
   List[(term[lab[A, B], C], term[lab[A, B], C])])):
      sum[(List[Char]) => List[Char], Unit] =
  binda[(List[Char]) => List[Char], Unit,
         Unit](valid,
                (_: Unit) =>
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](binda[(List[Char]) => List[Char], Unit,
       Unit](check_decr(snd[List[(term[lab[A, B], C], term[lab[A, B], C])],
                             List[(term[lab[A, B], C],
                                    term[lab[A, B], C])]](lRS)),
              (_: Unit) =>
                binda[(List[Char]) => List[Char], Unit,
                       Unit]((check_model_lab(fst[List[(term[lab[A, B], C],
                 term[lab[A, B], C])],
           List[(term[lab[A, B], C],
                  term[lab[A, B],
                        C])]](lRS)))(fst[List[(term[lab[A, B], C],
        term[lab[A, B], C])],
  List[(term[lab[A, B], C], term[lab[A, B], C])]](rs)),
                              (_: Unit) =>
                                (check_model_lab(snd[List[(term[lab[A, B], C],
                    term[lab[A, B], C])],
              List[(term[lab[A, B], C],
                     term[lab[A, B],
                           C])]](lRS)))(snd[List[(term[lab[A, B], C],
           term[lab[A, B], C])],
     List[(term[lab[A, B], C], term[lab[A, B], C])]](rs)))),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
                 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'l', 'a', 'b', 'e',
                 'l', 'e', 'd', ' ', 'T', 'R', 'S', ':')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_nl, x)))))

def rl_rel_tt[A: equal: show,
               B: equal: show](rs: (List[(term[lab[A, B], List[Char]],
   term[lab[A, B], List[Char]])],
                                     List[(term[lab[A, B], List[Char]],
    term[lab[A, B], List[Char]])]),
                                lRS: (List[(term[lab[A, B], List[Char]],
     term[lab[A, B], List[Char]])],
                                       List[(term[lab[A, B], List[Char]],
      term[lab[A, B], List[Char]])])):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val fs: List[lab[A, B]] =
      funs_trs_list[lab[A, B],
                     List[Char]](fst[List[(term[lab[A, B], List[Char]],
    term[lab[A, B], List[Char]])],
                                      List[(term[lab[A, B], List[Char]],
     term[lab[A, B], List[Char]])]](rs) ++
                                   snd[List[(term[lab[A, B], List[Char]],
      term[lab[A, B], List[Char]])],
List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]](rs));
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (nulla[lab[A, B]](fs)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('r', 'o', 'o', 't', '-', 'l', 'a', 'b',
 'e', 'l', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'a',
 't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ', 'f', 'u', 'n', 'c',
 't', 'i', 'o', 'n', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'i', 'n', ' ', 't',
 'h', 'e', ' ', 's', 'i', 'g', 'n', 'a', 't', 'u', 'r', 'e')),
                 shows_nl)),
                  (_: Unit) =>
                    {
                      val d: lab[A, B] = hd[lab[A, B]](fs);
                      check_sem_lab_rel_tt[A, B,
    List[Char]](model_check_valid[(List[Char]) => List[Char]],
                 model_check_decr[List[(term[lab[A, B], List[Char]],
 term[lab[A, B], List[Char]])],
                                   (List[Char]) => List[Char]],
                 (a: List[(term[lab[A, B], List[Char]],
                            term[lab[A, B], List[Char]])])
                   => (b: List[(term[lab[A, B], List[Char]],
                                 term[lab[A, B], List[Char]])])
                        => check_sl_lab_trs[A, B, lab[A, B],
     List[Char]]((aa: lab[A, B]) =>
                   (ba: List[lab[A, B]]) =>
                     rl_inter_list[lab[A, B], List[lab[A, B]]](fs, d, aa, ba),
                  (aa: lab[A, B]) =>
                    (ba: List[lab[A, B]]) =>
                      rl_fun[lab[A, B], List[lab[A, B]], B](aa, ba),
                  fs, model_cge[lab[A, B]], a, b),
                 (a: List[(term[lab[A, B], List[Char]],
                            term[lab[A, B], List[Char]])])
                   => (b: List[(term[lab[A, B], List[Char]],
                                 term[lab[A, B], List[Char]])])
                        => check_sl_model_lab_trs[A, B, lab[A, B],
           List[Char]]((aa: lab[A, B]) =>
                         (ba: List[lab[A, B]]) =>
                           rl_inter_list[lab[A, B],
  List[lab[A, B]]](fs, d, aa, ba),
                        (aa: lab[A, B]) =>
                          (ba: List[lab[A, B]]) =>
                            rl_fun[lab[A, B], List[lab[A, B]], B](aa, ba),
                        fs, model_cge[lab[A, B]], a, b),
                 model_LS[lab[A, B], nat, sum[B, List[lab[A, B]]]], rs, lRS)
                    })
  }

def rp_desc[A, B](x0: (sum[(List[Char]) => List[Char], Unit],
                        (((term[A, B], term[A, B])) =>
                           sum[(List[Char]) => List[Char], Unit],
                          (((term[A, B], term[A, B])) =>
                             sum[(List[Char]) => List[Char], Unit],
                            (A => nat => nat => Boolean,
                              (sum[(List[Char]) => List[Char], Unit],
                                (List[Char]) => List[Char])))))):
      (List[Char]) => List[Char] =
  x0 match {
  case (v, (s, (n, (a, (m, d))))) => d
}

def rp_mono[A, B](x0: (sum[(List[Char]) => List[Char], Unit],
                        (((term[A, B], term[A, B])) =>
                           sum[(List[Char]) => List[Char], Unit],
                          (((term[A, B], term[A, B])) =>
                             sum[(List[Char]) => List[Char], Unit],
                            (A => nat => nat => Boolean,
                              (sum[(List[Char]) => List[Char], Unit],
                                (List[Char]) => List[Char])))))):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case (v, (s, (n, (a, (m, d))))) => m
}

def rsym_impl[A, B](x0: ta_rule_impl[A, B]): (B, nat) = x0 match {
  case TA_rule_impl(f, qsa, q, qs) => (f, size_list[A](qsa))
}

def computeRuleMapH[A: equal, B](x0: List[(term[A, B], term[A, B])]):
      Option[List[((A, nat), List[(term[A, B], term[A, B])])]] =
  x0 match {
  case Nil => Some[List[((A, nat), List[(term[A, B], term[A, B])])]](Nil)
  case (Fun(f, ts), r) :: rules =>
    {
      val n: nat = size_list[term[A, B]](ts);
      computeRuleMapH[A, B](rules) match {
        case None => None
        case Some(rm) =>
          extract[((A, nat),
                    List[(term[A, B],
                           term[A, B])])]((a:
     ((A, nat), List[(term[A, B], term[A, B])]))
    => {
         val (fa, _): ((A, nat), List[(term[A, B], term[A, B])]) = a;
         equal_proda[A, nat](fa, (f, n))
       },
   rm)
            match {
            case None =>
              Some[List[((A, nat),
                          List[(term[A, B],
                                 term[A, B])])]](((f, n),
           List((Fun[A, B](f, ts), r))) ::
           rm)
            case Some((bef, ((fa, rls), aft))) =>
              Some[List[((A, nat),
                          List[(term[A, B],
                                 term[A, B])])]]((fa,
           (Fun[A, B](f, ts), r) :: rls) ::
           bef ++ aft)
          }
      }
    }
  case (Var(uu), uv) :: rules => None
}

def computeRuleMap[A: equal, B](rls: List[(term[A, B], term[A, B])]):
      Option[((A, nat)) => List[(term[A, B], term[A, B])]] =
  computeRuleMapH[A, B](rls) match {
    case None => None
    case Some(rm) =>
      Some[((A, nat)) =>
             List[(term[A, B],
                    term[A, B])]]((f: (A, nat)) =>
                                    lookupa[(A, nat),
     List[(term[A, B], term[A, B])]](f, rm)
                                      match {
                                      case None => Nil
                                      case Some(rlsa) => rlsa
                                    })
  }

def apply_af_entry[A, B](uu: A, x1: af_entry, ts: List[term[A, B]]):
      term[A, B] =
  (uu, x1, ts) match {
  case (uu, Collapse(i), ts) => nth[term[A, B]](ts, i)
  case (f, AFList(is), ts) =>
    Fun[A, B](f, mapa[nat, term[A, B]]((a: nat) => nth[term[A, B]](ts, a), is))
}

def afs_term[A, B](pi: A => nat => af_entry, x1: term[A, B]):
      term[filtered[A], B] =
  (pi, x1) match {
  case (pi, Fun(f, ts)) =>
    {
      val l: nat = size_list[term[A, B]](ts);
      apply_af_entry[filtered[A],
                      B](FPair[A](f, l), (pi(f))(l),
                          mapa[term[A, B],
                                term[filtered[A],
                                      B]]((a: term[A, B]) =>
    afs_term[A, B](pi, a),
   ts))
    }
  case (pi, Var(x)) => Var[B, filtered[A]](x)
}

def search_len(t: List[Char], s: List[Char], n: nat):
      Option[(List[Char], List[Char])] =
  (if (equal_lista[Char](take[Char](n, t), s))
    Some[(List[Char], List[Char])]((Nil, drop[Char](n, t)))
    else t match {
           case Nil => None
           case h :: tt =>
             search_len(tt, s, n) match {
               case None => None
               case Some((bef, aft)) =>
                 Some[(List[Char], List[Char])]((h :: bef, aft))
             }
         })

def reverse_trs[A, B](r: List[(term[A, B], term[A, B])]):
      List[(term[A, B], term[A, B])] =
  mapa[(term[A, B], term[A, B]),
        (term[A, B],
          term[A, B])]((lr: (term[A, B], term[A, B])) =>
                         (snd[term[A, B], term[A, B]](lr),
                           fst[term[A, B], term[A, B]](lr)),
                        r)

def check_graph_decomp[A](ss: A => (List[Char]) => List[Char],
                           g: ((A, A)) => Boolean,
                           rcs: List[(Boolean, List[A])]):
      sum[(List[Char]) => List[Char], Unit] =
  binda[(List[Char]) => List[Char], Unit,
         Unit](check_pairwise[List[A],
                               (List[Char]) =>
                                 List[Char]]((a: List[A]) =>
       (b: List[A]) => check_edges[A](ss, g, a, b),
      mapa[(Boolean, List[A]),
            List[A]]((a: (Boolean, List[A])) => snd[Boolean, List[A]](a), rcs)),
                (_: Unit) =>
                  catcha[(List[A], (List[Char]) => List[Char]), Unit,
                          (List[Char]) =>
                            List[Char]](forallm[List[A],
         (List[Char]) =>
           List[Char]]((c: List[A]) => check_edges[A](ss, g, c, c),
                        map_filter[(Boolean, List[A]),
                                    List[A]]((x: (Boolean, List[A])) =>
       (if (! (fst[Boolean, List[A]](x)))
         Some[List[A]](snd[Boolean, List[A]](x)) else None),
      rcs)),
 (x: (List[A], (List[Char]) => List[Char])) =>
   Inl[(List[Char]) => List[Char],
        Unit](snd[List[A], (List[Char]) => List[Char]](x))))

def add_varI[A: equal,
              B: equal](uu: A, uv: A => A => A, x: B, a: A, xa4: List[(B, A)]):
      List[(B, A)] =
  (uu, uv, x, a, xa4) match {
  case (uu, uv, x, a, Nil) => List((x, a))
  case (ze, pl, x, a, (y, b) :: vas) =>
    (if (eq[B](x, y))
      {
        val s: A = (pl(a))(b);
        (if (eq[A](s, ze)) vas else (x, s) :: vas)
      }
      else (y, b) :: add_varI[A, B](ze, pl, x, a, vas))
}

def sum_pvarsI[A: equal,
                B: equal](uu: A, uv: A => A => A, x2: List[(B, A)],
                           vbs: List[(B, A)]):
      List[(B, A)] =
  (uu, uv, x2, vbs) match {
  case (uu, uv, Nil, vbs) => vbs
  case (ze, pl, (x, a) :: vas, vbs) =>
    (if (eq[A](a, ze)) sum_pvarsI[A, B](ze, pl, vas, vbs)
      else sum_pvarsI[A, B](ze, pl, vas, add_varI[A, B](ze, pl, x, a, vbs)))
}

def sum_lpolyI[A: equal,
                B: equal](ze: A, pl: A => A => A, x2: l_poly[B, A],
                           x3: l_poly[B, A]):
      l_poly[B, A] =
  (ze, pl, x2, x3) match {
  case (ze, pl, LPoly(a, vas), LPoly(b, vbs)) =>
    LPoly[A, B]((pl(a))(b), sum_pvarsI[A, B](ze, pl, vas, vbs))
}

def list_sumI[A](ze: A, pl: A => A => A, x2: List[A]): A = (ze, pl, x2) match {
  case (ze, pl, Nil) => ze
  case (ze, pl, x :: xs) => (pl(x))(list_sumI[A](ze, pl, xs))
}

def mul_pvarsI[A: equal, B](uu: A, uv: A => A => A, a: A, x3: List[(B, A)]):
      List[(B, A)] =
  (uu, uv, a, x3) match {
  case (uu, uv, a, Nil) => Nil
  case (ze, ti, a, (x, b) :: vas) =>
    {
      val p: A = (ti(a))(b)
      val res: List[(B, A)] = mul_pvarsI[A, B](ze, ti, a, vas);
      (if (eq[A](p, ze)) res else (x, p) :: res)
    }
}

def mul_lpolyI[A: equal, B](ze: A, ti: A => A => A, a: A, x3: l_poly[B, A]):
      l_poly[B, A] =
  (ze, ti, a, x3) match {
  case (ze, ti, a, LPoly(b, vas)) =>
    LPoly[A, B]((ti(a))(b), mul_pvarsI[A, B](ze, ti, a, vas))
}

def pleftI[A: equal, B,
            C: equal](ze: A, on: A, pl: A => A => A, ti: A => A => A,
                       mz: A => A, pi: B => nat => (A, List[A]),
                       x6: term[B, C]):
      l_poly[C, A] =
  (ze, on, pl, ti, mz, pi, x6) match {
  case (ze, on, pl, ti, mz, pi, Var(x)) => LPoly[A, C](ze, List((x, on)))
  case (ze, on, pl, ti, mz, pi, Fun(f, ts)) =>
    {
      val (c, as): (A, List[A]) = (pi(f))(size_list[term[B, C]](ts));
      sum_lpolyI[A, C](ze, pl, LPoly[A, C](c, Nil),
                        list_sumI[l_poly[C,
  A]](LPoly[A, C](ze, Nil),
       (a: l_poly[C, A]) => (b: l_poly[C, A]) => sum_lpolyI[A, C](ze, pl, a, b),
       mapa[(A, term[B, C]),
             l_poly[C, A]]((at: (A, term[B, C])) =>
                             mul_lpolyI[A,
 C](ze, ti, fst[A, term[B, C]](at),
     pleftI[A, B, C](ze, on, pl, ti, mz, pi, snd[A, term[B, C]](at))),
                            zip[A, term[B, C]](as, ts))))
        match {
        case LPoly(d, Nil) => LPoly[A, C](mz(d), Nil)
        case LPoly(d, ab :: lista) => LPoly[A, C](d, ab :: lista)
      }
    }
}

implicit def linorder_list[A: equal: linorder]: linorder[List[A]] = new
  linorder[List[A]] {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

def check_DPs[A: equal: show,
               B: equal: show](unshp: A => A, r: List[(term[A, B], term[A, B])],
                                p: List[(term[A, B], term[A, B])]):
      sum[(term[A, B], term[A, B]), Unit] =
  {
    val d: List[(A, nat)] = defs_list[A, B](r);
    catcha[((term[A, B], term[A, B]), (term[A, B], term[A, B])), Unit,
            (term[A, B],
              term[A, B])](forallm[(term[A, B], term[A, B]),
                                    (term[A, B],
                                      term[A,
    B])]((x: (term[A, B], term[A, B])) =>
           (if ({
                  val (ll, rr): (term[A, B], term[A, B]) = x;
                  ! (is_var[A, B](rr)) &&
                    (! (membera[(A, nat)](d,
   (the[A](root[A, B](rr)), num_args[A, B](rr)))) &&
                      list_ex[(term[A, B],
                                term[A, B])]((a: (term[A, B], term[A, B])) =>
       {
         val (l, ra): (term[A, B], term[A, B]) = a;
         equal_terma[A, B](l, sharp_term[A, B](unshp, ll)) &&
           member[(term[A, B],
                    term[A, B])]((ra, sharp_term[A, B](unshp, rr)),
                                  supteq[A, B])
       },
      r))
                })
             Inr[Unit, (term[A, B], term[A, B])](())
             else Inl[(term[A, B], term[A, B]), Unit](x)),
          p),
                            (x: ((term[A, B], term[A, B]),
                                  (term[A, B], term[A, B])))
                              => Inl[(term[A, B], term[A, B]),
                                      Unit](snd[(term[A, B], term[A, B]),
         (term[A, B], term[A, B])](x)))
  }

def check_poly_ge[A: equal,
                   B: ordered_semiring_0](x0: List[(List[(A, nat)], B)],
   q: List[(List[(A, nat)], B)]):
      Boolean =
  (x0, q) match {
  case (Nil, q) =>
    list_all[(List[(A, nat)],
               B)]((a: (List[(A, nat)], B)) =>
                     {
                       val (_, aa): (List[(A, nat)], B) = a;
                       ge[B](zero[B], aa)
                     },
                    q)
  case ((m, c) :: p, q) =>
    extract[(List[(A, nat)],
              B)]((nd: (List[(A, nat)], B)) =>
                    eq_monom[A](fst[List[(A, nat)], B](nd), m),
                   q)
      match {
      case None => ge[B](c, zero[B]) && check_poly_ge[A, B](p, q)
      case Some((q1, ((_, d), q2))) =>
        ge[B](c, d) && check_poly_ge[A, B](p, q1 ++ q2)
    }
}

def check_poly_gt[A: ordered_semiring_0,
                   B: equal](gt: A => A => Boolean,
                              p: List[(List[(B, nat)], A)],
                              q: List[(List[(B, nat)], A)]):
      Boolean =
  {
    val (a1, p1): (A, List[(List[(B, nat)], A)]) = poly_split[B, A](Nil, p)
    val (b1, q1): (A, List[(List[(B, nat)], A)]) = poly_split[B, A](Nil, q);
    (gt(a1))(b1) && check_poly_ge[B, A](p1, q1)
  }

def rev_term[A, B](c: ctxt[A, B], x1: term[A, B]): term[A, B] = (c, x1) match {
  case (c, Var(x)) => ctxt_apply[A, B](c, Var[B, A](x))
  case (c, Fun(f, List(t))) => rev_term[A, B](More[A, B](f, Nil, c, Nil), t)
}

def rev_rule[A, B](x0: (term[A, B], term[A, B])): (term[A, B], term[A, B]) = x0
  match {
  case (l, r) => (rev_term[A, B](Hole, l), rev_term[A, B](Hole, r))
}

def rp_valid[A, B](x0: (sum[(List[Char]) => List[Char], Unit],
                         (((term[A, B], term[A, B])) =>
                            sum[(List[Char]) => List[Char], Unit],
                           (((term[A, B], term[A, B])) =>
                              sum[(List[Char]) => List[Char], Unit],
                             (A => nat => nat => Boolean,
                               (sum[(List[Char]) => List[Char], Unit],
                                 (List[Char]) => List[Char])))))):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case (v, (s, (n, (a, (m, d))))) => v
}

abstract sealed class ta_impl_ext[+A, +B, +C]
final case class
  Ta_impl_ext[A, B,
               C](a: rbt[A, Unit], b: rbt[(B, nat), List[ta_rule_impl[A, B]]],
                   c: List[A], d: rbt[A, Unit], e: A => rbt[A, Unit], f: C)
  extends ta_impl_ext[A, B, C]

def ta_rqs_set[A, B, C](x0: ta_impl_ext[A, B, C]): rbt[A, Unit] = x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, more)
    => ta_rqs_set
}

def check_no_defined_root[A: show,
                           B: show](rm: ((A, nat)) =>
  List[(term[A, B], term[A, B])],
                                     t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](nulla[(term[A, B],
                              term[A, B])](rm((the[A](root[A, B](t)),
        num_args[A, B](t)))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_prec_term[A, B](Zero_nat, t),
                     shows_string.apply(List(' ', 'i', 's', ' ', 'd', 'e', 'f',
      'i', 'n', 'e', 'd')))))

def uncurry_of_top_sig_list[A](a: A, m: nat, sml: List[((A, nat), List[A])],
                                sm: A => nat => List[A]):
      List[(term[A, List[Char]], term[A, List[Char]])] =
  maps[((A, nat), List[A]),
        (term[A, List[Char]],
          term[A, List[Char]])]((b: ((A, nat), List[A])) =>
                                  {
                                    val (ba, c): ((A, nat), List[A]) = b;
                                    ({
                                       val (f, n): (A, nat) = ba;
                                       (_: List[A]) =>
 {
   val g: nat => A = (aa: nat) => get_symbol[A](sm, f, n, aa);
   mapa[nat, (term[A, List[Char]],
               term[A, List[Char]])]((i: nat) =>
                                       (Fun[A,
     List[Char]](a, generate_f_xs[A](g(i), plus_nat(n, i)) ::
                      mapa[nat, term[A, List[Char]]]((ia: nat) =>
               Var[List[Char], A](generate_var(ia)),
              upt(plus_nat(n, i),
                   plus_nat(plus_nat(n, i), minus_nat(m, one_nat))))),
 generate_f_xs[A](g(Suc(i)), plus_nat(plus_nat(n, i), minus_nat(m, one_nat)))),
                                      upt(Zero_nat, aarity[A](sm, f, n)))
 }
                                     })(c)
                                  },
                                 sml)

def eta_closed_top_rules[A: equal: show,
                          B: equal: show](a: A, n: nat, sm: A => nat => List[A],
   r: List[(term[A, B], term[A, B])], p: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, B], term[A, B]),
                                 (List[Char]) =>
                                   List[Char]]((b: (term[A, B], term[A, B])) =>
         {
           val (l, ra): (term[A, B], term[A, B]) = b;
           l match {
             case Var(_) => Inr[Unit, (List[Char]) => List[Char]](())
             case Fun(ff, ls) =>
               check[(List[Char]) =>
                       List[Char]](equal_nata(aarity[A](sm, ff,
                 size_list[term[A, B]](ls)),
       Zero_nat) ||
                                     list_ex[(term[A, B],
       term[A, B])]((ba: (term[A, B], term[A, B])) =>
                      {
                        val (lll, rrr): (term[A, B], term[A, B]) = ba;
                        (lll, rrr) match {
                          case (Var(_), bb) => false
                          case (Fun(f, Nil), bb) => false
                          case (Fun(f, ll :: yy), Var(_)) => false
                          case (Fun(f, ll :: yy), Fun(g, Nil)) => false
                          case (Fun(f, ll :: yy), Fun(g, rr :: zz)) =>
                            eq[A](f, a) &&
                              (eq[A](g, a) &&
                                (equal_lista[term[A, B]](zz, yy) &&
                                  (equal_nata(size_list[term[A, B]](yy),
       minus_nat(n, one_nat)) &&
                                    (distinct[term[A, B]](yy) &&
                                      (list_all[term[A,
              B]]((aa: term[A, B]) => is_var[A, B](aa), yy) &&
(nulla[B](list_inter[B](mapa[term[A, B],
                              B]((aa: term[A, B]) => the_var[A, B](aa), yy),
                         vars_rule_lista[A, B]((ll, rr)))) &&
  instance_rule[A, B]((l, ra), (ll, rr))))))))
                        }
                      },
                     p),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('e', 't', 'a', ' ', 'e', 'x', 'p',
   'a', 'n', 's', 'i', 'o', 'n', ' ', 'o', 'f', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_rule[A, B]((l, ra)),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))
           }
         },
        r),
                         (x: ((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](x)))

def uncurry_top_rules[A: equal, B](a: A, n: nat, sm: A => nat => List[A]):
      (List[(term[A, B], term[A, B])]) => List[(term[A, B], term[A, B])] =
  (b: List[(term[A, B], term[A, B])]) =>
    mapa[(term[A, B], term[A, B]),
          (term[A, B],
            term[A, B])]((ba: (term[A, B], term[A, B])) =>
                           {
                             val (l, r): (term[A, B], term[A, B]) = ba;
                             (uncurry_top[A, B](a, n, sm, l),
                               uncurry_top[A, B](a, n, sm, r))
                           },
                          b)

def uncurry_top_proc[A: equal: show](uv: (A,
   (List[((A, nat), List[A])],
     (List[(term[A, List[Char]], term[A, List[Char]])],
       List[(term[A, List[Char]], term[A, List[Char]])]))),
                                      n: nat,
                                      fmap:
(List[((A, nat), List[A])]) => A => nat => A,
                                      check_inj:
A => nat =>
       (List[((A, nat), List[A])]) => sum[(List[Char]) => List[Char], Unit],
                                      pa:
List[(term[A, List[Char]], term[A, List[Char]])],
                                      ra:
List[(term[A, List[Char]], term[A, List[Char]])],
                                      p:
List[(term[A, List[Char]], term[A, List[Char]])],
                                      r:
List[(term[A, List[Char]], term[A, List[Char]])],
                                      x8:
Option[((A, nat)) => List[(term[A, List[Char]], term[A, List[Char]])]]):
      sum[(List[Char]) => List[Char], Unit] =
  (uv, n, fmap, check_inj, pa, ra, p, r, x8) match {
  case ((a, (sml, (uu, e))), n, fmap, check_inj, pa, ra, p, r, Some(rm)) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (equal_nata(n, Zero_nat)),
                                      shows_prec_list[Char](Zero_nat,
                     List('t', 'h', 'e', ' ', 'a', 'r', 'i', 't', 'y', ' ', 'o',
                           'f', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c', 'u',
                           'r', 'r', 'i', 'e', 'd', ' ', 's', 'y', 'm', 'b',
                           'o', 'l', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e',
                           ' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't', ' ',
                           '1'))),
                  (_: Unit) =>
                    {
                      val sm: A => nat => List[A] =
                        sig_list_to_sig_map[A](a, sml, fmap)
                      val p_eta:
                            List[(term[A, List[Char]], term[A, List[Char]])]
                        = e ++ pa
                      val u: List[(term[A, List[Char]], term[A, List[Char]])] =
                        map_rules_wa[A, A,
                                      List[Char]]((f: A) =>
            (na: nat) => get_symbol[A](sm, f, na, Zero_nat),
           ra);
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](((check_inj(a))(n))(sml),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], Unit,
     Unit](catcha[((term[A, List[Char]], term[A, List[Char]]),
                    (List[Char]) => List[Char]),
                   Unit,
                   (List[Char]) =>
                     List[Char]](forallm[(term[A, List[Char]],
   term[A, List[Char]]),
  (List[Char]) =>
    List[Char]]((b: (term[A, List[Char]], term[A, List[Char]])) =>
                  {
                    val (l, rb): (term[A, List[Char]], term[A, List[Char]]) = b;
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](hvf_top[A, List[Char]](a, n, l),
              comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
                     'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_term[A, List[Char]](l),
            shows_prec_list[Char](Zero_nat,
                                   List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l',
 'o', 'w', 'e', 'd'))))),
                                  (_: Unit) =>
                                    binda[(List[Char]) => List[Char], Unit,
   Unit](check_no_var[A, List[Char]](rb),
          (_: Unit) => check_no_defined_root[A, List[Char]](rm, rb)))
                  },
                 p_eta),
                                  (x: ((term[A, List[Char]],
 term[A, List[Char]]),
(List[Char]) => List[Char]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](snd[(term[A, List[Char]], term[A, List[Char]]),
               (List[Char]) => List[Char]](x))),
            (_: Unit) =>
              binda[(List[Char]) => List[Char], Unit,
                     Unit](check[(List[Char]) =>
                                   List[Char]](nulla[(term[A, List[Char]],
               term[A, List[Char]])](rm((a, n))),
        comp[List[Char], List[Char],
              List[Char]](shows_prec_list[Char](Zero_nat,
         List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ', 's',
               'y', 'm', 'b', 'o', 'l', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]]((b: List[Char]) =>
       shows_prec[A](Zero_nat, a, b),
      shows_prec_list[Char](Zero_nat,
                             List(' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't',
                                   ' ', 'b', 'e', ' ', 'd', 'e', 'f', 'i', 'n',
                                   'e', 'd', ' ', 'i', 'n', ' ', 'R'))))),
                            (_: Unit) =>
                              binda[(List[Char]) => List[Char], Unit,
                                     Unit]((if (list_ex[(term[A, List[Char]],
                  term[A, List[Char]])]((b:
   (term[A, List[Char]], term[A, List[Char]]))
  => {
       val (_, rb): (term[A, List[Char]], term[A, List[Char]]) = b;
       eq[A](the[A](root[A, List[Char]](rb)), a) &&
         (equal_nata(num_args[A, List[Char]](rb), n) &&
           equal_hterma[A, List[Char]](capRM2[A,
       List[Char]](rm, hd[term[A, List[Char]]](args[A, List[Char]](rb))),
HHole))
     },
 pa))
     binda[(List[Char]) => List[Char], Unit,
            Unit](catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                          (List[Char]) =>
                            List[Char]](check_CS_subseteq[A,
                   List[Char]](uncurry_of_top_sig_list[A](a, n, sml, sm), p),
 (x: (term[A, List[Char]], term[A, List[Char]])) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e',
                     ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_rule[A, List[Char]](x),
            shows_prec_list[Char](Zero_nat,
                                   List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
 'i', 'n', 'g', ' ', 'i', 'n', ' ', 'P')))))),
                   (_: Unit) =>
                     eta_closed_top_rules[A, List[Char]](a, n, sm, ra, p_eta))
     else Inr[Unit, (List[Char]) => List[Char]](())),
    (_: Unit) =>
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[(term[A, List[Char]], term[A, List[Char]]), Unit,
                           (List[Char]) =>
                             List[Char]](check_subseteq[(term[A, List[Char]],
                  term[A, List[Char]])]((uncurry_top_rules[A,
                    List[Char]](a, n, sm)).apply(p_eta),
 p),
  (x: (term[A, List[Char]], term[A, List[Char]])) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'p', 'a',
                      'i', 'r', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_rule[A, List[Char]](x),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
  'i', 'n', 'g')))))),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](catcha[(term[A, List[Char]],
    term[A, List[Char]]),
   Unit,
   (List[Char]) =>
     List[Char]](check_subseteq[(term[A, List[Char]],
                                  term[A, List[Char]])](u, r),
                  (x: (term[A, List[Char]], term[A, List[Char]])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](comp[List[Char], List[Char],
                                     List[Char]](shows_prec_list[Char](Zero_nat,
                                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e',
                                      'd', ' ', 'r', 'u', 'l', 'e', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_rule[A, List[Char]](x),
                             shows_prec_list[Char](Zero_nat,
            List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))),
                                    (_: Unit) =>
                                      catcha[(term[A, List[Char]],
       term[A, List[Char]]),
      Unit,
      (List[Char]) =>
        List[Char]](check_subseteq[(term[A, List[Char]],
                                     term[A, List[Char]])](r, u),
                     (x: (term[A, List[Char]], term[A, List[Char]])) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](shows_prec_list[Char](Zero_nat,
                                   List('a', 'd', 'd', 'i', 't', 'i', 'o', 'n',
 'a', 'l', ' ', 'r', 'u', 'l', 'e', ' ')),
             comp[List[Char], List[Char],
                   List[Char]](shows_rule[A, List[Char]](x),
                                shows_prec_list[Char](Zero_nat,
               List(' ', 'i', 'n', ' ', 'R', ' ', 'i', 's', ' ', 'n', 'o', 't',
                     ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd'))))))))))))
                    })
  case (uv, uw, ux, uy, uz, va, vb, vc, None) =>
    Inr[Unit, (List[Char]) => List[Char]](())
}

def doc_of_string(s: List[Char]): sum[List[Char], xmldoc] =
  binda[List[Char], (xmldoc, List[Char]),
         xmldoc](parse_doc.apply(s),
                  (a: (xmldoc, List[Char])) =>
                    {
                      val (doc, _): (xmldoc, List[Char]) = a;
                      Inr[xmldoc, List[Char]](doc)
                    })

def filtered_fun[A](x0: filtered[A]): A = x0 match {
  case FPair(f, n) => f
}

def shows_prec_filtered[A: show](d: nat, ff: filtered[A]):
      (List[Char]) => List[Char] =
  (a: List[Char]) => shows_prec[A](Zero_nat, filtered_fun[A](ff), a)

def shows_list_filtered[A: show](ffs: List[filtered[A]]):
      (List[Char]) => List[Char] =
  shows_list_aux[filtered[A]]((a: filtered[A]) =>
                                shows_prec_filtered[A](Zero_nat, a),
                               ffs)

implicit def show_filtered[A: show]: show[filtered[A]] = new show[filtered[A]] {
  val `Ceta.shows_prec` = (a: nat, b: filtered[A], c: List[Char]) =>
    (shows_prec_filtered[A](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[filtered[A]], b: List[Char]) =>
    (shows_list_filtered[A](a)).apply(b)
}

def afs_check[A: show,
               B: show](r: (List[Char]) => List[Char], pi: A => nat => af_entry,
                         g: ((term[filtered[A], B], term[filtered[A], B])) =>
                              sum[(List[Char]) => List[Char], Unit],
                         lr: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val pl: term[filtered[A], B] =
      afs_term[A, B](pi, fst[term[A, B], term[A, B]](lr))
    val pr: term[filtered[A], B] =
      afs_term[A, B](pi, snd[term[A, B], term[A, B]](lr));
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]](g((pl, pr)),
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
   'o', 't', ' ', 'o', 'r', 'i', 'e', 'n', 't', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_prec_term[A,
              B](Zero_nat, fst[term[A, B], term[A, B]](lr)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ')),
                 comp[List[Char], List[Char],
                       List[Char]](r, comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_prec_term[A,
            B](Zero_nat, snd[term[A, B], term[A, B]](lr)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl,
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('p', 'i', '(', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_prec_term[A, B](Zero_nat, fst[term[A, B], term[A, B]](lr)),
             comp[List[Char], List[Char],
                   List[Char]](shows_string.apply(List(' ', ')', ' ', '=',
                ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_term[filtered[A],
                           B](Zero_nat, pl),
           comp[List[Char], List[Char],
                 List[Char]](shows_nl,
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('p',
                                 'i', '(', ' ')),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_term[A,
    B](Zero_nat, snd[term[A, B], term[A, B]](lr)),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_string.apply(List(' ', ')',
                               ' ', '=', ' ')),
       comp[List[Char], List[Char],
             List[Char]](shows_prec_term[filtered[A], B](Zero_nat, pr),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl, x)))))))))))))))))))
  }

def shows_afs[A: show](af: List[((A, nat), af_entry)]):
      (List[Char]) => List[Char] =
  shows_map[((A, nat),
              af_entry)]((a: ((A, nat), af_entry)) =>
                           {
                             val (aa, b): ((A, nat), af_entry) = a;
                             ({
                                val (f, n): (A, nat) = aa;
                                (e: af_entry) =>
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('p', 'i', '(')),
             comp[List[Char], List[Char],
                   List[Char]]((ab: List[Char]) =>
                                 shows_prec[A](Zero_nat, f, ab),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('/')),
           comp[List[Char], List[Char],
                 List[Char]](shows_prec_nat(Zero_nat, n),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List(')',
                                 ' ', '=', ' ')),
         comp[List[Char], List[Char],
               List[Char]](e match {
                             case Collapse(i) =>
                               shows_prec_nat(Zero_nat, Suc(i))
                             case AFList(ids) =>
                               shows_prec_list[nat](Zero_nat,
             mapa[nat, nat]((ab: nat) => Suc(ab), ids))
                           },
                            shows_nl))))))
                              })(b)
                           },
                          af)

def xml2dpDisproof[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], dPDisproof[A, List[Char]]] =
  xml1element[dPDisproof[A, List[Char]],
               dPDisproof[A, List[Char]]](List('d', 'p', 'N', 'o', 'n', 't',
        'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
        'f'),
   (a: xml) =>
     xml_options[dPDisproof[A, List[Char]]](List((List('d', 'p', 'R', 'u', 'l',
                'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
           (aa: xml) =>
             xml3elements[List[(term[A, List[Char]], term[A, List[Char]])],
                           List[(term[A, List[Char]], term[A, List[Char]])],
                           dPDisproof[A, List[Char]],
                           dPDisproof[A, List[Char]]](List('d', 'p', 'R', 'u',
                    'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
               (ab: xml) =>
                 xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                              List[(term[A, List[Char]],
                                     term[A,
   List[Char]])]](List('d', 'p', 's'), rules[A](xml2name),
                   id[List[(term[A, List[Char]], term[A, List[Char]])]], ab),
               (ab: xml) =>
                 xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                              List[(term[A, List[Char]],
                                     term[A,
   List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                   id[List[(term[A, List[Char]], term[A, List[Char]])]], ab),
               (ab: xml) => xml2dpDisproof[A](xml2name, ab),
               (ab: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                 (b: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                   (c: dPDisproof[A, List[Char]]) =>
                     DPRuleRemoval[A, List[Char]](ab, b, c),
               aa)),
          (List('l', 'o', 'o', 'p'),
            (aa: xml) =>
              xml_change[(term[A, List[Char]],
                           (List[(pos, ((term[A, List[Char]],
  term[A, List[Char]]),
 (Boolean, term[A, List[Char]])))],
                             (List[(List[Char], term[A, List[Char]])],
                               ctxt[A, List[Char]]))),
                          dPDisproof[A, List[Char]]](loop[A](xml2name),
              (ab: (term[A, List[Char]],
                     (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                   (Boolean, term[A, List[Char]])))],
                       (List[(List[Char], term[A, List[Char]])],
                         ctxt[A, List[Char]]))))
                => {
                     val (s, (rseq, (ac, b))):
                           (term[A, List[Char]],
                             (List[(pos, ((term[A, List[Char]],
    term[A, List[Char]]),
   (Boolean, term[A, List[Char]])))],
                               (List[(List[Char], term[A, List[Char]])],
                                 ctxt[A, List[Char]])))
                       = ab;
                     DPLoop[A, List[Char]](s, rseq, ac, b)
                   },
              aa))),
     a),
   id[dPDisproof[A, List[Char]]], x)

def starts_with(t: List[Char], s: List[Char]): Boolean =
  (if (! (nulla[Char](t)) && membera[Char](wspace, hd[Char](t)))
    starts_with(tl[Char](t), s)
    else search_len(t, s, size_list[Char](s)) match {
           case None => false
           case Some(_) => true
         })

def forallm_index_aux[A, B](p: A => nat => sum[B, Unit], i: nat, x2: List[A]):
      sum[((A, nat), B), Unit] =
  (p, i, x2) match {
  case (p, i, Nil) => Inr[Unit, ((A, nat), B)](())
  case (p, i, x :: xs) =>
    binda[((A, nat), B), Unit,
           Unit](catcha[B, Unit,
                         ((A, nat),
                           B)]((p(x))(i),
                                (xa: B) =>
                                  Inl[((A, nat), B), Unit](((x, i), xa))),
                  (_: Unit) => forallm_index_aux[A, B](p, Suc(i), xs))
}

def forallm_index[A, B](p: A => nat => sum[B, Unit], xs: List[A]):
      sum[((A, nat), B), Unit] =
  forallm_index_aux[A, B](p, Zero_nat, xs)

implicit def
  linorder_lab[A: equal: linorder, B: equal: linorder]: linorder[lab[A, B]] =
  new linorder[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

def prightI[A: equal, B,
             C: equal](ze: A, on: A, pl: A => A => A, ti: A => A => A,
                        mz: A => A, pi: B => nat => (A, List[A]),
                        x6: term[B, C]):
      l_poly[C, A] =
  (ze, on, pl, ti, mz, pi, x6) match {
  case (ze, on, pl, ti, mz, pi, Var(x)) => LPoly[A, C](ze, List((x, on)))
  case (ze, on, pl, ti, mz, pi, Fun(f, ts)) =>
    {
      val (c, as): (A, List[A]) = (pi(f))(size_list[term[B, C]](ts))
      val (LPoly(d, a)): l_poly[C, A] =
        sum_lpolyI[A, C](ze, pl, LPoly[A, C](c, Nil),
                          list_sumI[l_poly[C,
    A]](LPoly[A, C](ze, Nil),
         (a: l_poly[C, A]) =>
           (b: l_poly[C, A]) => sum_lpolyI[A, C](ze, pl, a, b),
         mapa[(A, term[B, C]),
               l_poly[C, A]]((at: (A, term[B, C])) =>
                               mul_lpolyI[A,
   C](ze, ti, fst[A, term[B, C]](at),
       prightI[A, B, C](ze, on, pl, ti, mz, pi, snd[A, term[B, C]](at))),
                              zip[A, term[B, C]](as, ts))));
      LPoly[A, C](mz(d), a)
    }
}

def check_loop[A: equal: show,
                B: equal: show](s: term[A, B],
                                 rseq: List[(pos,
      ((term[A, B], term[A, B]), term[A, B]))],
                                 sigma: List[(B, term[A, B])], c: ctxt[A, B],
                                 r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  binda[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](! (nulla[(pos,
      ((term[A, B], term[A, B]), term[A, B]))](rseq)),
                                    shows_prec_list[Char](Zero_nat,
                   List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                         'u', 'e', 'n', 'c', 'e', ' ', 'm', 'u', 's', 't', ' ',
                         'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm', 'p', 't',
                         'y'))),
                (_: Unit) =>
                  check_rsteps[A, B](r, rseq, s,
                                      ctxt_apply[A,
          B](c, subst_apply_term[A, B, B](s, subst_of[B, A](sigma)))))

def poly_vars_list[A: equal, B](p: List[(List[(A, nat)], B)]): List[A] =
  remdups[A](maps[(List[(A, nat)], B),
                   A](comp[List[(A, nat)], List[A],
                            (List[(A, nat)],
                              B)]((a: List[(A, nat)]) =>
                                    mapa[(A, nat),
  A]((aa: (A, nat)) => fst[A, nat](aa), a),
                                   (a: (List[(A, nat)], B)) =>
                                     fst[List[(A, nat)], B](a)),
                       p))

def funas_term_list[A: equal, B]: (term[A, B]) => List[(A, nat)] =
  comp[List[(A, nat)], List[(A, nat)],
        term[A, B]]((a: List[(A, nat)]) => remdups[(A, nat)](a),
                     (a: term[A, B]) => funas_term_lista[A, B](a))

def rp_strict[A, B](x0: (sum[(List[Char]) => List[Char], Unit],
                          (((term[A, B], term[A, B])) =>
                             sum[(List[Char]) => List[Char], Unit],
                            (((term[A, B], term[A, B])) =>
                               sum[(List[Char]) => List[Char], Unit],
                              (A => nat => nat => Boolean,
                                (sum[(List[Char]) => List[Char], Unit],
                                  (List[Char]) => List[Char])))))):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case (v, (s, (n, (a, (m, d))))) => s
}

def rel_pow_impl[A, B](succ: (List[A]) => List[A], un: (List[A]) => B => B,
                        memb: A => B => Boolean, newa: List[A], have: B,
                        x5: nat):
      B =
  (succ, un, memb, newa, have, x5) match {
  case (succ, un, memb, newa, have, Zero_nat) => (un(newa))(have)
  case (succ, un, memb, newa, have, Suc(m)) =>
    (if (nulla[A](newa)) have
      else {
             val maybe: List[A] = succ(newa)
             val havea: B = (un(newa))(have)
             val newb: List[A] =
               filter[A]((n: A) => ! ((memb(n))(havea)), maybe);
             rel_pow_impl[A, B](succ, un, memb, newb, havea, m)
           })
}

def rtrancl_impl[A, B](gen_succ: (List[(A, A)]) => (List[A]) => List[A],
                        un: (List[A]) => B => B, memb: A => B => Boolean,
                        emp: B, rel: List[(A, A)]):
      (List[A]) => B =
  {
    val succ: (List[A]) => List[A] = gen_succ(rel)
    val n: nat = size_list[(A, A)](rel);
    (as: List[A]) => rel_pow_impl[A, B](succ, un, memb, as, emp, n)
  }

def elem_list_to_rm[A, B: linorder](key: A => B, x1: List[A]): rbt[B, List[A]] =
  (key, x1) match {
  case (key, Nil) => rm_empty[B, List[A]]
  case (key, d :: ds) =>
    {
      val rm: rbt[B, List[A]] = elem_list_to_rm[A, B](key, ds)
      val k: B = key(d);
      rm_alpha[B, List[A]].apply(rm).apply(k) match {
        case None => rm_update_dj[B, List[A]].apply(k).apply(List(d)).apply(rm)
        case Some(data) =>
          rm_update[B, List[A]].apply(k).apply(d :: data).apply(rm)
      }
    }
}

def rm_set_lookup[A: linorder, B](rm: rbt[A, List[B]]): A => List[B] =
  (a: A) =>
    rm_alpha[A, List[B]].apply(rm).apply(a) match {
      case None => Nil
      case Some(rules) => rules
    }

def rtrancl_rbt_impl[A: linorder]: (List[(A, A)]) => (List[A]) => rbt[A, Unit] =
  (a: List[(A, A)]) =>
    rtrancl_impl[A, rbt[A, Unit]]((rel: List[(A, A)]) =>
                                    {
                                      val rm: rbt[A, List[(A, A)]] =
elem_list_to_rm[(A, A), A]((aa: (A, A)) => fst[A, A](aa), rel);
                                      (as: List[A]) =>
rs_to_list[A].apply(foldl[rbt[A, Unit],
                           A]((aa: rbt[A, Unit]) =>
                                (x: A) =>
                                  rs_union[A].apply(aa).apply(list_to_rs[A].apply(mapa[(A,
         A),
        A]((ab: (A, A)) => snd[A, A](ab),
            (rm_set_lookup[A, (A, A)](rm)).apply(x)))),
                               rs_empty[A], as))
                                    },
                                   (as: List[A]) =>
                                     (bs: rbt[A, Unit]) =>
                                       rs_union[A].apply(bs).apply(list_to_rs[A].apply(as)),
                                   rs_memb[A], rs_empty[A], a)

def memo_rbt_rtrancl[A: linorder](rel: List[(A, A)]): A => rbt[A, Unit] =
  {
    val tr: (List[A]) => rbt[A, Unit] = rtrancl_rbt_impl[A].apply(rel)
    val rm: rbt[A, rbt[A, Unit]] =
      list_to_rm[A, rbt[A, Unit]].apply(mapa[A,
      (A, rbt[A, Unit])]((a: A) => (a, tr(List(a))),
                          (comp[List[A], List[A],
                                 List[(A,
A)]](comp[rbt[A, Unit], List[A], List[A]](rs_to_list[A], list_to_rs[A]),
      (a: List[(A, A)]) =>
        mapa[(A, A), A]((aa: (A, A)) => fst[A, A](aa), a))).apply(rel)));
    (a: A) =>
      rm_lookup[A, rbt[A, Unit]](a, rm) match {
        case None => list_to_rs[A].apply(List(a))
        case Some(as) => as
      }
  }

def conv_ta_rule[A, B](eps: A => rbt[A, Unit], x1: ta_rule[A, B]):
      ta_rule_impl[A, B] =
  (eps, x1) match {
  case (eps, TA_rule(f, qs, q)) => TA_rule_impl[B, A](f, qs, q, eps(q))
}

def generate_ta[A: linorder, B: equal: linorder](x0: tree_automaton[A, B]):
      ta_impl_ext[A, B, Unit] =
  x0 match {
  case Tree_Automaton(fin, rules, eps) =>
    {
      val ep: A => rbt[A, Unit] = memo_rbt_rtrancl[A](eps)
      val rqs_rs: rbt[A, Unit] =
        foldl[rbt[A, Unit],
               ta_rule[A, B]]((a: rbt[A, Unit]) =>
                                (x: ta_rule[A, B]) =>
                                  rs_union[A].apply(a).apply(ep(rq[A, B](x))),
                               rs_empty[A], rules)
      val rrules: List[ta_rule_impl[A, B]] =
        mapa[ta_rule[A, B],
              ta_rule_impl[A, B]]((a: ta_rule[A, B]) =>
                                    conv_ta_rule[A, B](ep, a),
                                   rules);
      Ta_impl_ext[A, B,
                   Unit](list_to_rs[A].apply(fin),
                          elem_list_to_rm[ta_rule_impl[A, B],
   (B, nat)]((a: ta_rule_impl[A, B]) => rsym_impl[A, B](a), rrules),
                          rs_to_list[A].apply(rqs_rs), rqs_rs, ep, ())
    }
}

def ta_eps_impl[A, B, C](x0: ta_impl_ext[A, B, C]): A => rbt[A, Unit] = x0 match
  {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, more)
    => ta_eps_impl
}

def ta_res_list[A: equal: linorder,
                 B: linorder](ta: rbt[(A, nat), List[ta_rule_impl[B, A]]],
                               eps: B => rbt[B, Unit], x2: term[A, B]):
      rbt[B, Unit] =
  (ta, eps, x2) match {
  case (ta, eps, Var(q)) => eps(q)
  case (ta, eps, Fun(f, ts)) =>
    {
      val rec: List[rbt[B, Unit]] =
        mapa[term[A, B],
              rbt[B, Unit]]((a: term[A, B]) => ta_res_list[A, B](ta, eps, a),
                             ts)
      val n: nat = size_list[term[A, B]](ts)
      val rules: List[ta_rule_impl[B, A]] =
        (rm_set_lookup[(A, nat), ta_rule_impl[B, A]](ta)).apply((f, n))
      val arules: List[ta_rule_impl[B, A]] =
        filter[ta_rule_impl[B, A]]((rule: ta_rule_impl[B, A]) =>
                                     {
                                       val qs: List[B] = rqs_impl[B, A](rule);
                                       all_interval_nat((i: nat) =>
                  rs_memb[B].apply(nth[B](qs,
   i)).apply(nth[rbt[B, Unit]](rec, i)),
                 Zero_nat, n)
                                     },
                                    rules)
      val a: List[rbt[B, Unit]] =
        mapa[ta_rule_impl[B, A],
              rbt[B, Unit]]((a: ta_rule_impl[B, A]) => rq_impl[B, A](a),
                             arules);
      foldl[rbt[B, Unit], rbt[B, Unit]](rs_union[B], rs_empty[B], a)
    }
}

def ta_rqs_impl[A, B, C](x0: ta_impl_ext[A, B, C]): List[A] = x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, more)
    => ta_rqs_impl
}

def check_not_wf_trs[A: show,
                      B: equal: show](r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (isOK[(List[Char]) => List[Char],
                               Unit](check_wf_trs[A, B](r))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('T', 'h', 'e',
                          ' ', 'T', 'R', 'S', ' ', 'i', 's', ' ', 'w', 'e', 'l',
                          'l', ' ', 'f', 'o', 'r', 'm', 'e', 'd')),
  shows_nl))

def uncurry_proc_both[A: equal: show,
                       B: equal: show](x0: Option[nat],
x1: (lab[A, B],
      (List[((lab[A, B], nat), List[lab[A, B]])],
        (List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])],
          List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])))):
      (List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
            (List[(term[lab[A, B], List[Char]],
                    term[lab[A, B], List[Char]])]) =>
              Option[((lab[A, B], nat)) =>
                       List[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]])]] =>
                sum[(List[Char]) => List[Char], Unit] =
  (x0, x1) match {
  case (None, (a, (sml, (u, e)))) =>
    (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
      (c: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (d: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (f: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])
            => (g: Option[((lab[A, B], nat)) =>
                            List[(term[lab[A, B], List[Char]],
                                   term[lab[A, B], List[Char]])]])
                 => uncurry_proc[lab[A, B]]((a, (sml, (u, e))),
     (ba: List[((lab[A, B], nat), List[lab[A, B]])]) =>
       fmap[A, B](a, nat(BigInt(2)), ba),
     (aa: lab[A, B]) =>
       (ba: nat) =>
         (ca: List[((lab[A, B], nat), List[lab[A, B]])]) =>
           check_inj[A, B](aa, ba, ca),
     b, c, d, f, g)
  case (Some(n), (a, (sml, (u, e)))) =>
    (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
      (c: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (d: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (f: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])
            => (g: Option[((lab[A, B], nat)) =>
                            List[(term[lab[A, B], List[Char]],
                                   term[lab[A, B], List[Char]])]])
                 => uncurry_top_proc[lab[A,
  B]]((a, (sml, (u, e))), n,
       (ba: List[((lab[A, B], nat), List[lab[A, B]])]) => fmap[A, B](a, n, ba),
       (aa: lab[A, B]) =>
         (ba: nat) =>
           (ca: List[((lab[A, B], nat), List[lab[A, B]])]) =>
             check_inj[A, B](aa, ba, ca),
       b, c, d, f, g)
}

def afs_list_to_af[A: equal](pi: List[((A, nat), af_entry)], f: A, n: nat):
      nat => Boolean =
  (i: nat) =>
    lookupa[(A, nat), af_entry]((f, n), pi) match {
      case None => true
      case Some(Collapse(a)) => equal_nata(i, a)
      case Some(AFList(ids)) => membera[nat](ids, i)
    }

def default_af_entry(n: nat): af_entry = AFList(upt(Zero_nat, n))

def afs_list_to_afs[A: equal](pi: List[((A, nat), af_entry)], f: A, n: nat):
      af_entry =
  lookupa[(A, nat), af_entry]((f, n), pi) match {
    case None => default_af_entry(n)
    case Some(entry) => entry
  }

def check_wf_afs[A: show](pi: List[((A, nat), af_entry)]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(((A, nat), af_entry), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[((A, nat), af_entry),
                                 (List[Char]) =>
                                   List[Char]]((a: ((A, nat), af_entry)) =>
         {
           val (aa, b): ((A, nat), af_entry) = a;
           ({
              val (f, n): (A, nat) = aa;
              (e: af_entry) =>
                catcha[nat, Unit,
                        (List[Char]) =>
                          List[Char]](e match {
case Collapse(i) => check[nat](less_nat(i, n), i)
case AFList(ids) =>
  catcha[(nat, nat), Unit,
          nat](forallm[nat, nat]((x: nat) =>
                                   (if (less_nat(x, n)) Inr[Unit, nat](())
                                     else Inl[nat, Unit](x)),
                                  ids),
                (x: (nat, nat)) => Inl[nat, Unit](snd[nat, nat](x)))
                                      },
                                       (x: nat) =>
 Inl[(List[Char]) => List[Char],
      Unit](comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('i', 'n', 'd', 'e', 'x',
               ' ')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_nat(Zero_nat, x),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(' ', 't', 'o', 'o', ' ',
             'h', 'i', 'g', 'h', ' ', 'f', 'o', 'r', ' ', 'A', 'F', '-', 'e',
             'n', 't', 'r', 'y', ' ', 'o', 'f')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_prec_prod[A,
                        nat](Zero_nat, (f, n)),
        shows_nl))))))
            })(b)
         },
        pi),
                         (x: (((A, nat), af_entry), (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[((A, nat), af_entry),
      (List[Char]) => List[Char]](x)))

def rp_non_strict[A, B](x0: (sum[(List[Char]) => List[Char], Unit],
                              (((term[A, B], term[A, B])) =>
                                 sum[(List[Char]) => List[Char], Unit],
                                (((term[A, B], term[A, B])) =>
                                   sum[(List[Char]) => List[Char], Unit],
                                  (A => nat => nat => Boolean,
                                    (sum[(List[Char]) => List[Char], Unit],
                                      (List[Char]) => List[Char])))))):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case (v, (s, (n, (a, (m, d))))) => n
}

def mono_af_entry(n: nat, x1: af_entry): Boolean = (n, x1) match {
  case (n, Collapse(i)) => less_eq_nat(n, one_nat)
  case (n, AFList(ids)) =>
    all_interval_nat((a: nat) => membera[nat](ids, a), Zero_nat, n)
}

def check_mono_afs_help[A: show, B, C](x0: ((A, B), C)):
      (List[Char]) => List[Char] =
  x0 match {
  case ((f, a), e) =>
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List('e', 'n', 't', 'r', 'y', ' ', 'f',
       'o', 'r', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]]((aa: List[Char]) =>
   shows_prec[A](Zero_nat, f, aa),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'n', 'o', 't', ' ', 'm', 'o',
     'n', 'o', 't', 'o', 'n', 'i', 'c')),
                     shows_nl)))
}

def check_mono_afs[A: show](pi: List[((A, nat), af_entry)]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((A, nat), af_entry), Unit,
          (List[Char]) =>
            List[Char]](catcha[(((A, nat), af_entry), ((A, nat), af_entry)),
                                Unit,
                                ((A, nat),
                                  af_entry)](forallm[((A, nat), af_entry),
              ((A, nat),
                af_entry)]((x: ((A, nat), af_entry)) =>
                             (if (mono_af_entry(snd[A,
             nat](fst[(A, nat), af_entry](x)),
         snd[(A, nat), af_entry](x)))
                               Inr[Unit, ((A, nat), af_entry)](())
                               else Inl[((A, nat), af_entry), Unit](x)),
                            pi),
      (x: (((A, nat), af_entry), ((A, nat), af_entry))) =>
        Inl[((A, nat), af_entry),
             Unit](snd[((A, nat), af_entry), ((A, nat), af_entry)](x))),
                         (x: ((A, nat), af_entry)) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](check_mono_afs_help[A, nat, af_entry](x)))

def af_redpair[A: equal: show,
                B: show](pi: List[((A, nat), af_entry)],
                          rp: (sum[(List[Char]) => List[Char], Unit],
                                (((term[filtered[A], B],
                                   term[filtered[A], B])) =>
                                   sum[(List[Char]) => List[Char], Unit],
                                  (((term[filtered[A], B],
                                     term[filtered[A], B])) =>
                                     sum[(List[Char]) => List[Char], Unit],
                                    ((filtered[A]) => nat => nat => Boolean,
                                      (sum[(List[Char]) => List[Char], Unit],
(List[Char]) => List[Char])))))):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
          (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
            (A => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  {
    val af: A => nat => nat => Boolean =
      (a: A) => (b: nat) => afs_list_to_af[A](pi, a, b)
    val afs: A => nat => af_entry =
      (a: A) => (b: nat) => afs_list_to_afs[A](pi, a, b);
    (binda[(List[Char]) => List[Char], Unit,
            Unit](check_wf_afs[A](pi),
                   (_: Unit) => rp_valid[filtered[A], B](rp)),
      ((a: (term[A, B], term[A, B])) =>
         afs_check[A, B](shows_string.apply(List('>')), afs,
                          rp_strict[filtered[A], B](rp), a),
        ((a: (term[A, B], term[A, B])) =>
           afs_check[A, B](shows_string.apply(List('>', '=')), afs,
                            rp_non_strict[filtered[A], B](rp), a),
          (af, (binda[(List[Char]) => List[Char], Unit,
                       Unit](check_mono_afs[A](pi),
                              (_: Unit) => rp_mono[filtered[A], B](rp)),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List('A', 'r', 'g', 'u',
                    'm', 'e', 'n', 't', ' ', 'F', 'i', 'l', 't', 'e', 'r', ':',
                    ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl,
               comp[List[Char], List[Char],
                     List[Char]](shows_afs[A](pi),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, rp_desc[filtered[A], B](rp))))))))))
  }

def xml2trsDisproof[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], tRSDisproof[A, List[Char]]] =
  xml1element[tRSDisproof[A, List[Char]],
               tRSDisproof[A, List[Char]]](List('t', 'r', 's', 'N', 'o', 'n',
         't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o',
         'o', 'f'),
    (a: xml) =>
      xml_options[tRSDisproof[A, List[Char]]](List((List('v', 'a', 'r', 'i',
                  'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't', 'i', 'o',
                  'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd'),
             (aa: xml) =>
               xml0elements[tRSDisproof[A,
 List[Char]]](List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd',
                    'i', 't', 'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e',
                    'd'),
               NotWellFormed, aa)),
            (List('l', 'o', 'o', 'p'),
              (aa: xml) =>
                xml_change[(term[A, List[Char]],
                             (List[(pos, ((term[A, List[Char]],
    term[A, List[Char]]),
   (Boolean, term[A, List[Char]])))],
                               (List[(List[Char], term[A, List[Char]])],
                                 ctxt[A, List[Char]]))),
                            tRSDisproof[A,
 List[Char]]](loop[A](xml2name),
               (ab: (term[A, List[Char]],
                      (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                    (Boolean, term[A, List[Char]])))],
                        (List[(List[Char], term[A, List[Char]])],
                          ctxt[A, List[Char]]))))
                 => {
                      val (s, (rseq, (ac, b))):
                            (term[A, List[Char]],
                              (List[(pos, ((term[A, List[Char]],
     term[A, List[Char]]),
    (Boolean, term[A, List[Char]])))],
                                (List[(List[Char], term[A, List[Char]])],
                                  ctxt[A, List[Char]])))
                        = ab;
                      Loop[A, List[Char]](s,
   mapa[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                (Boolean, term[A, List[Char]]))),
         (pos, ((term[A, List[Char]], term[A, List[Char]]),
                 term[A, List[Char]]))]((ad:
   (pos, ((term[A, List[Char]], term[A, List[Char]]),
           (Boolean, term[A, List[Char]]))))
  => {
       val (xa, (y, (_, z))):
             (pos, ((term[A, List[Char]], term[A, List[Char]]),
                     (Boolean, term[A, List[Char]])))
         = ad;
       (xa, (y, z))
     },
 rseq),
   ac, b)
                    },
               aa)),
            (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
              (aa: xml) =>
                xml2elements[List[(term[A, List[Char]], term[A, List[Char]])],
                              tRSDisproof[A, List[Char]],
                              tRSDisproof[A,
   List[Char]]](List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
                 (ab: xml) =>
                   xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                                List[(term[A, List[Char]],
                                       term[A,
     List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                     id[List[(term[A, List[Char]], term[A, List[Char]])]], ab),
                 (ab: xml) => xml2trsDisproof[A](xml2name, ab),
                 (ab: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                   (b: tRSDisproof[A, List[Char]]) =>
                     NTRuleRemoval[A, List[Char]](ab, b),
                 aa)),
            (List('d', 'p', 'T', 'r', 'a', 'n', 's'),
              (aa: xml) =>
                xml3elements[List[(term[A, List[Char]], term[A, List[Char]])],
                              Boolean, dPDisproof[A, List[Char]],
                              tRSDisproof[A,
   List[Char]]](List('d', 'p', 'T', 'r', 'a', 'n', 's'),
                 (ab: xml) =>
                   xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                                List[(term[A, List[Char]],
                                       term[A,
     List[Char]])]](List('d', 'p', 's'), rules[A](xml2name),
                     id[List[(term[A, List[Char]], term[A, List[Char]])]], ab),
                 (ab: xml) =>
                   bool(List('m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm', 'b',
                              'o', 'l', 's'),
                         ab),
                 (ab: xml) => xml2dpDisproof[A](xml2name, ab),
                 (p: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                   (_: Boolean) =>
                     (ab: dPDisproof[A, List[Char]]) =>
                       NTDPTrans[A, List[Char]](p, ab),
                 aa)),
            (List('s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's',
                   'a', 'l'),
              (aa: xml) =>
                xml2elements[List[(term[A, List[Char]], term[A, List[Char]])],
                              tRSDisproof[A, List[Char]],
                              tRSDisproof[A,
   List[Char]]](List('s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's',
                      'a', 'l'),
                 (ab: xml) =>
                   xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                                List[(term[A, List[Char]],
                                       term[A,
     List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                     id[List[(term[A, List[Char]], term[A, List[Char]])]], ab),
                 (ab: xml) => xml2trsDisproof[A](xml2name, ab),
                 (ab: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                   (b: tRSDisproof[A, List[Char]]) =>
                     NTStringReversal[A, List[Char]](ab, b),
                 aa))),
       a),
    id[tRSDisproof[A, List[Char]]], x)

def xml2relTrsProof[A: equal](xml2name:
                                xml => sum[List[Char], lab[A, List[nat]]],
                               x: xml):
      sum[List[Char], relTRSProof[A, List[nat], List[Char]]] =
  xml1element[relTRSProof[A, List[nat], List[Char]],
               relTRSProof[A, List[nat],
                            List[Char]]](List('r', 'e', 'l', 'a', 't', 'i', 'v',
       'e', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r',
       'o', 'o', 'f'),
  (a: xml) =>
    xml_options[relTRSProof[A, List[nat],
                             List[Char]]](List((List('r', 'I', 's', 'E', 'm',
              'p', 't', 'y'),
         (aa: xml) =>
           xml0elements[relTRSProof[A, List[nat],
                                     List[Char]]](List('r', 'I', 's', 'E', 'm',
                'p', 't', 'y'),
           RelRisEmpty, aa)),
        (List('s', 'I', 's', 'E', 'm', 'p', 't', 'y'),
          (aa: xml) =>
            xml1element[tRSProof[A, List[nat], List[Char]],
                         relTRSProof[A, List[nat],
                                      List[Char]]](List('s', 'I', 's', 'E', 'm',
                 'p', 't', 'y'),
            (ab: xml) => xml2trsProof[A](xml2name, ab),
            (ab: tRSProof[A, List[nat], List[Char]]) =>
              RelSisEmpty[A, List[nat], List[Char]](ab),
            aa)),
        (List('u', 'n', 'l', 'a', 'b'),
          (aa: xml) =>
            xml3elements[List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]], List[Char]])],
                          List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                          relTRSProof[A, List[nat], List[Char]],
                          relTRSProof[A, List[nat],
                                       List[Char]]](List('u', 'n', 'l', 'a',
                  'b'),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) => xml2relTrsProof[A](xml2name, ab),
             (ab: List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])])
               => (b: List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])])
                    => (c: relTRSProof[A, List[nat], List[Char]]) =>
                         RelUnlab[A, List[nat], List[Char]](ab, b, c),
             aa)),
        (List('s', 'e', 'm', 'l', 'a', 'b'),
          (aa: xml) =>
            xml4elements[sum[(sl_inter[A, List[nat]], Boolean),
                              Option[lab[A, List[nat]]]],
                          List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                          List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                          relTRSProof[A, List[nat], List[Char]],
                          relTRSProof[A, List[nat],
                                       List[Char]]](List('s', 'e', 'm', 'l',
                  'a', 'b'),
             sl_inter[A, List[nat]](xml2name),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) => xml2relTrsProof[A](xml2name, ab),
             (ab: sum[(sl_inter[A, List[nat]], Boolean),
                       Option[lab[A, List[nat]]]])
               => ab match {
                    case Inl((i, true)) =>
                      (ac: List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])])
                        => (b: List[(term[lab[A, List[nat]], List[Char]],
                                      term[lab[A, List[nat]], List[Char]])])
                             => (c: relTRSProof[A, List[nat], List[Char]]) =>
                                  RelSemlab[A, List[nat],
     List[Char]](i, Inl[(lab[A, List[nat]]) =>
                          (List[nat]) =>
                            sum[List[nat], List[lab[A, List[nat]]]],
                         List[Char]]((ad: lab[A, List[nat]]) =>
                                       (ba: List[nat]) =>
 id_labeling[A, nat](ad, ba)),
                  ac, b, c)
                    case Inl((i, false)) =>
                      (ac: List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])])
                        => (b: List[(term[lab[A, List[nat]], List[Char]],
                                      term[lab[A, List[nat]], List[Char]])])
                             => (c: relTRSProof[A, List[nat], List[Char]]) =>
                                  RelSemlab[A, List[nat],
     List[Char]](i, Inr[List[Char],
                         (lab[A, List[nat]]) =>
                           (List[nat]) =>
                             sum[List[nat], List[lab[A, List[nat]]]]](Nil),
                  ac, b, c)
                    case Inr(_) =>
                      (ac: List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])])
                        => (b: List[(term[lab[A, List[nat]], List[Char]],
                                      term[lab[A, List[nat]], List[Char]])])
                             => (c: relTRSProof[A, List[nat], List[Char]]) =>
                                  RelRootlab[A, List[nat], List[Char]](ac, b, c)
                  },
             aa)),
        (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
          (aa: xml) =>
            xml4elements[redpair[lab[A, List[nat]]],
                          List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                          List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                          relTRSProof[A, List[nat], List[Char]],
                          relTRSProof[A, List[nat],
                                       List[Char]]](List('r', 'u', 'l', 'e',
                  'R', 'e', 'm', 'o', 'v', 'a', 'l'),
             ordering_constraint_proof[lab[A, List[nat]]](xml2name),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) => xml2relTrsProof[A](xml2name, ab),
             (ab: redpair[lab[A, List[nat]]]) =>
               (b: List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])])
                 => (c: List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])])
                      => (d: relTRSProof[A, List[nat], List[Char]]) =>
                           RelRuleRemoval[A, List[nat],
   List[Char]](ab, b, c, d),
             aa)),
        (List('u', 'n', 'c', 'u', 'r', 'r', 'y'),
          (aa: xml) =>
            xml4elements[(lab[A, List[nat]],
                           (List[((lab[A, List[nat]], nat),
                                   List[lab[A, List[nat]]])],
                             (List[(term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]])],
                               List[(term[lab[A, List[nat]], List[Char]],
                                      term[lab[A, List[nat]], List[Char]])]))),
                          List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                          List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                          relTRSProof[A, List[nat], List[Char]],
                          relTRSProof[A, List[nat],
                                       List[Char]]](List('u', 'n', 'c', 'u',
                  'r', 'r', 'y'),
             uncurry_info[lab[A, List[nat]]](xml2name),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) =>
               xml1element[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])],
                            List[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])]](List('t', 'r', 's'), rules[lab[A, List[nat]]](xml2name),
                 id[List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])]],
                 ab),
             (ab: xml) => xml2relTrsProof[A](xml2name, ab),
             (ab: (lab[A, List[nat]],
                    (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
                      (List[(term[lab[A, List[nat]], List[Char]],
                              term[lab[A, List[nat]], List[Char]])],
                        List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])]))))
               => (b: List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])])
                    => (c: List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])])
                         => (d: relTRSProof[A, List[nat], List[Char]]) =>
                              RelUncurry[A, List[nat], List[Char]](ab, b, c, d),
             aa))),
   a),
  id[relTRSProof[A, List[nat], List[Char]]], x)

def xml2relTrsDisproof[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], relTRSDisproof[A, List[Char]]] =
  xml1element[relTRSDisproof[A, List[Char]],
               relTRSDisproof[A, List[Char]]](List('r', 'e', 'l', 'a', 't', 'i',
            'v', 'e', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't',
            'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
       (a: xml) =>
         xml_options[relTRSDisproof[A, List[Char]]](List((List('v', 'a', 'r',
                        'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't',
                        'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd'),
                   (aa: xml) =>
                     xml0elements[relTRSDisproof[A,
          List[Char]]](List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o',
                             'n', 'd', 'i', 't', 'i', 'o', 'n', 'V', 'i', 'o',
                             'l', 'a', 't', 'e', 'd'),
                        RelNotWellFormed, aa)),
                  (List('l', 'o', 'o', 'p'),
                    (aa: xml) =>
                      xml_change[(term[A, List[Char]],
                                   (List[(pos,
   ((term[A, List[Char]], term[A, List[Char]]),
     (Boolean, term[A, List[Char]])))],
                                     (List[(List[Char], term[A, List[Char]])],
                                       ctxt[A, List[Char]]))),
                                  relTRSDisproof[A,
          List[Char]]](loop[A](xml2name),
                        (ab: (term[A, List[Char]],
                               (List[(pos, ((term[A, List[Char]],
      term[A, List[Char]]),
     (Boolean, term[A, List[Char]])))],
                                 (List[(List[Char], term[A, List[Char]])],
                                   ctxt[A, List[Char]]))))
                          => {
                               val (s, (rseq, (ac, b))):
                                     (term[A, List[Char]],
                                       (List[(pos,
       ((term[A, List[Char]], term[A, List[Char]]),
         (Boolean, term[A, List[Char]])))],
 (List[(List[Char], term[A, List[Char]])], ctxt[A, List[Char]])))
                                 = ab;
                               RelLoop[A, List[Char]](s, rseq, ac, b)
                             },
                        aa)),
                  (List('t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i',
                         'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
                    (aa: xml) =>
                      xml_change[tRSDisproof[A, List[Char]],
                                  relTRSDisproof[A,
          List[Char]]]((ab: xml) => xml2trsDisproof[A](xml2name, ab),
                        (ab: tRSDisproof[A, List[Char]]) =>
                          RNotSN[A, List[Char]](ab),
                        aa)),
                  (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
                    (aa: xml) =>
                      xml3elements[List[(term[A, List[Char]],
  term[A, List[Char]])],
                                    List[(term[A, List[Char]],
   term[A, List[Char]])],
                                    relTRSDisproof[A, List[Char]],
                                    relTRSDisproof[A,
            List[Char]]](List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a',
                               'l'),
                          (ab: xml) =>
                            xml1element[List[(term[A, List[Char]],
       term[A, List[Char]])],
 List[(term[A, List[Char]],
        term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                                id[List[(term[A, List[Char]],
  term[A, List[Char]])]],
                                ab),
                          (ab: xml) =>
                            xml1element[List[(term[A, List[Char]],
       term[A, List[Char]])],
 List[(term[A, List[Char]],
        term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                                id[List[(term[A, List[Char]],
  term[A, List[Char]])]],
                                ab),
                          (ab: xml) => xml2relTrsDisproof[A](xml2name, ab),
                          (ab: List[(term[A, List[Char]], term[A, List[Char]])])
                            => (b: List[(term[A, List[Char]],
  term[A, List[Char]])])
                                 => (c: relTRSDisproof[A, List[Char]]) =>
                                      NTRelRuleRemoval[A, List[Char]](ab, b, c),
                          aa))),
             a),
       id[relTRSDisproof[A, List[Char]]], x)

def xml2certProblem[A: equal](xml2uname: xml => sum[List[Char], A],
                               xml2name:
                                 xml => sum[List[Char], lab[A, List[nat]]],
                               x2: xml):
      sum[List[Char], certificationProblem[A, List[nat], List[Char]]] =
  (xml2uname, xml2name, x2) match {
  case (xml2uname, xml2name, Xml(name, egal, cs, texta)) =>
    (if (equal_lista[Char](name,
                            List('c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a',
                                  't', 'i', 'o', 'n', 'P', 'r', 'o', 'b', 'l',
                                  'e', 'm')) &&
           (is_none[List[Char]](texta) &&
             less_eq_nat(nat(BigInt(3)), size_list[xml](cs))))
      binda[List[Char],
             (List[(term[A, List[Char]], term[A, List[Char]])],
               Option[List[(term[A, List[Char]], term[A, List[Char]])]]),
             certificationProblem[A, List[nat],
                                   List[Char]]]((xml2trsInput[A](xml2uname)).apply(nth[xml](cs,
             Zero_nat)),
         (a: (List[(term[A, List[Char]], term[A, List[Char]])],
               Option[List[(term[A, List[Char]], term[A, List[Char]])]]))
           => {
                val (r, s_opt):
                      (List[(term[A, List[Char]], term[A, List[Char]])],
                        Option[List[(term[A, List[Char]],
                                      term[A, List[Char]])]])
                  = a;
                binda[List[Char], List[Char],
                       certificationProblem[A, List[nat],
     List[Char]]](text(List('c', 'p', 'f', 'V', 'e', 'r', 's', 'i', 'o', 'n'),
                        nth[xml](cs, one_nat)),
                   (_: List[Char]) =>
                     s_opt match {
                       case None =>
                         xml1element[certificationProblem[A, List[nat],
                   List[Char]],
                                      certificationProblem[A, List[nat],
                    List[Char]]](List('p', 'r', 'o', 'o', 'f'),
                                  (aa: xml) =>
                                    xml_options[certificationProblem[A,
                              List[nat],
                              List[Char]]](List((List('t', 'r', 's', 'T', 'e',
               'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
               'f'),
          (ab: xml) =>
            xml_change[tRSProof[A, List[nat], List[Char]],
                        certificationProblem[A, List[nat],
      List[Char]]]((ac: xml) => xml2trsProof[A](xml2name, ac),
                    (ac: tRSProof[A, List[nat], List[Char]]) =>
                      TRSProof[A, List[Char], List[nat]](r, ac),
                    ab)),
         (List('t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a',
                't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
           (ab: xml) =>
             xml_change[tRSDisproof[lab[A, List[nat]], List[Char]],
                         certificationProblem[A, List[nat],
       List[Char]]]((ac: xml) =>
                      xml2trsDisproof[lab[A, List[nat]]](xml2name, ac),
                     (ac: tRSDisproof[lab[A, List[nat]], List[Char]]) =>
                       TRSDisproof[A, List[Char], List[nat]](r, ac),
                     ab))),
    aa),
                                  id[certificationProblem[A, List[nat],
                   List[Char]]],
                                  nth[xml](cs, nat(BigInt(2))))
                       case Some(s) =>
                         xml1element[certificationProblem[A, List[nat],
                   List[Char]],
                                      certificationProblem[A, List[nat],
                    List[Char]]](List('p', 'r', 'o', 'o', 'f'),
                                  (aa: xml) =>
                                    xml_options[certificationProblem[A,
                              List[nat],
                              List[Char]]](List((List('r', 'e', 'l', 'a', 't',
               'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
               'n', 'P', 'r', 'o', 'o', 'f'),
          (ab: xml) =>
            xml_change[relTRSProof[A, List[nat], List[Char]],
                        certificationProblem[A, List[nat],
      List[Char]]]((ac: xml) => xml2relTrsProof[A](xml2name, ac),
                    (ac: relTRSProof[A, List[nat], List[Char]]) =>
                      RelTRSProof[A, List[Char], List[nat]]((r, s), ac),
                    ab)),
         (List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o', 'n', 't', 'e',
                'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
                'f'),
           (ab: xml) =>
             xml_change[relTRSDisproof[lab[A, List[nat]], List[Char]],
                         certificationProblem[A, List[nat],
       List[Char]]]((ac: xml) =>
                      xml2relTrsDisproof[lab[A, List[nat]]](xml2name, ac),
                     (ac: relTRSDisproof[lab[A, List[nat]], List[Char]]) =>
                       RelTRSDisproof[A, List[Char], List[nat]]((r, s), ac),
                     ab))),
    aa),
                                  id[certificationProblem[A, List[nat],
                   List[Char]]],
                                  nth[xml](cs, nat(BigInt(2))))
                     })
              })
      else Inl[List[Char],
                certificationProblem[A, List[nat],
                                      List[Char]]](List('c', 'o', 'u', 'l', 'd',
                 ' ', 'n', 'o', 't', ' ', 'p', 'a', 'r', 's', 'e', ' ', 'r',
                 'o', 'o', 't', ' ', 'e', 'l', 'e', 'm', 'e', 'n', 't')))
}

def eqWhiteSpace(s1: List[Char], s2: List[Char]): Boolean =
  debug[Boolean](List('0'),
                  List('c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'w',
                        'h', 'e', 't', 'h', 'e', 'r', ' ', 'p', 'a', 'r', 's',
                        'e', 'd', ' ', 'i', 'n', 'p', 'u', 't', ' ', 'c', 'o',
                        'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', 's', ' ', 't',
                        'o', ' ', 'r', 'e', 'a', 'l', ' ', 'i', 'n', 'p', 'u',
                        't'),
                  equal_lista[Char](filter[Char]((c: Char) =>
           ! (membera[Char](wspace, c)),
          s1),
                                     filter[Char]((c: Char) =>
            ! (membera[Char](wspace, c)),
           s2)))

def xml_certificationProblem[A: show, B,
                              C: show](x0: certificationProblem[A, B, C]):
      xml =
  x0 match {
  case TRSProof(trs, uu) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         List(xml_rules[A, C](List('t', 'r', 's'), trs)), None)
  case TRSDisproof(trs, uv) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         List(xml_rules[A, C](List('t', 'r', 's'), trs)), None)
  case RelTRSProof((r, s), uw) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         List(xml_rules[A, C](List('t', 'r', 's'), r),
               xml_rules[A, C](List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R',
                                     'u', 'l', 'e', 's'),
                                s)),
         None)
  case RelTRSDisproof((r, s), ux) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         List(xml_rules[A, C](List('t', 'r', 's'), r),
               xml_rules[A, C](List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R',
                                     'u', 'l', 'e', 's'),
                                s)),
         None)
}

def check_DP[A: equal: show,
              B: equal: show](shp: A => A, r: List[(term[A, B], term[A, B])],
                               rm: Option[((A, nat)) =>
    List[(term[A, B], term[A, B])]],
                               p: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](binda[(List[Char]) => List[Char], Unit,
                               Unit](check_wf_trs[A, B](r),
                                      (_: Unit) =>
{
  val d: List[(A, nat)] = defs_list[A, B](r);
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[((A, nat), (List[Char]) => List[Char]), Unit,
                       (List[Char]) =>
                         List[Char]](forallm[(A, nat),
      (List[Char]) =>
        List[Char]]((a: (A, nat)) =>
                      {
                        val (f, n): (A, nat) = a;
                        check[(List[Char]) =>
                                List[Char]](! (membera[(A,
                 nat)](d, (shp(f), n))),
     comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('s', 'h', 'a', 'r', 'p', 'i',
        'n', 'g', ' ')),
                        comp[List[Char], List[Char],
                              List[Char]]((aa: List[Char]) =>
    shows_prec[A](Zero_nat, f, aa),
   comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List(' ', 'y', 'i', 'e', 'l', 'd', 's',
      ' ', 't', 'h', 'e', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd', ' ', 's', 'y',
      'm', 'b', 'o', 'l', ' ')),
                      (aa: List[Char]) =>
                        shows_prec[A](Zero_nat, shp(f), aa)))))
                      },
                     d),
                                      (x:
 ((A, nat), (List[Char]) => List[Char]))
=> Inl[(List[Char]) => List[Char],
        Unit](snd[(A, nat), (List[Char]) => List[Char]](x))),
                (_: Unit) =>
                  catcha[(term[A, B], term[A, B]), Unit,
                          (List[Char]) =>
                            List[Char]](check_subseteq[(term[A, B],
                 term[A, B])](dP_list[A, B](shp, r, rm), p),
 (x: (term[A, B], term[A, B])) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'D',
                 'P', ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_rule[A, B](x),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List(' ', 'd', 'o', 'e', 's',
               ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p', 'e', 'a', 'r', ' ', 'i',
               'n', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b',
               'l', 'e', 'm')),
                               shows_nl))))))
}),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'D', 'P', '-', 't',
 'r', 'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o', 'n', ' ', 'i', 's',
 ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p', 'l', 'i', 'e', 'd', ' ', 'c', 'o', 'r',
 'r', 'e', 'c', 't', 'l', 'y', '.')),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))

def internal_error: (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('i', 'n', 't', 'e', 'r', 'n', 'a',
     'l', ' ', 'e', 'r', 'r', 'o', 'r')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('p', 'l', 'e', 'a', 's', 'e', ' ',
   'c', 'o', 'n', 't', 'a', 'c', 't', ' ', 'o', 'n', 'e', ' ', 'o', 'f', ' ',
   't', 'h', 'e', ' ', 'C', 'e', 'T', 'A', ' ', 'm', 'e', 'm', 'b', 'e', 'r',
   's')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('h', 't', 't', 'p', ':', '/', '/', 'c',
 'l', '-', 'i', 'n', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'k', '.', 'u', 'i', 'b',
 'k', '.', 'a', 'c', '.', 'a', 't', '/', 's', 'o', 'f', 't', 'w', 'a', 'r', 'e',
 '/', 'c', 'e', 't', 'a')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('c', 'e', 't', 'a', '@', 'i', 'n', 'f',
                                       'o', 'r', 'm', 'a', 't', 'i', 'k', '.',
                                       'u', 'i', 'b', 'k', '.', 'a', 'c', '.',
                                       'a', 't')),
               shows_nl)))))))

def check_redpair_proc[A: equal: show,
                        B: equal: show](rp:
  (sum[(List[Char]) => List[Char], Unit],
    (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
      (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
        (A => nat => nat => Boolean,
          (sum[(List[Char]) => List[Char], Unit],
            (List[Char]) => List[Char]))))),
 p: List[(term[A, B], term[A, B])], rP: List[(term[A, B], term[A, B])],
 r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (_, nSp):
                                (List[(term[A, B], term[A, B])],
                                  List[(term[A, B], term[A, B])])
                            = partition[(term[A, B],
  term[A, B])]((foo: (term[A, B], term[A, B])) =>
                 isOK[(List[Char]) => List[Char],
                       Unit]((rp_strict[A, B](rp)).apply(foo)),
                p);
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](rp_valid[A, B](rp),
(_: Unit) =>
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](catcha[((term[A, B], term[A, B]),
      (List[Char]) => List[Char]),
     Unit,
     (List[Char]) =>
       List[Char]](forallm[(term[A, B], term[A, B]),
                            (List[Char]) =>
                              List[Char]](rp_non_strict[A, B](rp), r),
                    (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char]))
                      => Inl[(List[Char]) => List[Char],
                              Unit](snd[(term[A, B], term[A, B]),
 (List[Char]) => List[Char]](x))),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n',
              't', 'i', 'n', 'g', ' ', 'T', 'R', 'S')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  binda[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[A, B], term[A, B]),
    (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSp),
                                    (x: ((term[A, B], term[A, B]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[A, B], term[A, B]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'D', 'P', 's')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  catcha[(term[A, B], term[A, B]), Unit,
  (List[Char]) =>
    List[Char]](check_subseteq[(term[A, B], term[A, B])](nSp, rP),
                 (x: (term[A, B], term[A, B])) =>
                   Inl[(List[Char]) => List[Char],
                        Unit]((rp_strict[A, B](rp)).apply(x) match {
                                case Inl(err) => err
                                case Inr(_) => internal_error
                              })))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](rp_desc[A, B](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))

def to_lpoly_inter[A, B: equal](d: A, on: A, i: List[((B, nat), (A, List[A]))],
                                 f: B, n: nat):
      (A, List[A]) =
  lookupa[(B, nat), (A, List[A])]((f, n), i) match {
    case None => (d, replicate[A](n, on))
    case Some(cas) => cas
  }

def check_lpoly_coeffs[A: show,
                        B: show](wfe: A => Boolean, apos: A => Boolean,
                                  ch: A => Boolean,
                                  i: List[((B, nat), (A, List[A]))]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(((B, nat), (A, List[A])), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[((B, nat), (A, List[A])),
                                 (List[Char]) =>
                                   List[Char]]((a: ((B, nat), (A, List[A]))) =>
         {
           val (aa, b): ((B, nat), (A, List[A])) = a;
           ({
              val (f, n): (B, nat) = aa;
              (ab: (A, List[A])) =>
                {
                  val (c, cs): (A, List[A]) = ab;
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](binda[(List[Char]) => List[Char], Unit,
       Unit](check[(List[Char]) =>
                     List[Char]](wfe(c),
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ',
                                     'p', 'a', 'r', 't', ' ')),
             comp[List[Char], List[Char],
                   List[Char]]((ac: List[Char]) =>
                                 shows_prec[A](Zero_nat, c, ac),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List(' ',
                                   'i', 's', ' ', 'n', 'o', 't', ' ', 'w', 'e',
                                   'l', 'l', '-', 'f', 'o', 'r', 'm', 'e',
                                   'd')),
           shows_nl)))),
              (_: Unit) =>
                binda[(List[Char]) => List[Char], Unit,
                       Unit](check[(List[Char]) =>
                                     List[Char]](apos(c) ||
           list_ex[A](apos, take[A](n, cs)),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd',
             ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 'p', 'o', 's',
             'i', 't', 'i', 'v', 'e', ' ', 'e', 'n', 't', 'r', 'y', ' ', 'w',
             'h', 'i', 'c', 'h', ' ', 'i', 's', ' ', 'r', 'e', 'q', 'u', 'i',
             'r', 'e', 'd', ' ', 'f', 'o', 'r', ' ', 'a', 'r', 'c', 't', 'i',
             'c', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't',
             'i', 'o', 'n', 's')),
                             shows_nl)),
                              (_: Unit) =>
                                catcha[(A, (List[Char]) => List[Char]), Unit,
(List[Char]) =>
  List[Char]](forallm[A, (List[Char]) =>
                           List[Char]]((ac: A) =>
 check[(List[Char]) =>
         List[Char]](ch(ac) && wfe(ac),
                      comp[List[Char], List[Char],
                            List[Char]](shows_string.apply(List('c', 'o', 'e',
                         'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', ' ')),
 comp[List[Char], List[Char],
       List[Char]]((ba: List[Char]) => shows_prec[A](Zero_nat, ac, ba),
                    comp[List[Char], List[Char],
                          List[Char]](shows_string.apply(List(' ', 'i', 's',
                       ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
                       'd')),
                                       shows_nl)))),
cs),
               (x: (A, (List[Char]) => List[Char])) =>
                 Inl[(List[Char]) => List[Char],
                      Unit](snd[A, (List[Char]) => List[Char]](x))))),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
                 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'i', 'n', 't', 'e',
                 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o',
                 'f', ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]]((ac: List[Char]) =>
             shows_prec[B](Zero_nat, f, ac),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('/')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_nat(Zero_nat, n),
          comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))
                }
            })(b)
         },
        i),
                         (x: (((B, nat), (A, List[A])),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[((B, nat), (A, List[A])),
      (List[Char]) => List[Char]](x)))

def get_comp_wf_elem[A](x0: (A => Boolean,
                              (A, (A, (A,
(A => A => A,
  (A => A => A,
    (A => A,
      (A => A => Boolean,
        (A => A => Boolean,
          (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => Boolean =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => wfe
}

def get_comp_ge[A](x0: (A => Boolean,
                         (A, (A, (A, (A => A => A,
                                       (A => A => A,
 (A => A,
   (A => A => Boolean,
     (A => A => Boolean,
       (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => A => Boolean =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => ge
}

def lookup_rest[A: equal, B](x: A, xa1: List[(A, B)]):
      Option[(B, List[(A, B)])] =
  (x, xa1) match {
  case (x, Nil) => None
  case (x, (y, c) :: ycs) =>
    (if (eq[A](x, y)) Some[(B, List[(A, B)])]((c, ycs))
      else lookup_rest[A, B](x, ycs) match {
             case None => None
             case Some((d, yccs)) =>
               Some[(B, List[(A, B)])]((d, (y, c) :: yccs))
           })
}

def check_pvarsI[A, B: equal](ze: A, ge: A => A => Boolean, vas: List[(B, A)],
                               x3: List[(B, A)]):
      sum[B, Unit] =
  (ze, ge, vas, x3) match {
  case (ze, ge, vas, Nil) =>
    catcha[((B, A), B), Unit,
            B](forallm[(B, A),
                        B]((va: (B, A)) =>
                             check[B]((ge(snd[B, A](va)))(ze), fst[B, A](va)),
                            vas),
                (x: ((B, A), B)) => Inl[B, Unit](snd[(B, A), B](x)))
  case (ze, ge, vas, (x, b) :: vbs) =>
    {
      val (a, vasa): (A, List[(B, A)]) =
        lookup_rest[B, A](x, vas) match {
          case None => (ze, vas)
          case Some((a, ba)) => (a, ba)
        };
      binda[B, Unit,
             Unit](check[B]((ge(a))(b), x),
                    (_: Unit) => check_pvarsI[A, B](ze, ge, vasa, vbs))
    }
}

def show_pvars[A: equal: show, B: show](on: A, x1: List[(B, A)]):
      List[List[Char]] =
  (on, x1) match {
  case (on, Nil) => Nil
  case (on, (x, c) :: vas) =>
    (if (eq[A](on, c)) Nil else shows_prec[A](Zero_nat, c, Nil)) ++
      shows_prec[B](Zero_nat, x, Nil) ::
      show_pvars[A, B](on, vas)
}

def shows_lpoly[A: equal: show, B: show](ze: A, on: A, x2: l_poly[B, A]):
      (List[Char]) => List[Char] =
  (ze, on, x2) match {
  case (ze, on, LPoly(c, cs)) =>
    show_pvars[A, B](on, cs) match {
      case Nil => (a: List[Char]) => shows_prec[A](Zero_nat, c, a)
      case a :: list =>
        comp[List[Char], List[Char],
              List[Char]]((if (eq[A](c, ze)) id[List[Char]]
                            else comp[List[Char], List[Char],
                                       List[Char]]((aa: List[Char]) =>
             shows_prec[A](Zero_nat, c, aa),
            shows_string.apply(List(' ', '+', ' ')))),
                           shows_list_gen[List[Char]]((aa: List[Char]) =>
                shows_prec_list[Char](Zero_nat, aa),
               Nil, Nil, List(' ', '+', ' '), Nil, a :: list))
    }
}

def check_lpoly_sI[A: equal: show,
                    B: equal: show](mode: Boolean, ze: A, on: A,
                                     ge: A => A => Boolean,
                                     gt: A => A => Boolean, x5: l_poly[B, A],
                                     x6: l_poly[B, A]):
      sum[(List[Char]) => List[Char], Unit] =
  (mode, ze, on, ge, gt, x5, x6) match {
  case (mode, ze, on, ge, gt, LPoly(a, vas), LPoly(b, vbs)) =>
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]](binda[(List[Char]) => List[Char], Unit,
                                 Unit](check[(List[Char]) =>
       List[Char]]((gt(a))(b),
                    shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
     ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
     ' ', 'c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't'))),
(_: Unit) =>
  catcha[B, Unit,
          (List[Char]) =>
            List[Char]](check_pvarsI[A, B](ze, (if (mode) ge else gt), vas,
    vbs),
                         (x: B) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ', 'o', 'f', ' ', 'v',
 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ')),
                 (aa: List[Char]) => shows_prec[B](Zero_nat, x, aa))))),
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
   ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
   ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_lpoly[A,
          B](ze, on, LPoly[A, B](a, vas)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', '>', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_lpoly[A,
        B](ze, on, LPoly[A, B](b, vbs)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, x)))))))
}

def check_polo_sI[A: equal: show, B: show,
                   C: equal: show](mode: Boolean, ze: A, on: A, pl: A => A => A,
                                    ti: A => A => A, mz: A => A,
                                    ge: A => A => Boolean,
                                    gt: A => A => Boolean,
                                    pi: B => nat => (A, List[A]),
                                    x9: (term[B, C], term[B, C])):
      sum[(List[Char]) => List[Char], Unit] =
  (mode, ze, on, pl, ti, mz, ge, gt, pi, x9) match {
  case (mode, ze, on, pl, ti, mz, ge, gt, pi, (s, t)) =>
    {
      val left: l_poly[C, A] = pleftI[A, B, C](ze, on, pl, ti, mz, pi, s)
      val right: l_poly[C, A] = prightI[A, B, C](ze, on, pl, ti, mz, pi, t);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_lpoly_sI[A,
    C](mode, ze, on, ge, gt, left, right),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
     'o', 't', ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_prec_term[B, C](Zero_nat, s),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ', '>', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_prec_term[B, C](Zero_nat, t),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))))))
    }
}

def get_comp_max0[A](x0: (A => Boolean,
                           (A, (A, (A, (A => A => A,
 (A => A => A,
   (A => A,
     (A => A => Boolean,
       (A => A => Boolean,
         (Boolean, (List[Char], (A => Boolean, A => Boolean))))))))))))):
      A => A =
  x0 match {
  case (wfe, (ze, (on, (defa,
                         (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm))))))))))))
    => mz
}

def check_lpoly_nsI[A: equal: show,
                     B: equal: show](ze: A, on: A, ge: A => A => Boolean,
                                      x3: l_poly[B, A], x4: l_poly[B, A]):
      sum[(List[Char]) => List[Char], Unit] =
  (ze, on, ge, x3, x4) match {
  case (ze, on, ge, LPoly(a, vas), LPoly(b, vbs)) =>
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]](binda[(List[Char]) => List[Char], Unit,
                                 Unit](check[(List[Char]) =>
       List[Char]]((ge(a))(b),
                    shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
     ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
     ' ', 'c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't',
     's'))),
(_: Unit) =>
  catcha[B, Unit,
          (List[Char]) =>
            List[Char]](check_pvarsI[A, B](ze, ge, vas, vbs),
                         (x: B) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ', 'o', 'f', ' ', 'v',
 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ')),
                 (aa: List[Char]) => shows_prec[B](Zero_nat, x, aa))))),
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
   ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
   ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_lpoly[A,
          B](ze, on, LPoly[A, B](a, vas)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', '>', '=', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_lpoly[A,
        B](ze, on, LPoly[A, B](b, vbs)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, x)))))))
}

def check_polo_nsI[A: equal: show, B: show,
                    C: equal: show](ze: A, on: A, pl: A => A => A,
                                     ti: A => A => A, mz: A => A,
                                     ge: A => A => Boolean,
                                     pi: B => nat => (A, List[A]),
                                     x7: (term[B, C], term[B, C])):
      sum[(List[Char]) => List[Char], Unit] =
  (ze, on, pl, ti, mz, ge, pi, x7) match {
  case (ze, on, pl, ti, mz, ge, pi, (s, t)) =>
    {
      val left: l_poly[C, A] = pleftI[A, B, C](ze, on, pl, ti, mz, pi, s)
      val right: l_poly[C, A] = prightI[A, B, C](ze, on, pl, ti, mz, pi, t);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_lpoly_nsI[A, C](ze, on, ge, left, right),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
     'o', 't', ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_prec_term[B, C](Zero_nat, s),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ', '>', '=', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_prec_term[B, C](Zero_nat, t),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))))))
    }
}

def create_afI[A: equal,
                B: equal](ze: A, x1: List[((B, nat), (A, List[A]))], f: B,
                           n: nat, i: nat):
      Boolean =
  (ze, x1, f, n, i) match {
  case (ze, Nil, f, n, i) => true
  case (ze, ((g, m), (c, coeffs)) :: rest, f, n, i) =>
    (if (eq[B](f, g) && equal_nata(m, n))
      less_nat(i, size_list[A](coeffs)) && ! (eq[A](nth[A](coeffs, i), ze))
      else create_afI[A, B](ze, rest, f, n, i))
}

def check_poly_monoI[A: show,
                      B: show](ze: A, ge: A => A => Boolean, cm: A => Boolean):
      (List[((B, nat), (A, List[A]))]) =>
        sum[(List[Char]) => List[Char], Unit] =
  (xs: List[((B, nat), (A, List[A]))]) =>
    catcha[(((B, nat), (A, List[A])), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[((B, nat), (A, List[A])),
                                   (List[Char]) =>
                                     List[Char]]((a: ((B, nat), (A, List[A])))
           => {
                val (aa, b): ((B, nat), (A, List[A])) = a;
                ({
                   val (f, n): (B, nat) = aa;
                   (ab: (A, List[A])) =>
                     {
                       val (c, cs): (A, List[A]) = ab;
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](binda[(List[Char]) => List[Char],
            Unit,
            Unit](check[(List[Char]) =>
                          List[Char]]((ge(c))(ze),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('c', 'o', 'n', 's', 't', 'a', 'n', 't',
  ' ', 'p', 'a', 'r', 't', ' ')),
                  comp[List[Char], List[Char],
                        List[Char]]((ac: List[Char]) =>
                                      shows_prec[A](Zero_nat, c, ac),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 'm', 'u', 's', 't', ' ', 'b', 'e',
' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't', ' ')),
                comp[List[Char], List[Char],
                      List[Char]]((ac: List[Char]) =>
                                    shows_prec[A](Zero_nat, ze, ac),
                                   shows_nl))))),
                   (_: Unit) =>
                     binda[(List[Char]) => List[Char], Unit,
                            Unit](check[(List[Char]) =>
  List[Char]](less_eq_nat(n, size_list[A](cs)),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('t', 'h', 'e', ' ',
                  'l', 'a', 's', 't', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
                  't', ' ', 'i', 's', ' ', 'i', 'g', 'n', 'o', 'r', 'e', 'd')),
                                  shows_nl)),
                                   (_: Unit) =>
                                     catcha[(A, (List[Char]) => List[Char]),
     Unit,
     (List[Char]) =>
       List[Char]](forallm[A, (List[Char]) =>
                                List[Char]]((d: A) =>
      check[(List[Char]) =>
              List[Char]](cm(d),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't',
                              ' ')),
      comp[List[Char], List[Char],
            List[Char]]((ac: List[Char]) => shows_prec[A](Zero_nat, d, ac),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', 'i',
                            's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                            'w', 'e', 'd')),
    shows_nl)))),
     cs),
                    (x: (A, (List[Char]) => List[Char])) =>
                      Inl[(List[Char]) => List[Char],
                           Unit](snd[A, (List[Char]) => List[Char]](x))))),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o',
                      'n', 'o', 't', 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ',
                      'd', 'u', 'e', ' ', 't', 'o', ' ', 'i', 'n', 't', 'e',
                      'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ' ',
                      'o', 'f', ' ')),
                                      comp[List[Char], List[Char],
    List[Char]]((ac: List[Char]) => shows_prec[B](Zero_nat, f, ac),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List('/')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_nat(Zero_nat, n),
               comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))
                     }
                 })(b)
              },
          xs),
                           (x: (((B, nat), (A, List[A])),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[((B, nat), (A, List[A])),
        (List[Char]) => List[Char]](x)))

def create_lpoly_repr[A: equal: show,
                       B: equal: show](c:
 (A => Boolean,
   (A, (A, (A, (A => A => A,
                 (A => A => A,
                   (A => A,
                     (A => A => Boolean,
                       (A => A => Boolean,
                         (Boolean,
                           (List[Char], (A => Boolean, A => Boolean)))))))))))),
i: List[((B, nat), (A, List[A]))]):
      (List[Char]) => List[Char] =
  {
    val pi: B => nat => (A, List[A]) =
      (a: B) =>
        (b: nat) =>
          to_lpoly_inter[A, B](get_comp_def[A](c), get_comp_1[A](c), i, a, b);
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List('p', 'o', 'l', 'y', 'n', 'o', 'm',
       'i', 'a', 'l', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'r',
       'a', 't', 'i', 'o', 'n', ' ', 'o', 'v', 'e', 'r', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(get_comp_description[A](c)),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     shows_sep[(B, nat)]((a: (B, nat)) =>
   {
     val (f, n): (B, nat) = a
     val t: term[B, List[Char]] =
       Fun[B, List[Char]](f, mapa[List[Char],
                                   term[B,
 List[Char]]]((aa: List[Char]) => Var[List[Char], B](aa),
               fresh_strings_list(List('x', '_'), one_nat, Nil, n)));
     comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('P', 'o', 'l', '(')),
                        comp[List[Char], List[Char],
                              List[Char]](shows_prec_term[B,
                   List[Char]](Zero_nat, t),
   comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List(')', ' ', '=', ' ')),
                      shows_lpoly[A, List[Char]](get_comp_0[A](c),
          get_comp_1[A](c),
          pleftI[A, B,
                  List[Char]](get_comp_0[A](c), get_comp_1[A](c),
                               get_comp_plus[A](c), get_comp_times[A](c),
                               get_comp_max0[A](c), pi, t)))))
   },
  shows_nl,
  remdups[(B, nat)](mapa[((B, nat), (A, List[A])),
                          (B, nat)]((a: ((B, nat), (A, List[A]))) =>
                                      fst[(B, nat), (A, List[A])](a),
                                     i))))))
  }

def create_poly_redpair[A: equal: show, B: equal: show,
                         C: equal: show](cI:
   (List[((A, nat), (B, List[B]))]) => sum[(List[Char]) => List[Char], Unit],
  c: (B => Boolean,
       (B, (B, (B, (B => B => B,
                     (B => B => B,
                       (B => B,
                         (B => B => Boolean,
                           (B => B => Boolean,
                             (Boolean,
                               (List[Char],
                                 (B => Boolean, B => Boolean)))))))))))),
  i: List[((A, nat), (B, List[B]))]):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[A, C], term[A, C])) => sum[(List[Char]) => List[Char], Unit],
          (((term[A, C], term[A, C])) => sum[(List[Char]) => List[Char], Unit],
            (A => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  {
    val pi: A => nat => (B, List[B]) =
      (a: A) =>
        (b: nat) =>
          to_lpoly_inter[B, A](get_comp_def[B](c), get_comp_1[B](c), i, a, b);
    (binda[(List[Char]) => List[Char], Unit,
            Unit](cI(i),
                   (_: Unit) =>
                     check_lpoly_coeffs[B,
 A](get_comp_wf_elem[B](c), get_comp_apos[B](c),
     (x: B) => (get_comp_ge[B](c)).apply(x).apply(get_comp_0[B](c)), i)),
      ((a: (term[A, C], term[A, C])) =>
         check_polo_sI[B, A,
                        C](get_comp_mono[B](c), get_comp_0[B](c),
                            get_comp_1[B](c), get_comp_plus[B](c),
                            get_comp_times[B](c), get_comp_max0[B](c),
                            get_comp_ge[B](c), get_comp_gt[B](c), pi, a),
        ((a: (term[A, C], term[A, C])) =>
           check_polo_nsI[B, A,
                           C](get_comp_0[B](c), get_comp_1[B](c),
                               get_comp_plus[B](c), get_comp_times[B](c),
                               get_comp_max0[B](c), get_comp_ge[B](c), pi, a),
          ((a: A) =>
             (b: nat) =>
               (d: nat) => create_afI[B, A](get_comp_0[B](c), i, a, b, d),
            (binda[(List[Char]) => List[Char], Unit,
                    Unit](check[(List[Char]) =>
                                  List[Char]](get_comp_mono[B](c),
       shows_string.apply(List('m', 'o', 'n', 'o', 't', 'o', 'n', 'e', ' ', 'p',
                                'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l',
                                's', ' ', 'n', 'o', 't', ' ', 's', 'u', 'p',
                                'p', 'o', 'r', 't', ' ', 'f', 'o', 'r', ' ',
                                't', 'h', 'i', 's', ' ', 'd', 'o', 'm', 'a',
                                'i', 'n'))),
                           (_: Unit) =>
                             (check_poly_monoI[B,
        A](get_comp_0[B](c), get_comp_ge[B](c),
            get_comp_check_mono[B](c))).apply(i)),
              create_lpoly_repr[B, A](c, i))))))
  }

def poly_inter_list_to_inter[A: poly_carriera,
                              B: equal](defa: A,
 i: List[((B, nat), List[(List[(nat, nat)], A)])]):
      B => nat => List[(List[(nat, nat)], A)] =
  (f: B) =>
    (n: nat) =>
      lookupa[(B, nat), List[(List[(nat, nat)], A)]]((f, n), i) match {
        case None => default_I[A](defa, n)
        case Some(p) => p
      }

def check_poly_weak_mono_discrete[A: equal,
                                   B: equal: poly_carrier](p:
                     List[(List[(A, nat)], B)],
                    v: A):
      Boolean =
  check_poly_ge[A, B](poly_subst[A, A,
                                  B]((w: A) =>
                                       poly_of[A,
        B]((if (eq[A](w, v)) PSum[A, B](List(PNum[B, A](one[B]), PVar[A, B](v)))
             else PVar[A, B](w))),
                                      p),
                       p)

def check_poly_weak_mono_all[A, B: ordered_semiring_0](p:
                 List[(List[(A, nat)], B)]):
      Boolean =
  list_all[(List[(A, nat)],
             B)]((a: (List[(A, nat)], B)) =>
                   {
                     val (_, c): (List[(A, nat)], B) = a;
                     ge[B](c, zero[B])
                   },
                  p)

def check_poly_weak_mono_and_pos[A: equal,
                                  B: equal: poly_carrier](discrete: Boolean,
                   p: List[(List[(A, nat)], B)]):
      Boolean =
  (if (discrete)
    list_all[A]((a: A) => check_poly_weak_mono_discrete[A, B](p, a),
                 poly_vars_list[A, B](p)) &&
      ge[B](eval_poly[A, B]((_: A) => zero[B], p), zero[B])
    else check_poly_weak_mono_all[A, B](p))

def check_poly_inter_list[A, B: equal: poly_carriera](discrete: Boolean,
               i: List[((A, nat), List[(List[(nat, nat)], B)])]):
      sum[(A, List[(List[(nat, nat)], B)]), Unit] =
  catcha[((A, nat), List[(List[(nat, nat)], B)]), Unit,
          (A, List[(List[(nat, nat)],
                     B)])](catcha[(((A, nat), List[(List[(nat, nat)], B)]),
                                    ((A, nat), List[(List[(nat, nat)], B)])),
                                   Unit,
                                   ((A, nat),
                                     List[(List[(nat, nat)],
    B)])](forallm[((A, nat), List[(List[(nat, nat)], B)]),
                   ((A, nat),
                     List[(List[(nat, nat)],
                            B)])]((x: ((A, nat), List[(List[(nat, nat)], B)]))
                                    => (if ({
      val (_, a): ((A, nat), List[(List[(nat, nat)], B)]) = x;
      check_poly_weak_mono_and_pos[nat, B](discrete, a)
    })
 Inr[Unit, ((A, nat), List[(List[(nat, nat)], B)])](())
 else Inl[((A, nat), List[(List[(nat, nat)], B)]), Unit](x)),
                                   i),
           (x: (((A, nat), List[(List[(nat, nat)], B)]),
                 ((A, nat), List[(List[(nat, nat)], B)])))
             => Inl[((A, nat), List[(List[(nat, nat)], B)]),
                     Unit](snd[((A, nat), List[(List[(nat, nat)], B)]),
                                ((A, nat), List[(List[(nat, nat)], B)])](x))),
                            (x: ((A, nat), List[(List[(nat, nat)], B)])) =>
                              Inl[(A, List[(List[(nat, nat)], B)]),
                                   Unit]({
   val (a, b): ((A, nat), List[(List[(nat, nat)], B)]) = x;
   ({
      val (f, _): (A, nat) = a;
      (aa: List[(List[(nat, nat)], B)]) => (f, aa)
    })(b)
 }))

def poly_inter_to_af[A, B](i: A => nat => List[(List[(nat, nat)], B)], f: A,
                            n: nat):
      nat => Boolean =
  (a: nat) => poly_var[nat, B]((i(f))(n), a)

def check_poly_strict_mono_discrete[A: equal: poly_carrier,
                                     B: equal](gt: A => A => Boolean,
        p: List[(List[(B, nat)], A)], v: B):
      Boolean =
  check_poly_gt[A, B](gt, poly_subst[B, B,
                                      A]((w: B) =>
   poly_of[B, A]((if (eq[B](w, v))
                   PSum[B, A](List(PNum[A, B](one[A]), PVar[B, A](v)))
                   else PVar[B, A](w))),
  p),
                       p)

def check_monom_strict_mono[A: equal](pm: Boolean, m: List[(A, nat)], v: A):
      Boolean =
  ! (nulla[(A, nat)](m)) &&
    (nulla[(A, nat)](tl[(A, nat)](m)) &&
      (eq[A](fst[A, nat](hd[(A, nat)](m)), v) &&
        (if (pm) less_eq_nat(one_nat, snd[A, nat](hd[(A, nat)](m)))
          else equal_nata(snd[A, nat](hd[(A, nat)](m)), one_nat))))

def check_poly_strict_mono[A: equal,
                            B: poly_carrier](pm: Boolean,
      p: List[(List[(A, nat)], B)], v: A):
      Boolean =
  list_ex[(List[(A, nat)],
            B)]((a: (List[(A, nat)], B)) =>
                  {
                    val (m, c): (List[(A, nat)], B) = a;
                    ge[B](c, one[B]) && check_monom_strict_mono[A](pm, m, v)
                  },
                 p)

def check_poly_strict_mono_smart[A: equal: poly_carrier,
                                  B: equal](discrete: Boolean, pm: Boolean,
     gt: A => A => Boolean, p: List[(List[(B, nat)], A)], v: B):
      Boolean =
  (if (discrete) check_poly_strict_mono_discrete[A, B](gt, p, v)
    else check_poly_strict_mono[B, A](pm, p, v))

def create_nlpoly_redpair[A: equal: poly_carriera: show, B: equal: show,
                           C: equal: show](defa: A, gt: A => A => Boolean,
    power_mono: Boolean, discrete: Boolean,
    i: List[((B, nat), List[(List[(nat, nat)], A)])]):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
          (((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit],
            (B => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  {
    val j: B => nat => List[(List[(nat, nat)], A)] =
      poly_inter_list_to_inter[A, B](defa, i)
    val x: (List[(List[(nat, nat)], A)]) => List[(List[(List[Char], nat)], A)] =
      (a: List[(List[(nat, nat)], A)]) =>
        poly_subst[nat, List[Char],
                    A]((n: nat) =>
                         poly_of[List[Char],
                                  A](PVar[List[Char],
   A](List('x', '_') ++ (shows_prec_nat(Zero_nat, n)).apply(Nil))),
                        a);
    (catcha[(B, List[(List[(nat, nat)], A)]), Unit,
             (List[Char]) =>
               List[Char]](check_poly_inter_list[B, A](discrete, i),
                            (xa: (B, List[(List[(nat, nat)], A)])) =>
                              Inl[(List[Char]) => List[Char],
                                   Unit]({
   val (f, p): (B, List[(List[(nat, nat)], A)]) = xa;
   comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('i', 'n', 't', 'e', 'r', 'p', 'r',
      'e', 't', 'a', 't', 'i', 'o', 'n', ' ')),
                      comp[List[Char], List[Char],
                            List[Char]](shows_poly[List[Char], A](x(p)),
 comp[List[Char], List[Char],
       List[Char]](shows_string.apply(List(' ', 'o', 'f', ' ')),
                    comp[List[Char], List[Char],
                          List[Char]]((a: List[Char]) =>
shows_prec[B](Zero_nat, f, a),
                                       shows_string.apply(List(' ', 'i', 'n',
                        'v', 'a', 'l', 'i', 'd', ' '))))))
 })),
      ((a: (term[B, C], term[B, C])) =>
         {
           val (s, t): (term[B, C], term[B, C]) = a
           val p: List[(List[(C, nat)], A)] = eval_term[B, A, C](j, s)
           val q: List[(List[(C, nat)], A)] = eval_term[B, A, C](j, t);
           check[(List[Char]) =>
                   List[Char]](check_poly_gt[A, C](gt, p, q),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('c',
                                   'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ',
                                   'e', 'n', 's', 'u', 'r', 'e', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_poly[C, A](p),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List(' ',
                                 '>', ' ')),
         shows_poly[C, A](q)))))
         },
        ((a: (term[B, C], term[B, C])) =>
           {
             val (s, t): (term[B, C], term[B, C]) = a
             val p: List[(List[(C, nat)], A)] = eval_term[B, A, C](j, s)
             val q: List[(List[(C, nat)], A)] = eval_term[B, A, C](j, t);
             check[(List[Char]) =>
                     List[Char]](check_poly_ge[C, A](p, q),
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                                     ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ')),
             comp[List[Char], List[Char],
                   List[Char]](shows_poly[C, A](p),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List(' ',
                                   '>', '=', ' ')),
           shows_poly[C, A](q)))))
           },
          ((a: B) => (b: nat) => poly_inter_to_af[B, A](j, a, b),
            (catcha[((B, nat), List[(List[(nat, nat)], A)]), Unit,
                     (List[Char]) =>
                       List[Char]](catcha[(((B, nat),
     List[(List[(nat, nat)], A)]),
    ((B, nat), List[(List[(nat, nat)], A)])),
   Unit,
   ((B, nat),
     List[(List[(nat, nat)],
            A)])](forallm[((B, nat), List[(List[(nat, nat)], A)]),
                           ((B, nat),
                             List[(List[(nat, nat)],
                                    A)])]((xa:
     ((B, nat), List[(List[(nat, nat)], A)]))
    => (if ({
              val (a, b): ((B, nat), List[(List[(nat, nat)], A)]) = xa;
              ({
                 val (_, n): (B, nat) = a;
                 (p: List[(List[(nat, nat)], A)]) =>
                   all_interval_nat((aa: nat) =>
                                      check_poly_strict_mono_smart[A,
                            nat](discrete, power_mono, gt, p, aa),
                                     Zero_nat, n)
               })(b)
            })
         Inr[Unit, ((B, nat), List[(List[(nat, nat)], A)])](())
         else Inl[((B, nat), List[(List[(nat, nat)], A)]), Unit](xa)),
   i),
                   (xa: (((B, nat), List[(List[(nat, nat)], A)]),
                          ((B, nat), List[(List[(nat, nat)], A)])))
                     => Inl[((B, nat), List[(List[(nat, nat)], A)]),
                             Unit](snd[((B, nat), List[(List[(nat, nat)], A)]),
((B, nat), List[(List[(nat, nat)], A)])](xa))),
                                    (xa: ((B, nat),
   List[(List[(nat, nat)], A)]))
                                      => Inl[(List[Char]) => List[Char],
      Unit]({
              val (a, b): ((B, nat), List[(List[(nat, nat)], A)]) = xa;
              ({
                 val (f, _): (B, nat) = a;
                 (p: List[(List[(nat, nat)], A)]) =>
                   comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('c', 'o', 'u', 'l',
                      'd', ' ', 'n', 'o', 't', ' ', 'e', 'n', 's', 'u', 'r',
                      'e', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'i', 'c',
                      't', 'y', ' ', 'o', 'f', ' ')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_poly[List[Char], A](x(p)),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(' ', 'a', 's', ' ',
                    'i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i',
                    'o', 'n', ' ', 'o', 'f', ' ')),
                                    (aa: List[Char]) =>
                                      shows_prec[B](Zero_nat, f, aa))))
               })(b)
            })),
              comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('p', 'o', 'l', 'y', 'n',
                 'o', 'm', 'i', 'a', 'l', ' ', 'i', 'n', 't', 'e', 'r', 'p',
                 'r', 'e', 't', 'a', 't', 'i', 'o', 'n')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_nl,
            shows_sep[((B, nat),
                        List[(List[(nat, nat)],
                               A)])]((a:
((B, nat), List[(List[(nat, nat)], A)]))
                                       => {
    val (aa, b): ((B, nat), List[(List[(nat, nat)], A)]) = a;
    ({
       val (f, n): (B, nat) = aa;
       (p: List[(List[(nat, nat)], A)]) =>
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List('P', 'o', 'l', '(')),
                            comp[List[Char], List[Char],
                                  List[Char]]((ab: List[Char]) =>
        shows_prec[B](Zero_nat, f, ab),
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('/')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_prec_nat(Zero_nat, n),
     comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List(')', ' ', '=', ' ')),
                        shows_poly[List[Char], A](x(p)))))))
     })(b)
  },
                                      shows_nl, i))))))))
  }

implicit def ordered_semiring_0_int: ordered_semiring_0[BigInt] = new
  ordered_semiring_0[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
}

implicit def ordered_semiring_1_int: ordered_semiring_1[BigInt] = new
  ordered_semiring_1[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def poly_carrier_int: poly_carrier[BigInt] = new poly_carrier[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def max_ordered_semiring_1_int: max_ordered_semiring_1[BigInt] = new
  max_ordered_semiring_1[BigInt] {
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.max0` = (a: BigInt) => max0_int(a)
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
}

implicit def poly_carrier_inta: poly_carriera[BigInt] = new
  poly_carriera[BigInt] {
  val `Ceta.max0` = (a: BigInt) => max0_int(a)
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.zero` = zero_inta
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def ordered_semiring_0_rat: ordered_semiring_0[rat] = new
  ordered_semiring_0[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
}

implicit def ordered_semiring_1_rat: ordered_semiring_1[rat] = new
  ordered_semiring_1[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def poly_carrier_rat: poly_carrier[rat] = new poly_carrier[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def max_ordered_semiring_1_rat: max_ordered_semiring_1[rat] = new
  max_ordered_semiring_1[rat] {
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.max0` = (a: rat) => max0_rat(a)
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
}

implicit def poly_carrier_rata: poly_carriera[rat] = new poly_carriera[rat] {
  val `Ceta.max0` = (a: rat) => max0_rat(a)
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

def faulty_redpair_impl[A: show, B: show](s: List[Char]):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
          (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
            (A => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  (Inl[(List[Char]) => List[Char], Unit](shows_string.apply(s)),
    ((_: (term[A, B], term[A, B])) => Inr[Unit, (List[Char]) => List[Char]](()),
      ((_: (term[A, B], term[A, B])) =>
         Inr[Unit, (List[Char]) => List[Char]](()),
        ((a: A) => (b: nat) => (c: nat) => full_af[A](a, b, c),
          (Inr[Unit, (List[Char]) => List[Char]](()), id[List[Char]])))))

def shows_arctic(x0: arctic): (List[Char]) => List[Char] = x0 match {
  case Num_arc(i) => shows_prec_int(Zero_nat, i)
  case MinInfty => shows_string.apply(List('-', 'i', 'n', 'f'))
}

def shows_prec_arctic(d: nat, ai: arctic): (List[Char]) => List[Char] =
  shows_arctic(ai)

def shows_list_arctic(as: List[arctic]): (List[Char]) => List[Char] =
  shows_list_aux[arctic]((a: arctic) => shows_prec_arctic(Zero_nat, a), as)

implicit def show_arctic: show[arctic] = new show[arctic] {
  val `Ceta.shows_prec` = (a: nat, b: arctic, c: List[Char]) =>
    (shows_prec_arctic(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[arctic], b: List[Char]) =>
    (shows_list_arctic(a)).apply(b)
}

def shows_arctic_rat(x0: arctic_rat): (List[Char]) => List[Char] = x0 match {
  case Num_arc_rat(i) => shows_prec_rat(Zero_nat, i)
  case MinInfty_rat => shows_string.apply(List('-', 'i', 'n', 'f'))
}

def shows_prec_arctic_rat(d: nat, ari: arctic_rat): (List[Char]) => List[Char] =
  shows_arctic_rat(ari)

def shows_list_arctic_rat(as: List[arctic_rat]): (List[Char]) => List[Char] =
  shows_list_aux[arctic_rat]((a: arctic_rat) =>
                               shows_prec_arctic_rat(Zero_nat, a),
                              as)

implicit def show_arctic_rat: show[arctic_rat] = new show[arctic_rat] {
  val `Ceta.shows_prec` = (a: nat, b: arctic_rat, c: List[Char]) =>
    (shows_prec_arctic_rat(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[arctic_rat], b: List[Char]) =>
    (shows_list_arctic_rat(a)).apply(b)
}

def equal_filtereda[A: equal](x0: filtered[A], x1: filtered[A]): Boolean =
  (x0, x1) match {
  case (FPair(fa, nata), FPair(f, nat)) => eq[A](fa, f) && equal_nata(nata, nat)
}

implicit def equal_filtered[A: equal]: equal[filtered[A]] = new
  equal[filtered[A]] {
  val `Ceta.equal` = (a: filtered[A], b: filtered[A]) =>
    equal_filtereda[A](a, b)
}

def prec_repr_to_pr[A: equal](prs: List[((A, nat), (nat, order_tag))],
                               x1: filtered[A], uu: nat):
      nat =
  (prs, x1, uu) match {
  case (prs, FPair(f, a), uu) =>
    lookupa[(A, nat), (nat, order_tag)]((f, a), prs) match {
      case None => Zero_nat
      case Some(aa) => fst[nat, order_tag](aa)
    }
}

def prec_repr_to_status[A: equal](prs: List[((A, nat), (nat, order_tag))],
                                   x1: filtered[A], uu: nat):
      order_tag =
  (prs, x1, uu) match {
  case (prs, FPair(f, a), uu) =>
    lookupa[(A, nat), (nat, order_tag)]((f, a), prs) match {
      case None => Lex
      case Some(aa) => snd[nat, order_tag](aa)
    }
}

def get_redpair[A: equal: show, B: equal: show](x0: redpair[A]):
      (sum[(List[Char]) => List[Char], Unit],
        (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
          (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
            (A => nat => nat => Boolean,
              (sum[(List[Char]) => List[Char], Unit],
                (List[Char]) => List[Char]))))) =
  x0 match {
  case Int_carrier(i) =>
    create_poly_redpair[A, BigInt,
                         B]((_: List[((A, nat), (BigInt, List[BigInt]))]) =>
                              Inr[Unit, (List[Char]) => List[Char]](()),
                             ((_: BigInt) => true,
                               (BigInt(0),
                                 (BigInt(1),
                                   (BigInt(1),
                                     ((a: BigInt) => (b: BigInt) => a + b,
                                       ((a: BigInt) => (b: BigInt) => a * b,
 ((a: BigInt) => max0_int(a),
   ((a: BigInt) => (b: BigInt) => ge_int(a, b),
     ((x: BigInt) => (y: BigInt) => y < x,
       (true,
         (int_name,
           ((_: BigInt) => true, (a: BigInt) => int_mono(a))))))))))))),
                             i)
  case Int_nl_carrier(i) =>
    create_nlpoly_redpair[BigInt, A,
                           B](BigInt(1), (x: BigInt) => (y: BigInt) => y < x,
                               true, true, i)
  case Rat_carrier(d, i) =>
    create_poly_redpair[A, rat,
                         B]((_: List[((A, nat), (rat, List[rat]))]) =>
                              check[(List[Char]) =>
                                      List[Char]](less_rat(fract(BigInt(0),
                          BigInt(1)),
                    d),
           shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd',
                                    'e', 'l', 't', 'a'))),
                             ((_: rat) => true,
                               (fract(BigInt(0), BigInt(1)),
                                 (fract(BigInt(1), BigInt(1)),
                                   (d, ((a: rat) => (b: rat) => plus_rata(a, b),
 ((a: rat) => (b: rat) => times_rata(a, b),
   ((a: rat) => max0_rat(a),
     ((a: rat) => (b: rat) => ge_rat(a, b),
       (rat_gt(d),
         (true,
           (rat_name, ((_: rat) => true, (a: rat) => rat_mono(a))))))))))))),
                             i)
  case Rat_nl_carrier(d, i) =>
    (if (less_rat(fract(BigInt(0), BigInt(1)), d))
      create_nlpoly_redpair[rat, A,
                             B](d, rat_gt(d),
                                 less_eq_rat(fract(BigInt(1), BigInt(1)), d),
                                 false, i)
      else faulty_redpair_impl[A, B](List('d', 'e', 'l', 't', 'a', ' ', 'm',
   'u', 's', 't', ' ', 'b', 'e', ' ', 'l', 'a', 'r', 'g', 'e', 'r', ' ', 't',
   'h', 'a', 'n', ' ', 'z', 'e', 'r', 'o')))
  case Arctic_carrier(i) =>
    create_poly_redpair[A, arctic,
                         B]((_: List[((A, nat), (arctic, List[arctic]))]) =>
                              Inr[Unit, (List[Char]) => List[Char]](()),
                             ((_: arctic) => true,
                               (zero_arctic,
                                 (one_arctic,
                                   (one_arctic,
                                     ((a: arctic) =>
(b: arctic) => plus_arctic(a, b),
                                       ((a: arctic) =>
  (b: arctic) => times_arctic(a, b),
 (id[arctic],
   ((a: arctic) => (b: arctic) => ge_arctic(a, b),
     ((a: arctic) => (b: arctic) => gt_arctic(a, b),
       (false,
         (arctic_name,
           ((a: arctic) => pos_arctic(a), (_: arctic) => false)))))))))))),
                             i)
  case Arctic_rat_carrier(d, i) =>
    create_poly_redpair[A, arctic_rat,
                         B]((_: List[((A, nat),
                                       (arctic_rat, List[arctic_rat]))])
                              => check[(List[Char]) =>
 List[Char]](less_rat(fract(BigInt(0), BigInt(1)), d),
              shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i', 'd', ' ',
                                       'd', 'e', 'l', 't', 'a'))),
                             ((_: arctic_rat) => true,
                               (zero_arctic_rat,
                                 (one_arctic_rat,
                                   (one_arctic_rat,
                                     ((a: arctic_rat) =>
(b: arctic_rat) => plus_arctic_rat(a, b),
                                       ((a: arctic_rat) =>
  (b: arctic_rat) => times_arctic_rat(a, b),
 (id[arctic_rat],
   ((a: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(a, b),
     ((a: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, a, b),
       (false,
         (arctic_rat_name,
           ((a: arctic_rat) => pos_arctic_rat(a),
             (_: arctic_rat) => false)))))))))))),
                             i)
  case Int_mat_carrier(n, sd, i) =>
    create_poly_redpair[A, List[List[BigInt]],
                         B]((_: List[((A, nat),
                                       (List[List[BigInt]],
 List[List[List[BigInt]]]))])
                              => binda[(List[Char]) => List[Char], Unit,
Unit](Inr[Unit, (List[Char]) => List[Char]](()),
       (_: Unit) =>
         check[(List[Char]) =>
                 List[Char]](less_nat(Zero_nat, sd) && less_eq_nat(sd, n),
                              shows_string.apply(List('i', 'n', 'v', 'a', 'l',
               'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n')))),
                             ((a: List[List[BigInt]]) => mat[BigInt](n, n, a),
                               (mat0I[BigInt](BigInt(0), n, n),
                                 (mat1I[BigInt](BigInt(0), BigInt(1), n),
                                   (mat1I[BigInt](BigInt(0),
           get_comp_def[BigInt](((_: BigInt) => true,
                                  (BigInt(0),
                                    (BigInt(1),
                                      (BigInt(1),
((a: BigInt) => (b: BigInt) => a + b,
  ((a: BigInt) => (b: BigInt) => a * b,
    ((a: BigInt) => max0_int(a),
      ((a: BigInt) => (b: BigInt) => ge_int(a, b),
        ((x: BigInt) => (y: BigInt) => y < x,
          (true,
            (int_name,
              ((_: BigInt) => true, (a: BigInt) => int_mono(a)))))))))))))),
           n),
                                     ((a: List[List[BigInt]]) =>
(b: List[List[BigInt]]) =>
  mat_plusI[BigInt]((aa: BigInt) => (ba: BigInt) => aa + ba, a, b),
                                       ((a: List[List[BigInt]]) =>
  (b: List[List[BigInt]]) =>
    mat_multI[BigInt](BigInt(0), (aa: BigInt) => (ba: BigInt) => aa + ba,
                       (aa: BigInt) => (ba: BigInt) => aa * ba, n, a, b),
 (mat_max0[BigInt],
   (mat_ge[BigInt],
     ((a: List[List[BigInt]]) =>
        (b: List[List[BigInt]]) =>
          mat_gtI[BigInt](get_comp_gt[BigInt](((_: BigInt) => true,
        (BigInt(0),
          (BigInt(1),
            (BigInt(1),
              ((aa: BigInt) => (ba: BigInt) => aa + ba,
                ((aa: BigInt) => (ba: BigInt) => aa * ba,
                  ((aa: BigInt) => max0_int(aa),
                    ((aa: BigInt) => (ba: BigInt) => ge_int(aa, ba),
                      ((x: BigInt) => (y: BigInt) => y < x,
                        (true,
                          (int_name,
                            ((_: BigInt) => true,
                              (aa: BigInt) => int_mono(aa)))))))))))))),
                           sd, a, b),
       (get_comp_mono[BigInt](((_: BigInt) => true,
                                (BigInt(0),
                                  (BigInt(1),
                                    (BigInt(1),
                                      ((a: BigInt) => (b: BigInt) => a + b,
((a: BigInt) => (b: BigInt) => a * b,
  ((a: BigInt) => max0_int(a),
    ((a: BigInt) => (b: BigInt) => ge_int(a, b),
      ((x: BigInt) => (y: BigInt) => y < x,
        (true,
          (int_name,
            ((_: BigInt) => true, (a: BigInt) => int_mono(a)))))))))))))),
         (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'w', 'i', 't', 'h', ' ', 's', 't', 'r',
                                   'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n',
                                   's', 'i', 'o', 'n', ' '),
                             (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                             List(' ', 'o', 'v', 'e', 'r', ' '),
                             get_comp_description[BigInt](((_: BigInt) => true,
                    (BigInt(0),
                      (BigInt(1),
                        (BigInt(1),
                          ((a: BigInt) => (b: BigInt) => a + b,
                            ((a: BigInt) => (b: BigInt) => a * b,
                              ((a: BigInt) => max0_int(a),
                                ((a: BigInt) => (b: BigInt) => ge_int(a, b),
                                  ((x: BigInt) => (y: BigInt) => y < x,
                                    (true,
                                      (int_name,
((_: BigInt) => true, (a: BigInt) => int_mono(a)))))))))))))))),
           ((_: List[List[BigInt]]) => true,
             mat_mono[BigInt](sd, get_comp_check_mono[BigInt](((_: BigInt) =>
                         true,
                        (BigInt(0),
                          (BigInt(1),
                            (BigInt(1),
                              ((a: BigInt) => (b: BigInt) => a + b,
                                ((a: BigInt) => (b: BigInt) => a * b,
                                  ((a: BigInt) => max0_int(a),
                                    ((a: BigInt) => (b: BigInt) => ge_int(a, b),
                                      ((x: BigInt) => (y: BigInt) => y < x,
(true,
  (int_name,
    ((_: BigInt) => true, (a: BigInt) => int_mono(a))))))))))))))))))))))))))),
                             i)
  case Rat_mat_carrier(d, n, sd, i) =>
    create_poly_redpair[A, List[List[rat]],
                         B]((_: List[((A, nat),
                                       (List[List[rat]],
 List[List[List[rat]]]))])
                              => binda[(List[Char]) => List[Char], Unit,
Unit](check[(List[Char]) =>
              List[Char]](less_rat(fract(BigInt(0), BigInt(1)), d),
                           shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i',
            'd', ' ', 'd', 'e', 'l', 't', 'a'))),
       (_: Unit) =>
         check[(List[Char]) =>
                 List[Char]](less_nat(Zero_nat, sd) && less_eq_nat(sd, n),
                              shows_string.apply(List('i', 'n', 'v', 'a', 'l',
               'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n')))),
                             ((a: List[List[rat]]) => mat[rat](n, n, a),
                               (mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                                 (mat1I[rat](fract(BigInt(0), BigInt(1)),
      fract(BigInt(1), BigInt(1)), n),
                                   (mat1I[rat](fract(BigInt(0), BigInt(1)),
        get_comp_def[rat](((_: rat) => true,
                            (fract(BigInt(0), BigInt(1)),
                              (fract(BigInt(1), BigInt(1)),
                                (d, ((a: rat) => (b: rat) => plus_rata(a, b),
                                      ((a: rat) => (b: rat) => times_rata(a, b),
((a: rat) => max0_rat(a),
  ((a: rat) => (b: rat) => ge_rat(a, b),
    (rat_gt(d),
      (true, (rat_name, ((_: rat) => true, (a: rat) => rat_mono(a)))))))))))))),
        n),
                                     ((a: List[List[rat]]) =>
(b: List[List[rat]]) =>
  mat_plusI[rat]((aa: rat) => (ba: rat) => plus_rata(aa, ba), a, b),
                                       ((a: List[List[rat]]) =>
  (b: List[List[rat]]) =>
    mat_multI[rat](fract(BigInt(0), BigInt(1)),
                    (aa: rat) => (ba: rat) => plus_rata(aa, ba),
                    (aa: rat) => (ba: rat) => times_rata(aa, ba), n, a, b),
 (mat_max0[rat],
   (mat_ge[rat],
     ((a: List[List[rat]]) =>
        (b: List[List[rat]]) =>
          mat_gtI[rat](get_comp_gt[rat](((_: rat) => true,
  (fract(BigInt(0), BigInt(1)),
    (fract(BigInt(1), BigInt(1)),
      (d, ((aa: rat) => (ba: rat) => plus_rata(aa, ba),
            ((aa: rat) => (ba: rat) => times_rata(aa, ba),
              ((aa: rat) => max0_rat(aa),
                ((aa: rat) => (ba: rat) => ge_rat(aa, ba),
                  (rat_gt(d),
                    (true,
                      (rat_name,
                        ((_: rat) => true,
                          (aa: rat) => rat_mono(aa)))))))))))))),
                        sd, a, b),
       (get_comp_mono[rat](((_: rat) => true,
                             (fract(BigInt(0), BigInt(1)),
                               (fract(BigInt(1), BigInt(1)),
                                 (d, ((a: rat) => (b: rat) => plus_rata(a, b),
                                       ((a: rat) =>
  (b: rat) => times_rata(a, b),
 ((a: rat) => max0_rat(a),
   ((a: rat) => (b: rat) => ge_rat(a, b),
     (rat_gt(d),
       (true,
         (rat_name, ((_: rat) => true, (a: rat) => rat_mono(a)))))))))))))),
         (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'w', 'i', 't', 'h', ' ', 's', 't', 'r',
                                   'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n',
                                   's', 'i', 'o', 'n', ' '),
                             (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                             List(' ', 'o', 'v', 'e', 'r', ' '),
                             get_comp_description[rat](((_: rat) => true,
                 (fract(BigInt(0), BigInt(1)),
                   (fract(BigInt(1), BigInt(1)),
                     (d, ((a: rat) => (b: rat) => plus_rata(a, b),
                           ((a: rat) => (b: rat) => times_rata(a, b),
                             ((a: rat) => max0_rat(a),
                               ((a: rat) => (b: rat) => ge_rat(a, b),
                                 (rat_gt(d),
                                   (true,
                                     (rat_name,
                                       ((_: rat) => true,
 (a: rat) => rat_mono(a)))))))))))))))),
           ((_: List[List[rat]]) => true,
             mat_mono[rat](sd, get_comp_check_mono[rat](((_: rat) => true,
                  (fract(BigInt(0), BigInt(1)),
                    (fract(BigInt(1), BigInt(1)),
                      (d, ((a: rat) => (b: rat) => plus_rata(a, b),
                            ((a: rat) => (b: rat) => times_rata(a, b),
                              ((a: rat) => max0_rat(a),
                                ((a: rat) => (b: rat) => ge_rat(a, b),
                                  (rat_gt(d),
                                    (true,
                                      (rat_name,
((_: rat) => true, (a: rat) => rat_mono(a))))))))))))))))))))))))))),
                             i)
  case Arc_mat_carrier(n, i) =>
    create_poly_redpair[A, List[List[arctic]],
                         B]((_: List[((A, nat),
                                       (List[List[arctic]],
 List[List[List[arctic]]]))])
                              => binda[(List[Char]) => List[Char], Unit,
Unit](Inr[Unit, (List[Char]) => List[Char]](()),
       (_: Unit) =>
         check[(List[Char]) =>
                 List[Char]](less_nat(Zero_nat, n),
                              shows_string.apply(List('i', 'n', 'v', 'a', 'l',
               'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n')))),
                             ((a: List[List[arctic]]) => mat[arctic](n, n, a),
                               (mat0I[arctic](zero_arctic, n, n),
                                 (mat1I[arctic](zero_arctic, one_arctic, n),
                                   (mat1I[arctic](zero_arctic,
           get_comp_def[arctic](((_: arctic) => true,
                                  (zero_arctic,
                                    (one_arctic,
                                      (one_arctic,
((a: arctic) => (b: arctic) => plus_arctic(a, b),
  ((a: arctic) => (b: arctic) => times_arctic(a, b),
    (id[arctic],
      ((a: arctic) => (b: arctic) => ge_arctic(a, b),
        ((a: arctic) => (b: arctic) => gt_arctic(a, b),
          (false,
            (arctic_name,
              ((a: arctic) => pos_arctic(a), (_: arctic) => false))))))))))))),
           n),
                                     ((a: List[List[arctic]]) =>
(b: List[List[arctic]]) =>
  mat_plusI[arctic]((aa: arctic) => (ba: arctic) => plus_arctic(aa, ba), a, b),
                                       ((a: List[List[arctic]]) =>
  (b: List[List[arctic]]) =>
    mat_multI[arctic](zero_arctic,
                       (aa: arctic) => (ba: arctic) => plus_arctic(aa, ba),
                       (aa: arctic) => (ba: arctic) => times_arctic(aa, ba), n,
                       a, b),
 (id[List[List[arctic]]],
   (mat_ge[arctic],
     ((a: List[List[arctic]]) =>
        (b: List[List[arctic]]) =>
          mat_comp_all[arctic](get_comp_gt[arctic](((_: arctic) => true,
             (zero_arctic,
               (one_arctic,
                 (one_arctic,
                   ((aa: arctic) => (ba: arctic) => plus_arctic(aa, ba),
                     ((aa: arctic) => (ba: arctic) => times_arctic(aa, ba),
                       (id[arctic],
                         ((aa: arctic) => (ba: arctic) => ge_arctic(aa, ba),
                           ((aa: arctic) => (ba: arctic) => gt_arctic(aa, ba),
                             (false,
                               (arctic_name,
                                 ((aa: arctic) => pos_arctic(aa),
                                   (_: arctic) => false))))))))))))),
                                a, b),
       (false,
         (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'o', 'v', 'e', 'r', ' '),
                             get_comp_description[arctic](((_: arctic) => true,
                    (zero_arctic,
                      (one_arctic,
                        (one_arctic,
                          ((a: arctic) => (b: arctic) => plus_arctic(a, b),
                            ((a: arctic) => (b: arctic) => times_arctic(a, b),
                              (id[arctic],
                                ((a: arctic) => (b: arctic) => ge_arctic(a, b),
                                  ((a: arctic) =>
                                     (b: arctic) => gt_arctic(a, b),
                                    (false,
                                      (arctic_name,
((a: arctic) => pos_arctic(a), (_: arctic) => false))))))))))))))),
           ((a: List[List[arctic]]) =>
              mat_arc_posI[arctic](get_comp_apos[arctic](((_: arctic) => true,
                   (zero_arctic,
                     (one_arctic,
                       (one_arctic,
                         ((aa: arctic) => (b: arctic) => plus_arctic(aa, b),
                           ((aa: arctic) => (b: arctic) => times_arctic(aa, b),
                             (id[arctic],
                               ((aa: arctic) => (b: arctic) => ge_arctic(aa, b),
                                 ((aa: arctic) =>
                                    (b: arctic) => gt_arctic(aa, b),
                                   (false,
                                     (arctic_name,
                                       ((aa: arctic) => pos_arctic(aa),
 (_: arctic) => false))))))))))))),
                                    a),
             (_: List[List[arctic]]) => false)))))))))))),
                             i)
  case Arc_rat_mat_carrier(d, n, i) =>
    create_poly_redpair[A, List[List[arctic_rat]],
                         B]((_: List[((A, nat),
                                       (List[List[arctic_rat]],
 List[List[List[arctic_rat]]]))])
                              => binda[(List[Char]) => List[Char], Unit,
Unit](check[(List[Char]) =>
              List[Char]](less_rat(fract(BigInt(0), BigInt(1)), d),
                           shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i',
            'd', ' ', 'd', 'e', 'l', 't', 'a'))),
       (_: Unit) =>
         check[(List[Char]) =>
                 List[Char]](less_nat(Zero_nat, n),
                              shows_string.apply(List('i', 'n', 'v', 'a', 'l',
               'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n')))),
                             ((a: List[List[arctic_rat]]) =>
                                mat[arctic_rat](n, n, a),
                               (mat0I[arctic_rat](zero_arctic_rat, n, n),
                                 (mat1I[arctic_rat](zero_arctic_rat,
             one_arctic_rat, n),
                                   (mat1I[arctic_rat](zero_arctic_rat,
               get_comp_def[arctic_rat](((_: arctic_rat) => true,
  (zero_arctic_rat,
    (one_arctic_rat,
      (one_arctic_rat,
        ((a: arctic_rat) => (b: arctic_rat) => plus_arctic_rat(a, b),
          ((a: arctic_rat) => (b: arctic_rat) => times_arctic_rat(a, b),
            (id[arctic_rat],
              ((a: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(a, b),
                ((a: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, a, b),
                  (false,
                    (arctic_rat_name,
                      ((a: arctic_rat) => pos_arctic_rat(a),
                        (_: arctic_rat) => false))))))))))))),
               n),
                                     ((a: List[List[arctic_rat]]) =>
(b: List[List[arctic_rat]]) =>
  mat_plusI[arctic_rat]((aa: arctic_rat) =>
                          (ba: arctic_rat) => plus_arctic_rat(aa, ba),
                         a, b),
                                       ((a: List[List[arctic_rat]]) =>
  (b: List[List[arctic_rat]]) =>
    mat_multI[arctic_rat](zero_arctic_rat,
                           (aa: arctic_rat) =>
                             (ba: arctic_rat) => plus_arctic_rat(aa, ba),
                           (aa: arctic_rat) =>
                             (ba: arctic_rat) => times_arctic_rat(aa, ba),
                           n, a, b),
 (id[List[List[arctic_rat]]],
   (mat_ge[arctic_rat],
     ((a: List[List[arctic_rat]]) =>
        (b: List[List[arctic_rat]]) =>
          mat_comp_all[arctic_rat](get_comp_gt[arctic_rat](((_: arctic_rat) =>
                      true,
                     (zero_arctic_rat,
                       (one_arctic_rat,
                         (one_arctic_rat,
                           ((aa: arctic_rat) =>
                              (ba: arctic_rat) => plus_arctic_rat(aa, ba),
                             ((aa: arctic_rat) =>
                                (ba: arctic_rat) => times_arctic_rat(aa, ba),
                               (id[arctic_rat],
                                 ((aa: arctic_rat) =>
                                    (ba: arctic_rat) => ge_arctic_rat(aa, ba),
                                   ((aa: arctic_rat) =>
                                      (ba: arctic_rat) =>
gt_arctic_rat(d, aa, ba),
                                     (false,
                                       (arctic_rat_name,
 ((aa: arctic_rat) => pos_arctic_rat(aa), (_: arctic_rat) => false))))))))))))),
                                    a, b),
       (false,
         (concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'o', 'v', 'e', 'r', ' '),
                             get_comp_description[arctic_rat](((_: arctic_rat)
                         => true,
                        (zero_arctic_rat,
                          (one_arctic_rat,
                            (one_arctic_rat,
                              ((a: arctic_rat) =>
                                 (b: arctic_rat) => plus_arctic_rat(a, b),
                                ((a: arctic_rat) =>
                                   (b: arctic_rat) => times_arctic_rat(a, b),
                                  (id[arctic_rat],
                                    ((a: arctic_rat) =>
                                       (b: arctic_rat) => ge_arctic_rat(a, b),
                                      ((a: arctic_rat) =>
 (b: arctic_rat) => gt_arctic_rat(d, a, b),
(false,
  (arctic_rat_name,
    ((a: arctic_rat) => pos_arctic_rat(a),
      (_: arctic_rat) => false))))))))))))))),
           ((a: List[List[arctic_rat]]) =>
              mat_arc_posI[arctic_rat](get_comp_apos[arctic_rat](((_:
                             arctic_rat)
                            => true,
                           (zero_arctic_rat,
                             (one_arctic_rat,
                               (one_arctic_rat,
                                 ((aa: arctic_rat) =>
                                    (b: arctic_rat) => plus_arctic_rat(aa, b),
                                   ((aa: arctic_rat) =>
                                      (b: arctic_rat) =>
times_arctic_rat(aa, b),
                                     (id[arctic_rat],
                                       ((aa: arctic_rat) =>
  (b: arctic_rat) => ge_arctic_rat(aa, b),
 ((aa: arctic_rat) => (b: arctic_rat) => gt_arctic_rat(d, aa, b),
   (false,
     (arctic_rat_name,
       ((aa: arctic_rat) => pos_arctic_rat(aa),
         (_: arctic_rat) => false))))))))))))),
a),
             (_: List[List[arctic_rat]]) => false)))))))))))),
                             i)
  case Lpo(prec, pi) =>
    af_redpair[A, B](pi, create_LPO_redpair[A, filtered[A],
     B]((a: List[((A, nat), (nat, order_tag))]) =>
          (b: filtered[A]) => (c: nat) => prec_repr_to_pr[A](a, b, c),
         prec))
  case Mpo(prec, pi) =>
    af_redpair[A, B](pi, create_MPO_redpair[A, filtered[A],
     B]((a: List[((A, nat), (nat, order_tag))]) =>
          (b: filtered[A]) => (c: nat) => prec_repr_to_pr[A](a, b, c),
         prec))
  case Rpo(prec_tau, pi) =>
    af_redpair[A, B](pi, create_RPO_redpair[A, filtered[A],
     B]((pr: List[((A, nat), (nat, order_tag))]) =>
          ((a: filtered[A]) => (b: nat) => prec_repr_to_pr[A](pr, a, b),
            (a: filtered[A]) => (b: nat) => prec_repr_to_status[A](pr, a, b)),
         prec_tau))
}

def matchCapBelowI[A: equal,
                    B: equal](l: term[A, B], r: List[(term[A, B], term[A, B])],
                               x2: term[A, B]):
      Boolean =
  (l, r, x2) match {
  case (l, r, Fun(f, ts)) =>
    gcMatchI[A, B](HFun[A, B](f, mapa[term[A, B],
                                       hterm[A,
      B]]((a: term[A, B]) => capI[A, B](r, a), ts)),
                    l)
}

def check_ur_closed_term_af[A: equal: show,
                             B: equal: show](uu: List[(term[A, B], term[A, B])],
      uv: List[(term[A, B], term[A, B])], uw: A => nat => nat => Boolean,
      x3: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (uu, uv, uw, x3) match {
  case (uu, uv, uw, Var(x)) => Inr[Unit, (List[Char]) => List[Char]](())
  case (r, ur, pi, Fun(f, ts)) =>
    {
      val n: nat = size_list[term[A, B]](ts);
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[((term[A, B], nat), (List[Char]) => List[Char]), Unit,
                           (List[Char]) =>
                             List[Char]](forallm_index[term[A, B],
                (List[Char]) =>
                  List[Char]]((t: term[A, B]) =>
                                (i: nat) =>
                                  (if (((pi(f))(n))(i))
                                    check_ur_closed_term_af[A, B](r, ur, pi, t)
                                    else Inr[Unit,
      (List[Char]) => List[Char]](())),
                               ts),
  (x: ((term[A, B], nat), (List[Char]) => List[Char])) =>
    Inl[(List[Char]) => List[Char],
         Unit](snd[(term[A, B], nat), (List[Char]) => List[Char]](x))),
                    (_: Unit) =>
                      catcha[((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]),
                              Unit,
                              (List[Char]) =>
                                List[Char]](forallm[(term[A, B], term[A, B]),
             (List[Char]) =>
               List[Char]]((lr: (term[A, B], term[A, B])) =>
                             check[(List[Char]) =>
                                     List[Char]](membera[(term[A, B],
                   term[A, B])](ur, lr) ||
           ! (matchCapBelowI[A, B](fst[term[A, B], term[A, B]](lr), r,
                                    Fun[A, B](f, ts))),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List('d', 'u', 'e', ' ', 't',
             'o', ' ', 't', 'h', 'e', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm',
             ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_prec_term[A,
                        B](Zero_nat, Fun[A, B](f, ts)),
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(' ', 'o', 'f', ' ', 's', 'o',
           'm', 'e', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'h', 's', ',',
           ' ', 'r', 'u', 'l', 'e', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_rule[A, B](lr),
      shows_string.apply(List(' ', 's', 'h', 'o', 'u', 'l', 'd', ' ', 'b', 'e',
                               ' ', 'u', 's', 'a', 'b', 'l', 'e', '.'))))))),
                            r),
     (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
       Inl[(List[Char]) => List[Char],
            Unit](snd[(term[A, B], term[A, B]),
                       (List[Char]) => List[Char]](x))))
    }
}

def check_ur_P_closed_af[A: equal: show,
                          B: equal: show](r: List[(term[A, B], term[A, B])],
   ur: List[(term[A, B], term[A, B])], p: List[(term[A, B], term[A, B])],
   pi: A => nat => nat => Boolean):
      sum[(List[Char]) => List[Char], Unit] =
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](catcha[((term[A, B], term[A, B]),
      (List[Char]) => List[Char]),
     Unit,
     (List[Char]) =>
       List[Char]](forallm[(term[A, B], term[A, B]),
                            (List[Char]) =>
                              List[Char]]((lr: (term[A, B], term[A, B])) =>
    check_ur_closed_term_af[A, B](r, ur, pi, snd[term[A, B], term[A, B]](lr)),
   ur),
                    (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char]))
                      => Inl[(List[Char]) => List[Char],
                              Unit](snd[(term[A, B], term[A, B]),
 (List[Char]) => List[Char]](x))),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('e', 'r', 'r', 'o', 'r',
              ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n',
              'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p', 'r', 'o',
              'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o', 'f', ' ', 'r', 'h',
              's', ' ', 'o', 'f', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r',
              'u', 'l', 'e', 's')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](catcha[((term[A, B], term[A, B]),
         (List[Char]) => List[Char]),
        Unit,
        (List[Char]) =>
          List[Char]](forallm[(term[A, B], term[A, B]),
                               (List[Char]) =>
                                 List[Char]]((st: (term[A, B], term[A, B])) =>
       check_ur_closed_term_af[A, B](r, ur, pi,
                                      snd[term[A, B], term[A, B]](st)),
      p),
                       (x: ((term[A, B], term[A, B]),
                             (List[Char]) => List[Char]))
                         => Inl[(List[Char]) => List[Char],
                                 Unit](snd[(term[A, B], term[A, B]),
    (List[Char]) => List[Char]](x))),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('e', 'r', 'r', 'o', 'r',
                 ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i',
                 'n', 'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p',
                 'r', 'o', 'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o', 'f',
                 ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ', 'D', 'P', 's')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_nl, x)))))

def matchCapRMBelow[A: equal,
                     B: equal](l: term[A, B],
                                rm: Option[((A, nat)) =>
     List[(term[A, B], term[A, B])]],
                                x2: term[A, B]):
      Boolean =
  (l, rm, x2) match {
  case (l, rm, Fun(f, ts)) =>
    gcMatchI[A, B](HFun[A, B](f, mapa[term[A, B],
                                       hterm[A, B]](capRM[A, B](rm), ts)),
                    l)
}

def check_ur_closed_term_rm_af[A: equal: show,
                                B: equal: show](uu:
          ((A, nat)) => List[(term[A, B], term[A, B])],
         uv: List[(term[A, B], term[A, B])], uw: A => nat => nat => Boolean,
         x3: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (uu, uv, uw, x3) match {
  case (uu, uv, uw, Var(x)) => Inr[Unit, (List[Char]) => List[Char]](())
  case (rm, ur, pi, Fun(f, ts)) =>
    {
      val n: nat = size_list[term[A, B]](ts);
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[((term[A, B], nat), (List[Char]) => List[Char]), Unit,
                           (List[Char]) =>
                             List[Char]](forallm_index[term[A, B],
                (List[Char]) =>
                  List[Char]]((t: term[A, B]) =>
                                (i: nat) =>
                                  (if (((pi(f))(n))(i))
                                    check_ur_closed_term_rm_af[A,
                        B](rm, ur, pi, t)
                                    else Inr[Unit,
      (List[Char]) => List[Char]](())),
                               ts),
  (x: ((term[A, B], nat), (List[Char]) => List[Char])) =>
    Inl[(List[Char]) => List[Char],
         Unit](snd[(term[A, B], nat), (List[Char]) => List[Char]](x))),
                    (_: Unit) =>
                      catcha[((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]),
                              Unit,
                              (List[Char]) =>
                                List[Char]](forallm[(term[A, B], term[A, B]),
             (List[Char]) =>
               List[Char]]((lr: (term[A, B], term[A, B])) =>
                             check[(List[Char]) =>
                                     List[Char]](membera[(term[A, B],
                   term[A, B])](ur, lr) ||
           ! (matchCapRMBelow[A, B](fst[term[A, B], term[A, B]](lr),
                                     Some[((A, nat)) =>
    List[(term[A, B], term[A, B])]](rm),
                                     Fun[A, B](f, ts))),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List('d', 'u', 'e', ' ', 't',
             'o', ' ', 't', 'h', 'e', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm',
             ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_prec_term[A,
                        B](Zero_nat, Fun[A, B](f, ts)),
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(' ', 'o', 'f', ' ', 's', 'o',
           'm', 'e', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'h', 's', ',',
           ' ', 'r', 'u', 'l', 'e', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_rule[A, B](lr),
      shows_string.apply(List(' ', 's', 'h', 'o', 'u', 'l', 'd', ' ', 'b', 'e',
                               ' ', 'u', 's', 'a', 'b', 'l', 'e', '.'))))))),
                            rm((f, n))),
     (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
       Inl[(List[Char]) => List[Char],
            Unit](snd[(term[A, B], term[A, B]),
                       (List[Char]) => List[Char]](x))))
    }
}

def check_ur_P_closed_rm_af[A: equal: show,
                             B: equal: show](r: List[(term[A, B], term[A, B])],
      rmo: Option[((A, nat)) => List[(term[A, B], term[A, B])]],
      ur: List[(term[A, B], term[A, B])], p: List[(term[A, B], term[A, B])],
      pi: A => nat => nat => Boolean):
      sum[(List[Char]) => List[Char], Unit] =
  rmo match {
    case None => check_ur_P_closed_af[A, B](r, ur, p, pi)
    case Some(rm) =>
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[(List[Char]) => List[Char], Unit,
                           (List[Char]) =>
                             List[Char]](catcha[((term[A, B], term[A, B]),
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[(term[A, B], term[A, B]),
                                (List[Char]) =>
                                  List[Char]]((lr: (term[A, B], term[A, B])) =>
        check_ur_closed_term_rm_af[A, B](rm, ur, pi,
  snd[term[A, B], term[A, B]](lr)),
       ur),
                        (x: ((term[A, B], term[A, B]),
                              (List[Char]) => List[Char]))
                          => Inl[(List[Char]) => List[Char],
                                  Unit](snd[(term[A, B], term[A, B]),
     (List[Char]) => List[Char]](x))),
  (x: (List[Char]) => List[Char]) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('e', 'r', 'r', 'o',
                  'r', ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k',
                  'i', 'n', 'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ',
                  'p', 'r', 'o', 'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o',
                  'f', ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ', 'u', 's', 'a',
                  'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, x)))),
                    (_: Unit) =>
                      catcha[(List[Char]) => List[Char], Unit,
                              (List[Char]) =>
                                List[Char]](catcha[((term[A, B], term[A, B]),
             (List[Char]) => List[Char]),
            Unit,
            (List[Char]) =>
              List[Char]](forallm[(term[A, B], term[A, B]),
                                   (List[Char]) =>
                                     List[Char]]((st: (term[A, B], term[A, B]))
           => check_ur_closed_term_rm_af[A,
  B](rm, ur, pi, snd[term[A, B], term[A, B]](st)),
          p),
                           (x: ((term[A, B], term[A, B]),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[(term[A, B], term[A, B]),
        (List[Char]) => List[Char]](x))),
     (x: (List[Char]) => List[Char]) =>
       Inl[(List[Char]) => List[Char],
            Unit](comp[List[Char], List[Char],
                        List[Char]](shows_string.apply(List('e', 'r', 'r', 'o',
                     'r', ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k',
                     'i', 'n', 'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ',
                     'p', 'r', 'o', 'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o',
                     'f', ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ', 'D', 'P',
                     's')),
                                     comp[List[Char], List[Char],
   List[Char]](shows_nl, x)))))
  }

def check_redpair_ur_af_proc[A: equal: show,
                              B: equal: show](rp:
        (sum[(List[Char]) => List[Char], Unit],
          (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
            (((term[A, B], term[A, B])) =>
               sum[(List[Char]) => List[Char], Unit],
              (A => nat => nat => Boolean,
                (sum[(List[Char]) => List[Char], Unit],
                  (List[Char]) => List[Char]))))),
       p: List[(term[A, B], term[A, B])], rP: List[(term[A, B], term[A, B])],
       r: List[(term[A, B], term[A, B])],
       rm: Option[((A, nat)) => List[(term[A, B], term[A, B])]],
       ur: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (_, nSp):
                                (List[(term[A, B], term[A, B])],
                                  List[(term[A, B], term[A, B])])
                            = partition[(term[A, B],
  term[A, B])]((foo: (term[A, B], term[A, B])) =>
                 isOK[(List[Char]) => List[Char],
                       Unit]((rp_strict[A, B](rp)).apply(foo)),
                p);
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](rp_valid[A, B](rp),
(_: Unit) =>
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](check_ur_P_closed_rm_af[A,
                      B](r, rm, ur, p, rp_af[A, B](rp)),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'c', 'l', 'o', 's', 'u',
              'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e', 'r', 't', 'i', 'e', 's',
              ' ', 'o', 'f', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u',
              'l', 'e', 's', ' ', 'w', '.', 'r', '.', 't', '.', ' ', 'a', 'r',
              'g', 'u', 'm', 'e', 'n', 't', ' ', 'f', 'i', 'l', 't', 'e', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  binda[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[A, B], term[A, B]),
    (List[Char]) => List[Char]](rp_non_strict[A, B](rp), ur),
                                    (x: ((term[A, B], term[A, B]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[A, B], term[A, B]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u',
                              'l', 'e', 's')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  binda[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[A, B], term[A, B]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[A, B],
                     term[A, B]),
                    (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSp),
            (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[A, B], term[A, B]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'D', 'P', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          catcha[(term[A, B], term[A, B]), Unit,
                  (List[Char]) =>
                    List[Char]](check_subseteq[(term[A, B],
         term[A, B])](nSp, rP),
                                 (x: (term[A, B], term[A, B])) =>
                                   Inl[(List[Char]) => List[Char],
Unit]((rp_strict[A, B](rp)).apply(x) match {
        case Inl(err) => err
        case Inr(_) => internal_error
      }))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](rp_desc[A, B](rp),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'f',
                                     'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                                     ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                                     'r', 'u', 'l', 'e', 's')),
             comp[List[Char], List[Char],
                   List[Char]](shows_nl,
                                comp[List[Char], List[Char],
                                      List[Char]](shows_rules[A, B](ur),
           comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))))

def check_dep_graph_proc[A: equal: show,
                          B: equal: show](p: List[(term[A, B], term[A, B])],
   dps: List[(Boolean, List[(term[A, B], term[A, B])])],
   rm: Option[((A, nat)) => List[(term[A, B], term[A, B])]],
   rrm: Option[((A, nat)) => List[(term[A, B], term[A, B])]]):
      sum[(List[Char]) => List[Char], Unit] =
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(term[A, B], term[A, B]), Unit,
                       (List[Char]) =>
                         List[Char]](check_subseteq[(term[A, B],
              term[A, B])](p, maps[(Boolean, List[(term[A, B], term[A, B])]),
                                    (term[A, B],
                                      term[A,
    B])]((a: (Boolean, List[(term[A, B], term[A, B])])) =>
           snd[Boolean, List[(term[A, B], term[A, B])]](a),
          dps)),
                                      (x: (term[A, B], term[A, B])) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('D', 'e', 'p', 'e', 'n',
              'd', 'e', 'n', 'c', 'y', ' ', 'P', 'a', 'i', 'r', ' ')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_rule[A, B](x),
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'm', 'i',
            's', 's', 'i', 'n', 'g', ' ', 'i', 'n', ' ', 'd', 'e', 'c', 'o',
            'm', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n')),
                            shows_nl))))),
                (_: Unit) =>
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](check_graph_decomp[((term[A, B],
                      term[A, B]),
                     (hterm[A, B],
                       hterm[A, B]))](comp[(term[A, B], term[A, B]),
    (List[Char]) => List[Char],
    ((term[A, B], term[A, B]),
      (hterm[A, B],
        hterm[A, B]))]((a: (term[A, B], term[A, B])) =>
                         shows_prec_prod[term[A, B], term[A, B]](Zero_nat, a),
                        (a: ((term[A, B], term[A, B]),
                              (hterm[A, B], hterm[A, B])))
                          => fst[(term[A, B], term[A, B]),
                                  (hterm[A, B], hterm[A, B])](a)),
                                       (a:
  (((term[A, B], term[A, B]), (hterm[A, B], hterm[A, B])),
    ((term[A, B], term[A, B]), (hterm[A, B], hterm[A, B]))))
 => {
      val (aa, b):
            (((term[A, B], term[A, B]), (hterm[A, B], hterm[A, B])),
              ((term[A, B], term[A, B]), (hterm[A, B], hterm[A, B])))
        = a;
      ({
         val (ab, ba): ((term[A, B], term[A, B]), (hterm[A, B], hterm[A, B])) =
           aa;
         ({
            val (_, t): (term[A, B], term[A, B]) = ab;
            (ac: (hterm[A, B], hterm[A, B])) =>
              {
                val (_, ct): (hterm[A, B], hterm[A, B]) = ac;
                (ad: ((term[A, B], term[A, B]), (hterm[A, B], hterm[A, B]))) =>
                  {
                    val (ae, bb):
                          ((term[A, B], term[A, B]), (hterm[A, B], hterm[A, B]))
                      = ad;
                    ({
                       val (u, _): (term[A, B], term[A, B]) = ae;
                       (af: (hterm[A, B], hterm[A, B])) =>
                         {
                           val (cu, _): (hterm[A, B], hterm[A, B]) = af;
                           gcMatchI[A, B](ct, u) && gcMatchI[A, B](cu, t)
                         }
                     })(bb)
                  }
              }
          })(ba)
       })(b)
    },
                                       mapa[(Boolean,
      List[(term[A, B], term[A, B])]),
     (Boolean,
       List[((term[A, B], term[A, B]),
              (hterm[A, B],
                hterm[A, B]))])]((a: (Boolean, List[(term[A, B], term[A, B])]))
                                   => {
val (real, cs): (Boolean, List[(term[A, B], term[A, B])]) = a;
(real,
  mapa[(term[A, B], term[A, B]),
        ((term[A, B], term[A, B]),
          (hterm[A, B],
            hterm[A, B]))]((aa: (term[A, B], term[A, B])) =>
                             {
                               val (s, t): (term[A, B], term[A, B]) = aa;
                               ((s, t),
                                 ((capRM[A, B](rrm)).apply(s),
                                   (capRM[A, B](rm)).apply(t)))
                             },
                            cs))
                                      },
                                  dps)),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('o', 'u', 'r', ' ', 'e',
                 's', 't', 'i', 'm', 'a', 't', 'i', 'o', 'n', ' ', '(', 'E',
                 'D', 'G', '*', '*', '*', ')', ' ', 'c', 'o', 'u', 'l', 'd',
                 ' ', 'n', 'o', 't', ' ', 's', 'h', 'o', 'w', ' ', 't', 'h',
                 'a', 't', ' ', 'y', 'o', 'u', ' ', 'h', 'a', 'v', 'e', ' ',
                 'a', ' ', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'e', 'c', 'o',
                 'm', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_string.apply(List('d',
                                    'u', 'e', ' ', 't', 'o', ' ', 't', 'h', 'e',
                                    ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n',
                                    'g', ' ', 'r', 'e', 'a', 's', 'o', 'n')),
            comp[List[Char], List[Char], List[Char]](shows_nl, x))))))

def check_allm_gen_index[A, B](g: A => nat => nat, f: A => nat => sum[B, Unit],
                                n: nat, xs: List[A]):
      sum[B, Unit] =
  snd[nat, sum[B, Unit]](foldl[(nat, sum[B, Unit]),
                                A]((a: (nat, sum[B, Unit])) =>
                                     {
                                       val (i, m): (nat, sum[B, Unit]) = a;
                                       (x: A) =>
 ((g(x))(i), binda[B, Unit, Unit](m, (_: Unit) => (f(x))(i)))
                                     },
                                    (n, Inr[Unit, B](())), xs))

def check_mono_redpair_proc[A: equal: show,
                             B: equal: show](rp:
       (sum[(List[Char]) => List[Char], Unit],
         (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
           (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
             (A => nat => nat => Boolean,
               (sum[(List[Char]) => List[Char], Unit],
                 (List[Char]) => List[Char]))))),
      p: List[(term[A, B], term[A, B])], rP: List[(term[A, B], term[A, B])],
      r: List[(term[A, B], term[A, B])], rR: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (_, nSp):
                                (List[(term[A, B], term[A, B])],
                                  List[(term[A, B], term[A, B])])
                            = partition[(term[A, B],
  term[A, B])]((foo: (term[A, B], term[A, B])) =>
                 isOK[(List[Char]) => List[Char],
                       Unit]((rp_strict[A, B](rp)).apply(foo)),
                p)
                          val (_, nSr):
                                (List[(term[A, B], term[A, B])],
                                  List[(term[A, B], term[A, B])])
                            = partition[(term[A, B],
  term[A, B])]((foo: (term[A, B], term[A, B])) =>
                 isOK[(List[Char]) => List[Char],
                       Unit]((rp_strict[A, B](rp)).apply(foo)),
                r);
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](rp_valid[A, B](rp),
(_: Unit) =>
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](rp_mono[A, B](rp),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
              'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 's', 't',
              'r', 'i', 'c', 't', ' ', 'o', 'r', 'd', 'e', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  binda[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[A, B], term[A, B]),
    (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSr),
                                    (x: ((term[A, B], term[A, B]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[A, B], term[A, B]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'T', 'R', 'S')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  binda[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[A, B], term[A, B]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[A, B],
                     term[A, B]),
                    (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSp),
            (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[A, B], term[A, B]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'D', 'P', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          binda[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(term[A, B], term[A, B]), Unit,
                               (List[Char]) =>
                                 List[Char]](check_subseteq[(term[A, B],
                      term[A, B])](rR, r),
      (_: (term[A, B], term[A, B])) =>
        Inl[(List[Char]) => List[Char],
             Unit](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
    ':', ' ', 'a', 'd', 'd', 'i', 't', 'i', 'o', 'n', 'a', 'l', ' ')))),
                        (_: Unit) =>
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](catcha[(term[A, B], term[A, B]), Unit,
       (List[Char]) =>
         List[Char]](check_subseteq[(term[A, B], term[A, B])](nSr, rR),
                      (x: (term[A, B], term[A, B])) =>
                        Inl[(List[Char]) => List[Char],
                             Unit]((rp_strict[A, B](rp)).apply(x) match {
                                     case Inl(err) => err
                                     case Inr(_) => internal_error
                                   })),
(_: Unit) =>
  catcha[(term[A, B], term[A, B]), Unit,
          (List[Char]) =>
            List[Char]](check_subseteq[(term[A, B], term[A, B])](nSp, rP),
                         (x: (term[A, B], term[A, B])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit]((rp_strict[A, B](rp)).apply(x) match {
case Inl(err) => err
case Inr(_) => internal_error
                                      }))))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](rp_desc[A, B](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))

def check_mono_redpair_ur_proc[A: equal: show,
                                B: equal: show](rp:
          (sum[(List[Char]) => List[Char], Unit],
            (((term[A, B], term[A, B])) =>
               sum[(List[Char]) => List[Char], Unit],
              (((term[A, B], term[A, B])) =>
                 sum[(List[Char]) => List[Char], Unit],
                (A => nat => nat => Boolean,
                  (sum[(List[Char]) => List[Char], Unit],
                    (List[Char]) => List[Char]))))),
         p: List[(term[A, B], term[A, B])], rP: List[(term[A, B], term[A, B])],
         r: List[(term[A, B], term[A, B])],
         rm: Option[((A, nat)) => List[(term[A, B], term[A, B])]],
         rR: List[(term[A, B], term[A, B])],
         ur: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val us: List[(A, nat)] =
                            maps[(term[A, B], term[A, B]),
                                  (A, nat)](comp[term[A, B], List[(A, nat)],
          (term[A, B],
            term[A, B])](funas_term_list[A, B],
                          (a: (term[A, B], term[A, B])) =>
                            snd[term[A, B], term[A, B]](a)),
     p ++ ur)
                          val filt: (List[(term[A, B], term[A, B])]) =>
                                      List[(term[A, B], term[A, B])]
                            = (a: List[(term[A, B], term[A, B])]) =>
                                filter[(term[A, B],
 term[A, B])]((lr: (term[A, B], term[A, B])) =>
                wf_terms_list[A, B](us, fst[term[A, B], term[A, B]](lr)),
               a)
                          val (_, nSp):
                                (List[(term[A, B], term[A, B])],
                                  List[(term[A, B], term[A, B])])
                            = partition[(term[A, B],
  term[A, B])]((foo: (term[A, B], term[A, B])) =>
                 isOK[(List[Char]) => List[Char],
                       Unit]((rp_strict[A, B](rp)).apply(foo)),
                p)
                          val (_, nSr):
                                (List[(term[A, B], term[A, B])],
                                  List[(term[A, B], term[A, B])])
                            = partition[(term[A, B],
  term[A, B])]((foo: (term[A, B], term[A, B])) =>
                 isOK[(List[Char]) => List[Char],
                       Unit]((rp_strict[A, B](rp)).apply(foo)),
                ur);
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](rp_valid[A, B](rp),
(_: Unit) =>
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](rp_mono[A, B](rp),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
              'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 's', 't',
              'r', 'i', 'c', 't', ' ', 'o', 'r', 'd', 'e', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  binda[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](check_ur_P_closed_rm_af[A, B](r, rm, ur, p,
    (a: A) => (b: nat) => (c: nat) => pi_full[A](a, b, c)),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h',
                              ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p',
                              'r', 'o', 'p', 'e', 'r', 't', 'i', 'e', 's', ' ',
                              'o', 'f', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                              'r', 'u', 'l', 'e', 's')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  binda[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[A, B], term[A, B]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[A, B],
                     term[A, B]),
                    (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSr),
            (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[A, B], term[A, B]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          binda[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](catcha[((term[A, B], term[A, B]),
              (List[Char]) => List[Char]),
             Unit,
             (List[Char]) =>
               List[Char]](forallm[(term[A, B], term[A, B]),
                                    (List[Char]) =>
                                      List[Char]](rp_non_strict[A, B](rp), nSp),
                            (x: ((term[A, B], term[A, B]),
                                  (List[Char]) => List[Char]))
                              => Inl[(List[Char]) => List[Char],
                                      Unit](snd[(term[A, B], term[A, B]),
         (List[Char]) => List[Char]](x))),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
                      'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))),
                        (_: Unit) =>
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](catcha[(term[A, B], term[A, B]), Unit,
       (List[Char]) =>
         List[Char]](check_subseteq[(term[A, B], term[A, B])](rR, r),
                      (_: (term[A, B], term[A, B])) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ':',
                                      ' ', 'a', 'd', 'd', 'i', 't', 'i', 'o',
                                      'n', 'a', 'l', ' ', 'r', 'u', 'l', 'e',
                                      '(', 's', ')', ' ', 't', 'o', ' ', 'T',
                                      'R', 'S', ' ', 'a', 'd', 'd', 'e', 'd')),
              shows_nl))),
(_: Unit) =>
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(term[A, B], term[A, B]), Unit,
                       (List[Char]) =>
                         List[Char]](check_subseteq[(term[A, B],
              term[A, B])](filt(nSr), rR),
                                      (x: (term[A, B], term[A, B])) =>
Inl[(List[Char]) => List[Char],
     Unit]((rp_strict[A, B](rp)).apply(x) match {
             case Inl(err) => err
             case Inr(_) => internal_error
           })),
                (_: Unit) =>
                  catcha[(term[A, B], term[A, B]), Unit,
                          (List[Char]) =>
                            List[Char]](check_subseteq[(term[A, B],
                 term[A, B])](filt(nSp), rP),
 (x: (term[A, B], term[A, B])) =>
   Inl[(List[Char]) => List[Char],
        Unit]((rp_strict[A, B](rp)).apply(x) match {
                case Inl(err) => err
                case Inr(_) => internal_error
              })))))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', '(', 'm', 'o', 'n',
 'o', 't', 'o', 'n', 'i', 'c', ')', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r',
 ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w',
 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](rp_desc[A, B](rp),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'f',
                                     'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                                     ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                                     'r', 'u', 'l', 'e', 's')),
             comp[List[Char], List[Char],
                   List[Char]](shows_rules[A, B](ur),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_nl, x)))))))))

def unlab_of_trs_list[A: equal, B: equal,
                       C: equal](r: List[(term[lab[A, B], C],
   term[lab[A, B], C])]):
      List[(term[lab[A, B], C], term[lab[A, B], C])] =
  remdups[(term[lab[A, B], C],
            term[lab[A, B],
                  C])](map_trs_list[lab[A, B], lab[A, B],
                                     C]((a: lab[A, B]) => unlab[A, B](a),
 filter[(term[lab[A, B], C],
          term[lab[A, B],
                C])]((a: (term[lab[A, B], C], term[lab[A, B], C])) =>
                       {
                         val (l, ra): (term[lab[A, B], C], term[lab[A, B], C]) =
                           a;
                         equal_terma[lab[A, B], C](l, ra) ||
                           ! (equal_terma[lab[A, B],
   C](map_term[lab[A, B], lab[A, B], C]((aa: lab[A, B]) => unlab[A, B](aa), l),
       map_term[lab[A, B], lab[A, B],
                 C]((aa: lab[A, B]) => unlab[A, B](aa), ra)))
                       },
                      r)))

def check_unlab_proc[A: equal: show, B: equal: show,
                      C: equal: show](p:
List[(term[lab[A, B], C], term[lab[A, B], C])],
                                       r:
 List[(term[lab[A, B], C], term[lab[A, B], C])],
                                       ulP:
 List[(term[lab[A, B], C], term[lab[A, B], C])],
                                       ulR:
 List[(term[lab[A, B], C], term[lab[A, B], C])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val ulr: List[(term[lab[A, B], C], term[lab[A, B], C])] =
      unlab_of_trs_list[A, B, C](r);
    binda[(List[Char]) => List[Char], Unit,
           Unit](catcha[(term[lab[A, B], C], term[lab[A, B], C]), Unit,
                         (List[Char]) =>
                           List[Char]](check_subseteq[(term[lab[A, B], C],
                term[lab[A, B],
                      C])](map_trs_list[lab[A, B], lab[A, B],
 C]((a: lab[A, B]) => unlab[A, B](a), p),
                            ulP),
(x: (term[lab[A, B], C], term[lab[A, B], C])) =>
  Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_string.apply(List('u', 'n', 'l', 'a', 'b',
                'e', 'l', 'e', 'd', ' ', 'p', 'a', 'i', 'r', ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_rule[lab[A, B], C](x),
           comp[List[Char], List[Char],
                 List[Char]](shows_prec_list[Char](Zero_nat,
            List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')),
                              shows_nl))))),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](catcha[(term[lab[A, B], C],
  term[lab[A, B], C]),
 Unit,
 (List[Char]) =>
   List[Char]](check_subseteq[(term[lab[A, B], C],
                                term[lab[A, B], C])](ulr, ulR),
                (x: (term[lab[A, B], C], term[lab[A, B], C])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List('u', 'n',
                                'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r',
                                'u', 'l', 'e', ' ')),
        comp[List[Char], List[Char],
              List[Char]](shows_rule[lab[A, B], C](x),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_prec_list[Char](Zero_nat,
                            List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                  'n', 'g')),
      shows_nl))))),
                                  (_: Unit) =>
                                    catcha[(term[lab[A, B], C],
     term[lab[A, B], C]),
    Unit,
    (List[Char]) =>
      List[Char]](check_subseteq[(term[lab[A, B], C],
                                   term[lab[A, B], C])](ulR, ulr),
                   (x: (term[lab[A, B], C], term[lab[A, B], C])) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('r',
                                   'u', 'l', 'e', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_rule[lab[A, B], C](x),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_prec_list[Char](Zero_nat,
                               List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a',
                                     'n', ' ', 'u', 'n', 'l', 'a', 'b', 'e',
                                     'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e')),
         shows_nl)))))))
  }

def get_largest_element[A, B](x0: sl_inter[A, B]): nat = x0 match {
  case SL_Inter(n, uu) => n
}

def take_default[A](defa: A, x1: List[A], uu: nat): A = (defa, x1, uu) match {
  case (defa, Nil, uu) => defa
  case (uv, x :: xs, Zero_nat) => x
  case (defa, uw :: xs, Suc(i)) => take_default[A](defa, xs, i)
}

def eval_arithFun_unbound(nats: List[nat], x1: arithFun): nat = (nats, x1) match
  {
  case (nats, Arg(i)) => take_default[nat](Zero_nat, nats, i)
  case (nats, Const(n)) => n
  case (nats, Sum(Nil)) => Zero_nat
  case (nats, Sum(f :: fs)) =>
    plus_nat(eval_arithFun_unbound(nats, f),
              eval_arithFun_unbound(nats, Sum(fs)))
  case (nats, Prod(Nil)) => one_nat
  case (nats, Prod(f :: fs)) =>
    times_nat(eval_arithFun_unbound(nats, f),
               eval_arithFun_unbound(nats, Prod(fs)))
  case (nats, Max(List(f))) => eval_arithFun_unbound(nats, f)
  case (nats, Max(f :: v :: va)) =>
    max[nat].apply(eval_arithFun_unbound(nats,
  f)).apply(eval_arithFun_unbound(nats, Max(v :: va)))
  case (nats, Min(List(f))) => eval_arithFun_unbound(nats, f)
  case (nats, Min(f :: v :: va)) =>
    min[nat].apply(eval_arithFun_unbound(nats,
  f)).apply(eval_arithFun_unbound(nats, Min(v :: va)))
}

def eval_arithFun(c: nat, f: arithFun, nats: List[nat]): nat =
  mod_nat(eval_arithFun_unbound(nats, f), c)

def sl_inter_to_inter[A: equal,
                       B: equal](x0: sl_inter[A, B], fl: lab[A, B],
                                  cs: List[nat]):
      nat =
  (x0, fl, cs) match {
  case (SL_Inter(c, ls), fl, cs) =>
    lookupa[(lab[A, B], nat), arithFun]((fl, size_list[nat](cs)), ls) match {
      case None => Zero_nat
      case Some(f) => eval_arithFun(Suc(c), f, cs)
    }
}

def shows_arithFun(x0: arithFun): (List[Char]) => List[Char] = x0 match {
  case Arg(i) =>
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List('x', '_')),
                       shows_prec_nat(Zero_nat, Suc(i)))
  case Const(n) => shows_prec_nat(Zero_nat, n)
  case Sum(fs) =>
    shows_list_gen[arithFun]((a: arithFun) => shows_arithFun(a), Nil, Nil,
                              List(' ', '+', ' '), Nil, fs)
  case Prod(fs) =>
    shows_list_gen[arithFun]((a: arithFun) => shows_arithFun(a), Nil, Nil,
                              List(' ', '*', ' '), Nil, fs)
  case Min(fs) =>
    shows_list_gen[arithFun]((a: arithFun) => shows_arithFun(a), Nil,
                              List('m', 'i', 'n', '('), List(','), List(')'),
                              fs)
  case Max(fs) =>
    shows_list_gen[arithFun]((a: arithFun) => shows_arithFun(a), Nil,
                              List('m', 'a', 'x', '('), List(','), List(')'),
                              fs)
}

def shows_prec_arithFun(d: nat, f: arithFun): (List[Char]) => List[Char] =
  shows_arithFun(f)

def shows_sl_inter[A: show, B: show](x0: sl_inter[A, B]):
      (List[Char]) => List[Char] =
  x0 match {
  case SL_Inter(n, fnas) =>
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List('c', 'a', 'r', 'r', 'i', 'e', 'r',
       ' ', '{', '0', ',', '.', '.', '.', ',')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_nat(Zero_nat, n),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('}')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
shows_map[((lab[A, B], nat),
            arithFun)]((a: ((lab[A, B], nat), arithFun)) =>
                         {
                           val (aa, b): ((lab[A, B], nat), arithFun) = a;
                           ({
                              val (f, na): (lab[A, B], nat) = aa;
                              (ab: arithFun) =>
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('[')),
           comp[List[Char], List[Char],
                 List[Char]](shows_prec_lab[A, B](Zero_nat, f),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('/')),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_nat(Zero_nat, na),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_string.apply(List(']', ' ',
                               '=', ' ')),
       comp[List[Char], List[Char],
             List[Char]](shows_prec_arithFun(Zero_nat, ab), shows_nl))))))
                            })(b)
                         },
                        fnas)))))
}

def shows_prec_sl_inter[A: show, B: show](d: nat, sli: sl_inter[A, B]):
      (List[Char]) => List[Char] =
  shows_sl_inter[A, B](sli)

def qmodel_L[A: equal, B: equal, C, D](sig: List[(lab[A, B], nat)]):
      (lab[A, B]) => (List[C]) => sum[List[C], D] =
  (f: lab[A, B]) =>
    (cs: List[C]) =>
      (if (membera[(lab[A, B], nat)](sig, (f, size_list[C](cs))))
        Inl[List[C], D](cs) else Inl[List[C], D](Nil))

def enum_vectors_nat[A](c: List[A], x1: nat): List[List[A]] = (c, x1) match {
  case (c, Zero_nat) => List(Nil)
  case (c, Suc(n)) =>
    {
      val a: List[List[A]] = enum_vectors_nat[A](c, n);
      maps[List[A],
            List[A]]((vec: List[A]) =>
                       mapa[A, List[A]]((ca: A) => ca :: vec, c),
                      a)
    }
}

def qmodel_check_interpretation(f: arithFun, n: nat, c: nat):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val ca: List[nat] = upt(Zero_nat, Suc(c))
    val css: List[List[nat]] = enum_vectors_nat[nat](ca, n);
    catcha[(List[nat], (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[List[nat],
                                   (List[Char]) =>
                                     List[Char]]((cs: List[nat]) =>
           catcha[(nat, (List[Char]) => List[Char]), Unit,
                   (List[Char]) =>
                     List[Char]](forallm[nat,
  (List[Char]) =>
    List[Char]]((i: nat) =>
                  catcha[(nat, (List[Char]) => List[Char]), Unit,
                          (List[Char]) =>
                            List[Char]](forallm[nat,
         (List[Char]) =>
           List[Char]]((l: nat) =>
                         check[(List[Char]) =>
                                 List[Char]](less_eq_nat(eval_arithFun(Suc(c),
                                f, cs),
                  eval_arithFun(Suc(c), f, list_update[nat](cs, i, l))),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_list[Char](Zero_nat,
       List('n', 'o', 't', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'e', ' ',
             'i', 'n', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_prec_nat(Zero_nat, Suc(i)),
    shows_prec_list[Char](Zero_nat,
                           List('.', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
                                 't'))))),
                        upt(nth[nat](cs, i), Suc(c))),
 (x: (nat, (List[Char]) => List[Char])) =>
   Inl[(List[Char]) => List[Char],
        Unit](snd[nat, (List[Char]) => List[Char]](x))),
                 upt(Zero_nat, n)),
                                  (x: (nat, (List[Char]) => List[Char])) =>
                                    Inl[(List[Char]) => List[Char],
 Unit](snd[nat, (List[Char]) => List[Char]](x))),
          css),
                           (x: (List[nat], (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[List[nat],
     (List[Char]) => List[Char]](x)))
  }

def qmodel_check_valid[A: show, B: show](x0: sl_inter[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case SL_Inter(c, ls) =>
    catcha[(((lab[A, B], nat), arithFun), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[((lab[A, B], nat), arithFun),
                                   (List[Char]) =>
                                     List[Char]]((a:
            ((lab[A, B], nat), arithFun))
           => {
                val (aa, b): ((lab[A, B], nat), arithFun) = a;
                ({
                   val (f, n): (lab[A, B], nat) = aa;
                   (g: arithFun) =>
                     catcha[(List[Char]) => List[Char], Unit,
                             (List[Char]) =>
                               List[Char]](qmodel_check_interpretation(g, n, c),
    (x: (List[Char]) => List[Char]) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n', ' ',
                        'w', 'e', 'a', 'k', '-', 'm', 'o', 'n', 'o', 't', 'o',
                        'n', 'i', 'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 'i',
                        'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i',
                        'o', 'n', ' ', 'o', 'f', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_lab[A, B](Zero_nat, f),
               comp[List[Char], List[Char], List[Char]](shows_nl, x)))))
                 })(b)
              },
          ls),
                           (x: (((lab[A, B], nat), arithFun),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[((lab[A, B], nat), arithFun),
        (List[Char]) => List[Char]](x)))
}

def check_decr_present_aux_1[A: equal,
                              B: equal](r:
  List[(term[A, (B, nat)], term[A, (B, nat)])],
 v: B, f1: A, f2: A, n: nat):
      sum[(term[A, (B, nat)], term[A, (B, nat)]), Unit] =
  {
    val vs: List[term[A, (B, nat)]] =
      mapa[nat, term[A, (B, nat)]]((na: nat) => Var[(B, nat), A]((v, na)),
                                    upt(Zero_nat, n))
    val rule: (term[A, (B, nat)], term[A, (B, nat)]) =
      (Fun[A, (B, nat)](f1, vs), Fun[A, (B, nat)](f2, vs));
    check[(term[A, (B, nat)],
            term[A, (B, nat)])](! (is_none[(term[A, (B, nat)],
     term[A, (B, nat)])](find[(term[A, (B, nat)],
                                term[A, (B,
  nat)])]((a: (term[A, (B, nat)], term[A, (B, nat)])) =>
            instance_rule[A, (B, nat)](rule, a),
           r))),
                                 rule)
  }

def check_decr_present_aux_2[A: equal,
                              B: equal](r: List[(term[A, B], term[A, B])], v: B,
 req: List[(A, (A, nat))]):
      sum[(term[A, (B, nat)], term[A, (B, nat)]), Unit] =
  {
    val add_nats: (term[A, B]) => term[A, (B, nat)] =
      (a: term[A, B]) => map_vars[B, (B, nat), A]((va: B) => (va, Zero_nat), a)
    val ra: List[(term[A, (B, nat)], term[A, (B, nat)])] =
      mapa[(term[A, B], term[A, B]),
            (term[A, (B, nat)],
              term[A, (B, nat)])]((a: (term[A, B], term[A, B])) =>
                                    {
                                      val (l, ra): (term[A, B], term[A, B]) = a;
                                      (add_nats(l), add_nats(ra))
                                    },
                                   r);
    catcha[((A, (A, nat)), (term[A, (B, nat)], term[A, (B, nat)])), Unit,
            (term[A, (B, nat)],
              term[A, (B, nat)])](forallm[(A, (A, nat)),
   (term[A, (B, nat)],
     term[A, (B, nat)])]((a: (A, (A, nat))) =>
                           {
                             val (f1, (aa, b)): (A, (A, nat)) = a;
                             check_decr_present_aux_1[A, B](ra, v, f1, aa, b)
                           },
                          req),
                                   (x: ((A, (A, nat)),
 (term[A, (B, nat)], term[A, (B, nat)])))
                                     => Inl[(term[A, (B, nat)],
      term[A, (B, nat)]),
     Unit](snd[(A, (A, nat)), (term[A, (B, nat)], term[A, (B, nat)])](x)))
  }

def check_decr_present[A: equal,
                        B: equal](sig: List[(A, nat)], l: A => (List[nat]) => A,
                                   v: B, c: nat,
                                   r: List[(term[A, B], term[A, B])]):
      sum[(term[A, (B, nat)], term[A, (B, nat)]), Unit] =
  {
    val ca: List[nat] = upt(Zero_nat, Suc(c))
    val ls: ((A, nat)) => List[(A, (A, nat))] =
      (a: (A, nat)) =>
        {
          val (f, n): (A, nat) = a;
          maps[List[nat],
                (A, (A, nat))]((cs: List[nat]) =>
                                 maps[nat, (A,
     (A, nat))]((i: nat) =>
                  {
                    val ci: nat = nth[nat](cs, i);
                    (if (less_nat(ci, c))
                      List(((l(f))(list_update[nat](cs, i, Suc(ci))),
                             ((l(f))(cs), n)))
                      else Nil)
                  },
                 upt(Zero_nat, n)),
                                enum_vectors_nat[nat](ca, n))
        };
    check_decr_present_aux_2[A, B](r, v, maps[(A, nat), (A, (A, nat))](ls, sig))
  }

def qmodel_check_decr[A: equal: show,
                       B: equal](sig: List[(lab[A, List[nat]], nat)], v: B,
                                  c: nat):
      (List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
        sum[(List[Char]) => List[Char], Unit] =
  (lR: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
    catcha[(term[lab[A, List[nat]], (B, nat)],
             term[lab[A, List[nat]], (B, nat)]),
            Unit,
            (List[Char]) =>
              List[Char]](check_decr_present[lab[A, List[nat]],
      B](sig, (a: lab[A, List[nat]]) =>
                (b: List[nat]) => Lab[A, List[nat]](a, b),
          v, c, lR),
                           (x: (term[lab[A, List[nat]], (B, nat)],
                                 term[lab[A, List[nat]], (B, nat)]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit]({
     val display:
           (term[lab[A, List[nat]], (B, nat)]) =>
             term[lab[A, List[nat]], List[Char]]
       = (a: term[lab[A, List[nat]], (B, nat)]) =>
           map_vars[(B, nat), List[Char],
                     lab[A, List[nat]]]((aa: (B, nat)) =>
  {
    val (_, n): (B, nat) = aa;
    (comp[List[Char], List[Char],
           List[Char]](shows_prec_list[Char](Zero_nat, List('x')),
                        shows_prec_nat(Zero_nat, n))).apply(Nil)
  },
 a);
     comp[List[Char], List[Char],
           List[Char]](shows_prec_list[Char](Zero_nat,
      List('d', 'e', 'c', 'r', 'e', 'a', 's', 'i', 'n', 'g', ' ', 'r', 'u', 'l',
            'e', ' ')),
                        comp[List[Char], List[Char],
                              List[Char]](shows_rule[lab[A, List[nat]],
              List[Char]]((display(fst[term[lab[A, List[nat]], (B, nat)],
term[lab[A, List[nat]], (B, nat)]](x)),
                            display(snd[term[lab[A, List[nat]], (B, nat)],
 term[lab[A, List[nat]], (B, nat)]](x)))),
   shows_prec_list[Char](Zero_nat,
                          List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g'))))
   }))

def qmodel_cge: nat => nat => Boolean =
  (x: nat) => (y: nat) => less_eq_nat(y, x)

def pointwise_ext[A](f: A => A => (Boolean, Boolean), x1: List[A], x2: List[A]):
      (Boolean, Boolean) =
  (f, x1, x2) match {
  case (f, Nil, Nil) => (false, true)
  case (f, uu :: uv, Nil) => (false, false)
  case (f, Nil, uw :: ux) => (false, false)
  case (f, a :: as, b :: bs) =>
    (f(a))(b) match {
      case (stri, true) =>
        pointwise_ext[A](f, as, bs) match {
          case (strir, true) => (stri || strir, true)
          case (strir, false) => (false, false)
        }
      case (stri, false) => (false, false)
    }
}

def qmodel_lge[A, B, C, D](f: A, n: B):
      (sum[List[nat], C]) => (sum[List[nat], D]) => Boolean =
  (l: sum[List[nat], C]) =>
    (r: sum[List[nat], D]) =>
      (l, r) match {
        case (Inl(cs1), Inl(cs2)) =>
          snd[Boolean,
               Boolean](pointwise_ext[nat]((x: nat) =>
     (y: nat) => (less_nat(y, x), less_eq_nat(y, x)),
    cs1, cs2))
        case (Inl(cs1), Inr(_)) => false
        case (Inr(_), b) => false
      }

def qmodel_LS[A: equal, B: equal, C: equal](sig: List[(A, nat)], ls: List[B]):
      A => nat => (sum[List[B], C]) => Boolean =
  (f: A) =>
    (n: nat) =>
      (if (membera[(A, nat)](sig, (f, n)))
        set[sum[List[B],
                 C]](mapa[List[B],
                           sum[List[B],
                                C]]((a: List[B]) => Inl[List[B], C](a),
                                     enum_vectors_nat[B](ls, n)))
        else (a: sum[List[B], C]) =>
               inserta[sum[List[B],
                            C]](Inl[List[B], C](Nil),
                                 bot_fun[sum[List[B], C], Boolean], a))

def funas_root_term[A, B](x0: term[A, B]): List[(A, nat)] = x0 match {
  case Fun(f, ts) => List((f, size_list[term[A, B]](ts)))
  case Var(uu) => Nil
}

def funas_root_rule[A: equal, B](x0: (term[A, B], term[A, B])): List[(A, nat)] =
  x0 match {
  case (l, r) =>
    remdups[(A, nat)](funas_root_term[A, B](l) ++ funas_root_term[A, B](r))
}

def funas_root_rules[A: equal, B](r: List[(term[A, B], term[A, B])]):
      List[(A, nat)] =
  remdups[(A, nat)](maps[(term[A, B], term[A, B]),
                          (A, nat)]((a: (term[A, B], term[A, B])) =>
                                      funas_root_rule[A, B](a),
                                     r))

def check_sem_lab_root_proc[A, B: equal: show, C: equal: show, D: equal: show,
                             E, F, G: show, H: show, I: show, J: show, K,
                             L](valid: sum[(List[Char]) => List[Char], A],
                                 check_decr:
                                   (List[(term[lab[B, C], D],
   term[lab[B, C], D])]) =>
                                     sum[(List[Char]) => List[Char], E],
                                 check_lab:
                                   F => (List[(term[G, H], term[I, J])]) =>
  sum[(List[Char]) => List[Char], K],
                                 check_model_lab:
                                   (List[(term[lab[B, C], D],
   term[lab[B, C], D])]) =>
                                     (List[(term[lab[B, C], D],
     term[lab[B, C], D])]) =>
                                       sum[(List[Char]) => List[Char], L],
                                 lge: (lab[B, C]) =>
nat => (sum[C, List[lab[B, C]]]) => (sum[C, List[lab[B, C]]]) => Boolean,
                                 ls: (lab[B, C]) =>
                                       nat =>
 (sum[C, List[lab[B, C]]]) => Boolean,
                                 rm: Option[((I, nat)) =>
      List[(term[I, J], term[I, J])]],
                                 p: List[(term[G, H], term[I, J])],
                                 r: List[(term[lab[B, C], D],
   term[lab[B, C], D])],
                                 lP: F,
                                 lR: List[(term[lab[B, C], D],
    term[lab[B, C], D])]):
      sum[(List[Char]) => List[Char], Unit] =
  rm match {
    case None => Inr[Unit, (List[Char]) => List[Char]](())
    case Some(irm) =>
      binda[(List[Char]) => List[Char], A,
             Unit](valid,
                    (_: A) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](catcha[((term[G, H], term[I, J]),
    (List[Char]) => List[Char]),
   Unit,
   (List[Char]) =>
     List[Char]](forallm[(term[G, H], term[I, J]),
                          (List[Char]) =>
                            List[Char]]((a: (term[G, H], term[I, J])) =>
  {
    val (s, t): (term[G, H], term[I, J]) = a;
    binda[(List[Char]) => List[Char], Unit,
           Unit](check_no_var[G, H](s),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](check_no_var[I, J](t),
                                  (_: Unit) =>
                                    check_no_defined_root[I, J](irm, t)))
  },
 p),
                  (x: ((term[G, H], term[I, J]), (List[Char]) => List[Char])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](snd[(term[G, H], term[I, J]),
                                    (List[Char]) => List[Char]](x))),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], E,
     Unit](check_decr(lR),
            (_: E) =>
              binda[(List[Char]) => List[Char], K,
                     Unit](catcha[(List[Char]) => List[Char], K,
                                   (List[Char]) =>
                                     List[Char]]((check_lab(lP))(p),
          (x: (List[Char]) => List[Char]) =>
            Inl[(List[Char]) => List[Char],
                 K](comp[List[Char], List[Char],
                          List[Char]](shows_string.apply(List('n', 'o', 't',
                       ' ', 'a', 'l', 'l', ' ', 'l', 'a', 'b', 'e', 'l', 'e',
                       'd', ' ', 'D', 'P', 's', ' ', 'p', 'r', 'e', 's', 'e',
                       'n', 't', ':')),
                                       comp[List[Char], List[Char],
     List[Char]](shows_nl, x)))),
                            (_: K) =>
                              binda[(List[Char]) => List[Char], L,
                                     Unit](catcha[(List[Char]) => List[Char], L,
           (List[Char]) =>
             List[Char]]((check_model_lab(lR))(r),
                          (x: (List[Char]) => List[Char]) =>
                            Inl[(List[Char]) => List[Char],
                                 L](comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
                                       'w', 'i', 't', 'h', ' ', 'l', 'a', 'b',
                                       'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S',
                                       ':')),
               comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
    (_: L) =>
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]]((check_sl_only_lab_and_decr[B, C,
                 D](lge, ls, r)).apply(lR),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](x, shows_nl))))))))
  }

def lab_root[A, B, C, D,
              E](i: A => (List[B]) => B, la: A => (List[B]) => C,
                  l: A => (List[B]) => C, lc: A => nat => C => D, alpha: E => B,
                  x5: term[A, E]):
      term[D, E] =
  (i, la, l, lc, alpha, x5) match {
  case (i, la, l, lc, alpha, Fun(f, ts)) =>
    {
      val clts: List[(B, term[D, E])] =
        mapa[term[A, E],
              (B, term[D, E])]((a: term[A, E]) =>
                                 eval_lab[A, B, C, D, E](i, la, lc, alpha, a),
                                ts)
      val cs: List[B] =
        mapa[(B, term[D, E]),
              B]((a: (B, term[D, E])) => fst[B, term[D, E]](a), clts)
      val a: List[term[D, E]] =
        mapa[(B, term[D, E]),
              term[D, E]]((a: (B, term[D, E])) => snd[B, term[D, E]](a), clts);
      Fun[D, E](((lc(f))(size_list[term[A, E]](ts)))((l(f))(cs)), a)
    }
  case (uu, uv, uw, ux, uy, Var(x)) => Var[E, D](x)
}

def check_sl_rule_all_ass[A: show, B: show, C, D: equal: show,
                           E: equal: show](i: A => (List[B]) => B,
    lb: A => (List[B]) => C, la: A => (List[B]) => C, lc: A => nat => C => D,
    gen_smaller: C => List[C], lR: List[(term[D, E], term[D, E])],
    alpha: E => B, x7: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  (i, lb, la, lc, gen_smaller, lR, alpha, x7) match {
  case (i, lb, la, lc, gen_smaller, lR, alpha, (l, Fun(f, ts))) =>
    {
      val ll: term[D, E] = lab_root[A, B, C, D, E](i, lb, la, lc, alpha, l)
      val clts: List[(B, term[D, E])] =
        mapa[term[A, E],
              (B, term[D, E])]((a: term[A, E]) =>
                                 eval_lab[A, B, C, D, E](i, lb, lc, alpha, a),
                                ts)
      val lts: List[term[D, E]] =
        mapa[(B, term[D, E]),
              term[D, E]]((a: (B, term[D, E])) => snd[B, term[D, E]](a), clts)
      val ld: C =
        (la(f))(mapa[(B, term[D, E]),
                      B]((a: (B, term[D, E])) => fst[B, term[D, E]](a), clts))
      val n: nat = size_list[term[A, E]](ts)
      val small: List[C] = gen_smaller(ld);
      catcha[(C, (List[Char]) => List[Char]), Unit,
              (List[Char]) =>
                List[Char]](forallm[C, (List[Char]) =>
 List[Char]]((le: C) =>
               check[(List[Char]) =>
                       List[Char]](membera[(term[D, E],
     term[D, E])](lR, (ll, Fun[D, E](((lc(f))(n))(le), lts))),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('l', 'a', 'b', 'e', 'l', 'e', 'd', ' ',
                                       'r', 'u', 'l', 'e', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_rule[D,
     E]((ll, Fun[D, E](((lc(f))(n))(le), lts))),
                                  shows_string.apply(List(' ', 'm', 'i', 's',
                   's', 'i', 'n', 'g'))))),
              small),
                             (x: (C, (List[Char]) => List[Char])) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](snd[C,
       (List[Char]) => List[Char]](x)))
    }
  case (i, lb, la, lc, gen_smaller, lR, alpha, (l, Var(x))) =>
    {
      val ll: term[D, E] = lab_root[A, B, C, D, E](i, lb, la, lc, alpha, l)
      val lr: term[D, E] =
        lab_root[A, B, C, D, E](i, lb, la, lc, alpha, Var[E, A](x));
      check[(List[Char]) =>
              List[Char]](membera[(term[D, E], term[D, E])](lR, (ll, lr)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('l', 'a',
                              'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                              ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_rule[D, E]((ll, lr)),
                         shows_string.apply(List(' ', 'm', 'i', 's', 's', 'i',
          'n', 'g')))))
    }
}

def check_sl_rule_all[A: show, B: show, C, D: equal: show,
                       E: equal: show](i: A => (List[B]) => B,
lb: A => (List[B]) => C, la: A => (List[B]) => C, lc: A => nat => C => D,
c: List[B], gen_smaller: C => List[C], lR: List[(term[D, E], term[D, E])],
x7: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  (i, lb, la, lc, c, gen_smaller, lR, x7) match {
  case (i, lb, la, lc, c, gen_smaller, lR, (l, r)) =>
    catcha[(E => B, (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[E => B,
                                   (List[Char]) =>
                                     List[Char]]((alpha: E => B) =>
           check_sl_rule_all_ass[A, B, C, D,
                                  E](i, lb, la, lc, gen_smaller, lR, alpha,
                                      (l, r)),
          mapa[List[(E, B)],
                E => B]((a: List[(E, B)]) =>
                          (b: E) => assign_from_vector[E, B](a, b),
                         enum_vectors[B, E](c,
     list_union[E](vars_term_list[A, E].apply(l),
                    vars_term_list[A, E].apply(r))))),
                           (x: (E => B, (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[E => B,
     (List[Char]) => List[Char]](x)))
}

def check_sl_lab_all_trs[A: equal: show, B: equal: show, C: show,
                          D: equal: show](i: (lab[A, B]) => (List[C]) => C,
   la: (lab[A, B]) => (List[C]) => sum[B, List[lab[A, B]]],
   l: (lab[A, B]) => (List[C]) => sum[B, List[lab[A, B]]], c: List[C],
   gen: (sum[B, List[lab[A, B]]]) => List[sum[B, List[lab[A, B]]]],
   lP: List[(term[lab[A, B], D], term[lab[A, B], D])],
   p: List[(term[lab[A, B], D], term[lab[A, B], D])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[lab[A, B], D], term[lab[A, B], D]), (List[Char]) => List[Char]),
          Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[lab[A, B], D], term[lab[A, B], D]),
                                 (List[Char]) =>
                                   List[Char]]((a:
          (term[lab[A, B], D], term[lab[A, B], D]))
         => check_sl_rule_all[lab[A, B], C, sum[B, List[lab[A, B]]], lab[A, B],
                               D](i, la, l,
                                   (aa: lab[A, B]) =>
                                     (b: nat) =>
                                       (ca: sum[B, List[lab[A, B]]]) =>
 labela[A, B](aa, b, ca),
                                   c, gen, lP, a),
        p),
                         (x: ((term[lab[A, B], D], term[lab[A, B], D]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[lab[A, B], D],
       term[lab[A, B], D]),
      (List[Char]) => List[Char]](x)))

def qmodel_gen(x0: List[nat]): List[List[nat]] = x0 match {
  case Nil => List(Nil)
  case x :: xs =>
    {
      val ys: List[nat] = upt(Zero_nat, Suc(x))
      val zs: List[List[nat]] = qmodel_gen(xs);
      maps[nat, List[nat]]((y: nat) =>
                             mapa[List[nat],
                                   List[nat]]((a: List[nat]) => y :: a, zs),
                            ys)
    }
}

def qmodel_gen2[A, B](x0: sum[List[nat], A]): List[sum[List[nat], B]] = x0 match
  {
  case Inl(l) =>
    mapa[List[nat],
          sum[List[nat],
               B]]((a: List[nat]) => Inl[List[nat], B](a), qmodel_gen(l))
}

def create_sem_lab_proc[A: equal: show,
                         B: equal: show](sli: sl_inter[A, List[nat]],
  x1: sum[(lab[A, List[nat]]) =>
            (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]],
           (B, Boolean)],
  rm: Option[((lab[A, List[nat]], nat)) =>
               List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]],
  r1: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])],
  r2: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])],
  r3: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])],
  r4: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (sli, x1, rm, r1, r2, r3, r4) match {
  case (sli, Inl(l), rm, r1, r2, r3, r4) =>
    {
      val c: nat = get_largest_element[A, List[nat]](sli)
      val ca: List[nat] = upt(Zero_nat, Suc(c))
      val i: (lab[A, List[nat]]) => (List[nat]) => nat =
        (a: lab[A, List[nat]]) =>
          (b: List[nat]) => sl_inter_to_inter[A, List[nat]](sli, a, b);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_sem_lab_proc[A, List[nat],
        B](model_check_valid[(List[Char]) => List[Char]],
            model_check_decr[List[(term[lab[A, List[nat]], B],
                                    term[lab[A, List[nat]], B])],
                              (List[Char]) => List[Char]],
            (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => (b: List[(term[lab[A, List[nat]], B],
                            term[lab[A, List[nat]], B])])
                   => check_sl_lab_trs[A, List[nat], nat,
B](i, l, ca, model_cge[nat], a, b),
            (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => (b: List[(term[lab[A, List[nat]], B],
                            term[lab[A, List[nat]], B])])
                   => check_sl_model_lab_trs[A, List[nat], nat,
      B](i, l, ca, model_cge[nat], a, b),
            model_lge[lab[A, List[nat]], nat,
                       sum[List[nat], List[lab[A, List[nat]]]]],
            model_LS[lab[A, List[nat]], nat,
                      sum[List[nat], List[lab[A, List[nat]]]]],
            r1, r2, r3, r4),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_prec_sl_inter[A, List[nat]](Zero_nat, sli),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
    }
  case (sli, Inr((v, false)), rm, r1, r2, r3, r4) =>
    {
      val c: nat = get_largest_element[A, List[nat]](sli)
      val ca: List[nat] = upt(Zero_nat, Suc(c))
      val i: (lab[A, List[nat]]) => (List[nat]) => nat =
        (a: lab[A, List[nat]]) =>
          (b: List[nat]) => sl_inter_to_inter[A, List[nat]](sli, a, b)
      val sig: List[(lab[A, List[nat]], nat)] =
        funas_trs_list[lab[A, List[nat]], B](r1 ++ r2)
      val l: (lab[A, List[nat]]) =>
               (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]]
        = qmodel_L[A, List[nat], nat, List[lab[A, List[nat]]]](sig);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_sem_lab_proc[A, List[nat],
        B](qmodel_check_valid[A, List[nat]](sli),
            qmodel_check_decr[A, B](sig, v, c),
            (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => (b: List[(term[lab[A, List[nat]], B],
                            term[lab[A, List[nat]], B])])
                   => check_sl_lab_trs[A, List[nat], nat,
B](i, l, ca, qmodel_cge, a, b),
            (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => (b: List[(term[lab[A, List[nat]], B],
                            term[lab[A, List[nat]], B])])
                   => check_sl_model_lab_trs[A, List[nat], nat,
      B](i, l, ca, qmodel_cge, a, b),
            (a: lab[A, List[nat]]) =>
              (b: nat) =>
                qmodel_lge[lab[A, List[nat]], nat, List[lab[A, List[nat]]],
                            List[lab[A, List[nat]]]](a, b),
            qmodel_LS[lab[A, List[nat]], nat, List[lab[A, List[nat]]]](sig, ca),
            r1, r2, r3, r4),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_prec_sl_inter[A, List[nat]](Zero_nat, sli),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
    }
  case (sli, Inr((v, true)), rm, r1, r2, r3, r4) =>
    {
      val c: nat = get_largest_element[A, List[nat]](sli)
      val ca: List[nat] = upt(Zero_nat, Suc(c))
      val i: (lab[A, List[nat]]) => (List[nat]) => nat =
        (a: lab[A, List[nat]]) =>
          (b: List[nat]) => sl_inter_to_inter[A, List[nat]](sli, a, b)
      val sig1: List[(lab[A, List[nat]], nat)] =
        list_union[(lab[A, List[nat]],
                     nat)](funas_below_trs_list[lab[A, List[nat]], B](r1),
                            funas_trs_list[lab[A, List[nat]], B](r2))
      val sig2: List[(lab[A, List[nat]], nat)] =
        funas_root_rules[lab[A, List[nat]], B](r1)
      val l: (lab[A, List[nat]]) =>
               (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]]
        = qmodel_L[A, List[nat], nat, List[lab[A, List[nat]]]](sig1)
      val la: (lab[A, List[nat]]) =>
                (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]]
        = qmodel_L[A, List[nat], nat, List[lab[A, List[nat]]]](sig2);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_sem_lab_root_proc[Unit, A, List[nat], B, Unit,
             List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])],
             lab[A, List[nat]], B, lab[A, List[nat]], B, Unit,
             Unit](qmodel_check_valid[A, List[nat]](sli),
                    qmodel_check_decr[A, B](sig1, v, c),
                    (a: List[(term[lab[A, List[nat]], B],
                               term[lab[A, List[nat]], B])])
                      => (b: List[(term[lab[A, List[nat]], B],
                                    term[lab[A, List[nat]], B])])
                           => check_sl_lab_all_trs[A, List[nat], nat,
            B](i, l, la, ca,
                (aa: sum[List[nat], List[lab[A, List[nat]]]]) =>
                  qmodel_gen2[List[lab[A, List[nat]]],
                               List[lab[A, List[nat]]]](aa),
                a, b),
                    (a: List[(term[lab[A, List[nat]], B],
                               term[lab[A, List[nat]], B])])
                      => (b: List[(term[lab[A, List[nat]], B],
                                    term[lab[A, List[nat]], B])])
                           => check_sl_model_lab_trs[A, List[nat], nat,
              B](i, l, ca, qmodel_cge, a, b),
                    (a: lab[A, List[nat]]) =>
                      (b: nat) =>
                        qmodel_lge[lab[A, List[nat]], nat,
                                    List[lab[A, List[nat]]],
                                    List[lab[A, List[nat]]]](a, b),
                    qmodel_LS[lab[A, List[nat]], nat,
                               List[lab[A, List[nat]]]](sig1, ca),
                    rm, r1, r2, r3, r4),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_prec_sl_inter[A, List[nat]](Zero_nat, sli),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
    }
}

def create_rseq_map[A: equal,
                     B: equal](rseqs:
                                 List[((term[A, B], term[A, B]),
List[(pos, ((term[A, B], term[A, B]), term[A, B]))])],
                                r: (term[A, B], term[A, B])):
      Option[List[(pos, ((term[A, B], term[A, B]), term[A, B]))]] =
  lookupa[(term[A, B], term[A, B]),
           List[(pos, ((term[A, B], term[A, B]), term[A, B]))]](r, rseqs)

def check_strict[A: equal: show,
                  B: equal: show](ra: List[(term[A, B], term[A, B])],
                                   rseqm:
                                     ((term[A, B], term[A, B])) =>
                                       Option[List[(pos,
             ((term[A, B], term[A, B]), term[A, B]))]],
                                   p: A => nat, r: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val s: term[A, B] = proj_term[A, B](p, fst[term[A, B], term[A, B]](r))
    val t: term[A, B] = proj_term[A, B](p, snd[term[A, B], term[A, B]](r));
    rseqm(r) match {
      case None => check_supt[A, B](s, t)
      case Some(rseq) =>
        (if (equal_nata(size_list[(pos, ((term[A, B], term[A, B]),
  term[A, B]))](rseq),
                         Zero_nat))
          check_supt[A, B](s, t)
          else {
                 val u: term[A, B] =
                   snd[(term[A, B], term[A, B]),
                        term[A, B]](snd[pos,
 ((term[A, B], term[A, B]),
   term[A, B])](last[(pos, ((term[A, B], term[A, B]), term[A, B]))](rseq)));
                 binda[(List[Char]) => List[Char], Unit,
                        Unit](check_rsteps[A, B](ra, rseq, s, u),
                               (_: Unit) => check_supteq[A, B](u, t))
               })
    }
  }

def check_weak[A: equal: show,
                B: equal: show](p: A => nat, r: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](check[(List[Char]) =>
                                List[Char]](equal_terma[A,
                 B](proj_term[A, B](p, fst[term[A, B], term[A, B]](r)),
                     proj_term[A, B](p, snd[term[A, B], term[A, B]](r))),
     comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'p', 'r',
        'o', 'j', 'e', 'c', 't', 'e', 'd', ' ', 'l', 'h', 's', ' ', 'i', 's',
        ' ', 'n', 'o', 't', ' ', 'e', 'q', 'u', 'a', 'l', ' ', 't', 'o', ' ',
        't', 'h', 'e', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', ' ',
        'r', 'h', 's')),
                        shows_nl)),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('C', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'o', 'r', 'i', 'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_rule[A, B](r),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(',', ' ', 's', 'i', 'n', 'c', 'e')),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_prec_term[A, B](Zero_nat,
                                   proj_term[A,
      B](p, fst[term[A, B], term[A, B]](r))),
             comp[List[Char], List[Char],
                   List[Char]](shows_string.apply(List(' ', '!', '=', ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_term[A,
                           B](Zero_nat,
                               proj_term[A,
  B](p, snd[term[A, B], term[A, B]](r))),
           comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))))

def check_subterm_criterion_proc[A: equal: show,
                                  B: equal: show](pb: A => nat,
           rseqm:
             ((term[A, B], term[A, B])) =>
               Option[List[(pos, ((term[A, B], term[A, B]), term[A, B]))]],
           pa: List[(term[A, B], term[A, B])],
           p: List[(term[A, B], term[A, B])], r: List[(term[A, B], term[A, B])],
           x5: Option[((A, nat)) => List[(term[A, B], term[A, B])]]):
      sum[(List[Char]) => List[Char], Unit] =
  (pb, rseqm, pa, p, r, x5) match {
  case (pb, rseqm, pa, p, r, None) => Inr[Unit, (List[Char]) => List[Char]](())
  case (pb, rseqm, pa, p, r, Some(rm)) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]),
                         Unit,
                         (List[Char]) =>
                           List[Char]](forallm[(term[A, B], term[A, B]),
        (List[Char]) =>
          List[Char]]((a: (term[A, B], term[A, B])) =>
                        {
                          val (l, ra): (term[A, B], term[A, B]) = a;
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](check_no_var[A, B](l),
(_: Unit) =>
  binda[(List[Char]) => List[Char], Unit,
         Unit](check_no_var[A, B](ra),
                (_: Unit) => check_no_defined_root[A, B](rm, ra)))
                        },
                       pa),
(x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
  Inl[(List[Char]) => List[Char],
       Unit](snd[(term[A, B], term[A, B]), (List[Char]) => List[Char]](x))),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](catcha[((term[A, B], term[A, B]),
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[(term[A, B], term[A, B]),
                        (List[Char]) =>
                          List[Char]]((a: (term[A, B], term[A, B])) =>
check_strict[A, B](r, rseqm, pb, a),
                                       list_diff[(term[A, B],
           term[A, B])](pa, p)),
                (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[(term[A, B], term[A, B]),
                                  (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    catcha[((term[A, B], term[A, B]),
     (List[Char]) => List[Char]),
    Unit,
    (List[Char]) =>
      List[Char]](forallm[(term[A, B], term[A, B]),
                           (List[Char]) =>
                             List[Char]]((a: (term[A, B], term[A, B])) =>
   check_weak[A, B](pb, a),
  p),
                   (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char]))
                     => Inl[(List[Char]) => List[Char],
                             Unit](snd[(term[A, B], term[A, B]),
(List[Char]) => List[Char]](x)))))
}

def create_subterm_criterion_proc[A: equal: show,
                                   B: equal: show](pa: projL[A],
            rseqs:
              List[((term[A, B], term[A, B]),
                     List[(pos, ((term[A, B], term[A, B]), term[A, B]))])],
            p: List[(term[A, B], term[A, B])],
            rP: List[(term[A, B], term[A, B])],
            r: List[(term[A, B], term[A, B])],
            rm: Option[((A, nat)) => List[(term[A, B], term[A, B])]]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val pi: A => nat = (a: A) => create_proj[A](pa, a)
    val rseqm:
          ((term[A, B], term[A, B])) =>
            Option[List[(pos, ((term[A, B], term[A, B]), term[A, B]))]]
      = (a: (term[A, B], term[A, B])) => create_rseq_map[A, B](rseqs, a);
    check_subterm_criterion_proc[A, B](pi, rseqm, p, rP, r, rm)
  }

def get_arg[A, B](t: term[A, B], x1: nat): term[A, B] = (t, x1) match {
  case (t, Zero_nat) => t
  case (t, Suc(n)) => nth[term[A, B]](args[A, B](t), n)
}

abstract sealed class scg[+A, +B]
final case object Null extends scg[Nothing, Nothing]
final case class Scg[A, B](a: A, b: A, c: List[(B, B)], d: List[(B, B)]) extends
  scg[A, B]

def compa[A: equal](esa: List[(A, A)], es: List[(A, A)]): List[(A, A)] =
  remdups[(A, A)](maps[(A, A),
                        (A, A)]((a: (A, A)) =>
                                  {
                                    val (x, y): (A, A) = a;
                                    maps[(A, A),
  (A, A)]((aa: (A, A)) =>
            {
              val (ya, z): (A, A) = aa;
              (if (eq[A](y, ya)) List((x, z)) else Nil)
            },
           es)
                                  },
                                 esa))

def subtract_list[A: equal](xs: List[A], ys: List[A]): List[A] =
  filter[A]((x: A) => ! (membera[A](ys, x)), xs)

def scg_comp[A, B: equal](conn: A => A => Boolean, x1: scg[A, B], g: scg[A, B]):
      scg[A, B] =
  (conn, x1, g) match {
  case (conn, Scg(pa, qa, stra, wka), Scg(p, q, str, wk)) =>
    (if (! ((conn(qa))(p))) Null
      else {
             val strs: List[(B, B)] =
               remdups[(B, B)](compa[B](stra, str) ++
                                 (compa[B](stra, wk) ++ compa[B](wka, str)))
             val a: List[(B, B)] =
               subtract_list[(B, B)](remdups[(B, B)](compa[B](wka, wk)), strs);
             Scg[A, B](pa, q, strs, a)
           })
  case (conn, Null, g) => Null
  case (conn, Scg(v, va, vb, vc), Null) => Null
}

def equal_scg[A: equal, B: equal](x0: scg[A, B], x1: scg[A, B]): Boolean =
  (x0, x1) match {
  case (Scg(pp1, pp2, list1, list2), Null) => false
  case (Null, Scg(pp1, pp2, list1, list2)) => false
  case (Scg(pp1a, pp2a, list1a, list2a), Scg(pp1, pp2, list1, list2)) =>
    eq[A](pp1a, pp1) &&
      (eq[A](pp2a, pp2) &&
        (equal_lista[(B, B)](list1a, list1) &&
          equal_lista[(B, B)](list2a, list2)))
  case (Null, Null) => true
}

def subsumes[A: equal, B: equal](g: scg[A, B], x1: scg[A, B]): Boolean = (g, x1)
  match {
  case (Scg(pa, qa, stra, wka), Scg(p, q, str, wk)) =>
    eq[A](pa, p) &&
      (eq[A](qa, q) &&
        (list_all[(B, B)]((a: (B, B)) => membera[(B, B)](str, a), stra) &&
          list_all[(B, B)]((a: (B, B)) => membera[(B, B)](str ++ wk, a), wka)))
  case (g, Null) => true
  case (Null, Scg(v, va, vb, vc)) => false
}

def insert_scg[A: equal, B: equal](g: scg[A, B], gs: List[scg[A, B]]):
      Option[List[scg[A, B]]] =
  (if (equal_scg[A, B](g, Null)) None
    else (if (list_ex[scg[A, B]]((ga: scg[A, B]) => subsumes[A, B](ga, g), gs))
           None
           else Some[List[scg[A, B]]](g ::
filter[scg[A, B]]((ga: scg[A, B]) => ! (subsumes[A, B](g, ga)), gs))))

def mk_trancl[A: equal,
               B: equal](conn: A => A => Boolean, x1: List[scg[A, B]],
                          s: List[scg[A, B]]):
      List[scg[A, B]] =
  (conn, x1, s) match {
  case (conn, Nil, s) => s
  case (conn, g :: gs, s) =>
    insert_scg[A, B](g, s) match {
      case None => mk_trancl[A, B](conn, gs, s)
      case Some(sa) =>
        mk_trancl[A, B](conn,
                         mapa[scg[A, B],
                               scg[A, B]]((a: scg[A, B]) =>
    scg_comp[A, B](conn, g, a),
   sa) ++
                           (mapa[scg[A, B],
                                  scg[A, B]]((f: scg[A, B]) =>
       scg_comp[A, B](conn, f, g),
      sa) ++
                             gs),
                         sa)
    }
}

def in_situ[A, B: equal](x0: scg[A, B]): Boolean = x0 match {
  case Null => true
  case Scg(p, q, str, wk) =>
    list_ex[(B, B)]((a: (B, B)) => {
                                     val (aa, b): (B, B) = a;
                                     eq[B](aa, b)
                                   },
                     str)
}

def combine[A, B: equal](x0: scg[A, B], s: scg[A, B]): scg[A, B] = (x0, s) match
  {
  case (Scg(pa, qa, stra, wka), Scg(p, q, str, wk)) =>
    Scg[A, B](pa, qa, remdups[(B, B)](stra ++ str), remdups[(B, B)](wka ++ wk))
  case (Null, s) => Null
  case (Scg(v, va, vb, vc), Null) => Null
}

def sagiv[A: equal, B: equal](conn: A => A => Boolean, g: scg[A, B]): Boolean =
  (if (in_situ[A, B](g)) true
    else {
           val gg: scg[A, B] = scg_comp[A, B](conn, g, g);
           (if (subsumes[A, B](gg, g)) false
             else sagiv[A, B](conn, combine[A, B](g, gg)))
         })

def check_SCT[A: equal, B: equal](conn: A => A => Boolean, gs: List[scg[A, B]]):
      Boolean =
  {
    val a: List[scg[A, B]] = mk_trancl[A, B](conn, gs, Nil);
    list_all[scg[A, B]]((aa: scg[A, B]) => sagiv[A, B](conn, aa), a)
  }

def check_sct_subterm_proc[A: equal: show,
                            B: equal: show](p: List[(term[A, B], term[A, B])],
     gs: List[((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)]))],
     rm: Option[((A, nat)) => List[(term[A, B], term[A, B])]],
     rrm: Option[((A, nat)) => List[(term[A, B], term[A, B])]]):
      sum[(List[Char]) => List[Char], Unit] =
  rm match {
    case None => Inr[Unit, (List[Char]) => List[Char]](())
    case Some(irm) =>
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](binda[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[(term[A, B], term[A, B]), Unit,
         (List[Char]) =>
           List[Char]](check_subseteq[(term[A, B],
term[A, B])](p, mapa[((term[A, B], term[A, B]),
                       (List[(nat, nat)], List[(nat, nat)])),
                      (term[A, B],
                        term[A, B])]((a:
((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)])))
                                       => fst[(term[A, B], term[A, B]),
       (List[(nat, nat)], List[(nat, nat)])](a),
                                      gs)),
                        (x: (term[A, B], term[A, B])) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List('t', 'h', 'e', 'r', 'e', ' ', 'i', 's',
' ', 'n', 'o', ' ', 's', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ',
'g', 'r', 'a', 'p', 'h', ' ', 'f', 'o', 'r', ' ', 'D', 'P', ' ')),
                comp[List[Char], List[Char],
                      List[Char]](shows_rule[A, B](x), shows_nl)))),
  (_: Unit) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](catcha[(((term[A, B], term[A, B]),
                           (List[(nat, nat)], List[(nat, nat)])),
                          (List[Char]) => List[Char]),
                         Unit,
                         (List[Char]) =>
                           List[Char]](forallm[((term[A, B], term[A, B]),
         (List[(nat, nat)], List[(nat, nat)])),
        (List[Char]) =>
          List[Char]]((a: ((term[A, B], term[A, B]),
                            (List[(nat, nat)], List[(nat, nat)])))
                        => {
                             val (aa, b):
                                   ((term[A, B], term[A, B]),
                                     (List[(nat, nat)], List[(nat, nat)]))
                               = a;
                             ({
                                val (s, t): (term[A, B], term[A, B]) = aa;
                                (ab: (List[(nat, nat)], List[(nat, nat)])) =>
                                  {
                                    val (stri, nstri):
  (List[(nat, nat)], List[(nat, nat)])
                                      = ab;
                                    catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](binda[(List[Char]) => List[Char], Unit,
                         Unit](check_no_var[A, B](s),
                                (_: Unit) =>
                                  binda[(List[Char]) => List[Char], Unit,
 Unit](check_no_var[A, B](t),
        (_: Unit) =>
          binda[(List[Char]) => List[Char], Unit,
                 Unit](check_no_defined_root[A, B](irm, t),
                        (_: Unit) =>
                          {
                            val m: nat = num_args[A, B](t)
                            val n: nat = num_args[A, B](s);
                            binda[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[((nat, nat),
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[(nat, nat),
                                (List[Char]) =>
                                  List[Char]]((ac: (nat, nat)) =>
        {
          val (i, j): (nat, nat) = ac;
          check[(List[Char]) =>
                  List[Char]](less_eq_nat(i, n) &&
                                (less_eq_nat(j, m) &&
                                  isOK[(List[Char]) => List[Char],
Unit](check_supt[A, B](get_arg[A, B](s, i), get_arg[A, B](t, j)))),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_string.apply(List('p',
                                  'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                                  't', 'h', ' ', 'e', 'd', 'g', 'e', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_prec_nat(Zero_nat, i),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List(' ', '|',
                                '>', ' ')),
        comp[List[Char], List[Char],
              List[Char]](shows_prec_nat(Zero_nat, j), shows_nl)))))
        },
       stri),
                        (x: ((nat, nat), (List[Char]) => List[Char])) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](snd[(nat, nat),
  (List[Char]) => List[Char]](x))),
  (_: Unit) =>
    catcha[((nat, nat), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[(nat, nat),
                                   (List[Char]) =>
                                     List[Char]]((ac: (nat, nat)) =>
           {
             val (i, j): (nat, nat) = ac;
             check[(List[Char]) =>
                     List[Char]](less_eq_nat(i, n) &&
                                   (less_eq_nat(j, m) &&
                                     isOK[(List[Char]) => List[Char],
   Unit](check_supteq[A, B](get_arg[A, B](s, i), get_arg[A, B](t, j)))),
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'i', 't', 'h', ' ', 'e', 'd', 'g', 'e',
                                     ' ')),
             comp[List[Char], List[Char],
                   List[Char]](shows_prec_nat(Zero_nat, i),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List(' ',
                                   '|', '>', '=', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_prec_nat(Zero_nat, j), shows_nl)))))
           },
          nstri),
                           (x: ((nat, nat), (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[(nat, nat),
     (List[Char]) => List[Char]](x))))
                          }))),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('p',
                                   'r', 'o', 'b', 'l', 'e', 'm', 's', ' ', 'w',
                                   'i', 't', 'h', ' ', 'D', 'P', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_rule[A, B]((s, t)),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))))
                                  }
                              })(b)
                           },
                       gs),
(x: (((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)])),
      (List[Char]) => List[Char]))
  => Inl[(List[Char]) => List[Char],
          Unit](snd[((term[A, B], term[A, B]),
                      (List[(nat, nat)], List[(nat, nat)])),
                     (List[Char]) => List[Char]](x))),
                  (_: Unit) =>
                    check[(List[Char]) =>
                            List[Char]](check_SCT[(term[A, B],
            (term[A, B], (hterm[A, B], hterm[A, B]))),
           nat]((a: (term[A, B], (term[A, B], (hterm[A, B], hterm[A, B])))) =>
                  {
                    val (_, (t, (_, ct))):
                          (term[A, B], (term[A, B], (hterm[A, B], hterm[A, B])))
                      = a;
                    (aa: (term[A, B], (term[A, B], (hterm[A, B], hterm[A, B]))))
                      => {
                           val (u, (_, (cu, _))):
                                 (term[A, B],
                                   (term[A, B], (hterm[A, B], hterm[A, B])))
                             = aa;
                           gcMatchI[A, B](ct, u) && gcMatchI[A, B](cu, t)
                         }
                  },
                 mapa[((term[A, B], term[A, B]),
                        (List[(nat, nat)], List[(nat, nat)])),
                       scg[(term[A, B],
                             (term[A, B], (hterm[A, B], hterm[A, B]))),
                            nat]]((a: ((term[A, B], term[A, B]),
(List[(nat, nat)], List[(nat, nat)])))
                                    => {
 val (aa, b): ((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)])) =
   a;
 ({
    val (s, t): (term[A, B], term[A, B]) = aa;
    (ab: (List[(nat, nat)], List[(nat, nat)])) =>
      {
        val (stri, nstri): (List[(nat, nat)], List[(nat, nat)]) = ab
        val e: (term[A, B], (term[A, B], (hterm[A, B], hterm[A, B]))) =
          (s, (t, ((capRM[A, B](rrm)).apply(s), (capRM[A, B](rm)).apply(t))));
        Scg[(term[A, B], (term[A, B], (hterm[A, B], hterm[A, B]))),
             nat](e, e, stri, nstri)
      }
  })(b)
                                       },
                                   gs)),
 comp[List[Char], List[Char],
       List[Char]](shows_string.apply(List('s', 'i', 'z', 'e', '-', 'c', 'h',
    'a', 'n', 'g', 'e', ' ', 'a', 'n', 'a', 'l', 'y', 's', 'i', 's', ' ', 'f',
    'a', 'i', 'l', 'e', 'd')),
                    shows_nl)))),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
     'o', 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i',
     'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e',
     's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 's', 'u', 'b', 't', 'e',
     'r', 'm', '-', 'r', 'e', 'l', 'a', 't', 'i', 'o', 'n', ' ', 'f', 'o', 'r',
     ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ',
     'r', 'e', 'a', 's', 'o', 'n')),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
  }

def check_sct_ur_af_entry[A: equal: show, B: equal: show, C, D, E,
                           F](rm: ((A, nat)) => List[(term[A, B], term[A, B])],
                               ur: List[(term[A, B], term[A, B])],
                               pi: A => nat => nat => Boolean,
                               sa: ((term[A, B], term[A, B])) => sum[C, D],
                               ns: ((term[A, B], term[A, B])) => sum[E, F],
                               s: term[A, B], t: term[A, B],
                               stri: List[(nat, nat)], nstri: List[(nat, nat)]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](binda[(List[Char]) => List[Char], Unit,
                               Unit](check_no_var[A, B](s),
                                      (_: Unit) =>
binda[(List[Char]) => List[Char], Unit,
       Unit](check_no_var[A, B](t),
              (_: Unit) =>
                binda[(List[Char]) => List[Char], Unit,
                       Unit](check_no_defined_root[A, B](rm, t),
                              (_: Unit) =>
                                {
                                  val m: nat = num_args[A, B](t)
                                  val n: nat = num_args[A, B](s);
                                  binda[(List[Char]) => List[Char], Unit,
 Unit](catcha[(nat, (List[Char]) => List[Char]), Unit,
               (List[Char]) =>
                 List[Char]](forallm[nat, (List[Char]) =>
    List[Char]]((i: nat) =>
                  check[(List[Char]) =>
                          List[Char]](less_eq_nat(i, n),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('l', 'e', 'f', 't', '-', 'i', 'n', 'd',
  'e', 'x', ' ', 't', 'o', ' ', 'l', 'a', 'r', 'g', 'e')),
                  comp[List[Char], List[Char],
                        List[Char]](shows_prec_nat(Zero_nat, i), shows_nl))),
                 remdups[nat](mapa[(nat, nat),
                                    nat]((a: (nat, nat)) => fst[nat, nat](a),
  stri ++ nstri))),
                              (x: (nat, (List[Char]) => List[Char])) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](snd[nat,
        (List[Char]) => List[Char]](x))),
        (_: Unit) =>
          binda[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(nat, (List[Char]) => List[Char]), Unit,
                               (List[Char]) =>
                                 List[Char]](forallm[nat,
              (List[Char]) =>
                List[Char]]((j: nat) =>
                              check[(List[Char]) =>
                                      List[Char]](less_eq_nat(j, m) &&
            isOK[(List[Char]) => List[Char],
                  Unit](check_ur_closed_term_rm_af[A,
            B](rm, ur, pi, get_arg[A, B](t, j))),
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('r', 'i', 'g', 'h', 't',
              '-', 'i', 'n', 'd', 'e', 'x', ' ', 't', 'o', ' ', 'l', 'a', 'r',
              'g', 'e', ' ', 'o', 'r', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
              't', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 'u', 's',
              'a', 'b', 'l', 'e', '-', 'r', 'u', 'l', 'e', 's', ' ', 'c', 'o',
              'n', 'd', 'i', 't', 'i', 'o', 'n')),
                              shows_prec_nat(Zero_nat, j))),
                             remdups[nat](mapa[(nat, nat),
        nat]((a: (nat, nat)) => snd[nat, nat](a), stri ++ nstri))),
      (x: (nat, (List[Char]) => List[Char])) =>
        Inl[(List[Char]) => List[Char],
             Unit](snd[nat, (List[Char]) => List[Char]](x))),
                        (_: Unit) =>
                          {
                            args[A, B](s);
                            args[A, B](t);
                            binda[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[((nat, nat),
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[(nat, nat),
                                (List[Char]) =>
                                  List[Char]]((a: (nat, nat)) =>
        {
          val (i, j): (nat, nat) = a;
          check[(List[Char]) =>
                  List[Char]](isOK[C, D](sa((get_arg[A, B](s, i),
      get_arg[A, B](t, j)))),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_string.apply(List('p',
                                  'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                                  't', 'h', ' ', 'e', 'd', 'g', 'e', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_prec_nat(Zero_nat, i),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List(' ', '-',
                                'S', '-', '>', ' ')),
        shows_prec_nat(Zero_nat, j)))))
        },
       stri),
                        (x: ((nat, nat), (List[Char]) => List[Char])) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](snd[(nat, nat),
  (List[Char]) => List[Char]](x))),
  (_: Unit) =>
    catcha[((nat, nat), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[(nat, nat),
                                   (List[Char]) =>
                                     List[Char]]((a: (nat, nat)) =>
           {
             val (i, j): (nat, nat) = a;
             check[(List[Char]) =>
                     List[Char]](isOK[E, F](ns((get_arg[A, B](s, i),
         get_arg[A, B](t, j)))),
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'i', 't', 'h', ' ', 'e', 'd', 'g', 'e',
                                     ' ')),
             comp[List[Char], List[Char],
                   List[Char]](shows_prec_nat(Zero_nat, i),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List(' ',
                                   '-', 'N', 'S', '-', '>', ' ')),
           shows_prec_nat(Zero_nat, j)))))
           },
          nstri),
                           (x: ((nat, nat), (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[(nat, nat),
     (List[Char]) => List[Char]](x))))
                          }))
                                }))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', 's',
 ' ', 'w', 'i', 't', 'h', ' ', 'D', 'P', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_rule[A, B]((s, t)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, x)))))

def check_sct_ur_af_proc[A: equal: show,
                          B: equal: show](rp:
    (sum[(List[Char]) => List[Char], Unit],
      (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
        (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
          (A => nat => nat => Boolean,
            (sum[(List[Char]) => List[Char], Unit],
              (List[Char]) => List[Char]))))),
   p: List[(term[A, B], term[A, B])],
   gs: List[((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)]))],
   rm: Option[((A, nat)) => List[(term[A, B], term[A, B])]],
   rrm: Option[((A, nat)) => List[(term[A, B], term[A, B])]],
   ur: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  rm match {
    case None => Inr[Unit, (List[Char]) => List[Char]](())
    case Some(irm) =>
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]]({
                              val ggs: List[((term[A, B], term[A, B]),
      (List[(nat, nat)], List[(nat, nat)]))]
                                = filter[((term[A, B], term[A, B]),
   (List[(nat, nat)],
     List[(nat, nat)]))]((g: ((term[A, B], term[A, B]),
                               (List[(nat, nat)], List[(nat, nat)])))
                           => membera[(term[A, B],
term[A, B])](p, fst[(term[A, B], term[A, B]),
                     (List[(nat, nat)], List[(nat, nat)])](g)),
                          gs)
                              val pi: A => nat => nat => Boolean =
                                rp_af[A, B](rp)
                              val s: ((term[A, B], term[A, B])) =>
                                       sum[(List[Char]) => List[Char], Unit]
                                = rp_strict[A, B](rp)
                              val ns: ((term[A, B], term[A, B])) =>
sum[(List[Char]) => List[Char], Unit]
                                = rp_non_strict[A, B](rp);
                              binda[(List[Char]) => List[Char], Unit,
                                     Unit](rp_valid[A, B](rp),
    (_: Unit) =>
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[(List[Char]) => List[Char], Unit,
                           (List[Char]) =>
                             List[Char]](catcha[((term[A, B], term[A, B]),
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[(term[A, B], term[A, B]),
                                (List[Char]) =>
                                  List[Char]]((lr: (term[A, B], term[A, B])) =>
        check_ur_closed_term_rm_af[A, B](irm, ur, pi,
  snd[term[A, B], term[A, B]](lr)),
       ur),
                        (x: ((term[A, B], term[A, B]),
                              (List[Char]) => List[Char]))
                          => Inl[(List[Char]) => List[Char],
                                  Unit](snd[(term[A, B], term[A, B]),
     (List[Char]) => List[Char]](x))),
  (x: (List[Char]) => List[Char]) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('e', 'r', 'r', 'o',
                  'r', ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k',
                  'i', 'n', 'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ',
                  'p', 'r', 'o', 'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o',
                  'f', ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ', 'u', 's', 'a',
                  'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, x)))),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](catcha[(term[A, B], term[A, B]), Unit,
   (List[Char]) =>
     List[Char]](check_subseteq[(term[A, B],
                                  term[A,
B])](p, mapa[((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)])),
              (term[A, B],
                term[A, B])]((a: ((term[A, B], term[A, B]),
                                   (List[(nat, nat)], List[(nat, nat)])))
                               => fst[(term[A, B], term[A, B]),
                                       (List[(nat, nat)], List[(nat, nat)])](a),
                              ggs)),
                  (x: (term[A, B], term[A, B])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](comp[List[Char], List[Char],
                                     List[Char]](shows_string.apply(List('t',
                                  'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n',
                                  'o', ' ', 's', 'i', 'z', 'e', '-', 'c', 'h',
                                  'a', 'n', 'g', 'e', ' ', 'g', 'r', 'a', 'p',
                                  'h', ' ', 'f', 'o', 'r', ' ', 'D', 'P', ' ')),
          shows_rule[A, B](x)))),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], Unit,
     Unit](catcha[(((term[A, B], term[A, B]),
                     (List[(nat, nat)], List[(nat, nat)])),
                    (List[Char]) => List[Char]),
                   Unit,
                   (List[Char]) =>
                     List[Char]](forallm[((term[A, B], term[A, B]),
   (List[(nat, nat)], List[(nat, nat)])),
  (List[Char]) =>
    List[Char]]((a: ((term[A, B], term[A, B]),
                      (List[(nat, nat)], List[(nat, nat)])))
                  => {
                       val (aa, b):
                             ((term[A, B], term[A, B]),
                               (List[(nat, nat)], List[(nat, nat)]))
                         = a;
                       ({
                          val (sa, t): (term[A, B], term[A, B]) = aa;
                          (ab: (List[(nat, nat)], List[(nat, nat)])) =>
                            {
                              val (ac, ba): (List[(nat, nat)], List[(nat, nat)])
                                = ab;
                              check_sct_ur_af_entry[A, B,
             (List[Char]) => List[Char], Unit, (List[Char]) => List[Char],
             Unit](irm, ur, pi, s, ns, sa, t, ac, ba)
                            }
                        })(b)
                     },
                 ggs),
                                  (x: (((term[A, B], term[A, B]),
 (List[(nat, nat)], List[(nat, nat)])),
(List[Char]) => List[Char]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](snd[((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)])),
               (List[Char]) => List[Char]](x))),
            (_: Unit) =>
              binda[(List[Char]) => List[Char], Unit,
                     Unit](catcha[(List[Char]) => List[Char], Unit,
                                   (List[Char]) =>
                                     List[Char]](catcha[((term[A, B],
                   term[A, B]),
                  (List[Char]) => List[Char]),
                 Unit,
                 (List[Char]) =>
                   List[Char]](forallm[(term[A, B], term[A, B]),
(List[Char]) => List[Char]](rp_non_strict[A, B](rp), ur),
                                (x: ((term[A, B], term[A, B]),
                                      (List[Char]) => List[Char]))
                                  => Inl[(List[Char]) => List[Char],
  Unit](snd[(term[A, B], term[A, B]), (List[Char]) => List[Char]](x))),
          (x: (List[Char]) => List[Char]) =>
            Inl[(List[Char]) => List[Char],
                 Unit](comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('p', 'r', 'o',
                          'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
                          'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'u', 's',
                          'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
  comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                            (_: Unit) =>
                              check[(List[Char]) =>
                                      List[Char]](check_SCT[(term[A, B],
                      (term[A, B], (hterm[A, B], hterm[A, B]))),
                     nat]((a: (term[A, B],
                                (term[A, B], (hterm[A, B], hterm[A, B]))))
                            => {
                                 val (_, (t, (_, ct))):
                                       (term[A, B],
 (term[A, B], (hterm[A, B], hterm[A, B])))
                                   = a;
                                 (aa: (term[A, B],
(term[A, B], (hterm[A, B], hterm[A, B]))))
                                   => {
val (u, (_, (cu, _))): (term[A, B], (term[A, B], (hterm[A, B], hterm[A, B]))) =
  aa;
gcMatchI[A, B](ct, u) && gcMatchI[A, B](cu, t)
                                      }
                               },
                           mapa[((term[A, B], term[A, B]),
                                  (List[(nat, nat)], List[(nat, nat)])),
                                 scg[(term[A, B],
                                       (term[A, B],
 (hterm[A, B], hterm[A, B]))),
                                      nat]]((a:
       ((term[A, B], term[A, B]), (List[(nat, nat)], List[(nat, nat)])))
      => {
           val (aa, b):
                 ((term[A, B], term[A, B]),
                   (List[(nat, nat)], List[(nat, nat)]))
             = a;
           ({
              val (sa, t): (term[A, B], term[A, B]) = aa;
              (ab: (List[(nat, nat)], List[(nat, nat)])) =>
                {
                  val (stri, nstri): (List[(nat, nat)], List[(nat, nat)]) = ab
                  val e: (term[A, B], (term[A, B], (hterm[A, B], hterm[A, B])))
                    = (sa, (t, ((capRM[A, B](rrm)).apply(sa),
                                 (capRM[A, B](rm)).apply(t))));
                  Scg[(term[A, B], (term[A, B], (hterm[A, B], hterm[A, B]))),
                       nat](e, e, stri, nstri)
                }
            })(b)
         },
     ggs)),
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('s', 'i', 'z', 'e', '-',
              'c', 'h', 'a', 'n', 'g', 'e', ' ', 'a', 'n', 'a', 'l', 'y', 's',
              'i', 's', ' ', 'f', 'a', 'i', 'l', 'e', 'd')),
                              shows_nl)))))))
                            },
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
     'o', 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i',
     'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e',
     's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f',
     'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](rp_desc[A, B](rp),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('f', 'o', 'r', ' ', 't', 'h', 'e', ' ',
 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'r', 'e', 'a', 's', 'o',
 'n')),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))
  }

def checkDPProof[A: equal: show](i: (List[Char]) => List[Char],
                                  p: List[(term[lab[A, List[nat]], List[Char]],
    term[lab[A, List[nat]], List[Char]])],
                                  r: List[(term[lab[A, List[nat]], List[Char]],
    term[lab[A, List[nat]], List[Char]])],
                                  m: nat,
                                  rm: Option[((lab[A, List[nat]], nat)) =>
       List[(term[lab[A, List[nat]], List[Char]],
              term[lab[A, List[nat]], List[Char]])]],
                                  rrm: Option[((lab[A, List[nat]], nat)) =>
        List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])]],
                                  x6: dPProof[A, List[nat], List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  (i, p, r, m, rm, rrm, x6) match {
  case (i, p, r, m, rm, rrm, PisEmpty) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('P', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y'),
                       p match {
                         case Nil => Inr[Unit, (List[Char]) => List[Char]](())
                         case _ :: _ =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](i, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'P', ' ',
                      'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p', 't',
                      'y', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 'f', 'o',
                      'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'D', 'P', '-',
                      'p', 'r', 'o', 'b', 'l', 'e', 'm', ' ')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, shows_dpp[lab[A, List[nat]], List[Char]](p, r)))))
                       })
  case (i, p, r, m, rm, rrm, RedPairProc(redp, rP, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o',
                             'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_redpair_proc[lab[A, List[nat]],
                                      List[Char]](get_redpair[lab[A, List[nat]],
                       List[Char]](redp),
           p, rP, r),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c',
              't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o',
              'c', 'e', 's', 's', 'o', 'r', ' ', 't', 'o', ' ', 's', 'w', 'i',
              't', 'c', 'h', ' ', 'f', 'r', 'o', 'm')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_dpp[lab[A, List[nat]], List[Char]](p, r),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl,
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('t', 'o', ' ', 't', 'h', 'e',
          ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl,
     comp[List[Char], List[Char],
           List[Char]](shows_dpp[lab[A, List[nat]], List[Char]](rP, r),
                        comp[List[Char], List[Char],
                              List[Char]](shows_nl, x))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      rP, r, m, rm, rrm, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ',
                   'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                   'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, p, r, m, rm, rrm, RedPairUrProc(redp, rP, ur, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o',
                             'c', ' ', '(', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                             'r', 'u', 'l', 'e', 's', ' ', '+', ' ', 'a', 'f',
                             ')'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_redpair_ur_af_proc[lab[A, List[nat]],
    List[Char]](get_redpair[lab[A, List[nat]], List[Char]](redp), p, rP, r, rm,
                 ur),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c',
              't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('p',
                                 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ',
                                 '(', 'w', 'i', 't', 'h', ' ', 'u', 's', 'a',
                                 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's',
                                 ' ', 'a', 'n', 'd', ' ', 'a', 'f', ')', ' ',
                                 't', 'o', ' ', 's', 'w', 'i', 't', 'c', 'h',
                                 ' ', 'f', 'r', 'o', 'm', ' ')),
         comp[List[Char], List[Char],
               List[Char]](shows_nl,
                            comp[List[Char], List[Char],
                                  List[Char]](shows_dpp[lab[A, List[nat]],
                 List[Char]](p, r),
       comp[List[Char], List[Char],
             List[Char]](shows_nl,
                          comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List('t', 'o',
                             ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r',
                             'o', 'b', 'l', 'e', 'm', ' ')),
     comp[List[Char], List[Char],
           List[Char]](shows_nl,
                        comp[List[Char], List[Char],
                              List[Char]](shows_dpp[lab[A, List[nat]],
             List[Char]](rP, r),
   comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      rP, r, m, rm, rrm, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ',
                   'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                   'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, p, r, m, rm, rrm, DepGraphProc(edpts)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('D', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r',
                             'o', 'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_dep_graph_proc[lab[A, List[nat]],
List[Char]](p, mapa[(Option[dPProof[A, List[nat], List[Char]]],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])]),
                     (Boolean,
                       List[(term[lab[A, List[nat]], List[Char]],
                              term[lab[A, List[nat]],
                                    List[Char]])])]((x:
               (Option[dPProof[A, List[nat], List[Char]]],
                 List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])]))
              => (fst[Option[dPProof[A, List[nat], List[Char]]],
                       List[(term[lab[A, List[nat]], List[Char]],
                              term[lab[A, List[nat]], List[Char]])]](x)
                    match {
                    case None => false
                    case Some(_) => true
                  },
                   snd[Option[dPProof[A, List[nat], List[Char]]],
                        List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])]](x)),
             edpts),
             rm, rrm),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'i', 'l', 'e', ' ', 't', 'r', 'y', 'i',
              'n', 'g', ' ', 't', 'o', ' ', 'p', 'e', 'r', 'f', 'o', 'r', 'm',
              ' ', 'S', 'C', 'C', '-', 'd', 'e', 'c', 'o', 'm', 'p', 'o', 's',
              'i', 't', 'i', 'o', 'n', ' ', ' ', 'o', 'n', ' ')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_dpp[lab[A, List[nat]], List[Char]](p, r),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl, x))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_allm_gen_index[(Option[dPProof[A, List[nat],
                   List[Char]]],
    List[(term[lab[A, List[nat]], List[Char]],
           term[lab[A, List[nat]], List[Char]])]),
   (List[Char]) =>
     List[Char]]((rp: (Option[dPProof[A, List[nat], List[Char]]],
                        List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])]))
                   => (j: nat) =>
                        fst[Option[dPProof[A, List[nat], List[Char]]],
                             List[(term[lab[A, List[nat]], List[Char]],
                                    term[lab[A, List[nat]], List[Char]])]](rp)
                          match {
                          case None => j
                          case Some(_) => Suc(j)
                        },
                  (rp: (Option[dPProof[A, List[nat], List[Char]]],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]], List[Char]])]))
                    => (j: nat) =>
                         fst[Option[dPProof[A, List[nat], List[Char]]],
                              List[(term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]])]](rp)
                           match {
                           case None =>
                             Inr[Unit, (List[Char]) => List[Char]](())
                           case Some(a) =>
                             checkDPProof[A](comp[List[Char], List[Char],
           List[Char]](i, comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List('.')),
     shows_prec_nat(Zero_nat, j))),
      snd[Option[dPProof[A, List[nat], List[Char]]],
           List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])]](rp),
      r, m, rm, rrm, a)
                         },
                  one_nat, edpts),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'd', 'e', 'p', 'e', 'n', 'd', 'e', 'n', 'c', 'y',
                   ' ', 'g', 'r', 'a', 'p', 'h', ' ', 'p', 'r', 'o', 'c', 'e',
                   's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, p, r, m, rm, rrm, MonoRedPairProc(redp, rP, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i',
                             'r', 'P', 'r', 'o', 'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_mono_redpair_proc[lab[A, List[nat]],
   List[Char]](get_redpair[lab[A, List[nat]], List[Char]](redp), p, rP, r, rR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'n', 'o', 't',
              'o', 'n', 'i', 'c', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
              'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e', 's',
              's', 'o', 'r', ' ')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('t',
                                 'o', ' ', 's', 'w', 'i', 't', 'c', 'h', ' ',
                                 'f', 'r', 'o', 'm', ' ')),
         comp[List[Char], List[Char],
               List[Char]](shows_nl,
                            comp[List[Char], List[Char],
                                  List[Char]](shows_dpp[lab[A, List[nat]],
                 List[Char]](p, r),
       comp[List[Char], List[Char],
             List[Char]](shows_nl,
                          comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List('t', 'o',
                             ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r',
                             'o', 'b', 'l', 'e', 'm', ' ')),
     comp[List[Char], List[Char],
           List[Char]](shows_nl,
                        comp[List[Char], List[Char],
                              List[Char]](shows_dpp[lab[A, List[nat]],
             List[Char]](rP, rR),
   comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      rP, rR, m,
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](rR),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](rR)),
                                      prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'i', 'c', ' ',
                   'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a',
                   'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, p, r, m, rm, rrm, MonoRedPairUrProc(redp, rP, rR, ur, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i',
                             'r', 'P', 'r', 'o', 'c', ' ', '(', 'u', 's', 'a',
                             'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's', ')'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_mono_redpair_ur_proc[lab[A, List[nat]],
      List[Char]](get_redpair[lab[A, List[nat]], List[Char]](redp), p, rP, r,
                   rm, rR, ur),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'n', 'o', 't',
              'o', 'n', 'i', 'c', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
              'n', ' ', 'p', 'a', 'i', 'r', ' ')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('p',
                                 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ',
                                 '(', 'w', 'i', 't', 'h', ' ', 'u', 's', 'a',
                                 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's',
                                 ')', ' ', 't', 'o', ' ', 's', 'w', 'i', 't',
                                 'c', 'h', ' ', 'f', 'r', 'o', 'm', ' ')),
         comp[List[Char], List[Char],
               List[Char]](shows_nl,
                            comp[List[Char], List[Char],
                                  List[Char]](shows_dpp[lab[A, List[nat]],
                 List[Char]](p, r),
       comp[List[Char], List[Char],
             List[Char]](shows_nl,
                          comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List('t', 'o',
                             ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r',
                             'o', 'b', 'l', 'e', 'm')),
     comp[List[Char], List[Char],
           List[Char]](shows_nl,
                        comp[List[Char], List[Char],
                              List[Char]](shows_dpp[lab[A, List[nat]],
             List[Char]](rP, rR),
   comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      rP, rR, m,
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](rR),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](rR)),
                                      prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'i', 'c', ' ',
                   'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a',
                   'i', 'r', ' ')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List('(', 'w', 'i', 't', 'h', ' ', 'u', 's',
                                      'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l',
                                      'e', 's', ')', ' ', 'p', 'r', 'o', 'c',
                                      'e', 's', 's', 'o', 'r')),
              comp[List[Char], List[Char],
                    List[Char]](shows_nl,
                                 (a: List[Char]) => indent(x, a)))))))))
  case (i, p, r, Suc(m), rm, rrm, UnlabProc(ulP, ulR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('U', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_unlab_proc[A, List[nat], List[Char]](p, r, ulP, ulR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'l', 'a', 'b',
              'e', 'l', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
              'o', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      ulP, ulR, m,
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](ulR),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](ulR)),
                                      prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g',
                   ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, p, r, Zero_nat, rm, rrm, UnlabProc(ulP, ulR, prf)) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](i, comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r', ' ',
                                       'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p',
                                       'l', 'y', 'i', 'n', 'g', ' ', 't', 'h',
                                       'e', ' ', 'u', 'n', 'l', 'a', 'b', 'e',
                                       'l', 'i', 'n', 'g', ' ', 'p', 'r', 'o',
                                       'c', 'e', 's', 's', 'o', 'r')),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n',
                                     'g', ' ', 'n', 'o', 't', ' ', 'a', 'l',
                                     'l', 'o', 'w', 'e', 'd', ' ', 'd', 'u',
                                     'e', ' ', 't', 'o', ' ', 'l', 'a', 'b',
                                     'e', 'l', '-', 'd', 'e', 'p', 't', 'h',
                                     ' ', 'o', 'f', ' ', '0')),
             shows_nl)))))
  case (i, p, r, m, rm, rrm, SemlabProc(sli, l, lP, lR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('S', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](create_sem_lab_proc[A, List[Char]](sli, l, rm, p, r, lP, lR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 's', 'e', 'm', 'a', 'n',
              't', 'i', 'c', ' ', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ',
              'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      lP, lR, Suc(m),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](lR),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](lR)),
                                      prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i', 'c', ' ', 'l',
                   'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c',
                   'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, p, r, m, rm, rrm, RootlabProc(delt_opt, lP, lR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('R', 'o', 'o', 't', 'l', 'a', 'b', 'P', 'r', 'o',
                             'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](rl_proc[A, List[nat]](delt_opt, p, r, lP, lR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r', 'o', 'o', 't', ' ',
              'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c',
              'e', 's', 's', 'o', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      lP, lR, Suc(m),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](lR),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](lR)),
                                      prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e', 'l',
                   'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                   'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, pa, r, m, rm, rrm, SubtermCriterionProc(p, rseq, rP, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('S', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o',
                             'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](create_subterm_criterion_proc[lab[A, List[nat]],
         List[Char]](p, rseq, pa, rP, r, rm),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 's', 'u', 'b', 't', 'e',
              'r', 'm', ' ', 'c', 'r', 'i', 't', 'e', 'r', 'i', 'o', 'n', ' ',
              'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      rP, r, m, rm, rrm, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ', 'c', 'r',
                   'i', 't', 'e', 'r', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'c',
                   'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, p, r, m, rm, rrm, SizeChangeSubTerm(graphs)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e',
                             'S', 'u', 'b', 'T', 'e', 'r', 'm'),
                       check_sct_subterm_proc[lab[A, List[nat]],
       List[Char]](p, graphs, rm, rrm))
  case (i, p, r, m, rm, rrm, SizeChangeRedPair(redp, ur, graphs)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e',
                             'R', 'e', 'd', 'P', 'a', 'i', 'r'),
                       check_sct_ur_af_proc[lab[A, List[nat]],
     List[Char]](get_redpair[lab[A, List[nat]], List[Char]](redp), p, graphs,
                  rm, rrm, ur))
  case (i, p, r, m, rm, rrm, FccProc(f, fcs, rP, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('F', 'c', 'c', 'P', 'r', 'o', 'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](fcc_proc[A, List[nat]](m, p, r, f, fcs, rP, rR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'c', 'l', 'o', 's', 'u',
              'r', 'e', ' ', 'u', 'n', 'd', 'e', 'r', ' ', 'f', 'l', 'a', 't',
              ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', 's')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      rP, rR, m,
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](rR),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](rR)),
                                      prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 'f', 'l',
                   'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', ' ', 'c',
                   'l', 'o', 's', 'u', 'r', 'e', ' ', 'p', 'r', 'o', 'c', 'e',
                   's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, pa, ra, m, rm, rrm, UncurryProc(mode, u_info, p, r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('U', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o',
                             'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]]((uncurry_proc_both[A, List[nat]](mode,
            u_info)).apply(pa).apply(ra).apply(p).apply(r).apply(rm),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c', 'u', 'r',
              'r', 'y', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
              'o', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkDPProof[A](comp[List[Char], List[Char],
   List[Char]](i, shows_string.apply(List('.', '1'))),
                                      p, r, Zero_nat,
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](r),
                                      computeRuleMap[lab[A, List[nat]],
              List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](r)),
                                      prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g',
                   ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
}

def bounds_condition[A: show,
                      B: equal: show](x0: boundstype,
                                       uu: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (x0, uu) match {
  case (Roof, uu) => Inr[Unit, (List[Char]) => List[Char]](())
  case (Match, r) =>
    catcha[(term[A, B], term[A, B]), Unit,
            (List[Char]) =>
              List[Char]](catcha[((term[A, B], term[A, B]),
                                   (term[A, B], term[A, B])),
                                  Unit,
                                  (term[A, B],
                                    term[A,
  B])](forallm[(term[A, B], term[A, B]),
                (term[A, B],
                  term[A, B])]((x: (term[A, B], term[A, B])) =>
                                 (if ({
val (_, a): (term[A, B], term[A, B]) = x;
linear_term[A, B](a)
                                      })
                                   Inr[Unit, (term[A, B], term[A, B])](())
                                   else Inl[(term[A, B], term[A, B]), Unit](x)),
                                r),
        (x: ((term[A, B], term[A, B]), (term[A, B], term[A, B]))) =>
          Inl[(term[A, B], term[A, B]),
               Unit](snd[(term[A, B], term[A, B]),
                          (term[A, B], term[A, B])](x))),
                           (x: (term[A, B], term[A, B])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit]({
  val (l, ra): (term[A, B], term[A, B]) = x;
  comp[List[Char], List[Char],
        List[Char]](shows_prec_list[Char](Zero_nat,
   List('r', 'u', 'l', 'e', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_rule[A, B]((l, ra)),
shows_prec_list[Char](Zero_nat,
                       List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'r', 'i',
                             'g', 'h', 't', '-', 'l', 'i', 'n', 'e', 'a', 'r',
                             ' '))))
}))
}

def ta_rules_impl[A, B, C](x0: ta_impl_ext[A, B, C]):
      rbt[(B, nat), List[ta_rule_impl[A, B]]] =
  x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, more)
    => ta_rules_impl
}

def ta_final_impl[A, B, C](x0: ta_impl_ext[A, B, C]): rbt[A, Unit] = x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, more)
    => ta_final_impl
}

def ta_contains_impl[A: equal: linorder,
                      B: equal: linorder](f: List[(A, nat)],
   ta: ta_impl_ext[B, A, Unit], qs: List[B]):
      sum[sum[B, term[A, B]], Unit] =
  {
    val fin: rbt[B, Unit] = ta_final_impl[B, A, Unit](ta)
    val look: ((A, nat)) => List[ta_rule_impl[B, A]] =
      rm_set_lookup[(A, nat),
                     ta_rule_impl[B, A]](ta_rules_impl[B, A, Unit](ta));
    ta_eps_impl[B, A, Unit](ta);
    binda[sum[B, term[A, B]], Unit,
           Unit](catcha[B, Unit,
                         sum[B, term[A, B]]](catcha[(B, B), Unit,
             B](forallm[B, B]((q: B) =>
                                check[B](rs_memb[B].apply(q).apply(fin), q),
                               qs),
                 (x: (B, B)) => Inl[B, Unit](snd[B, B](x))),
      (x: B) => Inl[sum[B, term[A, B]], Unit](Inl[B, term[A, B]](x))),
                  (_: Unit) =>
                    catcha[term[A, B], Unit,
                            sum[B, term[A,
 B]]](catcha[((A, nat), term[A, B]), Unit,
              term[A, B]](forallm[(A, nat),
                                   term[A,
 B]]((a: (A, nat)) =>
       {
         val (fa, n): (A, nat) = a
         val rules: List[ta_rule_impl[B, A]] = look((fa, n));
         catcha[List[B], Unit,
                 term[A, B]](catcha[(List[B], List[B]), Unit,
                                     List[B]](forallm[List[B],
               List[B]]((x: List[B]) =>
                          (if (list_ex[ta_rule_impl[B,
             A]]((rule: ta_rule_impl[B, A]) =>
                   equal_lista[B](rqs_impl[B, A](rule), x) &&
                     {
                       val qq: rbt[B, Unit] = rq_impl[B, A](rule);
                       list_ex[B]((q: B) => rs_memb[B].apply(q).apply(qq), qs)
                     },
                  rules))
                            Inr[Unit, List[B]](()) else Inl[List[B], Unit](x)),
                         concat_lists[B](replicate[List[B]](n, qs))),
       (x: (List[B], List[B])) => Inl[List[B], Unit](snd[List[B], List[B]](x))),
                              (x: List[B]) =>
                                Inl[term[A, B],
                                     Unit](Fun[A,
        B](fa, mapa[B, term[A, B]]((aa: B) => Var[B, A](aa), x))))
       },
      f),
                           (x: ((A, nat), term[A, B])) =>
                             Inl[term[A, B],
                                  Unit](snd[(A, nat), term[A, B]](x))),
       (x: term[A, B]) => Inl[sum[B, term[A, B]], Unit](Inr[term[A, B], B](x))))
  }

def ta_rule_compatible_list[A: linorder, B: equal: linorder,
                             C: equal](ta: ta_impl_ext[A, B, Unit]):
      (((term[B, C], term[B, C]), List[C])) =>
        sum[((term[B, A], term[B, A]), A), Unit] =
  {
    val rm: rbt[(B, nat), List[ta_rule_impl[A, B]]] =
      ta_rules_impl[A, B, Unit](ta)
    val eps: A => rbt[A, Unit] = ta_eps_impl[A, B, Unit](ta);
    (a: ((term[B, C], term[B, C]), List[C])) =>
      {
        val (aa, b): ((term[B, C], term[B, C]), List[C]) = a;
        ({
           val (l, r): (term[B, C], term[B, C]) = aa;
           (xs: List[C]) =>
             catcha[(List[(C, A)], ((term[B, A], term[B, A]), A)), Unit,
                     ((term[B, A], term[B, A]),
                       A)](forallm[List[(C, A)],
                                    ((term[B, A], term[B, A]),
                                      A)]((vec: List[(C, A)]) =>
    {
      val tau: C => A = (ab: C) => assign_from_vector[C, A](vec, ab)
      val l_tau: term[B, A] = map_vars[C, A, B](tau, l)
      val r_tau: term[B, A] = map_vars[C, A, B](tau, r)
      val rhs: A => Boolean =
        rs_alpha[A].apply(ta_res_list[B, A](rm, eps, r_tau));
      catcha[(A, ((term[B, A], term[B, A]), A)), Unit,
              ((term[B, A], term[B, A]),
                A)](forallm[A, ((term[B, A], term[B, A]),
                                 A)]((q: A) =>
                                       check[((term[B, A], term[B, A]),
       A)](member[A](q, rhs), ((l_tau, r_tau), q)),
                                      rs_to_list[A].apply(ta_res_list[B,
                               A](rm, eps, l_tau))),
                     (x: (A, ((term[B, A], term[B, A]), A))) =>
                       Inl[((term[B, A], term[B, A]), A),
                            Unit](snd[A, ((term[B, A], term[B, A]), A)](x)))
    },
   enum_vectors[A, C](ta_rqs_impl[A, B, Unit](ta), xs)),
                            (x: (List[(C, A)], ((term[B, A], term[B, A]), A)))
                              => Inl[((term[B, A], term[B, A]), A),
                                      Unit](snd[List[(C, A)],
         ((term[B, A], term[B, A]), A)](x)))
         })(b)
      }
  }

def ta_trs_compatible_list[A: linorder, B: equal: linorder,
                            C: equal](ta: ta_impl_ext[A, B, Unit],
                                       r:
 List[((term[B, C], term[B, C]), List[C])]):
      sum[((term[B, C], term[B, C]), ((term[B, A], term[B, A]), A)), Unit] =
  {
    val check:
          (((term[B, C], term[B, C]), List[C])) =>
            sum[((term[B, A], term[B, A]), A), Unit]
      = ta_rule_compatible_list[A, B, C](ta);
    catcha[(((term[B, C], term[B, C]), List[C]),
             ((term[B, C], term[B, C]), ((term[B, A], term[B, A]), A))),
            Unit,
            ((term[B, C], term[B, C]),
              ((term[B, A], term[B, A]),
                A))](forallm[((term[B, C], term[B, C]), List[C]),
                              ((term[B, C], term[B, C]),
                                ((term[B, A], term[B, A]),
                                  A))]((lr: ((term[B, C], term[B, C]), List[C]))
 => catcha[((term[B, A], term[B, A]), A), Unit,
            ((term[B, C], term[B, C]),
              ((term[B, A], term[B, A]),
                A))](check(lr),
                      (x: ((term[B, A], term[B, A]), A)) =>
                        Inl[((term[B, C], term[B, C]),
                              ((term[B, A], term[B, A]), A)),
                             Unit]((fst[(term[B, C], term[B, C]), List[C]](lr),
                                     x))),
r),
                      (x: (((term[B, C], term[B, C]), List[C]),
                            ((term[B, C], term[B, C]),
                              ((term[B, A], term[B, A]), A))))
                        => Inl[((term[B, C], term[B, C]),
                                 ((term[B, A], term[B, A]), A)),
                                Unit](snd[((term[B, C], term[B, C]), List[C]),
   ((term[B, C], term[B, C]), ((term[B, A], term[B, A]), A))](x)))
  }

def flatten_term_enum_filter[A, B](f: (term[A, B]) => Boolean,
                                    x1: term[List[A], B]):
      List[term[A, B]] =
  (f, x1) match {
  case (f, Var(x)) =>
    {
      val tx: term[A, B] = Var[B, A](x);
      (if (f(tx)) List(tx) else Nil)
    }
  case (f, Fun(fs, ts)) =>
    {
      val lts: List[List[term[A, B]]] =
        mapa[term[List[A], B],
              List[term[A, B]]]((a: term[List[A], B]) =>
                                  flatten_term_enum_filter[A, B](f, a),
                                 ts);
      (if (list_ex[List[term[A, B]]]((a: List[term[A, B]]) =>
                                       nulla[term[A, B]](a),
                                      lts))
        Nil else {
                   val ss: List[List[term[A, B]]] =
                     concat_lists[term[A, B]](lts);
                   filter[term[A, B]](f, maps[A,
       term[A, B]]((fa: A) =>
                     mapa[List[term[A, B]],
                           term[A, B]]((a: List[term[A, B]]) =>
 Fun[A, B](fa, a),
ss),
                    fs))
                 })
    }
}

def inverse_base_term_filter[A, B](filt: (term[(A, nat), B]) => Boolean,
                                    l: term[A, B], c: nat):
      List[term[(A, nat), B]] =
  {
    val hs: List[nat] = upt(Zero_nat, Suc(c));
    flatten_term_enum_filter[(A, nat),
                              B](filt,
                                  map_term[A, List[(A, nat)],
    B]((f: A) => mapa[nat, (A, nat)]((h: nat) => lift[A](h, f), hs), l))
  }

def cover_bound_list_filter[A, B: equal](filt: (term[(A, nat), B]) => Boolean,
  ff: ((term[A, B], term[A, B])) => (term[A, B]) => Boolean, c: nat,
  r: List[(term[A, B], term[A, B])]):
      List[((term[(A, nat), B], term[(A, nat), B]), List[B])] =
  maps[(term[A, B], term[A, B]),
        ((term[(A, nat), B], term[(A, nat), B]),
          List[B])]((a: (term[A, B], term[A, B])) =>
                      {
                        val (l, ra): (term[A, B], term[A, B]) = a
                        val ee: (term[A, B]) => Boolean = ff((l, ra))
                        val xs: List[B] = vars_term_list[A, B].apply(l);
                        mapa[term[(A, nat), B],
                              ((term[(A, nat), B], term[(A, nat), B]),
                                List[B])]((la: term[(A, nat), B]) =>
    ((la, map_term[A, (A, nat),
                    B]((aa: A) =>
                         lift[A](Suc(min_list[nat](mapa[(A, nat),
                 nat]((ab: (A, nat)) => height[A](ab),
                       sym_collect[(A, nat),
                                    B]((t: term[(A, nat), B]) =>
 ee(map_term[(A, nat), A, B]((ab: (A, nat)) => base[A](ab), t)),
la)))),
                                  aa),
                        ra)),
      xs),
   inverse_base_term_filter[A, B](filt, l, c))
                      },
                     r)

def ta_res_list_all[A: linorder, B: equal: linorder,
                     C](q: rbt[A, Unit],
                         ta: rbt[(B, nat), List[ta_rule_impl[A, B]]],
                         x2: term[B, C]):
      rbt[A, Unit] =
  (q, ta, x2) match {
  case (q, ta, Var(uu)) => q
  case (q, ta, Fun(f, ts)) =>
    {
      val rec: List[rbt[A, Unit]] =
        mapa[term[B, C],
              rbt[A, Unit]]((a: term[B, C]) =>
                              ta_res_list_all[A, B, C](q, ta, a),
                             ts)
      val n: nat = size_list[term[B, C]](ts)
      val rules: List[ta_rule_impl[A, B]] =
        (rm_set_lookup[(B, nat), ta_rule_impl[A, B]](ta)).apply((f, n))
      val arules: List[ta_rule_impl[A, B]] =
        filter[ta_rule_impl[A, B]]((rule: ta_rule_impl[A, B]) =>
                                     {
                                       val qs: List[A] = rqs_impl[A, B](rule);
                                       all_interval_nat((i: nat) =>
                  rs_memb[A].apply(nth[A](qs,
   i)).apply(nth[rbt[A, Unit]](rec, i)),
                 Zero_nat, n)
                                     },
                                    rules)
      val a: List[rbt[A, Unit]] =
        mapa[ta_rule_impl[A, B],
              rbt[A, Unit]]((a: ta_rule_impl[A, B]) => rq_impl[A, B](a),
                             arules);
      foldl[rbt[A, Unit], rbt[A, Unit]](rs_union[A], rs_empty[A], a)
    }
}

def ta_rule_compatible_heuristic[A: equal: linorder, B: equal: linorder,
                                  C](ta: ta_impl_ext[A, B, Unit],
                                      l: term[B, C]):
      Boolean =
  rs_isEmpty[A].apply(ta_res_list_all[A, B,
                                       C](ta_rqs_set[A, B, Unit](ta),
   ta_rules_impl[A, B, Unit](ta), l))

def boundstype_fun[A, B: equal](x0: boundstype):
      ((term[A, B], term[A, B])) => (term[A, B]) => Boolean =
  x0 match {
  case Roof => (a: (term[A, B], term[A, B])) => roof[A, B](a)
  case Match => matcha[A, B]
}

def check_bounds[A: equal: linorder: show,
                  B: equal: linorder: show](x0: bounds_info[A, B],
     r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (x0, r) match {
  case (Bounds_Info(typea, c, qfin, preTA), r) =>
    {
      val ta: ta_impl_ext[B, (A, nat), Unit] = generate_ta[B, (A, nat)](preTA);
      binda[(List[Char]) => List[Char], Unit,
             Unit](check_wf_trs[A, B](r),
                    (_: Unit) =>
                      binda[(List[Char]) => List[Char], Unit,
                             Unit](check_left_linear_trs[A, B](r),
                                    (_: Unit) =>
                                      binda[(List[Char]) => List[Char], Unit,
     Unit](bounds_condition[A, B](typea, r),
            (_: Unit) =>
              binda[(List[Char]) => List[Char], Unit,
                     Unit](check[(List[Char]) =>
                                   List[Char]](list_ex[B]((q: B) =>
                    rs_memb[B].apply(q).apply(ta_rqs_set[B, (A, nat),
                  Unit](ta)),
                   qfin),
        shows_prec_list[Char](Zero_nat,
                               List('d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f',
                                     'i', 'n', 'd', ' ', 'm', 'e', 'n', 't',
                                     'i', 'o', 'n', 'e', 'd', ' ', 'f', 'i',
                                     'n', 'a', 'l', ' ', 's', 't', 'a', 't',
                                     'e', ' ', 'i', 'n', ' ', 'T', 'A'))),
                            (_: Unit) =>
                              binda[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[((A, nat), nat), Unit,
           (List[Char]) =>
             List[Char]](catcha[(((A, nat), nat), ((A, nat), nat)), Unit,
                                 ((A, nat),
                                   nat)](forallm[((A, nat), nat),
          ((A, nat),
            nat)]((x: ((A, nat), nat)) =>
                    (if ({
                           val (f, _): ((A, nat), nat) = x;
                           less_eq_nat(height[A](f), c)
                         })
                      Inr[Unit, ((A, nat), nat)](())
                      else Inl[((A, nat), nat), Unit](x)),
                   mapa[(((A, nat), nat), List[ta_rule_impl[B, (A, nat)]]),
                         ((A, nat),
                           nat)]((a: (((A, nat), nat),
                                       List[ta_rule_impl[B, (A, nat)]]))
                                   => fst[((A, nat), nat),
   List[ta_rule_impl[B, (A, nat)]]](a),
                                  rm_to_list[((A, nat), nat),
      List[ta_rule_impl[B, (A, nat)]]].apply(ta_rules_impl[B, (A, nat),
                    Unit](ta)))),
  (x: (((A, nat), nat), ((A, nat), nat))) =>
    Inl[((A, nat), nat), Unit](snd[((A, nat), nat), ((A, nat), nat)](x))),
                          (x: ((A, nat), nat)) =>
                            Inl[(List[Char]) => List[Char],
                                 Unit]({
 val (f, _): ((A, nat), nat) = x;
 comp[List[Char], List[Char],
       List[Char]](shows_prec_prod[A, nat](Zero_nat, f),
                    comp[List[Char], List[Char],
                          List[Char]](shows_prec_list[Char](Zero_nat,
                     List(' ', 'i', 's', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ',
                           'i', 'n', ' ', 'T', 'A', ' ', 'w', 'i', 't', 'h',
                           ' ', 'h', 'e', 'i', 'g', 'h', 't', ' ', 'l', 'a',
                           'r', 'g', 'e', 'r', ' ', 't', 'h', 'a', 'n', ' ',
                           'c', ' ', '=', ' ')),
                                       shows_prec_nat(Zero_nat, c)))
                                       })),
    (_: Unit) =>
      binda[(List[Char]) => List[Char], Unit,
             Unit](catcha[sum[B, term[(A, nat), B]], Unit,
                           (List[Char]) =>
                             List[Char]](ta_contains_impl[(A, nat),
                   B](mapa[(A, nat),
                            ((A, nat),
                              nat)]((a: (A, nat)) =>
                                      {
val (f, aa): (A, nat) = a;
(lift[A](Zero_nat, f), aa)
                                      },
                                     funas_trs_list[A, B](r)),
                       ta, qfin),
  (x: sum[B, term[(A, nat), B]]) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('i', 't', ' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                      ' ', 'b', 'e', ' ', 'g', 'u', 'a', 'r', 'a', 'n', 't',
                      'e', 'e', 'd', ' ', 't', 'h', 'a', 't', ' ', 'l', 'i',
                      'f', 't', '0', '(', 'T', '(', 'S', 'i', 'g', 'm', 'a',
                      ')', ')', ' ', 'i', 's', ' ', 'a', 'c', 'c', 'e', 'p',
                      't', 'e', 'd', ' ', 'b', 'y', ' ', 'T', 'A')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl,
             x match {
               case Inl(q) =>
                 comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('s', 't', 'a', 't', 'e', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, q, a),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 'i', 's', ' ', 'n', 'o', 't',
    ' ', 'f', 'i', 'n', 'a', 'l'))))
               case Inr(fqs) =>
                 comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o',
                        ' ', 't', 'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n',
                        ' ', 'f', 'r', 'o', 'm', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_term[(A, nat), B](Zero_nat, fqs),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 't', 'o', ' ', 'a', ' ', 'f',
    'i', 'n', 'a', 'l', ' ', 's', 't', 'a', 't', 'e'))))
             })))),
                    (_: Unit) =>
                      catcha[((term[(A, nat), B], term[(A, nat), B]),
                               ((term[(A, nat), B], term[(A, nat), B]), B)),
                              Unit,
                              (List[Char]) =>
                                List[Char]](ta_trs_compatible_list[B, (A, nat),
                            B](ta, cover_bound_list_filter[A,
                    B]((l: term[(A, nat), B]) =>
                         ! (ta_rule_compatible_heuristic[B, (A, nat),
                  B](ta, l)),
                        boundstype_fun[A, B](typea), c, r)),
     (x: ((term[(A, nat), B], term[(A, nat), B]),
           ((term[(A, nat), B], term[(A, nat), B]), B)))
       => Inl[(List[Char]) => List[Char],
               Unit]({
                       val (lr, (lrq, q)):
                             ((term[(A, nat), B], term[(A, nat), B]),
                               ((term[(A, nat), B], term[(A, nat), B]), B))
                         = x;
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_list[Char](Zero_nat,
                        List('T', 'A', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                              'c', 'o', 'm', 'p', 'a', 't', 'i', 'b', 'l', 'e',
                              ' ', 'w', 'i', 't', 'h', ' ', 'R')),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     comp[List[Char], List[Char],
                           List[Char]](shows_prec_list[Char](Zero_nat,
                      List('f', 'o', 'r', ' ', 'r', 'u', 'l', 'e', ' ')),
comp[List[Char], List[Char],
      List[Char]](shows_rule[(A, nat), B](lr),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_list[Char](Zero_nat,
                                       List('w', 'h', 'i', 'c', 'h', ' ', 'i',
     's', ' ', 'i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'e', 'd', ' ',
     'b', 'y', ' ', 's', 't', 'a', 't', 'e', 's', ' ', 't', 'o', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_rule[(A, nat), B](lrq),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl,
               comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('t', 'h', 'e', ' ', 's', 't', 'a', 't', 'e', ' ')),
                                  comp[List[Char], List[Char],
List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, q, a),
             comp[List[Char], List[Char],
                   List[Char]](shows_prec_list[Char](Zero_nat,
              List(' ', 'i', 's', ' ', 'o', 'n', 'l', 'y', ' ', 'r', 'e', 'a',
                    'c', 'h', 'a', 'b', 'l', 'e', ' ', 'f', 'r', 'o', 'm', ' ',
                    't', 'h', 'e', ' ', 'l', 'h', 's')),
                                shows_nl)))))))))))
                     }))))))))
    }
}

implicit def linorder_char: linorder[Char] = new linorder[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

def check_rule_removal_proc[A: equal: show,
                             B: equal: show](rp:
       (sum[(List[Char]) => List[Char], Unit],
         (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
           (((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit],
             (A => nat => nat => Boolean,
               (sum[(List[Char]) => List[Char], Unit],
                 (List[Char]) => List[Char]))))),
      r: List[(term[A, B], term[A, B])], rR: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (_, nSr):
                                (List[(term[A, B], term[A, B])],
                                  List[(term[A, B], term[A, B])])
                            = partition[(term[A, B],
  term[A, B])]((foo: (term[A, B], term[A, B])) =>
                 isOK[(List[Char]) => List[Char],
                       Unit]((rp_strict[A, B](rp)).apply(foo)),
                r);
                          binda[(List[Char]) => List[Char], Unit,
                                 Unit](rp_valid[A, B](rp),
(_: Unit) =>
  binda[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](rp_mono[A, B](rp),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
              'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 's', 't',
              'r', 'i', 'c', 't', ' ', 'o', 'r', 'd', 'e', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  binda[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(term[A, B], term[A, B]), Unit,
                                       (List[Char]) =>
 List[Char]](check_subseteq[(term[A, B], term[A, B])](rR, r),
              (x: (term[A, B], term[A, B])) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('n', 'o',
                              't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd', ' ',
                              't', 'o', ' ', 'i', 'n', 't', 'r', 'o', 'd', 'u',
                              'c', 'e', ' ', 'n', 'e', 'w', ' ', 'r', 'u', 'l',
                              'e', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_nl, shows_rule[A, B](x))))),
                                (_: Unit) =>
                                  binda[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[A, B], term[A, B]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[A, B],
                     term[A, B]),
                    (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSr),
            (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[A, B], term[A, B]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'r', 'u', 'l', 'e', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          catcha[(term[A, B], term[A, B]), Unit,
                  (List[Char]) =>
                    List[Char]](check_subseteq[(term[A, B],
         term[A, B])](nSr, rR),
                                 (x: (term[A, B], term[A, B])) =>
                                   Inl[(List[Char]) => List[Char],
Unit]((rp_strict[A, B](rp)).apply(x) match {
        case Inl(err) => err
        case Inr(_) => internal_error
      }))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e',
 ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](rp_desc[A, B](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))

def unlab_trs_tt[A: equal: show, B: equal: show,
                  C: equal: show](r: List[(term[lab[A, B], C],
    term[lab[A, B], C])],
                                   rR: List[(term[lab[A, B], C],
      term[lab[A, B], C])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val u: List[(term[lab[A, B], C], term[lab[A, B], C])] =
      unlab_of_trs_list[A, B, C](r);
    catcha[(term[lab[A, B], C], term[lab[A, B], C]), Unit,
            (List[Char]) =>
              List[Char]](binda[(term[lab[A, B], C], term[lab[A, B], C]), Unit,
                                 Unit](check_subseteq[(term[lab[A, B], C],
                term[lab[A, B], C])](rR, u),
(_: Unit) => check_subseteq[(term[lab[A, B], C], term[lab[A, B], C])](u, rR)),
                           (_: (term[lab[A, B], C], term[lab[A, B], C])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_rules[lab[A, B], C](rR),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
 't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ',
 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_rules[lab[A, B], C](u), shows_nl)))))))
  }

def create_sem_lab_tt[A: equal: show,
                       B: equal: show](sli: sl_inter[A, List[nat]],
x1: sum[(lab[A, List[nat]]) =>
          (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]],
         B],
r: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])],
lr: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (sli, x1, r, lr) match {
  case (sli, Inl(l), r, lr) =>
    {
      val c: nat = get_largest_element[A, List[nat]](sli)
      val ca: List[nat] = upt(Zero_nat, Suc(c))
      val i: (lab[A, List[nat]]) => (List[nat]) => nat =
        (a: lab[A, List[nat]]) =>
          (b: List[nat]) => sl_inter_to_inter[A, List[nat]](sli, a, b);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_sem_lab_tt[A, List[nat],
      B](model_check_valid[(List[Char]) => List[Char]],
          model_check_decr[List[(term[lab[A, List[nat]], B],
                                  term[lab[A, List[nat]], B])],
                            (List[Char]) => List[Char]],
          (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
            (b: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => check_sl_lab_trs[A, List[nat], nat,
                                   B](i, l, ca, model_cge[nat], a, b),
          (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
            (b: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => check_sl_model_lab_trs[A, List[nat], nat,
 B](i, l, ca, model_cge[nat], a, b),
          model_lge[lab[A, List[nat]], nat,
                     sum[List[nat], List[lab[A, List[nat]]]]],
          model_LS[lab[A, List[nat]], nat,
                    sum[List[nat], List[lab[A, List[nat]]]]],
          r, lr),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_prec_sl_inter[A, List[nat]](Zero_nat, sli),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
    }
  case (sli, Inr(v), r, lr) =>
    {
      val c: nat = get_largest_element[A, List[nat]](sli)
      val ca: List[nat] = upt(Zero_nat, Suc(c))
      val sig: List[(lab[A, List[nat]], nat)] =
        funas_trs_list[lab[A, List[nat]], B](r)
      val i: (lab[A, List[nat]]) => (List[nat]) => nat =
        (a: lab[A, List[nat]]) =>
          (b: List[nat]) => sl_inter_to_inter[A, List[nat]](sli, a, b)
      val l: (lab[A, List[nat]]) =>
               (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]]
        = qmodel_L[A, List[nat], nat, List[lab[A, List[nat]]]](sig);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_sem_lab_tt[A, List[nat],
      B](qmodel_check_valid[A, List[nat]](sli),
          qmodel_check_decr[A, B](sig, v, c),
          (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
            (b: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => check_sl_lab_trs[A, List[nat], nat,
                                   B](i, l, ca, qmodel_cge, a, b),
          (a: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
            (b: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])
              => check_sl_model_lab_trs[A, List[nat], nat,
 B](i, l, ca, qmodel_cge, a, b),
          (a: lab[A, List[nat]]) =>
            (b: nat) =>
              qmodel_lge[lab[A, List[nat]], nat, List[lab[A, List[nat]]],
                          List[lab[A, List[nat]]]](a, b),
          qmodel_LS[lab[A, List[nat]], nat, List[lab[A, List[nat]]]](sig, ca),
          r, lr),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_prec_sl_inter[A, List[nat]](Zero_nat, sli),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
    }
}

def unary_term[A, B](x0: term[A, B]): Boolean = x0 match {
  case Var(x) => true
  case Fun(f, List(t)) => unary_term[A, B](t)
  case Fun(v, Nil) => false
  case Fun(v, vb :: vd :: ve) => false
}

def check_unary_signature[A, B](r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(term[A, B], term[A, B]), Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[A, B], term[A, B]),
                                 (term[A, B], term[A, B])),
                                Unit,
                                (term[A, B],
                                  term[A,
B])](forallm[(term[A, B], term[A, B]),
              (term[A, B],
                term[A, B])]((x: (term[A, B], term[A, B])) =>
                               (if ({
                                      val (l, ra): (term[A, B], term[A, B]) = x;
                                      unary_term[A, B](l) &&
unary_term[A, B](ra)
                                    })
                                 Inr[Unit, (term[A, B], term[A, B])](())
                                 else Inl[(term[A, B], term[A, B]), Unit](x)),
                              r),
      (x: ((term[A, B], term[A, B]), (term[A, B], term[A, B]))) =>
        Inl[(term[A, B], term[A, B]),
             Unit](snd[(term[A, B], term[A, B]), (term[A, B], term[A, B])](x))),
                         (_: (term[A, B], term[A, B])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](shows_string.apply(List('t', 'h', 'e',
                       ' ', 't', 'r', 's', ' ', 'd', 'o', 'e', 's', ' ', 'n',
                       'o', 't', ' ', 'h', 'a', 'v', 'e', ' ', 'a', ' ', 'u',
                       'n', 'a', 'r', 'y', ' ', 's', 'i', 'g', 'n', 'a', 't',
                       'u', 'r', 'e'))))

def check_string_rev[A: equal: show,
                      B: equal: show](r: List[(term[A, B], term[A, B])],
                                       rR: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val rev: List[(term[A, B], term[A, B])] =
      mapa[(term[A, B], term[A, B]),
            (term[A, B],
              term[A, B])]((a: (term[A, B], term[A, B])) => rev_rule[A, B](a),
                            r);
    binda[(List[Char]) => List[Char], Unit,
           Unit](check_unary_signature[A, B](r),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](catcha[(term[A, B], term[A, B]), Unit,
 (List[Char]) =>
   List[Char]](check_subseteq[(term[A, B], term[A, B])](rev, rR),
                (x: (term[A, B], term[A, B])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List('t', 'h',
                                'e', ' ', 'r', 'u', 'l', 'e', ' ')),
        comp[List[Char], List[Char],
              List[Char]](shows_rule[A, B](x),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List(' ', 'i',
                              's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')),
      shows_nl))))),
                                  (_: Unit) =>
                                    catcha[(term[A, B], term[A, B]), Unit,
    (List[Char]) =>
      List[Char]](check_subseteq[(term[A, B], term[A, B])](rR, rev),
                   (x: (term[A, B], term[A, B])) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('t',
                                   'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_rule[A, B](x),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List(' ',
                                 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                                 'r', 'e', 'v', 'e', 'r', 's', 'e', 'd', ' ',
                                 'r', 'u', 'l', 'e')),
         shows_nl)))))))
  }

def checkTRSProof[A: equal: linorder: show](i: (List[Char]) => List[Char],
     r: List[(term[lab[A, List[nat]], List[Char]],
               term[lab[A, List[nat]], List[Char]])],
     m: nat, x3: tRSProof[A, List[nat], List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  (i, r, m, x3) match {
  case (i, r, m, DPTrans(p, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('D', 'P', ' ', 't', 'r', 'a', 'n', 's'),
                       {
                         val rm: Option[((lab[A, List[nat]], nat)) =>
  List[(term[lab[A, List[nat]], List[Char]],
         term[lab[A, List[nat]], List[Char]])]]
                           = computeRuleMap[lab[A, List[nat]], List[Char]](r);
                         binda[(List[Char]) => List[Char], Unit,
                                Unit](catcha[(List[Char]) => List[Char], Unit,
      (List[Char]) =>
        List[Char]](check_DP[lab[A, List[nat]],
                              List[Char]]((a: lab[A, List[nat]]) =>
    Sharp[A, List[nat]](a),
   r, rm, p),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](i, comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                  'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w', 'i',
                  't', 'c', 'h', 'i', 'n', 'g', ' ', 'f', 'r', 'o', 'm', ' ',
                  't', 'h', 'e', ' ', 'T', 'R', 'S')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl,
             comp[List[Char], List[Char],
                   List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_nl,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('t', 'o', ' ', 't', 'h',
              'e', ' ', 'i', 'n', 'i', 't', 'i', 'a', 'l', ' ', 'D', 'P', ' ',
              'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_dpp[lab[A, List[nat]], List[Char]](p, r),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl, x))))))))))),
                                       (_: Unit) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](checkDPProof[A](comp[List[Char], List[Char],
     List[Char]](i, shows_string.apply(List('.', '1'))),
p, r, Zero_nat, rm,
computeRuleMap[lab[A, List[nat]],
                List[Char]](reverse_trs[lab[A, List[nat]], List[Char]](r)),
prf),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](i, comp[List[Char], List[Char],
                        List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                     'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 'w',
                     'i', 't', 'c', 'h', ' ', 't', 'o', ' ', 'd', 'e', 'p', 'e',
                     'n', 'd', 'e', 'n', 'c', 'y', ' ', 'p', 'a', 'i', 'r',
                     's')),
                                     comp[List[Char], List[Char],
   List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))))
                       })
  case (i, r, m, Bounds(info)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('B', 'o', 'u', 'n', 'd', 's'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](check_bounds[lab[A, List[nat]],
                   List[Char]](info, r),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](i, comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
   'b', 'o', 'u', 'n', 'd', 's', ' ', 'o', 'n', ' ', 't', 'h', 'e', ' ', 'T',
   'R', 'S')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))
  case (i, r, m, RuleRemoval(rp, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_rule_removal_proc[lab[A, List[nat]],
   List[Char]](get_redpair[lab[A, List[nat]], List[Char]](rp), r, rR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', ' ', 'r', 'e',
              'm', 'o', 'v', 'a', 'l', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w',
              'i', 't', 'c', 'h', 'i', 'n', 'g', ' ', 'f', 'r', 'o', 'm', ' ',
              't', 'h', 'e', ' ', 'T', 'R', 'S')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl,
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('t', 'o', ' ', 't', 'h', 'e',
          ' ', 'T', 'R', 'S')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl,
     comp[List[Char], List[Char],
           List[Char]](shows_trs[lab[A, List[nat]], List[Char]](rR),
                        comp[List[Char], List[Char],
                              List[Char]](shows_nl, x))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       rR, m, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                   'a', 'l', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, Suc(m), Unlab(uR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('u', 'n', 'l', 'a', 'b', 'e', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](unlab_trs_tt[A, List[nat], List[Char]](r, uR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l',
              'i', 'n', 'g', ' ', 'o', 'f', ' ', 'T', 'R', 'S', ' ', 'w', 'h',
              'e', 'n', ' ', 's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g', ' ',
              'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl,
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('t', 'o', ' ', 't', 'h', 'e',
          ' ', 'T', 'R', 'S')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl,
     comp[List[Char], List[Char],
           List[Char]](shows_trs[lab[A, List[nat]], List[Char]](uR),
                        comp[List[Char], List[Char],
                              List[Char]](shows_nl, x))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       uR, m, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, Zero_nat, Unlab(uR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('u', 'n', 'l', 'a', 'b', 'e', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(term[lab[A, List[nat]], List[Char]],
     term[lab[A, List[nat]], List[Char]]),
    Unit,
    (List[Char]) =>
      List[Char]](check_subseteq[(term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]],
 List[Char]])](map_trs_list[lab[A, List[nat]], lab[A, List[nat]],
                             List[Char]]((a: lab[A, List[nat]]) =>
   unlab[A, List[nat]](a),
  r),
                uR),
                   (x: (term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]]))
                     => Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'i', 'n', ' ', 'u', 'n', 'l', 'a', 'b',
                   'e', 'l', 'i', 'n', 'g', ' ', 'o', 'f', ' ', 'T', 'R', 'S',
                   ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w', 'i', 't', 'c', 'h',
                   'i', 'n', 'g', ' ', 'f', 'r', 'o', 'm', ' ', 't', 'h', 'e',
                   ' ', 'T', 'R', 'S')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl,
              comp[List[Char], List[Char],
                    List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_nl,
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('t', 'o', ' ', 't', 'h',
               'e', ' ', 'T', 'R', 'S')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_nl,
          comp[List[Char], List[Char],
                List[Char]](shows_trs[lab[A, List[nat]], List[Char]](uR),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_nl,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List('u', 'n', 'l', 'a', 'b', 'e',
           'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_rule[lab[A, List[nat]],
                 List[Char]](x),
      shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'p', 'r',
                               'e', 's', 'e', 'n', 't'))))))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       uR, Zero_nat, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, m, RisEmpty) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('R', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y'),
                       check[(List[Char]) =>
                               List[Char]](nulla[(term[lab[A, List[nat]],
                List[Char]],
           term[lab[A, List[nat]], List[Char]])](r),
    comp[List[Char], List[Char],
          List[Char]](i, comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(':', ' ',
                            'R', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e',
                            'm', 'p', 't', 'y', ',', ' ', 'b', 'u', 't', ' ',
                            'i', 't', ' ', 'i', 's', ' ', 't', 'h', 'e', ' ',
                            'T', 'R', 'S')),
    comp[List[Char], List[Char],
          List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r), shows_nl)))))
  case (i, r, m, Semlab(sli, l, lR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('S', 'e', 'm', 'l', 'a', 'b'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](create_sem_lab_tt[A, List[Char]](sli, l, r, lR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i', 'c', ' ',
              'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       lR, Suc(m), prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 'e',
                   'm', 'a', 'n', 't', 'i', 'c', ' ', 'l', 'a', 'b', 'e', 'l',
                   'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, m, Rootlab(lR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('R', 'o', 'o', 't', 'l', 'a', 'b'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](rl_tt[A, List[nat]](r, lR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 'r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e',
              'l', 'i', 'n', 'g')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       lR, Suc(m), prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 'r', 'o',
                   'o', 't', '-', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, m, Fcc(fcs, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('F', 'c', 'c'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](fcc_tt[lab[A, List[nat]]](r, fcs, rR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't',
              'e', 'x', 't', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       rR, m, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 'f', 'l',
                   'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', ' ', 'c',
                   'l', 'o', 's', 'u', 'r', 'e')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, m, StringReversal(rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('S', 't', 'r', 'i', 'n', 'g', ' ', 'R', 'e', 'v',
                             'e', 'r', 's', 'a', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_string_rev[lab[A, List[nat]], List[Char]](r, rR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 's', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e',
              'v', 'e', 'r', 's', 'a', 'l')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       rR, m, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 't',
                   'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v', 'e', 'r', 's', 'a',
                   'l')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, m, Uncurry(u_info, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('U', 'n', 'c', 'u', 'r', 'r', 'y'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](uncurry_tt[lab[A, List[nat]]](u_info, r, rR),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g', ' ',
              't', 'r', 'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o',
              'n')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       rR, Zero_nat, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 'u', 'n',
                   'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
}

def init_lab[A, B](f: A): lab[A, B] = UnLab[A, B](f)

def check_dp_loop[A: equal: show,
                   B: equal: show](s: term[A, B],
                                    prseq:
                                      List[(pos,
     ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                    sigma: List[(B, term[A, B])], c: ctxt[A, B],
                                    p: List[(term[A, B], term[A, B])],
                                    r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (if (equal_ctxt[A, B](c, Hole))
    binda[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (nulla[(pos,
        ((term[A, B], term[A, B]), (Boolean, term[A, B])))](prseq)),
                                      shows_prec_list[Char](Zero_nat,
                     List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                           'u', 'e', 'n', 'c', 'e', ' ', 'm', 'u', 's', 't',
                           ' ', 'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm',
                           'p', 't', 'y'))),
                  (_: Unit) =>
                    check_steps[A, B](p, r, prseq, s,
                                       subst_apply_term[A, B,
                 B](s, subst_of[B, A](sigma))))
    else check_loop[A, B](s, mapa[(pos, ((term[A, B], term[A, B]),
  (Boolean, term[A, B]))),
                                   (pos, ((term[A, B], term[A, B]),
   term[A, B]))]((a: (pos, ((term[A, B], term[A, B]), (Boolean, term[A, B]))))
                   => {
                        val (x, (y, (_, z))):
                              (pos, ((term[A, B], term[A, B]),
                                      (Boolean, term[A, B])))
                          = a;
                        (x, (y, z))
                      },
                  prseq),
                           sigma, c, r))

def check_dp_disproof[A: equal: show,
                       B: equal: show](i: (List[Char]) => List[Char],
p: List[(term[A, B], term[A, B])], r: List[(term[A, B], term[A, B])],
x3: dPDisproof[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (i, p, r, x3) match {
  case (i, p, r, DPLoop(s, rseq, sigma, c)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('L', 'o', 'o', 'p'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](check_dp_loop[A,
                    B](s, rseq, sigma, c, p, r),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](i, comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'i', 'n', ' ', 'l', 'o', 'o', 'p', ' ', 'p', 'r', 'o', 'o', 'f')),
                   comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
  case (i, pa, ra, DPRuleRemoval(p, r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(term[A, B], term[A, B]), Unit,
    (List[Char]) =>
      List[Char]](check_subseteq[(term[A, B], term[A, B])](p, pa),
                   (x: (term[A, B], term[A, B])) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', '-', 'r', 'e',
              'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r', 'o', 'o', 'f')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_rule[A, B](x),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_string.apply(List(' ', 'i',
                               's', ' ', 'a', ' ', 'n', 'e', 'w', ' ', 'p', 'a',
                               'i', 'r', ' ', 'n', 'o', 't', ' ', 'o', 'c', 'c',
                               'u', 'r', 'r', 'i', 'n', 'g', ' ', 'i', 'n', ' ',
                               't', 'h', 'e', ' ', 'p', 'r', 'e', 'v', 'i', 'o',
                               'u', 's', ' ', 's', 'e', 't', ' ', 'o', 'f', ' ',
                               'p', 'a', 'i', 'r', 's')),
       shows_nl))))))),
                                     (_: Unit) =>
                                       binda[(List[Char]) => List[Char], Unit,
      Unit](catcha[(term[A, B], term[A, B]), Unit,
                    (List[Char]) =>
                      List[Char]](check_subseteq[(term[A, B],
           term[A, B])](r, ra),
                                   (x: (term[A, B], term[A, B])) =>
                                     Inl[(List[Char]) => List[Char],
  Unit](comp[List[Char], List[Char],
              List[Char]](i, comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List(':', ' ',
                                'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                                'r', 'u', 'l', 'e', '-', 'r', 'e', 'm', 'o',
                                'v', 'a', 'l', ' ', 'p', 'r', 'o', 'o', 'f')),
        comp[List[Char], List[Char],
              List[Char]](shows_nl,
                           comp[List[Char], List[Char],
                                 List[Char]](shows_rule[A, B](x),
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'a', ' ',
         'n', 'e', 'w', ' ', 'r', 'u', 'l', 'e', ' ', 'n', 'o', 't', ' ', 'o',
         'c', 'c', 'u', 'r', 'r', 'i', 'n', 'g', ' ', 'i', 'n', ' ', 't', 'h',
         'e', ' ', 'p', 'r', 'e', 'v', 'i', 'o', 'u', 's', ' ', 'T', 'R', 'S')),
                         shows_nl))))))),
             (_: Unit) =>
               check_dp_disproof[A, B](comp[List[Char], List[Char],
     List[Char]](i, shows_string.apply(List('.', '1'))),
p, r, prf))))
}

def checkTRSDisproof[A: equal: show, B: equal: show,
                      C: equal: show](i: (List[Char]) => List[Char],
                                       r:
 List[(term[lab[A, B], C], term[lab[A, B], C])],
                                       x2: tRSDisproof[lab[A, B], C]):
      sum[(List[Char]) => List[Char], Unit] =
  (i, r, x2) match {
  case (i, r, Loop(s, rseq, sigma, c)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('L', 'o', 'o', 'p'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](check_loop[lab[A, B],
                 C](s, rseq, sigma, c, r),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](i, comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'i', 'n', ' ', 'l', 'o', 'o', 'p', ' ', 'p', 'r', 'o', 'o', 'f')),
                   comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
  case (i, r, NotWellFormed) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f',
                             'o', 'r', 'm', 'e', 'd'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](check_not_wf_trs[lab[A, B], C](r),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](i, comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'i', 'n', ' ', 'n', 'o', 't', '-', 'w', 'e', 'l', 'l', '-', 'f', 'o',
   'r', 'm', 'e', 'd', ' ', 'p', 'r', 'o', 'o', 'f')),
                   comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
  case (i, ra, NTRuleRemoval(r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(term[lab[A, B], C],
     term[lab[A, B], C]),
    Unit,
    (List[Char]) =>
      List[Char]](check_subseteq[(term[lab[A, B], C],
                                   term[lab[A, B], C])](r, ra),
                   (x: (term[lab[A, B], C], term[lab[A, B], C])) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', '-', 'r', 'e',
              'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r', 'o', 'o', 'f')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_rule[lab[A, B], C](x),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_string.apply(List(' ', 'i',
                               's', ' ', 'a', ' ', 'n', 'e', 'w', ' ', 'r', 'u',
                               'l', 'e', ' ', 'n', 'o', 't', ' ', 'o', 'c', 'c',
                               'u', 'r', 'r', 'i', 'n', 'g', ' ', 'i', 'n', ' ',
                               't', 'h', 'e', ' ', 'p', 'r', 'e', 'v', 'i', 'o',
                               'u', 's', ' ', 'T', 'R', 'S')),
       shows_nl))))))),
                                     (_: Unit) =>
                                       checkTRSDisproof[A, B,
                 C](comp[List[Char], List[Char],
                          List[Char]](i, shows_string.apply(List('.', '1'))),
                     r, prf)))
  case (i, ra, NTStringReversal(r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('s', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v',
                             'e', 'r', 's', 'a', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_string_rev[lab[A, B], C](ra, r),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 's', 't', 'r', 'i', 'n', 'g', ' ',
              'r', 'e', 'v', 'e', 'r', 's', 'a', 'l', ' ', 'p', 'r', 'o', 'o',
              'f')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       checkTRSDisproof[A, B,
                 C](comp[List[Char], List[Char],
                          List[Char]](i, shows_string.apply(List('.', '1'))),
                     r, prf)))
  case (i, r, NTDPTrans(p, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('D', 'P', ' ', 't', 'r', 'a', 'n', 's', 'f', 'o',
                             'r', 'm', 'a', 't', 'i', 'o', 'n'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(term[lab[A, B], C],
     term[lab[A, B], C]),
    Unit,
    (List[Char]) =>
      List[Char]](check_DPs[lab[A, B],
                             C]((a: lab[A, B]) => unsharp[A, B](a), r, p),
                   (x: (term[lab[A, B], C], term[lab[A, B], C])) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'D', 'P', ' ', 't', 'r', 'a', 'n',
              's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o', 'n')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_prec_list[Char](Zero_nat,
          List('p', 'r', 'o', 'b', 'l', 'e', 'm', 'a', 't', 'i', 'c', ' ', 'r',
                'u', 'l', 'e', ' ')),
                            shows_rule[lab[A, B], C](x))))))),
                                     (_: Unit) =>
                                       check_dp_disproof[lab[A, B],
                  C](comp[List[Char], List[Char],
                           List[Char]](i, shows_string.apply(List('.', '1'))),
                      p, r, prf)))
}

def check_relative_rule_removal[A: equal: show,
                                 B: equal: show](rp:
           (sum[(List[Char]) => List[Char], Unit],
             (((term[A, B], term[A, B])) =>
                sum[(List[Char]) => List[Char], Unit],
               (((term[A, B], term[A, B])) =>
                  sum[(List[Char]) => List[Char], Unit],
                 (A => nat => nat => Boolean,
                   (sum[(List[Char]) => List[Char], Unit],
                     (List[Char]) => List[Char]))))),
          x1: (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]),
          x2: (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])):
      sum[(List[Char]) => List[Char], Unit] =
  (rp, x1, x2) match {
  case (rp, (r, s), (rR, rS)) =>
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]]({
                            val (_, nSr):
                                  (List[(term[A, B], term[A, B])],
                                    List[(term[A, B], term[A, B])])
                              = partition[(term[A, B],
    term[A, B])]((foo: (term[A, B], term[A, B])) =>
                   isOK[(List[Char]) => List[Char],
                         Unit]((rp_strict[A, B](rp)).apply(foo)),
                  r)
                            val (_, nSs):
                                  (List[(term[A, B], term[A, B])],
                                    List[(term[A, B], term[A, B])])
                              = partition[(term[A, B],
    term[A, B])]((foo: (term[A, B], term[A, B])) =>
                   isOK[(List[Char]) => List[Char],
                         Unit]((rp_strict[A, B](rp)).apply(foo)),
                  s);
                            binda[(List[Char]) => List[Char], Unit,
                                   Unit](rp_valid[A, B](rp),
  (_: Unit) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](catcha[(List[Char]) => List[Char], Unit,
                         (List[Char]) =>
                           List[Char]](rp_mono[A, B](rp),
(x: (List[Char]) => List[Char]) =>
  Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
                'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
                'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 's', 't',
                'r', 'i', 'c', 't', ' ', 'o', 'r', 'd', 'e', 'r')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_nl, x)))),
                  (_: Unit) =>
                    binda[(List[Char]) => List[Char], Unit,
                           Unit](catcha[(List[Char]) => List[Char], Unit,
 (List[Char]) =>
   List[Char]](catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]),
                       Unit,
                       (List[Char]) =>
                         List[Char]](forallm[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSr),
                                      (x:
 ((term[A, B], term[A, B]), (List[Char]) => List[Char]))
=> Inl[(List[Char]) => List[Char],
        Unit](snd[(term[A, B], term[A, B]), (List[Char]) => List[Char]](x))),
                (x: (List[Char]) => List[Char]) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List('p', 'r',
                                'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e',
                                'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i',
                                'n', 'g', ' ', 'r', 'u', 'l', 'e', 's')),
        comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                  (_: Unit) =>
                                    binda[(List[Char]) => List[Char], Unit,
   Unit](catcha[(List[Char]) => List[Char], Unit,
                 (List[Char]) =>
                   List[Char]](catcha[((term[A, B], term[A, B]),
(List[Char]) => List[Char]),
                                       Unit,
                                       (List[Char]) =>
 List[Char]](forallm[(term[A, B], term[A, B]),
                      (List[Char]) => List[Char]](rp_non_strict[A, B](rp), nSs),
              (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
                Inl[(List[Char]) => List[Char],
                     Unit](snd[(term[A, B], term[A, B]),
                                (List[Char]) => List[Char]](x))),
                                (x: (List[Char]) => List[Char]) =>
                                  Inl[(List[Char]) => List[Char],
                                       Unit](comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e',
        'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i',
        'n', 'g', ' ', 'r', 'u', 'l', 'e', 's')),
                        comp[List[Char], List[Char],
                              List[Char]](shows_nl, x)))),
          (_: Unit) =>
            binda[(List[Char]) => List[Char], Unit,
                   Unit](catcha[(term[A, B], term[A, B]), Unit,
                                 (List[Char]) =>
                                   List[Char]](check_subseteq[(term[A, B],
                        term[A, B])](nSs, rS),
        (x: (term[A, B], term[A, B])) =>
          Inl[(List[Char]) => List[Char],
               Unit]((rp_strict[A, B](rp)).apply(x) match {
                       case Inl(err) => err
                       case Inr(_) => internal_error
                     })),
                          (_: Unit) =>
                            catcha[(term[A, B], term[A, B]), Unit,
                                    (List[Char]) =>
                                      List[Char]](check_subseteq[(term[A, B],
                           term[A, B])](nSr, rR),
           (x: (term[A, B], term[A, B])) =>
             Inl[(List[Char]) => List[Char],
                  Unit]((rp_strict[A, B](rp)).apply(x) match {
                          case Inl(err) => err
                          case Inr(_) => internal_error
                        })))))))
                          },
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
   'o', 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'u',
   'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r', 'o', 'c',
   'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ',
   'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](rp_desc[A, B](rp),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
}

def unlab_reltrs_tt[A: equal: show, B: equal: show,
                     C: equal: show](rs: (List[(term[lab[A, B], C],
         term[lab[A, B], C])],
   List[(term[lab[A, B], C], term[lab[A, B], C])]),
                                      rRS:
(List[(term[lab[A, B], C], term[lab[A, B], C])],
  List[(term[lab[A, B], C], term[lab[A, B], C])])):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(term[lab[A, B], C], term[lab[A, B], C]), Unit,
          (List[Char]) =>
            List[Char]](binda[(term[lab[A, B], C], term[lab[A, B], C]), Unit,
                               Unit](check_subseteq[(term[lab[A, B], C],
              term[lab[A, B],
                    C])](map_trs_list[lab[A, B], lab[A, B],
                                       C]((a: lab[A, B]) => unlab[A, B](a),
   fst[List[(term[lab[A, B], C], term[lab[A, B], C])],
        List[(term[lab[A, B], C], term[lab[A, B], C])]](rs)),
                          fst[List[(term[lab[A, B], C], term[lab[A, B], C])],
                               List[(term[lab[A, B], C],
                                      term[lab[A, B], C])]](rRS)),
                                      (_: Unit) =>
catcha[((term[lab[A, B], C], term[lab[A, B], C]),
         (term[lab[A, B], C], term[lab[A, B], C])),
        Unit,
        (term[lab[A, B], C],
          term[lab[A, B],
                C])](forallm[(term[lab[A, B], C], term[lab[A, B], C]),
                              (term[lab[A, B], C],
                                term[lab[A, B],
                                      C])]((x:
      (term[lab[A, B], C], term[lab[A, B], C]))
     => (if (equal_terma[lab[A, B],
                          C](fst[term[lab[A, B], C], term[lab[A, B], C]](x),
                              snd[term[lab[A, B], C], term[lab[A, B], C]](x)) ||
               membera[(term[lab[A, B], C],
                         term[lab[A, B],
                               C])](snd[List[(term[lab[A, B], C],
       term[lab[A, B], C])],
 List[(term[lab[A, B], C], term[lab[A, B], C])]](rRS),
                                     x))
          Inr[Unit, (term[lab[A, B], C], term[lab[A, B], C])](())
          else Inl[(term[lab[A, B], C], term[lab[A, B], C]), Unit](x)),
    map_trs_list[lab[A, B], lab[A, B],
                  C]((a: lab[A, B]) => unlab[A, B](a),
                      snd[List[(term[lab[A, B], C], term[lab[A, B], C])],
                           List[(term[lab[A, B], C],
                                  term[lab[A, B], C])]](rs))),
                      (x: ((term[lab[A, B], C], term[lab[A, B], C]),
                            (term[lab[A, B], C], term[lab[A, B], C])))
                        => Inl[(term[lab[A, B], C], term[lab[A, B], C]),
                                Unit](snd[(term[lab[A, B], C],
    term[lab[A, B], C]),
   (term[lab[A, B], C], term[lab[A, B], C])](x)))),
                         (x: (term[lab[A, B], C], term[lab[A, B], C])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'a', 'n', 'n', 'o', 't', ' ', 'f',
 'i', 'n', 'd', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u',
 'l', 'e', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl, shows_rule[lab[A, B], C](x)))))

def create_sem_lab_rel_tt[A: equal: show,
                           B: equal: show](sli: sl_inter[A, List[nat]],
    x1: sum[(lab[A, List[nat]]) =>
              (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]],
             B],
    rs: (List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])],
          List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]),
    lrs: (List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])],
           List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])])):
      sum[(List[Char]) => List[Char], Unit] =
  (sli, x1, rs, lrs) match {
  case (sli, Inl(l), rs, lrs) =>
    {
      val c: nat = get_largest_element[A, List[nat]](sli)
      val ca: List[nat] = upt(Zero_nat, Suc(c))
      val i: (lab[A, List[nat]]) => (List[nat]) => nat =
        (a: lab[A, List[nat]]) =>
          (b: List[nat]) => sl_inter_to_inter[A, List[nat]](sli, a, b);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_sem_lab_rel_tt[A, List[nat],
          B](model_check_valid[(List[Char]) => List[Char]],
              model_check_decr[List[(term[lab[A, List[nat]], B],
                                      term[lab[A, List[nat]], B])],
                                (List[Char]) => List[Char]],
              (a: List[(term[lab[A, List[nat]], B],
                         term[lab[A, List[nat]], B])])
                => (b: List[(term[lab[A, List[nat]], B],
                              term[lab[A, List[nat]], B])])
                     => check_sl_lab_trs[A, List[nat], nat,
  B](i, l, ca, model_cge[nat], a, b),
              (a: List[(term[lab[A, List[nat]], B],
                         term[lab[A, List[nat]], B])])
                => (b: List[(term[lab[A, List[nat]], B],
                              term[lab[A, List[nat]], B])])
                     => check_sl_model_lab_trs[A, List[nat], nat,
        B](i, l, ca, model_cge[nat], a, b),
              model_LS[lab[A, List[nat]], nat,
                        sum[List[nat], List[lab[A, List[nat]]]]],
              rs, lrs),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_prec_sl_inter[A, List[nat]](Zero_nat, sli),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
    }
  case (sli, Inr(v), rs, lrs) =>
    {
      val c: nat = get_largest_element[A, List[nat]](sli)
      val ca: List[nat] = upt(Zero_nat, Suc(c))
      val i: (lab[A, List[nat]]) => (List[nat]) => nat =
        (a: lab[A, List[nat]]) =>
          (b: List[nat]) => sl_inter_to_inter[A, List[nat]](sli, a, b)
      val sig: List[(lab[A, List[nat]], nat)] =
        funas_trs_list[lab[A, List[nat]],
                        B](fst[List[(term[lab[A, List[nat]], B],
                                      term[lab[A, List[nat]], B])],
                                List[(term[lab[A, List[nat]], B],
                                       term[lab[A, List[nat]], B])]](rs) ++
                             snd[List[(term[lab[A, List[nat]], B],
term[lab[A, List[nat]], B])],
                                  List[(term[lab[A, List[nat]], B],
 term[lab[A, List[nat]], B])]](rs))
      val l: (lab[A, List[nat]]) =>
               (List[nat]) => sum[List[nat], List[lab[A, List[nat]]]]
        = qmodel_L[A, List[nat], nat, List[lab[A, List[nat]]]](sig);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_sem_lab_rel_tt[A, List[nat],
          B](qmodel_check_valid[A, List[nat]](sli),
              qmodel_check_decr[A, B](sig, v, c),
              (a: List[(term[lab[A, List[nat]], B],
                         term[lab[A, List[nat]], B])])
                => (b: List[(term[lab[A, List[nat]], B],
                              term[lab[A, List[nat]], B])])
                     => check_sl_lab_trs[A, List[nat], nat,
  B](i, l, ca, qmodel_cge, a, b),
              (a: List[(term[lab[A, List[nat]], B],
                         term[lab[A, List[nat]], B])])
                => (b: List[(term[lab[A, List[nat]], B],
                              term[lab[A, List[nat]], B])])
                     => check_sl_model_lab_trs[A, List[nat], nat,
        B](i, l, ca, qmodel_cge, a, b),
              qmodel_LS[lab[A, List[nat]], nat,
                         List[lab[A, List[nat]]]](sig, ca),
              rs, lrs),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_prec_sl_inter[A, List[nat]](Zero_nat, sli),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x))))
    }
}

def checkRelTRSProof[A: equal: linorder: show](i: (List[Char]) => List[Char],
        r: List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])],
        s: List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])],
        x3: relTRSProof[A, List[nat], List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  (i, r, s, x3) match {
  case (i, r, s, RelRisEmpty) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('R', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y'),
                       check[(List[Char]) =>
                               List[Char]](nulla[(term[lab[A, List[nat]],
                List[Char]],
           term[lab[A, List[nat]], List[Char]])](r),
    comp[List[Char], List[Char],
          List[Char]](i, comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(':', ' ',
                            'R', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e',
                            'm', 'p', 't', 'y', ',', ' ', 'b', 'u', 't', ' ',
                            'i', 't', ' ', 'i', 's', ' ', 't', 'h', 'e', ' ',
                            'T', 'R', 'S')),
    comp[List[Char], List[Char],
          List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r), shows_nl)))))
  case (i, r, s, RelSisEmpty(prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('S', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](check[(List[Char]) =>
    List[Char]](nulla[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])](s),
                 comp[List[Char], List[Char],
                       List[Char]](i, comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(':', ' ', 'S', ' ', 'i', 's', ' ', 'n',
 'o', 't', ' ', 'e', 'm', 'p', 't', 'y', ',', ' ', 'b', 'u', 't', ' ', 'i', 't',
 ' ', 'i', 's', ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_trs[lab[A, List[nat]], List[Char]](s),
                                    shows_nl)))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkTRSProof[A](comp[List[Char], List[Char],
    List[Char]](i, shows_string.apply(List('.', '1'))),
                                       r, Zero_nat, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                   'a', 'l', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, s, RelRuleRemoval(rp, rR, rS, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](check_relative_rule_removal[lab[A, List[nat]],
       List[Char]](get_redpair[lab[A, List[nat]], List[Char]](rp), (r, s),
                    (rR, rS)),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', ' ', 'r', 'e',
              'm', 'o', 'v', 'a', 'l', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w',
              'i', 't', 'c', 'h', 'i', 'n', 'g', ' ', 'f', 'r', 'o', 'm', ' ',
              't', 'h', 'e', ' ', 'T', 'R', 'S', 's')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl,
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List(' ', 'a', 'n', 'd', ' ')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl,
     comp[List[Char], List[Char],
           List[Char]](shows_trs[lab[A, List[nat]], List[Char]](s),
                        comp[List[Char], List[Char],
                              List[Char]](shows_nl,
   comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('t', 'o', ' ', 't', 'h', 'e', ' ',
      'T', 'R', 'S', 's')),
                      comp[List[Char], List[Char],
                            List[Char]](shows_nl,
 comp[List[Char], List[Char],
       List[Char]](shows_trs[lab[A, List[nat]], List[Char]](rR),
                    comp[List[Char], List[Char],
                          List[Char]](shows_nl,
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List(' ', 'a', 'n', 'd', ' ')),
                  comp[List[Char], List[Char],
                        List[Char]](shows_nl,
                                     comp[List[Char], List[Char],
   List[Char]](shows_trs[lab[A, List[nat]], List[Char]](rS),
                x)))))))))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkRelTRSProof[A](comp[List[Char], List[Char],
       List[Char]](i, shows_string.apply(List('.', '1'))),
  rR, rS, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                   'a', 'l', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, s, RelUnlab(uR, uS, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('u', 'n', 'l', 'a', 'b', 'e', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](unlab_reltrs_tt[A, List[nat], List[Char]]((r, s), (uR, uS)),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l',
              'i', 'n', 'g', ' ', 'o', 'f', ' ', 'T', 'R', 'S', ' ', 'w', 'h',
              'e', 'n', ' ', 's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g', ' ',
              'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S', ' ',
              'R')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_trs[lab[A, List[nat]], List[Char]](r),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl,
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('a', 'n', 'd', ' ', 'S')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl,
     comp[List[Char], List[Char],
           List[Char]](shows_trs[lab[A, List[nat]], List[Char]](s),
                        comp[List[Char], List[Char],
                              List[Char]](shows_string.apply(List('t', 'o', ' ',
                           't', 'h', 'e', ' ', 'T', 'R', 'S', 's', ' ')),
   comp[List[Char], List[Char],
         List[Char]](shows_nl,
                      comp[List[Char], List[Char],
                            List[Char]](shows_trs[lab[A, List[nat]],
           List[Char]](uR),
 comp[List[Char], List[Char],
       List[Char]](shows_nl,
                    comp[List[Char], List[Char],
                          List[Char]](shows_trs[lab[A, List[nat]],
         List[Char]](uS),
                                       x))))))))))))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkRelTRSProof[A](comp[List[Char], List[Char],
       List[Char]](i, shows_string.apply(List('.', '1'))),
  uR, uS, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, s, RelSemlab(sli, l, lR, lS, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('S', 'e', 'm', 'l', 'a', 'b'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](create_sem_lab_rel_tt[A,
 List[Char]](sli, l, (r, s), (lR, lS)),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i', 'c', ' ',
              'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkRelTRSProof[A](comp[List[Char], List[Char],
       List[Char]](i, shows_string.apply(List('.', '1'))),
  lR, lS, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 'e',
                   'm', 'a', 'n', 't', 'i', 'c', ' ', 'l', 'a', 'b', 'e', 'l',
                   'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, s, RelRootlab(lR, lS, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('R', 'o', 'o', 't', 'l', 'a', 'b'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](rl_rel_tt[A, List[nat]]((r, s), (lR, lS)),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 'r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e',
              'l', 'i', 'n', 'g')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkRelTRSProof[A](comp[List[Char], List[Char],
       List[Char]](i, shows_string.apply(List('.', '1'))),
  lR, lS, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 'e',
                   'm', 'a', 'n', 't', 'i', 'c', ' ', 'l', 'a', 'b', 'e', 'l',
                   'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (i, r, s, RelUncurry(info, uR, uS, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('U', 'n', 'c', 'u', 'r', 'r', 'y'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(List[Char]) => List[Char], Unit,
    (List[Char]) =>
      List[Char]](uncurry_rel_tt[lab[A, List[nat]]](info, r, s, uR, uS),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
              'i', 'n', 'g', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n',
              'g')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x))))),
                                     (_: Unit) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](checkRelTRSProof[A](comp[List[Char], List[Char],
       List[Char]](i, shows_string.apply(List('.', '1'))),
  uR, uS, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](i, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 'u', 'n',
                   'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
}

def check_rel_seq[A: equal: show,
                   B: equal: show](r: List[(term[A, B], term[A, B])],
                                    sa: List[(term[A, B], term[A, B])],
                                    x2: List[(pos,
       ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                    s: term[A, B], u: term[A, B], b: Boolean):
      sum[(List[Char]) => List[Char], Unit] =
  (r, sa, x2, s, u, b) match {
  case (r, sa, Nil, s, u, true) =>
    check[(List[Char]) =>
            List[Char]](equal_terma[A, B](s, u),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('t', 'h',
                            'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r',
                            'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'r',
                            'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                            'u', 'e', 'n', 'c', 'e')),
    comp[List[Char], List[Char],
          List[Char]](shows_nl,
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_term[A, B](Zero_nat, s),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(shows_nl.apply(List('d', 'o', 'e', 's',
                    ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p',
                    'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'g',
                    'o', 'a', 'l', ' ', 't', 'e', 'r', 'm'))),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_prec_term[A, B](Zero_nat, u), shows_nl)))))))
  case (r, sa, Nil, s, u, false) =>
    Inl[(List[Char]) => List[Char],
         Unit](shows_prec_list[Char](Zero_nat,
                                      List('d', 'i', 'd', ' ', 'n', 'o', 't',
    ' ', 'f', 'i', 'n', 'd', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 's', 't',
    'e', 'p', ' ', 'i', 'n', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's',
    'e', 'q', 'u', 'e', 'n', 'c', 'e')))
  case (ra, sa, (p, (r, (true, t))) :: prts, s, u, b) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](check_rstep[A, B](ra, p, r, s, t),
                  (_: Unit) => check_rel_seq[A, B](ra, sa, prts, t, u, true))
  case (ra, sa, (p, (r, (false, t))) :: prts, s, u, b) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](check_rstep[A, B](sa, p, r, s, t),
                  (_: Unit) => check_rel_seq[A, B](ra, sa, prts, t, u, b))
}

def check_rel_loop[A: equal: show,
                    B: equal: show](sa: term[A, B],
                                     rseq: List[(pos,
          ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                     sigma: List[(B, term[A, B])],
                                     c: ctxt[A, B],
                                     r: List[(term[A, B], term[A, B])],
                                     s: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  check_rel_seq[A, B](r, s, rseq, sa,
                       ctxt_apply[A, B](c,
 subst_apply_term[A, B, B](sa, subst_of[B, A](sigma))),
                       false)

def check_wf_rel_trs[A: show,
                      B: equal: show](x0:
(List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case (r, s) =>
    binda[(List[Char]) => List[Char], Unit,
           Unit](check_wf_trs[A, B](r),
                  (_: Unit) =>
                    (if (nulla[(term[A, B], term[A, B])](r))
                      Inr[Unit, (List[Char]) => List[Char]](())
                      else catcha[((term[A, B], term[A, B]),
                                    (List[Char]) => List[Char]),
                                   Unit,
                                   (List[Char]) =>
                                     List[Char]](forallm[(term[A, B],
                   term[A, B]),
                  (List[Char]) =>
                    List[Char]]((rule: (term[A, B], term[A, B])) =>
                                  catcha[(List[Char]) => List[Char], Unit,
  (List[Char]) =>
    List[Char]](catcha[B, Unit,
                        (List[Char]) =>
                          List[Char]](check_subseteq[B](vars_term_list[A,
                                B].apply(snd[term[A, B], term[A, B]](rule)),
                 vars_term_list[A, B].apply(fst[term[A, B], term[A, B]](rule))),
                                       (x: B) =>
 Inl[(List[Char]) => List[Char],
      Unit](comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('f', 'r', 'e', 'e', ' ',
               'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ')),
                               comp[List[Char], List[Char],
                                     List[Char]]((a: List[Char]) =>
           shows_prec[B](Zero_nat, x, a),
          shows_string.apply(List(' ', 'i', 'n', ' ', 'r', 'i', 'g', 'h', 't',
                                   '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd',
                                   'e', ' ', 'o', 'f')))))),
                 (x: (List[Char]) => List[Char]) =>
                   Inl[(List[Char]) => List[Char],
                        Unit](comp[List[Char], List[Char],
                                    List[Char]](x,
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List(' ', 'r', 'u', 'l', 'e',
            ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_rule[A, B](rule),
       shows_nl))))),
                                 s),
          (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
            Inl[(List[Char]) => List[Char],
                 Unit](snd[(term[A, B], term[A, B]),
                            (List[Char]) => List[Char]](x)))))
}

def check_not_wf_rel_trs[A: show,
                          B: equal: show](rs:
    (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (isOK[(List[Char]) => List[Char],
                               Unit](check_wf_rel_trs[A, B](rs))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('T', 'h', 'e',
                          ' ', 'T', 'R', 'S', 's', ' ', 'R', ' ', 'a', 'n', 'd',
                          ' ', 'S', ' ', 'a', 'r', 'e', ' ', 'w', 'e', 'l', 'l',
                          ' ', 'f', 'o', 'r', 'm', 'e', 'd')),
  shows_nl))

def checkRelTRSDisproof[A: equal: show, B: equal: show,
                         C: equal: show](i: (List[Char]) => List[Char],
  x1: (List[(term[lab[A, B], C], term[lab[A, B], C])],
        List[(term[lab[A, B], C], term[lab[A, B], C])]),
  x2: relTRSDisproof[lab[A, B], C]):
      sum[(List[Char]) => List[Char], Unit] =
  (i, x1, x2) match {
  case (i, (r, sa), RelLoop(s, rseq, sigma, c)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil), List('L', 'o', 'o', 'p'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](check_rel_loop[lab[A, B],
                     C](s, rseq, sigma, c, r, sa),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](i, comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'i', 'n', ' ', 'l', 'o', 'o', 'p', ' ', 'p', 'r', 'o', 'o', 'f')),
                   comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
  case (i, (r, s), RNotSN(prf)) =>
    checkTRSDisproof[A, B,
                      C](comp[List[Char], List[Char],
                               List[Char]](i,
    shows_string.apply(List('.', '1'))),
                          r, prf)
  case (i, (r, s), RelNotWellFormed) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f',
                             'o', 'r', 'm', 'e', 'd'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](check_not_wf_rel_trs[lab[A, B],
                           C]((r, s)),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](i, comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'i', 'n', ' ', 'n', 'o', 't', '-', 'w', 'e', 'l', 'l', '-', 'f', 'o',
   'r', 'm', 'e', 'd', ' ', 'p', 'r', 'o', 'o', 'f')),
                   comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
  case (i, (ra, sa), NTRelRuleRemoval(r, s, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](i(Nil),
                       List('r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       binda[(List[Char]) => List[Char], Unit,
                              Unit](catcha[(term[lab[A, B], C],
     term[lab[A, B], C]),
    Unit,
    (List[Char]) =>
      List[Char]](binda[(term[lab[A, B], C], term[lab[A, B], C]), Unit,
                         Unit](check_subseteq[(term[lab[A, B], C],
        term[lab[A, B], C])](r, ra),
                                (_: Unit) =>
                                  check_subseteq[(term[lab[A, B], C],
           term[lab[A, B], C])](s, sa)),
                   (x: (term[lab[A, B], C], term[lab[A, B], C])) =>
                     Inl[(List[Char]) => List[Char],
                          Unit](comp[List[Char], List[Char],
                                      List[Char]](i,
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
              'o', 'r', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', '-', 'r', 'e',
              'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r', 'o', 'o', 'f')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl,
         comp[List[Char], List[Char],
               List[Char]](shows_rule[lab[A, B], C](x),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_string.apply(List(' ', 'i',
                               's', ' ', 'a', ' ', 'n', 'e', 'w', ' ', 'r', 'u',
                               'l', 'e', ' ')),
       shows_nl))))))),
                                     (_: Unit) =>
                                       checkRelTRSDisproof[A, B,
                    C](comp[List[Char], List[Char],
                             List[Char]](i, shows_string.apply(List('.', '1'))),
                        (r, s), prf)))
}

def certifyCertProblem[A: equal: linorder: show](x0:
           certificationProblem[A, List[nat], List[Char]]):
      certificationResult =
  x0 match {
  case TRSProof(trs, prf) =>
    checkTRSProof[A](shows_string.apply(List('1')),
                      mapa[(term[A, List[Char]], term[A, List[Char]]),
                            (term[lab[A, List[nat]], List[Char]],
                              term[lab[A, List[nat]],
                                    List[Char]])]((a:
             (term[A, List[Char]], term[A, List[Char]]))
            => map_rule[A, lab[A, List[nat]],
                         List[Char]]((aa: A) => init_lab[A, List[nat]](aa), a),
           trs),
                      Zero_nat, prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('t', 'e', 'r', 'm',
                  'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'o',
                  'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't',
                  'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(List('<', 't', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a',
                        't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'))
    }
  case TRSDisproof(trs, prf) =>
    checkTRSDisproof[A, List[nat],
                      List[Char]](shows_string.apply(List('1')),
                                   mapa[(term[A, List[Char]],
  term[A, List[Char]]),
 (term[lab[A, List[nat]], List[Char]],
   term[lab[A, List[nat]],
         List[Char]])]((a: (term[A, List[Char]], term[A, List[Char]])) =>
                         map_rule[A, lab[A, List[nat]],
                                   List[Char]]((aa: A) =>
         init_lab[A, List[nat]](aa),
        a),
                        trs),
                                   prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('n', 'o', 'n', '-',
                  't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ',
                  'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c',
                  'c', 'e', 'p', 't', 'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(List('<', 't', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm',
                        'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
                        'f', '>'))
    }
  case RelTRSProof((r, s), prf) =>
    checkRelTRSProof[A](shows_string.apply(List('1')),
                         mapa[(term[A, List[Char]], term[A, List[Char]]),
                               (term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]],
                                       List[Char]])]((a:
                (term[A, List[Char]], term[A, List[Char]]))
               => map_rule[A, lab[A, List[nat]],
                            List[Char]]((aa: A) => init_lab[A, List[nat]](aa),
 a),
              r),
                         mapa[(term[A, List[Char]], term[A, List[Char]]),
                               (term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]],
                                       List[Char]])]((a:
                (term[A, List[Char]], term[A, List[Char]]))
               => map_rule[A, lab[A, List[nat]],
                            List[Char]]((aa: A) => init_lab[A, List[nat]](aa),
 a),
              s),
                         prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('r', 'e', 'l', 'a',
                  't', 'i', 'v', 'e', ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a',
                  't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'n',
                  'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(List('<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e',
                        'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r',
                        'o', 'o', 'f', '>'))
    }
  case RelTRSDisproof((r, s), prf) =>
    checkRelTRSDisproof[A, List[nat],
                         List[Char]](shows_string.apply(List('1')),
                                      (mapa[(term[A, List[Char]],
      term[A, List[Char]]),
     (term[lab[A, List[nat]], List[Char]],
       term[lab[A, List[nat]],
             List[Char]])]((a: (term[A, List[Char]], term[A, List[Char]])) =>
                             map_rule[A, lab[A, List[nat]],
                                       List[Char]]((aa: A) =>
             init_lab[A, List[nat]](aa),
            a),
                            r),
mapa[(term[A, List[Char]], term[A, List[Char]]),
      (term[lab[A, List[nat]], List[Char]],
        term[lab[A, List[nat]],
              List[Char]])]((a: (term[A, List[Char]], term[A, List[Char]])) =>
                              map_rule[A, lab[A, List[nat]],
List[Char]]((aa: A) => init_lab[A, List[nat]](aa), a),
                             s)),
                                      prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('r', 'e', 'l', 'a',
                  't', 'i', 'v', 'e', ' ', 'n', 'o', 'n', '-', 't', 'e', 'r',
                  'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o',
                  'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p',
                  't', 'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(List('<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o',
                        'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
                        'n', 'P', 'r', 'o', 'o', 'f', '>'))
    }
}

def extract_input_proof(s: List[Char]):
      Option[(List[Char], (List[Char], List[Char]))] =
  search_len(s, List('<', 'i', 'n', 'p', 'u', 't', '>'),
              size_list[Char](List('<', 'i', 'n', 'p', 'u', 't', '>')))
    match {
    case None => None
    case Some((_, minput)) =>
      search_len(minput, List('<', '/', 'i', 'n', 'p', 'u', 't', '>'),
                  size_list[Char](List('<', '/', 'i', 'n', 'p', 'u', 't', '>')))
        match {
        case None => None
        case Some((the_input, sa)) =>
          search_len(sa, List('<', 'p', 'r', 'o', 'o', 'f', '>'),
                      size_list[Char](List('<', 'p', 'r', 'o', 'o', 'f', '>')))
            match {
            case None => None
            case Some((version, prf)) =>
              Some[(List[Char],
                     (List[Char], List[Char]))]((the_input, (version, prf)))
          }
      }
  }

def xmldoc2certProblem(x0: xmldoc):
      sum[List[Char], certificationProblem[List[Char], List[nat], List[Char]]] =
  x0 match {
  case Xmldoc(header, xml) =>
    debug[sum[List[Char],
               certificationProblem[List[Char], List[nat],
                                     List[Char]]]](List('0'),
            List('p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 'x', 'm', 'l', ' ',
                  't', 'o', ' ', 'c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a',
                  't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'),
            xml2certProblem[List[Char]]((a: xml) =>
  text(List('n', 'a', 'm', 'e'), a),
 (a: xml) => plain_name(a), xml))
}

def parseCertProblem(s: List[Char]):
      sum[List[Char], certificationProblem[List[Char], List[nat], List[Char]]] =
  binda[List[Char], xmldoc,
         certificationProblem[List[Char], List[nat],
                               List[Char]]](doc_of_string(s),
     (a: xmldoc) => xmldoc2certProblem(a))

def certifyProof(problemString: List[Char]): certificationResult =
  parseCertProblem(problemString) match {
    case Inl(err) =>
      Unsupported(List('e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'i', 'l', 'e',
                        ' ', 'p', 'a', 'r', 's', 'i', 'n', 'g') ++
                    (List(newline) ++ err))
    case Inr(cp) =>
      extract_input_proof(problemString) match {
        case None =>
          Error(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                      'n', 'd', ' ', 'i', 'n', 'p', 'u', 't', ' ', 'a', 'n',
                      'd', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'i', 'n', ' ',
                      's', 't', 'r', 'i', 'n', 'g', ' ', 'f', 'o', 'r', ' ',
                      'c', 'o', 'm', 'p', 'a', 'r', 'i', 's', 'o', 'n', ' ',
                      'w', 'i', 't', 'h', ' ', 'p', 'a', 'r', 's', 'e', 'd',
                      ' ', 'r', 'e', 's', 'u', 'l', 't'))
        case Some((the_input, (_, the_proof))) =>
          (if (eqWhiteSpace(the_input,
                             (shows_prec_xml(Zero_nat,
      xml_certificationProblem[List[Char], List[nat],
                                List[Char]](cp))).apply(Nil)))
            certifyCertProblem[List[Char]](cp) match {
              case Certified(tag) =>
                (if (starts_with(the_proof, tag)) Certified(tag)
                  else Error(List('p', 'r', 'o', 'v', 'e', 'n', ' ', 'p', 'r',
                                   'o', 'p', 'e', 'r', 't', 'y', ' ') ++
                               (tag ++
                                 (List(' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ',
'p', 'r', 'o', 'o', 'f', ' ', 'i', 'n', ' ', 'i', 'n', 'p', 'u', 't', ' ', 's',
't', 'r', 'i', 'n', 'g', ':', ' ') ++
                                   take[Char](nat(BigInt(20)), the_proof)))))
              case Unsupported(a) => Unsupported(a)
              case Error(a) => Error(a)
            }
            else Error(concat[Char](List(List('p', 'a', 'r', 's', 'e', 'd', ' ',
       'p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'd', 'o', 'e', 's', ' ', 'n',
       'o', 't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', ' ',
       't', 'o', ' ', 'i', 'n', 'p', 'u', 't'),
  List(newline), List('i', 'n', 'p', 'u', 't', ':'), List(newline),
  filter[Char]((c: Char) => ! (membera[Char](wspace, c)), the_input),
  List(newline), List(newline), List('p', 'a', 'r', 's', 'e', 'd', ':'),
  List(newline),
  filter[Char]((c: Char) => ! (membera[Char](wspace, c)),
                (shows_prec_xml(Zero_nat,
                                 xml_certificationProblem[List[Char], List[nat],
                   List[Char]](cp))).apply(Nil)),
  List(newline)))))
      }
  }

} /* object Ceta */
